/*
 * Copyright (C) 2022 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "GeneratedSerializers.h"

#include "APIError.h"
#include "APIFrameHandle.h"
#include "APIGeometry.h"
#include "APIPageHandle.h"
#include "APIURL.h"
#include "APIURLRequest.h"
#include "APIURLResponse.h"
#include "ArgumentCoders.h"
#if USE(SOUP)
#include "ArgumentCodersGLib.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "AudioTrackPrivateRemoteConfiguration.h"
#endif
#include "AuxiliaryProcessCreationParameters.h"
#include "BackgroundFetchState.h"
#if ENABLE(SERVICE_WORKER)
#include "BackgroundFetchState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "BufferIdentifierSet.h"
#endif
#include "CacheModel.h"
#include "CallbackID.h"
#include "EditorState.h"
#include "FileSystemSyncAccessHandleInfo.h"
#include "FocusedElementInformation.h"
#include "FrameInfoData.h"
#include "FrameTreeCreationParameters.h"
#include "FrameTreeNodeData.h"
#if ENABLE(GPU_PROCESS)
#include "GPUProcessConnectionParameters.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "GPUProcessSessionParameters.h"
#endif
#if ENABLE(GAMEPAD)
#include "GamepadData.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "InitializationSegmentInfo.h"
#endif
#include "LayerTreeContext.h"
#include "LocalFrameCreationParameters.h"
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "MediaDescriptionInfo.h"
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
#include "ModelIdentifier.h"
#endif
#include "NavigationActionData.h"
#include "NetworkLoadParameters.h"
#include "NetworkProcessConnectionParameters.h"
#include "NetworkProcessCreationParameters.h"
#include "NetworkResourceLoadParameters.h"
#include "PlatformPopupMenuData.h"
#include "PolicyDecision.h"
#if ENABLE(GPU_PROCESS) && PLATFORM(COCOA)
#include "PrepareBackingStoreBuffersData.h"
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
#include "RemoteAudioSessionConfiguration.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMConfiguration.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteGPURequestAdapterResponse.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteMediaPlayerConfiguration.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteMediaPlayerProxyConfiguration.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "RemoteMediaPlayerState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderingBackendCreationParameters.h"
#endif
#include "RemoteWorkerType.h"
#include "SameDocumentNavigationType.h"
#include "SessionState.h"
#include "ShareableBitmap.h"
#include "StreamConnectionEncoder.h"
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "TextTrackPrivateRemoteConfiguration.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "TrackPrivateRemoteConfiguration.h"
#endif
#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
#include "UpdateInfo.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "VideoTrackPrivateRemoteConfiguration.h"
#endif
#include "WebCoreArgumentCoders.h"
#include "WebEvent.h"
#if ENABLE(GPU_PROCESS)
#include "WebGPUBindGroupDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBindGroupEntry.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBindGroupLayoutDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBindGroupLayoutEntry.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBlendComponent.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBlendState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBufferBinding.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBufferBindingLayout.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUBufferDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUCanvasConfiguration.h"
#endif
#include "WebGPUColor.h"
#if ENABLE(GPU_PROCESS)
#include "WebGPUColorTargetState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUCommandBufferDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUCommandEncoderDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUCompilationMessage.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUComputePassDescriptor.h"
#endif
#include "WebGPUComputePassTimestampWrites.h"
#if ENABLE(GPU_PROCESS)
#include "WebGPUComputePipelineDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUDepthStencilState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUDeviceDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUExtent3D.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUExternalTextureBindingLayout.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUExternalTextureDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUFragmentState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUImageCopyBuffer.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUImageCopyExternalImage.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUImageCopyTexture.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUImageCopyTextureTagged.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUImageDataLayout.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUMultisampleState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUObjectDescriptorBase.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUOrigin2D.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUOrigin3D.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUOutOfMemoryError.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUPipelineDescriptorBase.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUPipelineLayoutDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUPresentationContextDescriptor.h"
#endif
#include "WebGPUPrimitiveState.h"
#if ENABLE(GPU_PROCESS)
#include "WebGPUPrimitiveState.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUProgrammableStage.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUQuerySetDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPURenderBundleDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPURenderBundleEncoderDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPURenderPassColorAttachment.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPURenderPassDepthStencilAttachment.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPURenderPassDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPURenderPassLayout.h"
#endif
#include "WebGPURenderPassTimestampWrites.h"
#if ENABLE(GPU_PROCESS)
#include "WebGPURenderPipelineDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPURequestAdapterOptions.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUSamplerBindingLayout.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUSamplerDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUShaderModuleCompilationHint.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUShaderModuleDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUStencilFaceState.h"
#endif
#include "WebGPUStorageTextureBindingLayout.h"
#if ENABLE(GPU_PROCESS)
#include "WebGPUStorageTextureBindingLayout.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUSupportedFeatures.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUSupportedLimits.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUTextureBindingLayout.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUTextureDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUTextureViewDescriptor.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUValidationError.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUVertexAttribute.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUVertexBufferLayout.h"
#endif
#if ENABLE(GPU_PROCESS)
#include "WebGPUVertexState.h"
#endif
#if ENABLE(MAC_GESTURE_EVENTS)
#include "WebGestureEvent.h"
#endif
#include "WebHitTestResultData.h"
#include "WebIDBResult.h"
#include "WebKeyboardEvent.h"
#include "WebMouseEvent.h"
#include "WebPageCreationParameters.h"
#include "WebPopupItem.h"
#include "WebProcessCreationParameters.h"
#include "WebProcessDataStoreParameters.h"
#include "WebPushDaemonConnectionConfiguration.h"
#include "WebPushMessage.h"
#if ENABLE(SPEECH_SYNTHESIS)
#include "WebSpeechSynthesisVoice.h"
#endif
#if ENABLE(TOUCH_EVENTS)
#include "WebTouchEvent.h"
#endif
#include "WebWheelEvent.h"
#include "WebsiteData.h"
#include "WebsiteDataFetchOption.h"
#include "WebsiteDataStoreParameters.h"
#include "WebsiteDataType.h"
#include "WebsitePoliciesData.h"
#if USE(CG)
#include <CoreGraphics/CGAffineTransform.h>
#endif
#if USE(CG)
#include <CoreGraphics/CGGeometry.h>
#endif
#include <JavaScriptCore/ArrayBuffer.h>
#if USE(SYSTEM_PREVIEW)
#include <WebCore/ARKitBadgeSystemImage.h>
#endif
#if ENABLE(THREADED_ANIMATION_RESOLUTION)
#include <WebCore/AcceleratedEffect.h>
#endif
#if ENABLE(THREADED_ANIMATION_RESOLUTION)
#include <WebCore/AcceleratedEffectValues.h>
#endif
#include <WebCore/AffineTransform.h>
#if ENABLE(APP_HIGHLIGHTS)
#include <WebCore/AppHighlight.h>
#endif
#if USE(APPKIT)
#include <WebCore/AppKitControlSystemImage.h>
#endif
#if ENABLE(APPLE_PAY_AMS_UI) && ENABLE(PAYMENT_REQUEST)
#include <WebCore/ApplePayAMSUIRequest.h>
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
#include <WebCore/ApplePayAutomaticReloadPaymentRequest.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayButtonPart.h>
#endif
#if ENABLE(APPLE_PAY_COUPON_CODE)
#include <WebCore/ApplePayCouponCodeUpdate.h>
#endif
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
#include <WebCore/ApplePayDateComponents.h>
#endif
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
#include <WebCore/ApplePayDateComponentsRange.h>
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
#include <WebCore/ApplePayDeferredPaymentRequest.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayDetailsUpdateBase.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayError.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayFeature.h>
#endif
#if ENABLE(APPLE_PAY_INSTALLMENTS)
#include <WebCore/ApplePayInstallmentConfigurationWebCore.h>
#endif
#if ENABLE(APPLE_PAY_INSTALLMENTS)
#include <WebCore/ApplePayInstallmentItem.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayLineItem.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayLogoSystemImage.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayPaymentAuthorizationResult.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayPaymentMethodUpdate.h>
#endif
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
#include <WebCore/ApplePayPaymentOrderDetails.h>
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
#include <WebCore/ApplePayPaymentTokenContext.h>
#endif
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
#include <WebCore/ApplePayRecurringPaymentRequest.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePaySetupConfiguration.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayShippingContactUpdate.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayShippingMethod.h>
#endif
#if ENABLE(APPLE_PAY)
#include <WebCore/ApplePayShippingMethodUpdate.h>
#endif
#if ENABLE(APPLICATION_MANIFEST)
#include <WebCore/ApplicationManifest.h>
#endif
#include <WebCore/AttributionSecondsUntilSendData.h>
#include <WebCore/AudioConfiguration.h>
#include <WebCore/AudioIOCallback.h>
#if USE(AUDIO_SESSION)
#include <WebCore/AudioSession.h>
#endif
#include <WebCore/AuthenticationChallenge.h>
#if ENABLE(WEB_AUTHN)
#include <WebCore/AuthenticationExtensionsClientInputs.h>
#endif
#if ENABLE(WEB_AUTHN)
#include <WebCore/AuthenticationExtensionsClientOutputs.h>
#endif
#include <WebCore/AutocapitalizeTypes.h>
#include <WebCore/Autofill.h>
#if ENABLE(SERVICE_WORKER)
#include <WebCore/BackgroundFetchInformation.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/BackgroundFetchOptions.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/BackgroundFetchRecordInformation.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/BackgroundFetchRequest.h>
#endif
#if ENABLE(GPU_PROCESS)
#include <WebCore/BarcodeDetectorOptionsInterface.h>
#endif
#if ENABLE(GPU_PROCESS)
#include <WebCore/BarcodeFormatInterface.h>
#endif
#include <WebCore/BasicShapes.h>
#include <WebCore/BufferSource.h>
#if ENABLE(ENCRYPTED_MEDIA)
#include <WebCore/CDMKeySystemConfiguration.h>
#endif
#if ENABLE(ENCRYPTED_MEDIA)
#include <WebCore/CDMMediaCapability.h>
#endif
#if ENABLE(ENCRYPTED_MEDIA)
#include <WebCore/CDMRestrictions.h>
#endif
#include <WebCore/COEPInheritenceViolationReportBody.h>
#include <WebCore/CORPViolationReportBody.h>
#include <WebCore/CSPViolationReportBody.h>
#include <WebCore/CSSValueKey.h>
#include <WebCore/CacheQueryOptions.h>
#include <WebCore/CanvasActivityRecord.h>
#if ENABLE(MEDIA_STREAM)
#include <WebCore/CaptureDevice.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/CaptureDeviceWithCapabilities.h>
#endif
#if PLATFORM(MAC)
#include <WebCore/CaretAnimator.h>
#endif
#include <WebCore/CertificateInfo.h>
#include <WebCore/CharacterRange.h>
#include <WebCore/ClientOrigin.h>
#include <WebCore/ColorInterpolationMethod.h>
#include <WebCore/CompositionHighlight.h>
#include <WebCore/CompositionUnderline.h>
#include <WebCore/ContactInfo.h>
#include <WebCore/ContactsRequestData.h>
#if ENABLE(CONTENT_EXTENSIONS)
#include <WebCore/ContentRuleListResults.h>
#endif
#include <WebCore/ContentSecurityPolicy.h>
#include <WebCore/ContentType.h>
#include <WebCore/ControlStyle.h>
#include <WebCore/Cookie.h>
#include <WebCore/CookieConsentDecisionResult.h>
#include <WebCore/CookieRequestHeaderFieldProxy.h>
#include <WebCore/CookieStoreGetOptions.h>
#include <WebCore/CrossOriginEmbedderPolicy.h>
#include <WebCore/CrossOriginOpenerPolicy.h>
#include <WebCore/CustomHeaderFields.h>
#include <WebCore/DOMCacheEngine.h>
#include <WebCore/DOMPasteAccess.h>
#if ENABLE(DATA_DETECTION)
#include <WebCore/DataDetectorType.h>
#endif
#if ENABLE(DATALIST_ELEMENT)
#include <WebCore/DataListSuggestionInformation.h>
#endif
#include <WebCore/DatabaseDetails.h>
#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
#include <WebCore/DateTimeChooserParameters.h>
#endif
#include <WebCore/DecomposedGlyphs.h>
#include <WebCore/DeprecationReportBody.h>
#include <WebCore/DestinationColorSpace.h>
#if ENABLE(WEB_RTC)
#include <WebCore/DetachedRTCDataChannel.h>
#endif
#if ENABLE(GPU_PROCESS)
#include <WebCore/DetectedBarcodeInterface.h>
#endif
#if ENABLE(GPU_PROCESS)
#include <WebCore/DetectedFaceInterface.h>
#endif
#if ENABLE(GPU_PROCESS)
#include <WebCore/DetectedTextInterface.h>
#endif
#include <WebCore/DeviceOrientationOrMotionPermissionState.h>
#include <WebCore/DiagnosticLoggingDomain.h>
#include <WebCore/DictationAlternative.h>
#include <WebCore/DictionaryPopupInfo.h>
#if HAVE(SCREEN_CAPTURE_KIT)
#include <WebCore/DisplayCapturePromptType.h>
#endif
#include <WebCore/DisplayListItems.h>
#include <WebCore/DistantLightSource.h>
#include <WebCore/DocumentLoader.h>
#if ENABLE(TRACKING_PREVENTION)
#include <WebCore/DocumentStorageAccess.h>
#endif
#include <WebCore/DragData.h>
#if ENABLE(DRAG_SUPPORT)
#include <WebCore/DragData.h>
#endif
#include <WebCore/ElementAnimationContext.h>
#include <WebCore/ElementContext.h>
#include <WebCore/EventInit.h>
#include <WebCore/EventRegion.h>
#include <WebCore/EventTrackingRegions.h>
#include <WebCore/Exception.h>
#include <WebCore/ExceptionData.h>
#include <WebCore/ExceptionDetails.h>
#include <WebCore/FEBlend.h>
#include <WebCore/FEColorMatrix.h>
#include <WebCore/FEComponentTransfer.h>
#include <WebCore/FEComposite.h>
#include <WebCore/FEConvolveMatrix.h>
#include <WebCore/FEDiffuseLighting.h>
#include <WebCore/FEDisplacementMap.h>
#include <WebCore/FEDropShadow.h>
#include <WebCore/FEFlood.h>
#include <WebCore/FEGaussianBlur.h>
#include <WebCore/FEImage.h>
#include <WebCore/FEMerge.h>
#include <WebCore/FEMorphology.h>
#include <WebCore/FEOffset.h>
#include <WebCore/FESpecularLighting.h>
#include <WebCore/FETurbulence.h>
#if ENABLE(GPU_PROCESS)
#include <WebCore/FaceDetectorOptionsInterface.h>
#endif
#include <WebCore/FetchOptions.h>
#include <WebCore/FileChooser.h>
#include <WebCore/FilterEffectGeometry.h>
#include <WebCore/FloatPoint.h>
#include <WebCore/FloatPoint3D.h>
#include <WebCore/FloatQuad.h>
#include <WebCore/FloatRect.h>
#include <WebCore/FloatRoundedRect.h>
#include <WebCore/FloatSize.h>
#include <WebCore/FocusDirection.h>
#include <WebCore/FontAttributeChanges.h>
#include <WebCore/FontAttributes.h>
#include <WebCore/FontShadow.h>
#include <WebCore/FormData.h>
#include <WebCore/FourCC.h>
#include <WebCore/FrameLoaderTypes.h>
#if ENABLE(GAMEPAD)
#include <WebCore/GamepadEffectParameters.h>
#endif
#if ENABLE(GAMEPAD)
#include <WebCore/GamepadHapticEffectType.h>
#endif
#include <WebCore/GeolocationPositionData.h>
#include <WebCore/GlobalFrameIdentifier.h>
#include <WebCore/GlobalWindowIdentifier.h>
#include <WebCore/Gradient.h>
#include <WebCore/GradientColorStop.h>
#include <WebCore/GradientColorStops.h>
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#include <WebCore/GraphicsContextGLActiveInfo.h>
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#include <WebCore/GraphicsContextGLAttributes.h>
#endif
#include <WebCore/GraphicsStyle.h>
#include <WebCore/GraphicsTypes.h>
#include <WebCore/HTMLModelElementCamera.h>
#include <WebCore/HTTPCookieAcceptPolicy.h>
#include <WebCore/HTTPHeaderField.h>
#include <WebCore/HTTPHeaderMap.h>
#include <WebCore/IDBCursorInfo.h>
#include <WebCore/IDBCursorRecord.h>
#include <WebCore/IDBDatabaseIdentifier.h>
#include <WebCore/IDBDatabaseInfo.h>
#include <WebCore/IDBDatabaseNameAndVersion.h>
#include <WebCore/IDBError.h>
#include <WebCore/IDBGetAllRecordsData.h>
#include <WebCore/IDBGetAllResult.h>
#include <WebCore/IDBGetRecordData.h>
#include <WebCore/IDBGetResult.h>
#include <WebCore/IDBIndexInfo.h>
#include <WebCore/IDBIterateCursorData.h>
#include <WebCore/IDBKeyData.h>
#include <WebCore/IDBKeyRangeData.h>
#include <WebCore/IDBObjectStoreInfo.h>
#include <WebCore/IDBRequestData.h>
#include <WebCore/IDBResourceIdentifier.h>
#include <WebCore/IDBResultData.h>
#include <WebCore/IDBTransactionInfo.h>
#include <WebCore/IDBValue.h>
#include <WebCore/ISOVTTCue.h>
#include <WebCore/IdentityTransformOperation.h>
#include <WebCore/ImageBufferBackendParameters.h>
#include <WebCore/ImageDecoder.h>
#include <WebCore/ImageOrientation.h>
#include <WebCore/ImagePaintingOptions.h>
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ImageResource.h>
#endif
#if ENABLE(VIDEO)
#include <WebCore/InbandGenericCue.h>
#endif
#include <WebCore/IntPoint.h>
#include <WebCore/IntRect.h>
#include <WebCore/IntSize.h>
#include <WebCore/InteractionRegion.h>
#include <WebCore/KeyboardScroll.h>
#if ENABLE(GPU_PROCESS)
#include <WebCore/LandmarkInterface.h>
#endif
#if ENABLE(GPU_PROCESS)
#include <WebCore/LandmarkTypeInterface.h>
#endif
#include <WebCore/LayoutPoint.h>
#include <WebCore/LayoutRect.h>
#include <WebCore/LayoutSize.h>
#include <WebCore/LayoutUnit.h>
#if ENABLE(THREADED_ANIMATION_RESOLUTION)
#include <WebCore/LengthPoint.h>
#endif
#include <WebCore/LengthSize.h>
#include <WebCore/LinkIcon.h>
#include <WebCore/ListStyleType.h>
#include <WebCore/LoadSchedulingMode.h>
#include <WebCore/Matrix3DTransformOperation.h>
#include <WebCore/MatrixTransformOperation.h>
#include <WebCore/MediaCapabilitiesDecodingInfo.h>
#include <WebCore/MediaCapabilitiesEncodingInfo.h>
#include <WebCore/MediaCapabilitiesInfo.h>
#include <WebCore/MediaConfiguration.h>
#if ENABLE(MEDIA_STREAM)
#include <WebCore/MediaConstraints.h>
#endif
#if ENABLE(MEDIA_CONTROLS_CONTEXT_MENUS) && USE(UICONTEXTMENU)
#include <WebCore/MediaControlsContextMenuItem.h>
#endif
#include <WebCore/MediaDecodingConfiguration.h>
#include <WebCore/MediaDeviceHashSalts.h>
#include <WebCore/MediaEncodingConfiguration.h>
#if ENABLE(VIDEO)
#include <WebCore/MediaPlayer.h>
#endif
#if ENABLE(MEDIA_SESSION)
#include <WebCore/MediaPositionState.h>
#endif
#include <WebCore/MediaProducer.h>
#include <WebCore/MediaSelectionOption.h>
#if ENABLE(MEDIA_SESSION)
#include <WebCore/MediaSessionAction.h>
#endif
#if ENABLE(MEDIA_SESSION_COORDINATOR)
#include <WebCore/MediaSessionCoordinatorState.h>
#endif
#if ENABLE(MEDIA_SESSION)
#include <WebCore/MediaSessionPlaybackState.h>
#endif
#if ENABLE(MEDIA_SESSION_COORDINATOR)
#include <WebCore/MediaSessionReadyState.h>
#endif
#include <WebCore/MediaStreamRequest.h>
#if ! ENABLE(MEDIA_STREAM)
#include <WebCore/MediaStreamRequest.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/MediaStreamRequest.h>
#endif
#include <WebCore/MessagePortIdentifier.h>
#include <WebCore/MessageWithMessagePorts.h>
#include <WebCore/MeterPart.h>
#if ENABLE(CONTENT_FILTERING)
#include <WebCore/MockContentFilterSettings.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/MockMediaDevice.h>
#endif
#if ENABLE(WEB_AUTHN)
#include <WebCore/MockWebAuthenticationConfiguration.h>
#endif
#include <WebCore/ModalContainerTypes.h>
#include <WebCore/Model.h>
#if ENABLE(SERVICE_WORKER)
#include <WebCore/NavigationPreloadState.h>
#endif
#include <WebCore/NavigationRequester.h>
#include <WebCore/NetworkLoadInformation.h>
#include <WebCore/NotificationData.h>
#include <WebCore/NotificationResources.h>
#if ENABLE(THREADED_ANIMATION_RESOLUTION)
#include <WebCore/OffsetRotation.h>
#endif
#include <WebCore/Pasteboard.h>
#include <WebCore/PasteboardCustomData.h>
#include <WebCore/PasteboardItemInfo.h>
#include <WebCore/PathOperation.h>
#include <WebCore/PathSegment.h>
#include <WebCore/PathSegmentData.h>
#include <WebCore/Pattern.h>
#if ENABLE(APPLE_PAY_INSTALLMENTS)
#include <WebCore/PaymentInstallmentConfigurationWebCore.h>
#endif
#include <WebCore/PermissionDescriptor.h>
#include <WebCore/PermissionQuerySource.h>
#include <WebCore/PermissionState.h>
#include <WebCore/PerspectiveTransformOperation.h>
#include <WebCore/PixelBufferFormat.h>
#if ENABLE(VIDEO)
#include <WebCore/PlatformAudioTrackConfiguration.h>
#endif
#if !USE(CG)
#include <WebCore/PlatformColorSpace.h>
#endif
#include <WebCore/PlatformTimeRanges.h>
#if ENABLE(VIDEO)
#include <WebCore/PlatformTrackConfiguration.h>
#endif
#include <WebCore/PlatformVideoColorSpace.h>
#if ENABLE(VIDEO)
#include <WebCore/PlatformVideoTrackConfiguration.h>
#endif
#include <WebCore/PlatformWheelEvent.h>
#include <WebCore/PluginData.h>
#include <WebCore/PointLightSource.h>
#include <WebCore/PolicyContainer.h>
#include <WebCore/PositionedGlyphs.h>
#include <WebCore/PrivateClickMeasurement.h>
#include <WebCore/ProcessIdentity.h>
#include <WebCore/ProgressBarPart.h>
#include <WebCore/PromisedAttachmentInfo.h>
#include <WebCore/ProtectionSpace.h>
#include <WebCore/PublicKeyCredentialCreationOptions.h>
#if ENABLE(WEB_AUTHN)
#include <WebCore/PublicKeyCredentialDescriptor.h>
#endif
#include <WebCore/PublicKeyCredentialRequestOptions.h>
#if ENABLE(SERVICE_WORKER)
#include <WebCore/PushPermissionState.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/PushSubscriptionData.h>
#endif
#if ENABLE(WEB_RTC)
#include <WebCore/RTCDataChannelHandler.h>
#endif
#if ENABLE(WEB_RTC)
#include <WebCore/RTCDataChannelIdentifier.h>
#endif
#include <WebCore/RTCErrorDetailType.h>
#if ENABLE(MEDIA_STREAM)
#include <WebCore/RealtimeMediaSource.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/RealtimeMediaSourceCapabilities.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/RealtimeMediaSourceSettings.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/RealtimeMediaSourceSupportedConstraints.h>
#endif
#include <WebCore/RectEdges.h>
#include <WebCore/Region.h>
#include <WebCore/RegistrableDomain.h>
#include <WebCore/RenderStyleConstants.h>
#include <WebCore/RenderTreeAsText.h>
#include <WebCore/Report.h>
#include <WebCore/ReportBody.h>
#include <WebCore/ResourceLoadStatistics.h>
#include <WebCore/ResourceLoaderOptions.h>
#include <WebCore/ResourceRequest.h>
#if !USE(CURL) && !PLATFORM(COCOA)
#include <WebCore/ResourceRequest.h>
#endif
#if PLATFORM(COCOA)
#include <WebCore/ResourceRequest.h>
#endif
#if USE(CURL)
#include <WebCore/ResourceRequest.h>
#endif
#if USE(SOUP)
#include <WebCore/ResourceRequest.h>
#endif
#include <WebCore/ResourceResponse.h>
#include <WebCore/ResourceResponseBase.h>
#include <WebCore/RetrieveRecordsOptions.h>
#include <WebCore/RotateTransformOperation.h>
#include <WebCore/SVGFilterExpression.h>
#include <WebCore/SVGPathByteStream.h>
#include <WebCore/SVGPreserveAspectRatioValue.h>
#include <WebCore/SameSiteInfo.h>
#include <WebCore/ScaleTransformOperation.h>
#include <WebCore/ScreenOrientationLockType.h>
#include <WebCore/ScreenOrientationType.h>
#include <WebCore/ScreenProperties.h>
#include <WebCore/ScrollTypes.h>
#include <WebCore/ScrollingConstraints.h>
#include <WebCore/ScrollingCoordinatorTypes.h>
#if ENABLE(ASYNC_SCROLLING)
#include <WebCore/ScrollingStateNode.h>
#endif
#include <WebCore/SearchPopupMenu.h>
#include <WebCore/SecurityOrigin.h>
#include <WebCore/SecurityOriginData.h>
#include <WebCore/SecurityPolicyViolationEvent.h>
#include <WebCore/SecurityPolicyViolationEventDisposition.h>
#if PLATFORM(IOS_FAMILY)
#include <WebCore/SelectionGeometry.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerClientData.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerClientPendingMessage.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerClientQueryOptions.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerContextData.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerData.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerImportedScript.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerJobData.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerJobDataIdentifier.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerRegistrationData.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerRegistrationKey.h>
#endif
#if ENABLE(SERVICE_WORKER)
#include <WebCore/ServiceWorkerRegistrationOptions.h>
#endif
#include <WebCore/ShareData.h>
#include <WebCore/SharedWorkerKey.h>
#include <WebCore/ShouldTreatAsContinuingLoad.h>
#include <WebCore/SkewTransformOperation.h>
#include <WebCore/SliderTrackPart.h>
#include <WebCore/SourceBrush.h>
#include <WebCore/SourceImage.h>
#include <WebCore/SpeechRecognitionError.h>
#include <WebCore/SpeechRecognitionRequestInfo.h>
#include <WebCore/SpeechRecognitionResultData.h>
#include <WebCore/SpeechRecognitionUpdate.h>
#include <WebCore/SpotLightSource.h>
#include <WebCore/StorageEstimate.h>
#include <WebCore/StorageType.h>
#include <WebCore/SubstituteData.h>
#include <WebCore/TestReportBody.h>
#include <WebCore/TextChecking.h>
#include <WebCore/TextGranularity.h>
#include <WebCore/TextIndicator.h>
#include <WebCore/TextManipulationController.h>
#include <WebCore/TextManipulationItem.h>
#include <WebCore/TextManipulationToken.h>
#if ENABLE(IMAGE_ANALYSIS)
#include <WebCore/TextRecognitionResult.h>
#endif
#include <WebCore/TimingFunction.h>
#include <WebCore/TransformOperation.h>
#include <WebCore/TransformOperations.h>
#include <WebCore/TransformationMatrix.h>
#include <WebCore/TranslateTransformOperation.h>
#if HAVE(TRANSLATION_UI_SERVICES) && ENABLE(CONTEXT_MENUS)
#include <WebCore/TranslationContextMenuInfo.h>
#endif
#include <WebCore/UserStyleSheet.h>
#include <WebCore/VP9Utilities.h>
#include <WebCore/VelocityData.h>
#if ENABLE(VIDEO)
#include <WebCore/VideoFrame.h>
#endif
#if ENABLE(VIDEO) && USE(AVFOUNDATION)
#include <WebCore/VideoFrameCV.h>
#endif
#if ENABLE(VIDEO)
#include <WebCore/VideoFrameMetadata.h>
#endif
#include <WebCore/VideoFrameTimeMetadata.h>
#if ENABLE(VIDEO)
#include <WebCore/VideoPlaybackQualityMetrics.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/VideoPreset.h>
#endif
#include <WebCore/ViewportArguments.h>
#include <WebCore/VisibleSelection.h>
#if ENABLE(THREADED_ANIMATION_RESOLUTION)
#include <WebCore/WebAnimationTypes.h>
#endif
#include <WebCore/WebCodecsEncodedVideoChunk.h>
#if ENABLE(WEB_CODECS)
#include <WebCore/WebCodecsEncodedVideoChunkData.h>
#endif
#if ENABLE(GPU_PROCESS)
#include <WebCore/WebGPUExternalTextureDescriptor.h>
#endif
#include <WebCore/WebGPUPredefinedColorSpace.h>
#include <WebCore/WebGPUTextureAspect.h>
#include <WebCore/WebGPUTextureFormat.h>
#include <WebCore/WebLockManagerSnapshot.h>
#include <WebCore/WindowFeatures.h>
#include <WebCore/WorkerFetchResult.h>
#include <WebCore/WorkerInitializationData.h>
#include <WebCore/WorkerOptions.h>
#include <WebCore/WritingDirection.h>
#if ENABLE(WK_WEB_EXTENSIONS)
#include <WebKit/WebExtensionEventListenerType.h>
#endif
#include <wtf/MediaTime.h>
#include <wtf/MonotonicTime.h>
#if ENABLE(SERVICE_WORKER)
#include <wtf/RobinHoodHashTable.h>
#endif
#include <wtf/Seconds.h>
#include <wtf/URL.h>
#include <wtf/WallTime.h>
#include <wtf/text/AtomString.h>

template<size_t...> struct MembersInCorrectOrder;
template<size_t onlyOffset> struct MembersInCorrectOrder<onlyOffset> {
    static constexpr bool value = true;
};
template<size_t firstOffset, size_t secondOffset, size_t... remainingOffsets> struct MembersInCorrectOrder<firstOffset, secondOffset, remainingOffsets...> {
    static constexpr bool value = firstOffset > secondOffset ? false : MembersInCorrectOrder<secondOffset, remainingOffsets...>::value;
};

template<uint64_t...> struct BitsInIncreasingOrder;
template<uint64_t onlyBit> struct BitsInIncreasingOrder<onlyBit> {
    static constexpr bool value = true;
};
template<uint64_t firstBit, uint64_t secondBit, uint64_t... remainingBits> struct BitsInIncreasingOrder<firstBit, secondBit, remainingBits...> {
    static constexpr bool value = firstBit == secondBit >> 1 && BitsInIncreasingOrder<secondBit, remainingBits...>::value;
};

template<bool, bool> struct VirtualTableAndRefCountOverhead;
template<> struct VirtualTableAndRefCountOverhead<true, true> {
    virtual ~VirtualTableAndRefCountOverhead() { }
    unsigned refCount;
#if ASSERT_ENABLED
    bool m_isOwnedByMainThread;
    bool m_areThreadingChecksEnabled;
#endif
#if CHECK_REF_COUNTED_LIFECYCLE
    bool m_deletionHasBegun;
    bool m_adoptionIsRequired;
#endif
};
template<> struct VirtualTableAndRefCountOverhead<false, true> {
    unsigned refCount;
#if ASSERT_ENABLED
    bool m_isOwnedByMainThread;
    bool m_areThreadingChecksEnabled;
#endif
#if CHECK_REF_COUNTED_LIFECYCLE
    bool m_deletionHasBegun;
    bool m_adoptionIsRequired;
#endif
};
template<> struct VirtualTableAndRefCountOverhead<true, false> {
    virtual ~VirtualTableAndRefCountOverhead() { }
};
template<> struct VirtualTableAndRefCountOverhead<false, false> { };

#if COMPILER(GCC)
IGNORE_WARNINGS_BEGIN("invalid-offsetof")
#endif

namespace IPC {


#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
template<> struct ArgumentCoder<WebKit::InitializationSegmentInfo::TrackInformation> {
    static void encode(Encoder&, const WebKit::InitializationSegmentInfo::TrackInformation&);
    static std::optional<WebKit::InitializationSegmentInfo::TrackInformation> decode(Decoder&);
};
#endif

template<> struct ArgumentCoder<WebKit::EditorState::PostLayoutData> {
    static void encode(Encoder&, const WebKit::EditorState::PostLayoutData&);
    static std::optional<WebKit::EditorState::PostLayoutData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebKit::EditorState::VisualData> {
    static void encode(Encoder&, const WebKit::EditorState::VisualData&);
    static std::optional<WebKit::EditorState::VisualData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebKit::HTTPBody::Element> {
    static void encode(Encoder&, const WebKit::HTTPBody::Element&);
    static std::optional<WebKit::HTTPBody::Element> decode(Decoder&);
};

template<> struct ArgumentCoder<WebKit::HTTPBody::Element::FileData> {
    static void encode(Encoder&, const WebKit::HTTPBody::Element::FileData&);
    static std::optional<WebKit::HTTPBody::Element::FileData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::IDBKeyData::Invalid> {
    static void encode(Encoder&, const WebCore::IDBKeyData::Invalid&);
    static std::optional<WebCore::IDBKeyData::Invalid> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::IDBKeyData::Min> {
    static void encode(Encoder&, const WebCore::IDBKeyData::Min&);
    static std::optional<WebCore::IDBKeyData::Min> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::IDBKeyData::Max> {
    static void encode(Encoder&, const WebCore::IDBKeyData::Max&);
    static std::optional<WebCore::IDBKeyData::Max> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::IDBKeyData::Date> {
    static void encode(Encoder&, const WebCore::IDBKeyData::Date&);
    static std::optional<WebCore::IDBKeyData::Date> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::LinearTimingFunction::Point> {
    static void encode(Encoder&, const WebCore::LinearTimingFunction::Point&);
    static std::optional<WebCore::LinearTimingFunction::Point> decode(Decoder&);
};

#if ENABLE(APPLICATION_MANIFEST)
template<> struct ArgumentCoder<WebCore::ApplicationManifest::Icon> {
    static void encode(Encoder&, const WebCore::ApplicationManifest::Icon&);
    static std::optional<WebCore::ApplicationManifest::Icon> decode(Decoder&);
};
#endif

template<> struct ArgumentCoder<WebCore::WebLockManagerSnapshot::Info> {
    static void encode(Encoder&, const WebCore::WebLockManagerSnapshot::Info&);
    static std::optional<WebCore::WebLockManagerSnapshot::Info> decode(Decoder&);
};

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs> {
    static void encode(Encoder&, const WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs&);
    static std::optional<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput> {
    static void encode(Encoder&, const WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput&);
    static std::optional<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs> {
    static void encode(Encoder&, const WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs&);
    static std::optional<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::Parameters> {
    static void encode(Encoder&, const WebCore::PublicKeyCredentialCreationOptions::Parameters&);
    static std::optional<WebCore::PublicKeyCredentialCreationOptions::Parameters> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria> {
    static void encode(Encoder&, const WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria&);
    static std::optional<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::Entity> {
    static void encode(Encoder&, const WebCore::PublicKeyCredentialCreationOptions::Entity&);
    static std::optional<WebCore::PublicKeyCredentialCreationOptions::Entity> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::RpEntity> {
    static void encode(Encoder&, const WebCore::PublicKeyCredentialCreationOptions::RpEntity&);
    static std::optional<WebCore::PublicKeyCredentialCreationOptions::RpEntity> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::UserEntity> {
    static void encode(Encoder&, const WebCore::PublicKeyCredentialCreationOptions::UserEntity&);
    static std::optional<WebCore::PublicKeyCredentialCreationOptions::UserEntity> decode(Decoder&);
};
#endif

template<> struct ArgumentCoder<WebCore::HTTPHeaderMap::CommonHeader> {
    static void encode(Encoder&, const WebCore::HTTPHeaderMap::CommonHeader&);
    static std::optional<WebCore::HTTPHeaderMap::CommonHeader> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::HTTPHeaderMap::UncommonHeader> {
    static void encode(Encoder&, const WebCore::HTTPHeaderMap::UncommonHeader&);
    static std::optional<WebCore::HTTPHeaderMap::UncommonHeader> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ResourceRequest::RequestData> {
    static void encode(Encoder&, const WebCore::ResourceRequest::RequestData&);
    static std::optional<WebCore::ResourceRequest::RequestData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::FormDataElement::EncodedFileData> {
    static void encode(Encoder&, const WebCore::FormDataElement::EncodedFileData&);
    static std::optional<WebCore::FormDataElement::EncodedFileData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::FormDataElement::EncodedBlobData> {
    static void encode(Encoder&, const WebCore::FormDataElement::EncodedBlobData&);
    static std::optional<WebCore::FormDataElement::EncodedBlobData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::PlatformTimeRanges::Range> {
    static void encode(Encoder&, const WebCore::PlatformTimeRanges::Range&);
    static std::optional<WebCore::PlatformTimeRanges::Range> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::PasteboardCustomData::Entry> {
    static void encode(Encoder&, const WebCore::PasteboardCustomData::Entry&);
    static std::optional<WebCore::PasteboardCustomData::Entry> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::Gradient::LinearData> {
    static void encode(Encoder&, const WebCore::Gradient::LinearData&);
    static void encode(StreamConnectionEncoder&, const WebCore::Gradient::LinearData&);
    static std::optional<WebCore::Gradient::LinearData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::Gradient::RadialData> {
    static void encode(Encoder&, const WebCore::Gradient::RadialData&);
    static void encode(StreamConnectionEncoder&, const WebCore::Gradient::RadialData&);
    static std::optional<WebCore::Gradient::RadialData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::Gradient::ConicData> {
    static void encode(Encoder&, const WebCore::Gradient::ConicData&);
    static void encode(StreamConnectionEncoder&, const WebCore::Gradient::ConicData&);
    static std::optional<WebCore::Gradient::ConicData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::Pattern::Parameters> {
    static void encode(Encoder&, const WebCore::Pattern::Parameters&);
    static void encode(StreamConnectionEncoder&, const WebCore::Pattern::Parameters&);
    static std::optional<WebCore::Pattern::Parameters> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::SourceBrush::Brush> {
    static void encode(Encoder&, const WebCore::SourceBrush::Brush&);
    static void encode(StreamConnectionEncoder&, const WebCore::SourceBrush::Brush&);
    static std::optional<WebCore::SourceBrush::Brush> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ResourceResponseBase::ResponseData> {
    static void encode(Encoder&, const WebCore::ResourceResponseBase::ResponseData&);
    static std::optional<WebCore::ResourceResponseBase::ResponseData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ResourceResponseBase::CrossThreadData> {
    static void encode(Encoder&, const WebCore::ResourceResponseBase::CrossThreadData&);
    static std::optional<WebCore::ResourceResponseBase::CrossThreadData> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::ElementRule> {
    static void encode(Encoder&, const WebCore::TextManipulationControllerExclusionRule::ElementRule&);
    static std::optional<WebCore::TextManipulationControllerExclusionRule::ElementRule> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::AttributeRule> {
    static void encode(Encoder&, const WebCore::TextManipulationControllerExclusionRule::AttributeRule&);
    static std::optional<WebCore::TextManipulationControllerExclusionRule::AttributeRule> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::ClassRule> {
    static void encode(Encoder&, const WebCore::TextManipulationControllerExclusionRule::ClassRule&);
    static std::optional<WebCore::TextManipulationControllerExclusionRule::ClassRule> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::HSL> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::HSL&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::HSL&);
    static std::optional<WebCore::ColorInterpolationMethod::HSL> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::HWB> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::HWB&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::HWB&);
    static std::optional<WebCore::ColorInterpolationMethod::HWB> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::LCH> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::LCH&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::LCH&);
    static std::optional<WebCore::ColorInterpolationMethod::LCH> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::Lab> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::Lab&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::Lab&);
    static std::optional<WebCore::ColorInterpolationMethod::Lab> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::OKLCH> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::OKLCH&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::OKLCH&);
    static std::optional<WebCore::ColorInterpolationMethod::OKLCH> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::OKLab> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::OKLab&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::OKLab&);
    static std::optional<WebCore::ColorInterpolationMethod::OKLab> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::SRGB> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::SRGB&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::SRGB&);
    static std::optional<WebCore::ColorInterpolationMethod::SRGB> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::SRGBLinear> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::SRGBLinear&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::SRGBLinear&);
    static std::optional<WebCore::ColorInterpolationMethod::SRGBLinear> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD50> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::XYZD50&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::XYZD50&);
    static std::optional<WebCore::ColorInterpolationMethod::XYZD50> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD65> {
    static void encode(Encoder&, const WebCore::ColorInterpolationMethod::XYZD65&);
    static void encode(StreamConnectionEncoder&, const WebCore::ColorInterpolationMethod::XYZD65&);
    static std::optional<WebCore::ColorInterpolationMethod::XYZD65> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::SecurityOriginData::Tuple> {
    static void encode(Encoder&, const WebCore::SecurityOriginData::Tuple&);
    static std::optional<WebCore::SecurityOriginData::Tuple> decode(Decoder&);
};

#if ENABLE(CONTENT_EXTENSIONS)
template<> struct ArgumentCoder<WebCore::ContentRuleListResults::Result> {
    static void encode(Encoder&, const WebCore::ContentRuleListResults::Result&);
    static std::optional<WebCore::ContentRuleListResults::Result> decode(Decoder&);
};
#endif

#if ENABLE(CONTENT_EXTENSIONS)
template<> struct ArgumentCoder<WebCore::ContentRuleListResults::Summary> {
    static void encode(Encoder&, const WebCore::ContentRuleListResults::Summary&);
    static std::optional<WebCore::ContentRuleListResults::Summary> decode(Decoder&);
};
#endif

template<> struct ArgumentCoder<WebCore::Region::Span> {
    static void encode(Encoder&, const WebCore::Region::Span&);
    static std::optional<WebCore::Region::Span> decode(Decoder&);
};

template<> struct ArgumentCoder<WebCore::Region::Shape> {
    static void encode(Encoder&, const WebCore::Region::Shape&);
    static std::optional<WebCore::Region::Shape> decode(Decoder&);
};

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration> {
    static void encode(Encoder&, const WebCore::MockWebAuthenticationConfiguration::LocalConfiguration&);
    static std::optional<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::HidConfiguration> {
    static void encode(Encoder&, const WebCore::MockWebAuthenticationConfiguration::HidConfiguration&);
    static std::optional<WebCore::MockWebAuthenticationConfiguration::HidConfiguration> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration> {
    static void encode(Encoder&, const WebCore::MockWebAuthenticationConfiguration::NfcConfiguration&);
    static std::optional<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration> decode(Decoder&);
};
#endif

#if ENABLE(WEB_AUTHN)
template<> struct ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration> {
    static void encode(Encoder&, const WebCore::MockWebAuthenticationConfiguration::CcidConfiguration&);
    static std::optional<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration> decode(Decoder&);
};
#endif

template<> struct ArgumentCoder<WebKit::WebsiteData::Entry> {
    static void encode(Encoder&, const WebKit::WebsiteData::Entry&);
    static std::optional<WebKit::WebsiteData::Entry> decode(Decoder&);
};

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::GPUProcessSessionParameters>::encode(Encoder& encoder, const WebKit::GPUProcessSessionParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaCacheDirectory)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaCacheDirectorySandboxExtensionHandle)>, WebKit::SandboxExtension::Handle>);
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaKeysStorageDirectory)>, String>);
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaKeysStorageDirectorySandboxExtensionHandle)>, WebKit::SandboxExtension::Handle>);
#endif
    struct ShouldBeSameSizeAsGPUProcessSessionParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::GPUProcessSessionParameters>, false> {
        String mediaCacheDirectory;
        WebKit::SandboxExtension::Handle mediaCacheDirectorySandboxExtensionHandle;
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
        String mediaKeysStorageDirectory;
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
        WebKit::SandboxExtension::Handle mediaKeysStorageDirectorySandboxExtensionHandle;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsGPUProcessSessionParameters) == sizeof(WebKit::GPUProcessSessionParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::GPUProcessSessionParameters, mediaCacheDirectory)
        , offsetof(WebKit::GPUProcessSessionParameters, mediaCacheDirectorySandboxExtensionHandle)
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
        , offsetof(WebKit::GPUProcessSessionParameters, mediaKeysStorageDirectory)
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
        , offsetof(WebKit::GPUProcessSessionParameters, mediaKeysStorageDirectorySandboxExtensionHandle)
#endif
    >::value);
    encoder << instance.mediaCacheDirectory;
    encoder << instance.mediaCacheDirectorySandboxExtensionHandle;
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    encoder << instance.mediaKeysStorageDirectory;
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    encoder << instance.mediaKeysStorageDirectorySandboxExtensionHandle;
#endif
}

std::optional<WebKit::GPUProcessSessionParameters> ArgumentCoder<WebKit::GPUProcessSessionParameters>::decode(Decoder& decoder)
{
    auto mediaCacheDirectory = decoder.decode<String>();
    auto mediaCacheDirectorySandboxExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    auto mediaKeysStorageDirectory = decoder.decode<String>();
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    auto mediaKeysStorageDirectorySandboxExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::GPUProcessSessionParameters {
            WTFMove(*mediaCacheDirectory),
            WTFMove(*mediaCacheDirectorySandboxExtensionHandle),
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
            WTFMove(*mediaKeysStorageDirectory),
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
            WTFMove(*mediaKeysStorageDirectorySandboxExtensionHandle)
#endif
        }
    };
}

#endif

void ArgumentCoder<WebCore::PathMoveTo>::encode(Encoder& encoder, const WebCore::PathMoveTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathMoveTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathMoveTo>, false> {
        WebCore::FloatPoint point;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathMoveTo) == sizeof(WebCore::PathMoveTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathMoveTo, point)
    >::value);
    encoder << instance.point;
}

void ArgumentCoder<WebCore::PathMoveTo>::encode(StreamConnectionEncoder& encoder, const WebCore::PathMoveTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathMoveTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathMoveTo>, false> {
        WebCore::FloatPoint point;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathMoveTo) == sizeof(WebCore::PathMoveTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathMoveTo, point)
    >::value);
    encoder << instance.point;
}

std::optional<WebCore::PathMoveTo> ArgumentCoder<WebCore::PathMoveTo>::decode(Decoder& decoder)
{
    auto point = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathMoveTo {
            WTFMove(*point)
        }
    };
}

void ArgumentCoder<WebCore::PathLineTo>::encode(Encoder& encoder, const WebCore::PathLineTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathLineTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathLineTo>, false> {
        WebCore::FloatPoint point;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathLineTo) == sizeof(WebCore::PathLineTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathLineTo, point)
    >::value);
    encoder << instance.point;
}

void ArgumentCoder<WebCore::PathLineTo>::encode(StreamConnectionEncoder& encoder, const WebCore::PathLineTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathLineTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathLineTo>, false> {
        WebCore::FloatPoint point;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathLineTo) == sizeof(WebCore::PathLineTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathLineTo, point)
    >::value);
    encoder << instance.point;
}

std::optional<WebCore::PathLineTo> ArgumentCoder<WebCore::PathLineTo>::decode(Decoder& decoder)
{
    auto point = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathLineTo {
            WTFMove(*point)
        }
    };
}

void ArgumentCoder<WebCore::PathQuadCurveTo>::encode(Encoder& encoder, const WebCore::PathQuadCurveTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathQuadCurveTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathQuadCurveTo>, false> {
        WebCore::FloatPoint controlPoint;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathQuadCurveTo) == sizeof(WebCore::PathQuadCurveTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathQuadCurveTo, controlPoint)
        , offsetof(WebCore::PathQuadCurveTo, endPoint)
    >::value);
    encoder << instance.controlPoint;
    encoder << instance.endPoint;
}

void ArgumentCoder<WebCore::PathQuadCurveTo>::encode(StreamConnectionEncoder& encoder, const WebCore::PathQuadCurveTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathQuadCurveTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathQuadCurveTo>, false> {
        WebCore::FloatPoint controlPoint;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathQuadCurveTo) == sizeof(WebCore::PathQuadCurveTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathQuadCurveTo, controlPoint)
        , offsetof(WebCore::PathQuadCurveTo, endPoint)
    >::value);
    encoder << instance.controlPoint;
    encoder << instance.endPoint;
}

std::optional<WebCore::PathQuadCurveTo> ArgumentCoder<WebCore::PathQuadCurveTo>::decode(Decoder& decoder)
{
    auto controlPoint = decoder.decode<WebCore::FloatPoint>();
    auto endPoint = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathQuadCurveTo {
            WTFMove(*controlPoint),
            WTFMove(*endPoint)
        }
    };
}

void ArgumentCoder<WebCore::PathBezierCurveTo>::encode(Encoder& encoder, const WebCore::PathBezierCurveTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathBezierCurveTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathBezierCurveTo>, false> {
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathBezierCurveTo) == sizeof(WebCore::PathBezierCurveTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathBezierCurveTo, controlPoint1)
        , offsetof(WebCore::PathBezierCurveTo, controlPoint2)
        , offsetof(WebCore::PathBezierCurveTo, endPoint)
    >::value);
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.endPoint;
}

void ArgumentCoder<WebCore::PathBezierCurveTo>::encode(StreamConnectionEncoder& encoder, const WebCore::PathBezierCurveTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathBezierCurveTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathBezierCurveTo>, false> {
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathBezierCurveTo) == sizeof(WebCore::PathBezierCurveTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathBezierCurveTo, controlPoint1)
        , offsetof(WebCore::PathBezierCurveTo, controlPoint2)
        , offsetof(WebCore::PathBezierCurveTo, endPoint)
    >::value);
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.endPoint;
}

std::optional<WebCore::PathBezierCurveTo> ArgumentCoder<WebCore::PathBezierCurveTo>::decode(Decoder& decoder)
{
    auto controlPoint1 = decoder.decode<WebCore::FloatPoint>();
    auto controlPoint2 = decoder.decode<WebCore::FloatPoint>();
    auto endPoint = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathBezierCurveTo {
            WTFMove(*controlPoint1),
            WTFMove(*controlPoint2),
            WTFMove(*endPoint)
        }
    };
}

void ArgumentCoder<WebCore::PathArcTo>::encode(Encoder& encoder, const WebCore::PathArcTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    struct ShouldBeSameSizeAsPathArcTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathArcTo>, false> {
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        float radius;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathArcTo) == sizeof(WebCore::PathArcTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathArcTo, controlPoint1)
        , offsetof(WebCore::PathArcTo, controlPoint2)
        , offsetof(WebCore::PathArcTo, radius)
    >::value);
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.radius;
}

void ArgumentCoder<WebCore::PathArcTo>::encode(StreamConnectionEncoder& encoder, const WebCore::PathArcTo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    struct ShouldBeSameSizeAsPathArcTo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathArcTo>, false> {
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        float radius;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathArcTo) == sizeof(WebCore::PathArcTo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathArcTo, controlPoint1)
        , offsetof(WebCore::PathArcTo, controlPoint2)
        , offsetof(WebCore::PathArcTo, radius)
    >::value);
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.radius;
}

std::optional<WebCore::PathArcTo> ArgumentCoder<WebCore::PathArcTo>::decode(Decoder& decoder)
{
    auto controlPoint1 = decoder.decode<WebCore::FloatPoint>();
    auto controlPoint2 = decoder.decode<WebCore::FloatPoint>();
    auto radius = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathArcTo {
            WTFMove(*controlPoint1),
            WTFMove(*controlPoint2),
            WTFMove(*radius)
        }
    };
}

void ArgumentCoder<WebCore::PathArc>::encode(Encoder& encoder, const WebCore::PathArc& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.center)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction)>, WebCore::RotationDirection>);
    struct ShouldBeSameSizeAsPathArc : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathArc>, false> {
        WebCore::FloatPoint center;
        float radius;
        float startAngle;
        float endAngle;
        WebCore::RotationDirection direction;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathArc) == sizeof(WebCore::PathArc));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathArc, center)
        , offsetof(WebCore::PathArc, radius)
        , offsetof(WebCore::PathArc, startAngle)
        , offsetof(WebCore::PathArc, endAngle)
        , offsetof(WebCore::PathArc, direction)
    >::value);
    encoder << instance.center;
    encoder << instance.radius;
    encoder << instance.startAngle;
    encoder << instance.endAngle;
    encoder << instance.direction;
}

void ArgumentCoder<WebCore::PathArc>::encode(StreamConnectionEncoder& encoder, const WebCore::PathArc& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.center)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction)>, WebCore::RotationDirection>);
    struct ShouldBeSameSizeAsPathArc : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathArc>, false> {
        WebCore::FloatPoint center;
        float radius;
        float startAngle;
        float endAngle;
        WebCore::RotationDirection direction;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathArc) == sizeof(WebCore::PathArc));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathArc, center)
        , offsetof(WebCore::PathArc, radius)
        , offsetof(WebCore::PathArc, startAngle)
        , offsetof(WebCore::PathArc, endAngle)
        , offsetof(WebCore::PathArc, direction)
    >::value);
    encoder << instance.center;
    encoder << instance.radius;
    encoder << instance.startAngle;
    encoder << instance.endAngle;
    encoder << instance.direction;
}

std::optional<WebCore::PathArc> ArgumentCoder<WebCore::PathArc>::decode(Decoder& decoder)
{
    auto center = decoder.decode<WebCore::FloatPoint>();
    auto radius = decoder.decode<float>();
    auto startAngle = decoder.decode<float>();
    auto endAngle = decoder.decode<float>();
    auto direction = decoder.decode<WebCore::RotationDirection>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathArc {
            WTFMove(*center),
            WTFMove(*radius),
            WTFMove(*startAngle),
            WTFMove(*endAngle),
            WTFMove(*direction)
        }
    };
}

void ArgumentCoder<WebCore::PathEllipse>::encode(Encoder& encoder, const WebCore::PathEllipse& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.center)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusX)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusY)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rotation)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction)>, WebCore::RotationDirection>);
    struct ShouldBeSameSizeAsPathEllipse : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathEllipse>, false> {
        WebCore::FloatPoint center;
        float radiusX;
        float radiusY;
        float rotation;
        float startAngle;
        float endAngle;
        WebCore::RotationDirection direction;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathEllipse) == sizeof(WebCore::PathEllipse));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathEllipse, center)
        , offsetof(WebCore::PathEllipse, radiusX)
        , offsetof(WebCore::PathEllipse, radiusY)
        , offsetof(WebCore::PathEllipse, rotation)
        , offsetof(WebCore::PathEllipse, startAngle)
        , offsetof(WebCore::PathEllipse, endAngle)
        , offsetof(WebCore::PathEllipse, direction)
    >::value);
    encoder << instance.center;
    encoder << instance.radiusX;
    encoder << instance.radiusY;
    encoder << instance.rotation;
    encoder << instance.startAngle;
    encoder << instance.endAngle;
    encoder << instance.direction;
}

void ArgumentCoder<WebCore::PathEllipse>::encode(StreamConnectionEncoder& encoder, const WebCore::PathEllipse& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.center)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusX)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusY)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rotation)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endAngle)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction)>, WebCore::RotationDirection>);
    struct ShouldBeSameSizeAsPathEllipse : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathEllipse>, false> {
        WebCore::FloatPoint center;
        float radiusX;
        float radiusY;
        float rotation;
        float startAngle;
        float endAngle;
        WebCore::RotationDirection direction;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathEllipse) == sizeof(WebCore::PathEllipse));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathEllipse, center)
        , offsetof(WebCore::PathEllipse, radiusX)
        , offsetof(WebCore::PathEllipse, radiusY)
        , offsetof(WebCore::PathEllipse, rotation)
        , offsetof(WebCore::PathEllipse, startAngle)
        , offsetof(WebCore::PathEllipse, endAngle)
        , offsetof(WebCore::PathEllipse, direction)
    >::value);
    encoder << instance.center;
    encoder << instance.radiusX;
    encoder << instance.radiusY;
    encoder << instance.rotation;
    encoder << instance.startAngle;
    encoder << instance.endAngle;
    encoder << instance.direction;
}

std::optional<WebCore::PathEllipse> ArgumentCoder<WebCore::PathEllipse>::decode(Decoder& decoder)
{
    auto center = decoder.decode<WebCore::FloatPoint>();
    auto radiusX = decoder.decode<float>();
    auto radiusY = decoder.decode<float>();
    auto rotation = decoder.decode<float>();
    auto startAngle = decoder.decode<float>();
    auto endAngle = decoder.decode<float>();
    auto direction = decoder.decode<WebCore::RotationDirection>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathEllipse {
            WTFMove(*center),
            WTFMove(*radiusX),
            WTFMove(*radiusY),
            WTFMove(*rotation),
            WTFMove(*startAngle),
            WTFMove(*endAngle),
            WTFMove(*direction)
        }
    };
}

void ArgumentCoder<WebCore::PathEllipseInRect>::encode(Encoder& encoder, const WebCore::PathEllipseInRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect)>, WebCore::FloatRect>);
    struct ShouldBeSameSizeAsPathEllipseInRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathEllipseInRect>, false> {
        WebCore::FloatRect rect;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathEllipseInRect) == sizeof(WebCore::PathEllipseInRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathEllipseInRect, rect)
    >::value);
    encoder << instance.rect;
}

void ArgumentCoder<WebCore::PathEllipseInRect>::encode(StreamConnectionEncoder& encoder, const WebCore::PathEllipseInRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect)>, WebCore::FloatRect>);
    struct ShouldBeSameSizeAsPathEllipseInRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathEllipseInRect>, false> {
        WebCore::FloatRect rect;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathEllipseInRect) == sizeof(WebCore::PathEllipseInRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathEllipseInRect, rect)
    >::value);
    encoder << instance.rect;
}

std::optional<WebCore::PathEllipseInRect> ArgumentCoder<WebCore::PathEllipseInRect>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathEllipseInRect {
            WTFMove(*rect)
        }
    };
}

void ArgumentCoder<WebCore::PathRect>::encode(Encoder& encoder, const WebCore::PathRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect)>, WebCore::FloatRect>);
    struct ShouldBeSameSizeAsPathRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathRect>, false> {
        WebCore::FloatRect rect;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathRect) == sizeof(WebCore::PathRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathRect, rect)
    >::value);
    encoder << instance.rect;
}

void ArgumentCoder<WebCore::PathRect>::encode(StreamConnectionEncoder& encoder, const WebCore::PathRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect)>, WebCore::FloatRect>);
    struct ShouldBeSameSizeAsPathRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathRect>, false> {
        WebCore::FloatRect rect;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathRect) == sizeof(WebCore::PathRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathRect, rect)
    >::value);
    encoder << instance.rect;
}

std::optional<WebCore::PathRect> ArgumentCoder<WebCore::PathRect>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathRect {
            WTFMove(*rect)
        }
    };
}

void ArgumentCoder<WebCore::PathRoundedRect>::encode(Encoder& encoder, const WebCore::PathRoundedRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.roundedRect)>, WebCore::FloatRoundedRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.strategy)>, WebCore::PathRoundedRect::Strategy>);
    struct ShouldBeSameSizeAsPathRoundedRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathRoundedRect>, false> {
        WebCore::FloatRoundedRect roundedRect;
        WebCore::PathRoundedRect::Strategy strategy;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathRoundedRect) == sizeof(WebCore::PathRoundedRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathRoundedRect, roundedRect)
        , offsetof(WebCore::PathRoundedRect, strategy)
    >::value);
    encoder << instance.roundedRect;
    encoder << instance.strategy;
}

void ArgumentCoder<WebCore::PathRoundedRect>::encode(StreamConnectionEncoder& encoder, const WebCore::PathRoundedRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.roundedRect)>, WebCore::FloatRoundedRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.strategy)>, WebCore::PathRoundedRect::Strategy>);
    struct ShouldBeSameSizeAsPathRoundedRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathRoundedRect>, false> {
        WebCore::FloatRoundedRect roundedRect;
        WebCore::PathRoundedRect::Strategy strategy;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathRoundedRect) == sizeof(WebCore::PathRoundedRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathRoundedRect, roundedRect)
        , offsetof(WebCore::PathRoundedRect, strategy)
    >::value);
    encoder << instance.roundedRect;
    encoder << instance.strategy;
}

std::optional<WebCore::PathRoundedRect> ArgumentCoder<WebCore::PathRoundedRect>::decode(Decoder& decoder)
{
    auto roundedRect = decoder.decode<WebCore::FloatRoundedRect>();
    auto strategy = decoder.decode<WebCore::PathRoundedRect::Strategy>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathRoundedRect {
            WTFMove(*roundedRect),
            WTFMove(*strategy)
        }
    };
}

void ArgumentCoder<WebCore::PathDataLine>::encode(Encoder& encoder, const WebCore::PathDataLine& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.end)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathDataLine : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataLine>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint end;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataLine) == sizeof(WebCore::PathDataLine));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataLine, start)
        , offsetof(WebCore::PathDataLine, end)
    >::value);
    encoder << instance.start;
    encoder << instance.end;
}

void ArgumentCoder<WebCore::PathDataLine>::encode(StreamConnectionEncoder& encoder, const WebCore::PathDataLine& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.end)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathDataLine : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataLine>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint end;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataLine) == sizeof(WebCore::PathDataLine));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataLine, start)
        , offsetof(WebCore::PathDataLine, end)
    >::value);
    encoder << instance.start;
    encoder << instance.end;
}

std::optional<WebCore::PathDataLine> ArgumentCoder<WebCore::PathDataLine>::decode(Decoder& decoder)
{
    auto start = decoder.decode<WebCore::FloatPoint>();
    auto end = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathDataLine {
            WTFMove(*start),
            WTFMove(*end)
        }
    };
}

void ArgumentCoder<WebCore::PathDataQuadCurve>::encode(Encoder& encoder, const WebCore::PathDataQuadCurve& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathDataQuadCurve : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataQuadCurve>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint controlPoint;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataQuadCurve) == sizeof(WebCore::PathDataQuadCurve));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataQuadCurve, start)
        , offsetof(WebCore::PathDataQuadCurve, controlPoint)
        , offsetof(WebCore::PathDataQuadCurve, endPoint)
    >::value);
    encoder << instance.start;
    encoder << instance.controlPoint;
    encoder << instance.endPoint;
}

void ArgumentCoder<WebCore::PathDataQuadCurve>::encode(StreamConnectionEncoder& encoder, const WebCore::PathDataQuadCurve& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathDataQuadCurve : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataQuadCurve>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint controlPoint;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataQuadCurve) == sizeof(WebCore::PathDataQuadCurve));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataQuadCurve, start)
        , offsetof(WebCore::PathDataQuadCurve, controlPoint)
        , offsetof(WebCore::PathDataQuadCurve, endPoint)
    >::value);
    encoder << instance.start;
    encoder << instance.controlPoint;
    encoder << instance.endPoint;
}

std::optional<WebCore::PathDataQuadCurve> ArgumentCoder<WebCore::PathDataQuadCurve>::decode(Decoder& decoder)
{
    auto start = decoder.decode<WebCore::FloatPoint>();
    auto controlPoint = decoder.decode<WebCore::FloatPoint>();
    auto endPoint = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathDataQuadCurve {
            WTFMove(*start),
            WTFMove(*controlPoint),
            WTFMove(*endPoint)
        }
    };
}

void ArgumentCoder<WebCore::PathDataBezierCurve>::encode(Encoder& encoder, const WebCore::PathDataBezierCurve& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathDataBezierCurve : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataBezierCurve>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataBezierCurve) == sizeof(WebCore::PathDataBezierCurve));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataBezierCurve, start)
        , offsetof(WebCore::PathDataBezierCurve, controlPoint1)
        , offsetof(WebCore::PathDataBezierCurve, controlPoint2)
        , offsetof(WebCore::PathDataBezierCurve, endPoint)
    >::value);
    encoder << instance.start;
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.endPoint;
}

void ArgumentCoder<WebCore::PathDataBezierCurve>::encode(StreamConnectionEncoder& encoder, const WebCore::PathDataBezierCurve& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endPoint)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsPathDataBezierCurve : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataBezierCurve>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        WebCore::FloatPoint endPoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataBezierCurve) == sizeof(WebCore::PathDataBezierCurve));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataBezierCurve, start)
        , offsetof(WebCore::PathDataBezierCurve, controlPoint1)
        , offsetof(WebCore::PathDataBezierCurve, controlPoint2)
        , offsetof(WebCore::PathDataBezierCurve, endPoint)
    >::value);
    encoder << instance.start;
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.endPoint;
}

std::optional<WebCore::PathDataBezierCurve> ArgumentCoder<WebCore::PathDataBezierCurve>::decode(Decoder& decoder)
{
    auto start = decoder.decode<WebCore::FloatPoint>();
    auto controlPoint1 = decoder.decode<WebCore::FloatPoint>();
    auto controlPoint2 = decoder.decode<WebCore::FloatPoint>();
    auto endPoint = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathDataBezierCurve {
            WTFMove(*start),
            WTFMove(*controlPoint1),
            WTFMove(*controlPoint2),
            WTFMove(*endPoint)
        }
    };
}

void ArgumentCoder<WebCore::PathDataArc>::encode(Encoder& encoder, const WebCore::PathDataArc& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    struct ShouldBeSameSizeAsPathDataArc : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataArc>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        float radius;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataArc) == sizeof(WebCore::PathDataArc));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataArc, start)
        , offsetof(WebCore::PathDataArc, controlPoint1)
        , offsetof(WebCore::PathDataArc, controlPoint2)
        , offsetof(WebCore::PathDataArc, radius)
    >::value);
    encoder << instance.start;
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.radius;
}

void ArgumentCoder<WebCore::PathDataArc>::encode(StreamConnectionEncoder& encoder, const WebCore::PathDataArc& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlPoint2)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    struct ShouldBeSameSizeAsPathDataArc : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathDataArc>, false> {
        WebCore::FloatPoint start;
        WebCore::FloatPoint controlPoint1;
        WebCore::FloatPoint controlPoint2;
        float radius;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathDataArc) == sizeof(WebCore::PathDataArc));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PathDataArc, start)
        , offsetof(WebCore::PathDataArc, controlPoint1)
        , offsetof(WebCore::PathDataArc, controlPoint2)
        , offsetof(WebCore::PathDataArc, radius)
    >::value);
    encoder << instance.start;
    encoder << instance.controlPoint1;
    encoder << instance.controlPoint2;
    encoder << instance.radius;
}

std::optional<WebCore::PathDataArc> ArgumentCoder<WebCore::PathDataArc>::decode(Decoder& decoder)
{
    auto start = decoder.decode<WebCore::FloatPoint>();
    auto controlPoint1 = decoder.decode<WebCore::FloatPoint>();
    auto controlPoint2 = decoder.decode<WebCore::FloatPoint>();
    auto radius = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathDataArc {
            WTFMove(*start),
            WTFMove(*controlPoint1),
            WTFMove(*controlPoint2),
            WTFMove(*radius)
        }
    };
}

void ArgumentCoder<WebCore::PathCloseSubpath>::encode(Encoder& encoder, const WebCore::PathCloseSubpath& instance)
{
    struct ShouldBeSameSizeAsPathCloseSubpath : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathCloseSubpath>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathCloseSubpath) == sizeof(WebCore::PathCloseSubpath));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebCore::PathCloseSubpath>::encode(StreamConnectionEncoder& encoder, const WebCore::PathCloseSubpath& instance)
{
    struct ShouldBeSameSizeAsPathCloseSubpath : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PathCloseSubpath>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsPathCloseSubpath) == sizeof(WebCore::PathCloseSubpath));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::PathCloseSubpath> ArgumentCoder<WebCore::PathCloseSubpath>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathCloseSubpath {
        }
    };
}

void ArgumentCoder<WebCore::PathSegment>::encode(Encoder& encoder, const WebCore::PathSegment& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, WebCore::PathSegment::Data>);
    encoder << instance.data();
}

void ArgumentCoder<WebCore::PathSegment>::encode(StreamConnectionEncoder& encoder, const WebCore::PathSegment& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, WebCore::PathSegment::Data>);
    encoder << instance.data();
}

std::optional<WebCore::PathSegment> ArgumentCoder<WebCore::PathSegment>::decode(Decoder& decoder)
{
    auto data = decoder.decode<WebCore::PathSegment::Data>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PathSegment {
            WTFMove(*data)
        }
    };
}

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::RemoteRenderingBackendCreationParameters>::encode(Encoder& encoder, const WebKit::RemoteRenderingBackendCreationParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebKit::RenderingBackendIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageProxyID)>, WebKit::WebPageProxyIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageID)>, WebCore::PageIdentifier>);
    struct ShouldBeSameSizeAsRemoteRenderingBackendCreationParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteRenderingBackendCreationParameters>, false> {
        WebKit::RenderingBackendIdentifier identifier;
        WebKit::WebPageProxyIdentifier pageProxyID;
        WebCore::PageIdentifier pageID;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteRenderingBackendCreationParameters) == sizeof(WebKit::RemoteRenderingBackendCreationParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteRenderingBackendCreationParameters, identifier)
        , offsetof(WebKit::RemoteRenderingBackendCreationParameters, pageProxyID)
        , offsetof(WebKit::RemoteRenderingBackendCreationParameters, pageID)
    >::value);
    encoder << instance.identifier;
    encoder << instance.pageProxyID;
    encoder << instance.pageID;
}

std::optional<WebKit::RemoteRenderingBackendCreationParameters> ArgumentCoder<WebKit::RemoteRenderingBackendCreationParameters>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebKit::RenderingBackendIdentifier>();
    auto pageProxyID = decoder.decode<WebKit::WebPageProxyIdentifier>();
    auto pageID = decoder.decode<WebCore::PageIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::RemoteRenderingBackendCreationParameters {
            WTFMove(*identifier),
            WTFMove(*pageProxyID),
            WTFMove(*pageID)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::RemoteGPURequestAdapterResponse>::encode(Encoder& encoder, const WebKit::RemoteGPURequestAdapterResponse& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.features)>, WebKit::WebGPU::SupportedFeatures>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.limits)>, WebKit::WebGPU::SupportedLimits>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFallbackAdapter)>, bool>);
    struct ShouldBeSameSizeAsRemoteGPURequestAdapterResponse : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteGPURequestAdapterResponse>, false> {
        String name;
        WebKit::WebGPU::SupportedFeatures features;
        WebKit::WebGPU::SupportedLimits limits;
        bool isFallbackAdapter;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteGPURequestAdapterResponse) == sizeof(WebKit::RemoteGPURequestAdapterResponse));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, name)
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, features)
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, limits)
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, isFallbackAdapter)
    >::value);
    encoder << instance.name;
    encoder << instance.features;
    encoder << instance.limits;
    encoder << instance.isFallbackAdapter;
}

void ArgumentCoder<WebKit::RemoteGPURequestAdapterResponse>::encode(StreamConnectionEncoder& encoder, const WebKit::RemoteGPURequestAdapterResponse& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.features)>, WebKit::WebGPU::SupportedFeatures>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.limits)>, WebKit::WebGPU::SupportedLimits>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFallbackAdapter)>, bool>);
    struct ShouldBeSameSizeAsRemoteGPURequestAdapterResponse : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteGPURequestAdapterResponse>, false> {
        String name;
        WebKit::WebGPU::SupportedFeatures features;
        WebKit::WebGPU::SupportedLimits limits;
        bool isFallbackAdapter;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteGPURequestAdapterResponse) == sizeof(WebKit::RemoteGPURequestAdapterResponse));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, name)
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, features)
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, limits)
        , offsetof(WebKit::RemoteGPURequestAdapterResponse, isFallbackAdapter)
    >::value);
    encoder << instance.name;
    encoder << instance.features;
    encoder << instance.limits;
    encoder << instance.isFallbackAdapter;
}

std::optional<WebKit::RemoteGPURequestAdapterResponse> ArgumentCoder<WebKit::RemoteGPURequestAdapterResponse>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto features = decoder.decode<WebKit::WebGPU::SupportedFeatures>();
    auto limits = decoder.decode<WebKit::WebGPU::SupportedLimits>();
    auto isFallbackAdapter = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::RemoteGPURequestAdapterResponse {
            WTFMove(*name),
            WTFMove(*features),
            WTFMove(*limits),
            WTFMove(*isFallbackAdapter)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
void ArgumentCoder<WebKit::AudioTrackPrivateRemoteConfiguration>::encode(Encoder& encoder, const WebKit::AudioTrackPrivateRemoteConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackId)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.language)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startTimeVariance)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackIndex)>, int>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, trackId)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, label)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, language)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, startTimeVariance)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, trackIndex)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kind)>, WebCore::AudioTrackPrivate::Kind>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackConfiguration)>, WebCore::PlatformAudioTrackConfiguration>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::AudioTrackPrivateRemoteConfiguration, enabled)
        , offsetof(WebKit::AudioTrackPrivateRemoteConfiguration, kind)
        , offsetof(WebKit::AudioTrackPrivateRemoteConfiguration, trackConfiguration)
    >::value);
    encoder << instance.trackId;
    encoder << instance.label;
    encoder << instance.language;
    encoder << instance.startTimeVariance;
    encoder << instance.trackIndex;
    encoder << instance.enabled;
    encoder << instance.kind;
    encoder << instance.trackConfiguration;
}

std::optional<WebKit::AudioTrackPrivateRemoteConfiguration> ArgumentCoder<WebKit::AudioTrackPrivateRemoteConfiguration>::decode(Decoder& decoder)
{
    auto trackId = decoder.decode<AtomString>();
    auto label = decoder.decode<AtomString>();
    auto language = decoder.decode<AtomString>();
    auto startTimeVariance = decoder.decode<MediaTime>();
    auto trackIndex = decoder.decode<int>();
    auto enabled = decoder.decode<bool>();
    auto kind = decoder.decode<WebCore::AudioTrackPrivate::Kind>();
    auto trackConfiguration = decoder.decode<WebCore::PlatformAudioTrackConfiguration>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::AudioTrackPrivateRemoteConfiguration {
            WebKit::TrackPrivateRemoteConfiguration {
                WTFMove(*trackId),
                WTFMove(*label),
                WTFMove(*language),
                WTFMove(*startTimeVariance),
                WTFMove(*trackIndex)
            },
            WTFMove(*enabled),
            WTFMove(*kind),
            WTFMove(*trackConfiguration)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
void ArgumentCoder<WebKit::InitializationSegmentInfo>::encode(Encoder& encoder, const WebKit::InitializationSegmentInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.audioTracks)>, Vector<WebKit::InitializationSegmentInfo::TrackInformation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.videoTracks)>, Vector<WebKit::InitializationSegmentInfo::TrackInformation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textTracks)>, Vector<WebKit::InitializationSegmentInfo::TrackInformation>>);
    struct ShouldBeSameSizeAsInitializationSegmentInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::InitializationSegmentInfo>, false> {
        MediaTime duration;
        Vector<WebKit::InitializationSegmentInfo::TrackInformation> audioTracks;
        Vector<WebKit::InitializationSegmentInfo::TrackInformation> videoTracks;
        Vector<WebKit::InitializationSegmentInfo::TrackInformation> textTracks;
    };
    static_assert(sizeof(ShouldBeSameSizeAsInitializationSegmentInfo) == sizeof(WebKit::InitializationSegmentInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::InitializationSegmentInfo, duration)
        , offsetof(WebKit::InitializationSegmentInfo, audioTracks)
        , offsetof(WebKit::InitializationSegmentInfo, videoTracks)
        , offsetof(WebKit::InitializationSegmentInfo, textTracks)
    >::value);
    encoder << instance.duration;
    encoder << instance.audioTracks;
    encoder << instance.videoTracks;
    encoder << instance.textTracks;
}

std::optional<WebKit::InitializationSegmentInfo> ArgumentCoder<WebKit::InitializationSegmentInfo>::decode(Decoder& decoder)
{
    auto duration = decoder.decode<MediaTime>();
    auto audioTracks = decoder.decode<Vector<WebKit::InitializationSegmentInfo::TrackInformation>>();
    auto videoTracks = decoder.decode<Vector<WebKit::InitializationSegmentInfo::TrackInformation>>();
    auto textTracks = decoder.decode<Vector<WebKit::InitializationSegmentInfo::TrackInformation>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::InitializationSegmentInfo {
            WTFMove(*duration),
            WTFMove(*audioTracks),
            WTFMove(*videoTracks),
            WTFMove(*textTracks)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
void ArgumentCoder<WebKit::InitializationSegmentInfo::TrackInformation>::encode(Encoder& encoder, const WebKit::InitializationSegmentInfo::TrackInformation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.description)>, WebKit::MediaDescriptionInfo>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebKit::TrackPrivateRemoteIdentifier>);
    struct ShouldBeSameSizeAsTrackInformation : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::InitializationSegmentInfo::TrackInformation>, false> {
        WebKit::MediaDescriptionInfo description;
        WebKit::TrackPrivateRemoteIdentifier identifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTrackInformation) == sizeof(WebKit::InitializationSegmentInfo::TrackInformation));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::InitializationSegmentInfo::TrackInformation, description)
        , offsetof(WebKit::InitializationSegmentInfo::TrackInformation, identifier)
    >::value);
    encoder << instance.description;
    encoder << instance.identifier;
}

std::optional<WebKit::InitializationSegmentInfo::TrackInformation> ArgumentCoder<WebKit::InitializationSegmentInfo::TrackInformation>::decode(Decoder& decoder)
{
    auto description = decoder.decode<WebKit::MediaDescriptionInfo>();
    auto identifier = decoder.decode<WebKit::TrackPrivateRemoteIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::InitializationSegmentInfo::TrackInformation {
            WTFMove(*description),
            WTFMove(*identifier)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
void ArgumentCoder<WebKit::MediaDescriptionInfo>::encode(Encoder& encoder, const WebKit::MediaDescriptionInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_codec)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isVideo)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isAudio)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isText)>, bool>);
    struct ShouldBeSameSizeAsMediaDescriptionInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::MediaDescriptionInfo>, false> {
        AtomString m_codec;
        bool m_isVideo;
        bool m_isAudio;
        bool m_isText;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaDescriptionInfo) == sizeof(WebKit::MediaDescriptionInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::MediaDescriptionInfo, m_codec)
        , offsetof(WebKit::MediaDescriptionInfo, m_isVideo)
        , offsetof(WebKit::MediaDescriptionInfo, m_isAudio)
        , offsetof(WebKit::MediaDescriptionInfo, m_isText)
    >::value);
    encoder << instance.m_codec;
    encoder << instance.m_isVideo;
    encoder << instance.m_isAudio;
    encoder << instance.m_isText;
}

std::optional<WebKit::MediaDescriptionInfo> ArgumentCoder<WebKit::MediaDescriptionInfo>::decode(Decoder& decoder)
{
    auto m_codec = decoder.decode<AtomString>();
    auto m_isVideo = decoder.decode<bool>();
    auto m_isAudio = decoder.decode<bool>();
    auto m_isText = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::MediaDescriptionInfo {
            WTFMove(*m_codec),
            WTFMove(*m_isVideo),
            WTFMove(*m_isAudio),
            WTFMove(*m_isText)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::RemoteMediaPlayerProxyConfiguration>::encode(Encoder& encoder, const WebKit::RemoteMediaPlayerProxyConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrer)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userAgent)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceApplicationIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkInterfaceName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaContentTypesRequiringHardwareSupport)>, Vector<WebCore::ContentType>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaContainerTypes)>, std::optional<Vector<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaCodecTypes)>, std::optional<Vector<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaVideoCodecIDs)>, std::optional<Vector<WebCore::FourCC>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaAudioCodecIDs)>, std::optional<Vector<WebCore::FourCC>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaCaptionFormatTypes)>, std::optional<Vector<WebCore::FourCC>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.playerContentBoxRect)>, WebCore::LayoutRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preferredAudioCharacteristics)>, Vector<String>>);
#if ENABLE(AVF_CAPTIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.outOfBandTrackData)>, Vector<WebCore::PlatformTextTrackData>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentSecurityOrigin)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentationSize)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.videoInlineSize)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.logIdentifier)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldUsePersistentCache)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isVideo)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.renderingCanBeAccelerated)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.prefersSandboxedParsing)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldDisableHDR)>, bool>);
    struct ShouldBeSameSizeAsRemoteMediaPlayerProxyConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteMediaPlayerProxyConfiguration>, false> {
        String referrer;
        String userAgent;
        String sourceApplicationIdentifier;
        String networkInterfaceName;
        Vector<WebCore::ContentType> mediaContentTypesRequiringHardwareSupport;
        std::optional<Vector<String>> allowedMediaContainerTypes;
        std::optional<Vector<String>> allowedMediaCodecTypes;
        std::optional<Vector<WebCore::FourCC>> allowedMediaVideoCodecIDs;
        std::optional<Vector<WebCore::FourCC>> allowedMediaAudioCodecIDs;
        std::optional<Vector<WebCore::FourCC>> allowedMediaCaptionFormatTypes;
        WebCore::LayoutRect playerContentBoxRect;
        Vector<String> preferredAudioCharacteristics;
#if ENABLE(AVF_CAPTIONS)
        Vector<WebCore::PlatformTextTrackData> outOfBandTrackData;
#endif
        WebCore::SecurityOriginData documentSecurityOrigin;
        WebCore::IntSize presentationSize;
        WebCore::FloatSize videoInlineSize;
        uint64_t logIdentifier;
        bool shouldUsePersistentCache;
        bool isVideo;
        bool renderingCanBeAccelerated;
        bool prefersSandboxedParsing;
        bool shouldDisableHDR;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteMediaPlayerProxyConfiguration) == sizeof(WebKit::RemoteMediaPlayerProxyConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, referrer)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, userAgent)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, sourceApplicationIdentifier)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, networkInterfaceName)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, mediaContentTypesRequiringHardwareSupport)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, allowedMediaContainerTypes)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, allowedMediaCodecTypes)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, allowedMediaVideoCodecIDs)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, allowedMediaAudioCodecIDs)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, allowedMediaCaptionFormatTypes)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, playerContentBoxRect)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, preferredAudioCharacteristics)
#if ENABLE(AVF_CAPTIONS)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, outOfBandTrackData)
#endif
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, documentSecurityOrigin)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, presentationSize)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, videoInlineSize)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, logIdentifier)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, shouldUsePersistentCache)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, isVideo)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, renderingCanBeAccelerated)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, prefersSandboxedParsing)
        , offsetof(WebKit::RemoteMediaPlayerProxyConfiguration, shouldDisableHDR)
    >::value);
    encoder << instance.referrer;
    encoder << instance.userAgent;
    encoder << instance.sourceApplicationIdentifier;
    encoder << instance.networkInterfaceName;
    encoder << instance.mediaContentTypesRequiringHardwareSupport;
    encoder << instance.allowedMediaContainerTypes;
    encoder << instance.allowedMediaCodecTypes;
    encoder << instance.allowedMediaVideoCodecIDs;
    encoder << instance.allowedMediaAudioCodecIDs;
    encoder << instance.allowedMediaCaptionFormatTypes;
    encoder << instance.playerContentBoxRect;
    encoder << instance.preferredAudioCharacteristics;
#if ENABLE(AVF_CAPTIONS)
    encoder << instance.outOfBandTrackData;
#endif
    encoder << instance.documentSecurityOrigin;
    encoder << instance.presentationSize;
    encoder << instance.videoInlineSize;
    encoder << instance.logIdentifier;
    encoder << instance.shouldUsePersistentCache;
    encoder << instance.isVideo;
    encoder << instance.renderingCanBeAccelerated;
    encoder << instance.prefersSandboxedParsing;
    encoder << instance.shouldDisableHDR;
}

std::optional<WebKit::RemoteMediaPlayerProxyConfiguration> ArgumentCoder<WebKit::RemoteMediaPlayerProxyConfiguration>::decode(Decoder& decoder)
{
    auto referrer = decoder.decode<String>();
    auto userAgent = decoder.decode<String>();
    auto sourceApplicationIdentifier = decoder.decode<String>();
    auto networkInterfaceName = decoder.decode<String>();
    auto mediaContentTypesRequiringHardwareSupport = decoder.decode<Vector<WebCore::ContentType>>();
    auto allowedMediaContainerTypes = decoder.decode<std::optional<Vector<String>>>();
    auto allowedMediaCodecTypes = decoder.decode<std::optional<Vector<String>>>();
    auto allowedMediaVideoCodecIDs = decoder.decode<std::optional<Vector<WebCore::FourCC>>>();
    auto allowedMediaAudioCodecIDs = decoder.decode<std::optional<Vector<WebCore::FourCC>>>();
    auto allowedMediaCaptionFormatTypes = decoder.decode<std::optional<Vector<WebCore::FourCC>>>();
    auto playerContentBoxRect = decoder.decode<WebCore::LayoutRect>();
    auto preferredAudioCharacteristics = decoder.decode<Vector<String>>();
#if ENABLE(AVF_CAPTIONS)
    auto outOfBandTrackData = decoder.decode<Vector<WebCore::PlatformTextTrackData>>();
#endif
    auto documentSecurityOrigin = decoder.decode<WebCore::SecurityOriginData>();
    auto presentationSize = decoder.decode<WebCore::IntSize>();
    auto videoInlineSize = decoder.decode<WebCore::FloatSize>();
    auto logIdentifier = decoder.decode<uint64_t>();
    auto shouldUsePersistentCache = decoder.decode<bool>();
    auto isVideo = decoder.decode<bool>();
    auto renderingCanBeAccelerated = decoder.decode<bool>();
    auto prefersSandboxedParsing = decoder.decode<bool>();
    auto shouldDisableHDR = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::RemoteMediaPlayerProxyConfiguration {
            WTFMove(*referrer),
            WTFMove(*userAgent),
            WTFMove(*sourceApplicationIdentifier),
            WTFMove(*networkInterfaceName),
            WTFMove(*mediaContentTypesRequiringHardwareSupport),
            WTFMove(*allowedMediaContainerTypes),
            WTFMove(*allowedMediaCodecTypes),
            WTFMove(*allowedMediaVideoCodecIDs),
            WTFMove(*allowedMediaAudioCodecIDs),
            WTFMove(*allowedMediaCaptionFormatTypes),
            WTFMove(*playerContentBoxRect),
            WTFMove(*preferredAudioCharacteristics),
#if ENABLE(AVF_CAPTIONS)
            WTFMove(*outOfBandTrackData),
#endif
            WTFMove(*documentSecurityOrigin),
            WTFMove(*presentationSize),
            WTFMove(*videoInlineSize),
            WTFMove(*logIdentifier),
            WTFMove(*shouldUsePersistentCache),
            WTFMove(*isVideo),
            WTFMove(*renderingCanBeAccelerated),
            WTFMove(*prefersSandboxedParsing),
            WTFMove(*shouldDisableHDR)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
void ArgumentCoder<WebKit::TextTrackPrivateRemoteConfiguration>::encode(Encoder& encoder, const WebKit::TextTrackPrivateRemoteConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackId)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.language)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.inBandMetadataTrackDispatchType)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startTimeVariance)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackIndex)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cueFormat)>, WebCore::InbandTextTrackPrivate::CueFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kind)>, WebCore::InbandTextTrackPrivate::Kind>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isClosedCaptions)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSDH)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containsOnlyForcedSubtitles)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMainProgramContent)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isEasyToRead)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isDefault)>, bool>);
    struct ShouldBeSameSizeAsTextTrackPrivateRemoteConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::TextTrackPrivateRemoteConfiguration>, false> {
        AtomString trackId;
        AtomString label;
        AtomString language;
        AtomString inBandMetadataTrackDispatchType;
        MediaTime startTimeVariance;
        int trackIndex;
        WebCore::InbandTextTrackPrivate::CueFormat cueFormat;
        WebCore::InbandTextTrackPrivate::Kind kind;
        bool isClosedCaptions;
        bool isSDH;
        bool containsOnlyForcedSubtitles;
        bool isMainProgramContent;
        bool isEasyToRead;
        bool isDefault;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextTrackPrivateRemoteConfiguration) == sizeof(WebKit::TextTrackPrivateRemoteConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, trackId)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, label)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, language)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, inBandMetadataTrackDispatchType)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, startTimeVariance)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, trackIndex)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, cueFormat)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, kind)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, isClosedCaptions)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, isSDH)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, containsOnlyForcedSubtitles)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, isMainProgramContent)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, isEasyToRead)
        , offsetof(WebKit::TextTrackPrivateRemoteConfiguration, isDefault)
    >::value);
    encoder << instance.trackId;
    encoder << instance.label;
    encoder << instance.language;
    encoder << instance.inBandMetadataTrackDispatchType;
    encoder << instance.startTimeVariance;
    encoder << instance.trackIndex;
    encoder << instance.cueFormat;
    encoder << instance.kind;
    encoder << instance.isClosedCaptions;
    encoder << instance.isSDH;
    encoder << instance.containsOnlyForcedSubtitles;
    encoder << instance.isMainProgramContent;
    encoder << instance.isEasyToRead;
    encoder << instance.isDefault;
}

std::optional<WebKit::TextTrackPrivateRemoteConfiguration> ArgumentCoder<WebKit::TextTrackPrivateRemoteConfiguration>::decode(Decoder& decoder)
{
    auto trackId = decoder.decode<AtomString>();
    auto label = decoder.decode<AtomString>();
    auto language = decoder.decode<AtomString>();
    auto inBandMetadataTrackDispatchType = decoder.decode<AtomString>();
    auto startTimeVariance = decoder.decode<MediaTime>();
    auto trackIndex = decoder.decode<int>();
    auto cueFormat = decoder.decode<WebCore::InbandTextTrackPrivate::CueFormat>();
    auto kind = decoder.decode<WebCore::InbandTextTrackPrivate::Kind>();
    auto isClosedCaptions = decoder.decode<bool>();
    auto isSDH = decoder.decode<bool>();
    auto containsOnlyForcedSubtitles = decoder.decode<bool>();
    auto isMainProgramContent = decoder.decode<bool>();
    auto isEasyToRead = decoder.decode<bool>();
    auto isDefault = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::TextTrackPrivateRemoteConfiguration {
            WTFMove(*trackId),
            WTFMove(*label),
            WTFMove(*language),
            WTFMove(*inBandMetadataTrackDispatchType),
            WTFMove(*startTimeVariance),
            WTFMove(*trackIndex),
            WTFMove(*cueFormat),
            WTFMove(*kind),
            WTFMove(*isClosedCaptions),
            WTFMove(*isSDH),
            WTFMove(*containsOnlyForcedSubtitles),
            WTFMove(*isMainProgramContent),
            WTFMove(*isEasyToRead),
            WTFMove(*isDefault)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
void ArgumentCoder<WebKit::TrackPrivateRemoteConfiguration>::encode(Encoder& encoder, const WebKit::TrackPrivateRemoteConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackId)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.language)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startTimeVariance)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackIndex)>, int>);
    struct ShouldBeSameSizeAsTrackPrivateRemoteConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::TrackPrivateRemoteConfiguration>, false> {
        AtomString trackId;
        AtomString label;
        AtomString language;
        MediaTime startTimeVariance;
        int trackIndex;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTrackPrivateRemoteConfiguration) == sizeof(WebKit::TrackPrivateRemoteConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, trackId)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, label)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, language)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, startTimeVariance)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, trackIndex)
    >::value);
    encoder << instance.trackId;
    encoder << instance.label;
    encoder << instance.language;
    encoder << instance.startTimeVariance;
    encoder << instance.trackIndex;
}

std::optional<WebKit::TrackPrivateRemoteConfiguration> ArgumentCoder<WebKit::TrackPrivateRemoteConfiguration>::decode(Decoder& decoder)
{
    auto trackId = decoder.decode<AtomString>();
    auto label = decoder.decode<AtomString>();
    auto language = decoder.decode<AtomString>();
    auto startTimeVariance = decoder.decode<MediaTime>();
    auto trackIndex = decoder.decode<int>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::TrackPrivateRemoteConfiguration {
            WTFMove(*trackId),
            WTFMove(*label),
            WTFMove(*language),
            WTFMove(*startTimeVariance),
            WTFMove(*trackIndex)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
void ArgumentCoder<WebKit::VideoTrackPrivateRemoteConfiguration>::encode(Encoder& encoder, const WebKit::VideoTrackPrivateRemoteConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackId)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.language)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startTimeVariance)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackIndex)>, int>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, trackId)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, label)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, language)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, startTimeVariance)
        , offsetof(WebKit::TrackPrivateRemoteConfiguration, trackIndex)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selected)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kind)>, WebCore::VideoTrackPrivate::Kind>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackConfiguration)>, WebCore::PlatformVideoTrackConfiguration>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::VideoTrackPrivateRemoteConfiguration, selected)
        , offsetof(WebKit::VideoTrackPrivateRemoteConfiguration, kind)
        , offsetof(WebKit::VideoTrackPrivateRemoteConfiguration, trackConfiguration)
    >::value);
    encoder << instance.trackId;
    encoder << instance.label;
    encoder << instance.language;
    encoder << instance.startTimeVariance;
    encoder << instance.trackIndex;
    encoder << instance.selected;
    encoder << instance.kind;
    encoder << instance.trackConfiguration;
}

std::optional<WebKit::VideoTrackPrivateRemoteConfiguration> ArgumentCoder<WebKit::VideoTrackPrivateRemoteConfiguration>::decode(Decoder& decoder)
{
    auto trackId = decoder.decode<AtomString>();
    auto label = decoder.decode<AtomString>();
    auto language = decoder.decode<AtomString>();
    auto startTimeVariance = decoder.decode<MediaTime>();
    auto trackIndex = decoder.decode<int>();
    auto selected = decoder.decode<bool>();
    auto kind = decoder.decode<WebCore::VideoTrackPrivate::Kind>();
    auto trackConfiguration = decoder.decode<WebCore::PlatformVideoTrackConfiguration>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::VideoTrackPrivateRemoteConfiguration {
            WebKit::TrackPrivateRemoteConfiguration {
                WTFMove(*trackId),
                WTFMove(*label),
                WTFMove(*language),
                WTFMove(*startTimeVariance),
                WTFMove(*trackIndex)
            },
            WTFMove(*selected),
            WTFMove(*kind),
            WTFMove(*trackConfiguration)
        }
    };
}

#endif

void ArgumentCoder<WebKit::NetworkProcessCreationParameters>::encode(Encoder& encoder, const WebKit::NetworkProcessCreationParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.auxiliaryProcessParameters)>, WebKit::AuxiliaryProcessCreationParameters>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cacheModel)>, WebKit::CacheModel>);
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uiProcessCookieStorageIdentifier)>, Vector<uint8_t>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldSuppressMemoryPressureHandler)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredForCustomProtocols)>, Vector<String>>);
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uiProcessBundleIdentifier)>, String>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkATSContext)>, RetainPtr<CFDataRef>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.strictSecureDecodingForAllObjCEnabled)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isParentProcessFullWebBrowserOrRunningTest)>, bool>);
#endif
#if USE(SOUP)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cookieAcceptPolicy)>, WebCore::HTTPCookieAcceptPolicy>);
#endif
#if USE(SOUP)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.languages)>, Vector<String>>);
#endif
#if USE(SOUP)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.memoryPressureHandlerConfiguration)>, std::optional<MemoryPressureHandler::Configuration>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsSecure)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsBypassingContentSecurityPolicy)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsLocal)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsNoAccess)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enablePrivateClickMeasurement)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ftpEnabled)>, bool>);
#if ENABLE(BUILT_IN_NOTIFICATIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.builtInNotificationsEnabled)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.websiteDataStoreParameters)>, Vector<WebKit::WebsiteDataStoreParameters>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedFirstPartiesForCookies)>, Vector<std::pair<WebCore::ProcessIdentifier, WebCore::RegistrableDomain>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localhostAliasesForTesting)>, HashSet<String>>);
    struct ShouldBeSameSizeAsNetworkProcessCreationParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::NetworkProcessCreationParameters>, false> {
        WebKit::AuxiliaryProcessCreationParameters auxiliaryProcessParameters;
        WebKit::CacheModel cacheModel;
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
        Vector<uint8_t> uiProcessCookieStorageIdentifier;
#endif
        bool shouldSuppressMemoryPressureHandler;
        Vector<String> urlSchemesRegisteredForCustomProtocols;
#if PLATFORM(COCOA)
        String uiProcessBundleIdentifier;
#endif
#if PLATFORM(COCOA)
        RetainPtr<CFDataRef> networkATSContext;
#endif
#if PLATFORM(COCOA)
        bool strictSecureDecodingForAllObjCEnabled;
#endif
#if PLATFORM(COCOA)
        bool isParentProcessFullWebBrowserOrRunningTest;
#endif
#if USE(SOUP)
        WebCore::HTTPCookieAcceptPolicy cookieAcceptPolicy;
#endif
#if USE(SOUP)
        Vector<String> languages;
#endif
#if USE(SOUP)
        std::optional<MemoryPressureHandler::Configuration> memoryPressureHandlerConfiguration;
#endif
        Vector<String> urlSchemesRegisteredAsSecure;
        Vector<String> urlSchemesRegisteredAsBypassingContentSecurityPolicy;
        Vector<String> urlSchemesRegisteredAsLocal;
        Vector<String> urlSchemesRegisteredAsNoAccess;
        bool enablePrivateClickMeasurement;
        bool ftpEnabled;
#if ENABLE(BUILT_IN_NOTIFICATIONS)
        bool builtInNotificationsEnabled;
#endif
        Vector<WebKit::WebsiteDataStoreParameters> websiteDataStoreParameters;
        Vector<std::pair<WebCore::ProcessIdentifier, WebCore::RegistrableDomain>> allowedFirstPartiesForCookies;
        HashSet<String> localhostAliasesForTesting;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNetworkProcessCreationParameters) == sizeof(WebKit::NetworkProcessCreationParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::NetworkProcessCreationParameters, auxiliaryProcessParameters)
        , offsetof(WebKit::NetworkProcessCreationParameters, cacheModel)
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
        , offsetof(WebKit::NetworkProcessCreationParameters, uiProcessCookieStorageIdentifier)
#endif
        , offsetof(WebKit::NetworkProcessCreationParameters, shouldSuppressMemoryPressureHandler)
        , offsetof(WebKit::NetworkProcessCreationParameters, urlSchemesRegisteredForCustomProtocols)
#if PLATFORM(COCOA)
        , offsetof(WebKit::NetworkProcessCreationParameters, uiProcessBundleIdentifier)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::NetworkProcessCreationParameters, networkATSContext)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::NetworkProcessCreationParameters, strictSecureDecodingForAllObjCEnabled)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::NetworkProcessCreationParameters, isParentProcessFullWebBrowserOrRunningTest)
#endif
#if USE(SOUP)
        , offsetof(WebKit::NetworkProcessCreationParameters, cookieAcceptPolicy)
#endif
#if USE(SOUP)
        , offsetof(WebKit::NetworkProcessCreationParameters, languages)
#endif
#if USE(SOUP)
        , offsetof(WebKit::NetworkProcessCreationParameters, memoryPressureHandlerConfiguration)
#endif
        , offsetof(WebKit::NetworkProcessCreationParameters, urlSchemesRegisteredAsSecure)
        , offsetof(WebKit::NetworkProcessCreationParameters, urlSchemesRegisteredAsBypassingContentSecurityPolicy)
        , offsetof(WebKit::NetworkProcessCreationParameters, urlSchemesRegisteredAsLocal)
        , offsetof(WebKit::NetworkProcessCreationParameters, urlSchemesRegisteredAsNoAccess)
        , offsetof(WebKit::NetworkProcessCreationParameters, enablePrivateClickMeasurement)
        , offsetof(WebKit::NetworkProcessCreationParameters, ftpEnabled)
#if ENABLE(BUILT_IN_NOTIFICATIONS)
        , offsetof(WebKit::NetworkProcessCreationParameters, builtInNotificationsEnabled)
#endif
        , offsetof(WebKit::NetworkProcessCreationParameters, websiteDataStoreParameters)
        , offsetof(WebKit::NetworkProcessCreationParameters, allowedFirstPartiesForCookies)
        , offsetof(WebKit::NetworkProcessCreationParameters, localhostAliasesForTesting)
    >::value);
    encoder << instance.auxiliaryProcessParameters;
    encoder << instance.cacheModel;
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    encoder << instance.uiProcessCookieStorageIdentifier;
#endif
    encoder << instance.shouldSuppressMemoryPressureHandler;
    encoder << instance.urlSchemesRegisteredForCustomProtocols;
#if PLATFORM(COCOA)
    encoder << instance.uiProcessBundleIdentifier;
#endif
#if PLATFORM(COCOA)
    encoder << instance.networkATSContext;
#endif
#if PLATFORM(COCOA)
    encoder << instance.strictSecureDecodingForAllObjCEnabled;
#endif
#if PLATFORM(COCOA)
    encoder << instance.isParentProcessFullWebBrowserOrRunningTest;
#endif
#if USE(SOUP)
    encoder << instance.cookieAcceptPolicy;
#endif
#if USE(SOUP)
    encoder << instance.languages;
#endif
#if USE(SOUP)
    encoder << instance.memoryPressureHandlerConfiguration;
#endif
    encoder << instance.urlSchemesRegisteredAsSecure;
    encoder << instance.urlSchemesRegisteredAsBypassingContentSecurityPolicy;
    encoder << instance.urlSchemesRegisteredAsLocal;
    encoder << instance.urlSchemesRegisteredAsNoAccess;
    encoder << instance.enablePrivateClickMeasurement;
    encoder << instance.ftpEnabled;
#if ENABLE(BUILT_IN_NOTIFICATIONS)
    encoder << instance.builtInNotificationsEnabled;
#endif
    encoder << instance.websiteDataStoreParameters;
    encoder << instance.allowedFirstPartiesForCookies;
    encoder << instance.localhostAliasesForTesting;
}

std::optional<WebKit::NetworkProcessCreationParameters> ArgumentCoder<WebKit::NetworkProcessCreationParameters>::decode(Decoder& decoder)
{
    auto auxiliaryProcessParameters = decoder.decode<WebKit::AuxiliaryProcessCreationParameters>();
    auto cacheModel = decoder.decode<WebKit::CacheModel>();
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    auto uiProcessCookieStorageIdentifier = decoder.decode<Vector<uint8_t>>();
#endif
    auto shouldSuppressMemoryPressureHandler = decoder.decode<bool>();
    auto urlSchemesRegisteredForCustomProtocols = decoder.decode<Vector<String>>();
#if PLATFORM(COCOA)
    auto uiProcessBundleIdentifier = decoder.decode<String>();
#endif
#if PLATFORM(COCOA)
    auto networkATSContext = decoder.decode<RetainPtr<CFDataRef>>();
#endif
#if PLATFORM(COCOA)
    auto strictSecureDecodingForAllObjCEnabled = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto isParentProcessFullWebBrowserOrRunningTest = decoder.decode<bool>();
#endif
#if USE(SOUP)
    auto cookieAcceptPolicy = decoder.decode<WebCore::HTTPCookieAcceptPolicy>();
#endif
#if USE(SOUP)
    auto languages = decoder.decode<Vector<String>>();
#endif
#if USE(SOUP)
    auto memoryPressureHandlerConfiguration = decoder.decode<std::optional<MemoryPressureHandler::Configuration>>();
#endif
    auto urlSchemesRegisteredAsSecure = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsBypassingContentSecurityPolicy = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsLocal = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsNoAccess = decoder.decode<Vector<String>>();
    auto enablePrivateClickMeasurement = decoder.decode<bool>();
    auto ftpEnabled = decoder.decode<bool>();
#if ENABLE(BUILT_IN_NOTIFICATIONS)
    auto builtInNotificationsEnabled = decoder.decode<bool>();
#endif
    auto websiteDataStoreParameters = decoder.decode<Vector<WebKit::WebsiteDataStoreParameters>>();
    auto allowedFirstPartiesForCookies = decoder.decode<Vector<std::pair<WebCore::ProcessIdentifier, WebCore::RegistrableDomain>>>();
    auto localhostAliasesForTesting = decoder.decode<HashSet<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::NetworkProcessCreationParameters {
            WTFMove(*auxiliaryProcessParameters),
            WTFMove(*cacheModel),
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
            WTFMove(*uiProcessCookieStorageIdentifier),
#endif
            WTFMove(*shouldSuppressMemoryPressureHandler),
            WTFMove(*urlSchemesRegisteredForCustomProtocols),
#if PLATFORM(COCOA)
            WTFMove(*uiProcessBundleIdentifier),
#endif
#if PLATFORM(COCOA)
            WTFMove(*networkATSContext),
#endif
#if PLATFORM(COCOA)
            WTFMove(*strictSecureDecodingForAllObjCEnabled),
#endif
#if PLATFORM(COCOA)
            WTFMove(*isParentProcessFullWebBrowserOrRunningTest),
#endif
#if USE(SOUP)
            WTFMove(*cookieAcceptPolicy),
#endif
#if USE(SOUP)
            WTFMove(*languages),
#endif
#if USE(SOUP)
            WTFMove(*memoryPressureHandlerConfiguration),
#endif
            WTFMove(*urlSchemesRegisteredAsSecure),
            WTFMove(*urlSchemesRegisteredAsBypassingContentSecurityPolicy),
            WTFMove(*urlSchemesRegisteredAsLocal),
            WTFMove(*urlSchemesRegisteredAsNoAccess),
            WTFMove(*enablePrivateClickMeasurement),
            WTFMove(*ftpEnabled),
#if ENABLE(BUILT_IN_NOTIFICATIONS)
            WTFMove(*builtInNotificationsEnabled),
#endif
            WTFMove(*websiteDataStoreParameters),
            WTFMove(*allowedFirstPartiesForCookies),
            WTFMove(*localhostAliasesForTesting)
        }
    };
}

void ArgumentCoder<WebKit::NetworkLoadParameters>::encode(Encoder& encoder, const WebKit::NetworkLoadParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webPageProxyID)>, WebKit::WebPageProxyIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webPageID)>, WebCore::PageIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webFrameID)>, WebCore::FrameIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topOrigin)>, RefPtr<WebCore::SecurityOrigin>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceOrigin)>, RefPtr<WebCore::SecurityOrigin>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentPID)>, WTF::ProcessID>);
#if HAVE(AUDIT_TOKEN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkProcessAuditToken)>, std::optional<audit_token_t>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.request)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSniffingPolicy)>, WebCore::ContentSniffingPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentEncodingSniffingPolicy)>, WebCore::ContentEncodingSniffingPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.storedCredentialsPolicy)>, WebCore::StoredCredentialsPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientCredentialPolicy)>, WebCore::ClientCredentialPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldClearReferrerOnHTTPSToHTTPRedirect)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.needsCertificateInfo)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMainFrameNavigation)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMainResourceNavigationForAnyFrame)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRelaxThirdPartyCookieBlocking)>, WebCore::ShouldRelaxThirdPartyCookieBlocking>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blobFileReferences)>, Vector<RefPtr<WebCore::BlobDataFileReference>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldPreconnectOnly)>, WebKit::PreconnectOnly>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkActivityTracker)>, std::optional<WebKit::NetworkActivityTracker>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isNavigatingToAppBoundDomain)>, std::optional<WebKit::NavigatingToAppBoundDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hadMainFrameMainResourcePrivateRelayed)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowPrivacyProxy)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.advancedPrivacyProtections)>, OptionSet<WebCore::AdvancedPrivacyProtections>>);
    struct ShouldBeSameSizeAsNetworkLoadParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::NetworkLoadParameters>, false> {
        WebKit::WebPageProxyIdentifier webPageProxyID;
        WebCore::PageIdentifier webPageID;
        WebCore::FrameIdentifier webFrameID;
        RefPtr<WebCore::SecurityOrigin> topOrigin;
        RefPtr<WebCore::SecurityOrigin> sourceOrigin;
        WTF::ProcessID parentPID;
#if HAVE(AUDIT_TOKEN)
        std::optional<audit_token_t> networkProcessAuditToken;
#endif
        WebCore::ResourceRequest request;
        WebCore::ContentSniffingPolicy contentSniffingPolicy;
        WebCore::ContentEncodingSniffingPolicy contentEncodingSniffingPolicy;
        WebCore::StoredCredentialsPolicy storedCredentialsPolicy;
        WebCore::ClientCredentialPolicy clientCredentialPolicy;
        bool shouldClearReferrerOnHTTPSToHTTPRedirect;
        bool needsCertificateInfo;
        bool isMainFrameNavigation;
        bool isMainResourceNavigationForAnyFrame;
        WebCore::ShouldRelaxThirdPartyCookieBlocking shouldRelaxThirdPartyCookieBlocking;
        Vector<RefPtr<WebCore::BlobDataFileReference>> blobFileReferences;
        WebKit::PreconnectOnly shouldPreconnectOnly;
        std::optional<WebKit::NetworkActivityTracker> networkActivityTracker;
        std::optional<WebKit::NavigatingToAppBoundDomain> isNavigatingToAppBoundDomain;
        bool hadMainFrameMainResourcePrivateRelayed;
        bool allowPrivacyProxy;
        OptionSet<WebCore::AdvancedPrivacyProtections> advancedPrivacyProtections;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNetworkLoadParameters) == sizeof(WebKit::NetworkLoadParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::NetworkLoadParameters, webPageProxyID)
        , offsetof(WebKit::NetworkLoadParameters, webPageID)
        , offsetof(WebKit::NetworkLoadParameters, webFrameID)
        , offsetof(WebKit::NetworkLoadParameters, topOrigin)
        , offsetof(WebKit::NetworkLoadParameters, sourceOrigin)
        , offsetof(WebKit::NetworkLoadParameters, parentPID)
#if HAVE(AUDIT_TOKEN)
        , offsetof(WebKit::NetworkLoadParameters, networkProcessAuditToken)
#endif
        , offsetof(WebKit::NetworkLoadParameters, request)
        , offsetof(WebKit::NetworkLoadParameters, contentSniffingPolicy)
        , offsetof(WebKit::NetworkLoadParameters, contentEncodingSniffingPolicy)
        , offsetof(WebKit::NetworkLoadParameters, storedCredentialsPolicy)
        , offsetof(WebKit::NetworkLoadParameters, clientCredentialPolicy)
        , offsetof(WebKit::NetworkLoadParameters, shouldClearReferrerOnHTTPSToHTTPRedirect)
        , offsetof(WebKit::NetworkLoadParameters, needsCertificateInfo)
        , offsetof(WebKit::NetworkLoadParameters, isMainFrameNavigation)
        , offsetof(WebKit::NetworkLoadParameters, isMainResourceNavigationForAnyFrame)
        , offsetof(WebKit::NetworkLoadParameters, shouldRelaxThirdPartyCookieBlocking)
        , offsetof(WebKit::NetworkLoadParameters, blobFileReferences)
        , offsetof(WebKit::NetworkLoadParameters, shouldPreconnectOnly)
        , offsetof(WebKit::NetworkLoadParameters, networkActivityTracker)
        , offsetof(WebKit::NetworkLoadParameters, isNavigatingToAppBoundDomain)
        , offsetof(WebKit::NetworkLoadParameters, hadMainFrameMainResourcePrivateRelayed)
        , offsetof(WebKit::NetworkLoadParameters, allowPrivacyProxy)
        , offsetof(WebKit::NetworkLoadParameters, advancedPrivacyProtections)
    >::value);
    encoder << instance.webPageProxyID;
    encoder << instance.webPageID;
    encoder << instance.webFrameID;
    encoder << instance.topOrigin;
    encoder << instance.sourceOrigin;
    encoder << instance.parentPID;
    encoder << instance.request;
    encoder << instance.contentSniffingPolicy;
    encoder << instance.contentEncodingSniffingPolicy;
    encoder << instance.storedCredentialsPolicy;
    encoder << instance.clientCredentialPolicy;
    encoder << instance.shouldClearReferrerOnHTTPSToHTTPRedirect;
    encoder << instance.needsCertificateInfo;
    encoder << instance.isMainFrameNavigation;
    encoder << instance.isMainResourceNavigationForAnyFrame;
    encoder << instance.shouldRelaxThirdPartyCookieBlocking;
    encoder << instance.shouldPreconnectOnly;
    encoder << instance.isNavigatingToAppBoundDomain;
    encoder << instance.hadMainFrameMainResourcePrivateRelayed;
    encoder << instance.allowPrivacyProxy;
    encoder << instance.advancedPrivacyProtections;
}

std::optional<WebKit::NetworkLoadParameters> ArgumentCoder<WebKit::NetworkLoadParameters>::decode(Decoder& decoder)
{
    auto webPageProxyID = decoder.decode<WebKit::WebPageProxyIdentifier>();
    auto webPageID = decoder.decode<WebCore::PageIdentifier>();
    auto webFrameID = decoder.decode<WebCore::FrameIdentifier>();
    auto topOrigin = decoder.decode<RefPtr<WebCore::SecurityOrigin>>();
    auto sourceOrigin = decoder.decode<RefPtr<WebCore::SecurityOrigin>>();
    auto parentPID = decoder.decode<WTF::ProcessID>();
    auto request = decoder.decode<WebCore::ResourceRequest>();
    auto contentSniffingPolicy = decoder.decode<WebCore::ContentSniffingPolicy>();
    auto contentEncodingSniffingPolicy = decoder.decode<WebCore::ContentEncodingSniffingPolicy>();
    auto storedCredentialsPolicy = decoder.decode<WebCore::StoredCredentialsPolicy>();
    auto clientCredentialPolicy = decoder.decode<WebCore::ClientCredentialPolicy>();
    auto shouldClearReferrerOnHTTPSToHTTPRedirect = decoder.decode<bool>();
    auto needsCertificateInfo = decoder.decode<bool>();
    auto isMainFrameNavigation = decoder.decode<bool>();
    auto isMainResourceNavigationForAnyFrame = decoder.decode<bool>();
    auto shouldRelaxThirdPartyCookieBlocking = decoder.decode<WebCore::ShouldRelaxThirdPartyCookieBlocking>();
    auto shouldPreconnectOnly = decoder.decode<WebKit::PreconnectOnly>();
    auto isNavigatingToAppBoundDomain = decoder.decode<std::optional<WebKit::NavigatingToAppBoundDomain>>();
    auto hadMainFrameMainResourcePrivateRelayed = decoder.decode<bool>();
    auto allowPrivacyProxy = decoder.decode<bool>();
    auto advancedPrivacyProtections = decoder.decode<OptionSet<WebCore::AdvancedPrivacyProtections>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::NetworkLoadParameters {
            WTFMove(*webPageProxyID),
            WTFMove(*webPageID),
            WTFMove(*webFrameID),
            WTFMove(*topOrigin),
            WTFMove(*sourceOrigin),
            WTFMove(*parentPID),
            WTFMove(*request),
            WTFMove(*contentSniffingPolicy),
            WTFMove(*contentEncodingSniffingPolicy),
            WTFMove(*storedCredentialsPolicy),
            WTFMove(*clientCredentialPolicy),
            WTFMove(*shouldClearReferrerOnHTTPSToHTTPRedirect),
            WTFMove(*needsCertificateInfo),
            WTFMove(*isMainFrameNavigation),
            WTFMove(*isMainResourceNavigationForAnyFrame),
            WTFMove(*shouldRelaxThirdPartyCookieBlocking),
            WTFMove(*shouldPreconnectOnly),
            WTFMove(*isNavigatingToAppBoundDomain),
            WTFMove(*hadMainFrameMainResourcePrivateRelayed),
            WTFMove(*allowPrivacyProxy),
            WTFMove(*advancedPrivacyProtections)
        }
    };
}

void ArgumentCoder<WebKit::NetworkResourceLoadParameters>::encode(Encoder& encoder, const WebKit::NetworkResourceLoadParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webPageProxyID)>, WebKit::WebPageProxyIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webPageID)>, WebCore::PageIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webFrameID)>, WebCore::FrameIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topOrigin)>, RefPtr<WebCore::SecurityOrigin>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceOrigin)>, RefPtr<WebCore::SecurityOrigin>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentPID)>, WTF::ProcessID>);
#if HAVE(AUDIT_TOKEN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkProcessAuditToken)>, std::optional<audit_token_t>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.request)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSniffingPolicy)>, WebCore::ContentSniffingPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentEncodingSniffingPolicy)>, WebCore::ContentEncodingSniffingPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.storedCredentialsPolicy)>, WebCore::StoredCredentialsPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientCredentialPolicy)>, WebCore::ClientCredentialPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldClearReferrerOnHTTPSToHTTPRedirect)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.needsCertificateInfo)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMainFrameNavigation)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMainResourceNavigationForAnyFrame)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRelaxThirdPartyCookieBlocking)>, WebCore::ShouldRelaxThirdPartyCookieBlocking>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blobFileReferences)>, Vector<RefPtr<WebCore::BlobDataFileReference>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldPreconnectOnly)>, WebKit::PreconnectOnly>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkActivityTracker)>, std::optional<WebKit::NetworkActivityTracker>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isNavigatingToAppBoundDomain)>, std::optional<WebKit::NavigatingToAppBoundDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hadMainFrameMainResourcePrivateRelayed)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowPrivacyProxy)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.advancedPrivacyProtections)>, OptionSet<WebCore::AdvancedPrivacyProtections>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::NetworkLoadParameters, webPageProxyID)
        , offsetof(WebKit::NetworkLoadParameters, webPageID)
        , offsetof(WebKit::NetworkLoadParameters, webFrameID)
        , offsetof(WebKit::NetworkLoadParameters, topOrigin)
        , offsetof(WebKit::NetworkLoadParameters, sourceOrigin)
        , offsetof(WebKit::NetworkLoadParameters, parentPID)
#if HAVE(AUDIT_TOKEN)
        , offsetof(WebKit::NetworkLoadParameters, networkProcessAuditToken)
#endif
        , offsetof(WebKit::NetworkLoadParameters, request)
        , offsetof(WebKit::NetworkLoadParameters, contentSniffingPolicy)
        , offsetof(WebKit::NetworkLoadParameters, contentEncodingSniffingPolicy)
        , offsetof(WebKit::NetworkLoadParameters, storedCredentialsPolicy)
        , offsetof(WebKit::NetworkLoadParameters, clientCredentialPolicy)
        , offsetof(WebKit::NetworkLoadParameters, shouldClearReferrerOnHTTPSToHTTPRedirect)
        , offsetof(WebKit::NetworkLoadParameters, needsCertificateInfo)
        , offsetof(WebKit::NetworkLoadParameters, isMainFrameNavigation)
        , offsetof(WebKit::NetworkLoadParameters, isMainResourceNavigationForAnyFrame)
        , offsetof(WebKit::NetworkLoadParameters, shouldRelaxThirdPartyCookieBlocking)
        , offsetof(WebKit::NetworkLoadParameters, blobFileReferences)
        , offsetof(WebKit::NetworkLoadParameters, shouldPreconnectOnly)
        , offsetof(WebKit::NetworkLoadParameters, networkActivityTracker)
        , offsetof(WebKit::NetworkLoadParameters, isNavigatingToAppBoundDomain)
        , offsetof(WebKit::NetworkLoadParameters, hadMainFrameMainResourcePrivateRelayed)
        , offsetof(WebKit::NetworkLoadParameters, allowPrivacyProxy)
        , offsetof(WebKit::NetworkLoadParameters, advancedPrivacyProtections)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::ResourceLoaderIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.request.httpBody())>, RefPtr<WebCore::FormData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sandboxExtensionsIfHttpBody())>, std::optional<Vector<WebKit::SandboxExtension::Handle>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sandboxExtensionIflocalFile())>, std::optional<WebKit::SandboxExtension::Handle>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumBufferingTime)>, Seconds>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.options)>, WebCore::FetchOptions>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cspResponseHeaders)>, std::optional<WebCore::ContentSecurityPolicyResponseHeaders>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentFrameURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentCrossOriginEmbedderPolicy)>, WebCore::CrossOriginEmbedderPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginEmbedderPolicy)>, WebCore::CrossOriginEmbedderPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originalRequestHeaders)>, WebCore::HTTPHeaderMap>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRestrictHTTPResponseAccess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preflightPolicy)>, WebCore::PreflightPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldEnableCrossOriginResourcePolicy)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameAncestorOrigins)>, Vector<RefPtr<WebCore::SecurityOrigin>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageHasResourceLoadClient)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentFrameID)>, std::optional<WebCore::FrameIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginAccessControlCheckEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isCrossOriginOpenerPolicyEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isClearSiteDataHeaderEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isDisplayingInitialEmptyDocument)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.effectiveSandboxFlags)>, WebCore::SandboxFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.openerURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceCrossOriginOpenerPolicy)>, WebCore::CrossOriginOpenerPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.navigationID)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.navigationRequester)>, std::optional<WebCore::NavigationRequester>>);
#if ENABLE(SERVICE_WORKER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serviceWorkersMode)>, WebCore::ServiceWorkersMode>);
#endif
#if ENABLE(SERVICE_WORKER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serviceWorkerRegistrationIdentifier)>, std::optional<WebCore::ServiceWorkerRegistrationIdentifier>>);
#endif
#if ENABLE(SERVICE_WORKER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpHeadersToKeep)>, OptionSet<WebCore::HTTPHeadersToKeepFromCleaning>>);
#endif
#if ENABLE(SERVICE_WORKER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.navigationPreloadIdentifier)>, std::optional<WebCore::FetchIdentifier>>);
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mainDocumentURL)>, URL>);
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userContentControllerIdentifier)>, std::optional<WebKit::UserContentControllerIdentifier>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linkPreconnectEarlyHintsEnabled)>, bool>);
    encoder << instance.webPageProxyID;
    encoder << instance.webPageID;
    encoder << instance.webFrameID;
    encoder << instance.topOrigin;
    encoder << instance.sourceOrigin;
    encoder << instance.parentPID;
    encoder << instance.request;
    encoder << instance.contentSniffingPolicy;
    encoder << instance.contentEncodingSniffingPolicy;
    encoder << instance.storedCredentialsPolicy;
    encoder << instance.clientCredentialPolicy;
    encoder << instance.shouldClearReferrerOnHTTPSToHTTPRedirect;
    encoder << instance.needsCertificateInfo;
    encoder << instance.isMainFrameNavigation;
    encoder << instance.isMainResourceNavigationForAnyFrame;
    encoder << instance.shouldRelaxThirdPartyCookieBlocking;
    encoder << instance.shouldPreconnectOnly;
    encoder << instance.isNavigatingToAppBoundDomain;
    encoder << instance.hadMainFrameMainResourcePrivateRelayed;
    encoder << instance.allowPrivacyProxy;
    encoder << instance.advancedPrivacyProtections;
    encoder << instance.identifier;
    encoder << instance.request.httpBody();
    encoder << instance.sandboxExtensionsIfHttpBody();
    encoder << instance.sandboxExtensionIflocalFile();
    encoder << instance.maximumBufferingTime;
    encoder << instance.options;
    encoder << instance.cspResponseHeaders;
    encoder << instance.parentFrameURL;
    encoder << instance.frameURL;
    encoder << instance.parentCrossOriginEmbedderPolicy;
    encoder << instance.crossOriginEmbedderPolicy;
    encoder << instance.originalRequestHeaders;
    encoder << instance.shouldRestrictHTTPResponseAccess;
    encoder << instance.preflightPolicy;
    encoder << instance.shouldEnableCrossOriginResourcePolicy;
    encoder << instance.frameAncestorOrigins;
    encoder << instance.pageHasResourceLoadClient;
    encoder << instance.parentFrameID;
    encoder << instance.crossOriginAccessControlCheckEnabled;
    encoder << instance.documentURL;
    encoder << instance.isCrossOriginOpenerPolicyEnabled;
    encoder << instance.isClearSiteDataHeaderEnabled;
    encoder << instance.isDisplayingInitialEmptyDocument;
    encoder << instance.effectiveSandboxFlags;
    encoder << instance.openerURL;
    encoder << instance.sourceCrossOriginOpenerPolicy;
    encoder << instance.navigationID;
    encoder << instance.navigationRequester;
#if ENABLE(SERVICE_WORKER)
    encoder << instance.serviceWorkersMode;
#endif
#if ENABLE(SERVICE_WORKER)
    encoder << instance.serviceWorkerRegistrationIdentifier;
#endif
#if ENABLE(SERVICE_WORKER)
    encoder << instance.httpHeadersToKeep;
#endif
#if ENABLE(SERVICE_WORKER)
    encoder << instance.navigationPreloadIdentifier;
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    encoder << instance.mainDocumentURL;
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    encoder << instance.userContentControllerIdentifier;
#endif
    encoder << instance.linkPreconnectEarlyHintsEnabled;
}

std::optional<WebKit::NetworkResourceLoadParameters> ArgumentCoder<WebKit::NetworkResourceLoadParameters>::decode(Decoder& decoder)
{
    auto webPageProxyID = decoder.decode<WebKit::WebPageProxyIdentifier>();
    auto webPageID = decoder.decode<WebCore::PageIdentifier>();
    auto webFrameID = decoder.decode<WebCore::FrameIdentifier>();
    auto topOrigin = decoder.decode<RefPtr<WebCore::SecurityOrigin>>();
    auto sourceOrigin = decoder.decode<RefPtr<WebCore::SecurityOrigin>>();
    auto parentPID = decoder.decode<WTF::ProcessID>();
    auto request = decoder.decode<WebCore::ResourceRequest>();
    auto contentSniffingPolicy = decoder.decode<WebCore::ContentSniffingPolicy>();
    auto contentEncodingSniffingPolicy = decoder.decode<WebCore::ContentEncodingSniffingPolicy>();
    auto storedCredentialsPolicy = decoder.decode<WebCore::StoredCredentialsPolicy>();
    auto clientCredentialPolicy = decoder.decode<WebCore::ClientCredentialPolicy>();
    auto shouldClearReferrerOnHTTPSToHTTPRedirect = decoder.decode<bool>();
    auto needsCertificateInfo = decoder.decode<bool>();
    auto isMainFrameNavigation = decoder.decode<bool>();
    auto isMainResourceNavigationForAnyFrame = decoder.decode<bool>();
    auto shouldRelaxThirdPartyCookieBlocking = decoder.decode<WebCore::ShouldRelaxThirdPartyCookieBlocking>();
    auto shouldPreconnectOnly = decoder.decode<WebKit::PreconnectOnly>();
    auto isNavigatingToAppBoundDomain = decoder.decode<std::optional<WebKit::NavigatingToAppBoundDomain>>();
    auto hadMainFrameMainResourcePrivateRelayed = decoder.decode<bool>();
    auto allowPrivacyProxy = decoder.decode<bool>();
    auto advancedPrivacyProtections = decoder.decode<OptionSet<WebCore::AdvancedPrivacyProtections>>();
    auto identifier = decoder.decode<WebCore::ResourceLoaderIdentifier>();
    auto requesthttpBody = decoder.decode<RefPtr<WebCore::FormData>>();
    auto sandboxExtensionsIfHttpBody = decoder.decode<std::optional<Vector<WebKit::SandboxExtension::Handle>>>();
    auto sandboxExtensionIflocalFile = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
    auto maximumBufferingTime = decoder.decode<Seconds>();
    auto options = decoder.decode<WebCore::FetchOptions>();
    auto cspResponseHeaders = decoder.decode<std::optional<WebCore::ContentSecurityPolicyResponseHeaders>>();
    auto parentFrameURL = decoder.decode<URL>();
    auto frameURL = decoder.decode<URL>();
    auto parentCrossOriginEmbedderPolicy = decoder.decode<WebCore::CrossOriginEmbedderPolicy>();
    auto crossOriginEmbedderPolicy = decoder.decode<WebCore::CrossOriginEmbedderPolicy>();
    auto originalRequestHeaders = decoder.decode<WebCore::HTTPHeaderMap>();
    auto shouldRestrictHTTPResponseAccess = decoder.decode<bool>();
    auto preflightPolicy = decoder.decode<WebCore::PreflightPolicy>();
    auto shouldEnableCrossOriginResourcePolicy = decoder.decode<bool>();
    auto frameAncestorOrigins = decoder.decode<Vector<RefPtr<WebCore::SecurityOrigin>>>();
    auto pageHasResourceLoadClient = decoder.decode<bool>();
    auto parentFrameID = decoder.decode<std::optional<WebCore::FrameIdentifier>>();
    auto crossOriginAccessControlCheckEnabled = decoder.decode<bool>();
    auto documentURL = decoder.decode<URL>();
    auto isCrossOriginOpenerPolicyEnabled = decoder.decode<bool>();
    auto isClearSiteDataHeaderEnabled = decoder.decode<bool>();
    auto isDisplayingInitialEmptyDocument = decoder.decode<bool>();
    auto effectiveSandboxFlags = decoder.decode<WebCore::SandboxFlags>();
    auto openerURL = decoder.decode<URL>();
    auto sourceCrossOriginOpenerPolicy = decoder.decode<WebCore::CrossOriginOpenerPolicy>();
    auto navigationID = decoder.decode<uint64_t>();
    auto navigationRequester = decoder.decode<std::optional<WebCore::NavigationRequester>>();
#if ENABLE(SERVICE_WORKER)
    auto serviceWorkersMode = decoder.decode<WebCore::ServiceWorkersMode>();
#endif
#if ENABLE(SERVICE_WORKER)
    auto serviceWorkerRegistrationIdentifier = decoder.decode<std::optional<WebCore::ServiceWorkerRegistrationIdentifier>>();
#endif
#if ENABLE(SERVICE_WORKER)
    auto httpHeadersToKeep = decoder.decode<OptionSet<WebCore::HTTPHeadersToKeepFromCleaning>>();
#endif
#if ENABLE(SERVICE_WORKER)
    auto navigationPreloadIdentifier = decoder.decode<std::optional<WebCore::FetchIdentifier>>();
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    auto mainDocumentURL = decoder.decode<URL>();
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    auto userContentControllerIdentifier = decoder.decode<std::optional<WebKit::UserContentControllerIdentifier>>();
#endif
    auto linkPreconnectEarlyHintsEnabled = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::NetworkResourceLoadParameters {
            WebKit::NetworkLoadParameters {
                WTFMove(*webPageProxyID),
                WTFMove(*webPageID),
                WTFMove(*webFrameID),
                WTFMove(*topOrigin),
                WTFMove(*sourceOrigin),
                WTFMove(*parentPID),
                WTFMove(*request),
                WTFMove(*contentSniffingPolicy),
                WTFMove(*contentEncodingSniffingPolicy),
                WTFMove(*storedCredentialsPolicy),
                WTFMove(*clientCredentialPolicy),
                WTFMove(*shouldClearReferrerOnHTTPSToHTTPRedirect),
                WTFMove(*needsCertificateInfo),
                WTFMove(*isMainFrameNavigation),
                WTFMove(*isMainResourceNavigationForAnyFrame),
                WTFMove(*shouldRelaxThirdPartyCookieBlocking),
                WTFMove(*shouldPreconnectOnly),
                WTFMove(*isNavigatingToAppBoundDomain),
                WTFMove(*hadMainFrameMainResourcePrivateRelayed),
                WTFMove(*allowPrivacyProxy),
                WTFMove(*advancedPrivacyProtections)
            },
            WTFMove(*identifier),
            WTFMove(*requesthttpBody),
            WTFMove(*sandboxExtensionsIfHttpBody),
            WTFMove(*sandboxExtensionIflocalFile),
            WTFMove(*maximumBufferingTime),
            WTFMove(*options),
            WTFMove(*cspResponseHeaders),
            WTFMove(*parentFrameURL),
            WTFMove(*frameURL),
            WTFMove(*parentCrossOriginEmbedderPolicy),
            WTFMove(*crossOriginEmbedderPolicy),
            WTFMove(*originalRequestHeaders),
            WTFMove(*shouldRestrictHTTPResponseAccess),
            WTFMove(*preflightPolicy),
            WTFMove(*shouldEnableCrossOriginResourcePolicy),
            WTFMove(*frameAncestorOrigins),
            WTFMove(*pageHasResourceLoadClient),
            WTFMove(*parentFrameID),
            WTFMove(*crossOriginAccessControlCheckEnabled),
            WTFMove(*documentURL),
            WTFMove(*isCrossOriginOpenerPolicyEnabled),
            WTFMove(*isClearSiteDataHeaderEnabled),
            WTFMove(*isDisplayingInitialEmptyDocument),
            WTFMove(*effectiveSandboxFlags),
            WTFMove(*openerURL),
            WTFMove(*sourceCrossOriginOpenerPolicy),
            WTFMove(*navigationID),
            WTFMove(*navigationRequester),
#if ENABLE(SERVICE_WORKER)
            WTFMove(*serviceWorkersMode),
#endif
#if ENABLE(SERVICE_WORKER)
            WTFMove(*serviceWorkerRegistrationIdentifier),
#endif
#if ENABLE(SERVICE_WORKER)
            WTFMove(*httpHeadersToKeep),
#endif
#if ENABLE(SERVICE_WORKER)
            WTFMove(*navigationPreloadIdentifier),
#endif
#if ENABLE(CONTENT_EXTENSIONS)
            WTFMove(*mainDocumentURL),
#endif
#if ENABLE(CONTENT_EXTENSIONS)
            WTFMove(*userContentControllerIdentifier),
#endif
            WTFMove(*linkPreconnectEarlyHintsEnabled)
        }
    };
}

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebKit::BackgroundFetchState>::encode(Encoder& encoder, const WebKit::BackgroundFetchState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topOrigin)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scope)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.options)>, WebCore::BackgroundFetchOptions>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.downloadTotal)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.downloaded)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uploadTotal)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uploaded)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.result)>, WebCore::BackgroundFetchResult>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failureReason)>, WebCore::BackgroundFetchFailureReason>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isPaused)>, bool>);
    struct ShouldBeSameSizeAsBackgroundFetchState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::BackgroundFetchState>, false> {
        WebCore::SecurityOriginData topOrigin;
        URL scope;
        String identifier;
        WebCore::BackgroundFetchOptions options;
        uint64_t downloadTotal;
        uint64_t downloaded;
        uint64_t uploadTotal;
        uint64_t uploaded;
        WebCore::BackgroundFetchResult result;
        WebCore::BackgroundFetchFailureReason failureReason;
        bool isPaused;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBackgroundFetchState) == sizeof(WebKit::BackgroundFetchState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::BackgroundFetchState, topOrigin)
        , offsetof(WebKit::BackgroundFetchState, scope)
        , offsetof(WebKit::BackgroundFetchState, identifier)
        , offsetof(WebKit::BackgroundFetchState, options)
        , offsetof(WebKit::BackgroundFetchState, downloadTotal)
        , offsetof(WebKit::BackgroundFetchState, downloaded)
        , offsetof(WebKit::BackgroundFetchState, uploadTotal)
        , offsetof(WebKit::BackgroundFetchState, uploaded)
        , offsetof(WebKit::BackgroundFetchState, result)
        , offsetof(WebKit::BackgroundFetchState, failureReason)
        , offsetof(WebKit::BackgroundFetchState, isPaused)
    >::value);
    encoder << instance.topOrigin;
    encoder << instance.scope;
    encoder << instance.identifier;
    encoder << instance.options;
    encoder << instance.downloadTotal;
    encoder << instance.downloaded;
    encoder << instance.uploadTotal;
    encoder << instance.uploaded;
    encoder << instance.result;
    encoder << instance.failureReason;
    encoder << instance.isPaused;
}

std::optional<WebKit::BackgroundFetchState> ArgumentCoder<WebKit::BackgroundFetchState>::decode(Decoder& decoder)
{
    auto topOrigin = decoder.decode<WebCore::SecurityOriginData>();
    auto scope = decoder.decode<URL>();
    auto identifier = decoder.decode<String>();
    auto options = decoder.decode<WebCore::BackgroundFetchOptions>();
    auto downloadTotal = decoder.decode<uint64_t>();
    auto downloaded = decoder.decode<uint64_t>();
    auto uploadTotal = decoder.decode<uint64_t>();
    auto uploaded = decoder.decode<uint64_t>();
    auto result = decoder.decode<WebCore::BackgroundFetchResult>();
    auto failureReason = decoder.decode<WebCore::BackgroundFetchFailureReason>();
    auto isPaused = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::BackgroundFetchState {
            WTFMove(*topOrigin),
            WTFMove(*scope),
            WTFMove(*identifier),
            WTFMove(*options),
            WTFMove(*downloadTotal),
            WTFMove(*downloaded),
            WTFMove(*uploadTotal),
            WTFMove(*uploaded),
            WTFMove(*result),
            WTFMove(*failureReason),
            WTFMove(*isPaused)
        }
    };
}

#endif

void ArgumentCoder<WebKit::CallbackID>::encode(Encoder& encoder, const WebKit::CallbackID& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_id)>, uint64_t>);
    struct ShouldBeSameSizeAsCallbackID : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::CallbackID>, false> {
        uint64_t m_id;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCallbackID) == sizeof(WebKit::CallbackID));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::CallbackID, m_id)
    >::value);
    encoder << instance.m_id;
}

std::optional<WebKit::CallbackID> ArgumentCoder<WebKit::CallbackID>::decode(Decoder& decoder)
{
    auto m_id = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(WebKit::CallbackID::isValidCallbackID(*m_id)))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::CallbackID {
            WTFMove(*m_id)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::SetState>::encode(Encoder& encoder, const WebCore::DisplayList::SetState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.state())>, WebCore::GraphicsContextState>);
    encoder << instance.state();
}

void ArgumentCoder<WebCore::DisplayList::SetState>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::SetState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.state())>, WebCore::GraphicsContextState>);
    encoder << instance.state();
}

std::optional<WebCore::DisplayList::SetState> ArgumentCoder<WebCore::DisplayList::SetState>::decode(Decoder& decoder)
{
    auto state = decoder.decode<WebCore::GraphicsContextState>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::SetState {
            WTFMove(*state)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::SetLineDash>::encode(Encoder& encoder, const WebCore::DisplayList::SetLineDash& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dashArray())>, WebCore::DashArray>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dashOffset())>, float>);
    encoder << instance.dashArray();
    encoder << instance.dashOffset();
}

void ArgumentCoder<WebCore::DisplayList::SetLineDash>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::SetLineDash& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dashArray())>, WebCore::DashArray>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dashOffset())>, float>);
    encoder << instance.dashArray();
    encoder << instance.dashOffset();
}

std::optional<WebCore::DisplayList::SetLineDash> ArgumentCoder<WebCore::DisplayList::SetLineDash>::decode(Decoder& decoder)
{
    auto dashArray = decoder.decode<WebCore::DashArray>();
    auto dashOffset = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::SetLineDash {
            WTFMove(*dashArray),
            WTFMove(*dashOffset)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::ClipOutToPath>::encode(Encoder& encoder, const WebCore::DisplayList::ClipOutToPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    encoder << instance.path();
}

void ArgumentCoder<WebCore::DisplayList::ClipOutToPath>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::ClipOutToPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    encoder << instance.path();
}

std::optional<WebCore::DisplayList::ClipOutToPath> ArgumentCoder<WebCore::DisplayList::ClipOutToPath>::decode(Decoder& decoder)
{
    auto path = decoder.decode<WebCore::Path>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::ClipOutToPath {
            WTFMove(*path)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::ClipPath>::encode(Encoder& encoder, const WebCore::DisplayList::ClipPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.windRule())>, WebCore::WindRule>);
    encoder << instance.path();
    encoder << instance.windRule();
}

void ArgumentCoder<WebCore::DisplayList::ClipPath>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::ClipPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.windRule())>, WebCore::WindRule>);
    encoder << instance.path();
    encoder << instance.windRule();
}

std::optional<WebCore::DisplayList::ClipPath> ArgumentCoder<WebCore::DisplayList::ClipPath>::decode(Decoder& decoder)
{
    auto path = decoder.decode<WebCore::Path>();
    auto windRule = decoder.decode<WebCore::WindRule>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::ClipPath {
            WTFMove(*path),
            WTFMove(*windRule)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::DrawGlyphs>::encode(Encoder& encoder, const WebCore::DisplayList::DrawGlyphs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontIdentifier())>, WebCore::RenderingResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.positionedGlyphs())>, WebCore::PositionedGlyphs>);
    encoder << instance.fontIdentifier();
    encoder << instance.positionedGlyphs();
}

void ArgumentCoder<WebCore::DisplayList::DrawGlyphs>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::DrawGlyphs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontIdentifier())>, WebCore::RenderingResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.positionedGlyphs())>, WebCore::PositionedGlyphs>);
    encoder << instance.fontIdentifier();
    encoder << instance.positionedGlyphs();
}

std::optional<WebCore::DisplayList::DrawGlyphs> ArgumentCoder<WebCore::DisplayList::DrawGlyphs>::decode(Decoder& decoder)
{
    auto fontIdentifier = decoder.decode<WebCore::RenderingResourceIdentifier>();
    auto positionedGlyphs = decoder.decode<WebCore::PositionedGlyphs>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::DrawGlyphs {
            WTFMove(*fontIdentifier),
            WTFMove(*positionedGlyphs)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::DrawSystemImage>::encode(Encoder& encoder, const WebCore::DisplayList::DrawSystemImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.systemImage())>, Ref<WebCore::SystemImage>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationRect())>, WebCore::FloatRect>);
    encoder << instance.systemImage();
    encoder << instance.destinationRect();
}

void ArgumentCoder<WebCore::DisplayList::DrawSystemImage>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::DrawSystemImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.systemImage())>, Ref<WebCore::SystemImage>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationRect())>, WebCore::FloatRect>);
    encoder << instance.systemImage();
    encoder << instance.destinationRect();
}

std::optional<WebCore::DisplayList::DrawSystemImage> ArgumentCoder<WebCore::DisplayList::DrawSystemImage>::decode(Decoder& decoder)
{
    auto systemImage = decoder.decode<Ref<WebCore::SystemImage>>();
    auto destinationRect = decoder.decode<WebCore::FloatRect>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::DrawSystemImage {
            WTFMove(*systemImage),
            WTFMove(*destinationRect)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::DrawLinesForText>::encode(Encoder& encoder, const WebCore::DisplayList::DrawLinesForText& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockLocation())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localAnchor())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.widths())>, WebCore::DashArray>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.thickness())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isPrinting())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.doubleLines())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.style())>, WebCore::StrokeStyle>);
    encoder << instance.blockLocation();
    encoder << instance.localAnchor();
    encoder << instance.widths();
    encoder << instance.thickness();
    encoder << instance.isPrinting();
    encoder << instance.doubleLines();
    encoder << instance.style();
}

void ArgumentCoder<WebCore::DisplayList::DrawLinesForText>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::DrawLinesForText& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockLocation())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localAnchor())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.widths())>, WebCore::DashArray>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.thickness())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isPrinting())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.doubleLines())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.style())>, WebCore::StrokeStyle>);
    encoder << instance.blockLocation();
    encoder << instance.localAnchor();
    encoder << instance.widths();
    encoder << instance.thickness();
    encoder << instance.isPrinting();
    encoder << instance.doubleLines();
    encoder << instance.style();
}

std::optional<WebCore::DisplayList::DrawLinesForText> ArgumentCoder<WebCore::DisplayList::DrawLinesForText>::decode(Decoder& decoder)
{
    auto blockLocation = decoder.decode<WebCore::FloatPoint>();
    auto localAnchor = decoder.decode<WebCore::FloatSize>();
    auto widths = decoder.decode<WebCore::DashArray>();
    auto thickness = decoder.decode<float>();
    auto isPrinting = decoder.decode<bool>();
    auto doubleLines = decoder.decode<bool>();
    auto style = decoder.decode<WebCore::StrokeStyle>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::DrawLinesForText {
            WTFMove(*blockLocation),
            WTFMove(*localAnchor),
            WTFMove(*widths),
            WTFMove(*thickness),
            WTFMove(*isPrinting),
            WTFMove(*doubleLines),
            WTFMove(*style)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::DrawPath>::encode(Encoder& encoder, const WebCore::DisplayList::DrawPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    encoder << instance.path();
}

void ArgumentCoder<WebCore::DisplayList::DrawPath>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::DrawPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    encoder << instance.path();
}

std::optional<WebCore::DisplayList::DrawPath> ArgumentCoder<WebCore::DisplayList::DrawPath>::decode(Decoder& decoder)
{
    auto path = decoder.decode<WebCore::Path>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::DrawPath {
            WTFMove(*path)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::DrawFocusRingPath>::encode(Encoder& encoder, const WebCore::DisplayList::DrawFocusRingPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.outlineWidth())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    encoder << instance.path();
    encoder << instance.outlineWidth();
    encoder << instance.color();
}

void ArgumentCoder<WebCore::DisplayList::DrawFocusRingPath>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::DrawFocusRingPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.outlineWidth())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    encoder << instance.path();
    encoder << instance.outlineWidth();
    encoder << instance.color();
}

std::optional<WebCore::DisplayList::DrawFocusRingPath> ArgumentCoder<WebCore::DisplayList::DrawFocusRingPath>::decode(Decoder& decoder)
{
    auto path = decoder.decode<WebCore::Path>();
    auto outlineWidth = decoder.decode<float>();
    auto color = decoder.decode<WebCore::Color>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::DrawFocusRingPath {
            WTFMove(*path),
            WTFMove(*outlineWidth),
            WTFMove(*color)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::DrawFocusRingRects>::encode(Encoder& encoder, const WebCore::DisplayList::DrawFocusRingRects& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rects())>, Vector<WebCore::FloatRect>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.outlineOffset())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.outlineWidth())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    encoder << instance.rects();
    encoder << instance.outlineOffset();
    encoder << instance.outlineWidth();
    encoder << instance.color();
}

void ArgumentCoder<WebCore::DisplayList::DrawFocusRingRects>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::DrawFocusRingRects& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rects())>, Vector<WebCore::FloatRect>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.outlineOffset())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.outlineWidth())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    encoder << instance.rects();
    encoder << instance.outlineOffset();
    encoder << instance.outlineWidth();
    encoder << instance.color();
}

std::optional<WebCore::DisplayList::DrawFocusRingRects> ArgumentCoder<WebCore::DisplayList::DrawFocusRingRects>::decode(Decoder& decoder)
{
    auto rects = decoder.decode<Vector<WebCore::FloatRect>>();
    auto outlineOffset = decoder.decode<float>();
    auto outlineWidth = decoder.decode<float>();
    auto color = decoder.decode<WebCore::Color>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::DrawFocusRingRects {
            WTFMove(*rects),
            WTFMove(*outlineOffset),
            WTFMove(*outlineWidth),
            WTFMove(*color)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::FillRectWithColor>::encode(Encoder& encoder, const WebCore::DisplayList::FillRectWithColor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    encoder << instance.rect();
    encoder << instance.color();
}

void ArgumentCoder<WebCore::DisplayList::FillRectWithColor>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::FillRectWithColor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    encoder << instance.rect();
    encoder << instance.color();
}

std::optional<WebCore::DisplayList::FillRectWithColor> ArgumentCoder<WebCore::DisplayList::FillRectWithColor>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    auto color = decoder.decode<WebCore::Color>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::FillRectWithColor {
            WTFMove(*rect),
            WTFMove(*color)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::FillRectWithGradient>::encode(Encoder& encoder, const WebCore::DisplayList::FillRectWithGradient& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gradient())>, Ref<WebCore::Gradient>>);
    encoder << instance.rect();
    encoder << instance.gradient();
}

void ArgumentCoder<WebCore::DisplayList::FillRectWithGradient>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::FillRectWithGradient& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gradient())>, Ref<WebCore::Gradient>>);
    encoder << instance.rect();
    encoder << instance.gradient();
}

std::optional<WebCore::DisplayList::FillRectWithGradient> ArgumentCoder<WebCore::DisplayList::FillRectWithGradient>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    auto gradient = decoder.decode<Ref<WebCore::Gradient>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::FillRectWithGradient {
            WTFMove(*rect),
            WTFMove(*gradient)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::FillCompositedRect>::encode(Encoder& encoder, const WebCore::DisplayList::FillCompositedRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositeOperator())>, WebCore::CompositeOperator>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blendMode())>, WebCore::BlendMode>);
    encoder << instance.rect();
    encoder << instance.color();
    encoder << instance.compositeOperator();
    encoder << instance.blendMode();
}

void ArgumentCoder<WebCore::DisplayList::FillCompositedRect>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::FillCompositedRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositeOperator())>, WebCore::CompositeOperator>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blendMode())>, WebCore::BlendMode>);
    encoder << instance.rect();
    encoder << instance.color();
    encoder << instance.compositeOperator();
    encoder << instance.blendMode();
}

std::optional<WebCore::DisplayList::FillCompositedRect> ArgumentCoder<WebCore::DisplayList::FillCompositedRect>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    auto color = decoder.decode<WebCore::Color>();
    auto compositeOperator = decoder.decode<WebCore::CompositeOperator>();
    auto blendMode = decoder.decode<WebCore::BlendMode>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::FillCompositedRect {
            WTFMove(*rect),
            WTFMove(*color),
            WTFMove(*compositeOperator),
            WTFMove(*blendMode)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::FillRoundedRect>::encode(Encoder& encoder, const WebCore::DisplayList::FillRoundedRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.roundedRect())>, WebCore::FloatRoundedRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blendMode())>, WebCore::BlendMode>);
    encoder << instance.roundedRect();
    encoder << instance.color();
    encoder << instance.blendMode();
}

std::optional<WebCore::DisplayList::FillRoundedRect> ArgumentCoder<WebCore::DisplayList::FillRoundedRect>::decode(Decoder& decoder)
{
    auto roundedRect = decoder.decode<WebCore::FloatRoundedRect>();
    auto color = decoder.decode<WebCore::Color>();
    auto blendMode = decoder.decode<WebCore::BlendMode>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::FillRoundedRect {
            WTFMove(*roundedRect),
            WTFMove(*color),
            WTFMove(*blendMode)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::FillRectWithRoundedHole>::encode(Encoder& encoder, const WebCore::DisplayList::FillRectWithRoundedHole& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.roundedHoleRect())>, WebCore::FloatRoundedRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    encoder << instance.rect();
    encoder << instance.roundedHoleRect();
    encoder << instance.color();
}

std::optional<WebCore::DisplayList::FillRectWithRoundedHole> ArgumentCoder<WebCore::DisplayList::FillRectWithRoundedHole>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    auto roundedHoleRect = decoder.decode<WebCore::FloatRoundedRect>();
    auto color = decoder.decode<WebCore::Color>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::FillRectWithRoundedHole {
            WTFMove(*rect),
            WTFMove(*roundedHoleRect),
            WTFMove(*color)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::FillPath>::encode(Encoder& encoder, const WebCore::DisplayList::FillPath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    encoder << instance.path();
}

std::optional<WebCore::DisplayList::FillPath> ArgumentCoder<WebCore::DisplayList::FillPath>::decode(Decoder& decoder)
{
    auto path = decoder.decode<WebCore::Path>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::FillPath {
            WTFMove(*path)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::StrokePath>::encode(Encoder& encoder, const WebCore::DisplayList::StrokePath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    encoder << instance.path();
}

std::optional<WebCore::DisplayList::StrokePath> ArgumentCoder<WebCore::DisplayList::StrokePath>::decode(Decoder& decoder)
{
    auto path = decoder.decode<WebCore::Path>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::StrokePath {
            WTFMove(*path)
        }
    };
}

void ArgumentCoder<WebKit::EditorState>::encode(Encoder& encoder, const WebKit::EditorState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebKit::EditorStateIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldIgnoreSelectionChanges)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionIsNone)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionIsRange)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionIsRangeInsideImageOverlay)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionIsRangeInAutoFilledAndViewableField)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isContentEditable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isContentRichlyEditable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isInPasswordField)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isInPlugin)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasComposition)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.triggeredByAccessibilitySelectionChange)>, bool>);
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canEnableAutomaticSpellingCorrection)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.postLayoutData)>, std::optional<WebKit::EditorState::PostLayoutData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.visualData)>, std::optional<WebKit::EditorState::VisualData>>);
    struct ShouldBeSameSizeAsEditorState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::EditorState>, false> {
        WebKit::EditorStateIdentifier identifier;
        bool shouldIgnoreSelectionChanges;
        bool selectionIsNone;
        bool selectionIsRange;
        bool selectionIsRangeInsideImageOverlay;
        bool selectionIsRangeInAutoFilledAndViewableField;
        bool isContentEditable;
        bool isContentRichlyEditable;
        bool isInPasswordField;
        bool isInPlugin;
        bool hasComposition;
        bool triggeredByAccessibilitySelectionChange;
#if PLATFORM(MAC)
        bool canEnableAutomaticSpellingCorrection;
#endif
        std::optional<WebKit::EditorState::PostLayoutData> postLayoutData;
        std::optional<WebKit::EditorState::VisualData> visualData;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEditorState) == sizeof(WebKit::EditorState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::EditorState, identifier)
        , offsetof(WebKit::EditorState, shouldIgnoreSelectionChanges)
        , offsetof(WebKit::EditorState, selectionIsNone)
        , offsetof(WebKit::EditorState, selectionIsRange)
        , offsetof(WebKit::EditorState, selectionIsRangeInsideImageOverlay)
        , offsetof(WebKit::EditorState, selectionIsRangeInAutoFilledAndViewableField)
        , offsetof(WebKit::EditorState, isContentEditable)
        , offsetof(WebKit::EditorState, isContentRichlyEditable)
        , offsetof(WebKit::EditorState, isInPasswordField)
        , offsetof(WebKit::EditorState, isInPlugin)
        , offsetof(WebKit::EditorState, hasComposition)
        , offsetof(WebKit::EditorState, triggeredByAccessibilitySelectionChange)
#if PLATFORM(MAC)
        , offsetof(WebKit::EditorState, canEnableAutomaticSpellingCorrection)
#endif
        , offsetof(WebKit::EditorState, postLayoutData)
        , offsetof(WebKit::EditorState, visualData)
    >::value);
    encoder << instance.identifier;
    encoder << instance.shouldIgnoreSelectionChanges;
    encoder << instance.selectionIsNone;
    encoder << instance.selectionIsRange;
    encoder << instance.selectionIsRangeInsideImageOverlay;
    encoder << instance.selectionIsRangeInAutoFilledAndViewableField;
    encoder << instance.isContentEditable;
    encoder << instance.isContentRichlyEditable;
    encoder << instance.isInPasswordField;
    encoder << instance.isInPlugin;
    encoder << instance.hasComposition;
    encoder << instance.triggeredByAccessibilitySelectionChange;
#if PLATFORM(MAC)
    encoder << instance.canEnableAutomaticSpellingCorrection;
#endif
    encoder << instance.postLayoutData;
    encoder << instance.visualData;
}

std::optional<WebKit::EditorState> ArgumentCoder<WebKit::EditorState>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebKit::EditorStateIdentifier>();
    auto shouldIgnoreSelectionChanges = decoder.decode<bool>();
    auto selectionIsNone = decoder.decode<bool>();
    auto selectionIsRange = decoder.decode<bool>();
    auto selectionIsRangeInsideImageOverlay = decoder.decode<bool>();
    auto selectionIsRangeInAutoFilledAndViewableField = decoder.decode<bool>();
    auto isContentEditable = decoder.decode<bool>();
    auto isContentRichlyEditable = decoder.decode<bool>();
    auto isInPasswordField = decoder.decode<bool>();
    auto isInPlugin = decoder.decode<bool>();
    auto hasComposition = decoder.decode<bool>();
    auto triggeredByAccessibilitySelectionChange = decoder.decode<bool>();
#if PLATFORM(MAC)
    auto canEnableAutomaticSpellingCorrection = decoder.decode<bool>();
#endif
    auto postLayoutData = decoder.decode<std::optional<WebKit::EditorState::PostLayoutData>>();
    auto visualData = decoder.decode<std::optional<WebKit::EditorState::VisualData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::EditorState {
            WTFMove(*identifier),
            WTFMove(*shouldIgnoreSelectionChanges),
            WTFMove(*selectionIsNone),
            WTFMove(*selectionIsRange),
            WTFMove(*selectionIsRangeInsideImageOverlay),
            WTFMove(*selectionIsRangeInAutoFilledAndViewableField),
            WTFMove(*isContentEditable),
            WTFMove(*isContentRichlyEditable),
            WTFMove(*isInPasswordField),
            WTFMove(*isInPlugin),
            WTFMove(*hasComposition),
            WTFMove(*triggeredByAccessibilitySelectionChange),
#if PLATFORM(MAC)
            WTFMove(*canEnableAutomaticSpellingCorrection),
#endif
            WTFMove(*postLayoutData),
            WTFMove(*visualData)
        }
    };
}

void ArgumentCoder<WebKit::EditorState::PostLayoutData>::encode(Encoder& encoder, const WebKit::EditorState::PostLayoutData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.typingAttributes)>, OptionSet<WebKit::TypingAttribute>>);
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectedTextLength)>, uint64_t>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textAlignment)>, WebKit::TextAlignment>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textColor)>, WebCore::Color>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enclosingListType)>, WebKit::ListType>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseWritingDirection)>, WebCore::WritingDirection>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.editableRootIsTransparentOrFullyClipped)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.markedText)>, String>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wordAtSelection)>, String>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.characterAfterSelection)>, UChar32>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.characterBeforeSelection)>, UChar32>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.twoCharacterBeforeSelection)>, UChar32>);
#endif
#if PLATFORM(IOS_FAMILY) && USE(DICTATION_ALTERNATIVES)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dictationContextsForSelection)>, Vector<WebCore::DictationContext>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isReplaceAllowed)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasContent)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isStableStateUpdate)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.insideFixedPosition)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasPlainText)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.caretColor)>, WebCore::Color>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.atStartOfSentence)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionStartIsAtParagraphBoundary)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionEndIsAtParagraphBoundary)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasGrammarDocumentMarkers)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectedEditableImage)>, std::optional<WebCore::ElementContext>>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionBoundingRect)>, WebCore::IntRect>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.candidateRequestStartPosition)>, uint64_t>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paragraphContextForCandidateRequest)>, String>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stringForCandidateRequest)>, String>);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.surroundingContext)>, String>);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.surroundingContextCursorPosition)>, uint64_t>);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.surroundingContextSelectionPosition)>, uint64_t>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontAttributes)>, std::optional<WebCore::FontAttributes>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canCut)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canCopy)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canPaste)>, bool>);
    struct ShouldBeSameSizeAsPostLayoutData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::EditorState::PostLayoutData>, false> {
        OptionSet<WebKit::TypingAttribute> typingAttributes;
#if PLATFORM(COCOA)
        uint64_t selectedTextLength;
#endif
#if PLATFORM(COCOA)
        WebKit::TextAlignment textAlignment;
#endif
#if PLATFORM(COCOA)
        WebCore::Color textColor;
#endif
#if PLATFORM(COCOA)
        WebKit::ListType enclosingListType;
#endif
#if PLATFORM(COCOA)
        WebCore::WritingDirection baseWritingDirection;
#endif
#if PLATFORM(COCOA)
        bool editableRootIsTransparentOrFullyClipped;
#endif
#if PLATFORM(IOS_FAMILY)
        String markedText;
#endif
#if PLATFORM(IOS_FAMILY)
        String wordAtSelection;
#endif
#if PLATFORM(IOS_FAMILY)
        UChar32 characterAfterSelection;
#endif
#if PLATFORM(IOS_FAMILY)
        UChar32 characterBeforeSelection;
#endif
#if PLATFORM(IOS_FAMILY)
        UChar32 twoCharacterBeforeSelection;
#endif
#if PLATFORM(IOS_FAMILY) && USE(DICTATION_ALTERNATIVES)
        Vector<WebCore::DictationContext> dictationContextsForSelection;
#endif
#if PLATFORM(IOS_FAMILY)
        bool isReplaceAllowed;
#endif
#if PLATFORM(IOS_FAMILY)
        bool hasContent;
#endif
#if PLATFORM(IOS_FAMILY)
        bool isStableStateUpdate;
#endif
#if PLATFORM(IOS_FAMILY)
        bool insideFixedPosition;
#endif
#if PLATFORM(IOS_FAMILY)
        bool hasPlainText;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::Color caretColor;
#endif
#if PLATFORM(IOS_FAMILY)
        bool atStartOfSentence;
#endif
#if PLATFORM(IOS_FAMILY)
        bool selectionStartIsAtParagraphBoundary;
#endif
#if PLATFORM(IOS_FAMILY)
        bool selectionEndIsAtParagraphBoundary;
#endif
#if PLATFORM(IOS_FAMILY)
        bool hasGrammarDocumentMarkers;
#endif
#if PLATFORM(IOS_FAMILY)
        std::optional<WebCore::ElementContext> selectedEditableImage;
#endif
#if PLATFORM(MAC)
        WebCore::IntRect selectionBoundingRect;
#endif
#if PLATFORM(MAC)
        uint64_t candidateRequestStartPosition;
#endif
#if PLATFORM(MAC)
        String paragraphContextForCandidateRequest;
#endif
#if PLATFORM(MAC)
        String stringForCandidateRequest;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        String surroundingContext;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        uint64_t surroundingContextCursorPosition;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        uint64_t surroundingContextSelectionPosition;
#endif
        std::optional<WebCore::FontAttributes> fontAttributes;
        bool canCut;
        bool canCopy;
        bool canPaste;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPostLayoutData) == sizeof(WebKit::EditorState::PostLayoutData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::EditorState::PostLayoutData, typingAttributes)
#if PLATFORM(COCOA)
        , offsetof(WebKit::EditorState::PostLayoutData, selectedTextLength)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::EditorState::PostLayoutData, textAlignment)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::EditorState::PostLayoutData, textColor)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::EditorState::PostLayoutData, enclosingListType)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::EditorState::PostLayoutData, baseWritingDirection)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::EditorState::PostLayoutData, editableRootIsTransparentOrFullyClipped)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, markedText)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, wordAtSelection)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, characterAfterSelection)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, characterBeforeSelection)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, twoCharacterBeforeSelection)
#endif
#if PLATFORM(IOS_FAMILY) && USE(DICTATION_ALTERNATIVES)
        , offsetof(WebKit::EditorState::PostLayoutData, dictationContextsForSelection)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, isReplaceAllowed)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, hasContent)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, isStableStateUpdate)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, insideFixedPosition)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, hasPlainText)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, caretColor)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, atStartOfSentence)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, selectionStartIsAtParagraphBoundary)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, selectionEndIsAtParagraphBoundary)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, hasGrammarDocumentMarkers)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::PostLayoutData, selectedEditableImage)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::EditorState::PostLayoutData, selectionBoundingRect)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::EditorState::PostLayoutData, candidateRequestStartPosition)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::EditorState::PostLayoutData, paragraphContextForCandidateRequest)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::EditorState::PostLayoutData, stringForCandidateRequest)
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        , offsetof(WebKit::EditorState::PostLayoutData, surroundingContext)
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        , offsetof(WebKit::EditorState::PostLayoutData, surroundingContextCursorPosition)
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        , offsetof(WebKit::EditorState::PostLayoutData, surroundingContextSelectionPosition)
#endif
        , offsetof(WebKit::EditorState::PostLayoutData, fontAttributes)
        , offsetof(WebKit::EditorState::PostLayoutData, canCut)
        , offsetof(WebKit::EditorState::PostLayoutData, canCopy)
        , offsetof(WebKit::EditorState::PostLayoutData, canPaste)
    >::value);
    encoder << instance.typingAttributes;
#if PLATFORM(COCOA)
    encoder << instance.selectedTextLength;
#endif
#if PLATFORM(COCOA)
    encoder << instance.textAlignment;
#endif
#if PLATFORM(COCOA)
    encoder << instance.textColor;
#endif
#if PLATFORM(COCOA)
    encoder << instance.enclosingListType;
#endif
#if PLATFORM(COCOA)
    encoder << instance.baseWritingDirection;
#endif
#if PLATFORM(COCOA)
    encoder << instance.editableRootIsTransparentOrFullyClipped;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.markedText;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.wordAtSelection;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.characterAfterSelection;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.characterBeforeSelection;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.twoCharacterBeforeSelection;
#endif
#if PLATFORM(IOS_FAMILY) && USE(DICTATION_ALTERNATIVES)
    encoder << instance.dictationContextsForSelection;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.isReplaceAllowed;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.hasContent;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.isStableStateUpdate;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.insideFixedPosition;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.hasPlainText;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.caretColor;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.atStartOfSentence;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.selectionStartIsAtParagraphBoundary;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.selectionEndIsAtParagraphBoundary;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.hasGrammarDocumentMarkers;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.selectedEditableImage;
#endif
#if PLATFORM(MAC)
    encoder << instance.selectionBoundingRect;
#endif
#if PLATFORM(MAC)
    encoder << instance.candidateRequestStartPosition;
#endif
#if PLATFORM(MAC)
    encoder << instance.paragraphContextForCandidateRequest;
#endif
#if PLATFORM(MAC)
    encoder << instance.stringForCandidateRequest;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    encoder << instance.surroundingContext;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    encoder << instance.surroundingContextCursorPosition;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    encoder << instance.surroundingContextSelectionPosition;
#endif
    encoder << instance.fontAttributes;
    encoder << instance.canCut;
    encoder << instance.canCopy;
    encoder << instance.canPaste;
}

std::optional<WebKit::EditorState::PostLayoutData> ArgumentCoder<WebKit::EditorState::PostLayoutData>::decode(Decoder& decoder)
{
    auto typingAttributes = decoder.decode<OptionSet<WebKit::TypingAttribute>>();
#if PLATFORM(COCOA)
    auto selectedTextLength = decoder.decode<uint64_t>();
#endif
#if PLATFORM(COCOA)
    auto textAlignment = decoder.decode<WebKit::TextAlignment>();
#endif
#if PLATFORM(COCOA)
    auto textColor = decoder.decode<WebCore::Color>();
#endif
#if PLATFORM(COCOA)
    auto enclosingListType = decoder.decode<WebKit::ListType>();
#endif
#if PLATFORM(COCOA)
    auto baseWritingDirection = decoder.decode<WebCore::WritingDirection>();
#endif
#if PLATFORM(COCOA)
    auto editableRootIsTransparentOrFullyClipped = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto markedText = decoder.decode<String>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto wordAtSelection = decoder.decode<String>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto characterAfterSelection = decoder.decode<UChar32>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto characterBeforeSelection = decoder.decode<UChar32>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto twoCharacterBeforeSelection = decoder.decode<UChar32>();
#endif
#if PLATFORM(IOS_FAMILY) && USE(DICTATION_ALTERNATIVES)
    auto dictationContextsForSelection = decoder.decode<Vector<WebCore::DictationContext>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto isReplaceAllowed = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto hasContent = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto isStableStateUpdate = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto insideFixedPosition = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto hasPlainText = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto caretColor = decoder.decode<WebCore::Color>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto atStartOfSentence = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto selectionStartIsAtParagraphBoundary = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto selectionEndIsAtParagraphBoundary = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto hasGrammarDocumentMarkers = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto selectedEditableImage = decoder.decode<std::optional<WebCore::ElementContext>>();
#endif
#if PLATFORM(MAC)
    auto selectionBoundingRect = decoder.decode<WebCore::IntRect>();
#endif
#if PLATFORM(MAC)
    auto candidateRequestStartPosition = decoder.decode<uint64_t>();
#endif
#if PLATFORM(MAC)
    auto paragraphContextForCandidateRequest = decoder.decode<String>();
#endif
#if PLATFORM(MAC)
    auto stringForCandidateRequest = decoder.decode<String>();
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    auto surroundingContext = decoder.decode<String>();
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    auto surroundingContextCursorPosition = decoder.decode<uint64_t>();
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    auto surroundingContextSelectionPosition = decoder.decode<uint64_t>();
#endif
    auto fontAttributes = decoder.decode<std::optional<WebCore::FontAttributes>>();
    auto canCut = decoder.decode<bool>();
    auto canCopy = decoder.decode<bool>();
    auto canPaste = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::EditorState::PostLayoutData {
            WTFMove(*typingAttributes),
#if PLATFORM(COCOA)
            WTFMove(*selectedTextLength),
#endif
#if PLATFORM(COCOA)
            WTFMove(*textAlignment),
#endif
#if PLATFORM(COCOA)
            WTFMove(*textColor),
#endif
#if PLATFORM(COCOA)
            WTFMove(*enclosingListType),
#endif
#if PLATFORM(COCOA)
            WTFMove(*baseWritingDirection),
#endif
#if PLATFORM(COCOA)
            WTFMove(*editableRootIsTransparentOrFullyClipped),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*markedText),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*wordAtSelection),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*characterAfterSelection),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*characterBeforeSelection),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*twoCharacterBeforeSelection),
#endif
#if PLATFORM(IOS_FAMILY) && USE(DICTATION_ALTERNATIVES)
            WTFMove(*dictationContextsForSelection),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*isReplaceAllowed),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*hasContent),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*isStableStateUpdate),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*insideFixedPosition),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*hasPlainText),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*caretColor),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*atStartOfSentence),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*selectionStartIsAtParagraphBoundary),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*selectionEndIsAtParagraphBoundary),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*hasGrammarDocumentMarkers),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*selectedEditableImage),
#endif
#if PLATFORM(MAC)
            WTFMove(*selectionBoundingRect),
#endif
#if PLATFORM(MAC)
            WTFMove(*candidateRequestStartPosition),
#endif
#if PLATFORM(MAC)
            WTFMove(*paragraphContextForCandidateRequest),
#endif
#if PLATFORM(MAC)
            WTFMove(*stringForCandidateRequest),
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
            WTFMove(*surroundingContext),
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
            WTFMove(*surroundingContextCursorPosition),
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
            WTFMove(*surroundingContextSelectionPosition),
#endif
            WTFMove(*fontAttributes),
            WTFMove(*canCut),
            WTFMove(*canCopy),
            WTFMove(*canPaste)
        }
    };
}

void ArgumentCoder<WebKit::EditorState::VisualData>::encode(Encoder& encoder, const WebKit::EditorState::VisualData& instance)
{
#if PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || PLATFORM(WPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.caretRectAtStart)>, WebCore::IntRect>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionClipRect)>, WebCore::IntRect>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.caretRectAtEnd)>, WebCore::IntRect>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionGeometries)>, Vector<WebCore::SelectionGeometry>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.markedTextRects)>, Vector<WebCore::SelectionGeometry>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.markedTextCaretRectAtStart)>, WebCore::IntRect>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.markedTextCaretRectAtEnd)>, WebCore::IntRect>);
#endif
    struct ShouldBeSameSizeAsVisualData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::EditorState::VisualData>, false> {
#if PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || PLATFORM(WPE)
        WebCore::IntRect caretRectAtStart;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::IntRect selectionClipRect;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::IntRect caretRectAtEnd;
#endif
#if PLATFORM(IOS_FAMILY)
        Vector<WebCore::SelectionGeometry> selectionGeometries;
#endif
#if PLATFORM(IOS_FAMILY)
        Vector<WebCore::SelectionGeometry> markedTextRects;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::IntRect markedTextCaretRectAtStart;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::IntRect markedTextCaretRectAtEnd;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsVisualData) == sizeof(WebKit::EditorState::VisualData));
    static_assert(MembersInCorrectOrder < 0
#if PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || PLATFORM(WPE)
        , offsetof(WebKit::EditorState::VisualData, caretRectAtStart)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::VisualData, selectionClipRect)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::VisualData, caretRectAtEnd)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::VisualData, selectionGeometries)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::VisualData, markedTextRects)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::VisualData, markedTextCaretRectAtStart)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::EditorState::VisualData, markedTextCaretRectAtEnd)
#endif
    >::value);
#if PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || PLATFORM(WPE)
    encoder << instance.caretRectAtStart;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.selectionClipRect;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.caretRectAtEnd;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.selectionGeometries;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.markedTextRects;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.markedTextCaretRectAtStart;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.markedTextCaretRectAtEnd;
#endif
}

std::optional<WebKit::EditorState::VisualData> ArgumentCoder<WebKit::EditorState::VisualData>::decode(Decoder& decoder)
{
#if PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || PLATFORM(WPE)
    auto caretRectAtStart = decoder.decode<WebCore::IntRect>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto selectionClipRect = decoder.decode<WebCore::IntRect>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto caretRectAtEnd = decoder.decode<WebCore::IntRect>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto selectionGeometries = decoder.decode<Vector<WebCore::SelectionGeometry>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto markedTextRects = decoder.decode<Vector<WebCore::SelectionGeometry>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto markedTextCaretRectAtStart = decoder.decode<WebCore::IntRect>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto markedTextCaretRectAtEnd = decoder.decode<WebCore::IntRect>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::EditorState::VisualData {
#if PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || PLATFORM(WPE)
            WTFMove(*caretRectAtStart),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*selectionClipRect),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*caretRectAtEnd),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*selectionGeometries),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*markedTextRects),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*markedTextCaretRectAtStart),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*markedTextCaretRectAtEnd)
#endif
        }
    };
}

void ArgumentCoder<WebKit::FileSystemSyncAccessHandleInfo>::encode(Encoder& encoder, const WebKit::FileSystemSyncAccessHandleInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::FileSystemSyncAccessHandleIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.handle)>, IPC::SharedFileHandle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.capacity)>, uint64_t>);
    struct ShouldBeSameSizeAsFileSystemSyncAccessHandleInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::FileSystemSyncAccessHandleInfo>, false> {
        WebCore::FileSystemSyncAccessHandleIdentifier identifier;
        IPC::SharedFileHandle handle;
        uint64_t capacity;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFileSystemSyncAccessHandleInfo) == sizeof(WebKit::FileSystemSyncAccessHandleInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::FileSystemSyncAccessHandleInfo, identifier)
        , offsetof(WebKit::FileSystemSyncAccessHandleInfo, handle)
        , offsetof(WebKit::FileSystemSyncAccessHandleInfo, capacity)
    >::value);
    encoder << instance.identifier;
    encoder << instance.handle;
    encoder << instance.capacity;
}

std::optional<WebKit::FileSystemSyncAccessHandleInfo> ArgumentCoder<WebKit::FileSystemSyncAccessHandleInfo>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::FileSystemSyncAccessHandleIdentifier>();
    auto handle = decoder.decode<IPC::SharedFileHandle>();
    auto capacity = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::FileSystemSyncAccessHandleInfo {
            WTFMove(*identifier),
            WTFMove(*handle),
            WTFMove(*capacity)
        }
    };
}

#if PLATFORM(IOS_FAMILY)
void ArgumentCoder<WebKit::OptionItem>::encode(Encoder& encoder, const WebKit::OptionItem& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isGroup)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSelected)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.disabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentGroupID)>, int>);
    struct ShouldBeSameSizeAsOptionItem : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::OptionItem>, false> {
        String text;
        bool isGroup;
        bool isSelected;
        bool disabled;
        int parentGroupID;
    };
    static_assert(sizeof(ShouldBeSameSizeAsOptionItem) == sizeof(WebKit::OptionItem));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::OptionItem, text)
        , offsetof(WebKit::OptionItem, isGroup)
        , offsetof(WebKit::OptionItem, isSelected)
        , offsetof(WebKit::OptionItem, disabled)
        , offsetof(WebKit::OptionItem, parentGroupID)
    >::value);
    encoder << instance.text;
    encoder << instance.isGroup;
    encoder << instance.isSelected;
    encoder << instance.disabled;
    encoder << instance.parentGroupID;
}

std::optional<WebKit::OptionItem> ArgumentCoder<WebKit::OptionItem>::decode(Decoder& decoder)
{
    auto text = decoder.decode<String>();
    auto isGroup = decoder.decode<bool>();
    auto isSelected = decoder.decode<bool>();
    auto disabled = decoder.decode<bool>();
    auto parentGroupID = decoder.decode<int>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::OptionItem {
            WTFMove(*text),
            WTFMove(*isGroup),
            WTFMove(*isSelected),
            WTFMove(*disabled),
            WTFMove(*parentGroupID)
        }
    };
}

#endif

#if PLATFORM(IOS_FAMILY)
void ArgumentCoder<WebKit::FocusedElementInformation>::encode(Encoder& encoder, const WebKit::FocusedElementInformation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.interactionRect)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elementContext)>, WebCore::ElementContext>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastInteractionLocation)>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimumScaleFactor)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumScaleFactor)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumScaleFactorIgnoringAlwaysScalable)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.nodeFontSize)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasNextNode)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.nextNodeRect)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasPreviousNode)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.previousNodeRect)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isAutocorrect)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isRTL)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMultiSelect)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isReadOnly)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsUserScaling)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsUserScalingIgnoringAlwaysScalable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.insideFixedPosition)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.autocapitalizeType)>, WebCore::AutocapitalizeType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elementType)>, WebKit::InputType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.inputMode)>, WebCore::InputMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enterKeyHint)>, WebCore::EnterKeyHint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.formAction)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectOptions)>, Vector<WebKit::OptionItem>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectedIndex)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.valueAsNumber)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.title)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.acceptsAutofilledLoginCredentials)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isAutofillableUsernameField)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.representingPageURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.autofillFieldName)>, WebCore::AutofillFieldName>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.nonAutofillCredentialType)>, WebCore::NonAutofillCredentialType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.placeholder)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ariaLabel)>, String>);
#if ENABLE(DATALIST_ELEMENT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasSuggestions)>, bool>);
#endif
#if ENABLE(DATALIST_ELEMENT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFocusingWithDataListDropdown)>, bool>);
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorValue)>, WebCore::Color>);
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.suggestedColors)>, Vector<WebCore::Color>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasEverBeenPasswordField)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldSynthesizeKeyEventsForEditing)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSpellCheckingEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldAvoidResizingWhenInputViewBoundsChange)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldAvoidScrollingWhenFocusedContentIsVisible)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldUseLegacySelectPopoverDismissalBehaviorInDataActivation)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFocusingWithValidationMessage)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preventScroll)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebKit::FocusedElementInformationIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containerScrollingNodeID)>, WebCore::ScrollingNodeID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, WebCore::FrameIdentifier>);
    struct ShouldBeSameSizeAsFocusedElementInformation : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::FocusedElementInformation>, false> {
        WebCore::IntRect interactionRect;
        WebCore::ElementContext elementContext;
        WebCore::IntPoint lastInteractionLocation;
        double minimumScaleFactor;
        double maximumScaleFactor;
        double maximumScaleFactorIgnoringAlwaysScalable;
        double nodeFontSize;
        bool hasNextNode;
        WebCore::IntRect nextNodeRect;
        bool hasPreviousNode;
        WebCore::IntRect previousNodeRect;
        bool isAutocorrect;
        bool isRTL;
        bool isMultiSelect;
        bool isReadOnly;
        bool allowsUserScaling;
        bool allowsUserScalingIgnoringAlwaysScalable;
        bool insideFixedPosition;
        WebCore::AutocapitalizeType autocapitalizeType;
        WebKit::InputType elementType;
        WebCore::InputMode inputMode;
        WebCore::EnterKeyHint enterKeyHint;
        String formAction;
        Vector<WebKit::OptionItem> selectOptions;
        int selectedIndex;
        String value;
        double valueAsNumber;
        String title;
        bool acceptsAutofilledLoginCredentials;
        bool isAutofillableUsernameField;
        URL representingPageURL;
        WebCore::AutofillFieldName autofillFieldName;
        WebCore::NonAutofillCredentialType nonAutofillCredentialType;
        String placeholder;
        String label;
        String ariaLabel;
#if ENABLE(DATALIST_ELEMENT)
        bool hasSuggestions;
#endif
#if ENABLE(DATALIST_ELEMENT)
        bool isFocusingWithDataListDropdown;
#endif
#if ENABLE(INPUT_TYPE_COLOR)
        WebCore::Color colorValue;
#endif
#if ENABLE(INPUT_TYPE_COLOR)
        Vector<WebCore::Color> suggestedColors;
#endif
        bool hasEverBeenPasswordField;
        bool shouldSynthesizeKeyEventsForEditing;
        bool isSpellCheckingEnabled;
        bool shouldAvoidResizingWhenInputViewBoundsChange;
        bool shouldAvoidScrollingWhenFocusedContentIsVisible;
        bool shouldUseLegacySelectPopoverDismissalBehaviorInDataActivation;
        bool isFocusingWithValidationMessage;
        bool preventScroll;
        WebKit::FocusedElementInformationIdentifier identifier;
        WebCore::ScrollingNodeID containerScrollingNodeID;
        WebCore::FrameIdentifier frameID;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFocusedElementInformation) == sizeof(WebKit::FocusedElementInformation));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::FocusedElementInformation, interactionRect)
        , offsetof(WebKit::FocusedElementInformation, elementContext)
        , offsetof(WebKit::FocusedElementInformation, lastInteractionLocation)
        , offsetof(WebKit::FocusedElementInformation, minimumScaleFactor)
        , offsetof(WebKit::FocusedElementInformation, maximumScaleFactor)
        , offsetof(WebKit::FocusedElementInformation, maximumScaleFactorIgnoringAlwaysScalable)
        , offsetof(WebKit::FocusedElementInformation, nodeFontSize)
        , offsetof(WebKit::FocusedElementInformation, hasNextNode)
        , offsetof(WebKit::FocusedElementInformation, nextNodeRect)
        , offsetof(WebKit::FocusedElementInformation, hasPreviousNode)
        , offsetof(WebKit::FocusedElementInformation, previousNodeRect)
        , offsetof(WebKit::FocusedElementInformation, isAutocorrect)
        , offsetof(WebKit::FocusedElementInformation, isRTL)
        , offsetof(WebKit::FocusedElementInformation, isMultiSelect)
        , offsetof(WebKit::FocusedElementInformation, isReadOnly)
        , offsetof(WebKit::FocusedElementInformation, allowsUserScaling)
        , offsetof(WebKit::FocusedElementInformation, allowsUserScalingIgnoringAlwaysScalable)
        , offsetof(WebKit::FocusedElementInformation, insideFixedPosition)
        , offsetof(WebKit::FocusedElementInformation, autocapitalizeType)
        , offsetof(WebKit::FocusedElementInformation, elementType)
        , offsetof(WebKit::FocusedElementInformation, inputMode)
        , offsetof(WebKit::FocusedElementInformation, enterKeyHint)
        , offsetof(WebKit::FocusedElementInformation, formAction)
        , offsetof(WebKit::FocusedElementInformation, selectOptions)
        , offsetof(WebKit::FocusedElementInformation, selectedIndex)
        , offsetof(WebKit::FocusedElementInformation, value)
        , offsetof(WebKit::FocusedElementInformation, valueAsNumber)
        , offsetof(WebKit::FocusedElementInformation, title)
        , offsetof(WebKit::FocusedElementInformation, acceptsAutofilledLoginCredentials)
        , offsetof(WebKit::FocusedElementInformation, isAutofillableUsernameField)
        , offsetof(WebKit::FocusedElementInformation, representingPageURL)
        , offsetof(WebKit::FocusedElementInformation, autofillFieldName)
        , offsetof(WebKit::FocusedElementInformation, nonAutofillCredentialType)
        , offsetof(WebKit::FocusedElementInformation, placeholder)
        , offsetof(WebKit::FocusedElementInformation, label)
        , offsetof(WebKit::FocusedElementInformation, ariaLabel)
#if ENABLE(DATALIST_ELEMENT)
        , offsetof(WebKit::FocusedElementInformation, hasSuggestions)
#endif
#if ENABLE(DATALIST_ELEMENT)
        , offsetof(WebKit::FocusedElementInformation, isFocusingWithDataListDropdown)
#endif
#if ENABLE(INPUT_TYPE_COLOR)
        , offsetof(WebKit::FocusedElementInformation, colorValue)
#endif
#if ENABLE(INPUT_TYPE_COLOR)
        , offsetof(WebKit::FocusedElementInformation, suggestedColors)
#endif
        , offsetof(WebKit::FocusedElementInformation, hasEverBeenPasswordField)
        , offsetof(WebKit::FocusedElementInformation, shouldSynthesizeKeyEventsForEditing)
        , offsetof(WebKit::FocusedElementInformation, isSpellCheckingEnabled)
        , offsetof(WebKit::FocusedElementInformation, shouldAvoidResizingWhenInputViewBoundsChange)
        , offsetof(WebKit::FocusedElementInformation, shouldAvoidScrollingWhenFocusedContentIsVisible)
        , offsetof(WebKit::FocusedElementInformation, shouldUseLegacySelectPopoverDismissalBehaviorInDataActivation)
        , offsetof(WebKit::FocusedElementInformation, isFocusingWithValidationMessage)
        , offsetof(WebKit::FocusedElementInformation, preventScroll)
        , offsetof(WebKit::FocusedElementInformation, identifier)
        , offsetof(WebKit::FocusedElementInformation, containerScrollingNodeID)
        , offsetof(WebKit::FocusedElementInformation, frameID)
    >::value);
    encoder << instance.interactionRect;
    encoder << instance.elementContext;
    encoder << instance.lastInteractionLocation;
    encoder << instance.minimumScaleFactor;
    encoder << instance.maximumScaleFactor;
    encoder << instance.maximumScaleFactorIgnoringAlwaysScalable;
    encoder << instance.nodeFontSize;
    encoder << instance.hasNextNode;
    encoder << instance.nextNodeRect;
    encoder << instance.hasPreviousNode;
    encoder << instance.previousNodeRect;
    encoder << instance.isAutocorrect;
    encoder << instance.isRTL;
    encoder << instance.isMultiSelect;
    encoder << instance.isReadOnly;
    encoder << instance.allowsUserScaling;
    encoder << instance.allowsUserScalingIgnoringAlwaysScalable;
    encoder << instance.insideFixedPosition;
    encoder << instance.autocapitalizeType;
    encoder << instance.elementType;
    encoder << instance.inputMode;
    encoder << instance.enterKeyHint;
    encoder << instance.formAction;
    encoder << instance.selectOptions;
    encoder << instance.selectedIndex;
    encoder << instance.value;
    encoder << instance.valueAsNumber;
    encoder << instance.title;
    encoder << instance.acceptsAutofilledLoginCredentials;
    encoder << instance.isAutofillableUsernameField;
    encoder << instance.representingPageURL;
    encoder << instance.autofillFieldName;
    encoder << instance.nonAutofillCredentialType;
    encoder << instance.placeholder;
    encoder << instance.label;
    encoder << instance.ariaLabel;
#if ENABLE(DATALIST_ELEMENT)
    encoder << instance.hasSuggestions;
#endif
#if ENABLE(DATALIST_ELEMENT)
    encoder << instance.isFocusingWithDataListDropdown;
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    encoder << instance.colorValue;
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    encoder << instance.suggestedColors;
#endif
    encoder << instance.hasEverBeenPasswordField;
    encoder << instance.shouldSynthesizeKeyEventsForEditing;
    encoder << instance.isSpellCheckingEnabled;
    encoder << instance.shouldAvoidResizingWhenInputViewBoundsChange;
    encoder << instance.shouldAvoidScrollingWhenFocusedContentIsVisible;
    encoder << instance.shouldUseLegacySelectPopoverDismissalBehaviorInDataActivation;
    encoder << instance.isFocusingWithValidationMessage;
    encoder << instance.preventScroll;
    encoder << instance.identifier;
    encoder << instance.containerScrollingNodeID;
    encoder << instance.frameID;
}

std::optional<WebKit::FocusedElementInformation> ArgumentCoder<WebKit::FocusedElementInformation>::decode(Decoder& decoder)
{
    auto interactionRect = decoder.decode<WebCore::IntRect>();
    auto elementContext = decoder.decode<WebCore::ElementContext>();
    auto lastInteractionLocation = decoder.decode<WebCore::IntPoint>();
    auto minimumScaleFactor = decoder.decode<double>();
    auto maximumScaleFactor = decoder.decode<double>();
    auto maximumScaleFactorIgnoringAlwaysScalable = decoder.decode<double>();
    auto nodeFontSize = decoder.decode<double>();
    auto hasNextNode = decoder.decode<bool>();
    auto nextNodeRect = decoder.decode<WebCore::IntRect>();
    auto hasPreviousNode = decoder.decode<bool>();
    auto previousNodeRect = decoder.decode<WebCore::IntRect>();
    auto isAutocorrect = decoder.decode<bool>();
    auto isRTL = decoder.decode<bool>();
    auto isMultiSelect = decoder.decode<bool>();
    auto isReadOnly = decoder.decode<bool>();
    auto allowsUserScaling = decoder.decode<bool>();
    auto allowsUserScalingIgnoringAlwaysScalable = decoder.decode<bool>();
    auto insideFixedPosition = decoder.decode<bool>();
    auto autocapitalizeType = decoder.decode<WebCore::AutocapitalizeType>();
    auto elementType = decoder.decode<WebKit::InputType>();
    auto inputMode = decoder.decode<WebCore::InputMode>();
    auto enterKeyHint = decoder.decode<WebCore::EnterKeyHint>();
    auto formAction = decoder.decode<String>();
    auto selectOptions = decoder.decode<Vector<WebKit::OptionItem>>();
    auto selectedIndex = decoder.decode<int>();
    auto value = decoder.decode<String>();
    auto valueAsNumber = decoder.decode<double>();
    auto title = decoder.decode<String>();
    auto acceptsAutofilledLoginCredentials = decoder.decode<bool>();
    auto isAutofillableUsernameField = decoder.decode<bool>();
    auto representingPageURL = decoder.decode<URL>();
    auto autofillFieldName = decoder.decode<WebCore::AutofillFieldName>();
    auto nonAutofillCredentialType = decoder.decode<WebCore::NonAutofillCredentialType>();
    auto placeholder = decoder.decode<String>();
    auto label = decoder.decode<String>();
    auto ariaLabel = decoder.decode<String>();
#if ENABLE(DATALIST_ELEMENT)
    auto hasSuggestions = decoder.decode<bool>();
#endif
#if ENABLE(DATALIST_ELEMENT)
    auto isFocusingWithDataListDropdown = decoder.decode<bool>();
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    auto colorValue = decoder.decode<WebCore::Color>();
#endif
#if ENABLE(INPUT_TYPE_COLOR)
    auto suggestedColors = decoder.decode<Vector<WebCore::Color>>();
#endif
    auto hasEverBeenPasswordField = decoder.decode<bool>();
    auto shouldSynthesizeKeyEventsForEditing = decoder.decode<bool>();
    auto isSpellCheckingEnabled = decoder.decode<bool>();
    auto shouldAvoidResizingWhenInputViewBoundsChange = decoder.decode<bool>();
    auto shouldAvoidScrollingWhenFocusedContentIsVisible = decoder.decode<bool>();
    auto shouldUseLegacySelectPopoverDismissalBehaviorInDataActivation = decoder.decode<bool>();
    auto isFocusingWithValidationMessage = decoder.decode<bool>();
    auto preventScroll = decoder.decode<bool>();
    auto identifier = decoder.decode<WebKit::FocusedElementInformationIdentifier>();
    auto containerScrollingNodeID = decoder.decode<WebCore::ScrollingNodeID>();
    auto frameID = decoder.decode<WebCore::FrameIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::FocusedElementInformation {
            WTFMove(*interactionRect),
            WTFMove(*elementContext),
            WTFMove(*lastInteractionLocation),
            WTFMove(*minimumScaleFactor),
            WTFMove(*maximumScaleFactor),
            WTFMove(*maximumScaleFactorIgnoringAlwaysScalable),
            WTFMove(*nodeFontSize),
            WTFMove(*hasNextNode),
            WTFMove(*nextNodeRect),
            WTFMove(*hasPreviousNode),
            WTFMove(*previousNodeRect),
            WTFMove(*isAutocorrect),
            WTFMove(*isRTL),
            WTFMove(*isMultiSelect),
            WTFMove(*isReadOnly),
            WTFMove(*allowsUserScaling),
            WTFMove(*allowsUserScalingIgnoringAlwaysScalable),
            WTFMove(*insideFixedPosition),
            WTFMove(*autocapitalizeType),
            WTFMove(*elementType),
            WTFMove(*inputMode),
            WTFMove(*enterKeyHint),
            WTFMove(*formAction),
            WTFMove(*selectOptions),
            WTFMove(*selectedIndex),
            WTFMove(*value),
            WTFMove(*valueAsNumber),
            WTFMove(*title),
            WTFMove(*acceptsAutofilledLoginCredentials),
            WTFMove(*isAutofillableUsernameField),
            WTFMove(*representingPageURL),
            WTFMove(*autofillFieldName),
            WTFMove(*nonAutofillCredentialType),
            WTFMove(*placeholder),
            WTFMove(*label),
            WTFMove(*ariaLabel),
#if ENABLE(DATALIST_ELEMENT)
            WTFMove(*hasSuggestions),
#endif
#if ENABLE(DATALIST_ELEMENT)
            WTFMove(*isFocusingWithDataListDropdown),
#endif
#if ENABLE(INPUT_TYPE_COLOR)
            WTFMove(*colorValue),
#endif
#if ENABLE(INPUT_TYPE_COLOR)
            WTFMove(*suggestedColors),
#endif
            WTFMove(*hasEverBeenPasswordField),
            WTFMove(*shouldSynthesizeKeyEventsForEditing),
            WTFMove(*isSpellCheckingEnabled),
            WTFMove(*shouldAvoidResizingWhenInputViewBoundsChange),
            WTFMove(*shouldAvoidScrollingWhenFocusedContentIsVisible),
            WTFMove(*shouldUseLegacySelectPopoverDismissalBehaviorInDataActivation),
            WTFMove(*isFocusingWithValidationMessage),
            WTFMove(*preventScroll),
            WTFMove(*identifier),
            WTFMove(*containerScrollingNodeID),
            WTFMove(*frameID)
        }
    };
}

#endif

void ArgumentCoder<WebKit::FrameInfoData>::encode(Encoder& encoder, const WebKit::FrameInfoData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMainFrame)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameType)>, WebKit::FrameType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.request)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.securityOrigin)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, WebCore::FrameIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentFrameID)>, std::optional<WebCore::FrameIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processID)>, WTF::ProcessID>);
    struct ShouldBeSameSizeAsFrameInfoData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::FrameInfoData>, false> {
        bool isMainFrame;
        WebKit::FrameType frameType;
        WebCore::ResourceRequest request;
        WebCore::SecurityOriginData securityOrigin;
        String frameName;
        WebCore::FrameIdentifier frameID;
        std::optional<WebCore::FrameIdentifier> parentFrameID;
        WTF::ProcessID processID;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFrameInfoData) == sizeof(WebKit::FrameInfoData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::FrameInfoData, isMainFrame)
        , offsetof(WebKit::FrameInfoData, frameType)
        , offsetof(WebKit::FrameInfoData, request)
        , offsetof(WebKit::FrameInfoData, securityOrigin)
        , offsetof(WebKit::FrameInfoData, frameName)
        , offsetof(WebKit::FrameInfoData, frameID)
        , offsetof(WebKit::FrameInfoData, parentFrameID)
        , offsetof(WebKit::FrameInfoData, processID)
    >::value);
    encoder << instance.isMainFrame;
    encoder << instance.frameType;
    encoder << instance.request;
    encoder << instance.securityOrigin;
    encoder << instance.frameName;
    encoder << instance.frameID;
    encoder << instance.parentFrameID;
    encoder << instance.processID;
}

std::optional<WebKit::FrameInfoData> ArgumentCoder<WebKit::FrameInfoData>::decode(Decoder& decoder)
{
    auto isMainFrame = decoder.decode<bool>();
    auto frameType = decoder.decode<WebKit::FrameType>();
    auto request = decoder.decode<WebCore::ResourceRequest>();
    auto securityOrigin = decoder.decode<WebCore::SecurityOriginData>();
    auto frameName = decoder.decode<String>();
    auto frameID = decoder.decode<WebCore::FrameIdentifier>();
    auto parentFrameID = decoder.decode<std::optional<WebCore::FrameIdentifier>>();
    auto processID = decoder.decode<WTF::ProcessID>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::FrameInfoData {
            WTFMove(*isMainFrame),
            WTFMove(*frameType),
            WTFMove(*request),
            WTFMove(*securityOrigin),
            WTFMove(*frameName),
            WTFMove(*frameID),
            WTFMove(*parentFrameID),
            WTFMove(*processID)
        }
    };
}

void ArgumentCoder<WebKit::FrameTreeCreationParameters>::encode(Encoder& encoder, const WebKit::FrameTreeCreationParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, WebCore::FrameIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.children)>, Vector<WebKit::FrameTreeCreationParameters>>);
    struct ShouldBeSameSizeAsFrameTreeCreationParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::FrameTreeCreationParameters>, false> {
        WebCore::FrameIdentifier frameID;
        Vector<WebKit::FrameTreeCreationParameters> children;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFrameTreeCreationParameters) == sizeof(WebKit::FrameTreeCreationParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::FrameTreeCreationParameters, frameID)
        , offsetof(WebKit::FrameTreeCreationParameters, children)
    >::value);
    encoder << instance.frameID;
    encoder << instance.children;
}

std::optional<WebKit::FrameTreeCreationParameters> ArgumentCoder<WebKit::FrameTreeCreationParameters>::decode(Decoder& decoder)
{
    auto frameID = decoder.decode<WebCore::FrameIdentifier>();
    auto children = decoder.decode<Vector<WebKit::FrameTreeCreationParameters>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::FrameTreeCreationParameters {
            WTFMove(*frameID),
            WTFMove(*children)
        }
    };
}

void ArgumentCoder<WebKit::FrameTreeNodeData>::encode(Encoder& encoder, const WebKit::FrameTreeNodeData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.info)>, WebKit::FrameInfoData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.children)>, Vector<WebKit::FrameTreeNodeData>>);
    struct ShouldBeSameSizeAsFrameTreeNodeData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::FrameTreeNodeData>, false> {
        WebKit::FrameInfoData info;
        Vector<WebKit::FrameTreeNodeData> children;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFrameTreeNodeData) == sizeof(WebKit::FrameTreeNodeData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::FrameTreeNodeData, info)
        , offsetof(WebKit::FrameTreeNodeData, children)
    >::value);
    encoder << instance.info;
    encoder << instance.children;
}

std::optional<WebKit::FrameTreeNodeData> ArgumentCoder<WebKit::FrameTreeNodeData>::decode(Decoder& decoder)
{
    auto info = decoder.decode<WebKit::FrameInfoData>();
    auto children = decoder.decode<Vector<WebKit::FrameTreeNodeData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::FrameTreeNodeData {
            WTFMove(*info),
            WTFMove(*children)
        }
    };
}

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::GPUProcessConnectionParameters>::encode(Encoder& encoder, WebKit::GPUProcessConnectionParameters&& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webProcessIdentity)>, WebCore::ProcessIdentity>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isDOMRenderingEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isLockdownModeEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isWebGPUEnabled)>, bool>);
#if ENABLE(IPC_TESTING_API)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoreInvalidMessageForTesting)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowTestOnlyIPC)>, bool>);
#if HAVE(AUDIT_TOKEN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentingApplicationAuditToken)>, std::optional<audit_token_t>>);
#endif
#if ENABLE(VP9)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasVP9HardwareDecoder)>, std::optional<bool>>);
#endif
#if ENABLE(VP9)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasVP9ExtensionSupport)>, std::optional<bool>>);
#endif
    struct ShouldBeSameSizeAsGPUProcessConnectionParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::GPUProcessConnectionParameters>, false> {
        WebCore::ProcessIdentity webProcessIdentity;
        bool isDOMRenderingEnabled;
        bool isLockdownModeEnabled;
        bool isWebGPUEnabled;
#if ENABLE(IPC_TESTING_API)
        bool ignoreInvalidMessageForTesting;
#endif
        bool allowTestOnlyIPC;
#if HAVE(AUDIT_TOKEN)
        std::optional<audit_token_t> presentingApplicationAuditToken;
#endif
#if ENABLE(VP9)
        std::optional<bool> hasVP9HardwareDecoder;
#endif
#if ENABLE(VP9)
        std::optional<bool> hasVP9ExtensionSupport;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsGPUProcessConnectionParameters) == sizeof(WebKit::GPUProcessConnectionParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::GPUProcessConnectionParameters, webProcessIdentity)
        , offsetof(WebKit::GPUProcessConnectionParameters, isDOMRenderingEnabled)
        , offsetof(WebKit::GPUProcessConnectionParameters, isLockdownModeEnabled)
        , offsetof(WebKit::GPUProcessConnectionParameters, isWebGPUEnabled)
#if ENABLE(IPC_TESTING_API)
        , offsetof(WebKit::GPUProcessConnectionParameters, ignoreInvalidMessageForTesting)
#endif
        , offsetof(WebKit::GPUProcessConnectionParameters, allowTestOnlyIPC)
#if HAVE(AUDIT_TOKEN)
        , offsetof(WebKit::GPUProcessConnectionParameters, presentingApplicationAuditToken)
#endif
#if ENABLE(VP9)
        , offsetof(WebKit::GPUProcessConnectionParameters, hasVP9HardwareDecoder)
#endif
#if ENABLE(VP9)
        , offsetof(WebKit::GPUProcessConnectionParameters, hasVP9ExtensionSupport)
#endif
    >::value);
    encoder << WTFMove(instance.webProcessIdentity);
    encoder << WTFMove(instance.isDOMRenderingEnabled);
    encoder << WTFMove(instance.isLockdownModeEnabled);
    encoder << WTFMove(instance.isWebGPUEnabled);
#if ENABLE(IPC_TESTING_API)
    encoder << WTFMove(instance.ignoreInvalidMessageForTesting);
#endif
    encoder << WTFMove(instance.allowTestOnlyIPC);
#if HAVE(AUDIT_TOKEN)
    encoder << WTFMove(instance.presentingApplicationAuditToken);
#endif
#if ENABLE(VP9)
    encoder << WTFMove(instance.hasVP9HardwareDecoder);
#endif
#if ENABLE(VP9)
    encoder << WTFMove(instance.hasVP9ExtensionSupport);
#endif
}

std::optional<WebKit::GPUProcessConnectionParameters> ArgumentCoder<WebKit::GPUProcessConnectionParameters>::decode(Decoder& decoder)
{
    auto webProcessIdentity = decoder.decode<WebCore::ProcessIdentity>();
    auto isDOMRenderingEnabled = decoder.decode<bool>();
    auto isLockdownModeEnabled = decoder.decode<bool>();
    auto isWebGPUEnabled = decoder.decode<bool>();
#if ENABLE(IPC_TESTING_API)
    auto ignoreInvalidMessageForTesting = decoder.decode<bool>();
#endif
    auto allowTestOnlyIPC = decoder.decode<bool>();
#if HAVE(AUDIT_TOKEN)
    auto presentingApplicationAuditToken = decoder.decode<std::optional<audit_token_t>>();
#endif
#if ENABLE(VP9)
    auto hasVP9HardwareDecoder = decoder.decode<std::optional<bool>>();
#endif
#if ENABLE(VP9)
    auto hasVP9ExtensionSupport = decoder.decode<std::optional<bool>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::GPUProcessConnectionParameters {
            WTFMove(*webProcessIdentity),
            WTFMove(*isDOMRenderingEnabled),
            WTFMove(*isLockdownModeEnabled),
            WTFMove(*isWebGPUEnabled),
#if ENABLE(IPC_TESTING_API)
            WTFMove(*ignoreInvalidMessageForTesting),
#endif
            WTFMove(*allowTestOnlyIPC),
#if HAVE(AUDIT_TOKEN)
            WTFMove(*presentingApplicationAuditToken),
#endif
#if ENABLE(VP9)
            WTFMove(*hasVP9HardwareDecoder),
#endif
#if ENABLE(VP9)
            WTFMove(*hasVP9ExtensionSupport)
#endif
        }
    };
}

#endif

void ArgumentCoder<WebKit::LayerTreeContext>::encode(Encoder& encoder, const WebKit::LayerTreeContext& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contextID)>, uint64_t>);
    struct ShouldBeSameSizeAsLayerTreeContext : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::LayerTreeContext>, false> {
        uint64_t contextID;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLayerTreeContext) == sizeof(WebKit::LayerTreeContext));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::LayerTreeContext, contextID)
    >::value);
    encoder << instance.contextID;
}

std::optional<WebKit::LayerTreeContext> ArgumentCoder<WebKit::LayerTreeContext>::decode(Decoder& decoder)
{
    auto contextID = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::LayerTreeContext {
            WTFMove(*contextID)
        }
    };
}

void ArgumentCoder<WebKit::LocalFrameCreationParameters>::encode(Encoder& encoder, const WebKit::LocalFrameCreationParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layerHostingContextIdentifier)>, std::optional<WebCore::LayerHostingContextIdentifier>>);
    struct ShouldBeSameSizeAsLocalFrameCreationParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::LocalFrameCreationParameters>, false> {
        std::optional<WebCore::LayerHostingContextIdentifier> layerHostingContextIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLocalFrameCreationParameters) == sizeof(WebKit::LocalFrameCreationParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::LocalFrameCreationParameters, layerHostingContextIdentifier)
    >::value);
    encoder << instance.layerHostingContextIdentifier;
}

std::optional<WebKit::LocalFrameCreationParameters> ArgumentCoder<WebKit::LocalFrameCreationParameters>::decode(Decoder& decoder)
{
    auto layerHostingContextIdentifier = decoder.decode<std::optional<WebCore::LayerHostingContextIdentifier>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::LocalFrameCreationParameters {
            WTFMove(*layerHostingContextIdentifier)
        }
    };
}

#if ENABLE(ARKIT_INLINE_PREVIEW)
void ArgumentCoder<WebKit::ModelIdentifier>::encode(Encoder& encoder, const WebKit::ModelIdentifier& instance)
{
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uuid)>, String>);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layerIdentifier)>, WebCore::PlatformLayerIdentifier>);
#endif
    struct ShouldBeSameSizeAsModelIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::ModelIdentifier>, false> {
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
        String uuid;
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
        WebCore::PlatformLayerIdentifier layerIdentifier;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsModelIdentifier) == sizeof(WebKit::ModelIdentifier));
    static_assert(MembersInCorrectOrder < 0
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
        , offsetof(WebKit::ModelIdentifier, uuid)
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
        , offsetof(WebKit::ModelIdentifier, layerIdentifier)
#endif
    >::value);
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
    encoder << instance.uuid;
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
    encoder << instance.layerIdentifier;
#endif
}

std::optional<WebKit::ModelIdentifier> ArgumentCoder<WebKit::ModelIdentifier>::decode(Decoder& decoder)
{
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
    auto uuid = decoder.decode<String>();
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
    auto layerIdentifier = decoder.decode<WebCore::PlatformLayerIdentifier>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::ModelIdentifier {
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
            WTFMove(*uuid),
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
            WTFMove(*layerIdentifier)
#endif
        }
    };
}

#endif

void ArgumentCoder<WebKit::NavigationActionData>::encode(Encoder& encoder, const WebKit::NavigationActionData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.navigationType)>, WebCore::NavigationType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiers)>, OptionSet<WebKit::WebEventModifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mouseButton)>, WebKit::WebMouseEventButton>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.syntheticClickType)>, WebKit::WebMouseEventSyntheticClickType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userGestureTokenIdentifier)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userGestureAuthorizationToken)>, std::optional<WTF::UUID>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canHandleRequest)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldOpenExternalURLsPolicy)>, WebCore::ShouldOpenExternalURLsPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.downloadAttribute)>, WTF::String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clickLocationInRootViewCoordinates)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.redirectResponse)>, WebCore::ResourceResponse>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.treatAsSameOriginNavigation)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasOpenedFrames)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.openedByDOMWithOpener)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasOpener)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requesterOrigin)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.targetBackForwardItemIdentifier)>, std::optional<WebCore::BackForwardItemIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceBackForwardItemIdentifier)>, std::optional<WebCore::BackForwardItemIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lockHistory)>, WebCore::LockHistory>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lockBackForwardList)>, WebCore::LockBackForwardList>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientRedirectSourceForHistory)>, WTF::String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.effectiveSandboxFlags)>, WebCore::SandboxFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.privateClickMeasurement)>, std::optional<WebCore::PrivateClickMeasurement>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.advancedPrivacyProtections)>, OptionSet<WebCore::AdvancedPrivacyProtections>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originatorAdvancedPrivacyProtections)>, OptionSet<WebCore::AdvancedPrivacyProtections>>);
#if PLATFORM(MAC) || HAVE(UIKIT_WITH_MOUSE_SUPPORT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webHitTestResultData)>, std::optional<WebKit::WebHitTestResultData>>);
#endif
    struct ShouldBeSameSizeAsNavigationActionData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::NavigationActionData>, false> {
        WebCore::NavigationType navigationType;
        OptionSet<WebKit::WebEventModifier> modifiers;
        WebKit::WebMouseEventButton mouseButton;
        WebKit::WebMouseEventSyntheticClickType syntheticClickType;
        uint64_t userGestureTokenIdentifier;
        std::optional<WTF::UUID> userGestureAuthorizationToken;
        bool canHandleRequest;
        WebCore::ShouldOpenExternalURLsPolicy shouldOpenExternalURLsPolicy;
        WTF::String downloadAttribute;
        WebCore::FloatPoint clickLocationInRootViewCoordinates;
        WebCore::ResourceResponse redirectResponse;
        bool treatAsSameOriginNavigation;
        bool hasOpenedFrames;
        bool openedByDOMWithOpener;
        bool hasOpener;
        WebCore::SecurityOriginData requesterOrigin;
        std::optional<WebCore::BackForwardItemIdentifier> targetBackForwardItemIdentifier;
        std::optional<WebCore::BackForwardItemIdentifier> sourceBackForwardItemIdentifier;
        WebCore::LockHistory lockHistory;
        WebCore::LockBackForwardList lockBackForwardList;
        WTF::String clientRedirectSourceForHistory;
        WebCore::SandboxFlags effectiveSandboxFlags;
        std::optional<WebCore::PrivateClickMeasurement> privateClickMeasurement;
        OptionSet<WebCore::AdvancedPrivacyProtections> advancedPrivacyProtections;
        OptionSet<WebCore::AdvancedPrivacyProtections> originatorAdvancedPrivacyProtections;
#if PLATFORM(MAC) || HAVE(UIKIT_WITH_MOUSE_SUPPORT)
        std::optional<WebKit::WebHitTestResultData> webHitTestResultData;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsNavigationActionData) == sizeof(WebKit::NavigationActionData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::NavigationActionData, navigationType)
        , offsetof(WebKit::NavigationActionData, modifiers)
        , offsetof(WebKit::NavigationActionData, mouseButton)
        , offsetof(WebKit::NavigationActionData, syntheticClickType)
        , offsetof(WebKit::NavigationActionData, userGestureTokenIdentifier)
        , offsetof(WebKit::NavigationActionData, userGestureAuthorizationToken)
        , offsetof(WebKit::NavigationActionData, canHandleRequest)
        , offsetof(WebKit::NavigationActionData, shouldOpenExternalURLsPolicy)
        , offsetof(WebKit::NavigationActionData, downloadAttribute)
        , offsetof(WebKit::NavigationActionData, clickLocationInRootViewCoordinates)
        , offsetof(WebKit::NavigationActionData, redirectResponse)
        , offsetof(WebKit::NavigationActionData, treatAsSameOriginNavigation)
        , offsetof(WebKit::NavigationActionData, hasOpenedFrames)
        , offsetof(WebKit::NavigationActionData, openedByDOMWithOpener)
        , offsetof(WebKit::NavigationActionData, hasOpener)
        , offsetof(WebKit::NavigationActionData, requesterOrigin)
        , offsetof(WebKit::NavigationActionData, targetBackForwardItemIdentifier)
        , offsetof(WebKit::NavigationActionData, sourceBackForwardItemIdentifier)
        , offsetof(WebKit::NavigationActionData, lockHistory)
        , offsetof(WebKit::NavigationActionData, lockBackForwardList)
        , offsetof(WebKit::NavigationActionData, clientRedirectSourceForHistory)
        , offsetof(WebKit::NavigationActionData, effectiveSandboxFlags)
        , offsetof(WebKit::NavigationActionData, privateClickMeasurement)
        , offsetof(WebKit::NavigationActionData, advancedPrivacyProtections)
        , offsetof(WebKit::NavigationActionData, originatorAdvancedPrivacyProtections)
#if PLATFORM(MAC) || HAVE(UIKIT_WITH_MOUSE_SUPPORT)
        , offsetof(WebKit::NavigationActionData, webHitTestResultData)
#endif
    >::value);
    encoder << instance.navigationType;
    encoder << instance.modifiers;
    encoder << instance.mouseButton;
    encoder << instance.syntheticClickType;
    encoder << instance.userGestureTokenIdentifier;
    encoder << instance.userGestureAuthorizationToken;
    encoder << instance.canHandleRequest;
    encoder << instance.shouldOpenExternalURLsPolicy;
    encoder << instance.downloadAttribute;
    encoder << instance.clickLocationInRootViewCoordinates;
    encoder << instance.redirectResponse;
    encoder << instance.treatAsSameOriginNavigation;
    encoder << instance.hasOpenedFrames;
    encoder << instance.openedByDOMWithOpener;
    encoder << instance.hasOpener;
    encoder << instance.requesterOrigin;
    encoder << instance.targetBackForwardItemIdentifier;
    encoder << instance.sourceBackForwardItemIdentifier;
    encoder << instance.lockHistory;
    encoder << instance.lockBackForwardList;
    encoder << instance.clientRedirectSourceForHistory;
    encoder << instance.effectiveSandboxFlags;
    encoder << instance.privateClickMeasurement;
    encoder << instance.advancedPrivacyProtections;
    encoder << instance.originatorAdvancedPrivacyProtections;
#if PLATFORM(MAC) || HAVE(UIKIT_WITH_MOUSE_SUPPORT)
    encoder << instance.webHitTestResultData;
#endif
}

std::optional<WebKit::NavigationActionData> ArgumentCoder<WebKit::NavigationActionData>::decode(Decoder& decoder)
{
    auto navigationType = decoder.decode<WebCore::NavigationType>();
    auto modifiers = decoder.decode<OptionSet<WebKit::WebEventModifier>>();
    auto mouseButton = decoder.decode<WebKit::WebMouseEventButton>();
    auto syntheticClickType = decoder.decode<WebKit::WebMouseEventSyntheticClickType>();
    auto userGestureTokenIdentifier = decoder.decode<uint64_t>();
    auto userGestureAuthorizationToken = decoder.decode<std::optional<WTF::UUID>>();
    auto canHandleRequest = decoder.decode<bool>();
    auto shouldOpenExternalURLsPolicy = decoder.decode<WebCore::ShouldOpenExternalURLsPolicy>();
    auto downloadAttribute = decoder.decode<WTF::String>();
    auto clickLocationInRootViewCoordinates = decoder.decode<WebCore::FloatPoint>();
    auto redirectResponse = decoder.decode<WebCore::ResourceResponse>();
    auto treatAsSameOriginNavigation = decoder.decode<bool>();
    auto hasOpenedFrames = decoder.decode<bool>();
    auto openedByDOMWithOpener = decoder.decode<bool>();
    auto hasOpener = decoder.decode<bool>();
    auto requesterOrigin = decoder.decode<WebCore::SecurityOriginData>();
    auto targetBackForwardItemIdentifier = decoder.decode<std::optional<WebCore::BackForwardItemIdentifier>>();
    auto sourceBackForwardItemIdentifier = decoder.decode<std::optional<WebCore::BackForwardItemIdentifier>>();
    auto lockHistory = decoder.decode<WebCore::LockHistory>();
    auto lockBackForwardList = decoder.decode<WebCore::LockBackForwardList>();
    auto clientRedirectSourceForHistory = decoder.decode<WTF::String>();
    auto effectiveSandboxFlags = decoder.decode<WebCore::SandboxFlags>();
    auto privateClickMeasurement = decoder.decode<std::optional<WebCore::PrivateClickMeasurement>>();
    auto advancedPrivacyProtections = decoder.decode<OptionSet<WebCore::AdvancedPrivacyProtections>>();
    auto originatorAdvancedPrivacyProtections = decoder.decode<OptionSet<WebCore::AdvancedPrivacyProtections>>();
#if PLATFORM(MAC) || HAVE(UIKIT_WITH_MOUSE_SUPPORT)
    auto webHitTestResultData = decoder.decode<std::optional<WebKit::WebHitTestResultData>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::NavigationActionData {
            WTFMove(*navigationType),
            WTFMove(*modifiers),
            WTFMove(*mouseButton),
            WTFMove(*syntheticClickType),
            WTFMove(*userGestureTokenIdentifier),
            WTFMove(*userGestureAuthorizationToken),
            WTFMove(*canHandleRequest),
            WTFMove(*shouldOpenExternalURLsPolicy),
            WTFMove(*downloadAttribute),
            WTFMove(*clickLocationInRootViewCoordinates),
            WTFMove(*redirectResponse),
            WTFMove(*treatAsSameOriginNavigation),
            WTFMove(*hasOpenedFrames),
            WTFMove(*openedByDOMWithOpener),
            WTFMove(*hasOpener),
            WTFMove(*requesterOrigin),
            WTFMove(*targetBackForwardItemIdentifier),
            WTFMove(*sourceBackForwardItemIdentifier),
            WTFMove(*lockHistory),
            WTFMove(*lockBackForwardList),
            WTFMove(*clientRedirectSourceForHistory),
            WTFMove(*effectiveSandboxFlags),
            WTFMove(*privateClickMeasurement),
            WTFMove(*advancedPrivacyProtections),
            WTFMove(*originatorAdvancedPrivacyProtections),
#if PLATFORM(MAC) || HAVE(UIKIT_WITH_MOUSE_SUPPORT)
            WTFMove(*webHitTestResultData)
#endif
        }
    };
}

void ArgumentCoder<WebKit::NetworkProcessConnectionParameters>::encode(Encoder& encoder, const WebKit::NetworkProcessConnectionParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowTestOnlyIPC)>, bool>);
    struct ShouldBeSameSizeAsNetworkProcessConnectionParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::NetworkProcessConnectionParameters>, false> {
        bool allowTestOnlyIPC;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNetworkProcessConnectionParameters) == sizeof(WebKit::NetworkProcessConnectionParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::NetworkProcessConnectionParameters, allowTestOnlyIPC)
    >::value);
    encoder << instance.allowTestOnlyIPC;
}

std::optional<WebKit::NetworkProcessConnectionParameters> ArgumentCoder<WebKit::NetworkProcessConnectionParameters>::decode(Decoder& decoder)
{
    auto allowTestOnlyIPC = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::NetworkProcessConnectionParameters {
            WTFMove(*allowTestOnlyIPC)
        }
    };
}

void ArgumentCoder<WebCore::PasteboardImage>::encode(Encoder& encoder, const WebCore::PasteboardImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.image)>, RefPtr<WebCore::Image>>);
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInWebArchiveFormat)>, RefPtr<WebCore::SharedBuffer>>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInHTMLFormat)>, String>);
#endif
#if !PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, WebCore::PasteboardURL>);
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resourceData)>, RefPtr<WebCore::SharedBuffer>>);
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resourceMIMEType)>, String>);
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientTypes)>, Vector<String>>);
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientData)>, Vector<RefPtr<WebCore::SharedBuffer>>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.suggestedName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.imageSize)>, WebCore::FloatSize>);
    struct ShouldBeSameSizeAsPasteboardImage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PasteboardImage>, false> {
        RefPtr<WebCore::Image> image;
#if PLATFORM(MAC)
        RefPtr<WebCore::SharedBuffer> dataInWebArchiveFormat;
#endif
#if PLATFORM(MAC)
        String dataInHTMLFormat;
#endif
#if !PLATFORM(WIN)
        WebCore::PasteboardURL url;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        RefPtr<WebCore::SharedBuffer> resourceData;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        String resourceMIMEType;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        Vector<String> clientTypes;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        Vector<RefPtr<WebCore::SharedBuffer>> clientData;
#endif
        String suggestedName;
        WebCore::FloatSize imageSize;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPasteboardImage) == sizeof(WebCore::PasteboardImage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PasteboardImage, image)
#if PLATFORM(MAC)
        , offsetof(WebCore::PasteboardImage, dataInWebArchiveFormat)
#endif
#if PLATFORM(MAC)
        , offsetof(WebCore::PasteboardImage, dataInHTMLFormat)
#endif
#if !PLATFORM(WIN)
        , offsetof(WebCore::PasteboardImage, url)
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        , offsetof(WebCore::PasteboardImage, resourceData)
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        , offsetof(WebCore::PasteboardImage, resourceMIMEType)
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        , offsetof(WebCore::PasteboardImage, clientTypes)
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
        , offsetof(WebCore::PasteboardImage, clientData)
#endif
        , offsetof(WebCore::PasteboardImage, suggestedName)
        , offsetof(WebCore::PasteboardImage, imageSize)
    >::value);
    encoder << instance.image;
#if PLATFORM(MAC)
    encoder << instance.dataInWebArchiveFormat;
#endif
#if PLATFORM(MAC)
    encoder << instance.dataInHTMLFormat;
#endif
#if !PLATFORM(WIN)
    encoder << instance.url;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    encoder << instance.resourceData;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    encoder << instance.resourceMIMEType;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    encoder << instance.clientTypes;
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    encoder << instance.clientData;
#endif
    encoder << instance.suggestedName;
    encoder << instance.imageSize;
}

std::optional<WebCore::PasteboardImage> ArgumentCoder<WebCore::PasteboardImage>::decode(Decoder& decoder)
{
    auto image = decoder.decode<RefPtr<WebCore::Image>>();
#if PLATFORM(MAC)
    auto dataInWebArchiveFormat = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
#endif
#if PLATFORM(MAC)
    auto dataInHTMLFormat = decoder.decode<String>();
#endif
#if !PLATFORM(WIN)
    auto url = decoder.decode<WebCore::PasteboardURL>();
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    auto resourceData = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    auto resourceMIMEType = decoder.decode<String>();
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    auto clientTypes = decoder.decode<Vector<String>>();
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
    auto clientData = decoder.decode<Vector<RefPtr<WebCore::SharedBuffer>>>();
#endif
    auto suggestedName = decoder.decode<String>();
    auto imageSize = decoder.decode<WebCore::FloatSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PasteboardImage {
            WTFMove(*image),
#if PLATFORM(MAC)
            WTFMove(*dataInWebArchiveFormat),
#endif
#if PLATFORM(MAC)
            WTFMove(*dataInHTMLFormat),
#endif
#if !PLATFORM(WIN)
            WTFMove(*url),
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
            WTFMove(*resourceData),
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
            WTFMove(*resourceMIMEType),
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
            WTFMove(*clientTypes),
#endif
#if !(PLATFORM(GTK) || PLATFORM(WIN))
            WTFMove(*clientData),
#endif
            WTFMove(*suggestedName),
            WTFMove(*imageSize)
        }
    };
}

void ArgumentCoder<WebCore::PasteboardWebContent>::encode(Encoder& encoder, const WebCore::PasteboardWebContent& instance)
{
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentOrigin)>, String>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canSmartCopyOrDelete)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInWebArchiveFormat)>, RefPtr<WebCore::SharedBuffer>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInRTFDFormat)>, RefPtr<WebCore::SharedBuffer>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInRTFFormat)>, RefPtr<WebCore::SharedBuffer>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInAttributedStringFormat)>, RefPtr<WebCore::SharedBuffer>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInHTMLFormat)>, String>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataInStringFormat)>, String>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientTypes)>, Vector<String>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientData)>, Vector<RefPtr<WebCore::SharedBuffer>>>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentOrigin)>, String>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canSmartCopyOrDelete)>, bool>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, String>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.markup)>, String>);
#endif
#if USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, String>);
#endif
#if USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.markup)>, String>);
#endif
    struct ShouldBeSameSizeAsPasteboardWebContent : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PasteboardWebContent>, false> {
#if PLATFORM(COCOA)
        String contentOrigin;
#endif
#if PLATFORM(COCOA)
        bool canSmartCopyOrDelete;
#endif
#if PLATFORM(COCOA)
        RefPtr<WebCore::SharedBuffer> dataInWebArchiveFormat;
#endif
#if PLATFORM(COCOA)
        RefPtr<WebCore::SharedBuffer> dataInRTFDFormat;
#endif
#if PLATFORM(COCOA)
        RefPtr<WebCore::SharedBuffer> dataInRTFFormat;
#endif
#if PLATFORM(COCOA)
        RefPtr<WebCore::SharedBuffer> dataInAttributedStringFormat;
#endif
#if PLATFORM(COCOA)
        String dataInHTMLFormat;
#endif
#if PLATFORM(COCOA)
        String dataInStringFormat;
#endif
#if PLATFORM(COCOA)
        Vector<String> clientTypes;
#endif
#if PLATFORM(COCOA)
        Vector<RefPtr<WebCore::SharedBuffer>> clientData;
#endif
#if PLATFORM(GTK)
        String contentOrigin;
#endif
#if PLATFORM(GTK)
        bool canSmartCopyOrDelete;
#endif
#if PLATFORM(GTK)
        String text;
#endif
#if PLATFORM(GTK)
        String markup;
#endif
#if USE(LIBWPE)
        String text;
#endif
#if USE(LIBWPE)
        String markup;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsPasteboardWebContent) == sizeof(WebCore::PasteboardWebContent));
    static_assert(MembersInCorrectOrder < 0
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, contentOrigin)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, canSmartCopyOrDelete)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, dataInWebArchiveFormat)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, dataInRTFDFormat)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, dataInRTFFormat)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, dataInAttributedStringFormat)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, dataInHTMLFormat)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, dataInStringFormat)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, clientTypes)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardWebContent, clientData)
#endif
#if PLATFORM(GTK)
        , offsetof(WebCore::PasteboardWebContent, contentOrigin)
#endif
#if PLATFORM(GTK)
        , offsetof(WebCore::PasteboardWebContent, canSmartCopyOrDelete)
#endif
#if PLATFORM(GTK)
        , offsetof(WebCore::PasteboardWebContent, text)
#endif
#if PLATFORM(GTK)
        , offsetof(WebCore::PasteboardWebContent, markup)
#endif
#if USE(LIBWPE)
        , offsetof(WebCore::PasteboardWebContent, text)
#endif
#if USE(LIBWPE)
        , offsetof(WebCore::PasteboardWebContent, markup)
#endif
    >::value);
#if PLATFORM(COCOA)
    encoder << instance.contentOrigin;
#endif
#if PLATFORM(COCOA)
    encoder << instance.canSmartCopyOrDelete;
#endif
#if PLATFORM(COCOA)
    encoder << instance.dataInWebArchiveFormat;
#endif
#if PLATFORM(COCOA)
    encoder << instance.dataInRTFDFormat;
#endif
#if PLATFORM(COCOA)
    encoder << instance.dataInRTFFormat;
#endif
#if PLATFORM(COCOA)
    encoder << instance.dataInAttributedStringFormat;
#endif
#if PLATFORM(COCOA)
    encoder << instance.dataInHTMLFormat;
#endif
#if PLATFORM(COCOA)
    encoder << instance.dataInStringFormat;
#endif
#if PLATFORM(COCOA)
    encoder << instance.clientTypes;
#endif
#if PLATFORM(COCOA)
    encoder << instance.clientData;
#endif
#if PLATFORM(GTK)
    encoder << instance.contentOrigin;
#endif
#if PLATFORM(GTK)
    encoder << instance.canSmartCopyOrDelete;
#endif
#if PLATFORM(GTK)
    encoder << instance.text;
#endif
#if PLATFORM(GTK)
    encoder << instance.markup;
#endif
#if USE(LIBWPE)
    encoder << instance.text;
#endif
#if USE(LIBWPE)
    encoder << instance.markup;
#endif
}

std::optional<WebCore::PasteboardWebContent> ArgumentCoder<WebCore::PasteboardWebContent>::decode(Decoder& decoder)
{
#if PLATFORM(COCOA)
    auto contentOrigin = decoder.decode<String>();
#endif
#if PLATFORM(COCOA)
    auto canSmartCopyOrDelete = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto dataInWebArchiveFormat = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
#endif
#if PLATFORM(COCOA)
    auto dataInRTFDFormat = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
#endif
#if PLATFORM(COCOA)
    auto dataInRTFFormat = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
#endif
#if PLATFORM(COCOA)
    auto dataInAttributedStringFormat = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
#endif
#if PLATFORM(COCOA)
    auto dataInHTMLFormat = decoder.decode<String>();
#endif
#if PLATFORM(COCOA)
    auto dataInStringFormat = decoder.decode<String>();
#endif
#if PLATFORM(COCOA)
    auto clientTypes = decoder.decode<Vector<String>>();
#endif
#if PLATFORM(COCOA)
    auto clientData = decoder.decode<Vector<RefPtr<WebCore::SharedBuffer>>>();
#endif
#if PLATFORM(GTK)
    auto contentOrigin = decoder.decode<String>();
#endif
#if PLATFORM(GTK)
    auto canSmartCopyOrDelete = decoder.decode<bool>();
#endif
#if PLATFORM(GTK)
    auto text = decoder.decode<String>();
#endif
#if PLATFORM(GTK)
    auto markup = decoder.decode<String>();
#endif
#if USE(LIBWPE)
    auto text = decoder.decode<String>();
#endif
#if USE(LIBWPE)
    auto markup = decoder.decode<String>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PasteboardWebContent {
#if PLATFORM(COCOA)
            WTFMove(*contentOrigin),
#endif
#if PLATFORM(COCOA)
            WTFMove(*canSmartCopyOrDelete),
#endif
#if PLATFORM(COCOA)
            WTFMove(*dataInWebArchiveFormat),
#endif
#if PLATFORM(COCOA)
            WTFMove(*dataInRTFDFormat),
#endif
#if PLATFORM(COCOA)
            WTFMove(*dataInRTFFormat),
#endif
#if PLATFORM(COCOA)
            WTFMove(*dataInAttributedStringFormat),
#endif
#if PLATFORM(COCOA)
            WTFMove(*dataInHTMLFormat),
#endif
#if PLATFORM(COCOA)
            WTFMove(*dataInStringFormat),
#endif
#if PLATFORM(COCOA)
            WTFMove(*clientTypes),
#endif
#if PLATFORM(COCOA)
            WTFMove(*clientData),
#endif
#if PLATFORM(GTK)
            WTFMove(*contentOrigin),
#endif
#if PLATFORM(GTK)
            WTFMove(*canSmartCopyOrDelete),
#endif
#if PLATFORM(GTK)
            WTFMove(*text),
#endif
#if PLATFORM(GTK)
            WTFMove(*markup),
#endif
#if USE(LIBWPE)
            WTFMove(*text),
#endif
#if USE(LIBWPE)
            WTFMove(*markup)
#endif
        }
    };
}

void ArgumentCoder<WebCore::PasteboardURL>::encode(Encoder& encoder, const WebCore::PasteboardURL& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.title)>, String>);
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userVisibleForm)>, String>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.markup)>, String>);
#endif
    struct ShouldBeSameSizeAsPasteboardURL : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PasteboardURL>, false> {
        URL url;
        String title;
#if PLATFORM(MAC)
        String userVisibleForm;
#endif
#if PLATFORM(GTK)
        String markup;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsPasteboardURL) == sizeof(WebCore::PasteboardURL));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PasteboardURL, url)
        , offsetof(WebCore::PasteboardURL, title)
#if PLATFORM(MAC)
        , offsetof(WebCore::PasteboardURL, userVisibleForm)
#endif
#if PLATFORM(GTK)
        , offsetof(WebCore::PasteboardURL, markup)
#endif
    >::value);
    encoder << instance.url;
    encoder << instance.title;
#if PLATFORM(MAC)
    encoder << instance.userVisibleForm;
#endif
#if PLATFORM(GTK)
    encoder << instance.markup;
#endif
}

std::optional<WebCore::PasteboardURL> ArgumentCoder<WebCore::PasteboardURL>::decode(Decoder& decoder)
{
    auto url = decoder.decode<URL>();
    auto title = decoder.decode<String>();
#if PLATFORM(MAC)
    auto userVisibleForm = decoder.decode<String>();
#endif
#if PLATFORM(GTK)
    auto markup = decoder.decode<String>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PasteboardURL {
            WTFMove(*url),
            WTFMove(*title),
#if PLATFORM(MAC)
            WTFMove(*userVisibleForm),
#endif
#if PLATFORM(GTK)
            WTFMove(*markup)
#endif
        }
    };
}

void ArgumentCoder<WebCore::PasteboardBuffer>::encode(Encoder& encoder, const WebCore::PasteboardBuffer& instance)
{
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentOrigin)>, String>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data)>, RefPtr<WebCore::SharedBuffer>>);
    struct ShouldBeSameSizeAsPasteboardBuffer : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PasteboardBuffer>, false> {
#if PLATFORM(COCOA)
        String contentOrigin;
#endif
        String type;
        RefPtr<WebCore::SharedBuffer> data;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPasteboardBuffer) == sizeof(WebCore::PasteboardBuffer));
    static_assert(MembersInCorrectOrder < 0
#if PLATFORM(COCOA)
        , offsetof(WebCore::PasteboardBuffer, contentOrigin)
#endif
        , offsetof(WebCore::PasteboardBuffer, type)
        , offsetof(WebCore::PasteboardBuffer, data)
    >::value);
#if PLATFORM(COCOA)
    encoder << instance.contentOrigin;
#endif
    encoder << instance.type;
    encoder << instance.data;
}

std::optional<WebCore::PasteboardBuffer> ArgumentCoder<WebCore::PasteboardBuffer>::decode(Decoder& decoder)
{
#if PLATFORM(COCOA)
    auto contentOrigin = decoder.decode<String>();
#endif
    auto type = decoder.decode<String>();
    auto data = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PasteboardBuffer {
#if PLATFORM(COCOA)
            WTFMove(*contentOrigin),
#endif
            WTFMove(*type),
            WTFMove(*data)
        }
    };
}

void ArgumentCoder<WebKit::PlatformPopupMenuData>::encode(Encoder& encoder, const WebKit::PlatformPopupMenuData& instance)
{
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontInfo)>, WebKit::FontInfo>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldPopOver)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hideArrows)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.menuSize)>, WebCore::PopupMenuStyle::PopupMenuSize>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_clientPaddingLeft)>, int>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_clientPaddingRight)>, int>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_clientInsetLeft)>, int>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_clientInsetRight)>, int>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_popupWidth)>, int>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_itemHeight)>, int>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_notSelectedBackingStore)>, RefPtr<WebKit::ShareableBitmap>>);
#endif
#if PLATFORM(WIN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_selectedBackingStore)>, RefPtr<WebKit::ShareableBitmap>>);
#endif
    struct ShouldBeSameSizeAsPlatformPopupMenuData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::PlatformPopupMenuData>, false> {
#if PLATFORM(COCOA)
        WebKit::FontInfo fontInfo;
#endif
#if PLATFORM(COCOA)
        bool shouldPopOver;
#endif
#if PLATFORM(COCOA)
        bool hideArrows;
#endif
#if PLATFORM(COCOA)
        WebCore::PopupMenuStyle::PopupMenuSize menuSize;
#endif
#if PLATFORM(WIN)
        int m_clientPaddingLeft;
#endif
#if PLATFORM(WIN)
        int m_clientPaddingRight;
#endif
#if PLATFORM(WIN)
        int m_clientInsetLeft;
#endif
#if PLATFORM(WIN)
        int m_clientInsetRight;
#endif
#if PLATFORM(WIN)
        int m_popupWidth;
#endif
#if PLATFORM(WIN)
        int m_itemHeight;
#endif
#if PLATFORM(WIN)
        RefPtr<WebKit::ShareableBitmap> m_notSelectedBackingStore;
#endif
#if PLATFORM(WIN)
        RefPtr<WebKit::ShareableBitmap> m_selectedBackingStore;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsPlatformPopupMenuData) == sizeof(WebKit::PlatformPopupMenuData));
    static_assert(MembersInCorrectOrder < 0
#if PLATFORM(COCOA)
        , offsetof(WebKit::PlatformPopupMenuData, fontInfo)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::PlatformPopupMenuData, shouldPopOver)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::PlatformPopupMenuData, hideArrows)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::PlatformPopupMenuData, menuSize)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_clientPaddingLeft)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_clientPaddingRight)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_clientInsetLeft)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_clientInsetRight)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_popupWidth)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_itemHeight)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_notSelectedBackingStore)
#endif
#if PLATFORM(WIN)
        , offsetof(WebKit::PlatformPopupMenuData, m_selectedBackingStore)
#endif
    >::value);
#if PLATFORM(COCOA)
    encoder << instance.fontInfo;
#endif
#if PLATFORM(COCOA)
    encoder << instance.shouldPopOver;
#endif
#if PLATFORM(COCOA)
    encoder << instance.hideArrows;
#endif
#if PLATFORM(COCOA)
    encoder << instance.menuSize;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_clientPaddingLeft;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_clientPaddingRight;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_clientInsetLeft;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_clientInsetRight;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_popupWidth;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_itemHeight;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_notSelectedBackingStore;
#endif
#if PLATFORM(WIN)
    encoder << instance.m_selectedBackingStore;
#endif
}

std::optional<WebKit::PlatformPopupMenuData> ArgumentCoder<WebKit::PlatformPopupMenuData>::decode(Decoder& decoder)
{
#if PLATFORM(COCOA)
    auto fontInfo = decoder.decode<WebKit::FontInfo>();
#endif
#if PLATFORM(COCOA)
    auto shouldPopOver = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto hideArrows = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto menuSize = decoder.decode<WebCore::PopupMenuStyle::PopupMenuSize>();
#endif
#if PLATFORM(WIN)
    auto m_clientPaddingLeft = decoder.decode<int>();
#endif
#if PLATFORM(WIN)
    auto m_clientPaddingRight = decoder.decode<int>();
#endif
#if PLATFORM(WIN)
    auto m_clientInsetLeft = decoder.decode<int>();
#endif
#if PLATFORM(WIN)
    auto m_clientInsetRight = decoder.decode<int>();
#endif
#if PLATFORM(WIN)
    auto m_popupWidth = decoder.decode<int>();
#endif
#if PLATFORM(WIN)
    auto m_itemHeight = decoder.decode<int>();
#endif
#if PLATFORM(WIN)
    auto m_notSelectedBackingStore = decoder.decode<RefPtr<WebKit::ShareableBitmap>>();
#endif
#if PLATFORM(WIN)
    auto m_selectedBackingStore = decoder.decode<RefPtr<WebKit::ShareableBitmap>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::PlatformPopupMenuData {
#if PLATFORM(COCOA)
            WTFMove(*fontInfo),
#endif
#if PLATFORM(COCOA)
            WTFMove(*shouldPopOver),
#endif
#if PLATFORM(COCOA)
            WTFMove(*hideArrows),
#endif
#if PLATFORM(COCOA)
            WTFMove(*menuSize),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_clientPaddingLeft),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_clientPaddingRight),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_clientInsetLeft),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_clientInsetRight),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_popupWidth),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_itemHeight),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_notSelectedBackingStore),
#endif
#if PLATFORM(WIN)
            WTFMove(*m_selectedBackingStore)
#endif
        }
    };
}

void ArgumentCoder<WebKit::PolicyDecision>::encode(Encoder& encoder, const WebKit::PolicyDecision& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isNavigatingToAppBoundDomain)>, std::optional<WebKit::NavigatingToAppBoundDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.policyAction)>, WebCore::PolicyAction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.navigationID)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.downloadID)>, std::optional<WebKit::DownloadID>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.websitePoliciesData)>, std::optional<WebKit::WebsitePoliciesData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sandboxExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
    struct ShouldBeSameSizeAsPolicyDecision : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::PolicyDecision>, false> {
        std::optional<WebKit::NavigatingToAppBoundDomain> isNavigatingToAppBoundDomain;
        WebCore::PolicyAction policyAction;
        uint64_t navigationID;
        std::optional<WebKit::DownloadID> downloadID;
        std::optional<WebKit::WebsitePoliciesData> websitePoliciesData;
        std::optional<WebKit::SandboxExtension::Handle> sandboxExtensionHandle;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPolicyDecision) == sizeof(WebKit::PolicyDecision));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::PolicyDecision, isNavigatingToAppBoundDomain)
        , offsetof(WebKit::PolicyDecision, policyAction)
        , offsetof(WebKit::PolicyDecision, navigationID)
        , offsetof(WebKit::PolicyDecision, downloadID)
        , offsetof(WebKit::PolicyDecision, websitePoliciesData)
        , offsetof(WebKit::PolicyDecision, sandboxExtensionHandle)
    >::value);
    encoder << instance.isNavigatingToAppBoundDomain;
    encoder << instance.policyAction;
    encoder << instance.navigationID;
    encoder << instance.downloadID;
    encoder << instance.websitePoliciesData;
    encoder << instance.sandboxExtensionHandle;
}

std::optional<WebKit::PolicyDecision> ArgumentCoder<WebKit::PolicyDecision>::decode(Decoder& decoder)
{
    auto isNavigatingToAppBoundDomain = decoder.decode<std::optional<WebKit::NavigatingToAppBoundDomain>>();
    auto policyAction = decoder.decode<WebCore::PolicyAction>();
    auto navigationID = decoder.decode<uint64_t>();
    auto downloadID = decoder.decode<std::optional<WebKit::DownloadID>>();
    auto websitePoliciesData = decoder.decode<std::optional<WebKit::WebsitePoliciesData>>();
    auto sandboxExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::PolicyDecision {
            WTFMove(*isNavigatingToAppBoundDomain),
            WTFMove(*policyAction),
            WTFMove(*navigationID),
            WTFMove(*downloadID),
            WTFMove(*websitePoliciesData),
            WTFMove(*sandboxExtensionHandle)
        }
    };
}

void ArgumentCoder<WebKit::HTTPBody>::encode(Encoder& encoder, const WebKit::HTTPBody& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentType)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elements)>, Vector<WebKit::HTTPBody::Element>>);
    struct ShouldBeSameSizeAsHTTPBody : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::HTTPBody>, false> {
        String contentType;
        Vector<WebKit::HTTPBody::Element> elements;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHTTPBody) == sizeof(WebKit::HTTPBody));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::HTTPBody, contentType)
        , offsetof(WebKit::HTTPBody, elements)
    >::value);
    encoder << instance.contentType;
    encoder << instance.elements;
}

std::optional<WebKit::HTTPBody> ArgumentCoder<WebKit::HTTPBody>::decode(Decoder& decoder)
{
    auto contentType = decoder.decode<String>();
    auto elements = decoder.decode<Vector<WebKit::HTTPBody::Element>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::HTTPBody {
            WTFMove(*contentType),
            WTFMove(*elements)
        }
    };
}

void ArgumentCoder<WebKit::HTTPBody::Element>::encode(Encoder& encoder, const WebKit::HTTPBody::Element& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data)>, std::variant<Vector<uint8_t>, WebKit::HTTPBody::Element::FileData, String>>);
    struct ShouldBeSameSizeAsElement : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::HTTPBody::Element>, false> {
        std::variant<Vector<uint8_t>, WebKit::HTTPBody::Element::FileData, String> data;
    };
    static_assert(sizeof(ShouldBeSameSizeAsElement) == sizeof(WebKit::HTTPBody::Element));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::HTTPBody::Element, data)
    >::value);
    encoder << instance.data;
}

std::optional<WebKit::HTTPBody::Element> ArgumentCoder<WebKit::HTTPBody::Element>::decode(Decoder& decoder)
{
    auto data = decoder.decode<std::variant<Vector<uint8_t>, WebKit::HTTPBody::Element::FileData, String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::HTTPBody::Element {
            WTFMove(*data)
        }
    };
}

void ArgumentCoder<WebKit::HTTPBody::Element::FileData>::encode(Encoder& encoder, const WebKit::HTTPBody::Element::FileData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.filePath)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fileStart)>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fileLength)>, std::optional<int64_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expectedFileModificationTime)>, std::optional<WallTime>>);
    struct ShouldBeSameSizeAsFileData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::HTTPBody::Element::FileData>, false> {
        String filePath;
        int64_t fileStart;
        std::optional<int64_t> fileLength;
        std::optional<WallTime> expectedFileModificationTime;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFileData) == sizeof(WebKit::HTTPBody::Element::FileData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::HTTPBody::Element::FileData, filePath)
        , offsetof(WebKit::HTTPBody::Element::FileData, fileStart)
        , offsetof(WebKit::HTTPBody::Element::FileData, fileLength)
        , offsetof(WebKit::HTTPBody::Element::FileData, expectedFileModificationTime)
    >::value);
    encoder << instance.filePath;
    encoder << instance.fileStart;
    encoder << instance.fileLength;
    encoder << instance.expectedFileModificationTime;
}

std::optional<WebKit::HTTPBody::Element::FileData> ArgumentCoder<WebKit::HTTPBody::Element::FileData>::decode(Decoder& decoder)
{
    auto filePath = decoder.decode<String>();
    auto fileStart = decoder.decode<int64_t>();
    auto fileLength = decoder.decode<std::optional<int64_t>>();
    auto expectedFileModificationTime = decoder.decode<std::optional<WallTime>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::HTTPBody::Element::FileData {
            WTFMove(*filePath),
            WTFMove(*fileStart),
            WTFMove(*fileLength),
            WTFMove(*expectedFileModificationTime)
        }
    };
}

void ArgumentCoder<WebKit::FrameState>::encode(Encoder& encoder, const WebKit::FrameState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlString)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originalURLString)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrer)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.target)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stateObjectData)>, std::optional<Vector<uint8_t>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentSequenceNumber)>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.itemSequenceNumber)>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollPosition)>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRestoreScrollPosition)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageScaleFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpBody)>, std::optional<WebKit::HTTPBody>>);
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.exposedContentRect)>, WebCore::FloatRect>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.unobscuredContentRect)>, WebCore::IntRect>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimumLayoutSizeInScrollViewCoordinates)>, WebCore::FloatSize>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSize)>, WebCore::IntSize>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scaleIsInitial)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.obscuredInsets)>, WebCore::FloatBoxExtent>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.children)>, Vector<WebKit::FrameState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_documentState)>, Vector<AtomString>>);
    struct ShouldBeSameSizeAsFrameState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::FrameState>, false> {
        String urlString;
        String originalURLString;
        String referrer;
        AtomString target;
        std::optional<Vector<uint8_t>> stateObjectData;
        int64_t documentSequenceNumber;
        int64_t itemSequenceNumber;
        WebCore::IntPoint scrollPosition;
        bool shouldRestoreScrollPosition;
        float pageScaleFactor;
        std::optional<WebKit::HTTPBody> httpBody;
#if PLATFORM(IOS_FAMILY)
        WebCore::FloatRect exposedContentRect;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::IntRect unobscuredContentRect;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::FloatSize minimumLayoutSizeInScrollViewCoordinates;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::IntSize contentSize;
#endif
#if PLATFORM(IOS_FAMILY)
        bool scaleIsInitial;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::FloatBoxExtent obscuredInsets;
#endif
        Vector<WebKit::FrameState> children;
        Vector<AtomString> m_documentState;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFrameState) == sizeof(WebKit::FrameState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::FrameState, urlString)
        , offsetof(WebKit::FrameState, originalURLString)
        , offsetof(WebKit::FrameState, referrer)
        , offsetof(WebKit::FrameState, target)
        , offsetof(WebKit::FrameState, stateObjectData)
        , offsetof(WebKit::FrameState, documentSequenceNumber)
        , offsetof(WebKit::FrameState, itemSequenceNumber)
        , offsetof(WebKit::FrameState, scrollPosition)
        , offsetof(WebKit::FrameState, shouldRestoreScrollPosition)
        , offsetof(WebKit::FrameState, pageScaleFactor)
        , offsetof(WebKit::FrameState, httpBody)
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::FrameState, exposedContentRect)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::FrameState, unobscuredContentRect)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::FrameState, minimumLayoutSizeInScrollViewCoordinates)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::FrameState, contentSize)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::FrameState, scaleIsInitial)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::FrameState, obscuredInsets)
#endif
        , offsetof(WebKit::FrameState, children)
        , offsetof(WebKit::FrameState, m_documentState)
    >::value);
    encoder << instance.urlString;
    encoder << instance.originalURLString;
    encoder << instance.referrer;
    encoder << instance.target;
    encoder << instance.stateObjectData;
    encoder << instance.documentSequenceNumber;
    encoder << instance.itemSequenceNumber;
    encoder << instance.scrollPosition;
    encoder << instance.shouldRestoreScrollPosition;
    encoder << instance.pageScaleFactor;
    encoder << instance.httpBody;
#if PLATFORM(IOS_FAMILY)
    encoder << instance.exposedContentRect;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.unobscuredContentRect;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.minimumLayoutSizeInScrollViewCoordinates;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.contentSize;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.scaleIsInitial;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.obscuredInsets;
#endif
    encoder << instance.children;
    encoder << instance.m_documentState;
}

std::optional<WebKit::FrameState> ArgumentCoder<WebKit::FrameState>::decode(Decoder& decoder)
{
    auto urlString = decoder.decode<String>();
    auto originalURLString = decoder.decode<String>();
    auto referrer = decoder.decode<String>();
    auto target = decoder.decode<AtomString>();
    auto stateObjectData = decoder.decode<std::optional<Vector<uint8_t>>>();
    auto documentSequenceNumber = decoder.decode<int64_t>();
    auto itemSequenceNumber = decoder.decode<int64_t>();
    auto scrollPosition = decoder.decode<WebCore::IntPoint>();
    auto shouldRestoreScrollPosition = decoder.decode<bool>();
    auto pageScaleFactor = decoder.decode<float>();
    auto httpBody = decoder.decode<std::optional<WebKit::HTTPBody>>();
#if PLATFORM(IOS_FAMILY)
    auto exposedContentRect = decoder.decode<WebCore::FloatRect>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto unobscuredContentRect = decoder.decode<WebCore::IntRect>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto minimumLayoutSizeInScrollViewCoordinates = decoder.decode<WebCore::FloatSize>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto contentSize = decoder.decode<WebCore::IntSize>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto scaleIsInitial = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto obscuredInsets = decoder.decode<WebCore::FloatBoxExtent>();
#endif
    auto children = decoder.decode<Vector<WebKit::FrameState>>();
    auto m_documentState = decoder.decode<Vector<AtomString>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(WebKit::FrameState::validateDocumentState(*m_documentState)))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    WebKit::FrameState result;
    result.urlString = WTFMove(*urlString);
    result.originalURLString = WTFMove(*originalURLString);
    result.referrer = WTFMove(*referrer);
    result.target = WTFMove(*target);
    result.stateObjectData = WTFMove(*stateObjectData);
    result.documentSequenceNumber = WTFMove(*documentSequenceNumber);
    result.itemSequenceNumber = WTFMove(*itemSequenceNumber);
    result.scrollPosition = WTFMove(*scrollPosition);
    result.shouldRestoreScrollPosition = WTFMove(*shouldRestoreScrollPosition);
    result.pageScaleFactor = WTFMove(*pageScaleFactor);
    result.httpBody = WTFMove(*httpBody);
#if PLATFORM(IOS_FAMILY)
    result.exposedContentRect = WTFMove(*exposedContentRect);
#endif
#if PLATFORM(IOS_FAMILY)
    result.unobscuredContentRect = WTFMove(*unobscuredContentRect);
#endif
#if PLATFORM(IOS_FAMILY)
    result.minimumLayoutSizeInScrollViewCoordinates = WTFMove(*minimumLayoutSizeInScrollViewCoordinates);
#endif
#if PLATFORM(IOS_FAMILY)
    result.contentSize = WTFMove(*contentSize);
#endif
#if PLATFORM(IOS_FAMILY)
    result.scaleIsInitial = WTFMove(*scaleIsInitial);
#endif
#if PLATFORM(IOS_FAMILY)
    result.obscuredInsets = WTFMove(*obscuredInsets);
#endif
    result.children = WTFMove(*children);
    result.m_documentState = WTFMove(*m_documentState);
    return { WTFMove(result) };
}

void ArgumentCoder<WebKit::PageState>::encode(Encoder& encoder, const WebKit::PageState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.title)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mainFrameState)>, WebKit::FrameState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldOpenExternalURLsPolicy)>, WebCore::ShouldOpenExternalURLsPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sessionStateObject)>, RefPtr<WebCore::SerializedScriptValue>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wasCreatedByJSWithoutUserInteraction)>, bool>);
    struct ShouldBeSameSizeAsPageState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::PageState>, false> {
        String title;
        WebKit::FrameState mainFrameState;
        WebCore::ShouldOpenExternalURLsPolicy shouldOpenExternalURLsPolicy;
        RefPtr<WebCore::SerializedScriptValue> sessionStateObject;
        bool wasCreatedByJSWithoutUserInteraction;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPageState) == sizeof(WebKit::PageState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::PageState, title)
        , offsetof(WebKit::PageState, mainFrameState)
        , offsetof(WebKit::PageState, shouldOpenExternalURLsPolicy)
        , offsetof(WebKit::PageState, sessionStateObject)
        , offsetof(WebKit::PageState, wasCreatedByJSWithoutUserInteraction)
    >::value);
    encoder << instance.title;
    encoder << instance.mainFrameState;
    encoder << instance.shouldOpenExternalURLsPolicy;
    encoder << instance.sessionStateObject;
    encoder << instance.wasCreatedByJSWithoutUserInteraction;
}

std::optional<WebKit::PageState> ArgumentCoder<WebKit::PageState>::decode(Decoder& decoder)
{
    auto title = decoder.decode<String>();
    auto mainFrameState = decoder.decode<WebKit::FrameState>();
    auto shouldOpenExternalURLsPolicy = decoder.decode<WebCore::ShouldOpenExternalURLsPolicy>();
    auto sessionStateObject = decoder.decode<RefPtr<WebCore::SerializedScriptValue>>();
    auto wasCreatedByJSWithoutUserInteraction = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::PageState {
            WTFMove(*title),
            WTFMove(*mainFrameState),
            WTFMove(*shouldOpenExternalURLsPolicy),
            WTFMove(*sessionStateObject),
            WTFMove(*wasCreatedByJSWithoutUserInteraction)
        }
    };
}

void ArgumentCoder<WebKit::BackForwardListItemState>::encode(Encoder& encoder, const WebKit::BackForwardListItemState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::BackForwardItemIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageState)>, WebKit::PageState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasCachedPage)>, bool>);
#if PLATFORM(COCOA) || PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.snapshot)>, RefPtr<WebKit::ViewSnapshot>>);
#endif
    struct ShouldBeSameSizeAsBackForwardListItemState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::BackForwardListItemState>, false> {
        WebCore::BackForwardItemIdentifier identifier;
        WebKit::PageState pageState;
        bool hasCachedPage;
#if PLATFORM(COCOA) || PLATFORM(GTK)
        RefPtr<WebKit::ViewSnapshot> snapshot;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsBackForwardListItemState) == sizeof(WebKit::BackForwardListItemState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::BackForwardListItemState, identifier)
        , offsetof(WebKit::BackForwardListItemState, pageState)
        , offsetof(WebKit::BackForwardListItemState, hasCachedPage)
#if PLATFORM(COCOA) || PLATFORM(GTK)
        , offsetof(WebKit::BackForwardListItemState, snapshot)
#endif
    >::value);
    encoder << instance.identifier;
    encoder << instance.pageState;
    encoder << instance.hasCachedPage;
}

std::optional<WebKit::BackForwardListItemState> ArgumentCoder<WebKit::BackForwardListItemState>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::BackForwardItemIdentifier>();
    auto pageState = decoder.decode<WebKit::PageState>();
    auto hasCachedPage = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::BackForwardListItemState {
            WTFMove(*identifier),
            WTFMove(*pageState),
            WTFMove(*hasCachedPage)
        }
    };
}

void ArgumentCoder<WebKit::BackForwardListState>::encode(Encoder& encoder, const WebKit::BackForwardListState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.items)>, Vector<WebKit::BackForwardListItemState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.currentIndex)>, std::optional<uint32_t>>);
    struct ShouldBeSameSizeAsBackForwardListState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::BackForwardListState>, false> {
        Vector<WebKit::BackForwardListItemState> items;
        std::optional<uint32_t> currentIndex;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBackForwardListState) == sizeof(WebKit::BackForwardListState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::BackForwardListState, items)
        , offsetof(WebKit::BackForwardListState, currentIndex)
    >::value);
    encoder << instance.items;
    encoder << instance.currentIndex;
}

std::optional<WebKit::BackForwardListState> ArgumentCoder<WebKit::BackForwardListState>::decode(Decoder& decoder)
{
    auto items = decoder.decode<Vector<WebKit::BackForwardListItemState>>();
    auto currentIndex = decoder.decode<std::optional<uint32_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::BackForwardListState {
            WTFMove(*items),
            WTFMove(*currentIndex)
        }
    };
}

void ArgumentCoder<WebKit::ShareableBitmapConfiguration>::encode(Encoder& encoder, const WebKit::ShareableBitmapConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_size)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_colorSpace)>, std::optional<WebCore::DestinationColorSpace>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isOpaque)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bytesPerPixel())>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bytesPerRow())>, unsigned>);
#if USE(CG)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_bitmapInfo)>, CGBitmapInfo>);
#endif
    encoder << instance.m_size;
    encoder << instance.m_colorSpace;
    encoder << instance.m_isOpaque;
    encoder << instance.bytesPerPixel();
    encoder << instance.bytesPerRow();
#if USE(CG)
    encoder << instance.m_bitmapInfo;
#endif
}

std::optional<WebKit::ShareableBitmapConfiguration> ArgumentCoder<WebKit::ShareableBitmapConfiguration>::decode(Decoder& decoder)
{
    auto m_size = decoder.decode<WebCore::IntSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(m_size->width() >= 0 && m_size->height() >= 0))
        return std::nullopt;
    auto m_colorSpace = decoder.decode<std::optional<WebCore::DestinationColorSpace>>();
    auto m_isOpaque = decoder.decode<bool>();
    auto bytesPerPixel = decoder.decode<unsigned>();
    auto bytesPerRow = decoder.decode<unsigned>();
#if USE(CG)
    auto m_bitmapInfo = decoder.decode<CGBitmapInfo>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::ShareableBitmapConfiguration {
            WTFMove(*m_size),
            WTFMove(*m_colorSpace),
            WTFMove(*m_isOpaque),
            WTFMove(*bytesPerPixel),
            WTFMove(*bytesPerRow),
#if USE(CG)
            WTFMove(*m_bitmapInfo)
#endif
        }
    };
}

void ArgumentCoder<WebKit::ShareableBitmapHandle>::encode(Encoder& encoder, WebKit::ShareableBitmapHandle&& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_handle)>, WebKit::SharedMemoryHandle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_configuration)>, WebKit::ShareableBitmapConfiguration>);
    struct ShouldBeSameSizeAsShareableBitmapHandle : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::ShareableBitmapHandle>, false> {
        WebKit::SharedMemoryHandle m_handle;
        WebKit::ShareableBitmapConfiguration m_configuration;
    };
    static_assert(sizeof(ShouldBeSameSizeAsShareableBitmapHandle) == sizeof(WebKit::ShareableBitmapHandle));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::ShareableBitmapHandle, m_handle)
        , offsetof(WebKit::ShareableBitmapHandle, m_configuration)
    >::value);
    encoder << WTFMove(instance.m_handle);
    encoder << WTFMove(instance.m_configuration);
}

std::optional<WebKit::ShareableBitmapHandle> ArgumentCoder<WebKit::ShareableBitmapHandle>::decode(Decoder& decoder)
{
    auto m_handle = decoder.decode<WebKit::SharedMemoryHandle>();
    auto m_configuration = decoder.decode<WebKit::ShareableBitmapConfiguration>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(!m_configuration->sizeInBytes().hasOverflowed() && m_handle->size() >= m_configuration->sizeInBytes()))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::ShareableBitmapHandle {
            WTFMove(*m_handle),
            WTFMove(*m_configuration)
        }
    };
}

void ArgumentCoder<WebKit::ShareableBitmap>::encode(Encoder& encoder, const WebKit::ShareableBitmap& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.createReadOnlyHandle())>, std::optional<WebKit::ShareableBitmapHandle>>);
    encoder << instance.createReadOnlyHandle();
}

std::optional<Ref<WebKit::ShareableBitmap>> ArgumentCoder<WebKit::ShareableBitmap>::decode(Decoder& decoder)
{
    auto createReadOnlyHandle = decoder.decode<std::optional<WebKit::ShareableBitmapHandle>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::ShareableBitmap::createReadOnly(
            WTFMove(*createReadOnlyHandle)
        )
    };
}

#if ENABLE(IMAGE_ANALYSIS)
void ArgumentCoder<WebCore::TextRecognitionWordData>::encode(Encoder& encoder, const WebCore::TextRecognitionWordData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.normalizedQuad)>, WebCore::FloatQuad>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasLeadingWhitespace)>, bool>);
    struct ShouldBeSameSizeAsTextRecognitionWordData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextRecognitionWordData>, false> {
        String text;
        WebCore::FloatQuad normalizedQuad;
        bool hasLeadingWhitespace;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextRecognitionWordData) == sizeof(WebCore::TextRecognitionWordData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextRecognitionWordData, text)
        , offsetof(WebCore::TextRecognitionWordData, normalizedQuad)
        , offsetof(WebCore::TextRecognitionWordData, hasLeadingWhitespace)
    >::value);
    encoder << instance.text;
    encoder << instance.normalizedQuad;
    encoder << instance.hasLeadingWhitespace;
}

std::optional<WebCore::TextRecognitionWordData> ArgumentCoder<WebCore::TextRecognitionWordData>::decode(Decoder& decoder)
{
    auto text = decoder.decode<String>();
    auto normalizedQuad = decoder.decode<WebCore::FloatQuad>();
    auto hasLeadingWhitespace = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextRecognitionWordData {
            WTFMove(*text),
            WTFMove(*normalizedQuad),
            WTFMove(*hasLeadingWhitespace)
        }
    };
}

#endif

#if ENABLE(IMAGE_ANALYSIS)
void ArgumentCoder<WebCore::TextRecognitionLineData>::encode(Encoder& encoder, const WebCore::TextRecognitionLineData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.normalizedQuad)>, WebCore::FloatQuad>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.children)>, Vector<WebCore::TextRecognitionWordData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasTrailingNewline)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isVertical)>, bool>);
    struct ShouldBeSameSizeAsTextRecognitionLineData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextRecognitionLineData>, false> {
        WebCore::FloatQuad normalizedQuad;
        Vector<WebCore::TextRecognitionWordData> children;
        bool hasTrailingNewline;
        bool isVertical;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextRecognitionLineData) == sizeof(WebCore::TextRecognitionLineData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextRecognitionLineData, normalizedQuad)
        , offsetof(WebCore::TextRecognitionLineData, children)
        , offsetof(WebCore::TextRecognitionLineData, hasTrailingNewline)
        , offsetof(WebCore::TextRecognitionLineData, isVertical)
    >::value);
    encoder << instance.normalizedQuad;
    encoder << instance.children;
    encoder << instance.hasTrailingNewline;
    encoder << instance.isVertical;
}

std::optional<WebCore::TextRecognitionLineData> ArgumentCoder<WebCore::TextRecognitionLineData>::decode(Decoder& decoder)
{
    auto normalizedQuad = decoder.decode<WebCore::FloatQuad>();
    auto children = decoder.decode<Vector<WebCore::TextRecognitionWordData>>();
    auto hasTrailingNewline = decoder.decode<bool>();
    auto isVertical = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextRecognitionLineData {
            WTFMove(*normalizedQuad),
            WTFMove(*children),
            WTFMove(*hasTrailingNewline),
            WTFMove(*isVertical)
        }
    };
}

#endif

#if ENABLE(IMAGE_ANALYSIS)
void ArgumentCoder<WebCore::TextRecognitionBlockData>::encode(Encoder& encoder, const WebCore::TextRecognitionBlockData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.normalizedQuad)>, WebCore::FloatQuad>);
    struct ShouldBeSameSizeAsTextRecognitionBlockData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextRecognitionBlockData>, false> {
        String text;
        WebCore::FloatQuad normalizedQuad;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextRecognitionBlockData) == sizeof(WebCore::TextRecognitionBlockData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextRecognitionBlockData, text)
        , offsetof(WebCore::TextRecognitionBlockData, normalizedQuad)
    >::value);
    encoder << instance.text;
    encoder << instance.normalizedQuad;
}

std::optional<WebCore::TextRecognitionBlockData> ArgumentCoder<WebCore::TextRecognitionBlockData>::decode(Decoder& decoder)
{
    auto text = decoder.decode<String>();
    auto normalizedQuad = decoder.decode<WebCore::FloatQuad>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextRecognitionBlockData {
            WTFMove(*text),
            WTFMove(*normalizedQuad)
        }
    };
}

#endif

#if ENABLE(IMAGE_ANALYSIS)
void ArgumentCoder<WebCore::TextRecognitionResult>::encode(Encoder& encoder, const WebCore::TextRecognitionResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lines)>, Vector<WebCore::TextRecognitionLineData>>);
#if ENABLE(DATA_DETECTION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataDetectors)>, Vector<WebCore::TextRecognitionDataDetector>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blocks)>, Vector<WebCore::TextRecognitionBlockData>>);
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.platformData)>, RetainPtr<VKCImageAnalysis>>);
#endif
    struct ShouldBeSameSizeAsTextRecognitionResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextRecognitionResult>, false> {
        Vector<WebCore::TextRecognitionLineData> lines;
#if ENABLE(DATA_DETECTION)
        Vector<WebCore::TextRecognitionDataDetector> dataDetectors;
#endif
        Vector<WebCore::TextRecognitionBlockData> blocks;
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
        RetainPtr<VKCImageAnalysis> platformData;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextRecognitionResult) == sizeof(WebCore::TextRecognitionResult));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextRecognitionResult, lines)
#if ENABLE(DATA_DETECTION)
        , offsetof(WebCore::TextRecognitionResult, dataDetectors)
#endif
        , offsetof(WebCore::TextRecognitionResult, blocks)
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
        , offsetof(WebCore::TextRecognitionResult, platformData)
#endif
    >::value);
    encoder << instance.lines;
#if ENABLE(DATA_DETECTION)
    encoder << instance.dataDetectors;
#endif
    encoder << instance.blocks;
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
    encoder << instance.platformData;
#endif
}

std::optional<WebCore::TextRecognitionResult> ArgumentCoder<WebCore::TextRecognitionResult>::decode(Decoder& decoder)
{
    auto lines = decoder.decode<Vector<WebCore::TextRecognitionLineData>>();
#if ENABLE(DATA_DETECTION)
    auto dataDetectors = decoder.decode<Vector<WebCore::TextRecognitionDataDetector>>();
#endif
    auto blocks = decoder.decode<Vector<WebCore::TextRecognitionBlockData>>();
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
    auto platformData = decoder.decode<RetainPtr<VKCImageAnalysis>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextRecognitionResult {
            WTFMove(*lines),
#if ENABLE(DATA_DETECTION)
            WTFMove(*dataDetectors),
#endif
            WTFMove(*blocks),
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
            WTFMove(*platformData)
#endif
        }
    };
}

#endif

#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
void ArgumentCoder<WebKit::UpdateInfo>::encode(Encoder& encoder, WebKit::UpdateInfo&& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewSize)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deviceScaleFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollRect)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollOffset)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.updateRectBounds)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.updateRects)>, Vector<WebCore::IntRect>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.updateScaleFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bitmapHandle)>, WebKit::ShareableBitmapHandle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bitmapOffset)>, WebCore::IntPoint>);
    struct ShouldBeSameSizeAsUpdateInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::UpdateInfo>, false> {
        WebCore::IntSize viewSize;
        float deviceScaleFactor;
        WebCore::IntRect scrollRect;
        WebCore::IntSize scrollOffset;
        WebCore::IntRect updateRectBounds;
        Vector<WebCore::IntRect> updateRects;
        float updateScaleFactor;
        WebKit::ShareableBitmapHandle bitmapHandle;
        WebCore::IntPoint bitmapOffset;
    };
    static_assert(sizeof(ShouldBeSameSizeAsUpdateInfo) == sizeof(WebKit::UpdateInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::UpdateInfo, viewSize)
        , offsetof(WebKit::UpdateInfo, deviceScaleFactor)
        , offsetof(WebKit::UpdateInfo, scrollRect)
        , offsetof(WebKit::UpdateInfo, scrollOffset)
        , offsetof(WebKit::UpdateInfo, updateRectBounds)
        , offsetof(WebKit::UpdateInfo, updateRects)
        , offsetof(WebKit::UpdateInfo, updateScaleFactor)
        , offsetof(WebKit::UpdateInfo, bitmapHandle)
        , offsetof(WebKit::UpdateInfo, bitmapOffset)
    >::value);
    encoder << WTFMove(instance.viewSize);
    encoder << WTFMove(instance.deviceScaleFactor);
    encoder << WTFMove(instance.scrollRect);
    encoder << WTFMove(instance.scrollOffset);
    encoder << WTFMove(instance.updateRectBounds);
    encoder << WTFMove(instance.updateRects);
    encoder << WTFMove(instance.updateScaleFactor);
    encoder << WTFMove(instance.bitmapHandle);
    encoder << WTFMove(instance.bitmapOffset);
}

std::optional<WebKit::UpdateInfo> ArgumentCoder<WebKit::UpdateInfo>::decode(Decoder& decoder)
{
    auto viewSize = decoder.decode<WebCore::IntSize>();
    auto deviceScaleFactor = decoder.decode<float>();
    auto scrollRect = decoder.decode<WebCore::IntRect>();
    auto scrollOffset = decoder.decode<WebCore::IntSize>();
    auto updateRectBounds = decoder.decode<WebCore::IntRect>();
    auto updateRects = decoder.decode<Vector<WebCore::IntRect>>();
    auto updateScaleFactor = decoder.decode<float>();
    auto bitmapHandle = decoder.decode<WebKit::ShareableBitmapHandle>();
    auto bitmapOffset = decoder.decode<WebCore::IntPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::UpdateInfo {
            WTFMove(*viewSize),
            WTFMove(*deviceScaleFactor),
            WTFMove(*scrollRect),
            WTFMove(*scrollOffset),
            WTFMove(*updateRectBounds),
            WTFMove(*updateRects),
            WTFMove(*updateScaleFactor),
            WTFMove(*bitmapHandle),
            WTFMove(*bitmapOffset)
        }
    };
}

#endif

void ArgumentCoder<WTF::URL>::encode(Encoder& encoder, const WTF::URL& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.string())>, String>);
    encoder << instance.string();
}

std::optional<WTF::URL> ArgumentCoder<WTF::URL>::decode(Decoder& decoder)
{
    auto string = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WTF::URL {
            WTFMove(*string)
        }
    };
}

void ArgumentCoder<WTF::MediaTime>::encode(Encoder& encoder, const WTF::MediaTime& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeValue())>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeScale())>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeFlags())>, uint8_t>);
    encoder << instance.timeValue();
    encoder << instance.timeScale();
    encoder << instance.timeFlags();
}

void ArgumentCoder<WTF::MediaTime>::encode(StreamConnectionEncoder& encoder, const WTF::MediaTime& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeValue())>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeScale())>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeFlags())>, uint8_t>);
    encoder << instance.timeValue();
    encoder << instance.timeScale();
    encoder << instance.timeFlags();
}

std::optional<WTF::MediaTime> ArgumentCoder<WTF::MediaTime>::decode(Decoder& decoder)
{
    auto timeValue = decoder.decode<int64_t>();
    auto timeScale = decoder.decode<uint32_t>();
    auto timeFlags = decoder.decode<uint8_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WTF::MediaTime {
            WTFMove(*timeValue),
            WTFMove(*timeScale),
            WTFMove(*timeFlags)
        }
    };
}

void ArgumentCoder<WTF::Seconds>::encode(Encoder& encoder, const WTF::Seconds& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, double>);
    encoder << instance.value();
}

void ArgumentCoder<WTF::Seconds>::encode(StreamConnectionEncoder& encoder, const WTF::Seconds& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, double>);
    encoder << instance.value();
}

std::optional<WTF::Seconds> ArgumentCoder<WTF::Seconds>::decode(Decoder& decoder)
{
    auto value = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WTF::Seconds {
            WTFMove(*value)
        }
    };
}

void ArgumentCoder<WTF::MonotonicTime>::encode(Encoder& encoder, const WTF::MonotonicTime& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.secondsSinceEpoch().seconds())>, double>);
    encoder << instance.secondsSinceEpoch().seconds();
}

std::optional<WTF::MonotonicTime> ArgumentCoder<WTF::MonotonicTime>::decode(Decoder& decoder)
{
    auto secondsSinceEpochseconds = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WTF::MonotonicTime::fromRawSeconds(
            WTFMove(*secondsSinceEpochseconds)
        )
    };
}

void ArgumentCoder<WTF::WallTime>::encode(Encoder& encoder, const WTF::WallTime& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.secondsSinceEpoch().seconds())>, double>);
    encoder << instance.secondsSinceEpoch().seconds();
}

std::optional<WTF::WallTime> ArgumentCoder<WTF::WallTime>::decode(Decoder& decoder)
{
    auto secondsSinceEpochseconds = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WTF::WallTime::fromRawSeconds(
            WTFMove(*secondsSinceEpochseconds)
        )
    };
}

void ArgumentCoder<WTF::AtomString>::encode(Encoder& encoder, const WTF::AtomString& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.string())>, String>);
    encoder << instance.string();
}

void ArgumentCoder<WTF::AtomString>::encode(StreamConnectionEncoder& encoder, const WTF::AtomString& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.string())>, String>);
    encoder << instance.string();
}

std::optional<WTF::AtomString> ArgumentCoder<WTF::AtomString>::decode(Decoder& decoder)
{
    auto string = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WTF::AtomString {
            WTFMove(*string)
        }
    };
}

void ArgumentCoder<WebCore::DOMCacheEngine::CacheInfo>::encode(Encoder& encoder, const WebCore::DOMCacheEngine::CacheInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::DOMCacheIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    struct ShouldBeSameSizeAsCacheInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DOMCacheEngine::CacheInfo>, false> {
        WebCore::DOMCacheIdentifier identifier;
        String name;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCacheInfo) == sizeof(WebCore::DOMCacheEngine::CacheInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DOMCacheEngine::CacheInfo, identifier)
        , offsetof(WebCore::DOMCacheEngine::CacheInfo, name)
    >::value);
    encoder << instance.identifier;
    encoder << instance.name;
}

std::optional<WebCore::DOMCacheEngine::CacheInfo> ArgumentCoder<WebCore::DOMCacheEngine::CacheInfo>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::DOMCacheIdentifier>();
    auto name = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DOMCacheEngine::CacheInfo {
            WTFMove(*identifier),
            WTFMove(*name)
        }
    };
}

void ArgumentCoder<WebCore::DOMCacheEngine::CacheInfos>::encode(Encoder& encoder, const WebCore::DOMCacheEngine::CacheInfos& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.infos)>, Vector<WebCore::DOMCacheEngine::CacheInfo>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.updateCounter)>, uint64_t>);
    struct ShouldBeSameSizeAsCacheInfos : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DOMCacheEngine::CacheInfos>, false> {
        Vector<WebCore::DOMCacheEngine::CacheInfo> infos;
        uint64_t updateCounter;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCacheInfos) == sizeof(WebCore::DOMCacheEngine::CacheInfos));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DOMCacheEngine::CacheInfos, infos)
        , offsetof(WebCore::DOMCacheEngine::CacheInfos, updateCounter)
    >::value);
    encoder << instance.infos;
    encoder << instance.updateCounter;
}

std::optional<WebCore::DOMCacheEngine::CacheInfos> ArgumentCoder<WebCore::DOMCacheEngine::CacheInfos>::decode(Decoder& decoder)
{
    auto infos = decoder.decode<Vector<WebCore::DOMCacheEngine::CacheInfo>>();
    auto updateCounter = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DOMCacheEngine::CacheInfos {
            WTFMove(*infos),
            WTFMove(*updateCounter)
        }
    };
}

void ArgumentCoder<WebCore::DOMCacheEngine::CacheIdentifierOperationResult>::encode(Encoder& encoder, const WebCore::DOMCacheEngine::CacheIdentifierOperationResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::DOMCacheIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hadStorageError)>, bool>);
    struct ShouldBeSameSizeAsCacheIdentifierOperationResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DOMCacheEngine::CacheIdentifierOperationResult>, false> {
        WebCore::DOMCacheIdentifier identifier;
        bool hadStorageError;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCacheIdentifierOperationResult) == sizeof(WebCore::DOMCacheEngine::CacheIdentifierOperationResult));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DOMCacheEngine::CacheIdentifierOperationResult, identifier)
        , offsetof(WebCore::DOMCacheEngine::CacheIdentifierOperationResult, hadStorageError)
    >::value);
    encoder << instance.identifier;
    encoder << instance.hadStorageError;
}

std::optional<WebCore::DOMCacheEngine::CacheIdentifierOperationResult> ArgumentCoder<WebCore::DOMCacheEngine::CacheIdentifierOperationResult>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::DOMCacheIdentifier>();
    auto hadStorageError = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DOMCacheEngine::CacheIdentifierOperationResult {
            WTFMove(*identifier),
            WTFMove(*hadStorageError)
        }
    };
}

#if HAVE(TRANSLATION_UI_SERVICES) && ENABLE(CONTEXT_MENUS)
void ArgumentCoder<WebCore::TranslationContextMenuInfo>::encode(Encoder& encoder, const WebCore::TranslationContextMenuInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionBoundsInRootView)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.locationInRootView)>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mode)>, WebCore::TranslationContextMenuMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.source)>, WebCore::TranslationContextMenuSource>);
    struct ShouldBeSameSizeAsTranslationContextMenuInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TranslationContextMenuInfo>, false> {
        String text;
        WebCore::IntRect selectionBoundsInRootView;
        WebCore::IntPoint locationInRootView;
        WebCore::TranslationContextMenuMode mode;
        WebCore::TranslationContextMenuSource source;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTranslationContextMenuInfo) == sizeof(WebCore::TranslationContextMenuInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TranslationContextMenuInfo, text)
        , offsetof(WebCore::TranslationContextMenuInfo, selectionBoundsInRootView)
        , offsetof(WebCore::TranslationContextMenuInfo, locationInRootView)
        , offsetof(WebCore::TranslationContextMenuInfo, mode)
        , offsetof(WebCore::TranslationContextMenuInfo, source)
    >::value);
    encoder << instance.text;
    encoder << instance.selectionBoundsInRootView;
    encoder << instance.locationInRootView;
    encoder << instance.mode;
    encoder << instance.source;
}

std::optional<WebCore::TranslationContextMenuInfo> ArgumentCoder<WebCore::TranslationContextMenuInfo>::decode(Decoder& decoder)
{
    auto text = decoder.decode<String>();
    auto selectionBoundsInRootView = decoder.decode<WebCore::IntRect>();
    auto locationInRootView = decoder.decode<WebCore::IntPoint>();
    auto mode = decoder.decode<WebCore::TranslationContextMenuMode>();
    auto source = decoder.decode<WebCore::TranslationContextMenuSource>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TranslationContextMenuInfo {
            WTFMove(*text),
            WTFMove(*selectionBoundsInRootView),
            WTFMove(*locationInRootView),
            WTFMove(*mode),
            WTFMove(*source)
        }
    };
}

#endif

void ArgumentCoder<WebCore::DOMCacheEngine::CrossThreadRecord>::encode(Encoder& encoder, const WebCore::DOMCacheEngine::CrossThreadRecord& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.updateResponseCounter)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requestHeadersGuard)>, WebCore::FetchHeadersGuard>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.request)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.options)>, WebCore::FetchOptions>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrer)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.responseHeadersGuard)>, WebCore::FetchHeadersGuard>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.response)>, WebCore::ResourceResponse::CrossThreadData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.responseBody)>, std::variant<std::nullptr_t, Ref<WebCore::FormData>, Ref<WebCore::SharedBuffer>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.responseBodySize)>, uint64_t>);
    struct ShouldBeSameSizeAsCrossThreadRecord : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DOMCacheEngine::CrossThreadRecord>, false> {
        uint64_t identifier;
        uint64_t updateResponseCounter;
        WebCore::FetchHeadersGuard requestHeadersGuard;
        WebCore::ResourceRequest request;
        WebCore::FetchOptions options;
        String referrer;
        WebCore::FetchHeadersGuard responseHeadersGuard;
        WebCore::ResourceResponse::CrossThreadData response;
        std::variant<std::nullptr_t, Ref<WebCore::FormData>, Ref<WebCore::SharedBuffer>> responseBody;
        uint64_t responseBodySize;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCrossThreadRecord) == sizeof(WebCore::DOMCacheEngine::CrossThreadRecord));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, identifier)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, updateResponseCounter)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, requestHeadersGuard)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, request)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, options)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, referrer)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, responseHeadersGuard)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, response)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, responseBody)
        , offsetof(WebCore::DOMCacheEngine::CrossThreadRecord, responseBodySize)
    >::value);
    encoder << instance.identifier;
    encoder << instance.updateResponseCounter;
    encoder << instance.requestHeadersGuard;
    encoder << instance.request;
    encoder << instance.options;
    encoder << instance.referrer;
    encoder << instance.responseHeadersGuard;
    encoder << instance.response;
    encoder << instance.responseBody;
    encoder << instance.responseBodySize;
}

std::optional<WebCore::DOMCacheEngine::CrossThreadRecord> ArgumentCoder<WebCore::DOMCacheEngine::CrossThreadRecord>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<uint64_t>();
    auto updateResponseCounter = decoder.decode<uint64_t>();
    auto requestHeadersGuard = decoder.decode<WebCore::FetchHeadersGuard>();
    auto request = decoder.decode<WebCore::ResourceRequest>();
    auto options = decoder.decode<WebCore::FetchOptions>();
    auto referrer = decoder.decode<String>();
    auto responseHeadersGuard = decoder.decode<WebCore::FetchHeadersGuard>();
    auto response = decoder.decode<WebCore::ResourceResponse::CrossThreadData>();
    auto responseBody = decoder.decode<std::variant<std::nullptr_t, Ref<WebCore::FormData>, Ref<WebCore::SharedBuffer>>>();
    auto responseBodySize = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DOMCacheEngine::CrossThreadRecord {
            WTFMove(*identifier),
            WTFMove(*updateResponseCounter),
            WTFMove(*requestHeadersGuard),
            WTFMove(*request),
            WTFMove(*options),
            WTFMove(*referrer),
            WTFMove(*responseHeadersGuard),
            WTFMove(*response),
            WTFMove(*responseBody),
            WTFMove(*responseBodySize)
        }
    };
}

void ArgumentCoder<WebCore::TransformationMatrix>::encode(Encoder& encoder, const WebCore::TransformationMatrix& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m11())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m12())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m13())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m14())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m21())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m22())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m23())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m24())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m31())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m32())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m33())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m34())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m41())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m42())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m43())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m44())>, double>);
    encoder << instance.m11();
    encoder << instance.m12();
    encoder << instance.m13();
    encoder << instance.m14();
    encoder << instance.m21();
    encoder << instance.m22();
    encoder << instance.m23();
    encoder << instance.m24();
    encoder << instance.m31();
    encoder << instance.m32();
    encoder << instance.m33();
    encoder << instance.m34();
    encoder << instance.m41();
    encoder << instance.m42();
    encoder << instance.m43();
    encoder << instance.m44();
}

void ArgumentCoder<WebCore::TransformationMatrix>::encode(StreamConnectionEncoder& encoder, const WebCore::TransformationMatrix& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m11())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m12())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m13())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m14())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m21())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m22())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m23())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m24())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m31())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m32())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m33())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m34())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m41())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m42())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m43())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m44())>, double>);
    encoder << instance.m11();
    encoder << instance.m12();
    encoder << instance.m13();
    encoder << instance.m14();
    encoder << instance.m21();
    encoder << instance.m22();
    encoder << instance.m23();
    encoder << instance.m24();
    encoder << instance.m31();
    encoder << instance.m32();
    encoder << instance.m33();
    encoder << instance.m34();
    encoder << instance.m41();
    encoder << instance.m42();
    encoder << instance.m43();
    encoder << instance.m44();
}

std::optional<WebCore::TransformationMatrix> ArgumentCoder<WebCore::TransformationMatrix>::decode(Decoder& decoder)
{
    auto m11 = decoder.decode<double>();
    auto m12 = decoder.decode<double>();
    auto m13 = decoder.decode<double>();
    auto m14 = decoder.decode<double>();
    auto m21 = decoder.decode<double>();
    auto m22 = decoder.decode<double>();
    auto m23 = decoder.decode<double>();
    auto m24 = decoder.decode<double>();
    auto m31 = decoder.decode<double>();
    auto m32 = decoder.decode<double>();
    auto m33 = decoder.decode<double>();
    auto m34 = decoder.decode<double>();
    auto m41 = decoder.decode<double>();
    auto m42 = decoder.decode<double>();
    auto m43 = decoder.decode<double>();
    auto m44 = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TransformationMatrix {
            WTFMove(*m11),
            WTFMove(*m12),
            WTFMove(*m13),
            WTFMove(*m14),
            WTFMove(*m21),
            WTFMove(*m22),
            WTFMove(*m23),
            WTFMove(*m24),
            WTFMove(*m31),
            WTFMove(*m32),
            WTFMove(*m33),
            WTFMove(*m34),
            WTFMove(*m41),
            WTFMove(*m42),
            WTFMove(*m43),
            WTFMove(*m44)
        }
    };
}

void ArgumentCoder<WebCore::CacheQueryOptions>::encode(Encoder& encoder, const WebCore::CacheQueryOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoreSearch)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoreMethod)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoreVary)>, bool>);
    struct ShouldBeSameSizeAsCacheQueryOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CacheQueryOptions>, false> {
        bool ignoreSearch;
        bool ignoreMethod;
        bool ignoreVary;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCacheQueryOptions) == sizeof(WebCore::CacheQueryOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CacheQueryOptions, ignoreSearch)
        , offsetof(WebCore::CacheQueryOptions, ignoreMethod)
        , offsetof(WebCore::CacheQueryOptions, ignoreVary)
    >::value);
    encoder << instance.ignoreSearch;
    encoder << instance.ignoreMethod;
    encoder << instance.ignoreVary;
}

std::optional<WebCore::CacheQueryOptions> ArgumentCoder<WebCore::CacheQueryOptions>::decode(Decoder& decoder)
{
    auto ignoreSearch = decoder.decode<bool>();
    auto ignoreMethod = decoder.decode<bool>();
    auto ignoreVary = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CacheQueryOptions {
            WTFMove(*ignoreSearch),
            WTFMove(*ignoreMethod),
            WTFMove(*ignoreVary)
        }
    };
}

void ArgumentCoder<WebCore::CharacterRange>::encode(Encoder& encoder, const WebCore::CharacterRange& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.length)>, uint64_t>);
    struct ShouldBeSameSizeAsCharacterRange : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CharacterRange>, false> {
        uint64_t location;
        uint64_t length;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCharacterRange) == sizeof(WebCore::CharacterRange));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CharacterRange, location)
        , offsetof(WebCore::CharacterRange, length)
    >::value);
    encoder << instance.location;
    encoder << instance.length;
}

std::optional<WebCore::CharacterRange> ArgumentCoder<WebCore::CharacterRange>::decode(Decoder& decoder)
{
    auto location = decoder.decode<uint64_t>();
    auto length = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CharacterRange {
            WTFMove(*location),
            WTFMove(*length)
        }
    };
}

void ArgumentCoder<WebCore::AffineTransform>::encode(Encoder& encoder, const WebCore::AffineTransform& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.a())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.b())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.c())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.d())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.e())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.f())>, double>);
    encoder << instance.a();
    encoder << instance.b();
    encoder << instance.c();
    encoder << instance.d();
    encoder << instance.e();
    encoder << instance.f();
}

void ArgumentCoder<WebCore::AffineTransform>::encode(StreamConnectionEncoder& encoder, const WebCore::AffineTransform& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.a())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.b())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.c())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.d())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.e())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.f())>, double>);
    encoder << instance.a();
    encoder << instance.b();
    encoder << instance.c();
    encoder << instance.d();
    encoder << instance.e();
    encoder << instance.f();
}

std::optional<WebCore::AffineTransform> ArgumentCoder<WebCore::AffineTransform>::decode(Decoder& decoder)
{
    auto a = decoder.decode<double>();
    auto b = decoder.decode<double>();
    auto c = decoder.decode<double>();
    auto d = decoder.decode<double>();
    auto e = decoder.decode<double>();
    auto f = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AffineTransform {
            WTFMove(*a),
            WTFMove(*b),
            WTFMove(*c),
            WTFMove(*d),
            WTFMove(*e),
            WTFMove(*f)
        }
    };
}

void ArgumentCoder<WebCore::FloatPoint>::encode(Encoder& encoder, const WebCore::FloatPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, float>);
    encoder << instance.x();
    encoder << instance.y();
}

void ArgumentCoder<WebCore::FloatPoint>::encode(StreamConnectionEncoder& encoder, const WebCore::FloatPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, float>);
    encoder << instance.x();
    encoder << instance.y();
}

std::optional<WebCore::FloatPoint> ArgumentCoder<WebCore::FloatPoint>::decode(Decoder& decoder)
{
    auto x = decoder.decode<float>();
    auto y = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatPoint {
            WTFMove(*x),
            WTFMove(*y)
        }
    };
}

void ArgumentCoder<WebCore::FloatPoint3D>::encode(Encoder& encoder, const WebCore::FloatPoint3D& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.z())>, float>);
    encoder << instance.x();
    encoder << instance.y();
    encoder << instance.z();
}

void ArgumentCoder<WebCore::FloatPoint3D>::encode(StreamConnectionEncoder& encoder, const WebCore::FloatPoint3D& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.z())>, float>);
    encoder << instance.x();
    encoder << instance.y();
    encoder << instance.z();
}

std::optional<WebCore::FloatPoint3D> ArgumentCoder<WebCore::FloatPoint3D>::decode(Decoder& decoder)
{
    auto x = decoder.decode<float>();
    auto y = decoder.decode<float>();
    auto z = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatPoint3D {
            WTFMove(*x),
            WTFMove(*y),
            WTFMove(*z)
        }
    };
}

void ArgumentCoder<WebCore::FloatQuad>::encode(Encoder& encoder, const WebCore::FloatQuad& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.p1())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.p2())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.p3())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.p4())>, WebCore::FloatPoint>);
    encoder << instance.p1();
    encoder << instance.p2();
    encoder << instance.p3();
    encoder << instance.p4();
}

std::optional<WebCore::FloatQuad> ArgumentCoder<WebCore::FloatQuad>::decode(Decoder& decoder)
{
    auto p1 = decoder.decode<WebCore::FloatPoint>();
    auto p2 = decoder.decode<WebCore::FloatPoint>();
    auto p3 = decoder.decode<WebCore::FloatPoint>();
    auto p4 = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatQuad {
            WTFMove(*p1),
            WTFMove(*p2),
            WTFMove(*p3),
            WTFMove(*p4)
        }
    };
}

void ArgumentCoder<WebCore::IDBCursorRecord>::encode(Encoder& encoder, const WebCore::IDBCursorRecord& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.key)>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primaryKey)>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, WebCore::IDBValue>);
    struct ShouldBeSameSizeAsIDBCursorRecord : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBCursorRecord>, false> {
        WebCore::IDBKeyData key;
        WebCore::IDBKeyData primaryKey;
        WebCore::IDBValue value;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBCursorRecord) == sizeof(WebCore::IDBCursorRecord));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBCursorRecord, key)
        , offsetof(WebCore::IDBCursorRecord, primaryKey)
        , offsetof(WebCore::IDBCursorRecord, value)
    >::value);
    encoder << instance.key;
    encoder << instance.primaryKey;
    encoder << instance.value;
}

std::optional<WebCore::IDBCursorRecord> ArgumentCoder<WebCore::IDBCursorRecord>::decode(Decoder& decoder)
{
    auto key = decoder.decode<WebCore::IDBKeyData>();
    auto primaryKey = decoder.decode<WebCore::IDBKeyData>();
    auto value = decoder.decode<WebCore::IDBValue>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBCursorRecord {
            WTFMove(*key),
            WTFMove(*primaryKey),
            WTFMove(*value)
        }
    };
}

void ArgumentCoder<WebCore::IDBCursorInfo>::encode(Encoder& encoder, const WebCore::IDBCursorInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier())>, WebCore::IDBResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transactionIdentifier())>, WebCore::IDBResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.objectStoreIdentifier())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceIdentifier())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.range())>, WebCore::IDBKeyRangeData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cursorSource())>, WebCore::IndexedDB::CursorSource>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cursorDirection())>, WebCore::IndexedDB::CursorDirection>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cursorType())>, WebCore::IndexedDB::CursorType>);
    encoder << instance.identifier();
    encoder << instance.transactionIdentifier();
    encoder << instance.objectStoreIdentifier();
    encoder << instance.sourceIdentifier();
    encoder << instance.range();
    encoder << instance.cursorSource();
    encoder << instance.cursorDirection();
    encoder << instance.cursorType();
}

std::optional<WebCore::IDBCursorInfo> ArgumentCoder<WebCore::IDBCursorInfo>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::IDBResourceIdentifier>();
    auto transactionIdentifier = decoder.decode<WebCore::IDBResourceIdentifier>();
    auto objectStoreIdentifier = decoder.decode<uint64_t>();
    auto sourceIdentifier = decoder.decode<uint64_t>();
    auto range = decoder.decode<WebCore::IDBKeyRangeData>();
    auto cursorSource = decoder.decode<WebCore::IndexedDB::CursorSource>();
    auto cursorDirection = decoder.decode<WebCore::IndexedDB::CursorDirection>();
    auto cursorType = decoder.decode<WebCore::IndexedDB::CursorType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBCursorInfo {
            WTFMove(*identifier),
            WTFMove(*transactionIdentifier),
            WTFMove(*objectStoreIdentifier),
            WTFMove(*sourceIdentifier),
            WTFMove(*range),
            WTFMove(*cursorSource),
            WTFMove(*cursorDirection),
            WTFMove(*cursorType)
        }
    };
}

void ArgumentCoder<WebCore::IDBError>::encode(Encoder& encoder, const WebCore::IDBError& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.code())>, std::optional<WebCore::ExceptionCode>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.messageForSerialization())>, String>);
    encoder << instance.code();
    encoder << instance.messageForSerialization();
}

std::optional<WebCore::IDBError> ArgumentCoder<WebCore::IDBError>::decode(Decoder& decoder)
{
    auto code = decoder.decode<std::optional<WebCore::ExceptionCode>>();
    auto messageForSerialization = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBError {
            WTFMove(*code),
            WTFMove(*messageForSerialization)
        }
    };
}

void ArgumentCoder<WebCore::IDBGetAllRecordsData>::encode(Encoder& encoder, const WebCore::IDBGetAllRecordsData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyRangeData)>, WebCore::IDBKeyRangeData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getAllType)>, WebCore::IndexedDB::GetAllType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.count)>, std::optional<uint32_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.objectStoreIdentifier)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.indexIdentifier)>, uint64_t>);
    struct ShouldBeSameSizeAsIDBGetAllRecordsData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBGetAllRecordsData>, false> {
        WebCore::IDBKeyRangeData keyRangeData;
        WebCore::IndexedDB::GetAllType getAllType;
        std::optional<uint32_t> count;
        uint64_t objectStoreIdentifier;
        uint64_t indexIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBGetAllRecordsData) == sizeof(WebCore::IDBGetAllRecordsData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBGetAllRecordsData, keyRangeData)
        , offsetof(WebCore::IDBGetAllRecordsData, getAllType)
        , offsetof(WebCore::IDBGetAllRecordsData, count)
        , offsetof(WebCore::IDBGetAllRecordsData, objectStoreIdentifier)
        , offsetof(WebCore::IDBGetAllRecordsData, indexIdentifier)
    >::value);
    encoder << instance.keyRangeData;
    encoder << instance.getAllType;
    encoder << instance.count;
    encoder << instance.objectStoreIdentifier;
    encoder << instance.indexIdentifier;
}

std::optional<WebCore::IDBGetAllRecordsData> ArgumentCoder<WebCore::IDBGetAllRecordsData>::decode(Decoder& decoder)
{
    auto keyRangeData = decoder.decode<WebCore::IDBKeyRangeData>();
    auto getAllType = decoder.decode<WebCore::IndexedDB::GetAllType>();
    auto count = decoder.decode<std::optional<uint32_t>>();
    auto objectStoreIdentifier = decoder.decode<uint64_t>();
    auto indexIdentifier = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBGetAllRecordsData {
            WTFMove(*keyRangeData),
            WTFMove(*getAllType),
            WTFMove(*count),
            WTFMove(*objectStoreIdentifier),
            WTFMove(*indexIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::IDBGetResult>::encode(Encoder& encoder, const WebCore::IDBGetResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyData())>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primaryKeyData())>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, WebCore::IDBValue>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyPath())>, std::optional<WebCore::IDBKeyPath>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.prefetchedRecords())>, Vector<WebCore::IDBCursorRecord>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isDefined())>, bool>);
    encoder << instance.keyData();
    encoder << instance.primaryKeyData();
    encoder << instance.value();
    encoder << instance.keyPath();
    encoder << instance.prefetchedRecords();
    encoder << instance.isDefined();
}

std::optional<WebCore::IDBGetResult> ArgumentCoder<WebCore::IDBGetResult>::decode(Decoder& decoder)
{
    auto keyData = decoder.decode<WebCore::IDBKeyData>();
    auto primaryKeyData = decoder.decode<WebCore::IDBKeyData>();
    auto value = decoder.decode<WebCore::IDBValue>();
    auto keyPath = decoder.decode<std::optional<WebCore::IDBKeyPath>>();
    auto prefetchedRecords = decoder.decode<Vector<WebCore::IDBCursorRecord>>();
    auto isDefined = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBGetResult {
            WTFMove(*keyData),
            WTFMove(*primaryKeyData),
            WTFMove(*value),
            WTFMove(*keyPath),
            WTFMove(*prefetchedRecords),
            WTFMove(*isDefined)
        }
    };
}

void ArgumentCoder<WebCore::IDBGetAllResult>::encode(Encoder& encoder, const WebCore::IDBGetAllResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::IndexedDB::GetAllType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keys())>, Vector<WebCore::IDBKeyData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.values())>, Vector<WebCore::IDBValue>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyPath())>, std::optional<WebCore::IDBKeyPath>>);
    encoder << instance.type();
    encoder << instance.keys();
    encoder << instance.values();
    encoder << instance.keyPath();
}

std::optional<WebCore::IDBGetAllResult> ArgumentCoder<WebCore::IDBGetAllResult>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::IndexedDB::GetAllType>();
    auto keys = decoder.decode<Vector<WebCore::IDBKeyData>>();
    auto values = decoder.decode<Vector<WebCore::IDBValue>>();
    auto keyPath = decoder.decode<std::optional<WebCore::IDBKeyPath>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBGetAllResult {
            WTFMove(*type),
            WTFMove(*keys),
            WTFMove(*values),
            WTFMove(*keyPath)
        }
    };
}

void ArgumentCoder<WebCore::IDBDatabaseInfo>::encode(Encoder& encoder, const WebCore::IDBDatabaseInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_version)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_maxIndexID)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_maxObjectStoreID)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_objectStoreMap)>, HashMap<uint64_t, WebCore::IDBObjectStoreInfo>>);
    struct ShouldBeSameSizeAsIDBDatabaseInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBDatabaseInfo>, false> {
        String m_name;
        uint64_t m_version;
        uint64_t m_maxIndexID;
        uint64_t m_maxObjectStoreID;
        HashMap<uint64_t, WebCore::IDBObjectStoreInfo> m_objectStoreMap;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBDatabaseInfo) == sizeof(WebCore::IDBDatabaseInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBDatabaseInfo, m_name)
        , offsetof(WebCore::IDBDatabaseInfo, m_version)
        , offsetof(WebCore::IDBDatabaseInfo, m_maxIndexID)
        , offsetof(WebCore::IDBDatabaseInfo, m_maxObjectStoreID)
        , offsetof(WebCore::IDBDatabaseInfo, m_objectStoreMap)
    >::value);
    encoder << instance.m_name;
    encoder << instance.m_version;
    encoder << instance.m_maxIndexID;
    encoder << instance.m_maxObjectStoreID;
    encoder << instance.m_objectStoreMap;
}

std::optional<WebCore::IDBDatabaseInfo> ArgumentCoder<WebCore::IDBDatabaseInfo>::decode(Decoder& decoder)
{
    auto m_name = decoder.decode<String>();
    auto m_version = decoder.decode<uint64_t>();
    auto m_maxIndexID = decoder.decode<uint64_t>();
    auto m_maxObjectStoreID = decoder.decode<uint64_t>();
    auto m_objectStoreMap = decoder.decode<HashMap<uint64_t, WebCore::IDBObjectStoreInfo>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBDatabaseInfo {
            WTFMove(*m_name),
            WTFMove(*m_version),
            WTFMove(*m_maxIndexID),
            WTFMove(*m_maxObjectStoreID),
            WTFMove(*m_objectStoreMap)
        }
    };
}

void ArgumentCoder<WebCore::IDBKeyRangeData>::encode(Encoder& encoder, const WebCore::IDBKeyRangeData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lowerKey)>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.upperKey)>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lowerOpen)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.upperOpen)>, bool>);
    struct ShouldBeSameSizeAsIDBKeyRangeData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBKeyRangeData>, false> {
        WebCore::IDBKeyData lowerKey;
        WebCore::IDBKeyData upperKey;
        bool lowerOpen;
        bool upperOpen;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBKeyRangeData) == sizeof(WebCore::IDBKeyRangeData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBKeyRangeData, lowerKey)
        , offsetof(WebCore::IDBKeyRangeData, upperKey)
        , offsetof(WebCore::IDBKeyRangeData, lowerOpen)
        , offsetof(WebCore::IDBKeyRangeData, upperOpen)
    >::value);
    encoder << instance.lowerKey;
    encoder << instance.upperKey;
    encoder << instance.lowerOpen;
    encoder << instance.upperOpen;
}

std::optional<WebCore::IDBKeyRangeData> ArgumentCoder<WebCore::IDBKeyRangeData>::decode(Decoder& decoder)
{
    auto lowerKey = decoder.decode<WebCore::IDBKeyData>();
    auto upperKey = decoder.decode<WebCore::IDBKeyData>();
    auto lowerOpen = decoder.decode<bool>();
    auto upperOpen = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBKeyRangeData {
            WTFMove(*lowerKey),
            WTFMove(*upperKey),
            WTFMove(*lowerOpen),
            WTFMove(*upperOpen)
        }
    };
}

void ArgumentCoder<WebCore::IDBTransactionInfo>::encode(Encoder& encoder, const WebCore::IDBTransactionInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier())>, WebCore::IDBResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mode())>, WebCore::IDBTransactionMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.durability())>, WebCore::IDBTransactionDurability>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newVersion())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.objectStores())>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originalDatabaseInfo())>, std::unique_ptr<WebCore::IDBDatabaseInfo>>);
    encoder << instance.identifier();
    encoder << instance.mode();
    encoder << instance.durability();
    encoder << instance.newVersion();
    encoder << instance.objectStores();
    encoder << instance.originalDatabaseInfo();
}

std::optional<WebCore::IDBTransactionInfo> ArgumentCoder<WebCore::IDBTransactionInfo>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::IDBResourceIdentifier>();
    auto mode = decoder.decode<WebCore::IDBTransactionMode>();
    auto durability = decoder.decode<WebCore::IDBTransactionDurability>();
    auto newVersion = decoder.decode<uint64_t>();
    auto objectStores = decoder.decode<Vector<String>>();
    auto originalDatabaseInfo = decoder.decode<std::unique_ptr<WebCore::IDBDatabaseInfo>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBTransactionInfo {
            WTFMove(*identifier),
            WTFMove(*mode),
            WTFMove(*durability),
            WTFMove(*newVersion),
            WTFMove(*objectStores),
            WTFMove(*originalDatabaseInfo)
        }
    };
}

void ArgumentCoder<WebCore::IDBGetRecordData>::encode(Encoder& encoder, const WebCore::IDBGetRecordData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyRangeData)>, WebCore::IDBKeyRangeData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::IDBGetRecordDataType>);
    struct ShouldBeSameSizeAsIDBGetRecordData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBGetRecordData>, false> {
        WebCore::IDBKeyRangeData keyRangeData;
        WebCore::IDBGetRecordDataType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBGetRecordData) == sizeof(WebCore::IDBGetRecordData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBGetRecordData, keyRangeData)
        , offsetof(WebCore::IDBGetRecordData, type)
    >::value);
    encoder << instance.keyRangeData;
    encoder << instance.type;
}

std::optional<WebCore::IDBGetRecordData> ArgumentCoder<WebCore::IDBGetRecordData>::decode(Decoder& decoder)
{
    auto keyRangeData = decoder.decode<WebCore::IDBKeyRangeData>();
    auto type = decoder.decode<WebCore::IDBGetRecordDataType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBGetRecordData {
            WTFMove(*keyRangeData),
            WTFMove(*type)
        }
    };
}

void ArgumentCoder<WebCore::IDBIndexInfo>::encode(Encoder& encoder, const WebCore::IDBIndexInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.objectStoreIdentifier())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyPath())>, WebCore::IDBKeyPath>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.unique())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multiEntry())>, bool>);
    encoder << instance.identifier();
    encoder << instance.objectStoreIdentifier();
    encoder << instance.name();
    encoder << instance.keyPath();
    encoder << instance.unique();
    encoder << instance.multiEntry();
}

std::optional<WebCore::IDBIndexInfo> ArgumentCoder<WebCore::IDBIndexInfo>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<uint64_t>();
    auto objectStoreIdentifier = decoder.decode<uint64_t>();
    auto name = decoder.decode<String>();
    auto keyPath = decoder.decode<WebCore::IDBKeyPath>();
    auto unique = decoder.decode<bool>();
    auto multiEntry = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBIndexInfo {
            WTFMove(*identifier),
            WTFMove(*objectStoreIdentifier),
            WTFMove(*name),
            WTFMove(*keyPath),
            WTFMove(*unique),
            WTFMove(*multiEntry)
        }
    };
}

void ArgumentCoder<WebCore::IDBObjectStoreInfo>::encode(Encoder& encoder, const WebCore::IDBObjectStoreInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyPath())>, std::optional<WebCore::IDBKeyPath>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.autoIncrement())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.indexMap())>, HashMap<uint64_t, WebCore::IDBIndexInfo>>);
    encoder << instance.identifier();
    encoder << instance.name();
    encoder << instance.keyPath();
    encoder << instance.autoIncrement();
    encoder << instance.indexMap();
}

std::optional<WebCore::IDBObjectStoreInfo> ArgumentCoder<WebCore::IDBObjectStoreInfo>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<uint64_t>();
    auto name = decoder.decode<String>();
    auto keyPath = decoder.decode<std::optional<WebCore::IDBKeyPath>>();
    auto autoIncrement = decoder.decode<bool>();
    auto indexMap = decoder.decode<HashMap<uint64_t, WebCore::IDBIndexInfo>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBObjectStoreInfo {
            WTFMove(*identifier),
            WTFMove(*name),
            WTFMove(*keyPath),
            WTFMove(*autoIncrement),
            WTFMove(*indexMap)
        }
    };
}

void ArgumentCoder<WebCore::IDBIterateCursorData>::encode(Encoder& encoder, const WebCore::IDBIterateCursorData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyData)>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primaryKeyData)>, WebCore::IDBKeyData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.count)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.option)>, WebCore::IndexedDB::CursorIterateOption>);
    struct ShouldBeSameSizeAsIDBIterateCursorData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBIterateCursorData>, false> {
        WebCore::IDBKeyData keyData;
        WebCore::IDBKeyData primaryKeyData;
        unsigned count;
        WebCore::IndexedDB::CursorIterateOption option;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBIterateCursorData) == sizeof(WebCore::IDBIterateCursorData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBIterateCursorData, keyData)
        , offsetof(WebCore::IDBIterateCursorData, primaryKeyData)
        , offsetof(WebCore::IDBIterateCursorData, count)
        , offsetof(WebCore::IDBIterateCursorData, option)
    >::value);
    encoder << instance.keyData;
    encoder << instance.primaryKeyData;
    encoder << instance.count;
    encoder << instance.option;
}

std::optional<WebCore::IDBIterateCursorData> ArgumentCoder<WebCore::IDBIterateCursorData>::decode(Decoder& decoder)
{
    auto keyData = decoder.decode<WebCore::IDBKeyData>();
    auto primaryKeyData = decoder.decode<WebCore::IDBKeyData>();
    auto count = decoder.decode<unsigned>();
    auto option = decoder.decode<WebCore::IndexedDB::CursorIterateOption>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBIterateCursorData {
            WTFMove(*keyData),
            WTFMove(*primaryKeyData),
            WTFMove(*count),
            WTFMove(*option)
        }
    };
}

void ArgumentCoder<WebCore::IDBResourceIdentifier>::encode(Encoder& encoder, const WebCore::IDBResourceIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_idbConnectionIdentifier)>, WebCore::IDBConnectionIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_resourceNumber)>, uint64_t>);
    struct ShouldBeSameSizeAsIDBResourceIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBResourceIdentifier>, false> {
        WebCore::IDBConnectionIdentifier m_idbConnectionIdentifier;
        uint64_t m_resourceNumber;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBResourceIdentifier) == sizeof(WebCore::IDBResourceIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBResourceIdentifier, m_idbConnectionIdentifier)
        , offsetof(WebCore::IDBResourceIdentifier, m_resourceNumber)
    >::value);
    encoder << instance.m_idbConnectionIdentifier;
    encoder << instance.m_resourceNumber;
}

std::optional<WebCore::IDBResourceIdentifier> ArgumentCoder<WebCore::IDBResourceIdentifier>::decode(Decoder& decoder)
{
    auto m_idbConnectionIdentifier = decoder.decode<WebCore::IDBConnectionIdentifier>();
    auto m_resourceNumber = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBResourceIdentifier {
            WTFMove(*m_idbConnectionIdentifier),
            WTFMove(*m_resourceNumber)
        }
    };
}

void ArgumentCoder<WebCore::IDBValue>::encode(Encoder& encoder, const WebCore::IDBValue& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, WebCore::ThreadSafeDataBuffer>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blobURLs())>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blobFilePaths())>, Vector<String>>);
    encoder << instance.data();
    encoder << instance.blobURLs();
    encoder << instance.blobFilePaths();
}

std::optional<WebCore::IDBValue> ArgumentCoder<WebCore::IDBValue>::decode(Decoder& decoder)
{
    auto data = decoder.decode<WebCore::ThreadSafeDataBuffer>();
    auto blobURLs = decoder.decode<Vector<String>>();
    auto blobFilePaths = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBValue {
            WTFMove(*data),
            WTFMove(*blobURLs),
            WTFMove(*blobFilePaths)
        }
    };
}

void ArgumentCoder<WebCore::IDBRequestData>::encode(Encoder& encoder, const WebCore::IDBRequestData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_serverConnectionIdentifier)>, WebCore::IDBConnectionIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_requestIdentifier)>, WebCore::IDBResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_transactionIdentifier)>, std::optional<WebCore::IDBResourceIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_cursorIdentifier)>, std::optional<WebCore::IDBResourceIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_objectStoreIdentifier)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_indexIdentifier)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_indexRecordType)>, WebCore::IndexedDB::IndexRecordType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_databaseIdentifier)>, std::optional<WebCore::IDBDatabaseIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_requestedVersion)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_requestType)>, WebCore::IndexedDB::RequestType>);
    struct ShouldBeSameSizeAsIDBRequestData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBRequestData>, false> {
        WebCore::IDBConnectionIdentifier m_serverConnectionIdentifier;
        WebCore::IDBResourceIdentifier m_requestIdentifier;
        std::optional<WebCore::IDBResourceIdentifier> m_transactionIdentifier;
        std::optional<WebCore::IDBResourceIdentifier> m_cursorIdentifier;
        uint64_t m_objectStoreIdentifier;
        uint64_t m_indexIdentifier;
        WebCore::IndexedDB::IndexRecordType m_indexRecordType;
        std::optional<WebCore::IDBDatabaseIdentifier> m_databaseIdentifier;
        uint64_t m_requestedVersion;
        WebCore::IndexedDB::RequestType m_requestType;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBRequestData) == sizeof(WebCore::IDBRequestData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBRequestData, m_serverConnectionIdentifier)
        , offsetof(WebCore::IDBRequestData, m_requestIdentifier)
        , offsetof(WebCore::IDBRequestData, m_transactionIdentifier)
        , offsetof(WebCore::IDBRequestData, m_cursorIdentifier)
        , offsetof(WebCore::IDBRequestData, m_objectStoreIdentifier)
        , offsetof(WebCore::IDBRequestData, m_indexIdentifier)
        , offsetof(WebCore::IDBRequestData, m_indexRecordType)
        , offsetof(WebCore::IDBRequestData, m_databaseIdentifier)
        , offsetof(WebCore::IDBRequestData, m_requestedVersion)
        , offsetof(WebCore::IDBRequestData, m_requestType)
    >::value);
    encoder << instance.m_serverConnectionIdentifier;
    encoder << instance.m_requestIdentifier;
    encoder << instance.m_transactionIdentifier;
    encoder << instance.m_cursorIdentifier;
    encoder << instance.m_objectStoreIdentifier;
    encoder << instance.m_indexIdentifier;
    encoder << instance.m_indexRecordType;
    encoder << instance.m_databaseIdentifier;
    encoder << instance.m_requestedVersion;
    encoder << instance.m_requestType;
}

std::optional<WebCore::IDBRequestData> ArgumentCoder<WebCore::IDBRequestData>::decode(Decoder& decoder)
{
    auto m_serverConnectionIdentifier = decoder.decode<WebCore::IDBConnectionIdentifier>();
    auto m_requestIdentifier = decoder.decode<WebCore::IDBResourceIdentifier>();
    auto m_transactionIdentifier = decoder.decode<std::optional<WebCore::IDBResourceIdentifier>>();
    auto m_cursorIdentifier = decoder.decode<std::optional<WebCore::IDBResourceIdentifier>>();
    auto m_objectStoreIdentifier = decoder.decode<uint64_t>();
    auto m_indexIdentifier = decoder.decode<uint64_t>();
    auto m_indexRecordType = decoder.decode<WebCore::IndexedDB::IndexRecordType>();
    auto m_databaseIdentifier = decoder.decode<std::optional<WebCore::IDBDatabaseIdentifier>>();
    auto m_requestedVersion = decoder.decode<uint64_t>();
    auto m_requestType = decoder.decode<WebCore::IndexedDB::RequestType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBRequestData {
            WTFMove(*m_serverConnectionIdentifier),
            WTFMove(*m_requestIdentifier),
            WTFMove(*m_transactionIdentifier),
            WTFMove(*m_cursorIdentifier),
            WTFMove(*m_objectStoreIdentifier),
            WTFMove(*m_indexIdentifier),
            WTFMove(*m_indexRecordType),
            WTFMove(*m_databaseIdentifier),
            WTFMove(*m_requestedVersion),
            WTFMove(*m_requestType)
        }
    };
}

void ArgumentCoder<WebCore::IDBDatabaseIdentifier>::encode(Encoder& encoder, const WebCore::IDBDatabaseIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_databaseName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_origin)>, WebCore::ClientOrigin>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isTransient)>, bool>);
    struct ShouldBeSameSizeAsIDBDatabaseIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBDatabaseIdentifier>, false> {
        String m_databaseName;
        WebCore::ClientOrigin m_origin;
        bool m_isTransient;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBDatabaseIdentifier) == sizeof(WebCore::IDBDatabaseIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBDatabaseIdentifier, m_databaseName)
        , offsetof(WebCore::IDBDatabaseIdentifier, m_origin)
        , offsetof(WebCore::IDBDatabaseIdentifier, m_isTransient)
    >::value);
    encoder << instance.m_databaseName;
    encoder << instance.m_origin;
    encoder << instance.m_isTransient;
}

std::optional<WebCore::IDBDatabaseIdentifier> ArgumentCoder<WebCore::IDBDatabaseIdentifier>::decode(Decoder& decoder)
{
    auto m_databaseName = decoder.decode<String>();
    auto m_origin = decoder.decode<WebCore::ClientOrigin>();
    auto m_isTransient = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    WebCore::IDBDatabaseIdentifier result;
    result.m_databaseName = WTFMove(*m_databaseName);
    result.m_origin = WTFMove(*m_origin);
    result.m_isTransient = WTFMove(*m_isTransient);
    return { WTFMove(result) };
}

void ArgumentCoder<WebCore::IDBDatabaseNameAndVersion>::encode(Encoder& encoder, const WebCore::IDBDatabaseNameAndVersion& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.version)>, uint64_t>);
    struct ShouldBeSameSizeAsIDBDatabaseNameAndVersion : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBDatabaseNameAndVersion>, false> {
        String name;
        uint64_t version;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBDatabaseNameAndVersion) == sizeof(WebCore::IDBDatabaseNameAndVersion));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBDatabaseNameAndVersion, name)
        , offsetof(WebCore::IDBDatabaseNameAndVersion, version)
    >::value);
    encoder << instance.name;
    encoder << instance.version;
}

std::optional<WebCore::IDBDatabaseNameAndVersion> ArgumentCoder<WebCore::IDBDatabaseNameAndVersion>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto version = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBDatabaseNameAndVersion {
            WTFMove(*name),
            WTFMove(*version)
        }
    };
}

void ArgumentCoder<WebCore::IDBResultData>::encode(Encoder& encoder, const WebCore::IDBResultData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_type)>, WebCore::IDBResultType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_requestIdentifier)>, WebCore::IDBResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_error)>, WebCore::IDBError>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_databaseConnectionIdentifier)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_databaseInfo)>, std::unique_ptr<WebCore::IDBDatabaseInfo>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_transactionInfo)>, std::unique_ptr<WebCore::IDBTransactionInfo>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_resultKey)>, std::unique_ptr<WebCore::IDBKeyData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_getResult)>, std::unique_ptr<WebCore::IDBGetResult>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_getAllResult)>, std::unique_ptr<WebCore::IDBGetAllResult>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_resultInteger)>, uint64_t>);
    struct ShouldBeSameSizeAsIDBResultData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBResultData>, false> {
        WebCore::IDBResultType m_type;
        WebCore::IDBResourceIdentifier m_requestIdentifier;
        WebCore::IDBError m_error;
        uint64_t m_databaseConnectionIdentifier;
        std::unique_ptr<WebCore::IDBDatabaseInfo> m_databaseInfo;
        std::unique_ptr<WebCore::IDBTransactionInfo> m_transactionInfo;
        std::unique_ptr<WebCore::IDBKeyData> m_resultKey;
        std::unique_ptr<WebCore::IDBGetResult> m_getResult;
        std::unique_ptr<WebCore::IDBGetAllResult> m_getAllResult;
        uint64_t m_resultInteger;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIDBResultData) == sizeof(WebCore::IDBResultData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBResultData, m_type)
        , offsetof(WebCore::IDBResultData, m_requestIdentifier)
        , offsetof(WebCore::IDBResultData, m_error)
        , offsetof(WebCore::IDBResultData, m_databaseConnectionIdentifier)
        , offsetof(WebCore::IDBResultData, m_databaseInfo)
        , offsetof(WebCore::IDBResultData, m_transactionInfo)
        , offsetof(WebCore::IDBResultData, m_resultKey)
        , offsetof(WebCore::IDBResultData, m_getResult)
        , offsetof(WebCore::IDBResultData, m_getAllResult)
        , offsetof(WebCore::IDBResultData, m_resultInteger)
    >::value);
    encoder << instance.m_type;
    encoder << instance.m_requestIdentifier;
    encoder << instance.m_error;
    encoder << instance.m_databaseConnectionIdentifier;
    encoder << instance.m_databaseInfo;
    encoder << instance.m_transactionInfo;
    encoder << instance.m_resultKey;
    encoder << instance.m_getResult;
    encoder << instance.m_getAllResult;
    encoder << instance.m_resultInteger;
}

std::optional<WebCore::IDBResultData> ArgumentCoder<WebCore::IDBResultData>::decode(Decoder& decoder)
{
    auto m_type = decoder.decode<WebCore::IDBResultType>();
    auto m_requestIdentifier = decoder.decode<WebCore::IDBResourceIdentifier>();
    auto m_error = decoder.decode<WebCore::IDBError>();
    auto m_databaseConnectionIdentifier = decoder.decode<uint64_t>();
    auto m_databaseInfo = decoder.decode<std::unique_ptr<WebCore::IDBDatabaseInfo>>();
    auto m_transactionInfo = decoder.decode<std::unique_ptr<WebCore::IDBTransactionInfo>>();
    auto m_resultKey = decoder.decode<std::unique_ptr<WebCore::IDBKeyData>>();
    auto m_getResult = decoder.decode<std::unique_ptr<WebCore::IDBGetResult>>();
    auto m_getAllResult = decoder.decode<std::unique_ptr<WebCore::IDBGetAllResult>>();
    auto m_resultInteger = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    WebCore::IDBResultData result;
    result.m_type = WTFMove(*m_type);
    result.m_requestIdentifier = WTFMove(*m_requestIdentifier);
    result.m_error = WTFMove(*m_error);
    result.m_databaseConnectionIdentifier = WTFMove(*m_databaseConnectionIdentifier);
    result.m_databaseInfo = WTFMove(*m_databaseInfo);
    result.m_transactionInfo = WTFMove(*m_transactionInfo);
    result.m_resultKey = WTFMove(*m_resultKey);
    result.m_getResult = WTFMove(*m_getResult);
    result.m_getAllResult = WTFMove(*m_getAllResult);
    result.m_resultInteger = WTFMove(*m_resultInteger);
    return { WTFMove(result) };
}

void ArgumentCoder<WebCore::IDBKeyData>::encode(Encoder& encoder, const WebCore::IDBKeyData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, std::variant<std::nullptr_t, WebCore::IDBKeyData::Invalid, Vector<WebCore::IDBKeyData>, String, double, WebCore::IDBKeyData::Date, WebCore::ThreadSafeDataBuffer, WebCore::IDBKeyData::Min, WebCore::IDBKeyData::Max>>);
    encoder << instance.value();
}

std::optional<WebCore::IDBKeyData> ArgumentCoder<WebCore::IDBKeyData>::decode(Decoder& decoder)
{
    auto value = decoder.decode<std::variant<std::nullptr_t, WebCore::IDBKeyData::Invalid, Vector<WebCore::IDBKeyData>, String, double, WebCore::IDBKeyData::Date, WebCore::ThreadSafeDataBuffer, WebCore::IDBKeyData::Min, WebCore::IDBKeyData::Max>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBKeyData {
            WTFMove(*value)
        }
    };
}

void ArgumentCoder<WebCore::IDBKeyData::Invalid>::encode(Encoder& encoder, const WebCore::IDBKeyData::Invalid& instance)
{
    struct ShouldBeSameSizeAsInvalid : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBKeyData::Invalid>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsInvalid) == sizeof(WebCore::IDBKeyData::Invalid));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::IDBKeyData::Invalid> ArgumentCoder<WebCore::IDBKeyData::Invalid>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBKeyData::Invalid {
        }
    };
}

void ArgumentCoder<WebCore::IDBKeyData::Min>::encode(Encoder& encoder, const WebCore::IDBKeyData::Min& instance)
{
    struct ShouldBeSameSizeAsMin : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBKeyData::Min>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsMin) == sizeof(WebCore::IDBKeyData::Min));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::IDBKeyData::Min> ArgumentCoder<WebCore::IDBKeyData::Min>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBKeyData::Min {
        }
    };
}

void ArgumentCoder<WebCore::IDBKeyData::Max>::encode(Encoder& encoder, const WebCore::IDBKeyData::Max& instance)
{
    struct ShouldBeSameSizeAsMax : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBKeyData::Max>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsMax) == sizeof(WebCore::IDBKeyData::Max));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::IDBKeyData::Max> ArgumentCoder<WebCore::IDBKeyData::Max>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBKeyData::Max {
        }
    };
}

void ArgumentCoder<WebCore::IDBKeyData::Date>::encode(Encoder& encoder, const WebCore::IDBKeyData::Date& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, double>);
    struct ShouldBeSameSizeAsDate : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::IDBKeyData::Date>, false> {
        double value;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDate) == sizeof(WebCore::IDBKeyData::Date));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IDBKeyData::Date, value)
    >::value);
    encoder << instance.value;
}

std::optional<WebCore::IDBKeyData::Date> ArgumentCoder<WebCore::IDBKeyData::Date>::decode(Decoder& decoder)
{
    auto value = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IDBKeyData::Date {
            WTFMove(*value)
        }
    };
}

#if USE(CG)
void ArgumentCoder<CGSize>::encode(Encoder& encoder, const CGSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, CGFloat>);
    struct ShouldBeSameSizeAsCGSize : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<CGSize>, false> {
        CGFloat width;
        CGFloat height;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCGSize) == sizeof(CGSize));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(CGSize, width)
        , offsetof(CGSize, height)
    >::value);
    encoder << instance.width;
    encoder << instance.height;
}

void ArgumentCoder<CGSize>::encode(StreamConnectionEncoder& encoder, const CGSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, CGFloat>);
    struct ShouldBeSameSizeAsCGSize : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<CGSize>, false> {
        CGFloat width;
        CGFloat height;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCGSize) == sizeof(CGSize));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(CGSize, width)
        , offsetof(CGSize, height)
    >::value);
    encoder << instance.width;
    encoder << instance.height;
}

std::optional<CGSize> ArgumentCoder<CGSize>::decode(Decoder& decoder)
{
    auto width = decoder.decode<CGFloat>();
    auto height = decoder.decode<CGFloat>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        CGSize {
            WTFMove(*width),
            WTFMove(*height)
        }
    };
}

#endif

#if USE(CG)
void ArgumentCoder<CGPoint>::encode(Encoder& encoder, const CGPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y)>, CGFloat>);
    struct ShouldBeSameSizeAsCGPoint : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<CGPoint>, false> {
        CGFloat x;
        CGFloat y;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCGPoint) == sizeof(CGPoint));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(CGPoint, x)
        , offsetof(CGPoint, y)
    >::value);
    encoder << instance.x;
    encoder << instance.y;
}

std::optional<CGPoint> ArgumentCoder<CGPoint>::decode(Decoder& decoder)
{
    auto x = decoder.decode<CGFloat>();
    auto y = decoder.decode<CGFloat>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        CGPoint {
            WTFMove(*x),
            WTFMove(*y)
        }
    };
}

#endif

#if USE(CG)
void ArgumentCoder<CGRect>::encode(Encoder& encoder, const CGRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, CGPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, CGSize>);
    struct ShouldBeSameSizeAsCGRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<CGRect>, false> {
        CGPoint origin;
        CGSize size;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCGRect) == sizeof(CGRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(CGRect, origin)
        , offsetof(CGRect, size)
    >::value);
    encoder << instance.origin;
    encoder << instance.size;
}

std::optional<CGRect> ArgumentCoder<CGRect>::decode(Decoder& decoder)
{
    auto origin = decoder.decode<CGPoint>();
    auto size = decoder.decode<CGSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        CGRect {
            WTFMove(*origin),
            WTFMove(*size)
        }
    };
}

#endif

#if USE(CG)
void ArgumentCoder<CGAffineTransform>::encode(Encoder& encoder, const CGAffineTransform& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.a)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.b)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.c)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.d)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tx)>, CGFloat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ty)>, CGFloat>);
    struct ShouldBeSameSizeAsCGAffineTransform : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<CGAffineTransform>, false> {
        CGFloat a;
        CGFloat b;
        CGFloat c;
        CGFloat d;
        CGFloat tx;
        CGFloat ty;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCGAffineTransform) == sizeof(CGAffineTransform));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(CGAffineTransform, a)
        , offsetof(CGAffineTransform, b)
        , offsetof(CGAffineTransform, c)
        , offsetof(CGAffineTransform, d)
        , offsetof(CGAffineTransform, tx)
        , offsetof(CGAffineTransform, ty)
    >::value);
    encoder << instance.a;
    encoder << instance.b;
    encoder << instance.c;
    encoder << instance.d;
    encoder << instance.tx;
    encoder << instance.ty;
}

std::optional<CGAffineTransform> ArgumentCoder<CGAffineTransform>::decode(Decoder& decoder)
{
    auto a = decoder.decode<CGFloat>();
    auto b = decoder.decode<CGFloat>();
    auto c = decoder.decode<CGFloat>();
    auto d = decoder.decode<CGFloat>();
    auto tx = decoder.decode<CGFloat>();
    auto ty = decoder.decode<CGFloat>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        CGAffineTransform {
            WTFMove(*a),
            WTFMove(*b),
            WTFMove(*c),
            WTFMove(*d),
            WTFMove(*tx),
            WTFMove(*ty)
        }
    };
}

#endif

void ArgumentCoder<WebCore::FloatRect>::encode(Encoder& encoder, const WebCore::FloatRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size())>, WebCore::FloatSize>);
    encoder << instance.location();
    encoder << instance.size();
}

void ArgumentCoder<WebCore::FloatRect>::encode(StreamConnectionEncoder& encoder, const WebCore::FloatRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size())>, WebCore::FloatSize>);
    encoder << instance.location();
    encoder << instance.size();
}

std::optional<WebCore::FloatRect> ArgumentCoder<WebCore::FloatRect>::decode(Decoder& decoder)
{
    auto location = decoder.decode<WebCore::FloatPoint>();
    auto size = decoder.decode<WebCore::FloatSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatRect {
            WTFMove(*location),
            WTFMove(*size)
        }
    };
}

void ArgumentCoder<WebCore::LinearTimingFunction::Point>::encode(Encoder& encoder, const WebCore::LinearTimingFunction::Point& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.progress)>, double>);
    struct ShouldBeSameSizeAsPoint : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::LinearTimingFunction::Point>, false> {
        double value;
        double progress;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPoint) == sizeof(WebCore::LinearTimingFunction::Point));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::LinearTimingFunction::Point, value)
        , offsetof(WebCore::LinearTimingFunction::Point, progress)
    >::value);
    encoder << instance.value;
    encoder << instance.progress;
}

std::optional<WebCore::LinearTimingFunction::Point> ArgumentCoder<WebCore::LinearTimingFunction::Point>::decode(Decoder& decoder)
{
    auto value = decoder.decode<double>();
    auto progress = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LinearTimingFunction::Point {
            WTFMove(*value),
            WTFMove(*progress)
        }
    };
}

void ArgumentCoder<WebCore::LinearTimingFunction>::encode(Encoder& encoder, const WebCore::LinearTimingFunction& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.points())>, Vector<WebCore::LinearTimingFunction::Point>>);
    encoder << instance.points();
}

std::optional<Ref<WebCore::LinearTimingFunction>> ArgumentCoder<WebCore::LinearTimingFunction>::decode(Decoder& decoder)
{
    auto points = decoder.decode<Vector<WebCore::LinearTimingFunction::Point>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LinearTimingFunction::create(
            WTFMove(*points)
        )
    };
}

void ArgumentCoder<WebCore::CubicBezierTimingFunction>::encode(Encoder& encoder, const WebCore::CubicBezierTimingFunction& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timingFunctionPreset())>, WebCore::CubicBezierTimingFunction::TimingFunctionPreset>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x1())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y1())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x2())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y2())>, double>);
    encoder << instance.timingFunctionPreset();
    encoder << instance.x1();
    encoder << instance.y1();
    encoder << instance.x2();
    encoder << instance.y2();
}

std::optional<Ref<WebCore::CubicBezierTimingFunction>> ArgumentCoder<WebCore::CubicBezierTimingFunction>::decode(Decoder& decoder)
{
    auto timingFunctionPreset = decoder.decode<WebCore::CubicBezierTimingFunction::TimingFunctionPreset>();
    auto x1 = decoder.decode<double>();
    auto y1 = decoder.decode<double>();
    auto x2 = decoder.decode<double>();
    auto y2 = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CubicBezierTimingFunction::create(
            WTFMove(*timingFunctionPreset),
            WTFMove(*x1),
            WTFMove(*y1),
            WTFMove(*x2),
            WTFMove(*y2)
        )
    };
}

void ArgumentCoder<WebCore::StepsTimingFunction>::encode(Encoder& encoder, const WebCore::StepsTimingFunction& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.numberOfSteps())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stepPosition())>, std::optional<WebCore::StepsTimingFunction::StepPosition>>);
    encoder << instance.numberOfSteps();
    encoder << instance.stepPosition();
}

std::optional<Ref<WebCore::StepsTimingFunction>> ArgumentCoder<WebCore::StepsTimingFunction>::decode(Decoder& decoder)
{
    auto numberOfSteps = decoder.decode<int>();
    auto stepPosition = decoder.decode<std::optional<WebCore::StepsTimingFunction::StepPosition>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::StepsTimingFunction::create(
            WTFMove(*numberOfSteps),
            WTFMove(*stepPosition)
        )
    };
}

void ArgumentCoder<WebCore::SpringTimingFunction>::encode(Encoder& encoder, const WebCore::SpringTimingFunction& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mass())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stiffness())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.damping())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.initialVelocity())>, double>);
    encoder << instance.mass();
    encoder << instance.stiffness();
    encoder << instance.damping();
    encoder << instance.initialVelocity();
}

std::optional<Ref<WebCore::SpringTimingFunction>> ArgumentCoder<WebCore::SpringTimingFunction>::decode(Decoder& decoder)
{
    auto mass = decoder.decode<double>();
    auto stiffness = decoder.decode<double>();
    auto damping = decoder.decode<double>();
    auto initialVelocity = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SpringTimingFunction::create(
            WTFMove(*mass),
            WTFMove(*stiffness),
            WTFMove(*damping),
            WTFMove(*initialVelocity)
        )
    };
}

void ArgumentCoder<WebCore::ResourceLoadStatistics>::encode(Encoder& encoder, const WebCore::ResourceLoadStatistics& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrableDomain)>, WebCore::RegistrableDomain>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastSeen)>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hadUserInteraction)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mostRecentUserInteractionTime)>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.grandfathered)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.storageAccessUnderTopFrameDomains)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topFrameUniqueRedirectsTo)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topFrameUniqueRedirectsToSinceSameSiteStrictEnforcement)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topFrameUniqueRedirectsFrom)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topFrameLinkDecorationsFrom)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gotLinkDecorationFromPrevalentResource)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topFrameLoadedThirdPartyScripts)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subframeUnderTopFrameDomains)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subresourceUnderTopFrameDomains)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subresourceUniqueRedirectsTo)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subresourceUniqueRedirectsFrom)>, HashSet<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isPrevalentResource)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isVeryPrevalentResource)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataRecordsRemoved)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timesAccessedAsFirstPartyDueToUserInteraction)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timesAccessedAsFirstPartyDueToStorageAccessAPI)>, unsigned>);
#if ENABLE(WEB_API_STATISTICS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topFrameRegistrableDomainsWhichAccessedWebAPIs)>, HashSet<WebCore::RegistrableDomain>>);
#endif
#if ENABLE(WEB_API_STATISTICS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontsFailedToLoad)>, HashSet<String>>);
#endif
#if ENABLE(WEB_API_STATISTICS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontsSuccessfullyLoaded)>, HashSet<String>>);
#endif
#if ENABLE(WEB_API_STATISTICS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canvasActivityRecord)>, WebCore::CanvasActivityRecord>);
#endif
#if ENABLE(WEB_API_STATISTICS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.navigatorFunctionsAccessed)>, OptionSet<WebCore::NavigatorAPIsAccessed>>);
#endif
#if ENABLE(WEB_API_STATISTICS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenFunctionsAccessed)>, OptionSet<WebCore::ScreenAPIsAccessed>>);
#endif
    struct ShouldBeSameSizeAsResourceLoadStatistics : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ResourceLoadStatistics>, false> {
        WebCore::RegistrableDomain registrableDomain;
        WallTime lastSeen;
        bool hadUserInteraction;
        WallTime mostRecentUserInteractionTime;
        bool grandfathered;
        HashSet<WebCore::RegistrableDomain> storageAccessUnderTopFrameDomains;
        HashSet<WebCore::RegistrableDomain> topFrameUniqueRedirectsTo;
        HashSet<WebCore::RegistrableDomain> topFrameUniqueRedirectsToSinceSameSiteStrictEnforcement;
        HashSet<WebCore::RegistrableDomain> topFrameUniqueRedirectsFrom;
        HashSet<WebCore::RegistrableDomain> topFrameLinkDecorationsFrom;
        bool gotLinkDecorationFromPrevalentResource;
        HashSet<WebCore::RegistrableDomain> topFrameLoadedThirdPartyScripts;
        HashSet<WebCore::RegistrableDomain> subframeUnderTopFrameDomains;
        HashSet<WebCore::RegistrableDomain> subresourceUnderTopFrameDomains;
        HashSet<WebCore::RegistrableDomain> subresourceUniqueRedirectsTo;
        HashSet<WebCore::RegistrableDomain> subresourceUniqueRedirectsFrom;
        bool isPrevalentResource;
        bool isVeryPrevalentResource;
        unsigned dataRecordsRemoved;
        unsigned timesAccessedAsFirstPartyDueToUserInteraction;
        unsigned timesAccessedAsFirstPartyDueToStorageAccessAPI;
#if ENABLE(WEB_API_STATISTICS)
        HashSet<WebCore::RegistrableDomain> topFrameRegistrableDomainsWhichAccessedWebAPIs;
#endif
#if ENABLE(WEB_API_STATISTICS)
        HashSet<String> fontsFailedToLoad;
#endif
#if ENABLE(WEB_API_STATISTICS)
        HashSet<String> fontsSuccessfullyLoaded;
#endif
#if ENABLE(WEB_API_STATISTICS)
        WebCore::CanvasActivityRecord canvasActivityRecord;
#endif
#if ENABLE(WEB_API_STATISTICS)
        OptionSet<WebCore::NavigatorAPIsAccessed> navigatorFunctionsAccessed;
#endif
#if ENABLE(WEB_API_STATISTICS)
        OptionSet<WebCore::ScreenAPIsAccessed> screenFunctionsAccessed;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsResourceLoadStatistics) == sizeof(WebCore::ResourceLoadStatistics));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ResourceLoadStatistics, registrableDomain)
        , offsetof(WebCore::ResourceLoadStatistics, lastSeen)
        , offsetof(WebCore::ResourceLoadStatistics, hadUserInteraction)
        , offsetof(WebCore::ResourceLoadStatistics, mostRecentUserInteractionTime)
        , offsetof(WebCore::ResourceLoadStatistics, grandfathered)
        , offsetof(WebCore::ResourceLoadStatistics, storageAccessUnderTopFrameDomains)
        , offsetof(WebCore::ResourceLoadStatistics, topFrameUniqueRedirectsTo)
        , offsetof(WebCore::ResourceLoadStatistics, topFrameUniqueRedirectsToSinceSameSiteStrictEnforcement)
        , offsetof(WebCore::ResourceLoadStatistics, topFrameUniqueRedirectsFrom)
        , offsetof(WebCore::ResourceLoadStatistics, topFrameLinkDecorationsFrom)
        , offsetof(WebCore::ResourceLoadStatistics, gotLinkDecorationFromPrevalentResource)
        , offsetof(WebCore::ResourceLoadStatistics, topFrameLoadedThirdPartyScripts)
        , offsetof(WebCore::ResourceLoadStatistics, subframeUnderTopFrameDomains)
        , offsetof(WebCore::ResourceLoadStatistics, subresourceUnderTopFrameDomains)
        , offsetof(WebCore::ResourceLoadStatistics, subresourceUniqueRedirectsTo)
        , offsetof(WebCore::ResourceLoadStatistics, subresourceUniqueRedirectsFrom)
        , offsetof(WebCore::ResourceLoadStatistics, isPrevalentResource)
        , offsetof(WebCore::ResourceLoadStatistics, isVeryPrevalentResource)
        , offsetof(WebCore::ResourceLoadStatistics, dataRecordsRemoved)
        , offsetof(WebCore::ResourceLoadStatistics, timesAccessedAsFirstPartyDueToUserInteraction)
        , offsetof(WebCore::ResourceLoadStatistics, timesAccessedAsFirstPartyDueToStorageAccessAPI)
#if ENABLE(WEB_API_STATISTICS)
        , offsetof(WebCore::ResourceLoadStatistics, topFrameRegistrableDomainsWhichAccessedWebAPIs)
#endif
#if ENABLE(WEB_API_STATISTICS)
        , offsetof(WebCore::ResourceLoadStatistics, fontsFailedToLoad)
#endif
#if ENABLE(WEB_API_STATISTICS)
        , offsetof(WebCore::ResourceLoadStatistics, fontsSuccessfullyLoaded)
#endif
#if ENABLE(WEB_API_STATISTICS)
        , offsetof(WebCore::ResourceLoadStatistics, canvasActivityRecord)
#endif
#if ENABLE(WEB_API_STATISTICS)
        , offsetof(WebCore::ResourceLoadStatistics, navigatorFunctionsAccessed)
#endif
#if ENABLE(WEB_API_STATISTICS)
        , offsetof(WebCore::ResourceLoadStatistics, screenFunctionsAccessed)
#endif
    >::value);
    encoder << instance.registrableDomain;
    encoder << instance.lastSeen;
    encoder << instance.hadUserInteraction;
    encoder << instance.mostRecentUserInteractionTime;
    encoder << instance.grandfathered;
    encoder << instance.storageAccessUnderTopFrameDomains;
    encoder << instance.topFrameUniqueRedirectsTo;
    encoder << instance.topFrameUniqueRedirectsToSinceSameSiteStrictEnforcement;
    encoder << instance.topFrameUniqueRedirectsFrom;
    encoder << instance.topFrameLinkDecorationsFrom;
    encoder << instance.gotLinkDecorationFromPrevalentResource;
    encoder << instance.topFrameLoadedThirdPartyScripts;
    encoder << instance.subframeUnderTopFrameDomains;
    encoder << instance.subresourceUnderTopFrameDomains;
    encoder << instance.subresourceUniqueRedirectsTo;
    encoder << instance.subresourceUniqueRedirectsFrom;
    encoder << instance.isPrevalentResource;
    encoder << instance.isVeryPrevalentResource;
    encoder << instance.dataRecordsRemoved;
    encoder << instance.timesAccessedAsFirstPartyDueToUserInteraction;
    encoder << instance.timesAccessedAsFirstPartyDueToStorageAccessAPI;
#if ENABLE(WEB_API_STATISTICS)
    encoder << instance.topFrameRegistrableDomainsWhichAccessedWebAPIs;
#endif
#if ENABLE(WEB_API_STATISTICS)
    encoder << instance.fontsFailedToLoad;
#endif
#if ENABLE(WEB_API_STATISTICS)
    encoder << instance.fontsSuccessfullyLoaded;
#endif
#if ENABLE(WEB_API_STATISTICS)
    encoder << instance.canvasActivityRecord;
#endif
#if ENABLE(WEB_API_STATISTICS)
    encoder << instance.navigatorFunctionsAccessed;
#endif
#if ENABLE(WEB_API_STATISTICS)
    encoder << instance.screenFunctionsAccessed;
#endif
}

std::optional<WebCore::ResourceLoadStatistics> ArgumentCoder<WebCore::ResourceLoadStatistics>::decode(Decoder& decoder)
{
    auto registrableDomain = decoder.decode<WebCore::RegistrableDomain>();
    auto lastSeen = decoder.decode<WallTime>();
    auto hadUserInteraction = decoder.decode<bool>();
    auto mostRecentUserInteractionTime = decoder.decode<WallTime>();
    auto grandfathered = decoder.decode<bool>();
    auto storageAccessUnderTopFrameDomains = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto topFrameUniqueRedirectsTo = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto topFrameUniqueRedirectsToSinceSameSiteStrictEnforcement = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto topFrameUniqueRedirectsFrom = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto topFrameLinkDecorationsFrom = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto gotLinkDecorationFromPrevalentResource = decoder.decode<bool>();
    auto topFrameLoadedThirdPartyScripts = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto subframeUnderTopFrameDomains = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto subresourceUnderTopFrameDomains = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto subresourceUniqueRedirectsTo = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto subresourceUniqueRedirectsFrom = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
    auto isPrevalentResource = decoder.decode<bool>();
    auto isVeryPrevalentResource = decoder.decode<bool>();
    auto dataRecordsRemoved = decoder.decode<unsigned>();
    auto timesAccessedAsFirstPartyDueToUserInteraction = decoder.decode<unsigned>();
    auto timesAccessedAsFirstPartyDueToStorageAccessAPI = decoder.decode<unsigned>();
#if ENABLE(WEB_API_STATISTICS)
    auto topFrameRegistrableDomainsWhichAccessedWebAPIs = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
#endif
#if ENABLE(WEB_API_STATISTICS)
    auto fontsFailedToLoad = decoder.decode<HashSet<String>>();
#endif
#if ENABLE(WEB_API_STATISTICS)
    auto fontsSuccessfullyLoaded = decoder.decode<HashSet<String>>();
#endif
#if ENABLE(WEB_API_STATISTICS)
    auto canvasActivityRecord = decoder.decode<WebCore::CanvasActivityRecord>();
#endif
#if ENABLE(WEB_API_STATISTICS)
    auto navigatorFunctionsAccessed = decoder.decode<OptionSet<WebCore::NavigatorAPIsAccessed>>();
#endif
#if ENABLE(WEB_API_STATISTICS)
    auto screenFunctionsAccessed = decoder.decode<OptionSet<WebCore::ScreenAPIsAccessed>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    WebCore::ResourceLoadStatistics result;
    result.registrableDomain = WTFMove(*registrableDomain);
    result.lastSeen = WTFMove(*lastSeen);
    result.hadUserInteraction = WTFMove(*hadUserInteraction);
    result.mostRecentUserInteractionTime = WTFMove(*mostRecentUserInteractionTime);
    result.grandfathered = WTFMove(*grandfathered);
    result.storageAccessUnderTopFrameDomains = WTFMove(*storageAccessUnderTopFrameDomains);
    result.topFrameUniqueRedirectsTo = WTFMove(*topFrameUniqueRedirectsTo);
    result.topFrameUniqueRedirectsToSinceSameSiteStrictEnforcement = WTFMove(*topFrameUniqueRedirectsToSinceSameSiteStrictEnforcement);
    result.topFrameUniqueRedirectsFrom = WTFMove(*topFrameUniqueRedirectsFrom);
    result.topFrameLinkDecorationsFrom = WTFMove(*topFrameLinkDecorationsFrom);
    result.gotLinkDecorationFromPrevalentResource = WTFMove(*gotLinkDecorationFromPrevalentResource);
    result.topFrameLoadedThirdPartyScripts = WTFMove(*topFrameLoadedThirdPartyScripts);
    result.subframeUnderTopFrameDomains = WTFMove(*subframeUnderTopFrameDomains);
    result.subresourceUnderTopFrameDomains = WTFMove(*subresourceUnderTopFrameDomains);
    result.subresourceUniqueRedirectsTo = WTFMove(*subresourceUniqueRedirectsTo);
    result.subresourceUniqueRedirectsFrom = WTFMove(*subresourceUniqueRedirectsFrom);
    result.isPrevalentResource = WTFMove(*isPrevalentResource);
    result.isVeryPrevalentResource = WTFMove(*isVeryPrevalentResource);
    result.dataRecordsRemoved = WTFMove(*dataRecordsRemoved);
    result.timesAccessedAsFirstPartyDueToUserInteraction = WTFMove(*timesAccessedAsFirstPartyDueToUserInteraction);
    result.timesAccessedAsFirstPartyDueToStorageAccessAPI = WTFMove(*timesAccessedAsFirstPartyDueToStorageAccessAPI);
#if ENABLE(WEB_API_STATISTICS)
    result.topFrameRegistrableDomainsWhichAccessedWebAPIs = WTFMove(*topFrameRegistrableDomainsWhichAccessedWebAPIs);
#endif
#if ENABLE(WEB_API_STATISTICS)
    result.fontsFailedToLoad = WTFMove(*fontsFailedToLoad);
#endif
#if ENABLE(WEB_API_STATISTICS)
    result.fontsSuccessfullyLoaded = WTFMove(*fontsSuccessfullyLoaded);
#endif
#if ENABLE(WEB_API_STATISTICS)
    result.canvasActivityRecord = WTFMove(*canvasActivityRecord);
#endif
#if ENABLE(WEB_API_STATISTICS)
    result.navigatorFunctionsAccessed = WTFMove(*navigatorFunctionsAccessed);
#endif
#if ENABLE(WEB_API_STATISTICS)
    result.screenFunctionsAccessed = WTFMove(*screenFunctionsAccessed);
#endif
    return { WTFMove(result) };
}

#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
void ArgumentCoder<WebCore::GraphicsContextGLActiveInfo>::encode(Encoder& encoder, const WebCore::GraphicsContextGLActiveInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, GCGLenum>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, GCGLint>);
    struct ShouldBeSameSizeAsGraphicsContextGLActiveInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GraphicsContextGLActiveInfo>, false> {
        String name;
        GCGLenum type;
        GCGLint size;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGraphicsContextGLActiveInfo) == sizeof(WebCore::GraphicsContextGLActiveInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GraphicsContextGLActiveInfo, name)
        , offsetof(WebCore::GraphicsContextGLActiveInfo, type)
        , offsetof(WebCore::GraphicsContextGLActiveInfo, size)
    >::value);
    encoder << instance.name;
    encoder << instance.type;
    encoder << instance.size;
}

void ArgumentCoder<WebCore::GraphicsContextGLActiveInfo>::encode(StreamConnectionEncoder& encoder, const WebCore::GraphicsContextGLActiveInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, GCGLenum>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, GCGLint>);
    struct ShouldBeSameSizeAsGraphicsContextGLActiveInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GraphicsContextGLActiveInfo>, false> {
        String name;
        GCGLenum type;
        GCGLint size;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGraphicsContextGLActiveInfo) == sizeof(WebCore::GraphicsContextGLActiveInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GraphicsContextGLActiveInfo, name)
        , offsetof(WebCore::GraphicsContextGLActiveInfo, type)
        , offsetof(WebCore::GraphicsContextGLActiveInfo, size)
    >::value);
    encoder << instance.name;
    encoder << instance.type;
    encoder << instance.size;
}

std::optional<WebCore::GraphicsContextGLActiveInfo> ArgumentCoder<WebCore::GraphicsContextGLActiveInfo>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto type = decoder.decode<GCGLenum>();
    auto size = decoder.decode<GCGLint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GraphicsContextGLActiveInfo {
            WTFMove(*name),
            WTFMove(*type),
            WTFMove(*size)
        }
    };
}

#endif

void ArgumentCoder<WebCore::ViewportAttributes>::encode(Encoder& encoder, const WebCore::ViewportAttributes& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layoutSize)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.initialScale)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimumScale)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumScale)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userScalable)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.orientation)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shrinkToFit)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewportFit)>, WebCore::ViewportFit>);
    struct ShouldBeSameSizeAsViewportAttributes : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ViewportAttributes>, false> {
        WebCore::FloatSize layoutSize;
        float initialScale;
        float minimumScale;
        float maximumScale;
        float userScalable;
        float orientation;
        float shrinkToFit;
        WebCore::ViewportFit viewportFit;
    };
    static_assert(sizeof(ShouldBeSameSizeAsViewportAttributes) == sizeof(WebCore::ViewportAttributes));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ViewportAttributes, layoutSize)
        , offsetof(WebCore::ViewportAttributes, initialScale)
        , offsetof(WebCore::ViewportAttributes, minimumScale)
        , offsetof(WebCore::ViewportAttributes, maximumScale)
        , offsetof(WebCore::ViewportAttributes, userScalable)
        , offsetof(WebCore::ViewportAttributes, orientation)
        , offsetof(WebCore::ViewportAttributes, shrinkToFit)
        , offsetof(WebCore::ViewportAttributes, viewportFit)
    >::value);
    encoder << instance.layoutSize;
    encoder << instance.initialScale;
    encoder << instance.minimumScale;
    encoder << instance.maximumScale;
    encoder << instance.userScalable;
    encoder << instance.orientation;
    encoder << instance.shrinkToFit;
    encoder << instance.viewportFit;
}

std::optional<WebCore::ViewportAttributes> ArgumentCoder<WebCore::ViewportAttributes>::decode(Decoder& decoder)
{
    auto layoutSize = decoder.decode<WebCore::FloatSize>();
    auto initialScale = decoder.decode<float>();
    auto minimumScale = decoder.decode<float>();
    auto maximumScale = decoder.decode<float>();
    auto userScalable = decoder.decode<float>();
    auto orientation = decoder.decode<float>();
    auto shrinkToFit = decoder.decode<float>();
    auto viewportFit = decoder.decode<WebCore::ViewportFit>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ViewportAttributes {
            WTFMove(*layoutSize),
            WTFMove(*initialScale),
            WTFMove(*minimumScale),
            WTFMove(*maximumScale),
            WTFMove(*userScalable),
            WTFMove(*orientation),
            WTFMove(*shrinkToFit),
            WTFMove(*viewportFit)
        }
    };
}

void ArgumentCoder<WebCore::ShareDataWithParsedURL>::encode(Encoder& encoder, const WebCore::ShareDataWithParsedURL& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shareData)>, WebCore::ShareData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, std::optional<URL>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.files)>, Vector<WebCore::RawFile>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originator)>, WebCore::ShareDataOriginator>);
    struct ShouldBeSameSizeAsShareDataWithParsedURL : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShareDataWithParsedURL>, false> {
        WebCore::ShareData shareData;
        std::optional<URL> url;
        Vector<WebCore::RawFile> files;
        WebCore::ShareDataOriginator originator;
    };
    static_assert(sizeof(ShouldBeSameSizeAsShareDataWithParsedURL) == sizeof(WebCore::ShareDataWithParsedURL));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShareDataWithParsedURL, shareData)
        , offsetof(WebCore::ShareDataWithParsedURL, url)
        , offsetof(WebCore::ShareDataWithParsedURL, files)
        , offsetof(WebCore::ShareDataWithParsedURL, originator)
    >::value);
    encoder << instance.shareData;
    encoder << instance.url;
    encoder << instance.files;
    encoder << instance.originator;
}

std::optional<WebCore::ShareDataWithParsedURL> ArgumentCoder<WebCore::ShareDataWithParsedURL>::decode(Decoder& decoder)
{
    auto shareData = decoder.decode<WebCore::ShareData>();
    auto url = decoder.decode<std::optional<URL>>();
    auto files = decoder.decode<Vector<WebCore::RawFile>>();
    auto originator = decoder.decode<WebCore::ShareDataOriginator>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShareDataWithParsedURL {
            WTFMove(*shareData),
            WTFMove(*url),
            WTFMove(*files),
            WTFMove(*originator)
        }
    };
}

void ArgumentCoder<WebCore::ShareData>::encode(Encoder& encoder, const WebCore::ShareData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.title)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.files)>, Vector<RefPtr<WebCore::File>>>);
    struct ShouldBeSameSizeAsShareData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShareData>, false> {
        String title;
        String text;
        String url;
        Vector<RefPtr<WebCore::File>> files;
    };
    static_assert(sizeof(ShouldBeSameSizeAsShareData) == sizeof(WebCore::ShareData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShareData, title)
        , offsetof(WebCore::ShareData, text)
        , offsetof(WebCore::ShareData, url)
        , offsetof(WebCore::ShareData, files)
    >::value);
    encoder << instance.title;
    encoder << instance.text;
    encoder << instance.url;
}

std::optional<WebCore::ShareData> ArgumentCoder<WebCore::ShareData>::decode(Decoder& decoder)
{
    auto title = decoder.decode<String>();
    auto text = decoder.decode<String>();
    auto url = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShareData {
            WTFMove(*title),
            WTFMove(*text),
            WTFMove(*url)
        }
    };
}

void ArgumentCoder<WebCore::RawFile>::encode(Encoder& encoder, const WebCore::RawFile& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fileName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fileData)>, RefPtr<WebCore::SharedBuffer>>);
    struct ShouldBeSameSizeAsRawFile : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RawFile>, false> {
        String fileName;
        RefPtr<WebCore::SharedBuffer> fileData;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRawFile) == sizeof(WebCore::RawFile));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RawFile, fileName)
        , offsetof(WebCore::RawFile, fileData)
    >::value);
    encoder << instance.fileName;
    encoder << instance.fileData;
}

std::optional<WebCore::RawFile> ArgumentCoder<WebCore::RawFile>::decode(Decoder& decoder)
{
    auto fileName = decoder.decode<String>();
    auto fileData = decoder.decode<RefPtr<WebCore::SharedBuffer>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RawFile {
            WTFMove(*fileName),
            WTFMove(*fileData)
        }
    };
}

void ArgumentCoder<WebCore::FloatRoundedRect>::encode(Encoder& encoder, const WebCore::FloatRoundedRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().topLeft())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().topRight())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().bottomLeft())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().bottomRight())>, WebCore::FloatSize>);
    encoder << instance.rect();
    encoder << instance.radii().topLeft();
    encoder << instance.radii().topRight();
    encoder << instance.radii().bottomLeft();
    encoder << instance.radii().bottomRight();
}

void ArgumentCoder<WebCore::FloatRoundedRect>::encode(StreamConnectionEncoder& encoder, const WebCore::FloatRoundedRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().topLeft())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().topRight())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().bottomLeft())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radii().bottomRight())>, WebCore::FloatSize>);
    encoder << instance.rect();
    encoder << instance.radii().topLeft();
    encoder << instance.radii().topRight();
    encoder << instance.radii().bottomLeft();
    encoder << instance.radii().bottomRight();
}

std::optional<WebCore::FloatRoundedRect> ArgumentCoder<WebCore::FloatRoundedRect>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    auto radiitopLeft = decoder.decode<WebCore::FloatSize>();
    auto radiitopRight = decoder.decode<WebCore::FloatSize>();
    auto radiibottomLeft = decoder.decode<WebCore::FloatSize>();
    auto radiibottomRight = decoder.decode<WebCore::FloatSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatRoundedRect {
            WTFMove(*rect),
            WTFMove(*radiitopLeft),
            WTFMove(*radiitopRight),
            WTFMove(*radiibottomLeft),
            WTFMove(*radiibottomRight)
        }
    };
}

void ArgumentCoder<WebCore::IntRect>::encode(Encoder& encoder, const WebCore::IntRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size())>, WebCore::IntSize>);
    encoder << instance.location();
    encoder << instance.size();
}

void ArgumentCoder<WebCore::IntRect>::encode(StreamConnectionEncoder& encoder, const WebCore::IntRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size())>, WebCore::IntSize>);
    encoder << instance.location();
    encoder << instance.size();
}

std::optional<WebCore::IntRect> ArgumentCoder<WebCore::IntRect>::decode(Decoder& decoder)
{
    auto location = decoder.decode<WebCore::IntPoint>();
    auto size = decoder.decode<WebCore::IntSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(WebCore::IntRect { *location, *size }.isValid()))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IntRect {
            WTFMove(*location),
            WTFMove(*size)
        }
    };
}

void ArgumentCoder<WebCore::IntPoint>::encode(Encoder& encoder, const WebCore::IntPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, int>);
    encoder << instance.x();
    encoder << instance.y();
}

void ArgumentCoder<WebCore::IntPoint>::encode(StreamConnectionEncoder& encoder, const WebCore::IntPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, int>);
    encoder << instance.x();
    encoder << instance.y();
}

std::optional<WebCore::IntPoint> ArgumentCoder<WebCore::IntPoint>::decode(Decoder& decoder)
{
    auto x = decoder.decode<int>();
    auto y = decoder.decode<int>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IntPoint {
            WTFMove(*x),
            WTFMove(*y)
        }
    };
}

void ArgumentCoder<WebCore::IntSize>::encode(Encoder& encoder, const WebCore::IntSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, int>);
    encoder << instance.width();
    encoder << instance.height();
}

void ArgumentCoder<WebCore::IntSize>::encode(StreamConnectionEncoder& encoder, const WebCore::IntSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, int>);
    encoder << instance.width();
    encoder << instance.height();
}

std::optional<WebCore::IntSize> ArgumentCoder<WebCore::IntSize>::decode(Decoder& decoder)
{
    auto width = decoder.decode<int>();
    auto height = decoder.decode<int>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IntSize {
            WTFMove(*width),
            WTFMove(*height)
        }
    };
}

void ArgumentCoder<WebCore::FloatSize>::encode(Encoder& encoder, const WebCore::FloatSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, float>);
    encoder << instance.width();
    encoder << instance.height();
}

void ArgumentCoder<WebCore::FloatSize>::encode(StreamConnectionEncoder& encoder, const WebCore::FloatSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, float>);
    encoder << instance.width();
    encoder << instance.height();
}

std::optional<WebCore::FloatSize> ArgumentCoder<WebCore::FloatSize>::decode(Decoder& decoder)
{
    auto width = decoder.decode<float>();
    auto height = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatSize {
            WTFMove(*width),
            WTFMove(*height)
        }
    };
}

void ArgumentCoder<WebCore::LayoutUnit>::encode(Encoder& encoder, const WebCore::LayoutUnit& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rawValue())>, int>);
    encoder << instance.rawValue();
}

void ArgumentCoder<WebCore::LayoutUnit>::encode(StreamConnectionEncoder& encoder, const WebCore::LayoutUnit& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rawValue())>, int>);
    encoder << instance.rawValue();
}

std::optional<WebCore::LayoutUnit> ArgumentCoder<WebCore::LayoutUnit>::decode(Decoder& decoder)
{
    auto rawValue = decoder.decode<int>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LayoutUnit::fromRawValue(
            WTFMove(*rawValue)
        )
    };
}

void ArgumentCoder<WebCore::LayoutPoint>::encode(Encoder& encoder, const WebCore::LayoutPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, WebCore::LayoutUnit>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, WebCore::LayoutUnit>);
    encoder << instance.x();
    encoder << instance.y();
}

void ArgumentCoder<WebCore::LayoutPoint>::encode(StreamConnectionEncoder& encoder, const WebCore::LayoutPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, WebCore::LayoutUnit>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, WebCore::LayoutUnit>);
    encoder << instance.x();
    encoder << instance.y();
}

std::optional<WebCore::LayoutPoint> ArgumentCoder<WebCore::LayoutPoint>::decode(Decoder& decoder)
{
    auto x = decoder.decode<WebCore::LayoutUnit>();
    auto y = decoder.decode<WebCore::LayoutUnit>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LayoutPoint {
            WTFMove(*x),
            WTFMove(*y)
        }
    };
}

void ArgumentCoder<WebCore::LayoutSize>::encode(Encoder& encoder, const WebCore::LayoutSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, WebCore::LayoutUnit>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, WebCore::LayoutUnit>);
    encoder << instance.width();
    encoder << instance.height();
}

void ArgumentCoder<WebCore::LayoutSize>::encode(StreamConnectionEncoder& encoder, const WebCore::LayoutSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, WebCore::LayoutUnit>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, WebCore::LayoutUnit>);
    encoder << instance.width();
    encoder << instance.height();
}

std::optional<WebCore::LayoutSize> ArgumentCoder<WebCore::LayoutSize>::decode(Decoder& decoder)
{
    auto width = decoder.decode<WebCore::LayoutUnit>();
    auto height = decoder.decode<WebCore::LayoutUnit>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LayoutSize {
            WTFMove(*width),
            WTFMove(*height)
        }
    };
}

void ArgumentCoder<WebCore::DictionaryPopupInfo>::encode(Encoder& encoder, const WebCore::DictionaryPopupInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textIndicator)>, WebCore::TextIndicatorData>);
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.platformData)>, WebCore::DictionaryPopupInfoCocoa>);
#endif
    struct ShouldBeSameSizeAsDictionaryPopupInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DictionaryPopupInfo>, false> {
        WebCore::FloatPoint origin;
        WebCore::TextIndicatorData textIndicator;
#if PLATFORM(COCOA)
        WebCore::DictionaryPopupInfoCocoa platformData;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsDictionaryPopupInfo) == sizeof(WebCore::DictionaryPopupInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DictionaryPopupInfo, origin)
        , offsetof(WebCore::DictionaryPopupInfo, textIndicator)
#if PLATFORM(COCOA)
        , offsetof(WebCore::DictionaryPopupInfo, platformData)
#endif
    >::value);
    encoder << instance.origin;
    encoder << instance.textIndicator;
#if PLATFORM(COCOA)
    encoder << instance.platformData;
#endif
}

std::optional<WebCore::DictionaryPopupInfo> ArgumentCoder<WebCore::DictionaryPopupInfo>::decode(Decoder& decoder)
{
    auto origin = decoder.decode<WebCore::FloatPoint>();
    auto textIndicator = decoder.decode<WebCore::TextIndicatorData>();
#if PLATFORM(COCOA)
    auto platformData = decoder.decode<WebCore::DictionaryPopupInfoCocoa>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DictionaryPopupInfo {
            WTFMove(*origin),
            WTFMove(*textIndicator),
#if PLATFORM(COCOA)
            WTFMove(*platformData)
#endif
        }
    };
}

void ArgumentCoder<WebCore::PrivateClickMeasurement>::encode(Encoder& encoder, const WebCore::PrivateClickMeasurement& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceID())>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceSite())>, WebCore::PCM::SourceSite>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationSite())>, WebCore::PCM::AttributionDestinationSite>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeOfAdClick())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isEphemeral())>, WebCore::PCM::AttributionEphemeral>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.adamID())>, std::optional<uint64_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.attributionTriggerData())>, std::optional<WebCore::PCM::AttributionTriggerData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timesToSend())>, WebCore::PCM::AttributionTimeToSendData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ephemeralSourceNonce())>, std::optional<WebCore::PCM::EphemeralNonce>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceApplicationBundleID())>, String>);
    encoder << instance.sourceID();
    encoder << instance.sourceSite();
    encoder << instance.destinationSite();
    encoder << instance.timeOfAdClick();
    encoder << instance.isEphemeral();
    encoder << instance.adamID();
    encoder << instance.attributionTriggerData();
    encoder << instance.timesToSend();
    encoder << instance.ephemeralSourceNonce();
    encoder << instance.sourceApplicationBundleID();
}

std::optional<WebCore::PrivateClickMeasurement> ArgumentCoder<WebCore::PrivateClickMeasurement>::decode(Decoder& decoder)
{
    auto sourceID = decoder.decode<uint8_t>();
    auto sourceSite = decoder.decode<WebCore::PCM::SourceSite>();
    auto destinationSite = decoder.decode<WebCore::PCM::AttributionDestinationSite>();
    auto timeOfAdClick = decoder.decode<WallTime>();
    auto isEphemeral = decoder.decode<WebCore::PCM::AttributionEphemeral>();
    auto adamID = decoder.decode<std::optional<uint64_t>>();
    auto attributionTriggerData = decoder.decode<std::optional<WebCore::PCM::AttributionTriggerData>>();
    auto timesToSend = decoder.decode<WebCore::PCM::AttributionTimeToSendData>();
    auto ephemeralSourceNonce = decoder.decode<std::optional<WebCore::PCM::EphemeralNonce>>();
    auto sourceApplicationBundleID = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PrivateClickMeasurement {
            WTFMove(*sourceID),
            WTFMove(*sourceSite),
            WTFMove(*destinationSite),
            WTFMove(*timeOfAdClick),
            WTFMove(*isEphemeral),
            WTFMove(*adamID),
            WTFMove(*attributionTriggerData),
            WTFMove(*timesToSend),
            WTFMove(*ephemeralSourceNonce),
            WTFMove(*sourceApplicationBundleID)
        }
    };
}

void ArgumentCoder<WebCore::PCM::SourceSite>::encode(Encoder& encoder, const WebCore::PCM::SourceSite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrableDomain)>, WebCore::RegistrableDomain>);
    struct ShouldBeSameSizeAsSourceSite : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PCM::SourceSite>, false> {
        WebCore::RegistrableDomain registrableDomain;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSourceSite) == sizeof(WebCore::PCM::SourceSite));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PCM::SourceSite, registrableDomain)
    >::value);
    encoder << instance.registrableDomain;
}

std::optional<WebCore::PCM::SourceSite> ArgumentCoder<WebCore::PCM::SourceSite>::decode(Decoder& decoder)
{
    auto registrableDomain = decoder.decode<WebCore::RegistrableDomain>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PCM::SourceSite {
            WTFMove(*registrableDomain)
        }
    };
}

void ArgumentCoder<WebCore::PCM::AttributionDestinationSite>::encode(Encoder& encoder, const WebCore::PCM::AttributionDestinationSite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrableDomain)>, WebCore::RegistrableDomain>);
    struct ShouldBeSameSizeAsAttributionDestinationSite : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PCM::AttributionDestinationSite>, false> {
        WebCore::RegistrableDomain registrableDomain;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAttributionDestinationSite) == sizeof(WebCore::PCM::AttributionDestinationSite));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PCM::AttributionDestinationSite, registrableDomain)
    >::value);
    encoder << instance.registrableDomain;
}

std::optional<WebCore::PCM::AttributionDestinationSite> ArgumentCoder<WebCore::PCM::AttributionDestinationSite>::decode(Decoder& decoder)
{
    auto registrableDomain = decoder.decode<WebCore::RegistrableDomain>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PCM::AttributionDestinationSite {
            WTFMove(*registrableDomain)
        }
    };
}

void ArgumentCoder<WebCore::PCM::EphemeralNonce>::encode(Encoder& encoder, const WebCore::PCM::EphemeralNonce& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.nonce)>, String>);
    struct ShouldBeSameSizeAsEphemeralNonce : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PCM::EphemeralNonce>, false> {
        String nonce;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEphemeralNonce) == sizeof(WebCore::PCM::EphemeralNonce));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PCM::EphemeralNonce, nonce)
    >::value);
    encoder << instance.nonce;
}

std::optional<WebCore::PCM::EphemeralNonce> ArgumentCoder<WebCore::PCM::EphemeralNonce>::decode(Decoder& decoder)
{
    auto nonce = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PCM::EphemeralNonce {
            WTFMove(*nonce)
        }
    };
}

void ArgumentCoder<WebCore::PCM::AttributionTriggerData>::encode(Encoder& encoder, const WebCore::PCM::AttributionTriggerData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.priority)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wasSent)>, WebCore::PCM::WasSent>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceRegistrableDomain)>, std::optional<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ephemeralDestinationNonce)>, std::optional<WebCore::PCM::EphemeralNonce>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationSite)>, std::optional<WebCore::RegistrableDomain>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationUnlinkableToken)>, std::optional<WebCore::PCM::DestinationUnlinkableToken>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationSecretToken)>, std::optional<WebCore::PCM::DestinationSecretToken>>);
    struct ShouldBeSameSizeAsAttributionTriggerData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PCM::AttributionTriggerData>, false> {
        uint8_t data;
        uint8_t priority;
        WebCore::PCM::WasSent wasSent;
        std::optional<WebCore::RegistrableDomain> sourceRegistrableDomain;
        std::optional<WebCore::PCM::EphemeralNonce> ephemeralDestinationNonce;
        std::optional<WebCore::RegistrableDomain> destinationSite;
        std::optional<WebCore::PCM::DestinationUnlinkableToken> destinationUnlinkableToken;
        std::optional<WebCore::PCM::DestinationSecretToken> destinationSecretToken;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAttributionTriggerData) == sizeof(WebCore::PCM::AttributionTriggerData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PCM::AttributionTriggerData, data)
        , offsetof(WebCore::PCM::AttributionTriggerData, priority)
        , offsetof(WebCore::PCM::AttributionTriggerData, wasSent)
        , offsetof(WebCore::PCM::AttributionTriggerData, sourceRegistrableDomain)
        , offsetof(WebCore::PCM::AttributionTriggerData, ephemeralDestinationNonce)
        , offsetof(WebCore::PCM::AttributionTriggerData, destinationSite)
        , offsetof(WebCore::PCM::AttributionTriggerData, destinationUnlinkableToken)
        , offsetof(WebCore::PCM::AttributionTriggerData, destinationSecretToken)
    >::value);
    encoder << instance.data;
    encoder << instance.priority;
    encoder << instance.wasSent;
    encoder << instance.sourceRegistrableDomain;
    encoder << instance.ephemeralDestinationNonce;
    encoder << instance.destinationSite;
}

std::optional<WebCore::PCM::AttributionTriggerData> ArgumentCoder<WebCore::PCM::AttributionTriggerData>::decode(Decoder& decoder)
{
    auto data = decoder.decode<uint8_t>();
    auto priority = decoder.decode<uint8_t>();
    auto wasSent = decoder.decode<WebCore::PCM::WasSent>();
    auto sourceRegistrableDomain = decoder.decode<std::optional<WebCore::RegistrableDomain>>();
    auto ephemeralDestinationNonce = decoder.decode<std::optional<WebCore::PCM::EphemeralNonce>>();
    auto destinationSite = decoder.decode<std::optional<WebCore::RegistrableDomain>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PCM::AttributionTriggerData {
            WTFMove(*data),
            WTFMove(*priority),
            WTFMove(*wasSent),
            WTFMove(*sourceRegistrableDomain),
            WTFMove(*ephemeralDestinationNonce),
            WTFMove(*destinationSite)
        }
    };
}

void ArgumentCoder<WebCore::PCM::AttributionTimeToSendData>::encode(Encoder& encoder, const WebCore::PCM::AttributionTimeToSendData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceEarliestTimeToSend)>, std::optional<WallTime>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationEarliestTimeToSend)>, std::optional<WallTime>>);
    struct ShouldBeSameSizeAsAttributionTimeToSendData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PCM::AttributionTimeToSendData>, false> {
        std::optional<WallTime> sourceEarliestTimeToSend;
        std::optional<WallTime> destinationEarliestTimeToSend;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAttributionTimeToSendData) == sizeof(WebCore::PCM::AttributionTimeToSendData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PCM::AttributionTimeToSendData, sourceEarliestTimeToSend)
        , offsetof(WebCore::PCM::AttributionTimeToSendData, destinationEarliestTimeToSend)
    >::value);
    encoder << instance.sourceEarliestTimeToSend;
    encoder << instance.destinationEarliestTimeToSend;
}

std::optional<WebCore::PCM::AttributionTimeToSendData> ArgumentCoder<WebCore::PCM::AttributionTimeToSendData>::decode(Decoder& decoder)
{
    auto sourceEarliestTimeToSend = decoder.decode<std::optional<WallTime>>();
    auto destinationEarliestTimeToSend = decoder.decode<std::optional<WallTime>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PCM::AttributionTimeToSendData {
            WTFMove(*sourceEarliestTimeToSend),
            WTFMove(*destinationEarliestTimeToSend)
        }
    };
}

#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
void ArgumentCoder<WebCore::ApplePayRecurringPaymentRequest>::encode(Encoder& encoder, const WebCore::ApplePayRecurringPaymentRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paymentDescription)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.regularBilling)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trialBilling)>, std::optional<WebCore::ApplePayLineItem>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.billingAgreement)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.managementURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tokenNotificationURL)>, String>);
    struct ShouldBeSameSizeAsApplePayRecurringPaymentRequest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayRecurringPaymentRequest>, false> {
        String paymentDescription;
        WebCore::ApplePayLineItem regularBilling;
        std::optional<WebCore::ApplePayLineItem> trialBilling;
        String billingAgreement;
        String managementURL;
        String tokenNotificationURL;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayRecurringPaymentRequest) == sizeof(WebCore::ApplePayRecurringPaymentRequest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayRecurringPaymentRequest, paymentDescription)
        , offsetof(WebCore::ApplePayRecurringPaymentRequest, regularBilling)
        , offsetof(WebCore::ApplePayRecurringPaymentRequest, trialBilling)
        , offsetof(WebCore::ApplePayRecurringPaymentRequest, billingAgreement)
        , offsetof(WebCore::ApplePayRecurringPaymentRequest, managementURL)
        , offsetof(WebCore::ApplePayRecurringPaymentRequest, tokenNotificationURL)
    >::value);
    encoder << instance.paymentDescription;
    encoder << instance.regularBilling;
    encoder << instance.trialBilling;
    encoder << instance.billingAgreement;
    encoder << instance.managementURL;
    encoder << instance.tokenNotificationURL;
}

std::optional<WebCore::ApplePayRecurringPaymentRequest> ArgumentCoder<WebCore::ApplePayRecurringPaymentRequest>::decode(Decoder& decoder)
{
    auto paymentDescription = decoder.decode<String>();
    auto regularBilling = decoder.decode<WebCore::ApplePayLineItem>();
    auto trialBilling = decoder.decode<std::optional<WebCore::ApplePayLineItem>>();
    auto billingAgreement = decoder.decode<String>();
    auto managementURL = decoder.decode<String>();
    auto tokenNotificationURL = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayRecurringPaymentRequest {
            WTFMove(*paymentDescription),
            WTFMove(*regularBilling),
            WTFMove(*trialBilling),
            WTFMove(*billingAgreement),
            WTFMove(*managementURL),
            WTFMove(*tokenNotificationURL)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
void ArgumentCoder<WebCore::ApplePayPaymentTokenContext>::encode(Encoder& encoder, const WebCore::ApplePayPaymentTokenContext& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.merchantIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.externalIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.merchantName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.merchantDomain)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.amount)>, String>);
    struct ShouldBeSameSizeAsApplePayPaymentTokenContext : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayPaymentTokenContext>, false> {
        String merchantIdentifier;
        String externalIdentifier;
        String merchantName;
        String merchantDomain;
        String amount;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayPaymentTokenContext) == sizeof(WebCore::ApplePayPaymentTokenContext));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayPaymentTokenContext, merchantIdentifier)
        , offsetof(WebCore::ApplePayPaymentTokenContext, externalIdentifier)
        , offsetof(WebCore::ApplePayPaymentTokenContext, merchantName)
        , offsetof(WebCore::ApplePayPaymentTokenContext, merchantDomain)
        , offsetof(WebCore::ApplePayPaymentTokenContext, amount)
    >::value);
    encoder << instance.merchantIdentifier;
    encoder << instance.externalIdentifier;
    encoder << instance.merchantName;
    encoder << instance.merchantDomain;
    encoder << instance.amount;
}

std::optional<WebCore::ApplePayPaymentTokenContext> ArgumentCoder<WebCore::ApplePayPaymentTokenContext>::decode(Decoder& decoder)
{
    auto merchantIdentifier = decoder.decode<String>();
    auto externalIdentifier = decoder.decode<String>();
    auto merchantName = decoder.decode<String>();
    auto merchantDomain = decoder.decode<String>();
    auto amount = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayPaymentTokenContext {
            WTFMove(*merchantIdentifier),
            WTFMove(*externalIdentifier),
            WTFMove(*merchantName),
            WTFMove(*merchantDomain),
            WTFMove(*amount)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
void ArgumentCoder<WebCore::ApplePayDeferredPaymentRequest>::encode(Encoder& encoder, const WebCore::ApplePayDeferredPaymentRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.billingAgreement)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deferredBilling)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.freeCancellationDate)>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.freeCancellationDateTimeZone)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.managementURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paymentDescription)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tokenNotificationURL)>, String>);
    struct ShouldBeSameSizeAsApplePayDeferredPaymentRequest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayDeferredPaymentRequest>, false> {
        String billingAgreement;
        WebCore::ApplePayLineItem deferredBilling;
        WallTime freeCancellationDate;
        String freeCancellationDateTimeZone;
        String managementURL;
        String paymentDescription;
        String tokenNotificationURL;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayDeferredPaymentRequest) == sizeof(WebCore::ApplePayDeferredPaymentRequest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDeferredPaymentRequest, billingAgreement)
        , offsetof(WebCore::ApplePayDeferredPaymentRequest, deferredBilling)
        , offsetof(WebCore::ApplePayDeferredPaymentRequest, freeCancellationDate)
        , offsetof(WebCore::ApplePayDeferredPaymentRequest, freeCancellationDateTimeZone)
        , offsetof(WebCore::ApplePayDeferredPaymentRequest, managementURL)
        , offsetof(WebCore::ApplePayDeferredPaymentRequest, paymentDescription)
        , offsetof(WebCore::ApplePayDeferredPaymentRequest, tokenNotificationURL)
    >::value);
    encoder << instance.billingAgreement;
    encoder << instance.deferredBilling;
    encoder << instance.freeCancellationDate;
    encoder << instance.freeCancellationDateTimeZone;
    encoder << instance.managementURL;
    encoder << instance.paymentDescription;
    encoder << instance.tokenNotificationURL;
}

std::optional<WebCore::ApplePayDeferredPaymentRequest> ArgumentCoder<WebCore::ApplePayDeferredPaymentRequest>::decode(Decoder& decoder)
{
    auto billingAgreement = decoder.decode<String>();
    auto deferredBilling = decoder.decode<WebCore::ApplePayLineItem>();
    auto freeCancellationDate = decoder.decode<WallTime>();
    auto freeCancellationDateTimeZone = decoder.decode<String>();
    auto managementURL = decoder.decode<String>();
    auto paymentDescription = decoder.decode<String>();
    auto tokenNotificationURL = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayDeferredPaymentRequest {
            WTFMove(*billingAgreement),
            WTFMove(*deferredBilling),
            WTFMove(*freeCancellationDate),
            WTFMove(*freeCancellationDateTimeZone),
            WTFMove(*managementURL),
            WTFMove(*paymentDescription),
            WTFMove(*tokenNotificationURL)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
void ArgumentCoder<WebCore::ApplePayPaymentOrderDetails>::encode(Encoder& encoder, const WebCore::ApplePayPaymentOrderDetails& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.orderTypeIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.orderIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webServiceURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authenticationToken)>, String>);
    struct ShouldBeSameSizeAsApplePayPaymentOrderDetails : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayPaymentOrderDetails>, false> {
        String orderTypeIdentifier;
        String orderIdentifier;
        String webServiceURL;
        String authenticationToken;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayPaymentOrderDetails) == sizeof(WebCore::ApplePayPaymentOrderDetails));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayPaymentOrderDetails, orderTypeIdentifier)
        , offsetof(WebCore::ApplePayPaymentOrderDetails, orderIdentifier)
        , offsetof(WebCore::ApplePayPaymentOrderDetails, webServiceURL)
        , offsetof(WebCore::ApplePayPaymentOrderDetails, authenticationToken)
    >::value);
    encoder << instance.orderTypeIdentifier;
    encoder << instance.orderIdentifier;
    encoder << instance.webServiceURL;
    encoder << instance.authenticationToken;
}

std::optional<WebCore::ApplePayPaymentOrderDetails> ArgumentCoder<WebCore::ApplePayPaymentOrderDetails>::decode(Decoder& decoder)
{
    auto orderTypeIdentifier = decoder.decode<String>();
    auto orderIdentifier = decoder.decode<String>();
    auto webServiceURL = decoder.decode<String>();
    auto authenticationToken = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayPaymentOrderDetails {
            WTFMove(*orderTypeIdentifier),
            WTFMove(*orderIdentifier),
            WTFMove(*webServiceURL),
            WTFMove(*authenticationToken)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_AMS_UI) && ENABLE(PAYMENT_REQUEST)
void ArgumentCoder<WebCore::ApplePayAMSUIRequest>::encode(Encoder& encoder, const WebCore::ApplePayAMSUIRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.engagementRequest)>, String>);
    struct ShouldBeSameSizeAsApplePayAMSUIRequest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayAMSUIRequest>, false> {
        String engagementRequest;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayAMSUIRequest) == sizeof(WebCore::ApplePayAMSUIRequest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayAMSUIRequest, engagementRequest)
    >::value);
    encoder << instance.engagementRequest;
}

std::optional<WebCore::ApplePayAMSUIRequest> ArgumentCoder<WebCore::ApplePayAMSUIRequest>::decode(Decoder& decoder)
{
    auto engagementRequest = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayAMSUIRequest {
            WTFMove(*engagementRequest)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
void ArgumentCoder<WebCore::ApplePayAutomaticReloadPaymentRequest>::encode(Encoder& encoder, const WebCore::ApplePayAutomaticReloadPaymentRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paymentDescription)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.automaticReloadBilling)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.billingAgreement)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.managementURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tokenNotificationURL)>, String>);
    struct ShouldBeSameSizeAsApplePayAutomaticReloadPaymentRequest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayAutomaticReloadPaymentRequest>, false> {
        String paymentDescription;
        WebCore::ApplePayLineItem automaticReloadBilling;
        String billingAgreement;
        String managementURL;
        String tokenNotificationURL;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayAutomaticReloadPaymentRequest) == sizeof(WebCore::ApplePayAutomaticReloadPaymentRequest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayAutomaticReloadPaymentRequest, paymentDescription)
        , offsetof(WebCore::ApplePayAutomaticReloadPaymentRequest, automaticReloadBilling)
        , offsetof(WebCore::ApplePayAutomaticReloadPaymentRequest, billingAgreement)
        , offsetof(WebCore::ApplePayAutomaticReloadPaymentRequest, managementURL)
        , offsetof(WebCore::ApplePayAutomaticReloadPaymentRequest, tokenNotificationURL)
    >::value);
    encoder << instance.paymentDescription;
    encoder << instance.automaticReloadBilling;
    encoder << instance.billingAgreement;
    encoder << instance.managementURL;
    encoder << instance.tokenNotificationURL;
}

std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest> ArgumentCoder<WebCore::ApplePayAutomaticReloadPaymentRequest>::decode(Decoder& decoder)
{
    auto paymentDescription = decoder.decode<String>();
    auto automaticReloadBilling = decoder.decode<WebCore::ApplePayLineItem>();
    auto billingAgreement = decoder.decode<String>();
    auto managementURL = decoder.decode<String>();
    auto tokenNotificationURL = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayAutomaticReloadPaymentRequest {
            WTFMove(*paymentDescription),
            WTFMove(*automaticReloadBilling),
            WTFMove(*billingAgreement),
            WTFMove(*managementURL),
            WTFMove(*tokenNotificationURL)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
void ArgumentCoder<WebCore::ApplePayDateComponents>::encode(Encoder& encoder, const WebCore::ApplePayDateComponents& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.years)>, std::optional<unsigned>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.months)>, std::optional<unsigned>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.days)>, std::optional<unsigned>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hours)>, std::optional<unsigned>>);
    struct ShouldBeSameSizeAsApplePayDateComponents : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayDateComponents>, false> {
        std::optional<unsigned> years;
        std::optional<unsigned> months;
        std::optional<unsigned> days;
        std::optional<unsigned> hours;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayDateComponents) == sizeof(WebCore::ApplePayDateComponents));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDateComponents, years)
        , offsetof(WebCore::ApplePayDateComponents, months)
        , offsetof(WebCore::ApplePayDateComponents, days)
        , offsetof(WebCore::ApplePayDateComponents, hours)
    >::value);
    encoder << instance.years;
    encoder << instance.months;
    encoder << instance.days;
    encoder << instance.hours;
}

std::optional<WebCore::ApplePayDateComponents> ArgumentCoder<WebCore::ApplePayDateComponents>::decode(Decoder& decoder)
{
    auto years = decoder.decode<std::optional<unsigned>>();
    auto months = decoder.decode<std::optional<unsigned>>();
    auto days = decoder.decode<std::optional<unsigned>>();
    auto hours = decoder.decode<std::optional<unsigned>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayDateComponents {
            WTFMove(*years),
            WTFMove(*months),
            WTFMove(*days),
            WTFMove(*hours)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
void ArgumentCoder<WebCore::ApplePayDateComponentsRange>::encode(Encoder& encoder, const WebCore::ApplePayDateComponentsRange& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startDateComponents)>, WebCore::ApplePayDateComponents>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endDateComponents)>, WebCore::ApplePayDateComponents>);
    struct ShouldBeSameSizeAsApplePayDateComponentsRange : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayDateComponentsRange>, false> {
        WebCore::ApplePayDateComponents startDateComponents;
        WebCore::ApplePayDateComponents endDateComponents;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayDateComponentsRange) == sizeof(WebCore::ApplePayDateComponentsRange));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDateComponentsRange, startDateComponents)
        , offsetof(WebCore::ApplePayDateComponentsRange, endDateComponents)
    >::value);
    encoder << instance.startDateComponents;
    encoder << instance.endDateComponents;
}

std::optional<WebCore::ApplePayDateComponentsRange> ArgumentCoder<WebCore::ApplePayDateComponentsRange>::decode(Decoder& decoder)
{
    auto startDateComponents = decoder.decode<WebCore::ApplePayDateComponents>();
    auto endDateComponents = decoder.decode<WebCore::ApplePayDateComponents>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayDateComponentsRange {
            WTFMove(*startDateComponents),
            WTFMove(*endDateComponents)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePaySetupConfiguration>::encode(Encoder& encoder, const WebCore::ApplePaySetupConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.merchantIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrerIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.signature)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.signedFields)>, Vector<String>>);
    struct ShouldBeSameSizeAsApplePaySetupConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePaySetupConfiguration>, false> {
        String merchantIdentifier;
        String referrerIdentifier;
        String signature;
        Vector<String> signedFields;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePaySetupConfiguration) == sizeof(WebCore::ApplePaySetupConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePaySetupConfiguration, merchantIdentifier)
        , offsetof(WebCore::ApplePaySetupConfiguration, referrerIdentifier)
        , offsetof(WebCore::ApplePaySetupConfiguration, signature)
        , offsetof(WebCore::ApplePaySetupConfiguration, signedFields)
    >::value);
    encoder << instance.merchantIdentifier;
    encoder << instance.referrerIdentifier;
    encoder << instance.signature;
    encoder << instance.signedFields;
}

std::optional<WebCore::ApplePaySetupConfiguration> ArgumentCoder<WebCore::ApplePaySetupConfiguration>::decode(Decoder& decoder)
{
    auto merchantIdentifier = decoder.decode<String>();
    auto referrerIdentifier = decoder.decode<String>();
    auto signature = decoder.decode<String>();
    auto signedFields = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePaySetupConfiguration {
            WTFMove(*merchantIdentifier),
            WTFMove(*referrerIdentifier),
            WTFMove(*signature),
            WTFMove(*signedFields)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayLineItem>::encode(Encoder& encoder, const WebCore::ApplePayLineItem& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ApplePayLineItem::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.amount)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paymentTiming)>, WebCore::ApplePayPaymentTiming>);
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.recurringPaymentStartDate)>, WallTime>);
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.recurringPaymentIntervalUnit)>, WebCore::ApplePayRecurringPaymentDateUnit>);
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.recurringPaymentIntervalCount)>, unsigned>);
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.recurringPaymentEndDate)>, WallTime>);
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deferredPaymentDate)>, WallTime>);
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.automaticReloadPaymentThresholdAmount)>, String>);
#endif
    struct ShouldBeSameSizeAsApplePayLineItem : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayLineItem>, false> {
        WebCore::ApplePayLineItem::Type type;
        String label;
        String amount;
        WebCore::ApplePayPaymentTiming paymentTiming;
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        WallTime recurringPaymentStartDate;
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        WebCore::ApplePayRecurringPaymentDateUnit recurringPaymentIntervalUnit;
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        unsigned recurringPaymentIntervalCount;
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        WallTime recurringPaymentEndDate;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
        WallTime deferredPaymentDate;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
        String automaticReloadPaymentThresholdAmount;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayLineItem) == sizeof(WebCore::ApplePayLineItem));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayLineItem, type)
        , offsetof(WebCore::ApplePayLineItem, label)
        , offsetof(WebCore::ApplePayLineItem, amount)
        , offsetof(WebCore::ApplePayLineItem, paymentTiming)
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        , offsetof(WebCore::ApplePayLineItem, recurringPaymentStartDate)
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        , offsetof(WebCore::ApplePayLineItem, recurringPaymentIntervalUnit)
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        , offsetof(WebCore::ApplePayLineItem, recurringPaymentIntervalCount)
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
        , offsetof(WebCore::ApplePayLineItem, recurringPaymentEndDate)
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
        , offsetof(WebCore::ApplePayLineItem, deferredPaymentDate)
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
        , offsetof(WebCore::ApplePayLineItem, automaticReloadPaymentThresholdAmount)
#endif
    >::value);
    encoder << instance.type;
    encoder << instance.label;
    encoder << instance.amount;
    encoder << instance.paymentTiming;
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    encoder << instance.recurringPaymentStartDate;
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    encoder << instance.recurringPaymentIntervalUnit;
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    encoder << instance.recurringPaymentIntervalCount;
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    encoder << instance.recurringPaymentEndDate;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
    encoder << instance.deferredPaymentDate;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
    encoder << instance.automaticReloadPaymentThresholdAmount;
#endif
}

std::optional<WebCore::ApplePayLineItem> ArgumentCoder<WebCore::ApplePayLineItem>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::ApplePayLineItem::Type>();
    auto label = decoder.decode<String>();
    auto amount = decoder.decode<String>();
    auto paymentTiming = decoder.decode<WebCore::ApplePayPaymentTiming>();
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    auto recurringPaymentStartDate = decoder.decode<WallTime>();
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    auto recurringPaymentIntervalUnit = decoder.decode<WebCore::ApplePayRecurringPaymentDateUnit>();
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    auto recurringPaymentIntervalCount = decoder.decode<unsigned>();
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    auto recurringPaymentEndDate = decoder.decode<WallTime>();
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
    auto deferredPaymentDate = decoder.decode<WallTime>();
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
    auto automaticReloadPaymentThresholdAmount = decoder.decode<String>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayLineItem {
            WTFMove(*type),
            WTFMove(*label),
            WTFMove(*amount),
            WTFMove(*paymentTiming),
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
            WTFMove(*recurringPaymentStartDate),
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
            WTFMove(*recurringPaymentIntervalUnit),
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
            WTFMove(*recurringPaymentIntervalCount),
#endif
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
            WTFMove(*recurringPaymentEndDate),
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
            WTFMove(*deferredPaymentDate),
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
            WTFMove(*automaticReloadPaymentThresholdAmount)
#endif
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayShippingMethod>::encode(Encoder& encoder, const WebCore::ApplePayShippingMethod& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.detail)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.amount)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, String>);
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dateComponentsRange)>, std::optional<WebCore::ApplePayDateComponentsRange>>);
#endif
#if ENABLE(APPLE_PAY_SELECTED_SHIPPING_METHOD)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selected)>, bool>);
#endif
    struct ShouldBeSameSizeAsApplePayShippingMethod : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayShippingMethod>, false> {
        String label;
        String detail;
        String amount;
        String identifier;
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
        std::optional<WebCore::ApplePayDateComponentsRange> dateComponentsRange;
#endif
#if ENABLE(APPLE_PAY_SELECTED_SHIPPING_METHOD)
        bool selected;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayShippingMethod) == sizeof(WebCore::ApplePayShippingMethod));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayShippingMethod, label)
        , offsetof(WebCore::ApplePayShippingMethod, detail)
        , offsetof(WebCore::ApplePayShippingMethod, amount)
        , offsetof(WebCore::ApplePayShippingMethod, identifier)
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
        , offsetof(WebCore::ApplePayShippingMethod, dateComponentsRange)
#endif
#if ENABLE(APPLE_PAY_SELECTED_SHIPPING_METHOD)
        , offsetof(WebCore::ApplePayShippingMethod, selected)
#endif
    >::value);
    encoder << instance.label;
    encoder << instance.detail;
    encoder << instance.amount;
    encoder << instance.identifier;
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
    encoder << instance.dateComponentsRange;
#endif
#if ENABLE(APPLE_PAY_SELECTED_SHIPPING_METHOD)
    encoder << instance.selected;
#endif
}

std::optional<WebCore::ApplePayShippingMethod> ArgumentCoder<WebCore::ApplePayShippingMethod>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto detail = decoder.decode<String>();
    auto amount = decoder.decode<String>();
    auto identifier = decoder.decode<String>();
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
    auto dateComponentsRange = decoder.decode<std::optional<WebCore::ApplePayDateComponentsRange>>();
#endif
#if ENABLE(APPLE_PAY_SELECTED_SHIPPING_METHOD)
    auto selected = decoder.decode<bool>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayShippingMethod {
            WTFMove(*label),
            WTFMove(*detail),
            WTFMove(*amount),
            WTFMove(*identifier),
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
            WTFMove(*dateComponentsRange),
#endif
#if ENABLE(APPLE_PAY_SELECTED_SHIPPING_METHOD)
            WTFMove(*selected)
#endif
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayError>::encode(Encoder& encoder, const WebCore::ApplePayError& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.code())>, WebCore::ApplePayErrorCode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contactField())>, std::optional<WebCore::ApplePayErrorContactField>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message())>, String>);
    encoder << instance.code();
    encoder << instance.contactField();
    encoder << instance.message();
}

std::optional<Ref<WebCore::ApplePayError>> ArgumentCoder<WebCore::ApplePayError>::decode(Decoder& decoder)
{
    auto code = decoder.decode<WebCore::ApplePayErrorCode>();
    auto contactField = decoder.decode<std::optional<WebCore::ApplePayErrorContactField>>();
    auto message = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayError::create(
            WTFMove(*code),
            WTFMove(*contactField),
            WTFMove(*message)
        )
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayLogoSystemImage>::encode(Encoder& encoder, const WebCore::ApplePayLogoSystemImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applePayLogoStyle())>, WebCore::ApplePayLogoStyle>);
    encoder << instance.applePayLogoStyle();
}

void ArgumentCoder<WebCore::ApplePayLogoSystemImage>::encode(StreamConnectionEncoder& encoder, const WebCore::ApplePayLogoSystemImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applePayLogoStyle())>, WebCore::ApplePayLogoStyle>);
    encoder << instance.applePayLogoStyle();
}

std::optional<Ref<WebCore::ApplePayLogoSystemImage>> ArgumentCoder<WebCore::ApplePayLogoSystemImage>::decode(Decoder& decoder)
{
    auto applePayLogoStyle = decoder.decode<WebCore::ApplePayLogoStyle>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayLogoSystemImage::create(
            WTFMove(*applePayLogoStyle)
        )
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayDetailsUpdateBase>::encode(Encoder& encoder, const WebCore::ApplePayDetailsUpdateBase& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newTotal)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newLineItems)>, Vector<WebCore::ApplePayLineItem>>);
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newRecurringPaymentRequest)>, std::optional<WebCore::ApplePayRecurringPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newAutomaticReloadPaymentRequest)>, std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newMultiTokenContexts)>, std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>);
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newDeferredPaymentRequest)>, std::optional<WebCore::ApplePayDeferredPaymentRequest>>);
#endif
    struct ShouldBeSameSizeAsApplePayDetailsUpdateBase : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayDetailsUpdateBase>, false> {
        WebCore::ApplePayLineItem newTotal;
        Vector<WebCore::ApplePayLineItem> newLineItems;
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
        std::optional<WebCore::ApplePayRecurringPaymentRequest> newRecurringPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
        std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest> newAutomaticReloadPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
        std::optional<Vector<WebCore::ApplePayPaymentTokenContext>> newMultiTokenContexts;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
        std::optional<WebCore::ApplePayDeferredPaymentRequest> newDeferredPaymentRequest;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayDetailsUpdateBase) == sizeof(WebCore::ApplePayDetailsUpdateBase));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newTotal)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newLineItems)
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newRecurringPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newAutomaticReloadPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newMultiTokenContexts)
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newDeferredPaymentRequest)
#endif
    >::value);
    encoder << instance.newTotal;
    encoder << instance.newLineItems;
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    encoder << instance.newRecurringPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    encoder << instance.newAutomaticReloadPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    encoder << instance.newMultiTokenContexts;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    encoder << instance.newDeferredPaymentRequest;
#endif
}

std::optional<WebCore::ApplePayDetailsUpdateBase> ArgumentCoder<WebCore::ApplePayDetailsUpdateBase>::decode(Decoder& decoder)
{
    auto newTotal = decoder.decode<WebCore::ApplePayLineItem>();
    auto newLineItems = decoder.decode<Vector<WebCore::ApplePayLineItem>>();
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    auto newRecurringPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayRecurringPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    auto newAutomaticReloadPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    auto newMultiTokenContexts = decoder.decode<std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>();
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    auto newDeferredPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayDeferredPaymentRequest>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayDetailsUpdateBase {
            WTFMove(*newTotal),
            WTFMove(*newLineItems),
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
            WTFMove(*newRecurringPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
            WTFMove(*newAutomaticReloadPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
            WTFMove(*newMultiTokenContexts),
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
            WTFMove(*newDeferredPaymentRequest)
#endif
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayPaymentMethodUpdate>::encode(Encoder& encoder, const WebCore::ApplePayPaymentMethodUpdate& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newTotal)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newLineItems)>, Vector<WebCore::ApplePayLineItem>>);
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newRecurringPaymentRequest)>, std::optional<WebCore::ApplePayRecurringPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newAutomaticReloadPaymentRequest)>, std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newMultiTokenContexts)>, std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>);
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newDeferredPaymentRequest)>, std::optional<WebCore::ApplePayDeferredPaymentRequest>>);
#endif
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newTotal)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newLineItems)
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newRecurringPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newAutomaticReloadPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newMultiTokenContexts)
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newDeferredPaymentRequest)
#endif
    >::value);
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.errors)>, Vector<RefPtr<WebCore::ApplePayError>>>);
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newShippingMethods)>, Vector<WebCore::ApplePayShippingMethod>>);
#endif
#if ENABLE(APPLE_PAY_INSTALLMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.installmentGroupIdentifier)>, String>);
#endif
    static_assert(MembersInCorrectOrder < 0
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
        , offsetof(WebCore::ApplePayPaymentMethodUpdate, errors)
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
        , offsetof(WebCore::ApplePayPaymentMethodUpdate, newShippingMethods)
#endif
#if ENABLE(APPLE_PAY_INSTALLMENTS)
        , offsetof(WebCore::ApplePayPaymentMethodUpdate, installmentGroupIdentifier)
#endif
    >::value);
    encoder << instance.newTotal;
    encoder << instance.newLineItems;
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    encoder << instance.newRecurringPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    encoder << instance.newAutomaticReloadPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    encoder << instance.newMultiTokenContexts;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    encoder << instance.newDeferredPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    encoder << instance.errors;
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    encoder << instance.newShippingMethods;
#endif
#if ENABLE(APPLE_PAY_INSTALLMENTS)
    encoder << instance.installmentGroupIdentifier;
#endif
}

std::optional<WebCore::ApplePayPaymentMethodUpdate> ArgumentCoder<WebCore::ApplePayPaymentMethodUpdate>::decode(Decoder& decoder)
{
    auto newTotal = decoder.decode<WebCore::ApplePayLineItem>();
    auto newLineItems = decoder.decode<Vector<WebCore::ApplePayLineItem>>();
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    auto newRecurringPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayRecurringPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    auto newAutomaticReloadPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    auto newMultiTokenContexts = decoder.decode<std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>();
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    auto newDeferredPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayDeferredPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    auto errors = decoder.decode<Vector<RefPtr<WebCore::ApplePayError>>>();
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    auto newShippingMethods = decoder.decode<Vector<WebCore::ApplePayShippingMethod>>();
#endif
#if ENABLE(APPLE_PAY_INSTALLMENTS)
    auto installmentGroupIdentifier = decoder.decode<String>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayPaymentMethodUpdate {
            WebCore::ApplePayDetailsUpdateBase {
                WTFMove(*newTotal),
                WTFMove(*newLineItems),
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
                WTFMove(*newRecurringPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
                WTFMove(*newAutomaticReloadPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
                WTFMove(*newMultiTokenContexts),
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
                WTFMove(*newDeferredPaymentRequest)
#endif
            },
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
            WTFMove(*errors),
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
            WTFMove(*newShippingMethods),
#endif
#if ENABLE(APPLE_PAY_INSTALLMENTS)
            WTFMove(*installmentGroupIdentifier)
#endif
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayShippingContactUpdate>::encode(Encoder& encoder, const WebCore::ApplePayShippingContactUpdate& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newTotal)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newLineItems)>, Vector<WebCore::ApplePayLineItem>>);
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newRecurringPaymentRequest)>, std::optional<WebCore::ApplePayRecurringPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newAutomaticReloadPaymentRequest)>, std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newMultiTokenContexts)>, std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>);
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newDeferredPaymentRequest)>, std::optional<WebCore::ApplePayDeferredPaymentRequest>>);
#endif
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newTotal)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newLineItems)
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newRecurringPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newAutomaticReloadPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newMultiTokenContexts)
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newDeferredPaymentRequest)
#endif
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.errors)>, Vector<RefPtr<WebCore::ApplePayError>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newShippingMethods)>, Vector<WebCore::ApplePayShippingMethod>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayShippingContactUpdate, errors)
        , offsetof(WebCore::ApplePayShippingContactUpdate, newShippingMethods)
    >::value);
    encoder << instance.newTotal;
    encoder << instance.newLineItems;
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    encoder << instance.newRecurringPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    encoder << instance.newAutomaticReloadPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    encoder << instance.newMultiTokenContexts;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    encoder << instance.newDeferredPaymentRequest;
#endif
    encoder << instance.errors;
    encoder << instance.newShippingMethods;
}

std::optional<WebCore::ApplePayShippingContactUpdate> ArgumentCoder<WebCore::ApplePayShippingContactUpdate>::decode(Decoder& decoder)
{
    auto newTotal = decoder.decode<WebCore::ApplePayLineItem>();
    auto newLineItems = decoder.decode<Vector<WebCore::ApplePayLineItem>>();
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    auto newRecurringPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayRecurringPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    auto newAutomaticReloadPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    auto newMultiTokenContexts = decoder.decode<std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>();
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    auto newDeferredPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayDeferredPaymentRequest>>();
#endif
    auto errors = decoder.decode<Vector<RefPtr<WebCore::ApplePayError>>>();
    auto newShippingMethods = decoder.decode<Vector<WebCore::ApplePayShippingMethod>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayShippingContactUpdate {
            WebCore::ApplePayDetailsUpdateBase {
                WTFMove(*newTotal),
                WTFMove(*newLineItems),
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
                WTFMove(*newRecurringPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
                WTFMove(*newAutomaticReloadPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
                WTFMove(*newMultiTokenContexts),
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
                WTFMove(*newDeferredPaymentRequest)
#endif
            },
            WTFMove(*errors),
            WTFMove(*newShippingMethods)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayShippingMethodUpdate>::encode(Encoder& encoder, const WebCore::ApplePayShippingMethodUpdate& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newTotal)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newLineItems)>, Vector<WebCore::ApplePayLineItem>>);
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newRecurringPaymentRequest)>, std::optional<WebCore::ApplePayRecurringPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newAutomaticReloadPaymentRequest)>, std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newMultiTokenContexts)>, std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>);
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newDeferredPaymentRequest)>, std::optional<WebCore::ApplePayDeferredPaymentRequest>>);
#endif
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newTotal)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newLineItems)
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newRecurringPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newAutomaticReloadPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newMultiTokenContexts)
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newDeferredPaymentRequest)
#endif
    >::value);
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newShippingMethods)>, Vector<WebCore::ApplePayShippingMethod>>);
#endif
    static_assert(MembersInCorrectOrder < 0
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
        , offsetof(WebCore::ApplePayShippingMethodUpdate, newShippingMethods)
#endif
    >::value);
    encoder << instance.newTotal;
    encoder << instance.newLineItems;
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    encoder << instance.newRecurringPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    encoder << instance.newAutomaticReloadPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    encoder << instance.newMultiTokenContexts;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    encoder << instance.newDeferredPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    encoder << instance.newShippingMethods;
#endif
}

std::optional<WebCore::ApplePayShippingMethodUpdate> ArgumentCoder<WebCore::ApplePayShippingMethodUpdate>::decode(Decoder& decoder)
{
    auto newTotal = decoder.decode<WebCore::ApplePayLineItem>();
    auto newLineItems = decoder.decode<Vector<WebCore::ApplePayLineItem>>();
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    auto newRecurringPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayRecurringPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    auto newAutomaticReloadPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    auto newMultiTokenContexts = decoder.decode<std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>();
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    auto newDeferredPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayDeferredPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
    auto newShippingMethods = decoder.decode<Vector<WebCore::ApplePayShippingMethod>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayShippingMethodUpdate {
            WebCore::ApplePayDetailsUpdateBase {
                WTFMove(*newTotal),
                WTFMove(*newLineItems),
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
                WTFMove(*newRecurringPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
                WTFMove(*newAutomaticReloadPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
                WTFMove(*newMultiTokenContexts),
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
                WTFMove(*newDeferredPaymentRequest)
#endif
            },
#if ENABLE(APPLE_PAY_UPDATE_SHIPPING_METHODS_WHEN_CHANGING_LINE_ITEMS)
            WTFMove(*newShippingMethods)
#endif
        }
    };
}

#endif

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayPaymentAuthorizationResult>::encode(Encoder& encoder, const WebCore::ApplePayPaymentAuthorizationResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.status)>, unsigned short>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.errors)>, Vector<RefPtr<WebCore::ApplePayError>>>);
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.orderDetails)>, std::optional<WebCore::ApplePayPaymentOrderDetails>>);
#endif
    struct ShouldBeSameSizeAsApplePayPaymentAuthorizationResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayPaymentAuthorizationResult>, false> {
        unsigned short status;
        Vector<RefPtr<WebCore::ApplePayError>> errors;
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
        std::optional<WebCore::ApplePayPaymentOrderDetails> orderDetails;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayPaymentAuthorizationResult) == sizeof(WebCore::ApplePayPaymentAuthorizationResult));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayPaymentAuthorizationResult, status)
        , offsetof(WebCore::ApplePayPaymentAuthorizationResult, errors)
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
        , offsetof(WebCore::ApplePayPaymentAuthorizationResult, orderDetails)
#endif
    >::value);
    encoder << instance.status;
    encoder << instance.errors;
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
    encoder << instance.orderDetails;
#endif
}

std::optional<WebCore::ApplePayPaymentAuthorizationResult> ArgumentCoder<WebCore::ApplePayPaymentAuthorizationResult>::decode(Decoder& decoder)
{
    auto status = decoder.decode<unsigned short>();
    auto errors = decoder.decode<Vector<RefPtr<WebCore::ApplePayError>>>();
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
    auto orderDetails = decoder.decode<std::optional<WebCore::ApplePayPaymentOrderDetails>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayPaymentAuthorizationResult {
            WTFMove(*status),
            WTFMove(*errors),
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
            WTFMove(*orderDetails)
#endif
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_INSTALLMENTS)
void ArgumentCoder<WebCore::ApplePayInstallmentConfiguration>::encode(Encoder& encoder, const WebCore::ApplePayInstallmentConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.featureType)>, WebCore::ApplePaySetupFeatureType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.merchandisingImageData)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.openToBuyThresholdAmount)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bindingTotalAmount)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.currencyCode)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isInStorePurchase)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.merchantIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrerIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.items)>, Vector<WebCore::ApplePayInstallmentItem>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationMetadata)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.retailChannel)>, WebCore::ApplePayInstallmentRetailChannel>);
    struct ShouldBeSameSizeAsApplePayInstallmentConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayInstallmentConfiguration>, false> {
        WebCore::ApplePaySetupFeatureType featureType;
        String merchandisingImageData;
        String openToBuyThresholdAmount;
        String bindingTotalAmount;
        String currencyCode;
        bool isInStorePurchase;
        String merchantIdentifier;
        String referrerIdentifier;
        Vector<WebCore::ApplePayInstallmentItem> items;
        String applicationMetadata;
        WebCore::ApplePayInstallmentRetailChannel retailChannel;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayInstallmentConfiguration) == sizeof(WebCore::ApplePayInstallmentConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayInstallmentConfiguration, featureType)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, merchandisingImageData)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, openToBuyThresholdAmount)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, bindingTotalAmount)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, currencyCode)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, isInStorePurchase)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, merchantIdentifier)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, referrerIdentifier)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, items)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, applicationMetadata)
        , offsetof(WebCore::ApplePayInstallmentConfiguration, retailChannel)
    >::value);
    encoder << instance.featureType;
    encoder << instance.merchandisingImageData;
    encoder << instance.openToBuyThresholdAmount;
    encoder << instance.bindingTotalAmount;
    encoder << instance.currencyCode;
    encoder << instance.isInStorePurchase;
    encoder << instance.merchantIdentifier;
    encoder << instance.referrerIdentifier;
    encoder << instance.items;
    encoder << instance.applicationMetadata;
    encoder << instance.retailChannel;
}

std::optional<WebCore::ApplePayInstallmentConfiguration> ArgumentCoder<WebCore::ApplePayInstallmentConfiguration>::decode(Decoder& decoder)
{
    auto featureType = decoder.decode<WebCore::ApplePaySetupFeatureType>();
    auto merchandisingImageData = decoder.decode<String>();
    auto openToBuyThresholdAmount = decoder.decode<String>();
    auto bindingTotalAmount = decoder.decode<String>();
    auto currencyCode = decoder.decode<String>();
    auto isInStorePurchase = decoder.decode<bool>();
    auto merchantIdentifier = decoder.decode<String>();
    auto referrerIdentifier = decoder.decode<String>();
    auto items = decoder.decode<Vector<WebCore::ApplePayInstallmentItem>>();
    auto applicationMetadata = decoder.decode<String>();
    auto retailChannel = decoder.decode<WebCore::ApplePayInstallmentRetailChannel>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayInstallmentConfiguration {
            WTFMove(*featureType),
            WTFMove(*merchandisingImageData),
            WTFMove(*openToBuyThresholdAmount),
            WTFMove(*bindingTotalAmount),
            WTFMove(*currencyCode),
            WTFMove(*isInStorePurchase),
            WTFMove(*merchantIdentifier),
            WTFMove(*referrerIdentifier),
            WTFMove(*items),
            WTFMove(*applicationMetadata),
            WTFMove(*retailChannel)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_INSTALLMENTS)
void ArgumentCoder<WebCore::PaymentInstallmentConfiguration>::encode(Encoder& encoder, const WebCore::PaymentInstallmentConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applePayInstallmentConfiguration())>, std::optional<WebCore::ApplePayInstallmentConfiguration>>);
    encoder << instance.applePayInstallmentConfiguration();
}

std::optional<WebCore::PaymentInstallmentConfiguration> ArgumentCoder<WebCore::PaymentInstallmentConfiguration>::decode(Decoder& decoder)
{
    auto applePayInstallmentConfiguration = decoder.decode<std::optional<WebCore::ApplePayInstallmentConfiguration>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PaymentInstallmentConfiguration {
            WTFMove(*applePayInstallmentConfiguration)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_INSTALLMENTS)
void ArgumentCoder<WebCore::ApplePayInstallmentItem>::encode(Encoder& encoder, const WebCore::ApplePayInstallmentItem& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ApplePayInstallmentItemType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.amount)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.currencyCode)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.programIdentifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.apr)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.programTerms)>, String>);
    struct ShouldBeSameSizeAsApplePayInstallmentItem : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplePayInstallmentItem>, false> {
        WebCore::ApplePayInstallmentItemType type;
        String amount;
        String currencyCode;
        String programIdentifier;
        String apr;
        String programTerms;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplePayInstallmentItem) == sizeof(WebCore::ApplePayInstallmentItem));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayInstallmentItem, type)
        , offsetof(WebCore::ApplePayInstallmentItem, amount)
        , offsetof(WebCore::ApplePayInstallmentItem, currencyCode)
        , offsetof(WebCore::ApplePayInstallmentItem, programIdentifier)
        , offsetof(WebCore::ApplePayInstallmentItem, apr)
        , offsetof(WebCore::ApplePayInstallmentItem, programTerms)
    >::value);
    encoder << instance.type;
    encoder << instance.amount;
    encoder << instance.currencyCode;
    encoder << instance.programIdentifier;
    encoder << instance.apr;
    encoder << instance.programTerms;
}

std::optional<WebCore::ApplePayInstallmentItem> ArgumentCoder<WebCore::ApplePayInstallmentItem>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::ApplePayInstallmentItemType>();
    auto amount = decoder.decode<String>();
    auto currencyCode = decoder.decode<String>();
    auto programIdentifier = decoder.decode<String>();
    auto apr = decoder.decode<String>();
    auto programTerms = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayInstallmentItem {
            WTFMove(*type),
            WTFMove(*amount),
            WTFMove(*currencyCode),
            WTFMove(*programIdentifier),
            WTFMove(*apr),
            WTFMove(*programTerms)
        }
    };
}

#endif

#if ENABLE(APPLE_PAY_COUPON_CODE)
void ArgumentCoder<WebCore::ApplePayCouponCodeUpdate>::encode(Encoder& encoder, const WebCore::ApplePayCouponCodeUpdate& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newTotal)>, WebCore::ApplePayLineItem>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newLineItems)>, Vector<WebCore::ApplePayLineItem>>);
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newRecurringPaymentRequest)>, std::optional<WebCore::ApplePayRecurringPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newAutomaticReloadPaymentRequest)>, std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>);
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newMultiTokenContexts)>, std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>);
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newDeferredPaymentRequest)>, std::optional<WebCore::ApplePayDeferredPaymentRequest>>);
#endif
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newTotal)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newLineItems)
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newRecurringPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newAutomaticReloadPaymentRequest)
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newMultiTokenContexts)
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
        , offsetof(WebCore::ApplePayDetailsUpdateBase, newDeferredPaymentRequest)
#endif
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.errors)>, Vector<RefPtr<WebCore::ApplePayError>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.newShippingMethods)>, Vector<WebCore::ApplePayShippingMethod>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplePayCouponCodeUpdate, errors)
        , offsetof(WebCore::ApplePayCouponCodeUpdate, newShippingMethods)
    >::value);
    encoder << instance.newTotal;
    encoder << instance.newLineItems;
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    encoder << instance.newRecurringPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    encoder << instance.newAutomaticReloadPaymentRequest;
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    encoder << instance.newMultiTokenContexts;
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    encoder << instance.newDeferredPaymentRequest;
#endif
    encoder << instance.errors;
    encoder << instance.newShippingMethods;
}

std::optional<WebCore::ApplePayCouponCodeUpdate> ArgumentCoder<WebCore::ApplePayCouponCodeUpdate>::decode(Decoder& decoder)
{
    auto newTotal = decoder.decode<WebCore::ApplePayLineItem>();
    auto newLineItems = decoder.decode<Vector<WebCore::ApplePayLineItem>>();
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    auto newRecurringPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayRecurringPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    auto newAutomaticReloadPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayAutomaticReloadPaymentRequest>>();
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    auto newMultiTokenContexts = decoder.decode<std::optional<Vector<WebCore::ApplePayPaymentTokenContext>>>();
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    auto newDeferredPaymentRequest = decoder.decode<std::optional<WebCore::ApplePayDeferredPaymentRequest>>();
#endif
    auto errors = decoder.decode<Vector<RefPtr<WebCore::ApplePayError>>>();
    auto newShippingMethods = decoder.decode<Vector<WebCore::ApplePayShippingMethod>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayCouponCodeUpdate {
            WebCore::ApplePayDetailsUpdateBase {
                WTFMove(*newTotal),
                WTFMove(*newLineItems),
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
                WTFMove(*newRecurringPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
                WTFMove(*newAutomaticReloadPaymentRequest),
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
                WTFMove(*newMultiTokenContexts),
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
                WTFMove(*newDeferredPaymentRequest)
#endif
            },
            WTFMove(*errors),
            WTFMove(*newShippingMethods)
        }
    };
}

#endif

#if ENABLE(APPLICATION_MANIFEST)
void ArgumentCoder<WebCore::ApplicationManifest::Icon>::encode(Encoder& encoder, const WebCore::ApplicationManifest::Icon& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.src)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sizes)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.purposes)>, OptionSet<WebCore::ApplicationManifest::Icon::Purpose>>);
    struct ShouldBeSameSizeAsIcon : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplicationManifest::Icon>, false> {
        URL src;
        Vector<String> sizes;
        String type;
        OptionSet<WebCore::ApplicationManifest::Icon::Purpose> purposes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsIcon) == sizeof(WebCore::ApplicationManifest::Icon));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplicationManifest::Icon, src)
        , offsetof(WebCore::ApplicationManifest::Icon, sizes)
        , offsetof(WebCore::ApplicationManifest::Icon, type)
        , offsetof(WebCore::ApplicationManifest::Icon, purposes)
    >::value);
    encoder << instance.src;
    encoder << instance.sizes;
    encoder << instance.type;
    encoder << instance.purposes;
}

std::optional<WebCore::ApplicationManifest::Icon> ArgumentCoder<WebCore::ApplicationManifest::Icon>::decode(Decoder& decoder)
{
    auto src = decoder.decode<URL>();
    auto sizes = decoder.decode<Vector<String>>();
    auto type = decoder.decode<String>();
    auto purposes = decoder.decode<OptionSet<WebCore::ApplicationManifest::Icon::Purpose>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplicationManifest::Icon {
            WTFMove(*src),
            WTFMove(*sizes),
            WTFMove(*type),
            WTFMove(*purposes)
        }
    };
}

#endif

#if ENABLE(APPLICATION_MANIFEST)
void ArgumentCoder<WebCore::ApplicationManifest>::encode(Encoder& encoder, const WebCore::ApplicationManifest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shortName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.description)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scope)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.display)>, WebCore::ApplicationManifest::Display>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.orientation)>, std::optional<WebCore::ScreenOrientationLockType>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.backgroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.themeColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.icons)>, Vector<WebCore::ApplicationManifest::Icon>>);
    struct ShouldBeSameSizeAsApplicationManifest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ApplicationManifest>, false> {
        String name;
        String shortName;
        String description;
        URL scope;
        WebCore::ApplicationManifest::Display display;
        std::optional<WebCore::ScreenOrientationLockType> orientation;
        URL startURL;
        URL id;
        WebCore::Color backgroundColor;
        WebCore::Color themeColor;
        Vector<WebCore::ApplicationManifest::Icon> icons;
    };
    static_assert(sizeof(ShouldBeSameSizeAsApplicationManifest) == sizeof(WebCore::ApplicationManifest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ApplicationManifest, name)
        , offsetof(WebCore::ApplicationManifest, shortName)
        , offsetof(WebCore::ApplicationManifest, description)
        , offsetof(WebCore::ApplicationManifest, scope)
        , offsetof(WebCore::ApplicationManifest, display)
        , offsetof(WebCore::ApplicationManifest, orientation)
        , offsetof(WebCore::ApplicationManifest, startURL)
        , offsetof(WebCore::ApplicationManifest, id)
        , offsetof(WebCore::ApplicationManifest, backgroundColor)
        , offsetof(WebCore::ApplicationManifest, themeColor)
        , offsetof(WebCore::ApplicationManifest, icons)
    >::value);
    encoder << instance.name;
    encoder << instance.shortName;
    encoder << instance.description;
    encoder << instance.scope;
    encoder << instance.display;
    encoder << instance.orientation;
    encoder << instance.startURL;
    encoder << instance.id;
    encoder << instance.backgroundColor;
    encoder << instance.themeColor;
    encoder << instance.icons;
}

std::optional<WebCore::ApplicationManifest> ArgumentCoder<WebCore::ApplicationManifest>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto shortName = decoder.decode<String>();
    auto description = decoder.decode<String>();
    auto scope = decoder.decode<URL>();
    auto display = decoder.decode<WebCore::ApplicationManifest::Display>();
    auto orientation = decoder.decode<std::optional<WebCore::ScreenOrientationLockType>>();
    auto startURL = decoder.decode<URL>();
    auto id = decoder.decode<URL>();
    auto backgroundColor = decoder.decode<WebCore::Color>();
    auto themeColor = decoder.decode<WebCore::Color>();
    auto icons = decoder.decode<Vector<WebCore::ApplicationManifest::Icon>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplicationManifest {
            WTFMove(*name),
            WTFMove(*shortName),
            WTFMove(*description),
            WTFMove(*scope),
            WTFMove(*display),
            WTFMove(*orientation),
            WTFMove(*startURL),
            WTFMove(*id),
            WTFMove(*backgroundColor),
            WTFMove(*themeColor),
            WTFMove(*icons)
        }
    };
}

#endif

void ArgumentCoder<WebCore::RetrieveRecordsOptions>::encode(Encoder& encoder, const WebCore::RetrieveRecordsOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.request)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginEmbedderPolicy)>, WebCore::CrossOriginEmbedderPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceOrigin)>, Ref<WebCore::SecurityOrigin>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoreSearch)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoreMethod)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoreVary)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldProvideResponse)>, bool>);
    struct ShouldBeSameSizeAsRetrieveRecordsOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RetrieveRecordsOptions>, false> {
        WebCore::ResourceRequest request;
        WebCore::CrossOriginEmbedderPolicy crossOriginEmbedderPolicy;
        Ref<WebCore::SecurityOrigin> sourceOrigin;
        bool ignoreSearch;
        bool ignoreMethod;
        bool ignoreVary;
        bool shouldProvideResponse;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRetrieveRecordsOptions) == sizeof(WebCore::RetrieveRecordsOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RetrieveRecordsOptions, request)
        , offsetof(WebCore::RetrieveRecordsOptions, crossOriginEmbedderPolicy)
        , offsetof(WebCore::RetrieveRecordsOptions, sourceOrigin)
        , offsetof(WebCore::RetrieveRecordsOptions, ignoreSearch)
        , offsetof(WebCore::RetrieveRecordsOptions, ignoreMethod)
        , offsetof(WebCore::RetrieveRecordsOptions, ignoreVary)
        , offsetof(WebCore::RetrieveRecordsOptions, shouldProvideResponse)
    >::value);
    encoder << instance.request;
    encoder << instance.crossOriginEmbedderPolicy;
    encoder << instance.sourceOrigin;
    encoder << instance.ignoreSearch;
    encoder << instance.ignoreMethod;
    encoder << instance.ignoreVary;
    encoder << instance.shouldProvideResponse;
}

std::optional<WebCore::RetrieveRecordsOptions> ArgumentCoder<WebCore::RetrieveRecordsOptions>::decode(Decoder& decoder)
{
    auto request = decoder.decode<WebCore::ResourceRequest>();
    auto crossOriginEmbedderPolicy = decoder.decode<WebCore::CrossOriginEmbedderPolicy>();
    auto sourceOrigin = decoder.decode<Ref<WebCore::SecurityOrigin>>();
    auto ignoreSearch = decoder.decode<bool>();
    auto ignoreMethod = decoder.decode<bool>();
    auto ignoreVary = decoder.decode<bool>();
    auto shouldProvideResponse = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RetrieveRecordsOptions {
            WTFMove(*request),
            WTFMove(*crossOriginEmbedderPolicy),
            WTFMove(*sourceOrigin),
            WTFMove(*ignoreSearch),
            WTFMove(*ignoreMethod),
            WTFMove(*ignoreVary),
            WTFMove(*shouldProvideResponse)
        }
    };
}

void ArgumentCoder<WebCore::ContactInfo>::encode(Encoder& encoder, const WebCore::ContactInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.email)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tel)>, Vector<String>>);
    struct ShouldBeSameSizeAsContactInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ContactInfo>, false> {
        Vector<String> name;
        Vector<String> email;
        Vector<String> tel;
    };
    static_assert(sizeof(ShouldBeSameSizeAsContactInfo) == sizeof(WebCore::ContactInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ContactInfo, name)
        , offsetof(WebCore::ContactInfo, email)
        , offsetof(WebCore::ContactInfo, tel)
    >::value);
    encoder << instance.name;
    encoder << instance.email;
    encoder << instance.tel;
}

std::optional<WebCore::ContactInfo> ArgumentCoder<WebCore::ContactInfo>::decode(Decoder& decoder)
{
    auto name = decoder.decode<Vector<String>>();
    auto email = decoder.decode<Vector<String>>();
    auto tel = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ContactInfo {
            WTFMove(*name),
            WTFMove(*email),
            WTFMove(*tel)
        }
    };
}

void ArgumentCoder<WebCore::ContactsRequestData>::encode(Encoder& encoder, const WebCore::ContactsRequestData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.properties)>, Vector<WebCore::ContactProperty>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multiple)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, String>);
    struct ShouldBeSameSizeAsContactsRequestData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ContactsRequestData>, false> {
        Vector<WebCore::ContactProperty> properties;
        bool multiple;
        String url;
    };
    static_assert(sizeof(ShouldBeSameSizeAsContactsRequestData) == sizeof(WebCore::ContactsRequestData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ContactsRequestData, properties)
        , offsetof(WebCore::ContactsRequestData, multiple)
        , offsetof(WebCore::ContactsRequestData, url)
    >::value);
    encoder << instance.properties;
    encoder << instance.multiple;
    encoder << instance.url;
}

std::optional<WebCore::ContactsRequestData> ArgumentCoder<WebCore::ContactsRequestData>::decode(Decoder& decoder)
{
    auto properties = decoder.decode<Vector<WebCore::ContactProperty>>();
    auto multiple = decoder.decode<bool>();
    auto url = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ContactsRequestData {
            WTFMove(*properties),
            WTFMove(*multiple),
            WTFMove(*url)
        }
    };
}

#if ENABLE(MEDIA_SESSION)
void ArgumentCoder<WebCore::MediaPositionState>::encode(Encoder& encoder, const WebCore::MediaPositionState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.playbackRate)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position)>, double>);
    struct ShouldBeSameSizeAsMediaPositionState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaPositionState>, false> {
        double duration;
        double playbackRate;
        double position;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaPositionState) == sizeof(WebCore::MediaPositionState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaPositionState, duration)
        , offsetof(WebCore::MediaPositionState, playbackRate)
        , offsetof(WebCore::MediaPositionState, position)
    >::value);
    encoder << instance.duration;
    encoder << instance.playbackRate;
    encoder << instance.position;
}

std::optional<WebCore::MediaPositionState> ArgumentCoder<WebCore::MediaPositionState>::decode(Decoder& decoder)
{
    auto duration = decoder.decode<double>();
    auto playbackRate = decoder.decode<double>();
    auto position = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaPositionState {
            WTFMove(*duration),
            WTFMove(*playbackRate),
            WTFMove(*position)
        }
    };
}

#endif

#if ENABLE(WEB_RTC)
void ArgumentCoder<WebCore::DetachedRTCDataChannel>::encode(Encoder& encoder, const WebCore::DetachedRTCDataChannel& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::RTCDataChannelIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.options)>, WebCore::RTCDataChannelInit>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.state)>, WebCore::RTCDataChannelState>);
    struct ShouldBeSameSizeAsDetachedRTCDataChannel : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DetachedRTCDataChannel>, false> {
        WebCore::RTCDataChannelIdentifier identifier;
        String label;
        WebCore::RTCDataChannelInit options;
        WebCore::RTCDataChannelState state;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDetachedRTCDataChannel) == sizeof(WebCore::DetachedRTCDataChannel));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DetachedRTCDataChannel, identifier)
        , offsetof(WebCore::DetachedRTCDataChannel, label)
        , offsetof(WebCore::DetachedRTCDataChannel, options)
        , offsetof(WebCore::DetachedRTCDataChannel, state)
    >::value);
    encoder << instance.identifier;
    encoder << instance.label;
    encoder << instance.options;
    encoder << instance.state;
}

std::optional<WebCore::DetachedRTCDataChannel> ArgumentCoder<WebCore::DetachedRTCDataChannel>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::RTCDataChannelIdentifier>();
    auto label = decoder.decode<String>();
    auto options = decoder.decode<WebCore::RTCDataChannelInit>();
    auto state = decoder.decode<WebCore::RTCDataChannelState>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DetachedRTCDataChannel {
            WTFMove(*identifier),
            WTFMove(*label),
            WTFMove(*options),
            WTFMove(*state)
        }
    };
}

#endif

#if ENABLE(WEB_CODECS)
void ArgumentCoder<WebCore::WebCodecsEncodedVideoChunkData>::encode(Encoder& encoder, const WebCore::WebCodecsEncodedVideoChunkData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebCodecsEncodedVideoChunkType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp)>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration)>, std::optional<uint64_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffer)>, Vector<uint8_t>>);
    struct ShouldBeSameSizeAsWebCodecsEncodedVideoChunkData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WebCodecsEncodedVideoChunkData>, false> {
        WebCore::WebCodecsEncodedVideoChunkType type;
        int64_t timestamp;
        std::optional<uint64_t> duration;
        Vector<uint8_t> buffer;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebCodecsEncodedVideoChunkData) == sizeof(WebCore::WebCodecsEncodedVideoChunkData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WebCodecsEncodedVideoChunkData, type)
        , offsetof(WebCore::WebCodecsEncodedVideoChunkData, timestamp)
        , offsetof(WebCore::WebCodecsEncodedVideoChunkData, duration)
        , offsetof(WebCore::WebCodecsEncodedVideoChunkData, buffer)
    >::value);
    encoder << instance.type;
    encoder << instance.timestamp;
    encoder << instance.duration;
    encoder << instance.buffer;
}

std::optional<WebCore::WebCodecsEncodedVideoChunkData> ArgumentCoder<WebCore::WebCodecsEncodedVideoChunkData>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::WebCodecsEncodedVideoChunkType>();
    auto timestamp = decoder.decode<int64_t>();
    auto duration = decoder.decode<std::optional<uint64_t>>();
    auto buffer = decoder.decode<Vector<uint8_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WebCodecsEncodedVideoChunkData {
            WTFMove(*type),
            WTFMove(*timestamp),
            WTFMove(*duration),
            WTFMove(*buffer)
        }
    };
}

#endif

void ArgumentCoder<WebCore::HTMLModelElementCamera>::encode(Encoder& encoder, const WebCore::HTMLModelElementCamera& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pitch)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.yaw)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scale)>, double>);
    struct ShouldBeSameSizeAsHTMLModelElementCamera : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::HTMLModelElementCamera>, false> {
        double pitch;
        double yaw;
        double scale;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHTMLModelElementCamera) == sizeof(WebCore::HTMLModelElementCamera));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::HTMLModelElementCamera, pitch)
        , offsetof(WebCore::HTMLModelElementCamera, yaw)
        , offsetof(WebCore::HTMLModelElementCamera, scale)
    >::value);
    encoder << instance.pitch;
    encoder << instance.yaw;
    encoder << instance.scale;
}

std::optional<WebCore::HTMLModelElementCamera> ArgumentCoder<WebCore::HTMLModelElementCamera>::decode(Decoder& decoder)
{
    auto pitch = decoder.decode<double>();
    auto yaw = decoder.decode<double>();
    auto scale = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::HTMLModelElementCamera {
            WTFMove(*pitch),
            WTFMove(*yaw),
            WTFMove(*scale)
        }
    };
}

void ArgumentCoder<WebCore::NotificationData>::encode(Encoder& encoder, const WebCore::NotificationData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.title)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.body)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.iconURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tag)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.language)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction)>, WebCore::NotificationDirection>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originString)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serviceWorkerRegistrationURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.notificationID)>, WTF::UUID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contextIdentifier)>, WebCore::ScriptExecutionContextIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceSession)>, PAL::SessionID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.creationTime)>, MonotonicTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data)>, Vector<uint8_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.silent)>, std::optional<bool>>);
    struct ShouldBeSameSizeAsNotificationData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::NotificationData>, false> {
        String title;
        String body;
        String iconURL;
        String tag;
        String language;
        WebCore::NotificationDirection direction;
        String originString;
        URL serviceWorkerRegistrationURL;
        WTF::UUID notificationID;
        WebCore::ScriptExecutionContextIdentifier contextIdentifier;
        PAL::SessionID sourceSession;
        MonotonicTime creationTime;
        Vector<uint8_t> data;
        std::optional<bool> silent;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNotificationData) == sizeof(WebCore::NotificationData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::NotificationData, title)
        , offsetof(WebCore::NotificationData, body)
        , offsetof(WebCore::NotificationData, iconURL)
        , offsetof(WebCore::NotificationData, tag)
        , offsetof(WebCore::NotificationData, language)
        , offsetof(WebCore::NotificationData, direction)
        , offsetof(WebCore::NotificationData, originString)
        , offsetof(WebCore::NotificationData, serviceWorkerRegistrationURL)
        , offsetof(WebCore::NotificationData, notificationID)
        , offsetof(WebCore::NotificationData, contextIdentifier)
        , offsetof(WebCore::NotificationData, sourceSession)
        , offsetof(WebCore::NotificationData, creationTime)
        , offsetof(WebCore::NotificationData, data)
        , offsetof(WebCore::NotificationData, silent)
    >::value);
    encoder << instance.title;
    encoder << instance.body;
    encoder << instance.iconURL;
    encoder << instance.tag;
    encoder << instance.language;
    encoder << instance.direction;
    encoder << instance.originString;
    encoder << instance.serviceWorkerRegistrationURL;
    encoder << instance.notificationID;
    encoder << instance.contextIdentifier;
    encoder << instance.sourceSession;
    encoder << instance.creationTime;
    encoder << instance.data;
    encoder << instance.silent;
}

std::optional<WebCore::NotificationData> ArgumentCoder<WebCore::NotificationData>::decode(Decoder& decoder)
{
    auto title = decoder.decode<String>();
    auto body = decoder.decode<String>();
    auto iconURL = decoder.decode<String>();
    auto tag = decoder.decode<String>();
    auto language = decoder.decode<String>();
    auto direction = decoder.decode<WebCore::NotificationDirection>();
    auto originString = decoder.decode<String>();
    auto serviceWorkerRegistrationURL = decoder.decode<URL>();
    auto notificationID = decoder.decode<WTF::UUID>();
    auto contextIdentifier = decoder.decode<WebCore::ScriptExecutionContextIdentifier>();
    auto sourceSession = decoder.decode<PAL::SessionID>();
    auto creationTime = decoder.decode<MonotonicTime>();
    auto data = decoder.decode<Vector<uint8_t>>();
    auto silent = decoder.decode<std::optional<bool>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::NotificationData {
            WTFMove(*title),
            WTFMove(*body),
            WTFMove(*iconURL),
            WTFMove(*tag),
            WTFMove(*language),
            WTFMove(*direction),
            WTFMove(*originString),
            WTFMove(*serviceWorkerRegistrationURL),
            WTFMove(*notificationID),
            WTFMove(*contextIdentifier),
            WTFMove(*sourceSession),
            WTFMove(*creationTime),
            WTFMove(*data),
            WTFMove(*silent)
        }
    };
}

void ArgumentCoder<WebCore::PermissionDescriptor>::encode(Encoder& encoder, const WebCore::PermissionDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, WebCore::PermissionName>);
    struct ShouldBeSameSizeAsPermissionDescriptor : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PermissionDescriptor>, false> {
        WebCore::PermissionName name;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPermissionDescriptor) == sizeof(WebCore::PermissionDescriptor));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PermissionDescriptor, name)
    >::value);
    encoder << instance.name;
}

std::optional<WebCore::PermissionDescriptor> ArgumentCoder<WebCore::PermissionDescriptor>::decode(Decoder& decoder)
{
    auto name = decoder.decode<WebCore::PermissionName>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PermissionDescriptor {
            WTFMove(*name)
        }
    };
}

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::PushSubscriptionData>::encode(Encoder& encoder, const WebCore::PushSubscriptionData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::PushSubscriptionIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endpoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expirationTime)>, std::optional<WebCore::EpochTimeStamp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serverVAPIDPublicKey)>, Vector<uint8_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientECDHPublicKey)>, Vector<uint8_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sharedAuthenticationSecret)>, Vector<uint8_t>>);
    struct ShouldBeSameSizeAsPushSubscriptionData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PushSubscriptionData>, false> {
        WebCore::PushSubscriptionIdentifier identifier;
        String endpoint;
        std::optional<WebCore::EpochTimeStamp> expirationTime;
        Vector<uint8_t> serverVAPIDPublicKey;
        Vector<uint8_t> clientECDHPublicKey;
        Vector<uint8_t> sharedAuthenticationSecret;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPushSubscriptionData) == sizeof(WebCore::PushSubscriptionData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PushSubscriptionData, identifier)
        , offsetof(WebCore::PushSubscriptionData, endpoint)
        , offsetof(WebCore::PushSubscriptionData, expirationTime)
        , offsetof(WebCore::PushSubscriptionData, serverVAPIDPublicKey)
        , offsetof(WebCore::PushSubscriptionData, clientECDHPublicKey)
        , offsetof(WebCore::PushSubscriptionData, sharedAuthenticationSecret)
    >::value);
    encoder << instance.identifier;
    encoder << instance.endpoint;
    encoder << instance.expirationTime;
    encoder << instance.serverVAPIDPublicKey;
    encoder << instance.clientECDHPublicKey;
    encoder << instance.sharedAuthenticationSecret;
}

std::optional<WebCore::PushSubscriptionData> ArgumentCoder<WebCore::PushSubscriptionData>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::PushSubscriptionIdentifier>();
    auto endpoint = decoder.decode<String>();
    auto expirationTime = decoder.decode<std::optional<WebCore::EpochTimeStamp>>();
    auto serverVAPIDPublicKey = decoder.decode<Vector<uint8_t>>();
    auto clientECDHPublicKey = decoder.decode<Vector<uint8_t>>();
    auto sharedAuthenticationSecret = decoder.decode<Vector<uint8_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PushSubscriptionData {
            WTFMove(*identifier),
            WTFMove(*endpoint),
            WTFMove(*expirationTime),
            WTFMove(*serverVAPIDPublicKey),
            WTFMove(*clientECDHPublicKey),
            WTFMove(*sharedAuthenticationSecret)
        }
    };
}

#endif

void ArgumentCoder<WebCore::DeprecationReportBody>::encode(Encoder& encoder, const WebCore::DeprecationReportBody& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.anticipatedRemoval())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceFile())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lineNumber())>, std::optional<unsigned>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.columnNumber())>, std::optional<unsigned>>);
    encoder << instance.id();
    encoder << instance.anticipatedRemoval();
    encoder << instance.message();
    encoder << instance.sourceFile();
    encoder << instance.lineNumber();
    encoder << instance.columnNumber();
}

std::optional<Ref<WebCore::DeprecationReportBody>> ArgumentCoder<WebCore::DeprecationReportBody>::decode(Decoder& decoder)
{
    auto id = decoder.decode<String>();
    auto anticipatedRemoval = decoder.decode<WallTime>();
    auto message = decoder.decode<String>();
    auto sourceFile = decoder.decode<String>();
    auto lineNumber = decoder.decode<std::optional<unsigned>>();
    auto columnNumber = decoder.decode<std::optional<unsigned>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DeprecationReportBody::create(
            WTFMove(*id),
            WTFMove(*anticipatedRemoval),
            WTFMove(*message),
            WTFMove(*sourceFile),
            WTFMove(*lineNumber),
            WTFMove(*columnNumber)
        )
    };
}

void ArgumentCoder<WebCore::Report>::encode(Encoder& encoder, const WebCore::Report& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.body())>, RefPtr<WebCore::ReportBody>>);
    encoder << instance.type();
    encoder << instance.url();
    encoder << instance.body();
}

std::optional<Ref<WebCore::Report>> ArgumentCoder<WebCore::Report>::decode(Decoder& decoder)
{
    auto type = decoder.decode<String>();
    auto url = decoder.decode<String>();
    auto body = decoder.decode<RefPtr<WebCore::ReportBody>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Report::create(
            WTFMove(*type),
            WTFMove(*url),
            WTFMove(*body)
        )
    };
}

void ArgumentCoder<WebCore::TestReportBody>::encode(Encoder& encoder, const WebCore::TestReportBody& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message())>, String>);
    encoder << instance.message();
}

std::optional<Ref<WebCore::TestReportBody>> ArgumentCoder<WebCore::TestReportBody>::decode(Decoder& decoder)
{
    auto message = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TestReportBody::create(
            WTFMove(*message)
        )
    };
}

void ArgumentCoder<WebCore::SpeechRecognitionError>::encode(Encoder& encoder, const WebCore::SpeechRecognitionError& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::SpeechRecognitionErrorType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, String>);
    struct ShouldBeSameSizeAsSpeechRecognitionError : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SpeechRecognitionError>, false> {
        WebCore::SpeechRecognitionErrorType type;
        String message;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSpeechRecognitionError) == sizeof(WebCore::SpeechRecognitionError));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SpeechRecognitionError, type)
        , offsetof(WebCore::SpeechRecognitionError, message)
    >::value);
    encoder << instance.type;
    encoder << instance.message;
}

std::optional<WebCore::SpeechRecognitionError> ArgumentCoder<WebCore::SpeechRecognitionError>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::SpeechRecognitionErrorType>();
    auto message = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SpeechRecognitionError {
            WTFMove(*type),
            WTFMove(*message)
        }
    };
}

void ArgumentCoder<WebCore::SpeechRecognitionRequestInfo>::encode(Encoder& encoder, const WebCore::SpeechRecognitionRequestInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientIdentifier)>, WebCore::SpeechRecognitionConnectionClientIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lang)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.continuous)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.interimResults)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxAlternatives)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientOrigin)>, WebCore::ClientOrigin>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameIdentifier)>, WebCore::FrameIdentifier>);
    struct ShouldBeSameSizeAsSpeechRecognitionRequestInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SpeechRecognitionRequestInfo>, false> {
        WebCore::SpeechRecognitionConnectionClientIdentifier clientIdentifier;
        String lang;
        bool continuous;
        bool interimResults;
        uint64_t maxAlternatives;
        WebCore::ClientOrigin clientOrigin;
        WebCore::FrameIdentifier frameIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSpeechRecognitionRequestInfo) == sizeof(WebCore::SpeechRecognitionRequestInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SpeechRecognitionRequestInfo, clientIdentifier)
        , offsetof(WebCore::SpeechRecognitionRequestInfo, lang)
        , offsetof(WebCore::SpeechRecognitionRequestInfo, continuous)
        , offsetof(WebCore::SpeechRecognitionRequestInfo, interimResults)
        , offsetof(WebCore::SpeechRecognitionRequestInfo, maxAlternatives)
        , offsetof(WebCore::SpeechRecognitionRequestInfo, clientOrigin)
        , offsetof(WebCore::SpeechRecognitionRequestInfo, frameIdentifier)
    >::value);
    encoder << instance.clientIdentifier;
    encoder << instance.lang;
    encoder << instance.continuous;
    encoder << instance.interimResults;
    encoder << instance.maxAlternatives;
    encoder << instance.clientOrigin;
    encoder << instance.frameIdentifier;
}

std::optional<WebCore::SpeechRecognitionRequestInfo> ArgumentCoder<WebCore::SpeechRecognitionRequestInfo>::decode(Decoder& decoder)
{
    auto clientIdentifier = decoder.decode<WebCore::SpeechRecognitionConnectionClientIdentifier>();
    auto lang = decoder.decode<String>();
    auto continuous = decoder.decode<bool>();
    auto interimResults = decoder.decode<bool>();
    auto maxAlternatives = decoder.decode<uint64_t>();
    auto clientOrigin = decoder.decode<WebCore::ClientOrigin>();
    auto frameIdentifier = decoder.decode<WebCore::FrameIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SpeechRecognitionRequestInfo {
            WTFMove(*clientIdentifier),
            WTFMove(*lang),
            WTFMove(*continuous),
            WTFMove(*interimResults),
            WTFMove(*maxAlternatives),
            WTFMove(*clientOrigin),
            WTFMove(*frameIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::SpeechRecognitionAlternativeData>::encode(Encoder& encoder, const WebCore::SpeechRecognitionAlternativeData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transcript)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.confidence)>, double>);
    struct ShouldBeSameSizeAsSpeechRecognitionAlternativeData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SpeechRecognitionAlternativeData>, false> {
        String transcript;
        double confidence;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSpeechRecognitionAlternativeData) == sizeof(WebCore::SpeechRecognitionAlternativeData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SpeechRecognitionAlternativeData, transcript)
        , offsetof(WebCore::SpeechRecognitionAlternativeData, confidence)
    >::value);
    encoder << instance.transcript;
    encoder << instance.confidence;
}

std::optional<WebCore::SpeechRecognitionAlternativeData> ArgumentCoder<WebCore::SpeechRecognitionAlternativeData>::decode(Decoder& decoder)
{
    auto transcript = decoder.decode<String>();
    auto confidence = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SpeechRecognitionAlternativeData {
            WTFMove(*transcript),
            WTFMove(*confidence)
        }
    };
}

void ArgumentCoder<WebCore::SpeechRecognitionResultData>::encode(Encoder& encoder, const WebCore::SpeechRecognitionResultData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alternatives)>, Vector<WebCore::SpeechRecognitionAlternativeData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFinal)>, bool>);
    struct ShouldBeSameSizeAsSpeechRecognitionResultData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SpeechRecognitionResultData>, false> {
        Vector<WebCore::SpeechRecognitionAlternativeData> alternatives;
        bool isFinal;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSpeechRecognitionResultData) == sizeof(WebCore::SpeechRecognitionResultData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SpeechRecognitionResultData, alternatives)
        , offsetof(WebCore::SpeechRecognitionResultData, isFinal)
    >::value);
    encoder << instance.alternatives;
    encoder << instance.isFinal;
}

std::optional<WebCore::SpeechRecognitionResultData> ArgumentCoder<WebCore::SpeechRecognitionResultData>::decode(Decoder& decoder)
{
    auto alternatives = decoder.decode<Vector<WebCore::SpeechRecognitionAlternativeData>>();
    auto isFinal = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SpeechRecognitionResultData {
            WTFMove(*alternatives),
            WTFMove(*isFinal)
        }
    };
}

void ArgumentCoder<WebCore::WebLockManagerSnapshot>::encode(Encoder& encoder, const WebCore::WebLockManagerSnapshot& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.held)>, Vector<WebCore::WebLockManagerSnapshot::Info>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pending)>, Vector<WebCore::WebLockManagerSnapshot::Info>>);
    struct ShouldBeSameSizeAsWebLockManagerSnapshot : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WebLockManagerSnapshot>, false> {
        Vector<WebCore::WebLockManagerSnapshot::Info> held;
        Vector<WebCore::WebLockManagerSnapshot::Info> pending;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebLockManagerSnapshot) == sizeof(WebCore::WebLockManagerSnapshot));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WebLockManagerSnapshot, held)
        , offsetof(WebCore::WebLockManagerSnapshot, pending)
    >::value);
    encoder << instance.held;
    encoder << instance.pending;
}

std::optional<WebCore::WebLockManagerSnapshot> ArgumentCoder<WebCore::WebLockManagerSnapshot>::decode(Decoder& decoder)
{
    auto held = decoder.decode<Vector<WebCore::WebLockManagerSnapshot::Info>>();
    auto pending = decoder.decode<Vector<WebCore::WebLockManagerSnapshot::Info>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WebLockManagerSnapshot {
            WTFMove(*held),
            WTFMove(*pending)
        }
    };
}

void ArgumentCoder<WebCore::WebLockManagerSnapshot::Info>::encode(Encoder& encoder, const WebCore::WebLockManagerSnapshot::Info& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mode)>, WebCore::WebLockMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientId)>, String>);
    struct ShouldBeSameSizeAsInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WebLockManagerSnapshot::Info>, false> {
        String name;
        WebCore::WebLockMode mode;
        String clientId;
    };
    static_assert(sizeof(ShouldBeSameSizeAsInfo) == sizeof(WebCore::WebLockManagerSnapshot::Info));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WebLockManagerSnapshot::Info, name)
        , offsetof(WebCore::WebLockManagerSnapshot::Info, mode)
        , offsetof(WebCore::WebLockManagerSnapshot::Info, clientId)
    >::value);
    encoder << instance.name;
    encoder << instance.mode;
    encoder << instance.clientId;
}

std::optional<WebCore::WebLockManagerSnapshot::Info> ArgumentCoder<WebCore::WebLockManagerSnapshot::Info>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto mode = decoder.decode<WebCore::WebLockMode>();
    auto clientId = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WebLockManagerSnapshot::Info {
            WTFMove(*name),
            WTFMove(*mode),
            WTFMove(*clientId)
        }
    };
}

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs>::encode(Encoder& encoder, const WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.support)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.read)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.write)>, std::optional<WebCore::BufferSource>>);
    struct ShouldBeSameSizeAsLargeBlobInputs : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs>, false> {
        String support;
        std::optional<bool> read;
        std::optional<WebCore::BufferSource> write;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLargeBlobInputs) == sizeof(WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs, support)
        , offsetof(WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs, read)
        , offsetof(WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs, write)
    >::value);
    encoder << instance.support;
    encoder << instance.read;
    encoder << instance.write;
}

std::optional<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs> ArgumentCoder<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs>::decode(Decoder& decoder)
{
    auto support = decoder.decode<String>();
    auto read = decoder.decode<std::optional<bool>>();
    auto write = decoder.decode<std::optional<WebCore::BufferSource>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs {
            WTFMove(*support),
            WTFMove(*read),
            WTFMove(*write)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::AuthenticationExtensionsClientInputs>::encode(Encoder& encoder, const WebCore::AuthenticationExtensionsClientInputs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.appid)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.credProps)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.largeBlob)>, std::optional<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs>>);
    struct ShouldBeSameSizeAsAuthenticationExtensionsClientInputs : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AuthenticationExtensionsClientInputs>, false> {
        String appid;
        bool credProps;
        std::optional<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs> largeBlob;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAuthenticationExtensionsClientInputs) == sizeof(WebCore::AuthenticationExtensionsClientInputs));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AuthenticationExtensionsClientInputs, appid)
        , offsetof(WebCore::AuthenticationExtensionsClientInputs, credProps)
        , offsetof(WebCore::AuthenticationExtensionsClientInputs, largeBlob)
    >::value);
    encoder << instance.appid;
    encoder << instance.largeBlob;
}

std::optional<WebCore::AuthenticationExtensionsClientInputs> ArgumentCoder<WebCore::AuthenticationExtensionsClientInputs>::decode(Decoder& decoder)
{
    auto appid = decoder.decode<String>();
    auto largeBlob = decoder.decode<std::optional<WebCore::AuthenticationExtensionsClientInputs::LargeBlobInputs>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    WebCore::AuthenticationExtensionsClientInputs result;
    result.appid = WTFMove(*appid);
    result.largeBlob = WTFMove(*largeBlob);
    return { WTFMove(result) };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput>::encode(Encoder& encoder, const WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rk)>, bool>);
    struct ShouldBeSameSizeAsCredentialPropertiesOutput : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput>, false> {
        bool rk;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCredentialPropertiesOutput) == sizeof(WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput, rk)
    >::value);
    encoder << instance.rk;
}

std::optional<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput> ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput>::decode(Decoder& decoder)
{
    auto rk = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput {
            WTFMove(*rk)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs>::encode(Encoder& encoder, const WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supported)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blob)>, RefPtr<JSC::ArrayBuffer>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.written)>, std::optional<bool>>);
    struct ShouldBeSameSizeAsLargeBlobOutputs : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs>, false> {
        std::optional<bool> supported;
        RefPtr<JSC::ArrayBuffer> blob;
        std::optional<bool> written;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLargeBlobOutputs) == sizeof(WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs, supported)
        , offsetof(WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs, blob)
        , offsetof(WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs, written)
    >::value);
    encoder << instance.supported;
    encoder << instance.blob;
    encoder << instance.written;
}

std::optional<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs> ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs>::decode(Decoder& decoder)
{
    auto supported = decoder.decode<std::optional<bool>>();
    auto blob = decoder.decode<RefPtr<JSC::ArrayBuffer>>();
    auto written = decoder.decode<std::optional<bool>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs {
            WTFMove(*supported),
            WTFMove(*blob),
            WTFMove(*written)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs>::encode(Encoder& encoder, const WebCore::AuthenticationExtensionsClientOutputs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.appid)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.credProps)>, std::optional<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.largeBlob)>, std::optional<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs>>);
    struct ShouldBeSameSizeAsAuthenticationExtensionsClientOutputs : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AuthenticationExtensionsClientOutputs>, false> {
        std::optional<bool> appid;
        std::optional<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput> credProps;
        std::optional<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs> largeBlob;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAuthenticationExtensionsClientOutputs) == sizeof(WebCore::AuthenticationExtensionsClientOutputs));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AuthenticationExtensionsClientOutputs, appid)
        , offsetof(WebCore::AuthenticationExtensionsClientOutputs, credProps)
        , offsetof(WebCore::AuthenticationExtensionsClientOutputs, largeBlob)
    >::value);
    encoder << instance.appid;
    encoder << instance.credProps;
    encoder << instance.largeBlob;
}

std::optional<WebCore::AuthenticationExtensionsClientOutputs> ArgumentCoder<WebCore::AuthenticationExtensionsClientOutputs>::decode(Decoder& decoder)
{
    auto appid = decoder.decode<std::optional<bool>>();
    auto credProps = decoder.decode<std::optional<WebCore::AuthenticationExtensionsClientOutputs::CredentialPropertiesOutput>>();
    auto largeBlob = decoder.decode<std::optional<WebCore::AuthenticationExtensionsClientOutputs::LargeBlobOutputs>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AuthenticationExtensionsClientOutputs {
            WTFMove(*appid),
            WTFMove(*credProps),
            WTFMove(*largeBlob)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::Parameters>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialCreationOptions::Parameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::PublicKeyCredentialType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alg)>, int64_t>);
    struct ShouldBeSameSizeAsParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PublicKeyCredentialCreationOptions::Parameters>, false> {
        WebCore::PublicKeyCredentialType type;
        int64_t alg;
    };
    static_assert(sizeof(ShouldBeSameSizeAsParameters) == sizeof(WebCore::PublicKeyCredentialCreationOptions::Parameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Parameters, type)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Parameters, alg)
    >::value);
    encoder << instance.type;
    encoder << instance.alg;
}

std::optional<WebCore::PublicKeyCredentialCreationOptions::Parameters> ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::Parameters>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::PublicKeyCredentialType>();
    auto alg = decoder.decode<int64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialCreationOptions::Parameters {
            WTFMove(*type),
            WTFMove(*alg)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authenticatorAttachment)>, std::optional<WebCore::AuthenticatorAttachment>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.residentKey)>, std::optional<WebCore::ResidentKeyRequirement>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requireResidentKey)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userVerification)>, WebCore::UserVerificationRequirement>);
    struct ShouldBeSameSizeAsAuthenticatorSelectionCriteria : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria>, false> {
        std::optional<WebCore::AuthenticatorAttachment> authenticatorAttachment;
        std::optional<WebCore::ResidentKeyRequirement> residentKey;
        bool requireResidentKey;
        WebCore::UserVerificationRequirement userVerification;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAuthenticatorSelectionCriteria) == sizeof(WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria, authenticatorAttachment)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria, residentKey)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria, requireResidentKey)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria, userVerification)
    >::value);
    encoder << instance.authenticatorAttachment;
    encoder << instance.residentKey;
    encoder << instance.requireResidentKey;
    encoder << instance.userVerification;
}

std::optional<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria> ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria>::decode(Decoder& decoder)
{
    auto authenticatorAttachment = decoder.decode<std::optional<WebCore::AuthenticatorAttachment>>();
    auto residentKey = decoder.decode<std::optional<WebCore::ResidentKeyRequirement>>();
    auto requireResidentKey = decoder.decode<bool>();
    auto userVerification = decoder.decode<WebCore::UserVerificationRequirement>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria {
            WTFMove(*authenticatorAttachment),
            WTFMove(*residentKey),
            WTFMove(*requireResidentKey),
            WTFMove(*userVerification)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::Entity>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialCreationOptions::Entity& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.icon)>, String>);
    struct ShouldBeSameSizeAsEntity : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PublicKeyCredentialCreationOptions::Entity>, false> {
        String name;
        String icon;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEntity) == sizeof(WebCore::PublicKeyCredentialCreationOptions::Entity));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Entity, name)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Entity, icon)
    >::value);
    encoder << instance.name;
    encoder << instance.icon;
}

std::optional<WebCore::PublicKeyCredentialCreationOptions::Entity> ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::Entity>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto icon = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialCreationOptions::Entity {
            WTFMove(*name),
            WTFMove(*icon)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::RpEntity>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialCreationOptions::RpEntity& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.icon)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Entity, name)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Entity, icon)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id)>, std::optional<String>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::RpEntity, id)
    >::value);
    encoder << instance.name;
    encoder << instance.icon;
    encoder << instance.id;
}

std::optional<WebCore::PublicKeyCredentialCreationOptions::RpEntity> ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::RpEntity>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto icon = decoder.decode<String>();
    auto id = decoder.decode<std::optional<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialCreationOptions::RpEntity {
            WebCore::PublicKeyCredentialCreationOptions::Entity {
                WTFMove(*name),
                WTFMove(*icon)
            },
            WTFMove(*id)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::UserEntity>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialCreationOptions::UserEntity& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.icon)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Entity, name)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::Entity, icon)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id)>, WebCore::BufferSource>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.displayName)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::UserEntity, id)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions::UserEntity, displayName)
    >::value);
    encoder << instance.name;
    encoder << instance.icon;
    encoder << instance.id;
    encoder << instance.displayName;
}

std::optional<WebCore::PublicKeyCredentialCreationOptions::UserEntity> ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions::UserEntity>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto icon = decoder.decode<String>();
    auto id = decoder.decode<WebCore::BufferSource>();
    auto displayName = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialCreationOptions::UserEntity {
            WebCore::PublicKeyCredentialCreationOptions::Entity {
                WTFMove(*name),
                WTFMove(*icon)
            },
            WTFMove(*id),
            WTFMove(*displayName)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::PublicKeyCredentialDescriptor>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::PublicKeyCredentialType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id)>, WebCore::BufferSource>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transports)>, Vector<WebCore::AuthenticatorTransport>>);
    struct ShouldBeSameSizeAsPublicKeyCredentialDescriptor : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PublicKeyCredentialDescriptor>, false> {
        WebCore::PublicKeyCredentialType type;
        WebCore::BufferSource id;
        Vector<WebCore::AuthenticatorTransport> transports;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPublicKeyCredentialDescriptor) == sizeof(WebCore::PublicKeyCredentialDescriptor));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PublicKeyCredentialDescriptor, type)
        , offsetof(WebCore::PublicKeyCredentialDescriptor, id)
        , offsetof(WebCore::PublicKeyCredentialDescriptor, transports)
    >::value);
    encoder << instance.type;
    encoder << instance.id;
    encoder << instance.transports;
}

std::optional<WebCore::PublicKeyCredentialDescriptor> ArgumentCoder<WebCore::PublicKeyCredentialDescriptor>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::PublicKeyCredentialType>();
    auto id = decoder.decode<WebCore::BufferSource>();
    auto transports = decoder.decode<Vector<WebCore::AuthenticatorTransport>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialDescriptor {
            WTFMove(*type),
            WTFMove(*id),
            WTFMove(*transports)
        }
    };
}

#endif

void ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialCreationOptions& instance)
{
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rp)>, WebCore::PublicKeyCredentialCreationOptions::RpEntity>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.user)>, WebCore::PublicKeyCredentialCreationOptions::UserEntity>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.challenge)>, WebCore::BufferSource>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pubKeyCredParams)>, Vector<WebCore::PublicKeyCredentialCreationOptions::Parameters>>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeout)>, std::optional<unsigned>>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.excludeCredentials)>, Vector<WebCore::PublicKeyCredentialDescriptor>>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authenticatorSelection)>, std::optional<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria>>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.attestation)>, WebCore::AttestationConveyancePreference>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.extensions)>, std::optional<WebCore::AuthenticationExtensionsClientInputs>>);
#endif
    struct ShouldBeSameSizeAsPublicKeyCredentialCreationOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PublicKeyCredentialCreationOptions>, false> {
#if ENABLE(WEB_AUTHN)
        WebCore::PublicKeyCredentialCreationOptions::RpEntity rp;
#endif
#if ENABLE(WEB_AUTHN)
        WebCore::PublicKeyCredentialCreationOptions::UserEntity user;
#endif
#if ENABLE(WEB_AUTHN)
        WebCore::BufferSource challenge;
#endif
#if ENABLE(WEB_AUTHN)
        Vector<WebCore::PublicKeyCredentialCreationOptions::Parameters> pubKeyCredParams;
#endif
#if ENABLE(WEB_AUTHN)
        std::optional<unsigned> timeout;
#endif
#if ENABLE(WEB_AUTHN)
        Vector<WebCore::PublicKeyCredentialDescriptor> excludeCredentials;
#endif
#if ENABLE(WEB_AUTHN)
        std::optional<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria> authenticatorSelection;
#endif
#if ENABLE(WEB_AUTHN)
        WebCore::AttestationConveyancePreference attestation;
#endif
#if ENABLE(WEB_AUTHN)
        std::optional<WebCore::AuthenticationExtensionsClientInputs> extensions;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsPublicKeyCredentialCreationOptions) == sizeof(WebCore::PublicKeyCredentialCreationOptions));
    static_assert(MembersInCorrectOrder < 0
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, rp)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, user)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, challenge)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, pubKeyCredParams)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, timeout)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, excludeCredentials)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, authenticatorSelection)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, attestation)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialCreationOptions, extensions)
#endif
    >::value);
#if ENABLE(WEB_AUTHN)
    encoder << instance.rp;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.user;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.challenge;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.pubKeyCredParams;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.timeout;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.excludeCredentials;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.authenticatorSelection;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.attestation;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.extensions;
#endif
}

std::optional<WebCore::PublicKeyCredentialCreationOptions> ArgumentCoder<WebCore::PublicKeyCredentialCreationOptions>::decode(Decoder& decoder)
{
#if ENABLE(WEB_AUTHN)
    auto rp = decoder.decode<WebCore::PublicKeyCredentialCreationOptions::RpEntity>();
#endif
#if ENABLE(WEB_AUTHN)
    auto user = decoder.decode<WebCore::PublicKeyCredentialCreationOptions::UserEntity>();
#endif
#if ENABLE(WEB_AUTHN)
    auto challenge = decoder.decode<WebCore::BufferSource>();
#endif
#if ENABLE(WEB_AUTHN)
    auto pubKeyCredParams = decoder.decode<Vector<WebCore::PublicKeyCredentialCreationOptions::Parameters>>();
#endif
#if ENABLE(WEB_AUTHN)
    auto timeout = decoder.decode<std::optional<unsigned>>();
#endif
#if ENABLE(WEB_AUTHN)
    auto excludeCredentials = decoder.decode<Vector<WebCore::PublicKeyCredentialDescriptor>>();
#endif
#if ENABLE(WEB_AUTHN)
    auto authenticatorSelection = decoder.decode<std::optional<WebCore::PublicKeyCredentialCreationOptions::AuthenticatorSelectionCriteria>>();
#endif
#if ENABLE(WEB_AUTHN)
    auto attestation = decoder.decode<WebCore::AttestationConveyancePreference>();
#endif
#if ENABLE(WEB_AUTHN)
    auto extensions = decoder.decode<std::optional<WebCore::AuthenticationExtensionsClientInputs>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialCreationOptions {
#if ENABLE(WEB_AUTHN)
            WTFMove(*rp),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*user),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*challenge),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*pubKeyCredParams),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*timeout),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*excludeCredentials),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*authenticatorSelection),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*attestation),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*extensions)
#endif
        }
    };
}

void ArgumentCoder<WebCore::PublicKeyCredentialRequestOptions>::encode(Encoder& encoder, const WebCore::PublicKeyCredentialRequestOptions& instance)
{
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.challenge)>, WebCore::BufferSource>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeout)>, std::optional<unsigned>>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rpId)>, String>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowCredentials)>, Vector<WebCore::PublicKeyCredentialDescriptor>>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userVerification)>, WebCore::UserVerificationRequirement>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.extensions)>, std::optional<WebCore::AuthenticationExtensionsClientInputs>>);
#endif
#if ENABLE(WEB_AUTHN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authenticatorAttachment)>, std::optional<WebCore::AuthenticatorAttachment>>);
#endif
    struct ShouldBeSameSizeAsPublicKeyCredentialRequestOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PublicKeyCredentialRequestOptions>, false> {
#if ENABLE(WEB_AUTHN)
        WebCore::BufferSource challenge;
#endif
#if ENABLE(WEB_AUTHN)
        std::optional<unsigned> timeout;
#endif
#if ENABLE(WEB_AUTHN)
        String rpId;
#endif
#if ENABLE(WEB_AUTHN)
        Vector<WebCore::PublicKeyCredentialDescriptor> allowCredentials;
#endif
#if ENABLE(WEB_AUTHN)
        WebCore::UserVerificationRequirement userVerification;
#endif
#if ENABLE(WEB_AUTHN)
        std::optional<WebCore::AuthenticationExtensionsClientInputs> extensions;
#endif
#if ENABLE(WEB_AUTHN)
        std::optional<WebCore::AuthenticatorAttachment> authenticatorAttachment;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsPublicKeyCredentialRequestOptions) == sizeof(WebCore::PublicKeyCredentialRequestOptions));
    static_assert(MembersInCorrectOrder < 0
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialRequestOptions, challenge)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialRequestOptions, timeout)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialRequestOptions, rpId)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialRequestOptions, allowCredentials)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialRequestOptions, userVerification)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialRequestOptions, extensions)
#endif
#if ENABLE(WEB_AUTHN)
        , offsetof(WebCore::PublicKeyCredentialRequestOptions, authenticatorAttachment)
#endif
    >::value);
#if ENABLE(WEB_AUTHN)
    encoder << instance.challenge;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.timeout;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.rpId;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.allowCredentials;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.userVerification;
#endif
#if ENABLE(WEB_AUTHN)
    encoder << instance.extensions;
#endif
}

std::optional<WebCore::PublicKeyCredentialRequestOptions> ArgumentCoder<WebCore::PublicKeyCredentialRequestOptions>::decode(Decoder& decoder)
{
#if ENABLE(WEB_AUTHN)
    auto challenge = decoder.decode<WebCore::BufferSource>();
#endif
#if ENABLE(WEB_AUTHN)
    auto timeout = decoder.decode<std::optional<unsigned>>();
#endif
#if ENABLE(WEB_AUTHN)
    auto rpId = decoder.decode<String>();
#endif
#if ENABLE(WEB_AUTHN)
    auto allowCredentials = decoder.decode<Vector<WebCore::PublicKeyCredentialDescriptor>>();
#endif
#if ENABLE(WEB_AUTHN)
    auto userVerification = decoder.decode<WebCore::UserVerificationRequirement>();
#endif
#if ENABLE(WEB_AUTHN)
    auto extensions = decoder.decode<std::optional<WebCore::AuthenticationExtensionsClientInputs>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PublicKeyCredentialRequestOptions {
#if ENABLE(WEB_AUTHN)
            WTFMove(*challenge),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*timeout),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*rpId),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*allowCredentials),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*userVerification),
#endif
#if ENABLE(WEB_AUTHN)
            WTFMove(*extensions)
#endif
        }
    };
}

void ArgumentCoder<WebCore::RegistrableDomain>::encode(Encoder& encoder, const WebCore::RegistrableDomain& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.string())>, String>);
    encoder << instance.string();
}

std::optional<WebCore::RegistrableDomain> ArgumentCoder<WebCore::RegistrableDomain>::decode(Decoder& decoder)
{
    auto string = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RegistrableDomain::fromRawString(
            WTFMove(*string)
        )
    };
}

void ArgumentCoder<WebCore::AudioConfiguration>::encode(Encoder& encoder, const WebCore::AudioConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentType)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.channels)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bitrate)>, std::optional<uint64_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.samplerate)>, std::optional<uint32_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.spatialRendering)>, std::optional<bool>>);
    struct ShouldBeSameSizeAsAudioConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AudioConfiguration>, false> {
        String contentType;
        String channels;
        std::optional<uint64_t> bitrate;
        std::optional<uint32_t> samplerate;
        std::optional<bool> spatialRendering;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAudioConfiguration) == sizeof(WebCore::AudioConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AudioConfiguration, contentType)
        , offsetof(WebCore::AudioConfiguration, channels)
        , offsetof(WebCore::AudioConfiguration, bitrate)
        , offsetof(WebCore::AudioConfiguration, samplerate)
        , offsetof(WebCore::AudioConfiguration, spatialRendering)
    >::value);
    encoder << instance.contentType;
    encoder << instance.channels;
    encoder << instance.bitrate;
    encoder << instance.samplerate;
    encoder << instance.spatialRendering;
}

std::optional<WebCore::AudioConfiguration> ArgumentCoder<WebCore::AudioConfiguration>::decode(Decoder& decoder)
{
    auto contentType = decoder.decode<String>();
    auto channels = decoder.decode<String>();
    auto bitrate = decoder.decode<std::optional<uint64_t>>();
    auto samplerate = decoder.decode<std::optional<uint32_t>>();
    auto spatialRendering = decoder.decode<std::optional<bool>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AudioConfiguration {
            WTFMove(*contentType),
            WTFMove(*channels),
            WTFMove(*bitrate),
            WTFMove(*samplerate),
            WTFMove(*spatialRendering)
        }
    };
}

void ArgumentCoder<WebCore::Cookie>::encode(Encoder& encoder, const WebCore::Cookie& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.domain)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.created)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expires)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpOnly)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.secure)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.session)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.comment)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.commentURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ports)>, Vector<uint16_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sameSite)>, WebCore::Cookie::SameSitePolicy>);
    struct ShouldBeSameSizeAsCookie : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Cookie>, false> {
        String name;
        String value;
        String domain;
        String path;
        double created;
        std::optional<double> expires;
        bool httpOnly;
        bool secure;
        bool session;
        String comment;
        URL commentURL;
        Vector<uint16_t> ports;
        WebCore::Cookie::SameSitePolicy sameSite;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCookie) == sizeof(WebCore::Cookie));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Cookie, name)
        , offsetof(WebCore::Cookie, value)
        , offsetof(WebCore::Cookie, domain)
        , offsetof(WebCore::Cookie, path)
        , offsetof(WebCore::Cookie, created)
        , offsetof(WebCore::Cookie, expires)
        , offsetof(WebCore::Cookie, httpOnly)
        , offsetof(WebCore::Cookie, secure)
        , offsetof(WebCore::Cookie, session)
        , offsetof(WebCore::Cookie, comment)
        , offsetof(WebCore::Cookie, commentURL)
        , offsetof(WebCore::Cookie, ports)
        , offsetof(WebCore::Cookie, sameSite)
    >::value);
    encoder << instance.name;
    encoder << instance.value;
    encoder << instance.domain;
    encoder << instance.path;
    encoder << instance.created;
    encoder << instance.expires;
    encoder << instance.httpOnly;
    encoder << instance.secure;
    encoder << instance.session;
    encoder << instance.comment;
    encoder << instance.commentURL;
    encoder << instance.ports;
    encoder << instance.sameSite;
}

std::optional<WebCore::Cookie> ArgumentCoder<WebCore::Cookie>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto value = decoder.decode<String>();
    auto domain = decoder.decode<String>();
    auto path = decoder.decode<String>();
    auto created = decoder.decode<double>();
    auto expires = decoder.decode<std::optional<double>>();
    auto httpOnly = decoder.decode<bool>();
    auto secure = decoder.decode<bool>();
    auto session = decoder.decode<bool>();
    auto comment = decoder.decode<String>();
    auto commentURL = decoder.decode<URL>();
    auto ports = decoder.decode<Vector<uint16_t>>();
    auto sameSite = decoder.decode<WebCore::Cookie::SameSitePolicy>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Cookie {
            WTFMove(*name),
            WTFMove(*value),
            WTFMove(*domain),
            WTFMove(*path),
            WTFMove(*created),
            WTFMove(*expires),
            WTFMove(*httpOnly),
            WTFMove(*secure),
            WTFMove(*session),
            WTFMove(*comment),
            WTFMove(*commentURL),
            WTFMove(*ports),
            WTFMove(*sameSite)
        }
    };
}

#if ENABLE(VIDEO)
void ArgumentCoder<WebCore::VideoFrameMetadata>::encode(Encoder& encoder, const WebCore::VideoFrameMetadata& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentationTime)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expectedDisplayTime)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaTime)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentedFrames)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processingDuration)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.captureTime)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.receiveTime)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rtpTimestamp)>, std::optional<unsigned>>);
    struct ShouldBeSameSizeAsVideoFrameMetadata : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::VideoFrameMetadata>, false> {
        double presentationTime;
        double expectedDisplayTime;
        unsigned width;
        unsigned height;
        double mediaTime;
        unsigned presentedFrames;
        std::optional<double> processingDuration;
        std::optional<double> captureTime;
        std::optional<double> receiveTime;
        std::optional<unsigned> rtpTimestamp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVideoFrameMetadata) == sizeof(WebCore::VideoFrameMetadata));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::VideoFrameMetadata, presentationTime)
        , offsetof(WebCore::VideoFrameMetadata, expectedDisplayTime)
        , offsetof(WebCore::VideoFrameMetadata, width)
        , offsetof(WebCore::VideoFrameMetadata, height)
        , offsetof(WebCore::VideoFrameMetadata, mediaTime)
        , offsetof(WebCore::VideoFrameMetadata, presentedFrames)
        , offsetof(WebCore::VideoFrameMetadata, processingDuration)
        , offsetof(WebCore::VideoFrameMetadata, captureTime)
        , offsetof(WebCore::VideoFrameMetadata, receiveTime)
        , offsetof(WebCore::VideoFrameMetadata, rtpTimestamp)
    >::value);
    encoder << instance.presentationTime;
    encoder << instance.expectedDisplayTime;
    encoder << instance.width;
    encoder << instance.height;
    encoder << instance.mediaTime;
    encoder << instance.presentedFrames;
    encoder << instance.processingDuration;
    encoder << instance.captureTime;
    encoder << instance.receiveTime;
    encoder << instance.rtpTimestamp;
}

std::optional<WebCore::VideoFrameMetadata> ArgumentCoder<WebCore::VideoFrameMetadata>::decode(Decoder& decoder)
{
    auto presentationTime = decoder.decode<double>();
    auto expectedDisplayTime = decoder.decode<double>();
    auto width = decoder.decode<unsigned>();
    auto height = decoder.decode<unsigned>();
    auto mediaTime = decoder.decode<double>();
    auto presentedFrames = decoder.decode<unsigned>();
    auto processingDuration = decoder.decode<std::optional<double>>();
    auto captureTime = decoder.decode<std::optional<double>>();
    auto receiveTime = decoder.decode<std::optional<double>>();
    auto rtpTimestamp = decoder.decode<std::optional<unsigned>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::VideoFrameMetadata {
            WTFMove(*presentationTime),
            WTFMove(*expectedDisplayTime),
            WTFMove(*width),
            WTFMove(*height),
            WTFMove(*mediaTime),
            WTFMove(*presentedFrames),
            WTFMove(*processingDuration),
            WTFMove(*captureTime),
            WTFMove(*receiveTime),
            WTFMove(*rtpTimestamp)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::NavigationPreloadState>::encode(Encoder& encoder, const WebCore::NavigationPreloadState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.headerValue)>, String>);
    struct ShouldBeSameSizeAsNavigationPreloadState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::NavigationPreloadState>, false> {
        bool enabled;
        String headerValue;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNavigationPreloadState) == sizeof(WebCore::NavigationPreloadState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::NavigationPreloadState, enabled)
        , offsetof(WebCore::NavigationPreloadState, headerValue)
    >::value);
    encoder << instance.enabled;
    encoder << instance.headerValue;
}

std::optional<WebCore::NavigationPreloadState> ArgumentCoder<WebCore::NavigationPreloadState>::decode(Decoder& decoder)
{
    auto enabled = decoder.decode<bool>();
    auto headerValue = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::NavigationPreloadState {
            WTFMove(*enabled),
            WTFMove(*headerValue)
        }
    };
}

#endif

#if ENABLE(CONTENT_FILTERING)
void ArgumentCoder<WebCore::MockContentFilterSettings>::encode(Encoder& encoder, const WebCore::MockContentFilterSettings& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_enabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_decisionPoint)>, WebCore::MockContentFilterSettings::DecisionPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_decision)>, WebCore::MockContentFilterSettings::Decision>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_unblockRequestDecision)>, WebCore::MockContentFilterSettings::Decision>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_blockedString)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_modifiedRequestURL)>, String>);
    struct ShouldBeSameSizeAsMockContentFilterSettings : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MockContentFilterSettings>, false> {
        bool m_enabled;
        WebCore::MockContentFilterSettings::DecisionPoint m_decisionPoint;
        WebCore::MockContentFilterSettings::Decision m_decision;
        WebCore::MockContentFilterSettings::Decision m_unblockRequestDecision;
        String m_blockedString;
        String m_modifiedRequestURL;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMockContentFilterSettings) == sizeof(WebCore::MockContentFilterSettings));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MockContentFilterSettings, m_enabled)
        , offsetof(WebCore::MockContentFilterSettings, m_decisionPoint)
        , offsetof(WebCore::MockContentFilterSettings, m_decision)
        , offsetof(WebCore::MockContentFilterSettings, m_unblockRequestDecision)
        , offsetof(WebCore::MockContentFilterSettings, m_blockedString)
        , offsetof(WebCore::MockContentFilterSettings, m_modifiedRequestURL)
    >::value);
    encoder << instance.m_enabled;
    encoder << instance.m_decisionPoint;
    encoder << instance.m_decision;
    encoder << instance.m_unblockRequestDecision;
    encoder << instance.m_blockedString;
    encoder << instance.m_modifiedRequestURL;
}

std::optional<WebCore::MockContentFilterSettings> ArgumentCoder<WebCore::MockContentFilterSettings>::decode(Decoder& decoder)
{
    auto m_enabled = decoder.decode<bool>();
    auto m_decisionPoint = decoder.decode<WebCore::MockContentFilterSettings::DecisionPoint>();
    auto m_decision = decoder.decode<WebCore::MockContentFilterSettings::Decision>();
    auto m_unblockRequestDecision = decoder.decode<WebCore::MockContentFilterSettings::Decision>();
    auto m_blockedString = decoder.decode<String>();
    auto m_modifiedRequestURL = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    WebCore::MockContentFilterSettings result;
    result.m_enabled = WTFMove(*m_enabled);
    result.m_decisionPoint = WTFMove(*m_decisionPoint);
    result.m_decision = WTFMove(*m_decision);
    result.m_unblockRequestDecision = WTFMove(*m_unblockRequestDecision);
    result.m_blockedString = WTFMove(*m_blockedString);
    result.m_modifiedRequestURL = WTFMove(*m_modifiedRequestURL);
    return { WTFMove(result) };
}

#endif

void ArgumentCoder<WebCore::ExceptionDetails>::encode(Encoder& encoder, const WebCore::ExceptionDetails& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lineNumber)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.columnNumber)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ExceptionDetails::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceURL)>, String>);
    struct ShouldBeSameSizeAsExceptionDetails : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ExceptionDetails>, false> {
        String message;
        int lineNumber;
        int columnNumber;
        WebCore::ExceptionDetails::Type type;
        String sourceURL;
    };
    static_assert(sizeof(ShouldBeSameSizeAsExceptionDetails) == sizeof(WebCore::ExceptionDetails));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ExceptionDetails, message)
        , offsetof(WebCore::ExceptionDetails, lineNumber)
        , offsetof(WebCore::ExceptionDetails, columnNumber)
        , offsetof(WebCore::ExceptionDetails, type)
        , offsetof(WebCore::ExceptionDetails, sourceURL)
    >::value);
    encoder << instance.message;
    encoder << instance.lineNumber;
    encoder << instance.columnNumber;
    encoder << instance.type;
    encoder << instance.sourceURL;
}

std::optional<WebCore::ExceptionDetails> ArgumentCoder<WebCore::ExceptionDetails>::decode(Decoder& decoder)
{
    auto message = decoder.decode<String>();
    auto lineNumber = decoder.decode<int>();
    auto columnNumber = decoder.decode<int>();
    auto type = decoder.decode<WebCore::ExceptionDetails::Type>();
    auto sourceURL = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ExceptionDetails {
            WTFMove(*message),
            WTFMove(*lineNumber),
            WTFMove(*columnNumber),
            WTFMove(*type),
            WTFMove(*sourceURL)
        }
    };
}

void ArgumentCoder<WebCore::FloatBoxExtent>::encode(Encoder& encoder, const WebCore::FloatBoxExtent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.top())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.right())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bottom())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.left())>, float>);
    encoder << instance.top();
    encoder << instance.right();
    encoder << instance.bottom();
    encoder << instance.left();
}

void ArgumentCoder<WebCore::FloatBoxExtent>::encode(StreamConnectionEncoder& encoder, const WebCore::FloatBoxExtent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.top())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.right())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bottom())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.left())>, float>);
    encoder << instance.top();
    encoder << instance.right();
    encoder << instance.bottom();
    encoder << instance.left();
}

std::optional<WebCore::FloatBoxExtent> ArgumentCoder<WebCore::FloatBoxExtent>::decode(Decoder& decoder)
{
    auto top = decoder.decode<float>();
    auto right = decoder.decode<float>();
    auto bottom = decoder.decode<float>();
    auto left = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatBoxExtent {
            WTFMove(*top),
            WTFMove(*right),
            WTFMove(*bottom),
            WTFMove(*left)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::SetInlineFillColor>::encode(Encoder& encoder, const WebCore::DisplayList::SetInlineFillColor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().red)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().green)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().blue)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().alpha)>, uint8_t>);
    encoder << instance.colorData().resolved().red;
    encoder << instance.colorData().resolved().green;
    encoder << instance.colorData().resolved().blue;
    encoder << instance.colorData().resolved().alpha;
}

void ArgumentCoder<WebCore::DisplayList::SetInlineFillColor>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::SetInlineFillColor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().red)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().green)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().blue)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().alpha)>, uint8_t>);
    encoder << instance.colorData().resolved().red;
    encoder << instance.colorData().resolved().green;
    encoder << instance.colorData().resolved().blue;
    encoder << instance.colorData().resolved().alpha;
}

std::optional<WebCore::DisplayList::SetInlineFillColor> ArgumentCoder<WebCore::DisplayList::SetInlineFillColor>::decode(Decoder& decoder)
{
    auto colorDataresolvedred = decoder.decode<uint8_t>();
    auto colorDataresolvedgreen = decoder.decode<uint8_t>();
    auto colorDataresolvedblue = decoder.decode<uint8_t>();
    auto colorDataresolvedalpha = decoder.decode<uint8_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::SetInlineFillColor {
            WTFMove(*colorDataresolvedred),
            WTFMove(*colorDataresolvedgreen),
            WTFMove(*colorDataresolvedblue),
            WTFMove(*colorDataresolvedalpha)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::SetInlineStrokeColor>::encode(Encoder& encoder, const WebCore::DisplayList::SetInlineStrokeColor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().red)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().green)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().blue)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().alpha)>, uint8_t>);
    encoder << instance.colorData().resolved().red;
    encoder << instance.colorData().resolved().green;
    encoder << instance.colorData().resolved().blue;
    encoder << instance.colorData().resolved().alpha;
}

void ArgumentCoder<WebCore::DisplayList::SetInlineStrokeColor>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::SetInlineStrokeColor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().red)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().green)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().blue)>, uint8_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorData().resolved().alpha)>, uint8_t>);
    encoder << instance.colorData().resolved().red;
    encoder << instance.colorData().resolved().green;
    encoder << instance.colorData().resolved().blue;
    encoder << instance.colorData().resolved().alpha;
}

std::optional<WebCore::DisplayList::SetInlineStrokeColor> ArgumentCoder<WebCore::DisplayList::SetInlineStrokeColor>::decode(Decoder& decoder)
{
    auto colorDataresolvedred = decoder.decode<uint8_t>();
    auto colorDataresolvedgreen = decoder.decode<uint8_t>();
    auto colorDataresolvedblue = decoder.decode<uint8_t>();
    auto colorDataresolvedalpha = decoder.decode<uint8_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::SetInlineStrokeColor {
            WTFMove(*colorDataresolvedred),
            WTFMove(*colorDataresolvedgreen),
            WTFMove(*colorDataresolvedblue),
            WTFMove(*colorDataresolvedalpha)
        }
    };
}

void ArgumentCoder<WebCore::HTTPHeaderMap::CommonHeader>::encode(Encoder& encoder, const WebCore::HTTPHeaderMap::CommonHeader& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.key)>, WebCore::HTTPHeaderName>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, String>);
    struct ShouldBeSameSizeAsCommonHeader : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::HTTPHeaderMap::CommonHeader>, false> {
        WebCore::HTTPHeaderName key;
        String value;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCommonHeader) == sizeof(WebCore::HTTPHeaderMap::CommonHeader));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::HTTPHeaderMap::CommonHeader, key)
        , offsetof(WebCore::HTTPHeaderMap::CommonHeader, value)
    >::value);
    encoder << instance.key;
    encoder << instance.value;
}

std::optional<WebCore::HTTPHeaderMap::CommonHeader> ArgumentCoder<WebCore::HTTPHeaderMap::CommonHeader>::decode(Decoder& decoder)
{
    auto key = decoder.decode<WebCore::HTTPHeaderName>();
    auto value = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::HTTPHeaderMap::CommonHeader {
            WTFMove(*key),
            WTFMove(*value)
        }
    };
}

void ArgumentCoder<WebCore::HTTPHeaderMap::UncommonHeader>::encode(Encoder& encoder, const WebCore::HTTPHeaderMap::UncommonHeader& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.key)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, String>);
    struct ShouldBeSameSizeAsUncommonHeader : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::HTTPHeaderMap::UncommonHeader>, false> {
        String key;
        String value;
    };
    static_assert(sizeof(ShouldBeSameSizeAsUncommonHeader) == sizeof(WebCore::HTTPHeaderMap::UncommonHeader));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::HTTPHeaderMap::UncommonHeader, key)
        , offsetof(WebCore::HTTPHeaderMap::UncommonHeader, value)
    >::value);
    encoder << instance.key;
    encoder << instance.value;
}

std::optional<WebCore::HTTPHeaderMap::UncommonHeader> ArgumentCoder<WebCore::HTTPHeaderMap::UncommonHeader>::decode(Decoder& decoder)
{
    auto key = decoder.decode<String>();
    auto value = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::HTTPHeaderMap::UncommonHeader {
            WTFMove(*key),
            WTFMove(*value)
        }
    };
}

void ArgumentCoder<WebCore::HTTPHeaderMap>::encode(Encoder& encoder, const WebCore::HTTPHeaderMap& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.commonHeaders())>, Vector<WebCore::HTTPHeaderMap::CommonHeader, 0, CrashOnOverflow, 6>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uncommonHeaders())>, Vector<WebCore::HTTPHeaderMap::UncommonHeader, 0, CrashOnOverflow, 0>>);
    encoder << instance.commonHeaders();
    encoder << instance.uncommonHeaders();
}

std::optional<WebCore::HTTPHeaderMap> ArgumentCoder<WebCore::HTTPHeaderMap>::decode(Decoder& decoder)
{
    auto commonHeaders = decoder.decode<Vector<WebCore::HTTPHeaderMap::CommonHeader, 0, CrashOnOverflow, 6>>();
    auto uncommonHeaders = decoder.decode<Vector<WebCore::HTTPHeaderMap::UncommonHeader, 0, CrashOnOverflow, 0>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::HTTPHeaderMap {
            WTFMove(*commonHeaders),
            WTFMove(*uncommonHeaders)
        }
    };
}

void ArgumentCoder<WebCore::ElementContext>::encode(Encoder& encoder, const WebCore::ElementContext& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.boundingRect)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webPageIdentifier)>, WebCore::PageIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentIdentifier)>, WebCore::ScriptExecutionContextIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elementIdentifier)>, WebCore::ElementIdentifier>);
    struct ShouldBeSameSizeAsElementContext : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ElementContext>, false> {
        WebCore::FloatRect boundingRect;
        WebCore::PageIdentifier webPageIdentifier;
        WebCore::ScriptExecutionContextIdentifier documentIdentifier;
        WebCore::ElementIdentifier elementIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsElementContext) == sizeof(WebCore::ElementContext));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ElementContext, boundingRect)
        , offsetof(WebCore::ElementContext, webPageIdentifier)
        , offsetof(WebCore::ElementContext, documentIdentifier)
        , offsetof(WebCore::ElementContext, elementIdentifier)
    >::value);
    encoder << instance.boundingRect;
    encoder << instance.webPageIdentifier;
    encoder << instance.documentIdentifier;
    encoder << instance.elementIdentifier;
}

std::optional<WebCore::ElementContext> ArgumentCoder<WebCore::ElementContext>::decode(Decoder& decoder)
{
    auto boundingRect = decoder.decode<WebCore::FloatRect>();
    auto webPageIdentifier = decoder.decode<WebCore::PageIdentifier>();
    auto documentIdentifier = decoder.decode<WebCore::ScriptExecutionContextIdentifier>();
    auto elementIdentifier = decoder.decode<WebCore::ElementIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ElementContext {
            WTFMove(*boundingRect),
            WTFMove(*webPageIdentifier),
            WTFMove(*documentIdentifier),
            WTFMove(*elementIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::ElementAnimationContext>::encode(Encoder& encoder, const WebCore::ElementAnimationContext& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.element)>, WebCore::ElementContext>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isAnimating)>, bool>);
    struct ShouldBeSameSizeAsElementAnimationContext : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ElementAnimationContext>, false> {
        WebCore::ElementContext element;
        bool isAnimating;
    };
    static_assert(sizeof(ShouldBeSameSizeAsElementAnimationContext) == sizeof(WebCore::ElementAnimationContext));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ElementAnimationContext, element)
        , offsetof(WebCore::ElementAnimationContext, isAnimating)
    >::value);
    encoder << instance.element;
    encoder << instance.isAnimating;
}

std::optional<WebCore::ElementAnimationContext> ArgumentCoder<WebCore::ElementAnimationContext>::decode(Decoder& decoder)
{
    auto element = decoder.decode<WebCore::ElementContext>();
    auto isAnimating = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ElementAnimationContext {
            WTFMove(*element),
            WTFMove(*isAnimating)
        }
    };
}

void ArgumentCoder<WebCore::SystemPreviewInfo>::encode(Encoder& encoder, const WebCore::SystemPreviewInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.element)>, WebCore::ElementContext>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.previewRect)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isPreview)>, bool>);
    struct ShouldBeSameSizeAsSystemPreviewInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SystemPreviewInfo>, false> {
        WebCore::ElementContext element;
        WebCore::IntRect previewRect;
        bool isPreview;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSystemPreviewInfo) == sizeof(WebCore::SystemPreviewInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SystemPreviewInfo, element)
        , offsetof(WebCore::SystemPreviewInfo, previewRect)
        , offsetof(WebCore::SystemPreviewInfo, isPreview)
    >::value);
    encoder << instance.element;
    encoder << instance.previewRect;
    encoder << instance.isPreview;
}

std::optional<WebCore::SystemPreviewInfo> ArgumentCoder<WebCore::SystemPreviewInfo>::decode(Decoder& decoder)
{
    auto element = decoder.decode<WebCore::ElementContext>();
    auto previewRect = decoder.decode<WebCore::IntRect>();
    auto isPreview = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SystemPreviewInfo {
            WTFMove(*element),
            WTFMove(*previewRect),
            WTFMove(*isPreview)
        }
    };
}

void ArgumentCoder<WebCore::ResourceRequest::RequestData>::encode(Encoder& encoder, const WebCore::ResourceRequest::RequestData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_firstPartyForCookies)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_timeoutInterval)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_httpMethod)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_httpHeaderFields)>, WebCore::HTTPHeaderMap>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_responseContentDispositionEncodingFallbackArray)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_cachePolicy)>, WebCore::ResourceRequestCachePolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_sameSiteDisposition)>, WebCore::ResourceRequestBase::SameSiteDisposition>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_priority)>, WebCore::ResourceLoadPriority>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_requester)>, WebCore::ResourceRequestRequester>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_allowCookies)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isTopSite)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isAppInitiated)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_privacyProxyFailClosedForUnreachableNonMainHosts)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_useAdvancedPrivacyProtections)>, bool>);
    struct ShouldBeSameSizeAsRequestData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ResourceRequest::RequestData>, false> {
        URL m_url;
        URL m_firstPartyForCookies;
        double m_timeoutInterval;
        String m_httpMethod;
        WebCore::HTTPHeaderMap m_httpHeaderFields;
        Vector<String> m_responseContentDispositionEncodingFallbackArray;
        WebCore::ResourceRequestCachePolicy m_cachePolicy;
        WebCore::ResourceRequestBase::SameSiteDisposition m_sameSiteDisposition;
        WebCore::ResourceLoadPriority m_priority;
        WebCore::ResourceRequestRequester m_requester;
        bool m_allowCookies : 1;
        bool m_isTopSite : 1;
        bool m_isAppInitiated : 1;
        bool m_privacyProxyFailClosedForUnreachableNonMainHosts : 1;
        bool m_useAdvancedPrivacyProtections : 1;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRequestData) == sizeof(WebCore::ResourceRequest::RequestData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ResourceRequest::RequestData, m_url)
        , offsetof(WebCore::ResourceRequest::RequestData, m_firstPartyForCookies)
        , offsetof(WebCore::ResourceRequest::RequestData, m_timeoutInterval)
        , offsetof(WebCore::ResourceRequest::RequestData, m_httpMethod)
        , offsetof(WebCore::ResourceRequest::RequestData, m_httpHeaderFields)
        , offsetof(WebCore::ResourceRequest::RequestData, m_responseContentDispositionEncodingFallbackArray)
        , offsetof(WebCore::ResourceRequest::RequestData, m_cachePolicy)
        , offsetof(WebCore::ResourceRequest::RequestData, m_sameSiteDisposition)
        , offsetof(WebCore::ResourceRequest::RequestData, m_priority)
        , offsetof(WebCore::ResourceRequest::RequestData, m_requester)
    >::value);
    encoder << instance.m_url;
    encoder << instance.m_firstPartyForCookies;
    encoder << instance.m_timeoutInterval;
    encoder << instance.m_httpMethod;
    encoder << instance.m_httpHeaderFields;
    encoder << instance.m_responseContentDispositionEncodingFallbackArray;
    encoder << instance.m_cachePolicy;
    encoder << instance.m_sameSiteDisposition;
    encoder << instance.m_priority;
    encoder << instance.m_requester;
    encoder << instance.m_allowCookies;
    encoder << instance.m_isTopSite;
    encoder << instance.m_isAppInitiated;
    encoder << instance.m_privacyProxyFailClosedForUnreachableNonMainHosts;
    encoder << instance.m_useAdvancedPrivacyProtections;
}

std::optional<WebCore::ResourceRequest::RequestData> ArgumentCoder<WebCore::ResourceRequest::RequestData>::decode(Decoder& decoder)
{
    auto m_url = decoder.decode<URL>();
    auto m_firstPartyForCookies = decoder.decode<URL>();
    auto m_timeoutInterval = decoder.decode<double>();
    auto m_httpMethod = decoder.decode<String>();
    auto m_httpHeaderFields = decoder.decode<WebCore::HTTPHeaderMap>();
    auto m_responseContentDispositionEncodingFallbackArray = decoder.decode<Vector<String>>();
    auto m_cachePolicy = decoder.decode<WebCore::ResourceRequestCachePolicy>();
    auto m_sameSiteDisposition = decoder.decode<WebCore::ResourceRequestBase::SameSiteDisposition>();
    auto m_priority = decoder.decode<WebCore::ResourceLoadPriority>();
    auto m_requester = decoder.decode<WebCore::ResourceRequestRequester>();
    auto m_allowCookies = decoder.decode<bool>();
    auto m_isTopSite = decoder.decode<bool>();
    auto m_isAppInitiated = decoder.decode<bool>();
    auto m_privacyProxyFailClosedForUnreachableNonMainHosts = decoder.decode<bool>();
    auto m_useAdvancedPrivacyProtections = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceRequest::RequestData {
            WTFMove(*m_url),
            WTFMove(*m_firstPartyForCookies),
            WTFMove(*m_timeoutInterval),
            WTFMove(*m_httpMethod),
            WTFMove(*m_httpHeaderFields),
            WTFMove(*m_responseContentDispositionEncodingFallbackArray),
            WTFMove(*m_cachePolicy),
            WTFMove(*m_sameSiteDisposition),
            WTFMove(*m_priority),
            WTFMove(*m_requester),
            WTFMove(*m_allowCookies),
            WTFMove(*m_isTopSite),
            WTFMove(*m_isAppInitiated),
            WTFMove(*m_privacyProxyFailClosedForUnreachableNonMainHosts),
            WTFMove(*m_useAdvancedPrivacyProtections)
        }
    };
}

#if USE(SOUP)
void ArgumentCoder<WebCore::ResourceRequestPlatformData>::encode(Encoder& encoder, const WebCore::ResourceRequestPlatformData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requestData)>, WebCore::ResourceRequest::RequestData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.acceptEncoding)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.redirectCount)>, uint16_t>);
    struct ShouldBeSameSizeAsResourceRequestPlatformData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ResourceRequestPlatformData>, false> {
        WebCore::ResourceRequest::RequestData requestData;
        bool acceptEncoding;
        uint16_t redirectCount;
    };
    static_assert(sizeof(ShouldBeSameSizeAsResourceRequestPlatformData) == sizeof(WebCore::ResourceRequestPlatformData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ResourceRequestPlatformData, requestData)
        , offsetof(WebCore::ResourceRequestPlatformData, acceptEncoding)
        , offsetof(WebCore::ResourceRequestPlatformData, redirectCount)
    >::value);
    encoder << instance.requestData;
    encoder << instance.acceptEncoding;
    encoder << instance.redirectCount;
}

std::optional<WebCore::ResourceRequestPlatformData> ArgumentCoder<WebCore::ResourceRequestPlatformData>::decode(Decoder& decoder)
{
    auto requestData = decoder.decode<WebCore::ResourceRequest::RequestData>();
    auto acceptEncoding = decoder.decode<bool>();
    auto redirectCount = decoder.decode<uint16_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceRequestPlatformData {
            WTFMove(*requestData),
            WTFMove(*acceptEncoding),
            WTFMove(*redirectCount)
        }
    };
}

#endif

#if PLATFORM(COCOA)
void ArgumentCoder<WebCore::ResourceRequest>::encode(Encoder& encoder, const WebCore::ResourceRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getRequestDataToSerialize())>, std::variant<WebCore::ResourceRequest::RequestData, WebCore::ResourceRequestPlatformData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cachePartition())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hiddenFromInspector())>, bool>);
    encoder << instance.getRequestDataToSerialize();
    encoder << instance.cachePartition();
    encoder << instance.hiddenFromInspector();
}

std::optional<WebCore::ResourceRequest> ArgumentCoder<WebCore::ResourceRequest>::decode(Decoder& decoder)
{
    auto getRequestDataToSerialize = decoder.decode<std::variant<WebCore::ResourceRequest::RequestData, WebCore::ResourceRequestPlatformData>>();
    auto cachePartition = decoder.decode<String>();
    auto hiddenFromInspector = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceRequest::fromResourceRequestData(
            WTFMove(*getRequestDataToSerialize),
            WTFMove(*cachePartition),
            WTFMove(*hiddenFromInspector)
        )
    };
}

#endif

#if USE(CURL)
void ArgumentCoder<WebCore::ResourceRequest>::encode(Encoder& encoder, const WebCore::ResourceRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getRequestDataToSerialize())>, WebCore::ResourceRequest::RequestData>);
    encoder << instance.getRequestDataToSerialize();
}

std::optional<WebCore::ResourceRequest> ArgumentCoder<WebCore::ResourceRequest>::decode(Decoder& decoder)
{
    auto getRequestDataToSerialize = decoder.decode<WebCore::ResourceRequest::RequestData>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceRequest::fromResourceRequestData(
            WTFMove(*getRequestDataToSerialize)
        )
    };
}

#endif

#if !USE(CURL) && !PLATFORM(COCOA)
void ArgumentCoder<WebCore::ResourceRequest>::encode(Encoder& encoder, const WebCore::ResourceRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getRequestDataToSerialize())>, std::variant<WebCore::ResourceRequest::RequestData, WebCore::ResourceRequestPlatformData>>);
    encoder << instance.getRequestDataToSerialize();
}

std::optional<WebCore::ResourceRequest> ArgumentCoder<WebCore::ResourceRequest>::decode(Decoder& decoder)
{
    auto getRequestDataToSerialize = decoder.decode<std::variant<WebCore::ResourceRequest::RequestData, WebCore::ResourceRequestPlatformData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceRequest::fromResourceRequestData(
            WTFMove(*getRequestDataToSerialize)
        )
    };
}

#endif

#if !USE(CG)
void ArgumentCoder<WebCore::PlatformColorSpace>::encode(Encoder& encoder, const WebCore::PlatformColorSpace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.get())>, WebCore::PlatformColorSpace::Name>);
    encoder << instance.get();
}

void ArgumentCoder<WebCore::PlatformColorSpace>::encode(StreamConnectionEncoder& encoder, const WebCore::PlatformColorSpace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.get())>, WebCore::PlatformColorSpace::Name>);
    encoder << instance.get();
}

std::optional<WebCore::PlatformColorSpace> ArgumentCoder<WebCore::PlatformColorSpace>::decode(Decoder& decoder)
{
    auto get = decoder.decode<WebCore::PlatformColorSpace::Name>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PlatformColorSpace {
            WTFMove(*get)
        }
    };
}

#endif

void ArgumentCoder<WebCore::DestinationColorSpace>::encode(Encoder& encoder, const WebCore::DestinationColorSpace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serializableColorSpace())>, WebCore::PlatformColorSpace>);
    encoder << instance.serializableColorSpace();
}

void ArgumentCoder<WebCore::DestinationColorSpace>::encode(StreamConnectionEncoder& encoder, const WebCore::DestinationColorSpace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serializableColorSpace())>, WebCore::PlatformColorSpace>);
    encoder << instance.serializableColorSpace();
}

std::optional<WebCore::DestinationColorSpace> ArgumentCoder<WebCore::DestinationColorSpace>::decode(Decoder& decoder)
{
    auto serializableColorSpace = decoder.decode<WebCore::PlatformColorSpace>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DestinationColorSpace {
            WTFMove(*serializableColorSpace)
        }
    };
}

void ArgumentCoder<WebCore::WindowFeatures>::encode(Encoder& encoder, const WebCore::WindowFeatures& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x)>, std::optional<float>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y)>, std::optional<float>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, std::optional<float>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, std::optional<float>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.menuBarVisible)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.statusBarVisible)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.toolBarVisible)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.locationBarVisible)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollbarsVisible)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resizable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fullscreen)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dialog)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.noopener)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.noreferrer)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.additionalFeatures)>, Vector<String>>);
    struct ShouldBeSameSizeAsWindowFeatures : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WindowFeatures>, false> {
        std::optional<float> x;
        std::optional<float> y;
        std::optional<float> width;
        std::optional<float> height;
        bool menuBarVisible;
        bool statusBarVisible;
        bool toolBarVisible;
        bool locationBarVisible;
        bool scrollbarsVisible;
        bool resizable;
        bool fullscreen;
        bool dialog;
        bool noopener;
        bool noreferrer;
        Vector<String> additionalFeatures;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWindowFeatures) == sizeof(WebCore::WindowFeatures));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WindowFeatures, x)
        , offsetof(WebCore::WindowFeatures, y)
        , offsetof(WebCore::WindowFeatures, width)
        , offsetof(WebCore::WindowFeatures, height)
        , offsetof(WebCore::WindowFeatures, menuBarVisible)
        , offsetof(WebCore::WindowFeatures, statusBarVisible)
        , offsetof(WebCore::WindowFeatures, toolBarVisible)
        , offsetof(WebCore::WindowFeatures, locationBarVisible)
        , offsetof(WebCore::WindowFeatures, scrollbarsVisible)
        , offsetof(WebCore::WindowFeatures, resizable)
        , offsetof(WebCore::WindowFeatures, fullscreen)
        , offsetof(WebCore::WindowFeatures, dialog)
        , offsetof(WebCore::WindowFeatures, noopener)
        , offsetof(WebCore::WindowFeatures, noreferrer)
        , offsetof(WebCore::WindowFeatures, additionalFeatures)
    >::value);
    encoder << instance.x;
    encoder << instance.y;
    encoder << instance.width;
    encoder << instance.height;
    encoder << instance.menuBarVisible;
    encoder << instance.statusBarVisible;
    encoder << instance.toolBarVisible;
    encoder << instance.locationBarVisible;
    encoder << instance.scrollbarsVisible;
    encoder << instance.resizable;
    encoder << instance.fullscreen;
    encoder << instance.dialog;
}

std::optional<WebCore::WindowFeatures> ArgumentCoder<WebCore::WindowFeatures>::decode(Decoder& decoder)
{
    auto x = decoder.decode<std::optional<float>>();
    auto y = decoder.decode<std::optional<float>>();
    auto width = decoder.decode<std::optional<float>>();
    auto height = decoder.decode<std::optional<float>>();
    auto menuBarVisible = decoder.decode<bool>();
    auto statusBarVisible = decoder.decode<bool>();
    auto toolBarVisible = decoder.decode<bool>();
    auto locationBarVisible = decoder.decode<bool>();
    auto scrollbarsVisible = decoder.decode<bool>();
    auto resizable = decoder.decode<bool>();
    auto fullscreen = decoder.decode<bool>();
    auto dialog = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WindowFeatures {
            WTFMove(*x),
            WTFMove(*y),
            WTFMove(*width),
            WTFMove(*height),
            WTFMove(*menuBarVisible),
            WTFMove(*statusBarVisible),
            WTFMove(*toolBarVisible),
            WTFMove(*locationBarVisible),
            WTFMove(*scrollbarsVisible),
            WTFMove(*resizable),
            WTFMove(*fullscreen),
            WTFMove(*dialog)
        }
    };
}

void ArgumentCoder<WebCore::CompositionUnderline>::encode(Encoder& encoder, const WebCore::CompositionUnderline& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startOffset)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endOffset)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositionUnderlineColor)>, WebCore::CompositionUnderlineColor>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.thick)>, bool>);
    struct ShouldBeSameSizeAsCompositionUnderline : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CompositionUnderline>, false> {
        unsigned startOffset;
        unsigned endOffset;
        WebCore::CompositionUnderlineColor compositionUnderlineColor;
        WebCore::Color color;
        bool thick;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCompositionUnderline) == sizeof(WebCore::CompositionUnderline));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CompositionUnderline, startOffset)
        , offsetof(WebCore::CompositionUnderline, endOffset)
        , offsetof(WebCore::CompositionUnderline, compositionUnderlineColor)
        , offsetof(WebCore::CompositionUnderline, color)
        , offsetof(WebCore::CompositionUnderline, thick)
    >::value);
    encoder << instance.startOffset;
    encoder << instance.endOffset;
    encoder << instance.compositionUnderlineColor;
    encoder << instance.color;
    encoder << instance.thick;
}

std::optional<WebCore::CompositionUnderline> ArgumentCoder<WebCore::CompositionUnderline>::decode(Decoder& decoder)
{
    auto startOffset = decoder.decode<unsigned>();
    auto endOffset = decoder.decode<unsigned>();
    auto compositionUnderlineColor = decoder.decode<WebCore::CompositionUnderlineColor>();
    auto color = decoder.decode<WebCore::Color>();
    auto thick = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CompositionUnderline {
            WTFMove(*startOffset),
            WTFMove(*endOffset),
            WTFMove(*compositionUnderlineColor),
            WTFMove(*color),
            WTFMove(*thick)
        }
    };
}

void ArgumentCoder<WebCore::FileChooserSettings>::encode(Encoder& encoder, const WebCore::FileChooserSettings& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsDirectories)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsMultipleFiles)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.acceptMIMETypes)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.acceptFileExtensions)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectedFiles)>, Vector<String>>);
#if ENABLE(MEDIA_CAPTURE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaCaptureType)>, WebCore::MediaCaptureType>);
#endif
    struct ShouldBeSameSizeAsFileChooserSettings : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FileChooserSettings>, false> {
        bool allowsDirectories;
        bool allowsMultipleFiles;
        Vector<String> acceptMIMETypes;
        Vector<String> acceptFileExtensions;
        Vector<String> selectedFiles;
#if ENABLE(MEDIA_CAPTURE)
        WebCore::MediaCaptureType mediaCaptureType;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsFileChooserSettings) == sizeof(WebCore::FileChooserSettings));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FileChooserSettings, allowsDirectories)
        , offsetof(WebCore::FileChooserSettings, allowsMultipleFiles)
        , offsetof(WebCore::FileChooserSettings, acceptMIMETypes)
        , offsetof(WebCore::FileChooserSettings, acceptFileExtensions)
        , offsetof(WebCore::FileChooserSettings, selectedFiles)
#if ENABLE(MEDIA_CAPTURE)
        , offsetof(WebCore::FileChooserSettings, mediaCaptureType)
#endif
    >::value);
    encoder << instance.allowsDirectories;
    encoder << instance.allowsMultipleFiles;
    encoder << instance.acceptMIMETypes;
    encoder << instance.acceptFileExtensions;
    encoder << instance.selectedFiles;
#if ENABLE(MEDIA_CAPTURE)
    encoder << instance.mediaCaptureType;
#endif
}

std::optional<WebCore::FileChooserSettings> ArgumentCoder<WebCore::FileChooserSettings>::decode(Decoder& decoder)
{
    auto allowsDirectories = decoder.decode<bool>();
    auto allowsMultipleFiles = decoder.decode<bool>();
    auto acceptMIMETypes = decoder.decode<Vector<String>>();
    auto acceptFileExtensions = decoder.decode<Vector<String>>();
    auto selectedFiles = decoder.decode<Vector<String>>();
#if ENABLE(MEDIA_CAPTURE)
    auto mediaCaptureType = decoder.decode<WebCore::MediaCaptureType>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FileChooserSettings {
            WTFMove(*allowsDirectories),
            WTFMove(*allowsMultipleFiles),
            WTFMove(*acceptMIMETypes),
            WTFMove(*acceptFileExtensions),
            WTFMove(*selectedFiles),
#if ENABLE(MEDIA_CAPTURE)
            WTFMove(*mediaCaptureType)
#endif
        }
    };
}

void ArgumentCoder<WebCore::GrammarDetail>::encode(Encoder& encoder, const WebCore::GrammarDetail& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.range)>, WebCore::CharacterRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.guesses)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userDescription)>, String>);
    struct ShouldBeSameSizeAsGrammarDetail : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GrammarDetail>, false> {
        WebCore::CharacterRange range;
        Vector<String> guesses;
        String userDescription;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGrammarDetail) == sizeof(WebCore::GrammarDetail));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GrammarDetail, range)
        , offsetof(WebCore::GrammarDetail, guesses)
        , offsetof(WebCore::GrammarDetail, userDescription)
    >::value);
    encoder << instance.range;
    encoder << instance.guesses;
    encoder << instance.userDescription;
}

std::optional<WebCore::GrammarDetail> ArgumentCoder<WebCore::GrammarDetail>::decode(Decoder& decoder)
{
    auto range = decoder.decode<WebCore::CharacterRange>();
    auto guesses = decoder.decode<Vector<String>>();
    auto userDescription = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GrammarDetail {
            WTFMove(*range),
            WTFMove(*guesses),
            WTFMove(*userDescription)
        }
    };
}

void ArgumentCoder<WebCore::TextCheckingResult>::encode(Encoder& encoder, const WebCore::TextCheckingResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, OptionSet<WebCore::TextCheckingType>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.range)>, WebCore::CharacterRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.details)>, Vector<WebCore::GrammarDetail>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.replacement)>, String>);
    struct ShouldBeSameSizeAsTextCheckingResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextCheckingResult>, false> {
        OptionSet<WebCore::TextCheckingType> type;
        WebCore::CharacterRange range;
        Vector<WebCore::GrammarDetail> details;
        String replacement;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextCheckingResult) == sizeof(WebCore::TextCheckingResult));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextCheckingResult, type)
        , offsetof(WebCore::TextCheckingResult, range)
        , offsetof(WebCore::TextCheckingResult, details)
        , offsetof(WebCore::TextCheckingResult, replacement)
    >::value);
    encoder << instance.type;
    encoder << instance.range;
    encoder << instance.details;
    encoder << instance.replacement;
}

std::optional<WebCore::TextCheckingResult> ArgumentCoder<WebCore::TextCheckingResult>::decode(Decoder& decoder)
{
    auto type = decoder.decode<OptionSet<WebCore::TextCheckingType>>();
    auto range = decoder.decode<WebCore::CharacterRange>();
    auto details = decoder.decode<Vector<WebCore::GrammarDetail>>();
    auto replacement = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextCheckingResult {
            WTFMove(*type),
            WTFMove(*range),
            WTFMove(*details),
            WTFMove(*replacement)
        }
    };
}

void ArgumentCoder<WebCore::TextCheckingRequestData>::encode(Encoder& encoder, const WebCore::TextCheckingRequestData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier())>, std::optional<WebCore::TextCheckingRequestIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.checkingTypes())>, OptionSet<WebCore::TextCheckingType>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processType())>, WebCore::TextCheckingProcessType>);
    encoder << instance.identifier();
    encoder << instance.text();
    encoder << instance.checkingTypes();
    encoder << instance.processType();
}

std::optional<WebCore::TextCheckingRequestData> ArgumentCoder<WebCore::TextCheckingRequestData>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<std::optional<WebCore::TextCheckingRequestIdentifier>>();
    auto text = decoder.decode<String>();
    auto checkingTypes = decoder.decode<OptionSet<WebCore::TextCheckingType>>();
    auto processType = decoder.decode<WebCore::TextCheckingProcessType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextCheckingRequestData {
            WTFMove(*identifier),
            WTFMove(*text),
            WTFMove(*checkingTypes),
            WTFMove(*processType)
        }
    };
}

void ArgumentCoder<WebCore::LayoutRect>::encode(Encoder& encoder, const WebCore::LayoutRect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_location)>, WebCore::LayoutPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_size)>, WebCore::LayoutSize>);
    struct ShouldBeSameSizeAsLayoutRect : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::LayoutRect>, false> {
        WebCore::LayoutPoint m_location;
        WebCore::LayoutSize m_size;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLayoutRect) == sizeof(WebCore::LayoutRect));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::LayoutRect, m_location)
        , offsetof(WebCore::LayoutRect, m_size)
    >::value);
    encoder << instance.m_location;
    encoder << instance.m_size;
}

std::optional<WebCore::LayoutRect> ArgumentCoder<WebCore::LayoutRect>::decode(Decoder& decoder)
{
    auto m_location = decoder.decode<WebCore::LayoutPoint>();
    auto m_size = decoder.decode<WebCore::LayoutSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LayoutRect {
            WTFMove(*m_location),
            WTFMove(*m_size)
        }
    };
}

void ArgumentCoder<WebCore::ScreenDataOverrides>::encode(Encoder& encoder, const WebCore::ScreenDataOverrides& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scale)>, double>);
    struct ShouldBeSameSizeAsScreenDataOverrides : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ScreenDataOverrides>, false> {
        double width;
        double height;
        double scale;
    };
    static_assert(sizeof(ShouldBeSameSizeAsScreenDataOverrides) == sizeof(WebCore::ScreenDataOverrides));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ScreenDataOverrides, width)
        , offsetof(WebCore::ScreenDataOverrides, height)
        , offsetof(WebCore::ScreenDataOverrides, scale)
    >::value);
    encoder << instance.width;
    encoder << instance.height;
    encoder << instance.scale;
}

std::optional<WebCore::ScreenDataOverrides> ArgumentCoder<WebCore::ScreenDataOverrides>::decode(Decoder& decoder)
{
    auto width = decoder.decode<double>();
    auto height = decoder.decode<double>();
    auto scale = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ScreenDataOverrides {
            WTFMove(*width),
            WTFMove(*height),
            WTFMove(*scale)
        }
    };
}

void ArgumentCoder<WebCore::MediaCapabilitiesInfo>::encode(Encoder& encoder, const WebCore::MediaCapabilitiesInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supported)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.smooth)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.powerEfficient)>, bool>);
    struct ShouldBeSameSizeAsMediaCapabilitiesInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaCapabilitiesInfo>, false> {
        bool supported;
        bool smooth;
        bool powerEfficient;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaCapabilitiesInfo) == sizeof(WebCore::MediaCapabilitiesInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaCapabilitiesInfo, supported)
        , offsetof(WebCore::MediaCapabilitiesInfo, smooth)
        , offsetof(WebCore::MediaCapabilitiesInfo, powerEfficient)
    >::value);
    encoder << instance.supported;
    encoder << instance.smooth;
    encoder << instance.powerEfficient;
}

std::optional<WebCore::MediaCapabilitiesInfo> ArgumentCoder<WebCore::MediaCapabilitiesInfo>::decode(Decoder& decoder)
{
    auto supported = decoder.decode<bool>();
    auto smooth = decoder.decode<bool>();
    auto powerEfficient = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaCapabilitiesInfo {
            WTFMove(*supported),
            WTFMove(*smooth),
            WTFMove(*powerEfficient)
        }
    };
}

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::MediaStreamRequest>::encode(Encoder& encoder, const WebCore::MediaStreamRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::MediaStreamRequest::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.audioConstraints)>, WebCore::MediaConstraints>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.videoConstraints)>, WebCore::MediaConstraints>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isUserGesturePriviledged)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageIdentifier)>, WebCore::PageIdentifier>);
    struct ShouldBeSameSizeAsMediaStreamRequest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaStreamRequest>, false> {
        WebCore::MediaStreamRequest::Type type;
        WebCore::MediaConstraints audioConstraints;
        WebCore::MediaConstraints videoConstraints;
        bool isUserGesturePriviledged;
        WebCore::PageIdentifier pageIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaStreamRequest) == sizeof(WebCore::MediaStreamRequest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaStreamRequest, type)
        , offsetof(WebCore::MediaStreamRequest, audioConstraints)
        , offsetof(WebCore::MediaStreamRequest, videoConstraints)
        , offsetof(WebCore::MediaStreamRequest, isUserGesturePriviledged)
        , offsetof(WebCore::MediaStreamRequest, pageIdentifier)
    >::value);
    encoder << instance.type;
    encoder << instance.audioConstraints;
    encoder << instance.videoConstraints;
    encoder << instance.isUserGesturePriviledged;
    encoder << instance.pageIdentifier;
}

std::optional<WebCore::MediaStreamRequest> ArgumentCoder<WebCore::MediaStreamRequest>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::MediaStreamRequest::Type>();
    auto audioConstraints = decoder.decode<WebCore::MediaConstraints>();
    auto videoConstraints = decoder.decode<WebCore::MediaConstraints>();
    auto isUserGesturePriviledged = decoder.decode<bool>();
    auto pageIdentifier = decoder.decode<WebCore::PageIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaStreamRequest {
            WTFMove(*type),
            WTFMove(*audioConstraints),
            WTFMove(*videoConstraints),
            WTFMove(*isUserGesturePriviledged),
            WTFMove(*pageIdentifier)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::MediaTrackConstraintSetMap>::encode(Encoder& encoder, const WebCore::MediaTrackConstraintSetMap& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_width)>, std::optional<WebCore::IntConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_height)>, std::optional<WebCore::IntConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_sampleRate)>, std::optional<WebCore::IntConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_sampleSize)>, std::optional<WebCore::IntConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_aspectRatio)>, std::optional<WebCore::DoubleConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_zoom)>, std::optional<WebCore::DoubleConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_frameRate)>, std::optional<WebCore::DoubleConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_volume)>, std::optional<WebCore::DoubleConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_echoCancellation)>, std::optional<WebCore::BooleanConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_displaySurface)>, std::optional<WebCore::BooleanConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_logicalSurface)>, std::optional<WebCore::BooleanConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_facingMode)>, std::optional<WebCore::StringConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_deviceId)>, std::optional<WebCore::StringConstraint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_groupId)>, std::optional<WebCore::StringConstraint>>);
    struct ShouldBeSameSizeAsMediaTrackConstraintSetMap : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaTrackConstraintSetMap>, false> {
        std::optional<WebCore::IntConstraint> m_width;
        std::optional<WebCore::IntConstraint> m_height;
        std::optional<WebCore::IntConstraint> m_sampleRate;
        std::optional<WebCore::IntConstraint> m_sampleSize;
        std::optional<WebCore::DoubleConstraint> m_aspectRatio;
        std::optional<WebCore::DoubleConstraint> m_zoom;
        std::optional<WebCore::DoubleConstraint> m_frameRate;
        std::optional<WebCore::DoubleConstraint> m_volume;
        std::optional<WebCore::BooleanConstraint> m_echoCancellation;
        std::optional<WebCore::BooleanConstraint> m_displaySurface;
        std::optional<WebCore::BooleanConstraint> m_logicalSurface;
        std::optional<WebCore::StringConstraint> m_facingMode;
        std::optional<WebCore::StringConstraint> m_deviceId;
        std::optional<WebCore::StringConstraint> m_groupId;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaTrackConstraintSetMap) == sizeof(WebCore::MediaTrackConstraintSetMap));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_width)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_height)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_sampleRate)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_sampleSize)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_aspectRatio)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_zoom)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_frameRate)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_volume)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_echoCancellation)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_displaySurface)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_logicalSurface)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_facingMode)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_deviceId)
        , offsetof(WebCore::MediaTrackConstraintSetMap, m_groupId)
    >::value);
    encoder << instance.m_width;
    encoder << instance.m_height;
    encoder << instance.m_sampleRate;
    encoder << instance.m_sampleSize;
    encoder << instance.m_aspectRatio;
    encoder << instance.m_zoom;
    encoder << instance.m_frameRate;
    encoder << instance.m_volume;
    encoder << instance.m_echoCancellation;
    encoder << instance.m_displaySurface;
    encoder << instance.m_logicalSurface;
    encoder << instance.m_facingMode;
    encoder << instance.m_deviceId;
    encoder << instance.m_groupId;
}

std::optional<WebCore::MediaTrackConstraintSetMap> ArgumentCoder<WebCore::MediaTrackConstraintSetMap>::decode(Decoder& decoder)
{
    auto m_width = decoder.decode<std::optional<WebCore::IntConstraint>>();
    auto m_height = decoder.decode<std::optional<WebCore::IntConstraint>>();
    auto m_sampleRate = decoder.decode<std::optional<WebCore::IntConstraint>>();
    auto m_sampleSize = decoder.decode<std::optional<WebCore::IntConstraint>>();
    auto m_aspectRatio = decoder.decode<std::optional<WebCore::DoubleConstraint>>();
    auto m_zoom = decoder.decode<std::optional<WebCore::DoubleConstraint>>();
    auto m_frameRate = decoder.decode<std::optional<WebCore::DoubleConstraint>>();
    auto m_volume = decoder.decode<std::optional<WebCore::DoubleConstraint>>();
    auto m_echoCancellation = decoder.decode<std::optional<WebCore::BooleanConstraint>>();
    auto m_displaySurface = decoder.decode<std::optional<WebCore::BooleanConstraint>>();
    auto m_logicalSurface = decoder.decode<std::optional<WebCore::BooleanConstraint>>();
    auto m_facingMode = decoder.decode<std::optional<WebCore::StringConstraint>>();
    auto m_deviceId = decoder.decode<std::optional<WebCore::StringConstraint>>();
    auto m_groupId = decoder.decode<std::optional<WebCore::StringConstraint>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    WebCore::MediaTrackConstraintSetMap result;
    result.m_width = WTFMove(*m_width);
    result.m_height = WTFMove(*m_height);
    result.m_sampleRate = WTFMove(*m_sampleRate);
    result.m_sampleSize = WTFMove(*m_sampleSize);
    result.m_aspectRatio = WTFMove(*m_aspectRatio);
    result.m_zoom = WTFMove(*m_zoom);
    result.m_frameRate = WTFMove(*m_frameRate);
    result.m_volume = WTFMove(*m_volume);
    result.m_echoCancellation = WTFMove(*m_echoCancellation);
    result.m_displaySurface = WTFMove(*m_displaySurface);
    result.m_logicalSurface = WTFMove(*m_logicalSurface);
    result.m_facingMode = WTFMove(*m_facingMode);
    result.m_deviceId = WTFMove(*m_deviceId);
    result.m_groupId = WTFMove(*m_groupId);
    return { WTFMove(result) };
}

#endif

#if ! ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::MediaStreamRequest>::encode(Encoder& encoder, const WebCore::MediaStreamRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::MediaStreamRequest::Type>);
    struct ShouldBeSameSizeAsMediaStreamRequest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaStreamRequest>, false> {
        WebCore::MediaStreamRequest::Type type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaStreamRequest) == sizeof(WebCore::MediaStreamRequest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaStreamRequest, type)
    >::value);
    encoder << instance.type;
}

std::optional<WebCore::MediaStreamRequest> ArgumentCoder<WebCore::MediaStreamRequest>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::MediaStreamRequest::Type>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaStreamRequest {
            WTFMove(*type)
        }
    };
}

#endif

void ArgumentCoder<WebCore::MediaCapabilitiesDecodingInfo>::encode(Encoder& encoder, const WebCore::MediaCapabilitiesDecodingInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supported)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.smooth)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.powerEfficient)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaCapabilitiesInfo, supported)
        , offsetof(WebCore::MediaCapabilitiesInfo, smooth)
        , offsetof(WebCore::MediaCapabilitiesInfo, powerEfficient)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportedConfiguration)>, WebCore::MediaDecodingConfiguration>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaCapabilitiesDecodingInfo, supportedConfiguration)
    >::value);
    encoder << instance.supported;
    encoder << instance.smooth;
    encoder << instance.powerEfficient;
    encoder << instance.supportedConfiguration;
}

std::optional<WebCore::MediaCapabilitiesDecodingInfo> ArgumentCoder<WebCore::MediaCapabilitiesDecodingInfo>::decode(Decoder& decoder)
{
    auto supported = decoder.decode<bool>();
    auto smooth = decoder.decode<bool>();
    auto powerEfficient = decoder.decode<bool>();
    auto supportedConfiguration = decoder.decode<WebCore::MediaDecodingConfiguration>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaCapabilitiesDecodingInfo {
            WebCore::MediaCapabilitiesInfo {
                WTFMove(*supported),
                WTFMove(*smooth),
                WTFMove(*powerEfficient)
            },
            WTFMove(*supportedConfiguration)
        }
    };
}

void ArgumentCoder<WebCore::MediaCapabilitiesEncodingInfo>::encode(Encoder& encoder, const WebCore::MediaCapabilitiesEncodingInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supported)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.smooth)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.powerEfficient)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaCapabilitiesInfo, supported)
        , offsetof(WebCore::MediaCapabilitiesInfo, smooth)
        , offsetof(WebCore::MediaCapabilitiesInfo, powerEfficient)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportedConfiguration)>, WebCore::MediaEncodingConfiguration>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaCapabilitiesEncodingInfo, supportedConfiguration)
    >::value);
    encoder << instance.supported;
    encoder << instance.smooth;
    encoder << instance.powerEfficient;
    encoder << instance.supportedConfiguration;
}

std::optional<WebCore::MediaCapabilitiesEncodingInfo> ArgumentCoder<WebCore::MediaCapabilitiesEncodingInfo>::decode(Decoder& decoder)
{
    auto supported = decoder.decode<bool>();
    auto smooth = decoder.decode<bool>();
    auto powerEfficient = decoder.decode<bool>();
    auto supportedConfiguration = decoder.decode<WebCore::MediaEncodingConfiguration>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaCapabilitiesEncodingInfo {
            WebCore::MediaCapabilitiesInfo {
                WTFMove(*supported),
                WTFMove(*smooth),
                WTFMove(*powerEfficient)
            },
            WTFMove(*supportedConfiguration)
        }
    };
}

void ArgumentCoder<WebCore::FormDataElement>::encode(Encoder& encoder, const WebCore::FormDataElement& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data)>, std::variant<Vector<uint8_t>, WebCore::FormDataElement::EncodedFileData, WebCore::FormDataElement::EncodedBlobData>>);
    struct ShouldBeSameSizeAsFormDataElement : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FormDataElement>, false> {
        std::variant<Vector<uint8_t>, WebCore::FormDataElement::EncodedFileData, WebCore::FormDataElement::EncodedBlobData> data;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFormDataElement) == sizeof(WebCore::FormDataElement));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FormDataElement, data)
    >::value);
    encoder << instance.data;
}

std::optional<WebCore::FormDataElement> ArgumentCoder<WebCore::FormDataElement>::decode(Decoder& decoder)
{
    auto data = decoder.decode<std::variant<Vector<uint8_t>, WebCore::FormDataElement::EncodedFileData, WebCore::FormDataElement::EncodedBlobData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FormDataElement {
            WTFMove(*data)
        }
    };
}

void ArgumentCoder<WebCore::FormDataElement::EncodedFileData>::encode(Encoder& encoder, const WebCore::FormDataElement::EncodedFileData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.filename)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fileStart)>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fileLength)>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expectedFileModificationTime)>, std::optional<WallTime>>);
    struct ShouldBeSameSizeAsEncodedFileData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FormDataElement::EncodedFileData>, false> {
        String filename;
        int64_t fileStart;
        int64_t fileLength;
        std::optional<WallTime> expectedFileModificationTime;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEncodedFileData) == sizeof(WebCore::FormDataElement::EncodedFileData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FormDataElement::EncodedFileData, filename)
        , offsetof(WebCore::FormDataElement::EncodedFileData, fileStart)
        , offsetof(WebCore::FormDataElement::EncodedFileData, fileLength)
        , offsetof(WebCore::FormDataElement::EncodedFileData, expectedFileModificationTime)
    >::value);
    encoder << instance.filename;
    encoder << instance.fileStart;
    encoder << instance.fileLength;
    encoder << instance.expectedFileModificationTime;
}

std::optional<WebCore::FormDataElement::EncodedFileData> ArgumentCoder<WebCore::FormDataElement::EncodedFileData>::decode(Decoder& decoder)
{
    auto filename = decoder.decode<String>();
    auto fileStart = decoder.decode<int64_t>();
    auto fileLength = decoder.decode<int64_t>();
    auto expectedFileModificationTime = decoder.decode<std::optional<WallTime>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FormDataElement::EncodedFileData {
            WTFMove(*filename),
            WTFMove(*fileStart),
            WTFMove(*fileLength),
            WTFMove(*expectedFileModificationTime)
        }
    };
}

void ArgumentCoder<WebCore::FormDataElement::EncodedBlobData>::encode(Encoder& encoder, const WebCore::FormDataElement::EncodedBlobData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    struct ShouldBeSameSizeAsEncodedBlobData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FormDataElement::EncodedBlobData>, false> {
        URL url;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEncodedBlobData) == sizeof(WebCore::FormDataElement::EncodedBlobData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FormDataElement::EncodedBlobData, url)
    >::value);
    encoder << instance.url;
}

std::optional<WebCore::FormDataElement::EncodedBlobData> ArgumentCoder<WebCore::FormDataElement::EncodedBlobData>::decode(Decoder& decoder)
{
    auto url = decoder.decode<URL>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FormDataElement::EncodedBlobData {
            WTFMove(*url)
        }
    };
}

void ArgumentCoder<WebCore::NetworkTransactionInformation>::encode(Encoder& encoder, const WebCore::NetworkTransactionInformation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::NetworkTransactionInformation::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.request)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.response)>, WebCore::ResourceResponse>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.metrics)>, WebCore::NetworkLoadMetrics>);
    struct ShouldBeSameSizeAsNetworkTransactionInformation : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::NetworkTransactionInformation>, false> {
        WebCore::NetworkTransactionInformation::Type type;
        WebCore::ResourceRequest request;
        WebCore::ResourceResponse response;
        WebCore::NetworkLoadMetrics metrics;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNetworkTransactionInformation) == sizeof(WebCore::NetworkTransactionInformation));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::NetworkTransactionInformation, type)
        , offsetof(WebCore::NetworkTransactionInformation, request)
        , offsetof(WebCore::NetworkTransactionInformation, response)
        , offsetof(WebCore::NetworkTransactionInformation, metrics)
    >::value);
    encoder << instance.type;
    encoder << instance.request;
    encoder << instance.response;
    encoder << instance.metrics;
}

std::optional<WebCore::NetworkTransactionInformation> ArgumentCoder<WebCore::NetworkTransactionInformation>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::NetworkTransactionInformation::Type>();
    auto request = decoder.decode<WebCore::ResourceRequest>();
    auto response = decoder.decode<WebCore::ResourceResponse>();
    auto metrics = decoder.decode<WebCore::NetworkLoadMetrics>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::NetworkTransactionInformation {
            WTFMove(*type),
            WTFMove(*request),
            WTFMove(*response),
            WTFMove(*metrics)
        }
    };
}

void ArgumentCoder<WebCore::ContentType>::encode(Encoder& encoder, const WebCore::ContentType& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.raw())>, String>);
    encoder << instance.raw();
}

std::optional<WebCore::ContentType> ArgumentCoder<WebCore::ContentType>::decode(Decoder& decoder)
{
    auto raw = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ContentType {
            WTFMove(*raw)
        }
    };
}

#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
void ArgumentCoder<WebCore::DateTimeChooserParameters>::encode(Encoder& encoder, const WebCore::DateTimeChooserParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.anchorRectInRootView)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.locale)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.currentValue)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.suggestionValues)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localizedSuggestionValues)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.suggestionLabels)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimum)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximum)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.step)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stepBase)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.required)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isAnchorElementRTL)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useDarkAppearance)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasSecondField)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasMillisecondField)>, bool>);
    struct ShouldBeSameSizeAsDateTimeChooserParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DateTimeChooserParameters>, false> {
        AtomString type;
        WebCore::IntRect anchorRectInRootView;
        AtomString locale;
        String currentValue;
        Vector<String> suggestionValues;
        Vector<String> localizedSuggestionValues;
        Vector<String> suggestionLabels;
        double minimum;
        double maximum;
        double step;
        double stepBase;
        bool required;
        bool isAnchorElementRTL;
        bool useDarkAppearance;
        bool hasSecondField;
        bool hasMillisecondField;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDateTimeChooserParameters) == sizeof(WebCore::DateTimeChooserParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DateTimeChooserParameters, type)
        , offsetof(WebCore::DateTimeChooserParameters, anchorRectInRootView)
        , offsetof(WebCore::DateTimeChooserParameters, locale)
        , offsetof(WebCore::DateTimeChooserParameters, currentValue)
        , offsetof(WebCore::DateTimeChooserParameters, suggestionValues)
        , offsetof(WebCore::DateTimeChooserParameters, localizedSuggestionValues)
        , offsetof(WebCore::DateTimeChooserParameters, suggestionLabels)
        , offsetof(WebCore::DateTimeChooserParameters, minimum)
        , offsetof(WebCore::DateTimeChooserParameters, maximum)
        , offsetof(WebCore::DateTimeChooserParameters, step)
        , offsetof(WebCore::DateTimeChooserParameters, stepBase)
        , offsetof(WebCore::DateTimeChooserParameters, required)
        , offsetof(WebCore::DateTimeChooserParameters, isAnchorElementRTL)
        , offsetof(WebCore::DateTimeChooserParameters, useDarkAppearance)
        , offsetof(WebCore::DateTimeChooserParameters, hasSecondField)
        , offsetof(WebCore::DateTimeChooserParameters, hasMillisecondField)
    >::value);
    encoder << instance.type;
    encoder << instance.anchorRectInRootView;
    encoder << instance.locale;
    encoder << instance.currentValue;
    encoder << instance.suggestionValues;
    encoder << instance.localizedSuggestionValues;
    encoder << instance.suggestionLabels;
    encoder << instance.minimum;
    encoder << instance.maximum;
    encoder << instance.step;
    encoder << instance.stepBase;
    encoder << instance.required;
    encoder << instance.isAnchorElementRTL;
    encoder << instance.useDarkAppearance;
    encoder << instance.hasSecondField;
    encoder << instance.hasMillisecondField;
}

std::optional<WebCore::DateTimeChooserParameters> ArgumentCoder<WebCore::DateTimeChooserParameters>::decode(Decoder& decoder)
{
    auto type = decoder.decode<AtomString>();
    auto anchorRectInRootView = decoder.decode<WebCore::IntRect>();
    auto locale = decoder.decode<AtomString>();
    auto currentValue = decoder.decode<String>();
    auto suggestionValues = decoder.decode<Vector<String>>();
    auto localizedSuggestionValues = decoder.decode<Vector<String>>();
    auto suggestionLabels = decoder.decode<Vector<String>>();
    auto minimum = decoder.decode<double>();
    auto maximum = decoder.decode<double>();
    auto step = decoder.decode<double>();
    auto stepBase = decoder.decode<double>();
    auto required = decoder.decode<bool>();
    auto isAnchorElementRTL = decoder.decode<bool>();
    auto useDarkAppearance = decoder.decode<bool>();
    auto hasSecondField = decoder.decode<bool>();
    auto hasMillisecondField = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DateTimeChooserParameters {
            WTFMove(*type),
            WTFMove(*anchorRectInRootView),
            WTFMove(*locale),
            WTFMove(*currentValue),
            WTFMove(*suggestionValues),
            WTFMove(*localizedSuggestionValues),
            WTFMove(*suggestionLabels),
            WTFMove(*minimum),
            WTFMove(*maximum),
            WTFMove(*step),
            WTFMove(*stepBase),
            WTFMove(*required),
            WTFMove(*isAnchorElementRTL),
            WTFMove(*useDarkAppearance),
            WTFMove(*hasSecondField),
            WTFMove(*hasMillisecondField)
        }
    };
}

#endif

void ArgumentCoder<WebCore::ScreenData>::encode(Encoder& encoder, const WebCore::ScreenData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenAvailableRect)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenRect)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::DestinationColorSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenDepth)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenDepthPerComponent)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenSupportsExtendedColor)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenHasInvertedColors)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenSupportsHighDynamicRange)>, bool>);
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenIsMonochrome)>, bool>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.displayMask)>, uint32_t>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gpuID)>, WebCore::PlatformGPUID>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preferredDynamicRangeMode)>, WebCore::DynamicRangeMode>);
#endif
#if PLATFORM(MAC) || PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scaleFactor)>, float>);
#endif
    struct ShouldBeSameSizeAsScreenData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ScreenData>, false> {
        WebCore::FloatRect screenAvailableRect;
        WebCore::FloatRect screenRect;
        WebCore::DestinationColorSpace colorSpace;
        int screenDepth;
        int screenDepthPerComponent;
        bool screenSupportsExtendedColor;
        bool screenHasInvertedColors;
        bool screenSupportsHighDynamicRange;
#if PLATFORM(MAC)
        bool screenIsMonochrome;
#endif
#if PLATFORM(MAC)
        uint32_t displayMask;
#endif
#if PLATFORM(MAC)
        WebCore::PlatformGPUID gpuID;
#endif
#if PLATFORM(MAC)
        WebCore::DynamicRangeMode preferredDynamicRangeMode;
#endif
#if PLATFORM(MAC) || PLATFORM(IOS_FAMILY)
        float scaleFactor;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsScreenData) == sizeof(WebCore::ScreenData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ScreenData, screenAvailableRect)
        , offsetof(WebCore::ScreenData, screenRect)
        , offsetof(WebCore::ScreenData, colorSpace)
        , offsetof(WebCore::ScreenData, screenDepth)
        , offsetof(WebCore::ScreenData, screenDepthPerComponent)
        , offsetof(WebCore::ScreenData, screenSupportsExtendedColor)
        , offsetof(WebCore::ScreenData, screenHasInvertedColors)
        , offsetof(WebCore::ScreenData, screenSupportsHighDynamicRange)
#if PLATFORM(MAC)
        , offsetof(WebCore::ScreenData, screenIsMonochrome)
#endif
#if PLATFORM(MAC)
        , offsetof(WebCore::ScreenData, displayMask)
#endif
#if PLATFORM(MAC)
        , offsetof(WebCore::ScreenData, gpuID)
#endif
#if PLATFORM(MAC)
        , offsetof(WebCore::ScreenData, preferredDynamicRangeMode)
#endif
#if PLATFORM(MAC) || PLATFORM(IOS_FAMILY)
        , offsetof(WebCore::ScreenData, scaleFactor)
#endif
    >::value);
    encoder << instance.screenAvailableRect;
    encoder << instance.screenRect;
    encoder << instance.colorSpace;
    encoder << instance.screenDepth;
    encoder << instance.screenDepthPerComponent;
    encoder << instance.screenSupportsExtendedColor;
    encoder << instance.screenHasInvertedColors;
    encoder << instance.screenSupportsHighDynamicRange;
#if PLATFORM(MAC)
    encoder << instance.screenIsMonochrome;
#endif
#if PLATFORM(MAC)
    encoder << instance.displayMask;
#endif
#if PLATFORM(MAC)
    encoder << instance.gpuID;
#endif
#if PLATFORM(MAC)
    encoder << instance.preferredDynamicRangeMode;
#endif
#if PLATFORM(MAC) || PLATFORM(IOS_FAMILY)
    encoder << instance.scaleFactor;
#endif
}

std::optional<WebCore::ScreenData> ArgumentCoder<WebCore::ScreenData>::decode(Decoder& decoder)
{
    auto screenAvailableRect = decoder.decode<WebCore::FloatRect>();
    auto screenRect = decoder.decode<WebCore::FloatRect>();
    auto colorSpace = decoder.decode<WebCore::DestinationColorSpace>();
    auto screenDepth = decoder.decode<int>();
    auto screenDepthPerComponent = decoder.decode<int>();
    auto screenSupportsExtendedColor = decoder.decode<bool>();
    auto screenHasInvertedColors = decoder.decode<bool>();
    auto screenSupportsHighDynamicRange = decoder.decode<bool>();
#if PLATFORM(MAC)
    auto screenIsMonochrome = decoder.decode<bool>();
#endif
#if PLATFORM(MAC)
    auto displayMask = decoder.decode<uint32_t>();
#endif
#if PLATFORM(MAC)
    auto gpuID = decoder.decode<WebCore::PlatformGPUID>();
#endif
#if PLATFORM(MAC)
    auto preferredDynamicRangeMode = decoder.decode<WebCore::DynamicRangeMode>();
#endif
#if PLATFORM(MAC) || PLATFORM(IOS_FAMILY)
    auto scaleFactor = decoder.decode<float>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ScreenData {
            WTFMove(*screenAvailableRect),
            WTFMove(*screenRect),
            WTFMove(*colorSpace),
            WTFMove(*screenDepth),
            WTFMove(*screenDepthPerComponent),
            WTFMove(*screenSupportsExtendedColor),
            WTFMove(*screenHasInvertedColors),
            WTFMove(*screenSupportsHighDynamicRange),
#if PLATFORM(MAC)
            WTFMove(*screenIsMonochrome),
#endif
#if PLATFORM(MAC)
            WTFMove(*displayMask),
#endif
#if PLATFORM(MAC)
            WTFMove(*gpuID),
#endif
#if PLATFORM(MAC)
            WTFMove(*preferredDynamicRangeMode),
#endif
#if PLATFORM(MAC) || PLATFORM(IOS_FAMILY)
            WTFMove(*scaleFactor)
#endif
        }
    };
}

void ArgumentCoder<WebCore::ScreenProperties>::encode(Encoder& encoder, const WebCore::ScreenProperties& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primaryDisplayID)>, WebCore::PlatformDisplayID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenDataMap)>, HashMap<WebCore::PlatformDisplayID, WebCore::ScreenData>>);
    struct ShouldBeSameSizeAsScreenProperties : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ScreenProperties>, false> {
        WebCore::PlatformDisplayID primaryDisplayID;
        HashMap<WebCore::PlatformDisplayID, WebCore::ScreenData> screenDataMap;
    };
    static_assert(sizeof(ShouldBeSameSizeAsScreenProperties) == sizeof(WebCore::ScreenProperties));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ScreenProperties, primaryDisplayID)
        , offsetof(WebCore::ScreenProperties, screenDataMap)
    >::value);
    encoder << instance.primaryDisplayID;
    encoder << instance.screenDataMap;
}

std::optional<WebCore::ScreenProperties> ArgumentCoder<WebCore::ScreenProperties>::decode(Decoder& decoder)
{
    auto primaryDisplayID = decoder.decode<WebCore::PlatformDisplayID>();
    auto screenDataMap = decoder.decode<HashMap<WebCore::PlatformDisplayID, WebCore::ScreenData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ScreenProperties {
            WTFMove(*primaryDisplayID),
            WTFMove(*screenDataMap)
        }
    };
}

void ArgumentCoder<WebCore::PlatformTimeRanges>::encode(Encoder& encoder, const WebCore::PlatformTimeRanges& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_ranges)>, Vector<WebCore::PlatformTimeRanges::Range>>);
    struct ShouldBeSameSizeAsPlatformTimeRanges : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PlatformTimeRanges>, false> {
        Vector<WebCore::PlatformTimeRanges::Range> m_ranges;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPlatformTimeRanges) == sizeof(WebCore::PlatformTimeRanges));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformTimeRanges, m_ranges)
    >::value);
    encoder << instance.m_ranges;
}

std::optional<WebCore::PlatformTimeRanges> ArgumentCoder<WebCore::PlatformTimeRanges>::decode(Decoder& decoder)
{
    auto m_ranges = decoder.decode<Vector<WebCore::PlatformTimeRanges::Range>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PlatformTimeRanges {
            WTFMove(*m_ranges)
        }
    };
}

void ArgumentCoder<WebCore::PlatformTimeRanges::Range>::encode(Encoder& encoder, const WebCore::PlatformTimeRanges::Range& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.start)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.end)>, MediaTime>);
    struct ShouldBeSameSizeAsRange : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PlatformTimeRanges::Range>, false> {
        MediaTime start;
        MediaTime end;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRange) == sizeof(WebCore::PlatformTimeRanges::Range));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformTimeRanges::Range, start)
        , offsetof(WebCore::PlatformTimeRanges::Range, end)
    >::value);
    encoder << instance.start;
    encoder << instance.end;
}

std::optional<WebCore::PlatformTimeRanges::Range> ArgumentCoder<WebCore::PlatformTimeRanges::Range>::decode(Decoder& decoder)
{
    auto start = decoder.decode<MediaTime>();
    auto end = decoder.decode<MediaTime>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PlatformTimeRanges::Range {
            WTFMove(*start),
            WTFMove(*end)
        }
    };
}

#if ENABLE(VIDEO)
void ArgumentCoder<WebCore::VideoPlaybackQualityMetrics>::encode(Encoder& encoder, const WebCore::VideoPlaybackQualityMetrics& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.totalVideoFrames)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.droppedVideoFrames)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.corruptedVideoFrames)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.totalFrameDelay)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.displayCompositedVideoFrames)>, uint32_t>);
    struct ShouldBeSameSizeAsVideoPlaybackQualityMetrics : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::VideoPlaybackQualityMetrics>, false> {
        uint32_t totalVideoFrames;
        uint32_t droppedVideoFrames;
        uint32_t corruptedVideoFrames;
        double totalFrameDelay;
        uint32_t displayCompositedVideoFrames;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVideoPlaybackQualityMetrics) == sizeof(WebCore::VideoPlaybackQualityMetrics));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::VideoPlaybackQualityMetrics, totalVideoFrames)
        , offsetof(WebCore::VideoPlaybackQualityMetrics, droppedVideoFrames)
        , offsetof(WebCore::VideoPlaybackQualityMetrics, corruptedVideoFrames)
        , offsetof(WebCore::VideoPlaybackQualityMetrics, totalFrameDelay)
        , offsetof(WebCore::VideoPlaybackQualityMetrics, displayCompositedVideoFrames)
    >::value);
    encoder << instance.totalVideoFrames;
    encoder << instance.droppedVideoFrames;
    encoder << instance.corruptedVideoFrames;
    encoder << instance.totalFrameDelay;
    encoder << instance.displayCompositedVideoFrames;
}

std::optional<WebCore::VideoPlaybackQualityMetrics> ArgumentCoder<WebCore::VideoPlaybackQualityMetrics>::decode(Decoder& decoder)
{
    auto totalVideoFrames = decoder.decode<uint32_t>();
    auto droppedVideoFrames = decoder.decode<uint32_t>();
    auto corruptedVideoFrames = decoder.decode<uint32_t>();
    auto totalFrameDelay = decoder.decode<double>();
    auto displayCompositedVideoFrames = decoder.decode<uint32_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::VideoPlaybackQualityMetrics {
            WTFMove(*totalVideoFrames),
            WTFMove(*droppedVideoFrames),
            WTFMove(*corruptedVideoFrames),
            WTFMove(*totalFrameDelay),
            WTFMove(*displayCompositedVideoFrames)
        }
    };
}

#endif

#if PLATFORM(IOS_FAMILY)
void ArgumentCoder<WebCore::SelectionGeometry>::encode(Encoder& encoder, const WebCore::SelectionGeometry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.quad())>, WebCore::FloatQuad>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.behavior())>, WebCore::SelectionRenderingBehavior>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction())>, WebCore::TextDirection>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minX())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxX())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxY())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lineNumber())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isLineBreak())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFirstOnLine())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isLastOnLine())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containsStart())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containsEnd())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isHorizontal())>, bool>);
    encoder << instance.quad();
    encoder << instance.behavior();
    encoder << instance.direction();
    encoder << instance.minX();
    encoder << instance.maxX();
    encoder << instance.maxY();
    encoder << instance.lineNumber();
    encoder << instance.isLineBreak();
    encoder << instance.isFirstOnLine();
    encoder << instance.isLastOnLine();
    encoder << instance.containsStart();
    encoder << instance.containsEnd();
    encoder << instance.isHorizontal();
}

std::optional<WebCore::SelectionGeometry> ArgumentCoder<WebCore::SelectionGeometry>::decode(Decoder& decoder)
{
    auto quad = decoder.decode<WebCore::FloatQuad>();
    auto behavior = decoder.decode<WebCore::SelectionRenderingBehavior>();
    auto direction = decoder.decode<WebCore::TextDirection>();
    auto minX = decoder.decode<int>();
    auto maxX = decoder.decode<int>();
    auto maxY = decoder.decode<int>();
    auto lineNumber = decoder.decode<int>();
    auto isLineBreak = decoder.decode<bool>();
    auto isFirstOnLine = decoder.decode<bool>();
    auto isLastOnLine = decoder.decode<bool>();
    auto containsStart = decoder.decode<bool>();
    auto containsEnd = decoder.decode<bool>();
    auto isHorizontal = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SelectionGeometry {
            WTFMove(*quad),
            WTFMove(*behavior),
            WTFMove(*direction),
            WTFMove(*minX),
            WTFMove(*maxX),
            WTFMove(*maxY),
            WTFMove(*lineNumber),
            WTFMove(*isLineBreak),
            WTFMove(*isFirstOnLine),
            WTFMove(*isLastOnLine),
            WTFMove(*containsStart),
            WTFMove(*containsEnd),
            WTFMove(*isHorizontal)
        }
    };
}

#endif

#if ENABLE(DATALIST_ELEMENT)
void ArgumentCoder<WebCore::DataListSuggestion>::encode(Encoder& encoder, const WebCore::DataListSuggestion& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    struct ShouldBeSameSizeAsDataListSuggestion : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DataListSuggestion>, false> {
        String value;
        String label;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDataListSuggestion) == sizeof(WebCore::DataListSuggestion));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DataListSuggestion, value)
        , offsetof(WebCore::DataListSuggestion, label)
    >::value);
    encoder << instance.value;
    encoder << instance.label;
}

std::optional<WebCore::DataListSuggestion> ArgumentCoder<WebCore::DataListSuggestion>::decode(Decoder& decoder)
{
    auto value = decoder.decode<String>();
    auto label = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DataListSuggestion {
            WTFMove(*value),
            WTFMove(*label)
        }
    };
}

#endif

#if ENABLE(DATALIST_ELEMENT)
void ArgumentCoder<WebCore::DataListSuggestionInformation>::encode(Encoder& encoder, const WebCore::DataListSuggestionInformation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.activationType)>, WebCore::DataListSuggestionActivationType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.suggestions)>, Vector<WebCore::DataListSuggestion>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elementRect)>, WebCore::IntRect>);
    struct ShouldBeSameSizeAsDataListSuggestionInformation : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DataListSuggestionInformation>, false> {
        WebCore::DataListSuggestionActivationType activationType;
        Vector<WebCore::DataListSuggestion> suggestions;
        WebCore::IntRect elementRect;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDataListSuggestionInformation) == sizeof(WebCore::DataListSuggestionInformation));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DataListSuggestionInformation, activationType)
        , offsetof(WebCore::DataListSuggestionInformation, suggestions)
        , offsetof(WebCore::DataListSuggestionInformation, elementRect)
    >::value);
    encoder << instance.activationType;
    encoder << instance.suggestions;
    encoder << instance.elementRect;
}

std::optional<WebCore::DataListSuggestionInformation> ArgumentCoder<WebCore::DataListSuggestionInformation>::decode(Decoder& decoder)
{
    auto activationType = decoder.decode<WebCore::DataListSuggestionActivationType>();
    auto suggestions = decoder.decode<Vector<WebCore::DataListSuggestion>>();
    auto elementRect = decoder.decode<WebCore::IntRect>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DataListSuggestionInformation {
            WTFMove(*activationType),
            WTFMove(*suggestions),
            WTFMove(*elementRect)
        }
    };
}

#endif

void ArgumentCoder<WebCore::ClientOrigin>::encode(Encoder& encoder, const WebCore::ClientOrigin& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topOrigin)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientOrigin)>, WebCore::SecurityOriginData>);
    struct ShouldBeSameSizeAsClientOrigin : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ClientOrigin>, false> {
        WebCore::SecurityOriginData topOrigin;
        WebCore::SecurityOriginData clientOrigin;
    };
    static_assert(sizeof(ShouldBeSameSizeAsClientOrigin) == sizeof(WebCore::ClientOrigin));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ClientOrigin, topOrigin)
        , offsetof(WebCore::ClientOrigin, clientOrigin)
    >::value);
    encoder << instance.topOrigin;
    encoder << instance.clientOrigin;
}

std::optional<WebCore::ClientOrigin> ArgumentCoder<WebCore::ClientOrigin>::decode(Decoder& decoder)
{
    auto topOrigin = decoder.decode<WebCore::SecurityOriginData>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(!topOrigin->isNull()))
        return std::nullopt;
    auto clientOrigin = decoder.decode<WebCore::SecurityOriginData>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(!clientOrigin->isNull()))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ClientOrigin {
            WTFMove(*topOrigin),
            WTFMove(*clientOrigin)
        }
    };
}

void ArgumentCoder<WebCore::PixelBufferFormat>::encode(Encoder& encoder, const WebCore::PixelBufferFormat& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaFormat)>, WebCore::AlphaPremultiplication>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pixelFormat)>, WebCore::PixelFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::DestinationColorSpace>);
    struct ShouldBeSameSizeAsPixelBufferFormat : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PixelBufferFormat>, false> {
        WebCore::AlphaPremultiplication alphaFormat;
        WebCore::PixelFormat pixelFormat;
        WebCore::DestinationColorSpace colorSpace;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPixelBufferFormat) == sizeof(WebCore::PixelBufferFormat));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PixelBufferFormat, alphaFormat)
        , offsetof(WebCore::PixelBufferFormat, pixelFormat)
        , offsetof(WebCore::PixelBufferFormat, colorSpace)
    >::value);
    encoder << instance.alphaFormat;
    encoder << instance.pixelFormat;
    encoder << instance.colorSpace;
}

void ArgumentCoder<WebCore::PixelBufferFormat>::encode(StreamConnectionEncoder& encoder, const WebCore::PixelBufferFormat& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaFormat)>, WebCore::AlphaPremultiplication>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pixelFormat)>, WebCore::PixelFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::DestinationColorSpace>);
    struct ShouldBeSameSizeAsPixelBufferFormat : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PixelBufferFormat>, false> {
        WebCore::AlphaPremultiplication alphaFormat;
        WebCore::PixelFormat pixelFormat;
        WebCore::DestinationColorSpace colorSpace;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPixelBufferFormat) == sizeof(WebCore::PixelBufferFormat));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PixelBufferFormat, alphaFormat)
        , offsetof(WebCore::PixelBufferFormat, pixelFormat)
        , offsetof(WebCore::PixelBufferFormat, colorSpace)
    >::value);
    encoder << instance.alphaFormat;
    encoder << instance.pixelFormat;
    encoder << instance.colorSpace;
}

std::optional<WebCore::PixelBufferFormat> ArgumentCoder<WebCore::PixelBufferFormat>::decode(Decoder& decoder)
{
    auto alphaFormat = decoder.decode<WebCore::AlphaPremultiplication>();
    auto pixelFormat = decoder.decode<WebCore::PixelFormat>();
    auto colorSpace = decoder.decode<WebCore::DestinationColorSpace>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PixelBufferFormat {
            WTFMove(*alphaFormat),
            WTFMove(*pixelFormat),
            WTFMove(*colorSpace)
        }
    };
}

void ArgumentCoder<WebCore::TextIndicator>::encode(Encoder& encoder, const WebCore::TextIndicator& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, WebCore::TextIndicatorData>);
    encoder << instance.data();
}

std::optional<Ref<WebCore::TextIndicator>> ArgumentCoder<WebCore::TextIndicator>::decode(Decoder& decoder)
{
    auto data = decoder.decode<WebCore::TextIndicatorData>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextIndicator::create(
            WTFMove(*data)
        )
    };
}

void ArgumentCoder<WebCore::CertificateInfo>::encode(Encoder& encoder, const WebCore::CertificateInfo& instance)
{
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trust())>, RetainPtr<SecTrustRef>>);
#endif
#if USE(CURL)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.verificationError())>, int>);
#endif
#if USE(CURL)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.certificateChain())>, Vector<Vector<uint8_t>>>);
#endif
#if USE(SOUP)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.certificate())>, GRefPtr<GTlsCertificate>>);
#endif
#if USE(SOUP)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tlsErrors())>, GTlsCertificateFlags>);
#endif
#if PLATFORM(COCOA)
    encoder << instance.trust();
#endif
#if USE(CURL)
    encoder << instance.verificationError();
#endif
#if USE(CURL)
    encoder << instance.certificateChain();
#endif
#if USE(SOUP)
    encoder << instance.certificate();
#endif
#if USE(SOUP)
    encoder << instance.tlsErrors();
#endif
}

std::optional<WebCore::CertificateInfo> ArgumentCoder<WebCore::CertificateInfo>::decode(Decoder& decoder)
{
#if PLATFORM(COCOA)
    auto trust = decoder.decode<RetainPtr<SecTrustRef>>();
#endif
#if USE(CURL)
    auto verificationError = decoder.decode<int>();
#endif
#if USE(CURL)
    auto certificateChain = decoder.decode<Vector<Vector<uint8_t>>>();
#endif
#if USE(SOUP)
    auto certificate = decoder.decode<GRefPtr<GTlsCertificate>>();
#endif
#if USE(SOUP)
    auto tlsErrors = decoder.decode<GTlsCertificateFlags>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CertificateInfo {
#if PLATFORM(COCOA)
            WTFMove(*trust),
#endif
#if USE(CURL)
            WTFMove(*verificationError),
#endif
#if USE(CURL)
            WTFMove(*certificateChain),
#endif
#if USE(SOUP)
            WTFMove(*certificate),
#endif
#if USE(SOUP)
            WTFMove(*tlsErrors)
#endif
        }
    };
}

void ArgumentCoder<WebCore::PasteboardCustomData::Entry>::encode(Encoder& encoder, const WebCore::PasteboardCustomData::Entry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.customData)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.platformData)>, std::variant<String, Ref<WebCore::SharedBuffer>>>);
    struct ShouldBeSameSizeAsEntry : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PasteboardCustomData::Entry>, false> {
        String type;
        String customData;
        std::variant<String, Ref<WebCore::SharedBuffer>> platformData;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEntry) == sizeof(WebCore::PasteboardCustomData::Entry));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PasteboardCustomData::Entry, type)
        , offsetof(WebCore::PasteboardCustomData::Entry, customData)
        , offsetof(WebCore::PasteboardCustomData::Entry, platformData)
    >::value);
    encoder << instance.type;
    encoder << instance.customData;
    encoder << instance.platformData;
}

std::optional<WebCore::PasteboardCustomData::Entry> ArgumentCoder<WebCore::PasteboardCustomData::Entry>::decode(Decoder& decoder)
{
    auto type = decoder.decode<String>();
    auto customData = decoder.decode<String>();
    auto platformData = decoder.decode<std::variant<String, Ref<WebCore::SharedBuffer>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PasteboardCustomData::Entry {
            WTFMove(*type),
            WTFMove(*customData),
            WTFMove(*platformData)
        }
    };
}

void ArgumentCoder<WebCore::PasteboardCustomData>::encode(Encoder& encoder, const WebCore::PasteboardCustomData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, Vector<WebCore::PasteboardCustomData::Entry>>);
    encoder << instance.origin();
    encoder << instance.data();
}

std::optional<WebCore::PasteboardCustomData> ArgumentCoder<WebCore::PasteboardCustomData>::decode(Decoder& decoder)
{
    auto origin = decoder.decode<String>();
    auto data = decoder.decode<Vector<WebCore::PasteboardCustomData::Entry>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PasteboardCustomData {
            WTFMove(*origin),
            WTFMove(*data)
        }
    };
}

void ArgumentCoder<WebCore::DatabaseDetails>::encode(Encoder& encoder, const WebCore::DatabaseDetails& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.displayName())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expectedUsage())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.currentUsage())>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.creationTime())>, std::optional<WallTime>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modificationTime())>, std::optional<WallTime>>);
    encoder << instance.name();
    encoder << instance.displayName();
    encoder << instance.expectedUsage();
    encoder << instance.currentUsage();
    encoder << instance.creationTime();
    encoder << instance.modificationTime();
}

std::optional<WebCore::DatabaseDetails> ArgumentCoder<WebCore::DatabaseDetails>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto displayName = decoder.decode<String>();
    auto expectedUsage = decoder.decode<uint64_t>();
    auto currentUsage = decoder.decode<uint64_t>();
    auto creationTime = decoder.decode<std::optional<WallTime>>();
    auto modificationTime = decoder.decode<std::optional<WallTime>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DatabaseDetails {
            WTFMove(*name),
            WTFMove(*displayName),
            WTFMove(*expectedUsage),
            WTFMove(*currentUsage),
            WTFMove(*creationTime),
            WTFMove(*modificationTime)
        }
    };
}

void ArgumentCoder<WebCore::DecomposedGlyphs>::encode(Encoder& encoder, const WebCore::DecomposedGlyphs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.positionedGlyphs())>, WebCore::PositionedGlyphs>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.renderingResourceIdentifier())>, WebCore::RenderingResourceIdentifier>);
    encoder << instance.positionedGlyphs();
    encoder << instance.renderingResourceIdentifier();
}

std::optional<Ref<WebCore::DecomposedGlyphs>> ArgumentCoder<WebCore::DecomposedGlyphs>::decode(Decoder& decoder)
{
    auto positionedGlyphs = decoder.decode<WebCore::PositionedGlyphs>();
    auto renderingResourceIdentifier = decoder.decode<WebCore::RenderingResourceIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DecomposedGlyphs::create(
            WTFMove(*positionedGlyphs),
            WTFMove(*renderingResourceIdentifier)
        )
    };
}

void ArgumentCoder<WebCore::TextIndicatorData>::encode(Encoder& encoder, const WebCore::TextIndicatorData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.selectionRectInRootViewCoordinates)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textBoundingRectInRootViewCoordinates)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentImageWithoutSelectionRectInRootViewCoordinates)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textRectsInBoundingRectCoordinates)>, Vector<WebCore::FloatRect>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentImageScaleFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentImageWithHighlight)>, RefPtr<WebCore::Image>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentImageWithoutSelection)>, RefPtr<WebCore::Image>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentImage)>, RefPtr<WebCore::Image>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.estimatedBackgroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentationTransition)>, WebCore::TextIndicatorPresentationTransition>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.options)>, OptionSet<WebCore::TextIndicatorOption>>);
    struct ShouldBeSameSizeAsTextIndicatorData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextIndicatorData>, false> {
        WebCore::FloatRect selectionRectInRootViewCoordinates;
        WebCore::FloatRect textBoundingRectInRootViewCoordinates;
        WebCore::FloatRect contentImageWithoutSelectionRectInRootViewCoordinates;
        Vector<WebCore::FloatRect> textRectsInBoundingRectCoordinates;
        float contentImageScaleFactor;
        RefPtr<WebCore::Image> contentImageWithHighlight;
        RefPtr<WebCore::Image> contentImageWithoutSelection;
        RefPtr<WebCore::Image> contentImage;
        WebCore::Color estimatedBackgroundColor;
        WebCore::TextIndicatorPresentationTransition presentationTransition;
        OptionSet<WebCore::TextIndicatorOption> options;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextIndicatorData) == sizeof(WebCore::TextIndicatorData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextIndicatorData, selectionRectInRootViewCoordinates)
        , offsetof(WebCore::TextIndicatorData, textBoundingRectInRootViewCoordinates)
        , offsetof(WebCore::TextIndicatorData, contentImageWithoutSelectionRectInRootViewCoordinates)
        , offsetof(WebCore::TextIndicatorData, textRectsInBoundingRectCoordinates)
        , offsetof(WebCore::TextIndicatorData, contentImageScaleFactor)
        , offsetof(WebCore::TextIndicatorData, contentImageWithHighlight)
        , offsetof(WebCore::TextIndicatorData, contentImageWithoutSelection)
        , offsetof(WebCore::TextIndicatorData, contentImage)
        , offsetof(WebCore::TextIndicatorData, estimatedBackgroundColor)
        , offsetof(WebCore::TextIndicatorData, presentationTransition)
        , offsetof(WebCore::TextIndicatorData, options)
    >::value);
    encoder << instance.selectionRectInRootViewCoordinates;
    encoder << instance.textBoundingRectInRootViewCoordinates;
    encoder << instance.contentImageWithoutSelectionRectInRootViewCoordinates;
    encoder << instance.textRectsInBoundingRectCoordinates;
    encoder << instance.contentImageScaleFactor;
    encoder << instance.contentImageWithHighlight;
    encoder << instance.contentImageWithoutSelection;
    encoder << instance.contentImage;
    encoder << instance.estimatedBackgroundColor;
    encoder << instance.presentationTransition;
    encoder << instance.options;
}

std::optional<WebCore::TextIndicatorData> ArgumentCoder<WebCore::TextIndicatorData>::decode(Decoder& decoder)
{
    auto selectionRectInRootViewCoordinates = decoder.decode<WebCore::FloatRect>();
    auto textBoundingRectInRootViewCoordinates = decoder.decode<WebCore::FloatRect>();
    auto contentImageWithoutSelectionRectInRootViewCoordinates = decoder.decode<WebCore::FloatRect>();
    auto textRectsInBoundingRectCoordinates = decoder.decode<Vector<WebCore::FloatRect>>();
    auto contentImageScaleFactor = decoder.decode<float>();
    auto contentImageWithHighlight = decoder.decode<RefPtr<WebCore::Image>>();
    auto contentImageWithoutSelection = decoder.decode<RefPtr<WebCore::Image>>();
    auto contentImage = decoder.decode<RefPtr<WebCore::Image>>();
    auto estimatedBackgroundColor = decoder.decode<WebCore::Color>();
    auto presentationTransition = decoder.decode<WebCore::TextIndicatorPresentationTransition>();
    auto options = decoder.decode<OptionSet<WebCore::TextIndicatorOption>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextIndicatorData {
            WTFMove(*selectionRectInRootViewCoordinates),
            WTFMove(*textBoundingRectInRootViewCoordinates),
            WTFMove(*contentImageWithoutSelectionRectInRootViewCoordinates),
            WTFMove(*textRectsInBoundingRectCoordinates),
            WTFMove(*contentImageScaleFactor),
            WTFMove(*contentImageWithHighlight),
            WTFMove(*contentImageWithoutSelection),
            WTFMove(*contentImage),
            WTFMove(*estimatedBackgroundColor),
            WTFMove(*presentationTransition),
            WTFMove(*options)
        }
    };
}

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::MediaConstraints>::encode(Encoder& encoder, const WebCore::MediaConstraints& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mandatoryConstraints)>, WebCore::MediaTrackConstraintSetMap>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.advancedConstraints)>, Vector<WebCore::MediaTrackConstraintSetMap>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isValid)>, bool>);
    struct ShouldBeSameSizeAsMediaConstraints : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaConstraints>, false> {
        WebCore::MediaTrackConstraintSetMap mandatoryConstraints;
        Vector<WebCore::MediaTrackConstraintSetMap> advancedConstraints;
        bool isValid;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaConstraints) == sizeof(WebCore::MediaConstraints));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaConstraints, mandatoryConstraints)
        , offsetof(WebCore::MediaConstraints, advancedConstraints)
        , offsetof(WebCore::MediaConstraints, isValid)
    >::value);
    encoder << instance.mandatoryConstraints;
    encoder << instance.advancedConstraints;
    encoder << instance.isValid;
}

std::optional<WebCore::MediaConstraints> ArgumentCoder<WebCore::MediaConstraints>::decode(Decoder& decoder)
{
    auto mandatoryConstraints = decoder.decode<WebCore::MediaTrackConstraintSetMap>();
    auto advancedConstraints = decoder.decode<Vector<WebCore::MediaTrackConstraintSetMap>>();
    auto isValid = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaConstraints {
            WTFMove(*mandatoryConstraints),
            WTFMove(*advancedConstraints),
            WTFMove(*isValid)
        }
    };
}

#endif

void ArgumentCoder<WebCore::PromisedAttachmentInfo>::encode(Encoder& encoder, const WebCore::PromisedAttachmentInfo& instance)
{
#if ENABLE(ATTACHMENT_ELEMENT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.attachmentIdentifier)>, String>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.additionalTypes)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.additionalData)>, Vector<RefPtr<WebCore::SharedBuffer>>>);
    struct ShouldBeSameSizeAsPromisedAttachmentInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PromisedAttachmentInfo>, false> {
#if ENABLE(ATTACHMENT_ELEMENT)
        String attachmentIdentifier;
#endif
        Vector<String> additionalTypes;
        Vector<RefPtr<WebCore::SharedBuffer>> additionalData;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPromisedAttachmentInfo) == sizeof(WebCore::PromisedAttachmentInfo));
    static_assert(MembersInCorrectOrder < 0
#if ENABLE(ATTACHMENT_ELEMENT)
        , offsetof(WebCore::PromisedAttachmentInfo, attachmentIdentifier)
#endif
        , offsetof(WebCore::PromisedAttachmentInfo, additionalTypes)
        , offsetof(WebCore::PromisedAttachmentInfo, additionalData)
    >::value);
#if ENABLE(ATTACHMENT_ELEMENT)
    encoder << instance.attachmentIdentifier;
#endif
    encoder << instance.additionalTypes;
    encoder << instance.additionalData;
}

std::optional<WebCore::PromisedAttachmentInfo> ArgumentCoder<WebCore::PromisedAttachmentInfo>::decode(Decoder& decoder)
{
#if ENABLE(ATTACHMENT_ELEMENT)
    auto attachmentIdentifier = decoder.decode<String>();
#endif
    auto additionalTypes = decoder.decode<Vector<String>>();
    auto additionalData = decoder.decode<Vector<RefPtr<WebCore::SharedBuffer>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PromisedAttachmentInfo {
#if ENABLE(ATTACHMENT_ELEMENT)
            WTFMove(*attachmentIdentifier),
#endif
            WTFMove(*additionalTypes),
            WTFMove(*additionalData)
        }
    };
}

void ArgumentCoder<WebCore::RecentSearch>::encode(Encoder& encoder, const WebCore::RecentSearch& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.string)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.time)>, WallTime>);
    struct ShouldBeSameSizeAsRecentSearch : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RecentSearch>, false> {
        String string;
        WallTime time;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRecentSearch) == sizeof(WebCore::RecentSearch));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RecentSearch, string)
        , offsetof(WebCore::RecentSearch, time)
    >::value);
    encoder << instance.string;
    encoder << instance.time;
}

std::optional<WebCore::RecentSearch> ArgumentCoder<WebCore::RecentSearch>::decode(Decoder& decoder)
{
    auto string = decoder.decode<String>();
    auto time = decoder.decode<WallTime>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RecentSearch {
            WTFMove(*string),
            WTFMove(*time)
        }
    };
}

void ArgumentCoder<WebCore::EventTrackingRegions>::encode(Encoder& encoder, const WebCore::EventTrackingRegions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.asynchronousDispatchRegion)>, WebCore::Region>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.eventSpecificSynchronousDispatchRegions)>, HashMap<WebCore::EventTrackingRegions::EventType, WebCore::Region, WTF::IntHash<WebCore::EventTrackingRegions::EventType>, WTF::StrongEnumHashTraits<WebCore::EventTrackingRegions::EventType>>>);
    struct ShouldBeSameSizeAsEventTrackingRegions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::EventTrackingRegions>, false> {
        WebCore::Region asynchronousDispatchRegion;
        HashMap<WebCore::EventTrackingRegions::EventType, WebCore::Region, WTF::IntHash<WebCore::EventTrackingRegions::EventType>, WTF::StrongEnumHashTraits<WebCore::EventTrackingRegions::EventType>> eventSpecificSynchronousDispatchRegions;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEventTrackingRegions) == sizeof(WebCore::EventTrackingRegions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::EventTrackingRegions, asynchronousDispatchRegion)
        , offsetof(WebCore::EventTrackingRegions, eventSpecificSynchronousDispatchRegions)
    >::value);
    encoder << instance.asynchronousDispatchRegion;
    encoder << instance.eventSpecificSynchronousDispatchRegions;
}

std::optional<WebCore::EventTrackingRegions> ArgumentCoder<WebCore::EventTrackingRegions>::decode(Decoder& decoder)
{
    auto asynchronousDispatchRegion = decoder.decode<WebCore::Region>();
    auto eventSpecificSynchronousDispatchRegions = decoder.decode<HashMap<WebCore::EventTrackingRegions::EventType, WebCore::Region, WTF::IntHash<WebCore::EventTrackingRegions::EventType>, WTF::StrongEnumHashTraits<WebCore::EventTrackingRegions::EventType>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::EventTrackingRegions {
            WTFMove(*asynchronousDispatchRegion),
            WTFMove(*eventSpecificSynchronousDispatchRegions)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationItem>::encode(Encoder& encoder, const WebCore::TextManipulationItem& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, WebCore::FrameIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSubframe)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isCrossSiteSubframe)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::TextManipulationItemIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tokens)>, Vector<WebCore::TextManipulationToken>>);
    struct ShouldBeSameSizeAsTextManipulationItem : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationItem>, false> {
        WebCore::FrameIdentifier frameID;
        bool isSubframe;
        bool isCrossSiteSubframe;
        WebCore::TextManipulationItemIdentifier identifier;
        Vector<WebCore::TextManipulationToken> tokens;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextManipulationItem) == sizeof(WebCore::TextManipulationItem));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationItem, frameID)
        , offsetof(WebCore::TextManipulationItem, isSubframe)
        , offsetof(WebCore::TextManipulationItem, isCrossSiteSubframe)
        , offsetof(WebCore::TextManipulationItem, identifier)
        , offsetof(WebCore::TextManipulationItem, tokens)
    >::value);
    encoder << instance.frameID;
    encoder << instance.isSubframe;
    encoder << instance.isCrossSiteSubframe;
    encoder << instance.identifier;
    encoder << instance.tokens;
}

std::optional<WebCore::TextManipulationItem> ArgumentCoder<WebCore::TextManipulationItem>::decode(Decoder& decoder)
{
    auto frameID = decoder.decode<WebCore::FrameIdentifier>();
    auto isSubframe = decoder.decode<bool>();
    auto isCrossSiteSubframe = decoder.decode<bool>();
    auto identifier = decoder.decode<WebCore::TextManipulationItemIdentifier>();
    auto tokens = decoder.decode<Vector<WebCore::TextManipulationToken>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationItem {
            WTFMove(*frameID),
            WTFMove(*isSubframe),
            WTFMove(*isCrossSiteSubframe),
            WTFMove(*identifier),
            WTFMove(*tokens)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationToken>::encode(Encoder& encoder, const WebCore::TextManipulationToken& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::TextManipulationTokenIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.content)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.info)>, std::optional<WebCore::TextManipulationTokenInfo>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isExcluded)>, bool>);
    struct ShouldBeSameSizeAsTextManipulationToken : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationToken>, false> {
        WebCore::TextManipulationTokenIdentifier identifier;
        String content;
        std::optional<WebCore::TextManipulationTokenInfo> info;
        bool isExcluded;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextManipulationToken) == sizeof(WebCore::TextManipulationToken));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationToken, identifier)
        , offsetof(WebCore::TextManipulationToken, content)
        , offsetof(WebCore::TextManipulationToken, info)
        , offsetof(WebCore::TextManipulationToken, isExcluded)
    >::value);
    encoder << instance.identifier;
    encoder << instance.content;
    encoder << instance.info;
    encoder << instance.isExcluded;
}

std::optional<WebCore::TextManipulationToken> ArgumentCoder<WebCore::TextManipulationToken>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::TextManipulationTokenIdentifier>();
    auto content = decoder.decode<String>();
    auto info = decoder.decode<std::optional<WebCore::TextManipulationTokenInfo>>();
    auto isExcluded = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationToken {
            WTFMove(*identifier),
            WTFMove(*content),
            WTFMove(*info),
            WTFMove(*isExcluded)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationTokenInfo>::encode(Encoder& encoder, const WebCore::TextManipulationTokenInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tagName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.roleAttribute)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isVisible)>, bool>);
    struct ShouldBeSameSizeAsTextManipulationTokenInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationTokenInfo>, false> {
        String tagName;
        String roleAttribute;
        URL documentURL;
        bool isVisible;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextManipulationTokenInfo) == sizeof(WebCore::TextManipulationTokenInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationTokenInfo, tagName)
        , offsetof(WebCore::TextManipulationTokenInfo, roleAttribute)
        , offsetof(WebCore::TextManipulationTokenInfo, documentURL)
        , offsetof(WebCore::TextManipulationTokenInfo, isVisible)
    >::value);
    encoder << instance.tagName;
    encoder << instance.roleAttribute;
    encoder << instance.documentURL;
    encoder << instance.isVisible;
}

std::optional<WebCore::TextManipulationTokenInfo> ArgumentCoder<WebCore::TextManipulationTokenInfo>::decode(Decoder& decoder)
{
    auto tagName = decoder.decode<String>();
    auto roleAttribute = decoder.decode<String>();
    auto documentURL = decoder.decode<URL>();
    auto isVisible = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationTokenInfo {
            WTFMove(*tagName),
            WTFMove(*roleAttribute),
            WTFMove(*documentURL),
            WTFMove(*isVisible)
        }
    };
}

void ArgumentCoder<WebCore::VelocityData>::encode(Encoder& encoder, const WebCore::VelocityData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.horizontalVelocity)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.verticalVelocity)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scaleChangeRate)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastUpdateTime)>, MonotonicTime>);
    struct ShouldBeSameSizeAsVelocityData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::VelocityData>, false> {
        float horizontalVelocity;
        float verticalVelocity;
        float scaleChangeRate;
        MonotonicTime lastUpdateTime;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVelocityData) == sizeof(WebCore::VelocityData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::VelocityData, horizontalVelocity)
        , offsetof(WebCore::VelocityData, verticalVelocity)
        , offsetof(WebCore::VelocityData, scaleChangeRate)
        , offsetof(WebCore::VelocityData, lastUpdateTime)
    >::value);
    encoder << instance.horizontalVelocity;
    encoder << instance.verticalVelocity;
    encoder << instance.scaleChangeRate;
    encoder << instance.lastUpdateTime;
}

std::optional<WebCore::VelocityData> ArgumentCoder<WebCore::VelocityData>::decode(Decoder& decoder)
{
    auto horizontalVelocity = decoder.decode<float>();
    auto verticalVelocity = decoder.decode<float>();
    auto scaleChangeRate = decoder.decode<float>();
    auto lastUpdateTime = decoder.decode<MonotonicTime>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::VelocityData {
            WTFMove(*horizontalVelocity),
            WTFMove(*verticalVelocity),
            WTFMove(*scaleChangeRate),
            WTFMove(*lastUpdateTime)
        }
    };
}

void ArgumentCoder<WebCore::MimeClassInfo>::encode(Encoder& encoder, const WebCore::MimeClassInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.desc)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.extensions)>, Vector<String>>);
    struct ShouldBeSameSizeAsMimeClassInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MimeClassInfo>, false> {
        AtomString type;
        String desc;
        Vector<String> extensions;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMimeClassInfo) == sizeof(WebCore::MimeClassInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MimeClassInfo, type)
        , offsetof(WebCore::MimeClassInfo, desc)
        , offsetof(WebCore::MimeClassInfo, extensions)
    >::value);
    encoder << instance.type;
    encoder << instance.desc;
    encoder << instance.extensions;
}

std::optional<WebCore::MimeClassInfo> ArgumentCoder<WebCore::MimeClassInfo>::decode(Decoder& decoder)
{
    auto type = decoder.decode<AtomString>();
    auto desc = decoder.decode<String>();
    auto extensions = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MimeClassInfo {
            WTFMove(*type),
            WTFMove(*desc),
            WTFMove(*extensions)
        }
    };
}

void ArgumentCoder<WebCore::AuthenticationChallenge>::encode(Encoder& encoder, const WebCore::AuthenticationChallenge& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.protectionSpace())>, WebCore::ProtectionSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.proposedCredential())>, WebCore::Credential>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.previousFailureCount())>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failureResponse())>, WebCore::ResourceResponse>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.error())>, WebCore::ResourceError>);
#if USE(SOUP)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tlsPasswordFlags())>, uint32_t>);
#endif
    encoder << instance.protectionSpace();
    encoder << instance.proposedCredential();
    encoder << instance.previousFailureCount();
    encoder << instance.failureResponse();
    encoder << instance.error();
#if USE(SOUP)
    encoder << instance.tlsPasswordFlags();
#endif
}

std::optional<WebCore::AuthenticationChallenge> ArgumentCoder<WebCore::AuthenticationChallenge>::decode(Decoder& decoder)
{
    auto protectionSpace = decoder.decode<WebCore::ProtectionSpace>();
    auto proposedCredential = decoder.decode<WebCore::Credential>();
    auto previousFailureCount = decoder.decode<unsigned>();
    auto failureResponse = decoder.decode<WebCore::ResourceResponse>();
    auto error = decoder.decode<WebCore::ResourceError>();
#if USE(SOUP)
    auto tlsPasswordFlags = decoder.decode<uint32_t>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AuthenticationChallenge {
            WTFMove(*protectionSpace),
            WTFMove(*proposedCredential),
            WTFMove(*previousFailureCount),
            WTFMove(*failureResponse),
            WTFMove(*error),
#if USE(SOUP)
            WTFMove(*tlsPasswordFlags)
#endif
        }
    };
}

#if ENABLE(DRAG_SUPPORT)
void ArgumentCoder<WebCore::DragData>::encode(Encoder& encoder, const WebCore::DragData& instance)
{
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pasteboardName())>, String>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientPosition())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.globalPosition())>, WebCore::IntPoint>);
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fileNames())>, Vector<String>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.draggingSourceOperationMask())>, OptionSet<WebCore::DragOperation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.flags())>, OptionSet<WebCore::DragApplicationFlags>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dragDestinationActionMask())>, OptionSet<WebCore::DragDestinationAction>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageID())>, std::optional<WebCore::PageIdentifier>>);
#if PLATFORM(COCOA)
    encoder << instance.pasteboardName();
#endif
    encoder << instance.clientPosition();
    encoder << instance.globalPosition();
#if PLATFORM(COCOA)
    encoder << instance.fileNames();
#endif
    encoder << instance.draggingSourceOperationMask();
    encoder << instance.flags();
    encoder << instance.dragDestinationActionMask();
    encoder << instance.pageID();
}

std::optional<WebCore::DragData> ArgumentCoder<WebCore::DragData>::decode(Decoder& decoder)
{
#if PLATFORM(COCOA)
    auto pasteboardName = decoder.decode<String>();
#endif
    auto clientPosition = decoder.decode<WebCore::IntPoint>();
    auto globalPosition = decoder.decode<WebCore::IntPoint>();
#if PLATFORM(COCOA)
    auto fileNames = decoder.decode<Vector<String>>();
#endif
    auto draggingSourceOperationMask = decoder.decode<OptionSet<WebCore::DragOperation>>();
    auto flags = decoder.decode<OptionSet<WebCore::DragApplicationFlags>>();
    auto dragDestinationActionMask = decoder.decode<OptionSet<WebCore::DragDestinationAction>>();
    auto pageID = decoder.decode<std::optional<WebCore::PageIdentifier>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DragData {
#if PLATFORM(COCOA)
            WTFMove(*pasteboardName),
#endif
            WTFMove(*clientPosition),
            WTFMove(*globalPosition),
#if PLATFORM(COCOA)
            WTFMove(*fileNames),
#endif
            WTFMove(*draggingSourceOperationMask),
            WTFMove(*flags),
            WTFMove(*dragDestinationActionMask),
            WTFMove(*pageID)
        }
    };
}

#endif

void ArgumentCoder<WebCore::DictationAlternative>::encode(Encoder& encoder, const WebCore::DictationAlternative& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.range)>, WebCore::CharacterRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.context)>, WebCore::DictationContext>);
    struct ShouldBeSameSizeAsDictationAlternative : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DictationAlternative>, false> {
        WebCore::CharacterRange range;
        WebCore::DictationContext context;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDictationAlternative) == sizeof(WebCore::DictationAlternative));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DictationAlternative, range)
        , offsetof(WebCore::DictationAlternative, context)
    >::value);
    encoder << instance.range;
    encoder << instance.context;
}

std::optional<WebCore::DictationAlternative> ArgumentCoder<WebCore::DictationAlternative>::decode(Decoder& decoder)
{
    auto range = decoder.decode<WebCore::CharacterRange>();
    auto context = decoder.decode<WebCore::DictationContext>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DictationAlternative {
            WTFMove(*range),
            WTFMove(*context)
        }
    };
}

void ArgumentCoder<WebCore::UserStyleSheet>::encode(Encoder& encoder, const WebCore::UserStyleSheet& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.source())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url())>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowlist())>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blocklist())>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.injectedFrames())>, WebCore::UserContentInjectedFrames>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.level())>, WebCore::UserStyleLevel>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageID())>, std::optional<WebCore::PageIdentifier>>);
    encoder << instance.source();
    encoder << instance.url();
    encoder << instance.allowlist();
    encoder << instance.blocklist();
    encoder << instance.injectedFrames();
    encoder << instance.level();
    encoder << instance.pageID();
}

std::optional<WebCore::UserStyleSheet> ArgumentCoder<WebCore::UserStyleSheet>::decode(Decoder& decoder)
{
    auto source = decoder.decode<String>();
    auto url = decoder.decode<URL>();
    auto allowlist = decoder.decode<Vector<String>>();
    auto blocklist = decoder.decode<Vector<String>>();
    auto injectedFrames = decoder.decode<WebCore::UserContentInjectedFrames>();
    auto level = decoder.decode<WebCore::UserStyleLevel>();
    auto pageID = decoder.decode<std::optional<WebCore::PageIdentifier>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::UserStyleSheet {
            WTFMove(*source),
            WTFMove(*url),
            WTFMove(*allowlist),
            WTFMove(*blocklist),
            WTFMove(*injectedFrames),
            WTFMove(*level),
            WTFMove(*pageID)
        }
    };
}

void ArgumentCoder<WebCore::ScrollableAreaParameters>::encode(Encoder& encoder, const WebCore::ScrollableAreaParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.horizontalScrollElasticity)>, WebCore::ScrollElasticity>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.verticalScrollElasticity)>, WebCore::ScrollElasticity>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.horizontalScrollbarMode)>, WebCore::ScrollbarMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.verticalScrollbarMode)>, WebCore::ScrollbarMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.horizontalOverscrollBehavior)>, WebCore::OverscrollBehavior>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.verticalOverscrollBehavior)>, WebCore::OverscrollBehavior>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsHorizontalScrolling)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsVerticalScrolling)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.horizontalNativeScrollbarVisibility)>, WebCore::NativeScrollbarVisibility>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.verticalNativeScrollbarVisibility)>, WebCore::NativeScrollbarVisibility>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useDarkAppearanceForScrollbars)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollbarWidthStyle)>, WebCore::ScrollbarWidth>);
    struct ShouldBeSameSizeAsScrollableAreaParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ScrollableAreaParameters>, false> {
        WebCore::ScrollElasticity horizontalScrollElasticity;
        WebCore::ScrollElasticity verticalScrollElasticity;
        WebCore::ScrollbarMode horizontalScrollbarMode;
        WebCore::ScrollbarMode verticalScrollbarMode;
        WebCore::OverscrollBehavior horizontalOverscrollBehavior;
        WebCore::OverscrollBehavior verticalOverscrollBehavior;
        bool allowsHorizontalScrolling;
        bool allowsVerticalScrolling;
        WebCore::NativeScrollbarVisibility horizontalNativeScrollbarVisibility;
        WebCore::NativeScrollbarVisibility verticalNativeScrollbarVisibility;
        bool useDarkAppearanceForScrollbars;
        WebCore::ScrollbarWidth scrollbarWidthStyle;
    };
    static_assert(sizeof(ShouldBeSameSizeAsScrollableAreaParameters) == sizeof(WebCore::ScrollableAreaParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ScrollableAreaParameters, horizontalScrollElasticity)
        , offsetof(WebCore::ScrollableAreaParameters, verticalScrollElasticity)
        , offsetof(WebCore::ScrollableAreaParameters, horizontalScrollbarMode)
        , offsetof(WebCore::ScrollableAreaParameters, verticalScrollbarMode)
        , offsetof(WebCore::ScrollableAreaParameters, horizontalOverscrollBehavior)
        , offsetof(WebCore::ScrollableAreaParameters, verticalOverscrollBehavior)
        , offsetof(WebCore::ScrollableAreaParameters, allowsHorizontalScrolling)
        , offsetof(WebCore::ScrollableAreaParameters, allowsVerticalScrolling)
        , offsetof(WebCore::ScrollableAreaParameters, horizontalNativeScrollbarVisibility)
        , offsetof(WebCore::ScrollableAreaParameters, verticalNativeScrollbarVisibility)
        , offsetof(WebCore::ScrollableAreaParameters, useDarkAppearanceForScrollbars)
        , offsetof(WebCore::ScrollableAreaParameters, scrollbarWidthStyle)
    >::value);
    encoder << instance.horizontalScrollElasticity;
    encoder << instance.verticalScrollElasticity;
    encoder << instance.horizontalScrollbarMode;
    encoder << instance.verticalScrollbarMode;
    encoder << instance.horizontalOverscrollBehavior;
    encoder << instance.verticalOverscrollBehavior;
    encoder << instance.allowsHorizontalScrolling;
    encoder << instance.allowsVerticalScrolling;
    encoder << instance.horizontalNativeScrollbarVisibility;
    encoder << instance.verticalNativeScrollbarVisibility;
    encoder << instance.useDarkAppearanceForScrollbars;
    encoder << instance.scrollbarWidthStyle;
}

std::optional<WebCore::ScrollableAreaParameters> ArgumentCoder<WebCore::ScrollableAreaParameters>::decode(Decoder& decoder)
{
    auto horizontalScrollElasticity = decoder.decode<WebCore::ScrollElasticity>();
    auto verticalScrollElasticity = decoder.decode<WebCore::ScrollElasticity>();
    auto horizontalScrollbarMode = decoder.decode<WebCore::ScrollbarMode>();
    auto verticalScrollbarMode = decoder.decode<WebCore::ScrollbarMode>();
    auto horizontalOverscrollBehavior = decoder.decode<WebCore::OverscrollBehavior>();
    auto verticalOverscrollBehavior = decoder.decode<WebCore::OverscrollBehavior>();
    auto allowsHorizontalScrolling = decoder.decode<bool>();
    auto allowsVerticalScrolling = decoder.decode<bool>();
    auto horizontalNativeScrollbarVisibility = decoder.decode<WebCore::NativeScrollbarVisibility>();
    auto verticalNativeScrollbarVisibility = decoder.decode<WebCore::NativeScrollbarVisibility>();
    auto useDarkAppearanceForScrollbars = decoder.decode<bool>();
    auto scrollbarWidthStyle = decoder.decode<WebCore::ScrollbarWidth>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ScrollableAreaParameters {
            WTFMove(*horizontalScrollElasticity),
            WTFMove(*verticalScrollElasticity),
            WTFMove(*horizontalScrollbarMode),
            WTFMove(*verticalScrollbarMode),
            WTFMove(*horizontalOverscrollBehavior),
            WTFMove(*verticalOverscrollBehavior),
            WTFMove(*allowsHorizontalScrolling),
            WTFMove(*allowsVerticalScrolling),
            WTFMove(*horizontalNativeScrollbarVisibility),
            WTFMove(*verticalNativeScrollbarVisibility),
            WTFMove(*useDarkAppearanceForScrollbars),
            WTFMove(*scrollbarWidthStyle)
        }
    };
}

void ArgumentCoder<WebCore::RequestedKeyboardScrollData>::encode(Encoder& encoder, const WebCore::RequestedKeyboardScrollData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.action)>, WebCore::KeyboardScrollAction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyboardScroll)>, std::optional<WebCore::KeyboardScroll>>);
    struct ShouldBeSameSizeAsRequestedKeyboardScrollData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RequestedKeyboardScrollData>, false> {
        WebCore::KeyboardScrollAction action;
        std::optional<WebCore::KeyboardScroll> keyboardScroll;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRequestedKeyboardScrollData) == sizeof(WebCore::RequestedKeyboardScrollData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RequestedKeyboardScrollData, action)
        , offsetof(WebCore::RequestedKeyboardScrollData, keyboardScroll)
    >::value);
    encoder << instance.action;
    encoder << instance.keyboardScroll;
}

std::optional<WebCore::RequestedKeyboardScrollData> ArgumentCoder<WebCore::RequestedKeyboardScrollData>::decode(Decoder& decoder)
{
    auto action = decoder.decode<WebCore::KeyboardScrollAction>();
    auto keyboardScroll = decoder.decode<std::optional<WebCore::KeyboardScroll>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RequestedKeyboardScrollData {
            WTFMove(*action),
            WTFMove(*keyboardScroll)
        }
    };
}

void ArgumentCoder<WebCore::AbsolutePositionConstraints>::encode(Encoder& encoder, const WebCore::AbsolutePositionConstraints& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alignmentOffset())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layerPositionAtLastLayout())>, WebCore::FloatPoint>);
    encoder << instance.alignmentOffset();
    encoder << instance.layerPositionAtLastLayout();
}

std::optional<WebCore::AbsolutePositionConstraints> ArgumentCoder<WebCore::AbsolutePositionConstraints>::decode(Decoder& decoder)
{
    auto alignmentOffset = decoder.decode<WebCore::FloatSize>();
    auto layerPositionAtLastLayout = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AbsolutePositionConstraints {
            WTFMove(*alignmentOffset),
            WTFMove(*layerPositionAtLastLayout)
        }
    };
}

void ArgumentCoder<WebCore::KeyboardScroll>::encode(Encoder& encoder, const WebCore::KeyboardScroll& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumVelocity)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.force)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.granularity)>, WebCore::ScrollGranularity>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction)>, WebCore::ScrollDirection>);
    struct ShouldBeSameSizeAsKeyboardScroll : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::KeyboardScroll>, false> {
        WebCore::FloatSize offset;
        WebCore::FloatSize maximumVelocity;
        WebCore::FloatSize force;
        WebCore::ScrollGranularity granularity;
        WebCore::ScrollDirection direction;
    };
    static_assert(sizeof(ShouldBeSameSizeAsKeyboardScroll) == sizeof(WebCore::KeyboardScroll));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::KeyboardScroll, offset)
        , offsetof(WebCore::KeyboardScroll, maximumVelocity)
        , offsetof(WebCore::KeyboardScroll, force)
        , offsetof(WebCore::KeyboardScroll, granularity)
        , offsetof(WebCore::KeyboardScroll, direction)
    >::value);
    encoder << instance.offset;
    encoder << instance.maximumVelocity;
    encoder << instance.force;
    encoder << instance.granularity;
    encoder << instance.direction;
}

std::optional<WebCore::KeyboardScroll> ArgumentCoder<WebCore::KeyboardScroll>::decode(Decoder& decoder)
{
    auto offset = decoder.decode<WebCore::FloatSize>();
    auto maximumVelocity = decoder.decode<WebCore::FloatSize>();
    auto force = decoder.decode<WebCore::FloatSize>();
    auto granularity = decoder.decode<WebCore::ScrollGranularity>();
    auto direction = decoder.decode<WebCore::ScrollDirection>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::KeyboardScroll {
            WTFMove(*offset),
            WTFMove(*maximumVelocity),
            WTFMove(*force),
            WTFMove(*granularity),
            WTFMove(*direction)
        }
    };
}

void ArgumentCoder<WebCore::NotificationResources>::encode(Encoder& encoder, const WebCore::NotificationResources& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.icon())>, RefPtr<WebCore::Image>>);
    encoder << instance.icon();
}

std::optional<Ref<WebCore::NotificationResources>> ArgumentCoder<WebCore::NotificationResources>::decode(Decoder& decoder)
{
    auto icon = decoder.decode<RefPtr<WebCore::Image>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::NotificationResources::create(
            WTFMove(*icon)
        )
    };
}

void ArgumentCoder<WebCore::IdentityTransformOperation>::encode(Encoder& encoder, const WebCore::IdentityTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::TransformOperation::Type>);
}

std::optional<Ref<WebCore::IdentityTransformOperation>> ArgumentCoder<WebCore::IdentityTransformOperation>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IdentityTransformOperation::create(
        )
    };
}

void ArgumentCoder<WebCore::TranslateTransformOperation>::encode(Encoder& encoder, const WebCore::TranslateTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.z())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::TransformOperation::Type>);
    encoder << instance.x();
    encoder << instance.y();
    encoder << instance.z();
    encoder << instance.type();
}

std::optional<Ref<WebCore::TranslateTransformOperation>> ArgumentCoder<WebCore::TranslateTransformOperation>::decode(Decoder& decoder)
{
    auto x = decoder.decode<WebCore::Length>();
    auto y = decoder.decode<WebCore::Length>();
    auto z = decoder.decode<WebCore::Length>();
    auto type = decoder.decode<WebCore::TransformOperation::Type>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(WebCore::TransformOperation::isTranslateTransformOperationType(*type)))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TranslateTransformOperation::create(
            WTFMove(*x),
            WTFMove(*y),
            WTFMove(*z),
            WTFMove(*type)
        )
    };
}

void ArgumentCoder<WebCore::RotateTransformOperation>::encode(Encoder& encoder, const WebCore::RotateTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.z())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.angle())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::TransformOperation::Type>);
    encoder << instance.x();
    encoder << instance.y();
    encoder << instance.z();
    encoder << instance.angle();
    encoder << instance.type();
}

std::optional<Ref<WebCore::RotateTransformOperation>> ArgumentCoder<WebCore::RotateTransformOperation>::decode(Decoder& decoder)
{
    auto x = decoder.decode<double>();
    auto y = decoder.decode<double>();
    auto z = decoder.decode<double>();
    auto angle = decoder.decode<double>();
    auto type = decoder.decode<WebCore::TransformOperation::Type>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(WebCore::TransformOperation::isRotateTransformOperationType(*type)))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RotateTransformOperation::create(
            WTFMove(*x),
            WTFMove(*y),
            WTFMove(*z),
            WTFMove(*angle),
            WTFMove(*type)
        )
    };
}

void ArgumentCoder<WebCore::ScaleTransformOperation>::encode(Encoder& encoder, const WebCore::ScaleTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.z())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::TransformOperation::Type>);
    encoder << instance.x();
    encoder << instance.y();
    encoder << instance.z();
    encoder << instance.type();
}

std::optional<Ref<WebCore::ScaleTransformOperation>> ArgumentCoder<WebCore::ScaleTransformOperation>::decode(Decoder& decoder)
{
    auto x = decoder.decode<double>();
    auto y = decoder.decode<double>();
    auto z = decoder.decode<double>();
    auto type = decoder.decode<WebCore::TransformOperation::Type>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(WebCore::TransformOperation::isScaleTransformOperationType(*type)))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ScaleTransformOperation::create(
            WTFMove(*x),
            WTFMove(*y),
            WTFMove(*z),
            WTFMove(*type)
        )
    };
}

void ArgumentCoder<WebCore::SkewTransformOperation>::encode(Encoder& encoder, const WebCore::SkewTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.angleX())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.angleY())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::TransformOperation::Type>);
    encoder << instance.angleX();
    encoder << instance.angleY();
    encoder << instance.type();
}

std::optional<Ref<WebCore::SkewTransformOperation>> ArgumentCoder<WebCore::SkewTransformOperation>::decode(Decoder& decoder)
{
    auto angleX = decoder.decode<double>();
    auto angleY = decoder.decode<double>();
    auto type = decoder.decode<WebCore::TransformOperation::Type>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(WebCore::TransformOperation::isSkewTransformOperationType(*type)))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SkewTransformOperation::create(
            WTFMove(*angleX),
            WTFMove(*angleY),
            WTFMove(*type)
        )
    };
}

void ArgumentCoder<WebCore::PerspectiveTransformOperation>::encode(Encoder& encoder, const WebCore::PerspectiveTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.perspective())>, std::optional<WebCore::Length>>);
    encoder << instance.perspective();
}

std::optional<Ref<WebCore::PerspectiveTransformOperation>> ArgumentCoder<WebCore::PerspectiveTransformOperation>::decode(Decoder& decoder)
{
    auto perspective = decoder.decode<std::optional<WebCore::Length>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PerspectiveTransformOperation::create(
            WTFMove(*perspective)
        )
    };
}

void ArgumentCoder<WebCore::MatrixTransformOperation>::encode(Encoder& encoder, const WebCore::MatrixTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.matrix())>, WebCore::TransformationMatrix>);
    encoder << instance.matrix();
}

std::optional<Ref<WebCore::MatrixTransformOperation>> ArgumentCoder<WebCore::MatrixTransformOperation>::decode(Decoder& decoder)
{
    auto matrix = decoder.decode<WebCore::TransformationMatrix>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MatrixTransformOperation::create(
            WTFMove(*matrix)
        )
    };
}

void ArgumentCoder<WebCore::Matrix3DTransformOperation>::encode(Encoder& encoder, const WebCore::Matrix3DTransformOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.matrix())>, WebCore::TransformationMatrix>);
    encoder << instance.matrix();
}

std::optional<Ref<WebCore::Matrix3DTransformOperation>> ArgumentCoder<WebCore::Matrix3DTransformOperation>::decode(Decoder& decoder)
{
    auto matrix = decoder.decode<WebCore::TransformationMatrix>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Matrix3DTransformOperation::create(
            WTFMove(*matrix)
        )
    };
}

void ArgumentCoder<WebCore::TransformOperations>::encode(Encoder& encoder, const WebCore::TransformOperations& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.operations())>, Vector<RefPtr<WebCore::TransformOperation>>>);
    encoder << instance.operations();
}

std::optional<WebCore::TransformOperations> ArgumentCoder<WebCore::TransformOperations>::decode(Decoder& decoder)
{
    auto operations = decoder.decode<Vector<RefPtr<WebCore::TransformOperation>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TransformOperations {
            WTFMove(*operations)
        }
    };
}

void ArgumentCoder<WebCore::Gradient::LinearData>::encode(Encoder& encoder, const WebCore::Gradient::LinearData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point0)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point1)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsLinearData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Gradient::LinearData>, false> {
        WebCore::FloatPoint point0;
        WebCore::FloatPoint point1;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLinearData) == sizeof(WebCore::Gradient::LinearData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Gradient::LinearData, point0)
        , offsetof(WebCore::Gradient::LinearData, point1)
    >::value);
    encoder << instance.point0;
    encoder << instance.point1;
}

void ArgumentCoder<WebCore::Gradient::LinearData>::encode(StreamConnectionEncoder& encoder, const WebCore::Gradient::LinearData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point0)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point1)>, WebCore::FloatPoint>);
    struct ShouldBeSameSizeAsLinearData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Gradient::LinearData>, false> {
        WebCore::FloatPoint point0;
        WebCore::FloatPoint point1;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLinearData) == sizeof(WebCore::Gradient::LinearData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Gradient::LinearData, point0)
        , offsetof(WebCore::Gradient::LinearData, point1)
    >::value);
    encoder << instance.point0;
    encoder << instance.point1;
}

std::optional<WebCore::Gradient::LinearData> ArgumentCoder<WebCore::Gradient::LinearData>::decode(Decoder& decoder)
{
    auto point0 = decoder.decode<WebCore::FloatPoint>();
    auto point1 = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Gradient::LinearData {
            WTFMove(*point0),
            WTFMove(*point1)
        }
    };
}

void ArgumentCoder<WebCore::Gradient::RadialData>::encode(Encoder& encoder, const WebCore::Gradient::RadialData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point0)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startRadius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endRadius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspectRatio)>, float>);
    struct ShouldBeSameSizeAsRadialData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Gradient::RadialData>, false> {
        WebCore::FloatPoint point0;
        WebCore::FloatPoint point1;
        float startRadius;
        float endRadius;
        float aspectRatio;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRadialData) == sizeof(WebCore::Gradient::RadialData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Gradient::RadialData, point0)
        , offsetof(WebCore::Gradient::RadialData, point1)
        , offsetof(WebCore::Gradient::RadialData, startRadius)
        , offsetof(WebCore::Gradient::RadialData, endRadius)
        , offsetof(WebCore::Gradient::RadialData, aspectRatio)
    >::value);
    encoder << instance.point0;
    encoder << instance.point1;
    encoder << instance.startRadius;
    encoder << instance.endRadius;
    encoder << instance.aspectRatio;
}

void ArgumentCoder<WebCore::Gradient::RadialData>::encode(StreamConnectionEncoder& encoder, const WebCore::Gradient::RadialData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point0)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point1)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startRadius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endRadius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspectRatio)>, float>);
    struct ShouldBeSameSizeAsRadialData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Gradient::RadialData>, false> {
        WebCore::FloatPoint point0;
        WebCore::FloatPoint point1;
        float startRadius;
        float endRadius;
        float aspectRatio;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRadialData) == sizeof(WebCore::Gradient::RadialData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Gradient::RadialData, point0)
        , offsetof(WebCore::Gradient::RadialData, point1)
        , offsetof(WebCore::Gradient::RadialData, startRadius)
        , offsetof(WebCore::Gradient::RadialData, endRadius)
        , offsetof(WebCore::Gradient::RadialData, aspectRatio)
    >::value);
    encoder << instance.point0;
    encoder << instance.point1;
    encoder << instance.startRadius;
    encoder << instance.endRadius;
    encoder << instance.aspectRatio;
}

std::optional<WebCore::Gradient::RadialData> ArgumentCoder<WebCore::Gradient::RadialData>::decode(Decoder& decoder)
{
    auto point0 = decoder.decode<WebCore::FloatPoint>();
    auto point1 = decoder.decode<WebCore::FloatPoint>();
    auto startRadius = decoder.decode<float>();
    auto endRadius = decoder.decode<float>();
    auto aspectRatio = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Gradient::RadialData {
            WTFMove(*point0),
            WTFMove(*point1),
            WTFMove(*startRadius),
            WTFMove(*endRadius),
            WTFMove(*aspectRatio)
        }
    };
}

void ArgumentCoder<WebCore::Gradient::ConicData>::encode(Encoder& encoder, const WebCore::Gradient::ConicData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point0)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.angleRadians)>, float>);
    struct ShouldBeSameSizeAsConicData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Gradient::ConicData>, false> {
        WebCore::FloatPoint point0;
        float angleRadians;
    };
    static_assert(sizeof(ShouldBeSameSizeAsConicData) == sizeof(WebCore::Gradient::ConicData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Gradient::ConicData, point0)
        , offsetof(WebCore::Gradient::ConicData, angleRadians)
    >::value);
    encoder << instance.point0;
    encoder << instance.angleRadians;
}

void ArgumentCoder<WebCore::Gradient::ConicData>::encode(StreamConnectionEncoder& encoder, const WebCore::Gradient::ConicData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point0)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.angleRadians)>, float>);
    struct ShouldBeSameSizeAsConicData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Gradient::ConicData>, false> {
        WebCore::FloatPoint point0;
        float angleRadians;
    };
    static_assert(sizeof(ShouldBeSameSizeAsConicData) == sizeof(WebCore::Gradient::ConicData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Gradient::ConicData, point0)
        , offsetof(WebCore::Gradient::ConicData, angleRadians)
    >::value);
    encoder << instance.point0;
    encoder << instance.angleRadians;
}

std::optional<WebCore::Gradient::ConicData> ArgumentCoder<WebCore::Gradient::ConicData>::decode(Decoder& decoder)
{
    auto point0 = decoder.decode<WebCore::FloatPoint>();
    auto angleRadians = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Gradient::ConicData {
            WTFMove(*point0),
            WTFMove(*angleRadians)
        }
    };
}

void ArgumentCoder<WebCore::Gradient>::encode(Encoder& encoder, const WebCore::Gradient& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, std::variant<WebCore::Gradient::LinearData, WebCore::Gradient::RadialData, WebCore::Gradient::ConicData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorInterpolationMethod())>, WebCore::ColorInterpolationMethod>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.spreadMethod())>, WebCore::GradientSpreadMethod>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stops())>, WebCore::GradientColorStops>);
    encoder << instance.data();
    encoder << instance.colorInterpolationMethod();
    encoder << instance.spreadMethod();
    encoder << instance.stops();
}

void ArgumentCoder<WebCore::Gradient>::encode(StreamConnectionEncoder& encoder, const WebCore::Gradient& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, std::variant<WebCore::Gradient::LinearData, WebCore::Gradient::RadialData, WebCore::Gradient::ConicData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorInterpolationMethod())>, WebCore::ColorInterpolationMethod>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.spreadMethod())>, WebCore::GradientSpreadMethod>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stops())>, WebCore::GradientColorStops>);
    encoder << instance.data();
    encoder << instance.colorInterpolationMethod();
    encoder << instance.spreadMethod();
    encoder << instance.stops();
}

std::optional<Ref<WebCore::Gradient>> ArgumentCoder<WebCore::Gradient>::decode(Decoder& decoder)
{
    auto data = decoder.decode<std::variant<WebCore::Gradient::LinearData, WebCore::Gradient::RadialData, WebCore::Gradient::ConicData>>();
    auto colorInterpolationMethod = decoder.decode<WebCore::ColorInterpolationMethod>();
    auto spreadMethod = decoder.decode<WebCore::GradientSpreadMethod>();
    auto stops = decoder.decode<WebCore::GradientColorStops>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Gradient::create(
            WTFMove(*data),
            WTFMove(*colorInterpolationMethod),
            WTFMove(*spreadMethod),
            WTFMove(*stops)
        )
    };
}

void ArgumentCoder<WebCore::Pattern::Parameters>::encode(Encoder& encoder, const WebCore::Pattern::Parameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.repeatX)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.repeatY)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.patternSpaceTransform)>, WebCore::AffineTransform>);
    struct ShouldBeSameSizeAsParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Pattern::Parameters>, false> {
        bool repeatX;
        bool repeatY;
        WebCore::AffineTransform patternSpaceTransform;
    };
    static_assert(sizeof(ShouldBeSameSizeAsParameters) == sizeof(WebCore::Pattern::Parameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Pattern::Parameters, repeatX)
        , offsetof(WebCore::Pattern::Parameters, repeatY)
        , offsetof(WebCore::Pattern::Parameters, patternSpaceTransform)
    >::value);
    encoder << instance.repeatX;
    encoder << instance.repeatY;
    encoder << instance.patternSpaceTransform;
}

void ArgumentCoder<WebCore::Pattern::Parameters>::encode(StreamConnectionEncoder& encoder, const WebCore::Pattern::Parameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.repeatX)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.repeatY)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.patternSpaceTransform)>, WebCore::AffineTransform>);
    struct ShouldBeSameSizeAsParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Pattern::Parameters>, false> {
        bool repeatX;
        bool repeatY;
        WebCore::AffineTransform patternSpaceTransform;
    };
    static_assert(sizeof(ShouldBeSameSizeAsParameters) == sizeof(WebCore::Pattern::Parameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Pattern::Parameters, repeatX)
        , offsetof(WebCore::Pattern::Parameters, repeatY)
        , offsetof(WebCore::Pattern::Parameters, patternSpaceTransform)
    >::value);
    encoder << instance.repeatX;
    encoder << instance.repeatY;
    encoder << instance.patternSpaceTransform;
}

std::optional<WebCore::Pattern::Parameters> ArgumentCoder<WebCore::Pattern::Parameters>::decode(Decoder& decoder)
{
    auto repeatX = decoder.decode<bool>();
    auto repeatY = decoder.decode<bool>();
    auto patternSpaceTransform = decoder.decode<WebCore::AffineTransform>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Pattern::Parameters {
            WTFMove(*repeatX),
            WTFMove(*repeatY),
            WTFMove(*patternSpaceTransform)
        }
    };
}

void ArgumentCoder<WebCore::Pattern>::encode(Encoder& encoder, const WebCore::Pattern& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tileImage())>, WebCore::SourceImage>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parameters())>, WebCore::Pattern::Parameters>);
    encoder << instance.tileImage();
    encoder << instance.parameters();
}

void ArgumentCoder<WebCore::Pattern>::encode(StreamConnectionEncoder& encoder, const WebCore::Pattern& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tileImage())>, WebCore::SourceImage>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parameters())>, WebCore::Pattern::Parameters>);
    encoder << instance.tileImage();
    encoder << instance.parameters();
}

std::optional<Ref<WebCore::Pattern>> ArgumentCoder<WebCore::Pattern>::decode(Decoder& decoder)
{
    auto tileImage = decoder.decode<WebCore::SourceImage>();
    auto parameters = decoder.decode<WebCore::Pattern::Parameters>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Pattern::create(
            WTFMove(*tileImage),
            WTFMove(*parameters)
        )
    };
}

void ArgumentCoder<WebCore::SourceBrush::Brush>::encode(Encoder& encoder, const WebCore::SourceBrush::Brush& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.brush)>, std::variant<WebCore::SourceBrush::Brush::LogicalGradient, Ref<WebCore::Pattern>>>);
    struct ShouldBeSameSizeAsBrush : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SourceBrush::Brush>, false> {
        std::variant<WebCore::SourceBrush::Brush::LogicalGradient, Ref<WebCore::Pattern>> brush;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBrush) == sizeof(WebCore::SourceBrush::Brush));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SourceBrush::Brush, brush)
    >::value);
    encoder << instance.brush;
}

void ArgumentCoder<WebCore::SourceBrush::Brush>::encode(StreamConnectionEncoder& encoder, const WebCore::SourceBrush::Brush& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.brush)>, std::variant<WebCore::SourceBrush::Brush::LogicalGradient, Ref<WebCore::Pattern>>>);
    struct ShouldBeSameSizeAsBrush : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SourceBrush::Brush>, false> {
        std::variant<WebCore::SourceBrush::Brush::LogicalGradient, Ref<WebCore::Pattern>> brush;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBrush) == sizeof(WebCore::SourceBrush::Brush));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SourceBrush::Brush, brush)
    >::value);
    encoder << instance.brush;
}

std::optional<WebCore::SourceBrush::Brush> ArgumentCoder<WebCore::SourceBrush::Brush>::decode(Decoder& decoder)
{
    auto brush = decoder.decode<std::variant<WebCore::SourceBrush::Brush::LogicalGradient, Ref<WebCore::Pattern>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SourceBrush::Brush {
            WTFMove(*brush)
        }
    };
}

void ArgumentCoder<WebCore::SourceBrush>::encode(Encoder& encoder, const WebCore::SourceBrush& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.brush())>, std::optional<WebCore::SourceBrush::Brush>>);
    encoder << instance.color();
    encoder << instance.brush();
}

void ArgumentCoder<WebCore::SourceBrush>::encode(StreamConnectionEncoder& encoder, const WebCore::SourceBrush& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.brush())>, std::optional<WebCore::SourceBrush::Brush>>);
    encoder << instance.color();
    encoder << instance.brush();
}

std::optional<WebCore::SourceBrush> ArgumentCoder<WebCore::SourceBrush>::decode(Decoder& decoder)
{
    auto color = decoder.decode<WebCore::Color>();
    auto brush = decoder.decode<std::optional<WebCore::SourceBrush::Brush>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SourceBrush {
            WTFMove(*color),
            WTFMove(*brush)
        }
    };
}

#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
void ArgumentCoder<WebCore::GraphicsContextGLAttributes>::encode(Encoder& encoder, const WebCore::GraphicsContextGLAttributes& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alpha)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depth)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencil)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.antialias)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.premultipliedAlpha)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preserveDrawingBuffer)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failIfMajorPerformanceCaveat)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.powerPreference)>, WebCore::GraphicsContextGLPowerPreference>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shareResources)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.noExtensions)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.devicePixelRatio)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.initialPowerPreference)>, WebCore::GraphicsContextGLPowerPreference>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webGLVersion)>, WebCore::GraphicsContextGLWebGLVersion>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.forceRequestForHighPerformanceGPU)>, bool>);
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.windowGPUID)>, uint64_t>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useMetal)>, bool>);
#endif
#if ENABLE(WEBXR)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.xrCompatible)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failPlatformContextCreationForTesting)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.remoteIPCBufferSizeLog2ForTesting)>, unsigned>);
    struct ShouldBeSameSizeAsGraphicsContextGLAttributes : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GraphicsContextGLAttributes>, false> {
        bool alpha;
        bool depth;
        bool stencil;
        bool antialias;
        bool premultipliedAlpha;
        bool preserveDrawingBuffer;
        bool failIfMajorPerformanceCaveat;
        WebCore::GraphicsContextGLPowerPreference powerPreference;
        bool shareResources;
        bool noExtensions;
        float devicePixelRatio;
        WebCore::GraphicsContextGLPowerPreference initialPowerPreference;
        WebCore::GraphicsContextGLWebGLVersion webGLVersion;
        bool forceRequestForHighPerformanceGPU;
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
        uint64_t windowGPUID;
#endif
#if PLATFORM(COCOA)
        bool useMetal;
#endif
#if ENABLE(WEBXR)
        bool xrCompatible;
#endif
        bool failPlatformContextCreationForTesting;
        unsigned remoteIPCBufferSizeLog2ForTesting;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGraphicsContextGLAttributes) == sizeof(WebCore::GraphicsContextGLAttributes));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GraphicsContextGLAttributes, alpha)
        , offsetof(WebCore::GraphicsContextGLAttributes, depth)
        , offsetof(WebCore::GraphicsContextGLAttributes, stencil)
        , offsetof(WebCore::GraphicsContextGLAttributes, antialias)
        , offsetof(WebCore::GraphicsContextGLAttributes, premultipliedAlpha)
        , offsetof(WebCore::GraphicsContextGLAttributes, preserveDrawingBuffer)
        , offsetof(WebCore::GraphicsContextGLAttributes, failIfMajorPerformanceCaveat)
        , offsetof(WebCore::GraphicsContextGLAttributes, powerPreference)
        , offsetof(WebCore::GraphicsContextGLAttributes, shareResources)
        , offsetof(WebCore::GraphicsContextGLAttributes, noExtensions)
        , offsetof(WebCore::GraphicsContextGLAttributes, devicePixelRatio)
        , offsetof(WebCore::GraphicsContextGLAttributes, initialPowerPreference)
        , offsetof(WebCore::GraphicsContextGLAttributes, webGLVersion)
        , offsetof(WebCore::GraphicsContextGLAttributes, forceRequestForHighPerformanceGPU)
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
        , offsetof(WebCore::GraphicsContextGLAttributes, windowGPUID)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebCore::GraphicsContextGLAttributes, useMetal)
#endif
#if ENABLE(WEBXR)
        , offsetof(WebCore::GraphicsContextGLAttributes, xrCompatible)
#endif
        , offsetof(WebCore::GraphicsContextGLAttributes, failPlatformContextCreationForTesting)
        , offsetof(WebCore::GraphicsContextGLAttributes, remoteIPCBufferSizeLog2ForTesting)
    >::value);
    encoder << instance.alpha;
    encoder << instance.depth;
    encoder << instance.stencil;
    encoder << instance.antialias;
    encoder << instance.premultipliedAlpha;
    encoder << instance.preserveDrawingBuffer;
    encoder << instance.failIfMajorPerformanceCaveat;
    encoder << instance.powerPreference;
    encoder << instance.shareResources;
    encoder << instance.noExtensions;
    encoder << instance.devicePixelRatio;
    encoder << instance.initialPowerPreference;
    encoder << instance.webGLVersion;
    encoder << instance.forceRequestForHighPerformanceGPU;
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    encoder << instance.windowGPUID;
#endif
#if PLATFORM(COCOA)
    encoder << instance.useMetal;
#endif
#if ENABLE(WEBXR)
    encoder << instance.xrCompatible;
#endif
    encoder << instance.failPlatformContextCreationForTesting;
}

std::optional<WebCore::GraphicsContextGLAttributes> ArgumentCoder<WebCore::GraphicsContextGLAttributes>::decode(Decoder& decoder)
{
    auto alpha = decoder.decode<bool>();
    auto depth = decoder.decode<bool>();
    auto stencil = decoder.decode<bool>();
    auto antialias = decoder.decode<bool>();
    auto premultipliedAlpha = decoder.decode<bool>();
    auto preserveDrawingBuffer = decoder.decode<bool>();
    auto failIfMajorPerformanceCaveat = decoder.decode<bool>();
    auto powerPreference = decoder.decode<WebCore::GraphicsContextGLPowerPreference>();
    auto shareResources = decoder.decode<bool>();
    auto noExtensions = decoder.decode<bool>();
    auto devicePixelRatio = decoder.decode<float>();
    auto initialPowerPreference = decoder.decode<WebCore::GraphicsContextGLPowerPreference>();
    auto webGLVersion = decoder.decode<WebCore::GraphicsContextGLWebGLVersion>();
    auto forceRequestForHighPerformanceGPU = decoder.decode<bool>();
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    auto windowGPUID = decoder.decode<uint64_t>();
#endif
#if PLATFORM(COCOA)
    auto useMetal = decoder.decode<bool>();
#endif
#if ENABLE(WEBXR)
    auto xrCompatible = decoder.decode<bool>();
#endif
    auto failPlatformContextCreationForTesting = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GraphicsContextGLAttributes {
            WTFMove(*alpha),
            WTFMove(*depth),
            WTFMove(*stencil),
            WTFMove(*antialias),
            WTFMove(*premultipliedAlpha),
            WTFMove(*preserveDrawingBuffer),
            WTFMove(*failIfMajorPerformanceCaveat),
            WTFMove(*powerPreference),
            WTFMove(*shareResources),
            WTFMove(*noExtensions),
            WTFMove(*devicePixelRatio),
            WTFMove(*initialPowerPreference),
            WTFMove(*webGLVersion),
            WTFMove(*forceRequestForHighPerformanceGPU),
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
            WTFMove(*windowGPUID),
#endif
#if PLATFORM(COCOA)
            WTFMove(*useMetal),
#endif
#if ENABLE(WEBXR)
            WTFMove(*xrCompatible),
#endif
            WTFMove(*failPlatformContextCreationForTesting)
        }
    };
}

#endif

enum class WebCore_TimingFunction_Subclass : IPC::EncodedVariantIndex {
    LinearTimingFunction,
    CubicBezierTimingFunction,
    StepsTimingFunction,
    SpringTimingFunction
};

void ArgumentCoder<WebCore::TimingFunction>::encode(Encoder& encoder, const WebCore::TimingFunction& instance)
{
    if (auto* subclass = dynamicDowncast<WebCore::LinearTimingFunction>(instance)) {
        encoder << WebCore_TimingFunction_Subclass::LinearTimingFunction;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::CubicBezierTimingFunction>(instance)) {
        encoder << WebCore_TimingFunction_Subclass::CubicBezierTimingFunction;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::StepsTimingFunction>(instance)) {
        encoder << WebCore_TimingFunction_Subclass::StepsTimingFunction;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::SpringTimingFunction>(instance)) {
        encoder << WebCore_TimingFunction_Subclass::SpringTimingFunction;
        encoder << *subclass;
        return;
    }
    ASSERT_NOT_REACHED();
}

std::optional<Ref<WebCore::TimingFunction>> ArgumentCoder<WebCore::TimingFunction>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore_TimingFunction_Subclass>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (type == WebCore_TimingFunction_Subclass::LinearTimingFunction) {
        auto result = decoder.decode<Ref<WebCore::LinearTimingFunction>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TimingFunction_Subclass::CubicBezierTimingFunction) {
        auto result = decoder.decode<Ref<WebCore::CubicBezierTimingFunction>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TimingFunction_Subclass::StepsTimingFunction) {
        auto result = decoder.decode<Ref<WebCore::StepsTimingFunction>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TimingFunction_Subclass::SpringTimingFunction) {
        auto result = decoder.decode<Ref<WebCore::SpringTimingFunction>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    ASSERT_NOT_REACHED();
    return std::nullopt;
}

enum class WebCore_TransformOperation_Subclass : IPC::EncodedVariantIndex {
    TranslateTransformOperation,
    ScaleTransformOperation,
    RotateTransformOperation,
    SkewTransformOperation,
    MatrixTransformOperation,
    Matrix3DTransformOperation,
    PerspectiveTransformOperation,
    IdentityTransformOperation
};

void ArgumentCoder<WebCore::TransformOperation>::encode(Encoder& encoder, const WebCore::TransformOperation& instance)
{
    if (auto* subclass = dynamicDowncast<WebCore::TranslateTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::TranslateTransformOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::ScaleTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::ScaleTransformOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::RotateTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::RotateTransformOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::SkewTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::SkewTransformOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::MatrixTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::MatrixTransformOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::Matrix3DTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::Matrix3DTransformOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::PerspectiveTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::PerspectiveTransformOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::IdentityTransformOperation>(instance)) {
        encoder << WebCore_TransformOperation_Subclass::IdentityTransformOperation;
        encoder << *subclass;
        return;
    }
    ASSERT_NOT_REACHED();
}

std::optional<Ref<WebCore::TransformOperation>> ArgumentCoder<WebCore::TransformOperation>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore_TransformOperation_Subclass>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (type == WebCore_TransformOperation_Subclass::TranslateTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::TranslateTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TransformOperation_Subclass::ScaleTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::ScaleTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TransformOperation_Subclass::RotateTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::RotateTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TransformOperation_Subclass::SkewTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::SkewTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TransformOperation_Subclass::MatrixTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::MatrixTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TransformOperation_Subclass::Matrix3DTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::Matrix3DTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TransformOperation_Subclass::PerspectiveTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::PerspectiveTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_TransformOperation_Subclass::IdentityTransformOperation) {
        auto result = decoder.decode<Ref<WebCore::IdentityTransformOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    ASSERT_NOT_REACHED();
    return std::nullopt;
}

void ArgumentCoder<WebCore::LengthSize>::encode(Encoder& encoder, const WebCore::LengthSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, WebCore::Length>);
    struct ShouldBeSameSizeAsLengthSize : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::LengthSize>, false> {
        WebCore::Length width;
        WebCore::Length height;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLengthSize) == sizeof(WebCore::LengthSize));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::LengthSize, width)
        , offsetof(WebCore::LengthSize, height)
    >::value);
    encoder << instance.width;
    encoder << instance.height;
}

std::optional<WebCore::LengthSize> ArgumentCoder<WebCore::LengthSize>::decode(Decoder& decoder)
{
    auto width = decoder.decode<WebCore::Length>();
    auto height = decoder.decode<WebCore::Length>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LengthSize {
            WTFMove(*width),
            WTFMove(*height)
        }
    };
}

void ArgumentCoder<WebCore::SVGPathByteStream>::encode(Encoder& encoder, const WebCore::SVGPathByteStream& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, Vector<unsigned char>>);
    encoder << instance.data();
}

std::optional<WebCore::SVGPathByteStream> ArgumentCoder<WebCore::SVGPathByteStream>::decode(Decoder& decoder)
{
    auto data = decoder.decode<Vector<unsigned char>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SVGPathByteStream {
            WTFMove(*data)
        }
    };
}

void ArgumentCoder<WebCore::BasicShapeCenterCoordinate>::encode(Encoder& encoder, const WebCore::BasicShapeCenterCoordinate& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction())>, WebCore::BasicShapeCenterCoordinate::Direction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.length())>, WebCore::Length>);
    encoder << instance.direction();
    encoder << instance.length();
}

std::optional<WebCore::BasicShapeCenterCoordinate> ArgumentCoder<WebCore::BasicShapeCenterCoordinate>::decode(Decoder& decoder)
{
    auto direction = decoder.decode<WebCore::BasicShapeCenterCoordinate::Direction>();
    auto length = decoder.decode<WebCore::Length>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BasicShapeCenterCoordinate {
            WTFMove(*direction),
            WTFMove(*length)
        }
    };
}

void ArgumentCoder<WebCore::BasicShapeRadius>::encode(Encoder& encoder, const WebCore::BasicShapeRadius& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::BasicShapeRadius::Type>);
    encoder << instance.value();
    encoder << instance.type();
}

std::optional<WebCore::BasicShapeRadius> ArgumentCoder<WebCore::BasicShapeRadius>::decode(Decoder& decoder)
{
    auto value = decoder.decode<WebCore::Length>();
    auto type = decoder.decode<WebCore::BasicShapeRadius::Type>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BasicShapeRadius {
            WTFMove(*value),
            WTFMove(*type)
        }
    };
}

void ArgumentCoder<WebCore::BasicShapeCircle>::encode(Encoder& encoder, const WebCore::BasicShapeCircle& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.centerX())>, WebCore::BasicShapeCenterCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.centerY())>, WebCore::BasicShapeCenterCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius())>, WebCore::BasicShapeRadius>);
    encoder << instance.centerX();
    encoder << instance.centerY();
    encoder << instance.radius();
}

std::optional<Ref<WebCore::BasicShapeCircle>> ArgumentCoder<WebCore::BasicShapeCircle>::decode(Decoder& decoder)
{
    auto centerX = decoder.decode<WebCore::BasicShapeCenterCoordinate>();
    auto centerY = decoder.decode<WebCore::BasicShapeCenterCoordinate>();
    auto radius = decoder.decode<WebCore::BasicShapeRadius>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BasicShapeCircle::create(
            WTFMove(*centerX),
            WTFMove(*centerY),
            WTFMove(*radius)
        )
    };
}

void ArgumentCoder<WebCore::BasicShapeEllipse>::encode(Encoder& encoder, const WebCore::BasicShapeEllipse& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.centerX())>, WebCore::BasicShapeCenterCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.centerY())>, WebCore::BasicShapeCenterCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusX())>, WebCore::BasicShapeRadius>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusY())>, WebCore::BasicShapeRadius>);
    encoder << instance.centerX();
    encoder << instance.centerY();
    encoder << instance.radiusX();
    encoder << instance.radiusY();
}

std::optional<Ref<WebCore::BasicShapeEllipse>> ArgumentCoder<WebCore::BasicShapeEllipse>::decode(Decoder& decoder)
{
    auto centerX = decoder.decode<WebCore::BasicShapeCenterCoordinate>();
    auto centerY = decoder.decode<WebCore::BasicShapeCenterCoordinate>();
    auto radiusX = decoder.decode<WebCore::BasicShapeRadius>();
    auto radiusY = decoder.decode<WebCore::BasicShapeRadius>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BasicShapeEllipse::create(
            WTFMove(*centerX),
            WTFMove(*centerY),
            WTFMove(*radiusX),
            WTFMove(*radiusY)
        )
    };
}

void ArgumentCoder<WebCore::BasicShapePolygon>::encode(Encoder& encoder, const WebCore::BasicShapePolygon& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.windRule())>, WebCore::WindRule>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.values())>, Vector<WebCore::Length>>);
    encoder << instance.windRule();
    encoder << instance.values();
}

std::optional<Ref<WebCore::BasicShapePolygon>> ArgumentCoder<WebCore::BasicShapePolygon>::decode(Decoder& decoder)
{
    auto windRule = decoder.decode<WebCore::WindRule>();
    auto values = decoder.decode<Vector<WebCore::Length>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BasicShapePolygon::create(
            WTFMove(*windRule),
            WTFMove(*values)
        )
    };
}

void ArgumentCoder<WebCore::BasicShapePath>::encode(Encoder& encoder, const WebCore::BasicShapePath& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.byteStream())>, std::unique_ptr<WebCore::SVGPathByteStream>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.zoom())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.windRule())>, WebCore::WindRule>);
    encoder << instance.byteStream();
    encoder << instance.zoom();
    encoder << instance.windRule();
}

std::optional<Ref<WebCore::BasicShapePath>> ArgumentCoder<WebCore::BasicShapePath>::decode(Decoder& decoder)
{
    auto byteStream = decoder.decode<std::unique_ptr<WebCore::SVGPathByteStream>>();
    auto zoom = decoder.decode<float>();
    auto windRule = decoder.decode<WebCore::WindRule>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BasicShapePath::create(
            WTFMove(*byteStream),
            WTFMove(*zoom),
            WTFMove(*windRule)
        )
    };
}

void ArgumentCoder<WebCore::BasicShapeInset>::encode(Encoder& encoder, const WebCore::BasicShapeInset& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.right())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.top())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bottom())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.left())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topLeftRadius())>, WebCore::LengthSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topRightRadius())>, WebCore::LengthSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bottomRightRadius())>, WebCore::LengthSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bottomLeftRadius())>, WebCore::LengthSize>);
    encoder << instance.right();
    encoder << instance.top();
    encoder << instance.bottom();
    encoder << instance.left();
    encoder << instance.topLeftRadius();
    encoder << instance.topRightRadius();
    encoder << instance.bottomRightRadius();
    encoder << instance.bottomLeftRadius();
}

std::optional<Ref<WebCore::BasicShapeInset>> ArgumentCoder<WebCore::BasicShapeInset>::decode(Decoder& decoder)
{
    auto right = decoder.decode<WebCore::Length>();
    auto top = decoder.decode<WebCore::Length>();
    auto bottom = decoder.decode<WebCore::Length>();
    auto left = decoder.decode<WebCore::Length>();
    auto topLeftRadius = decoder.decode<WebCore::LengthSize>();
    auto topRightRadius = decoder.decode<WebCore::LengthSize>();
    auto bottomRightRadius = decoder.decode<WebCore::LengthSize>();
    auto bottomLeftRadius = decoder.decode<WebCore::LengthSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BasicShapeInset::create(
            WTFMove(*right),
            WTFMove(*top),
            WTFMove(*bottom),
            WTFMove(*left),
            WTFMove(*topLeftRadius),
            WTFMove(*topRightRadius),
            WTFMove(*bottomRightRadius),
            WTFMove(*bottomLeftRadius)
        )
    };
}

enum class WebCore_BasicShape_Subclass : IPC::EncodedVariantIndex {
    BasicShapeCircle,
    BasicShapeEllipse,
    BasicShapePolygon,
    BasicShapePath,
    BasicShapeInset
};

void ArgumentCoder<WebCore::BasicShape>::encode(Encoder& encoder, const WebCore::BasicShape& instance)
{
    if (auto* subclass = dynamicDowncast<WebCore::BasicShapeCircle>(instance)) {
        encoder << WebCore_BasicShape_Subclass::BasicShapeCircle;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::BasicShapeEllipse>(instance)) {
        encoder << WebCore_BasicShape_Subclass::BasicShapeEllipse;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::BasicShapePolygon>(instance)) {
        encoder << WebCore_BasicShape_Subclass::BasicShapePolygon;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::BasicShapePath>(instance)) {
        encoder << WebCore_BasicShape_Subclass::BasicShapePath;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::BasicShapeInset>(instance)) {
        encoder << WebCore_BasicShape_Subclass::BasicShapeInset;
        encoder << *subclass;
        return;
    }
    ASSERT_NOT_REACHED();
}

std::optional<Ref<WebCore::BasicShape>> ArgumentCoder<WebCore::BasicShape>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore_BasicShape_Subclass>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (type == WebCore_BasicShape_Subclass::BasicShapeCircle) {
        auto result = decoder.decode<Ref<WebCore::BasicShapeCircle>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_BasicShape_Subclass::BasicShapeEllipse) {
        auto result = decoder.decode<Ref<WebCore::BasicShapeEllipse>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_BasicShape_Subclass::BasicShapePolygon) {
        auto result = decoder.decode<Ref<WebCore::BasicShapePolygon>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_BasicShape_Subclass::BasicShapePath) {
        auto result = decoder.decode<Ref<WebCore::BasicShapePath>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_BasicShape_Subclass::BasicShapeInset) {
        auto result = decoder.decode<Ref<WebCore::BasicShapeInset>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    ASSERT_NOT_REACHED();
    return std::nullopt;
}

void ArgumentCoder<WebCore::ReferencePathOperation>::encode(Encoder& encoder, const WebCore::ReferencePathOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, std::optional<WebCore::Path>>);
    encoder << instance.path();
}

std::optional<Ref<WebCore::ReferencePathOperation>> ArgumentCoder<WebCore::ReferencePathOperation>::decode(Decoder& decoder)
{
    auto path = decoder.decode<std::optional<WebCore::Path>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ReferencePathOperation::create(
            WTFMove(*path)
        )
    };
}

void ArgumentCoder<WebCore::ShapePathOperation>::encode(Encoder& encoder, const WebCore::ShapePathOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shape())>, Ref<WebCore::BasicShape>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referenceBox())>, WebCore::CSSBoxType>);
    encoder << instance.shape();
    encoder << instance.referenceBox();
}

std::optional<Ref<WebCore::ShapePathOperation>> ArgumentCoder<WebCore::ShapePathOperation>::decode(Decoder& decoder)
{
    auto shape = decoder.decode<Ref<WebCore::BasicShape>>();
    auto referenceBox = decoder.decode<WebCore::CSSBoxType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShapePathOperation::create(
            WTFMove(*shape),
            WTFMove(*referenceBox)
        )
    };
}

void ArgumentCoder<WebCore::BoxPathOperation>::encode(Encoder& encoder, const WebCore::BoxPathOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.path())>, WebCore::Path>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referenceBox())>, WebCore::CSSBoxType>);
    encoder << instance.path();
    encoder << instance.referenceBox();
}

std::optional<Ref<WebCore::BoxPathOperation>> ArgumentCoder<WebCore::BoxPathOperation>::decode(Decoder& decoder)
{
    auto path = decoder.decode<WebCore::Path>();
    auto referenceBox = decoder.decode<WebCore::CSSBoxType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BoxPathOperation::create(
            WTFMove(*path),
            WTFMove(*referenceBox)
        )
    };
}

void ArgumentCoder<WebCore::RayPathOperation>::encode(Encoder& encoder, const WebCore::RayPathOperation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.angle())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size())>, WebCore::RayPathOperation::Size>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isContaining())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containingBlockBoundingRect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::FloatPoint>);
    encoder << instance.angle();
    encoder << instance.size();
    encoder << instance.isContaining();
    encoder << instance.containingBlockBoundingRect();
    encoder << instance.position();
}

std::optional<Ref<WebCore::RayPathOperation>> ArgumentCoder<WebCore::RayPathOperation>::decode(Decoder& decoder)
{
    auto angle = decoder.decode<float>();
    auto size = decoder.decode<WebCore::RayPathOperation::Size>();
    auto isContaining = decoder.decode<bool>();
    auto containingBlockBoundingRect = decoder.decode<WebCore::FloatRect>();
    auto position = decoder.decode<WebCore::FloatPoint>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RayPathOperation::create(
            WTFMove(*angle),
            WTFMove(*size),
            WTFMove(*isContaining),
            WTFMove(*containingBlockBoundingRect),
            WTFMove(*position)
        )
    };
}

enum class WebCore_PathOperation_Subclass : IPC::EncodedVariantIndex {
    ReferencePathOperation,
    ShapePathOperation,
    BoxPathOperation,
    RayPathOperation
};

void ArgumentCoder<WebCore::PathOperation>::encode(Encoder& encoder, const WebCore::PathOperation& instance)
{
    if (auto* subclass = dynamicDowncast<WebCore::ReferencePathOperation>(instance)) {
        encoder << WebCore_PathOperation_Subclass::ReferencePathOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::ShapePathOperation>(instance)) {
        encoder << WebCore_PathOperation_Subclass::ShapePathOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::BoxPathOperation>(instance)) {
        encoder << WebCore_PathOperation_Subclass::BoxPathOperation;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::RayPathOperation>(instance)) {
        encoder << WebCore_PathOperation_Subclass::RayPathOperation;
        encoder << *subclass;
        return;
    }
    ASSERT_NOT_REACHED();
}

std::optional<Ref<WebCore::PathOperation>> ArgumentCoder<WebCore::PathOperation>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore_PathOperation_Subclass>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (type == WebCore_PathOperation_Subclass::ReferencePathOperation) {
        auto result = decoder.decode<Ref<WebCore::ReferencePathOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_PathOperation_Subclass::ShapePathOperation) {
        auto result = decoder.decode<Ref<WebCore::ShapePathOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_PathOperation_Subclass::BoxPathOperation) {
        auto result = decoder.decode<Ref<WebCore::BoxPathOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_PathOperation_Subclass::RayPathOperation) {
        auto result = decoder.decode<Ref<WebCore::RayPathOperation>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    ASSERT_NOT_REACHED();
    return std::nullopt;
}

#if ENABLE(ENCRYPTED_MEDIA)
void ArgumentCoder<WebCore::CDMKeySystemConfiguration>::encode(Encoder& encoder, const WebCore::CDMKeySystemConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.initDataTypes)>, Vector<AtomString>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.audioCapabilities)>, Vector<WebCore::CDMMediaCapability>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.videoCapabilities)>, Vector<WebCore::CDMMediaCapability>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.distinctiveIdentifier)>, WebCore::CDMRequirement>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.persistentState)>, WebCore::CDMRequirement>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sessionTypes)>, Vector<WebCore::CDMSessionType>>);
    struct ShouldBeSameSizeAsCDMKeySystemConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CDMKeySystemConfiguration>, false> {
        String label;
        Vector<AtomString> initDataTypes;
        Vector<WebCore::CDMMediaCapability> audioCapabilities;
        Vector<WebCore::CDMMediaCapability> videoCapabilities;
        WebCore::CDMRequirement distinctiveIdentifier;
        WebCore::CDMRequirement persistentState;
        Vector<WebCore::CDMSessionType> sessionTypes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCDMKeySystemConfiguration) == sizeof(WebCore::CDMKeySystemConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CDMKeySystemConfiguration, label)
        , offsetof(WebCore::CDMKeySystemConfiguration, initDataTypes)
        , offsetof(WebCore::CDMKeySystemConfiguration, audioCapabilities)
        , offsetof(WebCore::CDMKeySystemConfiguration, videoCapabilities)
        , offsetof(WebCore::CDMKeySystemConfiguration, distinctiveIdentifier)
        , offsetof(WebCore::CDMKeySystemConfiguration, persistentState)
        , offsetof(WebCore::CDMKeySystemConfiguration, sessionTypes)
    >::value);
    encoder << instance.label;
    encoder << instance.initDataTypes;
    encoder << instance.audioCapabilities;
    encoder << instance.videoCapabilities;
    encoder << instance.distinctiveIdentifier;
    encoder << instance.persistentState;
    encoder << instance.sessionTypes;
}

std::optional<WebCore::CDMKeySystemConfiguration> ArgumentCoder<WebCore::CDMKeySystemConfiguration>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto initDataTypes = decoder.decode<Vector<AtomString>>();
    auto audioCapabilities = decoder.decode<Vector<WebCore::CDMMediaCapability>>();
    auto videoCapabilities = decoder.decode<Vector<WebCore::CDMMediaCapability>>();
    auto distinctiveIdentifier = decoder.decode<WebCore::CDMRequirement>();
    auto persistentState = decoder.decode<WebCore::CDMRequirement>();
    auto sessionTypes = decoder.decode<Vector<WebCore::CDMSessionType>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CDMKeySystemConfiguration {
            WTFMove(*label),
            WTFMove(*initDataTypes),
            WTFMove(*audioCapabilities),
            WTFMove(*videoCapabilities),
            WTFMove(*distinctiveIdentifier),
            WTFMove(*persistentState),
            WTFMove(*sessionTypes)
        }
    };
}

#endif

void ArgumentCoder<WebCore::MediaConfiguration>::encode(Encoder& encoder, const WebCore::MediaConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.video)>, std::optional<WebCore::VideoConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.audio)>, std::optional<WebCore::AudioConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaContainerTypes)>, std::optional<Vector<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaCodecTypes)>, std::optional<Vector<String>>>);
    struct ShouldBeSameSizeAsMediaConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaConfiguration>, false> {
        std::optional<WebCore::VideoConfiguration> video;
        std::optional<WebCore::AudioConfiguration> audio;
        std::optional<Vector<String>> allowedMediaContainerTypes;
        std::optional<Vector<String>> allowedMediaCodecTypes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaConfiguration) == sizeof(WebCore::MediaConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaConfiguration, video)
        , offsetof(WebCore::MediaConfiguration, audio)
        , offsetof(WebCore::MediaConfiguration, allowedMediaContainerTypes)
        , offsetof(WebCore::MediaConfiguration, allowedMediaCodecTypes)
    >::value);
    encoder << instance.video;
    encoder << instance.audio;
    encoder << instance.allowedMediaContainerTypes;
    encoder << instance.allowedMediaCodecTypes;
}

std::optional<WebCore::MediaConfiguration> ArgumentCoder<WebCore::MediaConfiguration>::decode(Decoder& decoder)
{
    auto video = decoder.decode<std::optional<WebCore::VideoConfiguration>>();
    auto audio = decoder.decode<std::optional<WebCore::AudioConfiguration>>();
    auto allowedMediaContainerTypes = decoder.decode<std::optional<Vector<String>>>();
    auto allowedMediaCodecTypes = decoder.decode<std::optional<Vector<String>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaConfiguration {
            WTFMove(*video),
            WTFMove(*audio),
            WTFMove(*allowedMediaContainerTypes),
            WTFMove(*allowedMediaCodecTypes)
        }
    };
}

void ArgumentCoder<WebCore::MediaEncodingConfiguration>::encode(Encoder& encoder, const WebCore::MediaEncodingConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.video)>, std::optional<WebCore::VideoConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.audio)>, std::optional<WebCore::AudioConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaContainerTypes)>, std::optional<Vector<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaCodecTypes)>, std::optional<Vector<String>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaConfiguration, video)
        , offsetof(WebCore::MediaConfiguration, audio)
        , offsetof(WebCore::MediaConfiguration, allowedMediaContainerTypes)
        , offsetof(WebCore::MediaConfiguration, allowedMediaCodecTypes)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::MediaEncodingType>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaEncodingConfiguration, type)
    >::value);
    encoder << instance.video;
    encoder << instance.audio;
    encoder << instance.allowedMediaContainerTypes;
    encoder << instance.allowedMediaCodecTypes;
    encoder << instance.type;
}

std::optional<WebCore::MediaEncodingConfiguration> ArgumentCoder<WebCore::MediaEncodingConfiguration>::decode(Decoder& decoder)
{
    auto video = decoder.decode<std::optional<WebCore::VideoConfiguration>>();
    auto audio = decoder.decode<std::optional<WebCore::AudioConfiguration>>();
    auto allowedMediaContainerTypes = decoder.decode<std::optional<Vector<String>>>();
    auto allowedMediaCodecTypes = decoder.decode<std::optional<Vector<String>>>();
    auto type = decoder.decode<WebCore::MediaEncodingType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaEncodingConfiguration {
            WebCore::MediaConfiguration {
                WTFMove(*video),
                WTFMove(*audio),
                WTFMove(*allowedMediaContainerTypes),
                WTFMove(*allowedMediaCodecTypes)
            },
            WTFMove(*type)
        }
    };
}

void ArgumentCoder<WebCore::MediaDecodingConfiguration>::encode(Encoder& encoder, const WebCore::MediaDecodingConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.video)>, std::optional<WebCore::VideoConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.audio)>, std::optional<WebCore::AudioConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaContainerTypes)>, std::optional<Vector<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaCodecTypes)>, std::optional<Vector<String>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaConfiguration, video)
        , offsetof(WebCore::MediaConfiguration, audio)
        , offsetof(WebCore::MediaConfiguration, allowedMediaContainerTypes)
        , offsetof(WebCore::MediaConfiguration, allowedMediaCodecTypes)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::MediaDecodingType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canExposeVP9)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaDecodingConfiguration, type)
        , offsetof(WebCore::MediaDecodingConfiguration, canExposeVP9)
    >::value);
    encoder << instance.video;
    encoder << instance.audio;
    encoder << instance.allowedMediaContainerTypes;
    encoder << instance.allowedMediaCodecTypes;
    encoder << instance.type;
    encoder << instance.canExposeVP9;
}

std::optional<WebCore::MediaDecodingConfiguration> ArgumentCoder<WebCore::MediaDecodingConfiguration>::decode(Decoder& decoder)
{
    auto video = decoder.decode<std::optional<WebCore::VideoConfiguration>>();
    auto audio = decoder.decode<std::optional<WebCore::AudioConfiguration>>();
    auto allowedMediaContainerTypes = decoder.decode<std::optional<Vector<String>>>();
    auto allowedMediaCodecTypes = decoder.decode<std::optional<Vector<String>>>();
    auto type = decoder.decode<WebCore::MediaDecodingType>();
    auto canExposeVP9 = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaDecodingConfiguration {
            WebCore::MediaConfiguration {
                WTFMove(*video),
                WTFMove(*audio),
                WTFMove(*allowedMediaContainerTypes),
                WTFMove(*allowedMediaCodecTypes)
            },
            WTFMove(*type),
            WTFMove(*canExposeVP9)
        }
    };
}

void ArgumentCoder<WebCore::ResourceResponseBase::ResponseData>::encode(Encoder& encoder, const WebCore::ResourceResponseBase::ResponseData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_mimeType)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_expectedContentLength)>, long long>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_textEncodingName)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_httpStatusText)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_httpVersion)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_httpHeaderFields)>, WebCore::HTTPHeaderMap>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_networkLoadMetrics)>, Box<WebCore::NetworkLoadMetrics>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_httpStatusCode)>, short>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_certificateInfo)>, std::optional<WebCore::CertificateInfo>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_source)>, WebCore::ResourceResponseBase::Source>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_type)>, WebCore::ResourceResponseBase::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_tainting)>, WebCore::ResourceResponseBase::Tainting>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isRedirected)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_usedLegacyTLS)>, WebCore::UsedLegacyTLS>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_wasPrivateRelayed)>, WebCore::WasPrivateRelayed>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isRangeRequested)>, bool>);
    struct ShouldBeSameSizeAsResponseData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ResourceResponseBase::ResponseData>, false> {
        URL m_url;
        AtomString m_mimeType;
        long long m_expectedContentLength;
        AtomString m_textEncodingName;
        AtomString m_httpStatusText;
        AtomString m_httpVersion;
        WebCore::HTTPHeaderMap m_httpHeaderFields;
        Box<WebCore::NetworkLoadMetrics> m_networkLoadMetrics;
        short m_httpStatusCode;
        std::optional<WebCore::CertificateInfo> m_certificateInfo;
        WebCore::ResourceResponseBase::Source m_source;
        WebCore::ResourceResponseBase::Type m_type;
        WebCore::ResourceResponseBase::Tainting m_tainting;
        bool m_isRedirected;
        WebCore::UsedLegacyTLS m_usedLegacyTLS;
        WebCore::WasPrivateRelayed m_wasPrivateRelayed;
        bool m_isRangeRequested;
    };
    static_assert(sizeof(ShouldBeSameSizeAsResponseData) == sizeof(WebCore::ResourceResponseBase::ResponseData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_url)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_mimeType)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_expectedContentLength)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_textEncodingName)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_httpStatusText)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_httpVersion)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_httpHeaderFields)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_networkLoadMetrics)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_httpStatusCode)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_certificateInfo)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_source)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_type)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_tainting)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_isRedirected)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_usedLegacyTLS)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_wasPrivateRelayed)
        , offsetof(WebCore::ResourceResponseBase::ResponseData, m_isRangeRequested)
    >::value);
    encoder << instance.m_url;
    encoder << instance.m_mimeType;
    encoder << instance.m_expectedContentLength;
    encoder << instance.m_textEncodingName;
    encoder << instance.m_httpStatusText;
    encoder << instance.m_httpVersion;
    encoder << instance.m_httpHeaderFields;
    encoder << instance.m_networkLoadMetrics;
    encoder << instance.m_httpStatusCode;
    encoder << instance.m_certificateInfo;
    encoder << instance.m_source;
    encoder << instance.m_type;
    encoder << instance.m_tainting;
    encoder << instance.m_isRedirected;
    encoder << instance.m_usedLegacyTLS;
    encoder << instance.m_wasPrivateRelayed;
    encoder << instance.m_isRangeRequested;
}

std::optional<WebCore::ResourceResponseBase::ResponseData> ArgumentCoder<WebCore::ResourceResponseBase::ResponseData>::decode(Decoder& decoder)
{
    auto m_url = decoder.decode<URL>();
    auto m_mimeType = decoder.decode<AtomString>();
    auto m_expectedContentLength = decoder.decode<long long>();
    auto m_textEncodingName = decoder.decode<AtomString>();
    auto m_httpStatusText = decoder.decode<AtomString>();
    auto m_httpVersion = decoder.decode<AtomString>();
    auto m_httpHeaderFields = decoder.decode<WebCore::HTTPHeaderMap>();
    auto m_networkLoadMetrics = decoder.decode<Box<WebCore::NetworkLoadMetrics>>();
    auto m_httpStatusCode = decoder.decode<short>();
    auto m_certificateInfo = decoder.decode<std::optional<WebCore::CertificateInfo>>();
    auto m_source = decoder.decode<WebCore::ResourceResponseBase::Source>();
    auto m_type = decoder.decode<WebCore::ResourceResponseBase::Type>();
    auto m_tainting = decoder.decode<WebCore::ResourceResponseBase::Tainting>();
    auto m_isRedirected = decoder.decode<bool>();
    auto m_usedLegacyTLS = decoder.decode<WebCore::UsedLegacyTLS>();
    auto m_wasPrivateRelayed = decoder.decode<WebCore::WasPrivateRelayed>();
    auto m_isRangeRequested = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceResponseBase::ResponseData {
            WTFMove(*m_url),
            WTFMove(*m_mimeType),
            WTFMove(*m_expectedContentLength),
            WTFMove(*m_textEncodingName),
            WTFMove(*m_httpStatusText),
            WTFMove(*m_httpVersion),
            WTFMove(*m_httpHeaderFields),
            WTFMove(*m_networkLoadMetrics),
            WTFMove(*m_httpStatusCode),
            WTFMove(*m_certificateInfo),
            WTFMove(*m_source),
            WTFMove(*m_type),
            WTFMove(*m_tainting),
            WTFMove(*m_isRedirected),
            WTFMove(*m_usedLegacyTLS),
            WTFMove(*m_wasPrivateRelayed),
            WTFMove(*m_isRangeRequested)
        }
    };
}

void ArgumentCoder<WebCore::ResourceResponseBase>::encode(Encoder& encoder, const WebCore::ResourceResponseBase& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getResponseData())>, std::optional<WebCore::ResourceResponseBase::ResponseData>>);
    encoder << instance.getResponseData();
}

std::optional<WebCore::ResourceResponseBase> ArgumentCoder<WebCore::ResourceResponseBase>::decode(Decoder& decoder)
{
    auto getResponseData = decoder.decode<std::optional<WebCore::ResourceResponseBase::ResponseData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceResponseBase {
            WTFMove(*getResponseData)
        }
    };
}

void ArgumentCoder<WebCore::ResourceResponse>::encode(Encoder& encoder, const WebCore::ResourceResponse& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getResponseData())>, std::optional<WebCore::ResourceResponseBase::ResponseData>>);
    static_assert(MembersInCorrectOrder < 0
    >::value);
    encoder << instance.getResponseData();
}

std::optional<WebCore::ResourceResponse> ArgumentCoder<WebCore::ResourceResponse>::decode(Decoder& decoder)
{
    auto getResponseData = decoder.decode<std::optional<WebCore::ResourceResponseBase::ResponseData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceResponse {
            WebCore::ResourceResponseBase {
                WTFMove(*getResponseData)
            }
        }
    };
}

void ArgumentCoder<WebCore::ResourceResponseBase::CrossThreadData>::encode(Encoder& encoder, const WebCore::ResourceResponseBase::CrossThreadData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mimeType)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expectedContentLength)>, long long>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textEncodingName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpStatusCode)>, short>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpStatusText)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpVersion)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpHeaderFields)>, WebCore::HTTPHeaderMap>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkLoadMetrics)>, std::optional<WebCore::NetworkLoadMetrics>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.source)>, WebCore::ResourceResponseBase::Source>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ResourceResponseBase::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tainting)>, WebCore::ResourceResponseBase::Tainting>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isRedirected)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usedLegacyTLS)>, WebCore::UsedLegacyTLS>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wasPrivateRelayed)>, WebCore::WasPrivateRelayed>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isRangeRequested)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.certificateInfo)>, std::optional<WebCore::CertificateInfo>>);
    struct ShouldBeSameSizeAsCrossThreadData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ResourceResponseBase::CrossThreadData>, false> {
        URL url;
        String mimeType;
        long long expectedContentLength;
        String textEncodingName;
        short httpStatusCode;
        String httpStatusText;
        String httpVersion;
        WebCore::HTTPHeaderMap httpHeaderFields;
        std::optional<WebCore::NetworkLoadMetrics> networkLoadMetrics;
        WebCore::ResourceResponseBase::Source source;
        WebCore::ResourceResponseBase::Type type;
        WebCore::ResourceResponseBase::Tainting tainting;
        bool isRedirected;
        WebCore::UsedLegacyTLS usedLegacyTLS;
        WebCore::WasPrivateRelayed wasPrivateRelayed;
        bool isRangeRequested;
        std::optional<WebCore::CertificateInfo> certificateInfo;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCrossThreadData) == sizeof(WebCore::ResourceResponseBase::CrossThreadData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, url)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, mimeType)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, expectedContentLength)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, textEncodingName)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, httpStatusCode)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, httpStatusText)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, httpVersion)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, httpHeaderFields)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, networkLoadMetrics)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, source)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, type)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, tainting)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, isRedirected)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, usedLegacyTLS)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, wasPrivateRelayed)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, isRangeRequested)
        , offsetof(WebCore::ResourceResponseBase::CrossThreadData, certificateInfo)
    >::value);
    encoder << instance.url;
    encoder << instance.mimeType;
    encoder << instance.expectedContentLength;
    encoder << instance.textEncodingName;
    encoder << instance.httpStatusCode;
    encoder << instance.httpStatusText;
    encoder << instance.httpVersion;
    encoder << instance.httpHeaderFields;
    encoder << instance.networkLoadMetrics;
    encoder << instance.source;
    encoder << instance.type;
    encoder << instance.tainting;
    encoder << instance.isRedirected;
    encoder << instance.usedLegacyTLS;
    encoder << instance.wasPrivateRelayed;
    encoder << instance.isRangeRequested;
    encoder << instance.certificateInfo;
}

std::optional<WebCore::ResourceResponseBase::CrossThreadData> ArgumentCoder<WebCore::ResourceResponseBase::CrossThreadData>::decode(Decoder& decoder)
{
    auto url = decoder.decode<URL>();
    auto mimeType = decoder.decode<String>();
    auto expectedContentLength = decoder.decode<long long>();
    auto textEncodingName = decoder.decode<String>();
    auto httpStatusCode = decoder.decode<short>();
    auto httpStatusText = decoder.decode<String>();
    auto httpVersion = decoder.decode<String>();
    auto httpHeaderFields = decoder.decode<WebCore::HTTPHeaderMap>();
    auto networkLoadMetrics = decoder.decode<std::optional<WebCore::NetworkLoadMetrics>>();
    auto source = decoder.decode<WebCore::ResourceResponseBase::Source>();
    auto type = decoder.decode<WebCore::ResourceResponseBase::Type>();
    auto tainting = decoder.decode<WebCore::ResourceResponseBase::Tainting>();
    auto isRedirected = decoder.decode<bool>();
    auto usedLegacyTLS = decoder.decode<WebCore::UsedLegacyTLS>();
    auto wasPrivateRelayed = decoder.decode<WebCore::WasPrivateRelayed>();
    auto isRangeRequested = decoder.decode<bool>();
    auto certificateInfo = decoder.decode<std::optional<WebCore::CertificateInfo>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ResourceResponseBase::CrossThreadData {
            WTFMove(*url),
            WTFMove(*mimeType),
            WTFMove(*expectedContentLength),
            WTFMove(*textEncodingName),
            WTFMove(*httpStatusCode),
            WTFMove(*httpStatusText),
            WTFMove(*httpVersion),
            WTFMove(*httpHeaderFields),
            WTFMove(*networkLoadMetrics),
            WTFMove(*source),
            WTFMove(*type),
            WTFMove(*tainting),
            WTFMove(*isRedirected),
            WTFMove(*usedLegacyTLS),
            WTFMove(*wasPrivateRelayed),
            WTFMove(*isRangeRequested),
            WTFMove(*certificateInfo)
        }
    };
}

void ArgumentCoder<WebCore::FetchOptions>::encode(Encoder& encoder, const WebCore::FetchOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destination)>, WebCore::FetchOptionsDestination>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mode)>, WebCore::FetchOptionsMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.credentials)>, WebCore::FetchOptionsCredentials>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cache)>, WebCore::FetchOptionsCache>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.redirect)>, WebCore::FetchOptionsRedirect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrerPolicy)>, WebCore::ReferrerPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keepAlive)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.integrity)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientIdentifier)>, Markable<WTF::UUID>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resultingClientIdentifier)>, Markable<WTF::UUID>>);
    struct ShouldBeSameSizeAsFetchOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FetchOptions>, false> {
        WebCore::FetchOptionsDestination destination;
        WebCore::FetchOptionsMode mode;
        WebCore::FetchOptionsCredentials credentials;
        WebCore::FetchOptionsCache cache;
        WebCore::FetchOptionsRedirect redirect;
        WebCore::ReferrerPolicy referrerPolicy;
        bool keepAlive;
        String integrity;
        Markable<WTF::UUID> clientIdentifier;
        Markable<WTF::UUID> resultingClientIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFetchOptions) == sizeof(WebCore::FetchOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FetchOptions, destination)
        , offsetof(WebCore::FetchOptions, mode)
        , offsetof(WebCore::FetchOptions, credentials)
        , offsetof(WebCore::FetchOptions, cache)
        , offsetof(WebCore::FetchOptions, redirect)
        , offsetof(WebCore::FetchOptions, referrerPolicy)
        , offsetof(WebCore::FetchOptions, keepAlive)
        , offsetof(WebCore::FetchOptions, integrity)
        , offsetof(WebCore::FetchOptions, clientIdentifier)
        , offsetof(WebCore::FetchOptions, resultingClientIdentifier)
    >::value);
    encoder << instance.destination;
    encoder << instance.mode;
    encoder << instance.credentials;
    encoder << instance.cache;
    encoder << instance.redirect;
    encoder << instance.referrerPolicy;
    encoder << instance.keepAlive;
    encoder << instance.integrity;
    encoder << instance.clientIdentifier;
    encoder << instance.resultingClientIdentifier;
}

std::optional<WebCore::FetchOptions> ArgumentCoder<WebCore::FetchOptions>::decode(Decoder& decoder)
{
    auto destination = decoder.decode<WebCore::FetchOptionsDestination>();
    auto mode = decoder.decode<WebCore::FetchOptionsMode>();
    auto credentials = decoder.decode<WebCore::FetchOptionsCredentials>();
    auto cache = decoder.decode<WebCore::FetchOptionsCache>();
    auto redirect = decoder.decode<WebCore::FetchOptionsRedirect>();
    auto referrerPolicy = decoder.decode<WebCore::ReferrerPolicy>();
    auto keepAlive = decoder.decode<bool>();
    auto integrity = decoder.decode<String>();
    auto clientIdentifier = decoder.decode<Markable<WTF::UUID>>();
    auto resultingClientIdentifier = decoder.decode<Markable<WTF::UUID>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FetchOptions {
            WTFMove(*destination),
            WTFMove(*mode),
            WTFMove(*credentials),
            WTFMove(*cache),
            WTFMove(*redirect),
            WTFMove(*referrerPolicy),
            WTFMove(*keepAlive),
            WTFMove(*integrity),
            WTFMove(*clientIdentifier),
            WTFMove(*resultingClientIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::ControlStyle>::encode(Encoder& encoder, const WebCore::ControlStyle& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.states)>, OptionSet<WebCore::ControlStyle::State>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontSize)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.zoomFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.accentColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.borderWidth)>, WebCore::FloatBoxExtent>);
    struct ShouldBeSameSizeAsControlStyle : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ControlStyle>, false> {
        OptionSet<WebCore::ControlStyle::State> states;
        float fontSize;
        float zoomFactor;
        WebCore::Color accentColor;
        WebCore::Color textColor;
        WebCore::FloatBoxExtent borderWidth;
    };
    static_assert(sizeof(ShouldBeSameSizeAsControlStyle) == sizeof(WebCore::ControlStyle));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ControlStyle, states)
        , offsetof(WebCore::ControlStyle, fontSize)
        , offsetof(WebCore::ControlStyle, zoomFactor)
        , offsetof(WebCore::ControlStyle, accentColor)
        , offsetof(WebCore::ControlStyle, textColor)
        , offsetof(WebCore::ControlStyle, borderWidth)
    >::value);
    encoder << instance.states;
    encoder << instance.fontSize;
    encoder << instance.zoomFactor;
    encoder << instance.accentColor;
    encoder << instance.textColor;
    encoder << instance.borderWidth;
}

void ArgumentCoder<WebCore::ControlStyle>::encode(StreamConnectionEncoder& encoder, const WebCore::ControlStyle& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.states)>, OptionSet<WebCore::ControlStyle::State>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontSize)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.zoomFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.accentColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.borderWidth)>, WebCore::FloatBoxExtent>);
    struct ShouldBeSameSizeAsControlStyle : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ControlStyle>, false> {
        OptionSet<WebCore::ControlStyle::State> states;
        float fontSize;
        float zoomFactor;
        WebCore::Color accentColor;
        WebCore::Color textColor;
        WebCore::FloatBoxExtent borderWidth;
    };
    static_assert(sizeof(ShouldBeSameSizeAsControlStyle) == sizeof(WebCore::ControlStyle));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ControlStyle, states)
        , offsetof(WebCore::ControlStyle, fontSize)
        , offsetof(WebCore::ControlStyle, zoomFactor)
        , offsetof(WebCore::ControlStyle, accentColor)
        , offsetof(WebCore::ControlStyle, textColor)
        , offsetof(WebCore::ControlStyle, borderWidth)
    >::value);
    encoder << instance.states;
    encoder << instance.fontSize;
    encoder << instance.zoomFactor;
    encoder << instance.accentColor;
    encoder << instance.textColor;
    encoder << instance.borderWidth;
}

std::optional<WebCore::ControlStyle> ArgumentCoder<WebCore::ControlStyle>::decode(Decoder& decoder)
{
    auto states = decoder.decode<OptionSet<WebCore::ControlStyle::State>>();
    auto fontSize = decoder.decode<float>();
    auto zoomFactor = decoder.decode<float>();
    auto accentColor = decoder.decode<WebCore::Color>();
    auto textColor = decoder.decode<WebCore::Color>();
    auto borderWidth = decoder.decode<WebCore::FloatBoxExtent>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ControlStyle {
            WTFMove(*states),
            WTFMove(*fontSize),
            WTFMove(*zoomFactor),
            WTFMove(*accentColor),
            WTFMove(*textColor),
            WTFMove(*borderWidth)
        }
    };
}

#if ENABLE(APPLE_PAY)
void ArgumentCoder<WebCore::ApplePayButtonPart>::encode(Encoder& encoder, const WebCore::ApplePayButtonPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buttonType())>, WebCore::ApplePayButtonType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buttonStyle())>, WebCore::ApplePayButtonStyle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.locale())>, String>);
    encoder << instance.buttonType();
    encoder << instance.buttonStyle();
    encoder << instance.locale();
}

void ArgumentCoder<WebCore::ApplePayButtonPart>::encode(StreamConnectionEncoder& encoder, const WebCore::ApplePayButtonPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buttonType())>, WebCore::ApplePayButtonType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buttonStyle())>, WebCore::ApplePayButtonStyle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.locale())>, String>);
    encoder << instance.buttonType();
    encoder << instance.buttonStyle();
    encoder << instance.locale();
}

std::optional<Ref<WebCore::ApplePayButtonPart>> ArgumentCoder<WebCore::ApplePayButtonPart>::decode(Decoder& decoder)
{
    auto buttonType = decoder.decode<WebCore::ApplePayButtonType>();
    auto buttonStyle = decoder.decode<WebCore::ApplePayButtonStyle>();
    auto locale = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ApplePayButtonPart::create(
            WTFMove(*buttonType),
            WTFMove(*buttonStyle),
            WTFMove(*locale)
        )
    };
}

#endif

void ArgumentCoder<WebCore::MeterPart>::encode(Encoder& encoder, const WebCore::MeterPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gaugeRegion())>, WebCore::MeterPart::GaugeRegion>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimum())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximum())>, double>);
    encoder << instance.gaugeRegion();
    encoder << instance.value();
    encoder << instance.minimum();
    encoder << instance.maximum();
}

void ArgumentCoder<WebCore::MeterPart>::encode(StreamConnectionEncoder& encoder, const WebCore::MeterPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gaugeRegion())>, WebCore::MeterPart::GaugeRegion>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimum())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximum())>, double>);
    encoder << instance.gaugeRegion();
    encoder << instance.value();
    encoder << instance.minimum();
    encoder << instance.maximum();
}

std::optional<Ref<WebCore::MeterPart>> ArgumentCoder<WebCore::MeterPart>::decode(Decoder& decoder)
{
    auto gaugeRegion = decoder.decode<WebCore::MeterPart::GaugeRegion>();
    auto value = decoder.decode<double>();
    auto minimum = decoder.decode<double>();
    auto maximum = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MeterPart::create(
            WTFMove(*gaugeRegion),
            WTFMove(*value),
            WTFMove(*minimum),
            WTFMove(*maximum)
        )
    };
}

void ArgumentCoder<WebCore::ProgressBarPart>::encode(Encoder& encoder, const WebCore::ProgressBarPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.animationStartTime())>, Seconds>);
    encoder << instance.position();
    encoder << instance.animationStartTime();
}

void ArgumentCoder<WebCore::ProgressBarPart>::encode(StreamConnectionEncoder& encoder, const WebCore::ProgressBarPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.animationStartTime())>, Seconds>);
    encoder << instance.position();
    encoder << instance.animationStartTime();
}

std::optional<Ref<WebCore::ProgressBarPart>> ArgumentCoder<WebCore::ProgressBarPart>::decode(Decoder& decoder)
{
    auto position = decoder.decode<double>();
    auto animationStartTime = decoder.decode<Seconds>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ProgressBarPart::create(
            WTFMove(*position),
            WTFMove(*animationStartTime)
        )
    };
}

void ArgumentCoder<WebCore::SliderTrackPart>::encode(Encoder& encoder, const WebCore::SliderTrackPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::StyleAppearance>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.thumbSize())>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackBounds())>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tickRatios())>, Vector<double>>);
    encoder << instance.type();
    encoder << instance.thumbSize();
    encoder << instance.trackBounds();
    encoder << instance.tickRatios();
}

void ArgumentCoder<WebCore::SliderTrackPart>::encode(StreamConnectionEncoder& encoder, const WebCore::SliderTrackPart& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::StyleAppearance>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.thumbSize())>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackBounds())>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tickRatios())>, Vector<double>>);
    encoder << instance.type();
    encoder << instance.thumbSize();
    encoder << instance.trackBounds();
    encoder << instance.tickRatios();
}

std::optional<Ref<WebCore::SliderTrackPart>> ArgumentCoder<WebCore::SliderTrackPart>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::StyleAppearance>();
    auto thumbSize = decoder.decode<WebCore::IntSize>();
    auto trackBounds = decoder.decode<WebCore::IntRect>();
    auto tickRatios = decoder.decode<Vector<double>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SliderTrackPart::create(
            WTFMove(*type),
            WTFMove(*thumbSize),
            WTFMove(*trackBounds),
            WTFMove(*tickRatios)
        )
    };
}

void ArgumentCoder<WebCore::DistantLightSource>::encode(Encoder& encoder, const WebCore::DistantLightSource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.azimuth())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elevation())>, float>);
    encoder << instance.azimuth();
    encoder << instance.elevation();
}

void ArgumentCoder<WebCore::DistantLightSource>::encode(StreamConnectionEncoder& encoder, const WebCore::DistantLightSource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.azimuth())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elevation())>, float>);
    encoder << instance.azimuth();
    encoder << instance.elevation();
}

std::optional<Ref<WebCore::DistantLightSource>> ArgumentCoder<WebCore::DistantLightSource>::decode(Decoder& decoder)
{
    auto azimuth = decoder.decode<float>();
    auto elevation = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DistantLightSource::create(
            WTFMove(*azimuth),
            WTFMove(*elevation)
        )
    };
}

void ArgumentCoder<WebCore::PointLightSource>::encode(Encoder& encoder, const WebCore::PointLightSource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::FloatPoint3D>);
    encoder << instance.position();
}

void ArgumentCoder<WebCore::PointLightSource>::encode(StreamConnectionEncoder& encoder, const WebCore::PointLightSource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::FloatPoint3D>);
    encoder << instance.position();
}

std::optional<Ref<WebCore::PointLightSource>> ArgumentCoder<WebCore::PointLightSource>::decode(Decoder& decoder)
{
    auto position = decoder.decode<WebCore::FloatPoint3D>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PointLightSource::create(
            WTFMove(*position)
        )
    };
}

void ArgumentCoder<WebCore::SpotLightSource>::encode(Encoder& encoder, const WebCore::SpotLightSource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::FloatPoint3D>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction())>, WebCore::FloatPoint3D>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.specularExponent())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.limitingConeAngle())>, float>);
    encoder << instance.position();
    encoder << instance.direction();
    encoder << instance.specularExponent();
    encoder << instance.limitingConeAngle();
}

void ArgumentCoder<WebCore::SpotLightSource>::encode(StreamConnectionEncoder& encoder, const WebCore::SpotLightSource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::FloatPoint3D>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction())>, WebCore::FloatPoint3D>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.specularExponent())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.limitingConeAngle())>, float>);
    encoder << instance.position();
    encoder << instance.direction();
    encoder << instance.specularExponent();
    encoder << instance.limitingConeAngle();
}

std::optional<Ref<WebCore::SpotLightSource>> ArgumentCoder<WebCore::SpotLightSource>::decode(Decoder& decoder)
{
    auto position = decoder.decode<WebCore::FloatPoint3D>();
    auto direction = decoder.decode<WebCore::FloatPoint3D>();
    auto specularExponent = decoder.decode<float>();
    auto limitingConeAngle = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SpotLightSource::create(
            WTFMove(*position),
            WTFMove(*direction),
            WTFMove(*specularExponent),
            WTFMove(*limitingConeAngle)
        )
    };
}

void ArgumentCoder<WebCore::FEBlend>::encode(Encoder& encoder, const WebCore::FEBlend& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blendMode())>, WebCore::BlendMode>);
    encoder << instance.blendMode();
}

void ArgumentCoder<WebCore::FEBlend>::encode(StreamConnectionEncoder& encoder, const WebCore::FEBlend& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blendMode())>, WebCore::BlendMode>);
    encoder << instance.blendMode();
}

std::optional<Ref<WebCore::FEBlend>> ArgumentCoder<WebCore::FEBlend>::decode(Decoder& decoder)
{
    auto blendMode = decoder.decode<WebCore::BlendMode>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEBlend::create(
            WTFMove(*blendMode)
        )
    };
}

void ArgumentCoder<WebCore::FEColorMatrix>::encode(Encoder& encoder, const WebCore::FEColorMatrix& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::ColorMatrixType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.values())>, Vector<float>>);
    encoder << instance.type();
    encoder << instance.values();
}

void ArgumentCoder<WebCore::FEColorMatrix>::encode(StreamConnectionEncoder& encoder, const WebCore::FEColorMatrix& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::ColorMatrixType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.values())>, Vector<float>>);
    encoder << instance.type();
    encoder << instance.values();
}

std::optional<Ref<WebCore::FEColorMatrix>> ArgumentCoder<WebCore::FEColorMatrix>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::ColorMatrixType>();
    auto values = decoder.decode<Vector<float>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEColorMatrix::create(
            WTFMove(*type),
            WTFMove(*values)
        )
    };
}

void ArgumentCoder<WebCore::ComponentTransferFunction>::encode(Encoder& encoder, const WebCore::ComponentTransferFunction& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ComponentTransferType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.slope)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.intercept)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.amplitude)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.exponent)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tableValues)>, Vector<float>>);
    struct ShouldBeSameSizeAsComponentTransferFunction : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ComponentTransferFunction>, false> {
        WebCore::ComponentTransferType type;
        float slope;
        float intercept;
        float amplitude;
        float exponent;
        float offset;
        Vector<float> tableValues;
    };
    static_assert(sizeof(ShouldBeSameSizeAsComponentTransferFunction) == sizeof(WebCore::ComponentTransferFunction));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ComponentTransferFunction, type)
        , offsetof(WebCore::ComponentTransferFunction, slope)
        , offsetof(WebCore::ComponentTransferFunction, intercept)
        , offsetof(WebCore::ComponentTransferFunction, amplitude)
        , offsetof(WebCore::ComponentTransferFunction, exponent)
        , offsetof(WebCore::ComponentTransferFunction, offset)
        , offsetof(WebCore::ComponentTransferFunction, tableValues)
    >::value);
    encoder << instance.type;
    encoder << instance.slope;
    encoder << instance.intercept;
    encoder << instance.amplitude;
    encoder << instance.exponent;
    encoder << instance.offset;
    encoder << instance.tableValues;
}

void ArgumentCoder<WebCore::ComponentTransferFunction>::encode(StreamConnectionEncoder& encoder, const WebCore::ComponentTransferFunction& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ComponentTransferType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.slope)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.intercept)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.amplitude)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.exponent)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tableValues)>, Vector<float>>);
    struct ShouldBeSameSizeAsComponentTransferFunction : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ComponentTransferFunction>, false> {
        WebCore::ComponentTransferType type;
        float slope;
        float intercept;
        float amplitude;
        float exponent;
        float offset;
        Vector<float> tableValues;
    };
    static_assert(sizeof(ShouldBeSameSizeAsComponentTransferFunction) == sizeof(WebCore::ComponentTransferFunction));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ComponentTransferFunction, type)
        , offsetof(WebCore::ComponentTransferFunction, slope)
        , offsetof(WebCore::ComponentTransferFunction, intercept)
        , offsetof(WebCore::ComponentTransferFunction, amplitude)
        , offsetof(WebCore::ComponentTransferFunction, exponent)
        , offsetof(WebCore::ComponentTransferFunction, offset)
        , offsetof(WebCore::ComponentTransferFunction, tableValues)
    >::value);
    encoder << instance.type;
    encoder << instance.slope;
    encoder << instance.intercept;
    encoder << instance.amplitude;
    encoder << instance.exponent;
    encoder << instance.offset;
    encoder << instance.tableValues;
}

std::optional<WebCore::ComponentTransferFunction> ArgumentCoder<WebCore::ComponentTransferFunction>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::ComponentTransferType>();
    auto slope = decoder.decode<float>();
    auto intercept = decoder.decode<float>();
    auto amplitude = decoder.decode<float>();
    auto exponent = decoder.decode<float>();
    auto offset = decoder.decode<float>();
    auto tableValues = decoder.decode<Vector<float>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ComponentTransferFunction {
            WTFMove(*type),
            WTFMove(*slope),
            WTFMove(*intercept),
            WTFMove(*amplitude),
            WTFMove(*exponent),
            WTFMove(*offset),
            WTFMove(*tableValues)
        }
    };
}

void ArgumentCoder<WebCore::FEComponentTransfer>::encode(Encoder& encoder, const WebCore::FEComponentTransfer& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.redFunction())>, WebCore::ComponentTransferFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.greenFunction())>, WebCore::ComponentTransferFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blueFunction())>, WebCore::ComponentTransferFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaFunction())>, WebCore::ComponentTransferFunction>);
    encoder << instance.redFunction();
    encoder << instance.greenFunction();
    encoder << instance.blueFunction();
    encoder << instance.alphaFunction();
}

void ArgumentCoder<WebCore::FEComponentTransfer>::encode(StreamConnectionEncoder& encoder, const WebCore::FEComponentTransfer& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.redFunction())>, WebCore::ComponentTransferFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.greenFunction())>, WebCore::ComponentTransferFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blueFunction())>, WebCore::ComponentTransferFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaFunction())>, WebCore::ComponentTransferFunction>);
    encoder << instance.redFunction();
    encoder << instance.greenFunction();
    encoder << instance.blueFunction();
    encoder << instance.alphaFunction();
}

std::optional<Ref<WebCore::FEComponentTransfer>> ArgumentCoder<WebCore::FEComponentTransfer>::decode(Decoder& decoder)
{
    auto redFunction = decoder.decode<WebCore::ComponentTransferFunction>();
    auto greenFunction = decoder.decode<WebCore::ComponentTransferFunction>();
    auto blueFunction = decoder.decode<WebCore::ComponentTransferFunction>();
    auto alphaFunction = decoder.decode<WebCore::ComponentTransferFunction>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEComponentTransfer::create(
            WTFMove(*redFunction),
            WTFMove(*greenFunction),
            WTFMove(*blueFunction),
            WTFMove(*alphaFunction)
        )
    };
}

void ArgumentCoder<WebCore::FEComposite>::encode(Encoder& encoder, const WebCore::FEComposite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.operation())>, WebCore::CompositeOperationType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k1())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k2())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k3())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k4())>, float>);
    encoder << instance.operation();
    encoder << instance.k1();
    encoder << instance.k2();
    encoder << instance.k3();
    encoder << instance.k4();
}

void ArgumentCoder<WebCore::FEComposite>::encode(StreamConnectionEncoder& encoder, const WebCore::FEComposite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.operation())>, WebCore::CompositeOperationType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k1())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k2())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k3())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.k4())>, float>);
    encoder << instance.operation();
    encoder << instance.k1();
    encoder << instance.k2();
    encoder << instance.k3();
    encoder << instance.k4();
}

std::optional<Ref<WebCore::FEComposite>> ArgumentCoder<WebCore::FEComposite>::decode(Decoder& decoder)
{
    auto operation = decoder.decode<WebCore::CompositeOperationType>();
    auto k1 = decoder.decode<float>();
    auto k2 = decoder.decode<float>();
    auto k3 = decoder.decode<float>();
    auto k4 = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEComposite::create(
            WTFMove(*operation),
            WTFMove(*k1),
            WTFMove(*k2),
            WTFMove(*k3),
            WTFMove(*k4)
        )
    };
}

void ArgumentCoder<WebCore::FEConvolveMatrix>::encode(Encoder& encoder, const WebCore::FEConvolveMatrix& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelSize())>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.divisor())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bias())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.targetOffset())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.edgeMode())>, WebCore::EdgeModeType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLength())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preserveAlpha())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernel())>, Vector<float>>);
    encoder << instance.kernelSize();
    encoder << instance.divisor();
    encoder << instance.bias();
    encoder << instance.targetOffset();
    encoder << instance.edgeMode();
    encoder << instance.kernelUnitLength();
    encoder << instance.preserveAlpha();
    encoder << instance.kernel();
}

void ArgumentCoder<WebCore::FEConvolveMatrix>::encode(StreamConnectionEncoder& encoder, const WebCore::FEConvolveMatrix& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelSize())>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.divisor())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bias())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.targetOffset())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.edgeMode())>, WebCore::EdgeModeType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLength())>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preserveAlpha())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernel())>, Vector<float>>);
    encoder << instance.kernelSize();
    encoder << instance.divisor();
    encoder << instance.bias();
    encoder << instance.targetOffset();
    encoder << instance.edgeMode();
    encoder << instance.kernelUnitLength();
    encoder << instance.preserveAlpha();
    encoder << instance.kernel();
}

std::optional<Ref<WebCore::FEConvolveMatrix>> ArgumentCoder<WebCore::FEConvolveMatrix>::decode(Decoder& decoder)
{
    auto kernelSize = decoder.decode<WebCore::IntSize>();
    auto divisor = decoder.decode<float>();
    auto bias = decoder.decode<float>();
    auto targetOffset = decoder.decode<WebCore::IntPoint>();
    auto edgeMode = decoder.decode<WebCore::EdgeModeType>();
    auto kernelUnitLength = decoder.decode<WebCore::FloatPoint>();
    auto preserveAlpha = decoder.decode<bool>();
    auto kernel = decoder.decode<Vector<float>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEConvolveMatrix::create(
            WTFMove(*kernelSize),
            WTFMove(*divisor),
            WTFMove(*bias),
            WTFMove(*targetOffset),
            WTFMove(*edgeMode),
            WTFMove(*kernelUnitLength),
            WTFMove(*preserveAlpha),
            WTFMove(*kernel)
        )
    };
}

void ArgumentCoder<WebCore::FEDiffuseLighting>::encode(Encoder& encoder, const WebCore::FEDiffuseLighting& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightingColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.surfaceScale())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.diffuseConstant())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightSource())>, Ref<WebCore::LightSource>>);
    encoder << instance.lightingColor();
    encoder << instance.surfaceScale();
    encoder << instance.diffuseConstant();
    encoder << instance.kernelUnitLengthX();
    encoder << instance.kernelUnitLengthY();
    encoder << instance.lightSource();
}

void ArgumentCoder<WebCore::FEDiffuseLighting>::encode(StreamConnectionEncoder& encoder, const WebCore::FEDiffuseLighting& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightingColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.surfaceScale())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.diffuseConstant())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightSource())>, Ref<WebCore::LightSource>>);
    encoder << instance.lightingColor();
    encoder << instance.surfaceScale();
    encoder << instance.diffuseConstant();
    encoder << instance.kernelUnitLengthX();
    encoder << instance.kernelUnitLengthY();
    encoder << instance.lightSource();
}

std::optional<Ref<WebCore::FEDiffuseLighting>> ArgumentCoder<WebCore::FEDiffuseLighting>::decode(Decoder& decoder)
{
    auto lightingColor = decoder.decode<WebCore::Color>();
    auto surfaceScale = decoder.decode<float>();
    auto diffuseConstant = decoder.decode<float>();
    auto kernelUnitLengthX = decoder.decode<float>();
    auto kernelUnitLengthY = decoder.decode<float>();
    auto lightSource = decoder.decode<Ref<WebCore::LightSource>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEDiffuseLighting::create(
            WTFMove(*lightingColor),
            WTFMove(*surfaceScale),
            WTFMove(*diffuseConstant),
            WTFMove(*kernelUnitLengthX),
            WTFMove(*kernelUnitLengthY),
            WTFMove(*lightSource)
        )
    };
}

void ArgumentCoder<WebCore::FEDisplacementMap>::encode(Encoder& encoder, const WebCore::FEDisplacementMap& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.xChannelSelector())>, WebCore::ChannelSelectorType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.yChannelSelector())>, WebCore::ChannelSelectorType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scale())>, float>);
    encoder << instance.xChannelSelector();
    encoder << instance.yChannelSelector();
    encoder << instance.scale();
}

void ArgumentCoder<WebCore::FEDisplacementMap>::encode(StreamConnectionEncoder& encoder, const WebCore::FEDisplacementMap& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.xChannelSelector())>, WebCore::ChannelSelectorType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.yChannelSelector())>, WebCore::ChannelSelectorType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scale())>, float>);
    encoder << instance.xChannelSelector();
    encoder << instance.yChannelSelector();
    encoder << instance.scale();
}

std::optional<Ref<WebCore::FEDisplacementMap>> ArgumentCoder<WebCore::FEDisplacementMap>::decode(Decoder& decoder)
{
    auto xChannelSelector = decoder.decode<WebCore::ChannelSelectorType>();
    auto yChannelSelector = decoder.decode<WebCore::ChannelSelectorType>();
    auto scale = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEDisplacementMap::create(
            WTFMove(*xChannelSelector),
            WTFMove(*yChannelSelector),
            WTFMove(*scale)
        )
    };
}

void ArgumentCoder<WebCore::FEDropShadow>::encode(Encoder& encoder, const WebCore::FEDropShadow& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dx())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dy())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shadowColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shadowOpacity())>, float>);
    encoder << instance.stdDeviationX();
    encoder << instance.stdDeviationY();
    encoder << instance.dx();
    encoder << instance.dy();
    encoder << instance.shadowColor();
    encoder << instance.shadowOpacity();
}

void ArgumentCoder<WebCore::FEDropShadow>::encode(StreamConnectionEncoder& encoder, const WebCore::FEDropShadow& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dx())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dy())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shadowColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shadowOpacity())>, float>);
    encoder << instance.stdDeviationX();
    encoder << instance.stdDeviationY();
    encoder << instance.dx();
    encoder << instance.dy();
    encoder << instance.shadowColor();
    encoder << instance.shadowOpacity();
}

std::optional<Ref<WebCore::FEDropShadow>> ArgumentCoder<WebCore::FEDropShadow>::decode(Decoder& decoder)
{
    auto stdDeviationX = decoder.decode<float>();
    auto stdDeviationY = decoder.decode<float>();
    auto dx = decoder.decode<float>();
    auto dy = decoder.decode<float>();
    auto shadowColor = decoder.decode<WebCore::Color>();
    auto shadowOpacity = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEDropShadow::create(
            WTFMove(*stdDeviationX),
            WTFMove(*stdDeviationY),
            WTFMove(*dx),
            WTFMove(*dy),
            WTFMove(*shadowColor),
            WTFMove(*shadowOpacity)
        )
    };
}

void ArgumentCoder<WebCore::FEFlood>::encode(Encoder& encoder, const WebCore::FEFlood& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.floodColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.floodOpacity())>, float>);
    encoder << instance.floodColor();
    encoder << instance.floodOpacity();
}

void ArgumentCoder<WebCore::FEFlood>::encode(StreamConnectionEncoder& encoder, const WebCore::FEFlood& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.floodColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.floodOpacity())>, float>);
    encoder << instance.floodColor();
    encoder << instance.floodOpacity();
}

std::optional<Ref<WebCore::FEFlood>> ArgumentCoder<WebCore::FEFlood>::decode(Decoder& decoder)
{
    auto floodColor = decoder.decode<WebCore::Color>();
    auto floodOpacity = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEFlood::create(
            WTFMove(*floodColor),
            WTFMove(*floodOpacity)
        )
    };
}

void ArgumentCoder<WebCore::FEGaussianBlur>::encode(Encoder& encoder, const WebCore::FEGaussianBlur& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.edgeMode())>, WebCore::EdgeModeType>);
    encoder << instance.stdDeviationX();
    encoder << instance.stdDeviationY();
    encoder << instance.edgeMode();
}

void ArgumentCoder<WebCore::FEGaussianBlur>::encode(StreamConnectionEncoder& encoder, const WebCore::FEGaussianBlur& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stdDeviationY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.edgeMode())>, WebCore::EdgeModeType>);
    encoder << instance.stdDeviationX();
    encoder << instance.stdDeviationY();
    encoder << instance.edgeMode();
}

std::optional<Ref<WebCore::FEGaussianBlur>> ArgumentCoder<WebCore::FEGaussianBlur>::decode(Decoder& decoder)
{
    auto stdDeviationX = decoder.decode<float>();
    auto stdDeviationY = decoder.decode<float>();
    auto edgeMode = decoder.decode<WebCore::EdgeModeType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEGaussianBlur::create(
            WTFMove(*stdDeviationX),
            WTFMove(*stdDeviationY),
            WTFMove(*edgeMode)
        )
    };
}

void ArgumentCoder<WebCore::FEImage>::encode(Encoder& encoder, const WebCore::FEImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceImage())>, WebCore::SourceImage>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceImageRect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preserveAspectRatio())>, WebCore::SVGPreserveAspectRatioValue>);
    encoder << instance.sourceImage();
    encoder << instance.sourceImageRect();
    encoder << instance.preserveAspectRatio();
}

void ArgumentCoder<WebCore::FEImage>::encode(StreamConnectionEncoder& encoder, const WebCore::FEImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceImage())>, WebCore::SourceImage>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceImageRect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preserveAspectRatio())>, WebCore::SVGPreserveAspectRatioValue>);
    encoder << instance.sourceImage();
    encoder << instance.sourceImageRect();
    encoder << instance.preserveAspectRatio();
}

std::optional<Ref<WebCore::FEImage>> ArgumentCoder<WebCore::FEImage>::decode(Decoder& decoder)
{
    auto sourceImage = decoder.decode<WebCore::SourceImage>();
    auto sourceImageRect = decoder.decode<WebCore::FloatRect>();
    auto preserveAspectRatio = decoder.decode<WebCore::SVGPreserveAspectRatioValue>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEImage::create(
            WTFMove(*sourceImage),
            WTFMove(*sourceImageRect),
            WTFMove(*preserveAspectRatio)
        )
    };
}

void ArgumentCoder<WebCore::FEMerge>::encode(Encoder& encoder, const WebCore::FEMerge& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.numberOfEffectInputs())>, unsigned>);
    encoder << instance.numberOfEffectInputs();
}

void ArgumentCoder<WebCore::FEMerge>::encode(StreamConnectionEncoder& encoder, const WebCore::FEMerge& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.numberOfEffectInputs())>, unsigned>);
    encoder << instance.numberOfEffectInputs();
}

std::optional<Ref<WebCore::FEMerge>> ArgumentCoder<WebCore::FEMerge>::decode(Decoder& decoder)
{
    auto numberOfEffectInputs = decoder.decode<unsigned>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEMerge::create(
            WTFMove(*numberOfEffectInputs)
        )
    };
}

void ArgumentCoder<WebCore::FEMorphology>::encode(Encoder& encoder, const WebCore::FEMorphology& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.morphologyOperator())>, WebCore::MorphologyOperatorType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusY())>, float>);
    encoder << instance.morphologyOperator();
    encoder << instance.radiusX();
    encoder << instance.radiusY();
}

void ArgumentCoder<WebCore::FEMorphology>::encode(StreamConnectionEncoder& encoder, const WebCore::FEMorphology& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.morphologyOperator())>, WebCore::MorphologyOperatorType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusY())>, float>);
    encoder << instance.morphologyOperator();
    encoder << instance.radiusX();
    encoder << instance.radiusY();
}

std::optional<Ref<WebCore::FEMorphology>> ArgumentCoder<WebCore::FEMorphology>::decode(Decoder& decoder)
{
    auto morphologyOperator = decoder.decode<WebCore::MorphologyOperatorType>();
    auto radiusX = decoder.decode<float>();
    auto radiusY = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEMorphology::create(
            WTFMove(*morphologyOperator),
            WTFMove(*radiusX),
            WTFMove(*radiusY)
        )
    };
}

void ArgumentCoder<WebCore::FEOffset>::encode(Encoder& encoder, const WebCore::FEOffset& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dx())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dy())>, float>);
    encoder << instance.dx();
    encoder << instance.dy();
}

void ArgumentCoder<WebCore::FEOffset>::encode(StreamConnectionEncoder& encoder, const WebCore::FEOffset& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dx())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dy())>, float>);
    encoder << instance.dx();
    encoder << instance.dy();
}

std::optional<Ref<WebCore::FEOffset>> ArgumentCoder<WebCore::FEOffset>::decode(Decoder& decoder)
{
    auto dx = decoder.decode<float>();
    auto dy = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FEOffset::create(
            WTFMove(*dx),
            WTFMove(*dy)
        )
    };
}

void ArgumentCoder<WebCore::FESpecularLighting>::encode(Encoder& encoder, const WebCore::FESpecularLighting& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightingColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.surfaceScale())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.specularConstant())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.specularExponent())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightSource())>, Ref<WebCore::LightSource>>);
    encoder << instance.lightingColor();
    encoder << instance.surfaceScale();
    encoder << instance.specularConstant();
    encoder << instance.specularExponent();
    encoder << instance.kernelUnitLengthX();
    encoder << instance.kernelUnitLengthY();
    encoder << instance.lightSource();
}

void ArgumentCoder<WebCore::FESpecularLighting>::encode(StreamConnectionEncoder& encoder, const WebCore::FESpecularLighting& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightingColor())>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.surfaceScale())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.specularConstant())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.specularExponent())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.kernelUnitLengthY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lightSource())>, Ref<WebCore::LightSource>>);
    encoder << instance.lightingColor();
    encoder << instance.surfaceScale();
    encoder << instance.specularConstant();
    encoder << instance.specularExponent();
    encoder << instance.kernelUnitLengthX();
    encoder << instance.kernelUnitLengthY();
    encoder << instance.lightSource();
}

std::optional<Ref<WebCore::FESpecularLighting>> ArgumentCoder<WebCore::FESpecularLighting>::decode(Decoder& decoder)
{
    auto lightingColor = decoder.decode<WebCore::Color>();
    auto surfaceScale = decoder.decode<float>();
    auto specularConstant = decoder.decode<float>();
    auto specularExponent = decoder.decode<float>();
    auto kernelUnitLengthX = decoder.decode<float>();
    auto kernelUnitLengthY = decoder.decode<float>();
    auto lightSource = decoder.decode<Ref<WebCore::LightSource>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FESpecularLighting::create(
            WTFMove(*lightingColor),
            WTFMove(*surfaceScale),
            WTFMove(*specularConstant),
            WTFMove(*specularExponent),
            WTFMove(*kernelUnitLengthX),
            WTFMove(*kernelUnitLengthY),
            WTFMove(*lightSource)
        )
    };
}

void ArgumentCoder<WebCore::FETurbulence>::encode(Encoder& encoder, const WebCore::FETurbulence& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::TurbulenceType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseFrequencyX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseFrequencyY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.numOctaves())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.seed())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stitchTiles())>, bool>);
    encoder << instance.type();
    encoder << instance.baseFrequencyX();
    encoder << instance.baseFrequencyY();
    encoder << instance.numOctaves();
    encoder << instance.seed();
    encoder << instance.stitchTiles();
}

void ArgumentCoder<WebCore::FETurbulence>::encode(StreamConnectionEncoder& encoder, const WebCore::FETurbulence& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::TurbulenceType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseFrequencyX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseFrequencyY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.numOctaves())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.seed())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stitchTiles())>, bool>);
    encoder << instance.type();
    encoder << instance.baseFrequencyX();
    encoder << instance.baseFrequencyY();
    encoder << instance.numOctaves();
    encoder << instance.seed();
    encoder << instance.stitchTiles();
}

std::optional<Ref<WebCore::FETurbulence>> ArgumentCoder<WebCore::FETurbulence>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::TurbulenceType>();
    auto baseFrequencyX = decoder.decode<float>();
    auto baseFrequencyY = decoder.decode<float>();
    auto numOctaves = decoder.decode<int>();
    auto seed = decoder.decode<float>();
    auto stitchTiles = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FETurbulence::create(
            WTFMove(*type),
            WTFMove(*baseFrequencyX),
            WTFMove(*baseFrequencyY),
            WTFMove(*numOctaves),
            WTFMove(*seed),
            WTFMove(*stitchTiles)
        )
    };
}

void ArgumentCoder<WebCore::FilterEffectGeometry>::encode(Encoder& encoder, const WebCore::FilterEffectGeometry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_boundaries)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_flags)>, OptionSet<WebCore::FilterEffectGeometry::Flags>>);
    struct ShouldBeSameSizeAsFilterEffectGeometry : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FilterEffectGeometry>, false> {
        WebCore::FloatRect m_boundaries;
        OptionSet<WebCore::FilterEffectGeometry::Flags> m_flags;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFilterEffectGeometry) == sizeof(WebCore::FilterEffectGeometry));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FilterEffectGeometry, m_boundaries)
        , offsetof(WebCore::FilterEffectGeometry, m_flags)
    >::value);
    encoder << instance.m_boundaries;
    encoder << instance.m_flags;
}

void ArgumentCoder<WebCore::FilterEffectGeometry>::encode(StreamConnectionEncoder& encoder, const WebCore::FilterEffectGeometry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_boundaries)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_flags)>, OptionSet<WebCore::FilterEffectGeometry::Flags>>);
    struct ShouldBeSameSizeAsFilterEffectGeometry : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FilterEffectGeometry>, false> {
        WebCore::FloatRect m_boundaries;
        OptionSet<WebCore::FilterEffectGeometry::Flags> m_flags;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFilterEffectGeometry) == sizeof(WebCore::FilterEffectGeometry));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FilterEffectGeometry, m_boundaries)
        , offsetof(WebCore::FilterEffectGeometry, m_flags)
    >::value);
    encoder << instance.m_boundaries;
    encoder << instance.m_flags;
}

std::optional<WebCore::FilterEffectGeometry> ArgumentCoder<WebCore::FilterEffectGeometry>::decode(Decoder& decoder)
{
    auto m_boundaries = decoder.decode<WebCore::FloatRect>();
    auto m_flags = decoder.decode<OptionSet<WebCore::FilterEffectGeometry::Flags>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FilterEffectGeometry {
            WTFMove(*m_boundaries),
            WTFMove(*m_flags)
        }
    };
}

void ArgumentCoder<WebCore::SVGFilterExpressionTerm>::encode(Encoder& encoder, const WebCore::SVGFilterExpressionTerm& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.index)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.level)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.geometry)>, std::optional<WebCore::FilterEffectGeometry>>);
    struct ShouldBeSameSizeAsSVGFilterExpressionTerm : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SVGFilterExpressionTerm>, false> {
        unsigned index;
        unsigned level;
        std::optional<WebCore::FilterEffectGeometry> geometry;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSVGFilterExpressionTerm) == sizeof(WebCore::SVGFilterExpressionTerm));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SVGFilterExpressionTerm, index)
        , offsetof(WebCore::SVGFilterExpressionTerm, level)
        , offsetof(WebCore::SVGFilterExpressionTerm, geometry)
    >::value);
    encoder << instance.index;
    encoder << instance.level;
    encoder << instance.geometry;
}

void ArgumentCoder<WebCore::SVGFilterExpressionTerm>::encode(StreamConnectionEncoder& encoder, const WebCore::SVGFilterExpressionTerm& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.index)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.level)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.geometry)>, std::optional<WebCore::FilterEffectGeometry>>);
    struct ShouldBeSameSizeAsSVGFilterExpressionTerm : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SVGFilterExpressionTerm>, false> {
        unsigned index;
        unsigned level;
        std::optional<WebCore::FilterEffectGeometry> geometry;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSVGFilterExpressionTerm) == sizeof(WebCore::SVGFilterExpressionTerm));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SVGFilterExpressionTerm, index)
        , offsetof(WebCore::SVGFilterExpressionTerm, level)
        , offsetof(WebCore::SVGFilterExpressionTerm, geometry)
    >::value);
    encoder << instance.index;
    encoder << instance.level;
    encoder << instance.geometry;
}

std::optional<WebCore::SVGFilterExpressionTerm> ArgumentCoder<WebCore::SVGFilterExpressionTerm>::decode(Decoder& decoder)
{
    auto index = decoder.decode<unsigned>();
    auto level = decoder.decode<unsigned>();
    auto geometry = decoder.decode<std::optional<WebCore::FilterEffectGeometry>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SVGFilterExpressionTerm {
            WTFMove(*index),
            WTFMove(*level),
            WTFMove(*geometry)
        }
    };
}

void ArgumentCoder<WebCore::SVGPreserveAspectRatioValue>::encode(Encoder& encoder, const WebCore::SVGPreserveAspectRatioValue& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_align)>, WebCore::SVGPreserveAspectRatioValue::SVGPreserveAspectRatioType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_meetOrSlice)>, WebCore::SVGPreserveAspectRatioValue::SVGMeetOrSliceType>);
    struct ShouldBeSameSizeAsSVGPreserveAspectRatioValue : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SVGPreserveAspectRatioValue>, false> {
        WebCore::SVGPreserveAspectRatioValue::SVGPreserveAspectRatioType m_align;
        WebCore::SVGPreserveAspectRatioValue::SVGMeetOrSliceType m_meetOrSlice;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSVGPreserveAspectRatioValue) == sizeof(WebCore::SVGPreserveAspectRatioValue));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SVGPreserveAspectRatioValue, m_align)
        , offsetof(WebCore::SVGPreserveAspectRatioValue, m_meetOrSlice)
    >::value);
    encoder << instance.m_align;
    encoder << instance.m_meetOrSlice;
}

void ArgumentCoder<WebCore::SVGPreserveAspectRatioValue>::encode(StreamConnectionEncoder& encoder, const WebCore::SVGPreserveAspectRatioValue& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_align)>, WebCore::SVGPreserveAspectRatioValue::SVGPreserveAspectRatioType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_meetOrSlice)>, WebCore::SVGPreserveAspectRatioValue::SVGMeetOrSliceType>);
    struct ShouldBeSameSizeAsSVGPreserveAspectRatioValue : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SVGPreserveAspectRatioValue>, false> {
        WebCore::SVGPreserveAspectRatioValue::SVGPreserveAspectRatioType m_align;
        WebCore::SVGPreserveAspectRatioValue::SVGMeetOrSliceType m_meetOrSlice;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSVGPreserveAspectRatioValue) == sizeof(WebCore::SVGPreserveAspectRatioValue));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SVGPreserveAspectRatioValue, m_align)
        , offsetof(WebCore::SVGPreserveAspectRatioValue, m_meetOrSlice)
    >::value);
    encoder << instance.m_align;
    encoder << instance.m_meetOrSlice;
}

std::optional<WebCore::SVGPreserveAspectRatioValue> ArgumentCoder<WebCore::SVGPreserveAspectRatioValue>::decode(Decoder& decoder)
{
    auto m_align = decoder.decode<WebCore::SVGPreserveAspectRatioValue::SVGPreserveAspectRatioType>();
    auto m_meetOrSlice = decoder.decode<WebCore::SVGPreserveAspectRatioValue::SVGMeetOrSliceType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SVGPreserveAspectRatioValue {
            WTFMove(*m_align),
            WTFMove(*m_meetOrSlice)
        }
    };
}

#if ENABLE(GAMEPAD)
void ArgumentCoder<WebCore::GamepadEffectParameters>::encode(Encoder& encoder, const WebCore::GamepadEffectParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startDelay)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.strongMagnitude)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.weakMagnitude)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.leftTrigger)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rightTrigger)>, double>);
    struct ShouldBeSameSizeAsGamepadEffectParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GamepadEffectParameters>, false> {
        double duration;
        double startDelay;
        double strongMagnitude;
        double weakMagnitude;
        double leftTrigger;
        double rightTrigger;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGamepadEffectParameters) == sizeof(WebCore::GamepadEffectParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GamepadEffectParameters, duration)
        , offsetof(WebCore::GamepadEffectParameters, startDelay)
        , offsetof(WebCore::GamepadEffectParameters, strongMagnitude)
        , offsetof(WebCore::GamepadEffectParameters, weakMagnitude)
        , offsetof(WebCore::GamepadEffectParameters, leftTrigger)
        , offsetof(WebCore::GamepadEffectParameters, rightTrigger)
    >::value);
    encoder << instance.duration;
    encoder << instance.startDelay;
    encoder << instance.strongMagnitude;
    encoder << instance.weakMagnitude;
    encoder << instance.leftTrigger;
    encoder << instance.rightTrigger;
}

std::optional<WebCore::GamepadEffectParameters> ArgumentCoder<WebCore::GamepadEffectParameters>::decode(Decoder& decoder)
{
    auto duration = decoder.decode<double>();
    auto startDelay = decoder.decode<double>();
    auto strongMagnitude = decoder.decode<double>();
    auto weakMagnitude = decoder.decode<double>();
    auto leftTrigger = decoder.decode<double>();
    auto rightTrigger = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GamepadEffectParameters {
            WTFMove(*duration),
            WTFMove(*startDelay),
            WTFMove(*strongMagnitude),
            WTFMove(*weakMagnitude),
            WTFMove(*leftTrigger),
            WTFMove(*rightTrigger)
        }
    };
}

#endif

void ArgumentCoder<WebCore::StorageEstimate>::encode(Encoder& encoder, const WebCore::StorageEstimate& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usage)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.quota)>, uint64_t>);
    struct ShouldBeSameSizeAsStorageEstimate : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::StorageEstimate>, false> {
        uint64_t usage;
        uint64_t quota;
    };
    static_assert(sizeof(ShouldBeSameSizeAsStorageEstimate) == sizeof(WebCore::StorageEstimate));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::StorageEstimate, usage)
        , offsetof(WebCore::StorageEstimate, quota)
    >::value);
    encoder << instance.usage;
    encoder << instance.quota;
}

std::optional<WebCore::StorageEstimate> ArgumentCoder<WebCore::StorageEstimate>::decode(Decoder& decoder)
{
    auto usage = decoder.decode<uint64_t>();
    auto quota = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::StorageEstimate {
            WTFMove(*usage),
            WTFMove(*quota)
        }
    };
}

void ArgumentCoder<WebCore::ListStyleType>::encode(Encoder& encoder, const WebCore::ListStyleType& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ListStyleType::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, AtomString>);
    struct ShouldBeSameSizeAsListStyleType : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ListStyleType>, false> {
        WebCore::ListStyleType::Type type;
        AtomString identifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsListStyleType) == sizeof(WebCore::ListStyleType));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ListStyleType, type)
        , offsetof(WebCore::ListStyleType, identifier)
    >::value);
    encoder << instance.type;
    encoder << instance.identifier;
}

std::optional<WebCore::ListStyleType> ArgumentCoder<WebCore::ListStyleType>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::ListStyleType::Type>();
    auto identifier = decoder.decode<AtomString>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ListStyleType {
            WTFMove(*type),
            WTFMove(*identifier)
        }
    };
}

void ArgumentCoder<WebCore::TextList>::encode(Encoder& encoder, const WebCore::TextList& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.styleType)>, WebCore::ListStyleType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startingItemNumber)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ordered)>, bool>);
    struct ShouldBeSameSizeAsTextList : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextList>, false> {
        WebCore::ListStyleType styleType;
        int startingItemNumber;
        bool ordered;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextList) == sizeof(WebCore::TextList));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextList, styleType)
        , offsetof(WebCore::TextList, startingItemNumber)
        , offsetof(WebCore::TextList, ordered)
    >::value);
    encoder << instance.styleType;
    encoder << instance.startingItemNumber;
    encoder << instance.ordered;
}

std::optional<WebCore::TextList> ArgumentCoder<WebCore::TextList>::decode(Decoder& decoder)
{
    auto styleType = decoder.decode<WebCore::ListStyleType>();
    auto startingItemNumber = decoder.decode<int>();
    auto ordered = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextList {
            WTFMove(*styleType),
            WTFMove(*startingItemNumber),
            WTFMove(*ordered)
        }
    };
}

void ArgumentCoder<WebCore::FontAttributes>::encode(Encoder& encoder, const WebCore::FontAttributes& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.font)>, RefPtr<WebCore::Font>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.backgroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.foregroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontShadow)>, WebCore::FontShadow>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subscriptOrSuperscript)>, WebCore::FontAttributes::SubscriptOrSuperscript>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.horizontalAlignment)>, WebCore::FontAttributes::HorizontalAlignment>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textLists)>, Vector<WebCore::TextList>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasUnderline)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasStrikeThrough)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasMultipleFonts)>, bool>);
    struct ShouldBeSameSizeAsFontAttributes : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FontAttributes>, false> {
        RefPtr<WebCore::Font> font;
        WebCore::Color backgroundColor;
        WebCore::Color foregroundColor;
        WebCore::FontShadow fontShadow;
        WebCore::FontAttributes::SubscriptOrSuperscript subscriptOrSuperscript;
        WebCore::FontAttributes::HorizontalAlignment horizontalAlignment;
        Vector<WebCore::TextList> textLists;
        bool hasUnderline;
        bool hasStrikeThrough;
        bool hasMultipleFonts;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFontAttributes) == sizeof(WebCore::FontAttributes));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FontAttributes, font)
        , offsetof(WebCore::FontAttributes, backgroundColor)
        , offsetof(WebCore::FontAttributes, foregroundColor)
        , offsetof(WebCore::FontAttributes, fontShadow)
        , offsetof(WebCore::FontAttributes, subscriptOrSuperscript)
        , offsetof(WebCore::FontAttributes, horizontalAlignment)
        , offsetof(WebCore::FontAttributes, textLists)
        , offsetof(WebCore::FontAttributes, hasUnderline)
        , offsetof(WebCore::FontAttributes, hasStrikeThrough)
        , offsetof(WebCore::FontAttributes, hasMultipleFonts)
    >::value);
    encoder << instance.font;
    encoder << instance.backgroundColor;
    encoder << instance.foregroundColor;
    encoder << instance.fontShadow;
    encoder << instance.subscriptOrSuperscript;
    encoder << instance.horizontalAlignment;
    encoder << instance.textLists;
    encoder << instance.hasUnderline;
    encoder << instance.hasStrikeThrough;
    encoder << instance.hasMultipleFonts;
}

std::optional<WebCore::FontAttributes> ArgumentCoder<WebCore::FontAttributes>::decode(Decoder& decoder)
{
    auto font = decoder.decode<RefPtr<WebCore::Font>>();
    auto backgroundColor = decoder.decode<WebCore::Color>();
    auto foregroundColor = decoder.decode<WebCore::Color>();
    auto fontShadow = decoder.decode<WebCore::FontShadow>();
    auto subscriptOrSuperscript = decoder.decode<WebCore::FontAttributes::SubscriptOrSuperscript>();
    auto horizontalAlignment = decoder.decode<WebCore::FontAttributes::HorizontalAlignment>();
    auto textLists = decoder.decode<Vector<WebCore::TextList>>();
    auto hasUnderline = decoder.decode<bool>();
    auto hasStrikeThrough = decoder.decode<bool>();
    auto hasMultipleFonts = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FontAttributes {
            WTFMove(*font),
            WTFMove(*backgroundColor),
            WTFMove(*foregroundColor),
            WTFMove(*fontShadow),
            WTFMove(*subscriptOrSuperscript),
            WTFMove(*horizontalAlignment),
            WTFMove(*textLists),
            WTFMove(*hasUnderline),
            WTFMove(*hasStrikeThrough),
            WTFMove(*hasMultipleFonts)
        }
    };
}

void ArgumentCoder<WebCore::CrossOriginEmbedderPolicy>::encode(Encoder& encoder, const WebCore::CrossOriginEmbedderPolicy& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, WebCore::CrossOriginEmbedderPolicyValue>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.reportOnlyValue)>, WebCore::CrossOriginEmbedderPolicyValue>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.reportingEndpoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.reportOnlyReportingEndpoint)>, String>);
    struct ShouldBeSameSizeAsCrossOriginEmbedderPolicy : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CrossOriginEmbedderPolicy>, false> {
        WebCore::CrossOriginEmbedderPolicyValue value;
        WebCore::CrossOriginEmbedderPolicyValue reportOnlyValue;
        String reportingEndpoint;
        String reportOnlyReportingEndpoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCrossOriginEmbedderPolicy) == sizeof(WebCore::CrossOriginEmbedderPolicy));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CrossOriginEmbedderPolicy, value)
        , offsetof(WebCore::CrossOriginEmbedderPolicy, reportOnlyValue)
        , offsetof(WebCore::CrossOriginEmbedderPolicy, reportingEndpoint)
        , offsetof(WebCore::CrossOriginEmbedderPolicy, reportOnlyReportingEndpoint)
    >::value);
    encoder << instance.value;
    encoder << instance.reportOnlyValue;
    encoder << instance.reportingEndpoint;
    encoder << instance.reportOnlyReportingEndpoint;
}

std::optional<WebCore::CrossOriginEmbedderPolicy> ArgumentCoder<WebCore::CrossOriginEmbedderPolicy>::decode(Decoder& decoder)
{
    auto value = decoder.decode<WebCore::CrossOriginEmbedderPolicyValue>();
    auto reportOnlyValue = decoder.decode<WebCore::CrossOriginEmbedderPolicyValue>();
    auto reportingEndpoint = decoder.decode<String>();
    auto reportOnlyReportingEndpoint = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CrossOriginEmbedderPolicy {
            WTFMove(*value),
            WTFMove(*reportOnlyValue),
            WTFMove(*reportingEndpoint),
            WTFMove(*reportOnlyReportingEndpoint)
        }
    };
}

void ArgumentCoder<WebCore::SameSiteInfo>::encode(Encoder& encoder, const WebCore::SameSiteInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSameSite)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isTopSite)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSafeHTTPMethod)>, bool>);
    struct ShouldBeSameSizeAsSameSiteInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SameSiteInfo>, false> {
        bool isSameSite;
        bool isTopSite;
        bool isSafeHTTPMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSameSiteInfo) == sizeof(WebCore::SameSiteInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SameSiteInfo, isSameSite)
        , offsetof(WebCore::SameSiteInfo, isTopSite)
        , offsetof(WebCore::SameSiteInfo, isSafeHTTPMethod)
    >::value);
    encoder << instance.isSameSite;
    encoder << instance.isTopSite;
    encoder << instance.isSafeHTTPMethod;
}

std::optional<WebCore::SameSiteInfo> ArgumentCoder<WebCore::SameSiteInfo>::decode(Decoder& decoder)
{
    auto isSameSite = decoder.decode<bool>();
    auto isTopSite = decoder.decode<bool>();
    auto isSafeHTTPMethod = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SameSiteInfo {
            WTFMove(*isSameSite),
            WTFMove(*isTopSite),
            WTFMove(*isSafeHTTPMethod)
        }
    };
}

void ArgumentCoder<WebCore::SecurityOrigin>::encode(Encoder& encoder, const WebCore::SecurityOrigin& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_data)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_domain)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_filePath)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_universalAccess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_domainWasSetInDOM)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_canLoadLocalResources)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_enforcesFilePathSeparation)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_needsStorageAccessFromFileURLsQuirk)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isPotentiallyTrustworthy)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isLocal)>, bool>);
    struct ShouldBeSameSizeAsSecurityOrigin : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SecurityOrigin>, true> {
        WebCore::SecurityOriginData m_data;
        String m_domain;
        String m_filePath;
        bool m_universalAccess;
        bool m_domainWasSetInDOM;
        bool m_canLoadLocalResources;
        bool m_enforcesFilePathSeparation;
        bool m_needsStorageAccessFromFileURLsQuirk;
        std::optional<bool> m_isPotentiallyTrustworthy;
        bool m_isLocal;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSecurityOrigin) == sizeof(WebCore::SecurityOrigin));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SecurityOrigin, m_data)
        , offsetof(WebCore::SecurityOrigin, m_domain)
        , offsetof(WebCore::SecurityOrigin, m_filePath)
        , offsetof(WebCore::SecurityOrigin, m_universalAccess)
        , offsetof(WebCore::SecurityOrigin, m_domainWasSetInDOM)
        , offsetof(WebCore::SecurityOrigin, m_canLoadLocalResources)
        , offsetof(WebCore::SecurityOrigin, m_enforcesFilePathSeparation)
        , offsetof(WebCore::SecurityOrigin, m_needsStorageAccessFromFileURLsQuirk)
        , offsetof(WebCore::SecurityOrigin, m_isPotentiallyTrustworthy)
        , offsetof(WebCore::SecurityOrigin, m_isLocal)
    >::value);
    encoder << instance.m_data;
    encoder << instance.m_domain;
    encoder << instance.m_filePath;
    encoder << instance.m_universalAccess;
    encoder << instance.m_domainWasSetInDOM;
    encoder << instance.m_canLoadLocalResources;
    encoder << instance.m_enforcesFilePathSeparation;
    encoder << instance.m_needsStorageAccessFromFileURLsQuirk;
    encoder << instance.m_isPotentiallyTrustworthy;
    encoder << instance.m_isLocal;
}

std::optional<Ref<WebCore::SecurityOrigin>> ArgumentCoder<WebCore::SecurityOrigin>::decode(Decoder& decoder)
{
    auto m_data = decoder.decode<WebCore::SecurityOriginData>();
    auto m_domain = decoder.decode<String>();
    auto m_filePath = decoder.decode<String>();
    auto m_universalAccess = decoder.decode<bool>();
    auto m_domainWasSetInDOM = decoder.decode<bool>();
    auto m_canLoadLocalResources = decoder.decode<bool>();
    auto m_enforcesFilePathSeparation = decoder.decode<bool>();
    auto m_needsStorageAccessFromFileURLsQuirk = decoder.decode<bool>();
    auto m_isPotentiallyTrustworthy = decoder.decode<std::optional<bool>>();
    auto m_isLocal = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SecurityOrigin::create(
            WTFMove(*m_data),
            WTFMove(*m_domain),
            WTFMove(*m_filePath),
            WTFMove(*m_universalAccess),
            WTFMove(*m_domainWasSetInDOM),
            WTFMove(*m_canLoadLocalResources),
            WTFMove(*m_enforcesFilePathSeparation),
            WTFMove(*m_needsStorageAccessFromFileURLsQuirk),
            WTFMove(*m_isPotentiallyTrustworthy),
            WTFMove(*m_isLocal)
        )
    };
}

void ArgumentCoder<WebCore::CookieRequestHeaderFieldProxy>::encode(Encoder& encoder, const WebCore::CookieRequestHeaderFieldProxy& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.firstParty)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sameSiteInfo)>, WebCore::SameSiteInfo>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, std::optional<WebCore::FrameIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageID)>, std::optional<WebCore::PageIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.includeSecureCookies)>, WebCore::IncludeSecureCookies>);
    struct ShouldBeSameSizeAsCookieRequestHeaderFieldProxy : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CookieRequestHeaderFieldProxy>, false> {
        URL firstParty;
        WebCore::SameSiteInfo sameSiteInfo;
        URL url;
        std::optional<WebCore::FrameIdentifier> frameID;
        std::optional<WebCore::PageIdentifier> pageID;
        WebCore::IncludeSecureCookies includeSecureCookies;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCookieRequestHeaderFieldProxy) == sizeof(WebCore::CookieRequestHeaderFieldProxy));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CookieRequestHeaderFieldProxy, firstParty)
        , offsetof(WebCore::CookieRequestHeaderFieldProxy, sameSiteInfo)
        , offsetof(WebCore::CookieRequestHeaderFieldProxy, url)
        , offsetof(WebCore::CookieRequestHeaderFieldProxy, frameID)
        , offsetof(WebCore::CookieRequestHeaderFieldProxy, pageID)
        , offsetof(WebCore::CookieRequestHeaderFieldProxy, includeSecureCookies)
    >::value);
    encoder << instance.firstParty;
    encoder << instance.sameSiteInfo;
    encoder << instance.url;
    encoder << instance.frameID;
    encoder << instance.pageID;
    encoder << instance.includeSecureCookies;
}

std::optional<WebCore::CookieRequestHeaderFieldProxy> ArgumentCoder<WebCore::CookieRequestHeaderFieldProxy>::decode(Decoder& decoder)
{
    auto firstParty = decoder.decode<URL>();
    auto sameSiteInfo = decoder.decode<WebCore::SameSiteInfo>();
    auto url = decoder.decode<URL>();
    auto frameID = decoder.decode<std::optional<WebCore::FrameIdentifier>>();
    auto pageID = decoder.decode<std::optional<WebCore::PageIdentifier>>();
    auto includeSecureCookies = decoder.decode<WebCore::IncludeSecureCookies>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CookieRequestHeaderFieldProxy {
            WTFMove(*firstParty),
            WTFMove(*sameSiteInfo),
            WTFMove(*url),
            WTFMove(*frameID),
            WTFMove(*pageID),
            WTFMove(*includeSecureCookies)
        }
    };
}

void ArgumentCoder<WebCore::FormData>::encode(Encoder& encoder, const WebCore::FormData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_elements)>, Vector<WebCore::FormDataElement>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_identifier)>, int64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_alwaysStream)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_boundary)>, Vector<char>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_lengthInBytes)>, std::optional<uint64_t>>);
    struct ShouldBeSameSizeAsFormData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FormData>, true> {
        Vector<WebCore::FormDataElement> m_elements;
        int64_t m_identifier;
        bool m_alwaysStream;
        Vector<char> m_boundary;
        std::optional<uint64_t> m_lengthInBytes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFormData) == sizeof(WebCore::FormData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FormData, m_elements)
        , offsetof(WebCore::FormData, m_identifier)
        , offsetof(WebCore::FormData, m_alwaysStream)
        , offsetof(WebCore::FormData, m_boundary)
        , offsetof(WebCore::FormData, m_lengthInBytes)
    >::value);
    encoder << instance.m_elements;
    encoder << instance.m_identifier;
    encoder << instance.m_alwaysStream;
    encoder << instance.m_boundary;
}

std::optional<Ref<WebCore::FormData>> ArgumentCoder<WebCore::FormData>::decode(Decoder& decoder)
{
    auto m_elements = decoder.decode<Vector<WebCore::FormDataElement>>();
    auto m_identifier = decoder.decode<int64_t>();
    auto m_alwaysStream = decoder.decode<bool>();
    auto m_boundary = decoder.decode<Vector<char>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FormData::create(
            WTFMove(*m_elements),
            WTFMove(*m_identifier),
            WTFMove(*m_alwaysStream),
            WTFMove(*m_boundary)
        )
    };
}

void ArgumentCoder<WebCore::RequestedScrollData>::encode(Encoder& encoder, const WebCore::RequestedScrollData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requestType)>, WebCore::ScrollRequestType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollPositionOrDelta)>, std::variant<WebCore::FloatPoint, WebCore::FloatSize>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollType)>, WebCore::ScrollType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clamping)>, WebCore::ScrollClamping>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.animated)>, WebCore::ScrollIsAnimated>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requestedDataBeforeAnimatedScroll)>, std::optional<std::tuple<WebCore::ScrollRequestType, std::variant<WebCore::FloatPoint, WebCore::FloatSize>, WebCore::ScrollType, WebCore::ScrollClamping>>>);
    struct ShouldBeSameSizeAsRequestedScrollData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RequestedScrollData>, false> {
        WebCore::ScrollRequestType requestType;
        std::variant<WebCore::FloatPoint, WebCore::FloatSize> scrollPositionOrDelta;
        WebCore::ScrollType scrollType;
        WebCore::ScrollClamping clamping;
        WebCore::ScrollIsAnimated animated;
        std::optional<std::tuple<WebCore::ScrollRequestType, std::variant<WebCore::FloatPoint, WebCore::FloatSize>, WebCore::ScrollType, WebCore::ScrollClamping>> requestedDataBeforeAnimatedScroll;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRequestedScrollData) == sizeof(WebCore::RequestedScrollData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RequestedScrollData, requestType)
        , offsetof(WebCore::RequestedScrollData, scrollPositionOrDelta)
        , offsetof(WebCore::RequestedScrollData, scrollType)
        , offsetof(WebCore::RequestedScrollData, clamping)
        , offsetof(WebCore::RequestedScrollData, animated)
        , offsetof(WebCore::RequestedScrollData, requestedDataBeforeAnimatedScroll)
    >::value);
    encoder << instance.requestType;
    encoder << instance.scrollPositionOrDelta;
    encoder << instance.scrollType;
    encoder << instance.clamping;
    encoder << instance.animated;
    encoder << instance.requestedDataBeforeAnimatedScroll;
}

std::optional<WebCore::RequestedScrollData> ArgumentCoder<WebCore::RequestedScrollData>::decode(Decoder& decoder)
{
    auto requestType = decoder.decode<WebCore::ScrollRequestType>();
    auto scrollPositionOrDelta = decoder.decode<std::variant<WebCore::FloatPoint, WebCore::FloatSize>>();
    auto scrollType = decoder.decode<WebCore::ScrollType>();
    auto clamping = decoder.decode<WebCore::ScrollClamping>();
    auto animated = decoder.decode<WebCore::ScrollIsAnimated>();
    auto requestedDataBeforeAnimatedScroll = decoder.decode<std::optional<std::tuple<WebCore::ScrollRequestType, std::variant<WebCore::FloatPoint, WebCore::FloatSize>, WebCore::ScrollType, WebCore::ScrollClamping>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RequestedScrollData {
            WTFMove(*requestType),
            WTFMove(*scrollPositionOrDelta),
            WTFMove(*scrollType),
            WTFMove(*clamping),
            WTFMove(*animated),
            WTFMove(*requestedDataBeforeAnimatedScroll)
        }
    };
}

void ArgumentCoder<WebCore::FloatScrollSnapOffsetsInfo>::encode(Encoder& encoder, const WebCore::FloatScrollSnapOffsetsInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.strictness)>, WebCore::ScrollSnapStrictness>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.horizontalSnapOffsets)>, Vector<WebCore::FloatSnapOffset>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.verticalSnapOffsets)>, Vector<WebCore::FloatSnapOffset>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.snapAreas)>, Vector<WebCore::FloatRect>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.snapAreasIDs)>, Vector<WebCore::ElementIdentifier>>);
    struct ShouldBeSameSizeAsFloatScrollSnapOffsetsInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FloatScrollSnapOffsetsInfo>, false> {
        WebCore::ScrollSnapStrictness strictness;
        Vector<WebCore::FloatSnapOffset> horizontalSnapOffsets;
        Vector<WebCore::FloatSnapOffset> verticalSnapOffsets;
        Vector<WebCore::FloatRect> snapAreas;
        Vector<WebCore::ElementIdentifier> snapAreasIDs;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFloatScrollSnapOffsetsInfo) == sizeof(WebCore::FloatScrollSnapOffsetsInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FloatScrollSnapOffsetsInfo, strictness)
        , offsetof(WebCore::FloatScrollSnapOffsetsInfo, horizontalSnapOffsets)
        , offsetof(WebCore::FloatScrollSnapOffsetsInfo, verticalSnapOffsets)
        , offsetof(WebCore::FloatScrollSnapOffsetsInfo, snapAreas)
        , offsetof(WebCore::FloatScrollSnapOffsetsInfo, snapAreasIDs)
    >::value);
    encoder << instance.strictness;
    encoder << instance.horizontalSnapOffsets;
    encoder << instance.verticalSnapOffsets;
    encoder << instance.snapAreas;
    encoder << instance.snapAreasIDs;
}

std::optional<WebCore::FloatScrollSnapOffsetsInfo> ArgumentCoder<WebCore::FloatScrollSnapOffsetsInfo>::decode(Decoder& decoder)
{
    auto strictness = decoder.decode<WebCore::ScrollSnapStrictness>();
    auto horizontalSnapOffsets = decoder.decode<Vector<WebCore::FloatSnapOffset>>();
    auto verticalSnapOffsets = decoder.decode<Vector<WebCore::FloatSnapOffset>>();
    auto snapAreas = decoder.decode<Vector<WebCore::FloatRect>>();
    auto snapAreasIDs = decoder.decode<Vector<WebCore::ElementIdentifier>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatScrollSnapOffsetsInfo {
            WTFMove(*strictness),
            WTFMove(*horizontalSnapOffsets),
            WTFMove(*verticalSnapOffsets),
            WTFMove(*snapAreas),
            WTFMove(*snapAreasIDs)
        }
    };
}

void ArgumentCoder<WebCore::FloatSnapOffset>::encode(Encoder& encoder, const WebCore::FloatSnapOffset& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stop)>, WebCore::ScrollSnapStop>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasSnapAreaLargerThanViewport)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.snapTargetID)>, WebCore::ElementIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFocused)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.snapAreaIndices)>, Vector<size_t>>);
    struct ShouldBeSameSizeAsFloatSnapOffset : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FloatSnapOffset>, false> {
        float offset;
        WebCore::ScrollSnapStop stop;
        bool hasSnapAreaLargerThanViewport;
        WebCore::ElementIdentifier snapTargetID;
        bool isFocused;
        Vector<size_t> snapAreaIndices;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFloatSnapOffset) == sizeof(WebCore::FloatSnapOffset));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FloatSnapOffset, offset)
        , offsetof(WebCore::FloatSnapOffset, stop)
        , offsetof(WebCore::FloatSnapOffset, hasSnapAreaLargerThanViewport)
        , offsetof(WebCore::FloatSnapOffset, snapTargetID)
        , offsetof(WebCore::FloatSnapOffset, isFocused)
        , offsetof(WebCore::FloatSnapOffset, snapAreaIndices)
    >::value);
    encoder << instance.offset;
    encoder << instance.stop;
    encoder << instance.hasSnapAreaLargerThanViewport;
    encoder << instance.snapTargetID;
    encoder << instance.isFocused;
    encoder << instance.snapAreaIndices;
}

std::optional<WebCore::FloatSnapOffset> ArgumentCoder<WebCore::FloatSnapOffset>::decode(Decoder& decoder)
{
    auto offset = decoder.decode<float>();
    auto stop = decoder.decode<WebCore::ScrollSnapStop>();
    auto hasSnapAreaLargerThanViewport = decoder.decode<bool>();
    auto snapTargetID = decoder.decode<WebCore::ElementIdentifier>();
    auto isFocused = decoder.decode<bool>();
    auto snapAreaIndices = decoder.decode<Vector<size_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FloatSnapOffset {
            WTFMove(*offset),
            WTFMove(*stop),
            WTFMove(*hasSnapAreaLargerThanViewport),
            WTFMove(*snapTargetID),
            WTFMove(*isFocused),
            WTFMove(*snapAreaIndices)
        }
    };
}

#if ENABLE(ENCRYPTED_MEDIA)
void ArgumentCoder<WebCore::CDMMediaCapability>::encode(Encoder& encoder, const WebCore::CDMMediaCapability& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentType)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.robustness)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.encryptionScheme)>, std::optional<WebCore::CDMEncryptionScheme>>);
    struct ShouldBeSameSizeAsCDMMediaCapability : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CDMMediaCapability>, false> {
        String contentType;
        String robustness;
        std::optional<WebCore::CDMEncryptionScheme> encryptionScheme;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCDMMediaCapability) == sizeof(WebCore::CDMMediaCapability));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CDMMediaCapability, contentType)
        , offsetof(WebCore::CDMMediaCapability, robustness)
        , offsetof(WebCore::CDMMediaCapability, encryptionScheme)
    >::value);
    encoder << instance.contentType;
    encoder << instance.robustness;
    encoder << instance.encryptionScheme;
}

std::optional<WebCore::CDMMediaCapability> ArgumentCoder<WebCore::CDMMediaCapability>::decode(Decoder& decoder)
{
    auto contentType = decoder.decode<String>();
    auto robustness = decoder.decode<String>();
    auto encryptionScheme = decoder.decode<std::optional<WebCore::CDMEncryptionScheme>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CDMMediaCapability {
            WTFMove(*contentType),
            WTFMove(*robustness),
            WTFMove(*encryptionScheme)
        }
    };
}

#endif

#if ENABLE(ENCRYPTED_MEDIA)
void ArgumentCoder<WebCore::CDMRestrictions>::encode(Encoder& encoder, const WebCore::CDMRestrictions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.distinctiveIdentifierDenied)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.persistentStateDenied)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deniedSessionTypes)>, HashSet<WebCore::CDMSessionType, IntHash<WebCore::CDMSessionType>, WTF::StrongEnumHashTraits<WebCore::CDMSessionType>>>);
    struct ShouldBeSameSizeAsCDMRestrictions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CDMRestrictions>, false> {
        bool distinctiveIdentifierDenied;
        bool persistentStateDenied;
        HashSet<WebCore::CDMSessionType, IntHash<WebCore::CDMSessionType>, WTF::StrongEnumHashTraits<WebCore::CDMSessionType>> deniedSessionTypes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCDMRestrictions) == sizeof(WebCore::CDMRestrictions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CDMRestrictions, distinctiveIdentifierDenied)
        , offsetof(WebCore::CDMRestrictions, persistentStateDenied)
        , offsetof(WebCore::CDMRestrictions, deniedSessionTypes)
    >::value);
    encoder << instance.distinctiveIdentifierDenied;
    encoder << instance.persistentStateDenied;
    encoder << instance.deniedSessionTypes;
}

std::optional<WebCore::CDMRestrictions> ArgumentCoder<WebCore::CDMRestrictions>::decode(Decoder& decoder)
{
    auto distinctiveIdentifierDenied = decoder.decode<bool>();
    auto persistentStateDenied = decoder.decode<bool>();
    auto deniedSessionTypes = decoder.decode<HashSet<WebCore::CDMSessionType, IntHash<WebCore::CDMSessionType>, WTF::StrongEnumHashTraits<WebCore::CDMSessionType>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CDMRestrictions {
            WTFMove(*distinctiveIdentifierDenied),
            WTFMove(*persistentStateDenied),
            WTFMove(*deniedSessionTypes)
        }
    };
}

#endif

void ArgumentCoder<WebCore::MediaSelectionOption>::encode(Encoder& encoder, const WebCore::MediaSelectionOption& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaType)>, WebCore::MediaSelectionOption::MediaType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.displayName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.legibleType)>, WebCore::MediaSelectionOption::LegibleType>);
    struct ShouldBeSameSizeAsMediaSelectionOption : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaSelectionOption>, false> {
        WebCore::MediaSelectionOption::MediaType mediaType;
        String displayName;
        WebCore::MediaSelectionOption::LegibleType legibleType;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaSelectionOption) == sizeof(WebCore::MediaSelectionOption));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaSelectionOption, mediaType)
        , offsetof(WebCore::MediaSelectionOption, displayName)
        , offsetof(WebCore::MediaSelectionOption, legibleType)
    >::value);
    encoder << instance.mediaType;
    encoder << instance.displayName;
    encoder << instance.legibleType;
}

std::optional<WebCore::MediaSelectionOption> ArgumentCoder<WebCore::MediaSelectionOption>::decode(Decoder& decoder)
{
    auto mediaType = decoder.decode<WebCore::MediaSelectionOption::MediaType>();
    auto displayName = decoder.decode<String>();
    auto legibleType = decoder.decode<WebCore::MediaSelectionOption::LegibleType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaSelectionOption {
            WTFMove(*mediaType),
            WTFMove(*displayName),
            WTFMove(*legibleType)
        }
    };
}

void ArgumentCoder<WebCore::BufferSource>::encode(Encoder& encoder, const WebCore::BufferSource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.span())>, std::span<const uint8_t>>);
    encoder << instance.span();
}

std::optional<WebCore::BufferSource> ArgumentCoder<WebCore::BufferSource>::decode(Decoder& decoder)
{
    auto span = decoder.decode<std::span<const uint8_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BufferSource {
            WTFMove(*span)
        }
    };
}

void ArgumentCoder<WebCore::FontShadow>::encode(Encoder& encoder, const WebCore::FontShadow& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blurRadius)>, double>);
    struct ShouldBeSameSizeAsFontShadow : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FontShadow>, false> {
        WebCore::Color color;
        WebCore::FloatSize offset;
        double blurRadius;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFontShadow) == sizeof(WebCore::FontShadow));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FontShadow, color)
        , offsetof(WebCore::FontShadow, offset)
        , offsetof(WebCore::FontShadow, blurRadius)
    >::value);
    encoder << instance.color;
    encoder << instance.offset;
    encoder << instance.blurRadius;
}

std::optional<WebCore::FontShadow> ArgumentCoder<WebCore::FontShadow>::decode(Decoder& decoder)
{
    auto color = decoder.decode<WebCore::Color>();
    auto offset = decoder.decode<WebCore::FloatSize>();
    auto blurRadius = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FontShadow {
            WTFMove(*color),
            WTFMove(*offset),
            WTFMove(*blurRadius)
        }
    };
}

void ArgumentCoder<WebCore::CompositionHighlight>::encode(Encoder& encoder, const WebCore::CompositionHighlight& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startOffset)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endOffset)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.backgroundColor)>, std::optional<WebCore::Color>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.foregroundColor)>, std::optional<WebCore::Color>>);
    struct ShouldBeSameSizeAsCompositionHighlight : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CompositionHighlight>, false> {
        unsigned startOffset;
        unsigned endOffset;
        std::optional<WebCore::Color> backgroundColor;
        std::optional<WebCore::Color> foregroundColor;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCompositionHighlight) == sizeof(WebCore::CompositionHighlight));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CompositionHighlight, startOffset)
        , offsetof(WebCore::CompositionHighlight, endOffset)
        , offsetof(WebCore::CompositionHighlight, backgroundColor)
        , offsetof(WebCore::CompositionHighlight, foregroundColor)
    >::value);
    encoder << instance.startOffset;
    encoder << instance.endOffset;
    encoder << instance.backgroundColor;
    encoder << instance.foregroundColor;
}

std::optional<WebCore::CompositionHighlight> ArgumentCoder<WebCore::CompositionHighlight>::decode(Decoder& decoder)
{
    auto startOffset = decoder.decode<unsigned>();
    auto endOffset = decoder.decode<unsigned>();
    auto backgroundColor = decoder.decode<std::optional<WebCore::Color>>();
    auto foregroundColor = decoder.decode<std::optional<WebCore::Color>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CompositionHighlight {
            WTFMove(*startOffset),
            WTFMove(*endOffset),
            WTFMove(*backgroundColor),
            WTFMove(*foregroundColor)
        }
    };
}

void ArgumentCoder<WebCore::FontChanges>::encode(Encoder& encoder, const WebCore::FontChanges& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_fontName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_fontFamily)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_fontSize)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_fontSizeDelta)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_bold)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_italic)>, std::optional<bool>>);
    struct ShouldBeSameSizeAsFontChanges : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FontChanges>, false> {
        String m_fontName;
        String m_fontFamily;
        std::optional<double> m_fontSize;
        std::optional<double> m_fontSizeDelta;
        std::optional<bool> m_bold;
        std::optional<bool> m_italic;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFontChanges) == sizeof(WebCore::FontChanges));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FontChanges, m_fontName)
        , offsetof(WebCore::FontChanges, m_fontFamily)
        , offsetof(WebCore::FontChanges, m_fontSize)
        , offsetof(WebCore::FontChanges, m_fontSizeDelta)
        , offsetof(WebCore::FontChanges, m_bold)
        , offsetof(WebCore::FontChanges, m_italic)
    >::value);
    encoder << instance.m_fontName;
    encoder << instance.m_fontFamily;
    encoder << instance.m_fontSize;
    encoder << instance.m_fontSizeDelta;
    encoder << instance.m_bold;
    encoder << instance.m_italic;
}

std::optional<WebCore::FontChanges> ArgumentCoder<WebCore::FontChanges>::decode(Decoder& decoder)
{
    auto m_fontName = decoder.decode<String>();
    auto m_fontFamily = decoder.decode<String>();
    auto m_fontSize = decoder.decode<std::optional<double>>();
    auto m_fontSizeDelta = decoder.decode<std::optional<double>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(!*m_fontSize || !*m_fontSizeDelta))
        return std::nullopt;
    auto m_bold = decoder.decode<std::optional<bool>>();
    auto m_italic = decoder.decode<std::optional<bool>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FontChanges {
            WTFMove(*m_fontName),
            WTFMove(*m_fontFamily),
            WTFMove(*m_fontSize),
            WTFMove(*m_fontSizeDelta),
            WTFMove(*m_bold),
            WTFMove(*m_italic)
        }
    };
}

void ArgumentCoder<WebCore::FontAttributeChanges>::encode(Encoder& encoder, const WebCore::FontAttributeChanges& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_verticalAlign)>, std::optional<WebCore::VerticalAlignChange>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_backgroundColor)>, std::optional<WebCore::Color>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_foregroundColor)>, std::optional<WebCore::Color>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_shadow)>, std::optional<WebCore::FontShadow>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_strikeThrough)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_underline)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_fontChanges)>, WebCore::FontChanges>);
    struct ShouldBeSameSizeAsFontAttributeChanges : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FontAttributeChanges>, false> {
        std::optional<WebCore::VerticalAlignChange> m_verticalAlign;
        std::optional<WebCore::Color> m_backgroundColor;
        std::optional<WebCore::Color> m_foregroundColor;
        std::optional<WebCore::FontShadow> m_shadow;
        std::optional<bool> m_strikeThrough;
        std::optional<bool> m_underline;
        WebCore::FontChanges m_fontChanges;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFontAttributeChanges) == sizeof(WebCore::FontAttributeChanges));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FontAttributeChanges, m_verticalAlign)
        , offsetof(WebCore::FontAttributeChanges, m_backgroundColor)
        , offsetof(WebCore::FontAttributeChanges, m_foregroundColor)
        , offsetof(WebCore::FontAttributeChanges, m_shadow)
        , offsetof(WebCore::FontAttributeChanges, m_strikeThrough)
        , offsetof(WebCore::FontAttributeChanges, m_underline)
        , offsetof(WebCore::FontAttributeChanges, m_fontChanges)
    >::value);
    encoder << instance.m_verticalAlign;
    encoder << instance.m_backgroundColor;
    encoder << instance.m_foregroundColor;
    encoder << instance.m_shadow;
    encoder << instance.m_strikeThrough;
    encoder << instance.m_underline;
    encoder << instance.m_fontChanges;
}

std::optional<WebCore::FontAttributeChanges> ArgumentCoder<WebCore::FontAttributeChanges>::decode(Decoder& decoder)
{
    auto m_verticalAlign = decoder.decode<std::optional<WebCore::VerticalAlignChange>>();
    auto m_backgroundColor = decoder.decode<std::optional<WebCore::Color>>();
    auto m_foregroundColor = decoder.decode<std::optional<WebCore::Color>>();
    auto m_shadow = decoder.decode<std::optional<WebCore::FontShadow>>();
    auto m_strikeThrough = decoder.decode<std::optional<bool>>();
    auto m_underline = decoder.decode<std::optional<bool>>();
    auto m_fontChanges = decoder.decode<WebCore::FontChanges>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FontAttributeChanges {
            WTFMove(*m_verticalAlign),
            WTFMove(*m_backgroundColor),
            WTFMove(*m_foregroundColor),
            WTFMove(*m_shadow),
            WTFMove(*m_strikeThrough),
            WTFMove(*m_underline),
            WTFMove(*m_fontChanges)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::ElementRule>::encode(Encoder& encoder, const WebCore::TextManipulationControllerExclusionRule::ElementRule& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localName)>, AtomString>);
    struct ShouldBeSameSizeAsElementRule : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationControllerExclusionRule::ElementRule>, false> {
        AtomString localName;
    };
    static_assert(sizeof(ShouldBeSameSizeAsElementRule) == sizeof(WebCore::TextManipulationControllerExclusionRule::ElementRule));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationControllerExclusionRule::ElementRule, localName)
    >::value);
    encoder << instance.localName;
}

std::optional<WebCore::TextManipulationControllerExclusionRule::ElementRule> ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::ElementRule>::decode(Decoder& decoder)
{
    auto localName = decoder.decode<AtomString>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationControllerExclusionRule::ElementRule {
            WTFMove(*localName)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::AttributeRule>::encode(Encoder& encoder, const WebCore::TextManipulationControllerExclusionRule::AttributeRule& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, String>);
    struct ShouldBeSameSizeAsAttributeRule : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationControllerExclusionRule::AttributeRule>, false> {
        AtomString name;
        String value;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAttributeRule) == sizeof(WebCore::TextManipulationControllerExclusionRule::AttributeRule));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationControllerExclusionRule::AttributeRule, name)
        , offsetof(WebCore::TextManipulationControllerExclusionRule::AttributeRule, value)
    >::value);
    encoder << instance.name;
    encoder << instance.value;
}

std::optional<WebCore::TextManipulationControllerExclusionRule::AttributeRule> ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::AttributeRule>::decode(Decoder& decoder)
{
    auto name = decoder.decode<AtomString>();
    auto value = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationControllerExclusionRule::AttributeRule {
            WTFMove(*name),
            WTFMove(*value)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::ClassRule>::encode(Encoder& encoder, const WebCore::TextManipulationControllerExclusionRule::ClassRule& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.className)>, AtomString>);
    struct ShouldBeSameSizeAsClassRule : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationControllerExclusionRule::ClassRule>, false> {
        AtomString className;
    };
    static_assert(sizeof(ShouldBeSameSizeAsClassRule) == sizeof(WebCore::TextManipulationControllerExclusionRule::ClassRule));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationControllerExclusionRule::ClassRule, className)
    >::value);
    encoder << instance.className;
}

std::optional<WebCore::TextManipulationControllerExclusionRule::ClassRule> ArgumentCoder<WebCore::TextManipulationControllerExclusionRule::ClassRule>::decode(Decoder& decoder)
{
    auto className = decoder.decode<AtomString>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationControllerExclusionRule::ClassRule {
            WTFMove(*className)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationControllerExclusionRule>::encode(Encoder& encoder, const WebCore::TextManipulationControllerExclusionRule& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::TextManipulationControllerExclusionRule::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rule)>, std::variant<WebCore::TextManipulationControllerExclusionRule::ElementRule, WebCore::TextManipulationControllerExclusionRule::AttributeRule, WebCore::TextManipulationControllerExclusionRule::ClassRule>>);
    struct ShouldBeSameSizeAsTextManipulationControllerExclusionRule : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationControllerExclusionRule>, false> {
        WebCore::TextManipulationControllerExclusionRule::Type type;
        std::variant<WebCore::TextManipulationControllerExclusionRule::ElementRule, WebCore::TextManipulationControllerExclusionRule::AttributeRule, WebCore::TextManipulationControllerExclusionRule::ClassRule> rule;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextManipulationControllerExclusionRule) == sizeof(WebCore::TextManipulationControllerExclusionRule));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationControllerExclusionRule, type)
        , offsetof(WebCore::TextManipulationControllerExclusionRule, rule)
    >::value);
    encoder << instance.type;
    encoder << instance.rule;
}

std::optional<WebCore::TextManipulationControllerExclusionRule> ArgumentCoder<WebCore::TextManipulationControllerExclusionRule>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::TextManipulationControllerExclusionRule::Type>();
    auto rule = decoder.decode<std::variant<WebCore::TextManipulationControllerExclusionRule::ElementRule, WebCore::TextManipulationControllerExclusionRule::AttributeRule, WebCore::TextManipulationControllerExclusionRule::ClassRule>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationControllerExclusionRule {
            WTFMove(*type),
            WTFMove(*rule)
        }
    };
}

void ArgumentCoder<WebCore::TextManipulationControllerManipulationFailure>::encode(Encoder& encoder, const WebCore::TextManipulationControllerManipulationFailure& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, WebCore::FrameIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::TextManipulationItemIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.index)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::TextManipulationControllerManipulationFailure::Type>);
    struct ShouldBeSameSizeAsTextManipulationControllerManipulationFailure : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::TextManipulationControllerManipulationFailure>, false> {
        WebCore::FrameIdentifier frameID;
        WebCore::TextManipulationItemIdentifier identifier;
        uint64_t index;
        WebCore::TextManipulationControllerManipulationFailure::Type type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextManipulationControllerManipulationFailure) == sizeof(WebCore::TextManipulationControllerManipulationFailure));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::TextManipulationControllerManipulationFailure, frameID)
        , offsetof(WebCore::TextManipulationControllerManipulationFailure, identifier)
        , offsetof(WebCore::TextManipulationControllerManipulationFailure, index)
        , offsetof(WebCore::TextManipulationControllerManipulationFailure, type)
    >::value);
    encoder << instance.frameID;
    encoder << instance.identifier;
    encoder << instance.index;
    encoder << instance.type;
}

std::optional<WebCore::TextManipulationControllerManipulationFailure> ArgumentCoder<WebCore::TextManipulationControllerManipulationFailure>::decode(Decoder& decoder)
{
    auto frameID = decoder.decode<WebCore::FrameIdentifier>();
    auto identifier = decoder.decode<WebCore::TextManipulationItemIdentifier>();
    auto index = decoder.decode<uint64_t>();
    auto type = decoder.decode<WebCore::TextManipulationControllerManipulationFailure::Type>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::TextManipulationControllerManipulationFailure {
            WTFMove(*frameID),
            WTFMove(*identifier),
            WTFMove(*index),
            WTFMove(*type)
        }
    };
}

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::BackgroundFetchInformation>::encode(Encoder& encoder, const WebCore::BackgroundFetchInformation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrationIdentifier)>, WebCore::ServiceWorkerRegistrationIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uploadTotal)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uploaded)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.downloadTotal)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.downloaded)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.result)>, WebCore::BackgroundFetchResult>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failureReason)>, WebCore::BackgroundFetchFailureReason>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.recordsAvailable)>, bool>);
    struct ShouldBeSameSizeAsBackgroundFetchInformation : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::BackgroundFetchInformation>, false> {
        WebCore::ServiceWorkerRegistrationIdentifier registrationIdentifier;
        String identifier;
        uint64_t uploadTotal;
        uint64_t uploaded;
        uint64_t downloadTotal;
        uint64_t downloaded;
        WebCore::BackgroundFetchResult result;
        WebCore::BackgroundFetchFailureReason failureReason;
        bool recordsAvailable;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBackgroundFetchInformation) == sizeof(WebCore::BackgroundFetchInformation));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::BackgroundFetchInformation, registrationIdentifier)
        , offsetof(WebCore::BackgroundFetchInformation, identifier)
        , offsetof(WebCore::BackgroundFetchInformation, uploadTotal)
        , offsetof(WebCore::BackgroundFetchInformation, uploaded)
        , offsetof(WebCore::BackgroundFetchInformation, downloadTotal)
        , offsetof(WebCore::BackgroundFetchInformation, downloaded)
        , offsetof(WebCore::BackgroundFetchInformation, result)
        , offsetof(WebCore::BackgroundFetchInformation, failureReason)
        , offsetof(WebCore::BackgroundFetchInformation, recordsAvailable)
    >::value);
    encoder << instance.registrationIdentifier;
    encoder << instance.identifier;
    encoder << instance.uploadTotal;
    encoder << instance.uploaded;
    encoder << instance.downloadTotal;
    encoder << instance.downloaded;
    encoder << instance.result;
    encoder << instance.failureReason;
    encoder << instance.recordsAvailable;
}

std::optional<WebCore::BackgroundFetchInformation> ArgumentCoder<WebCore::BackgroundFetchInformation>::decode(Decoder& decoder)
{
    auto registrationIdentifier = decoder.decode<WebCore::ServiceWorkerRegistrationIdentifier>();
    auto identifier = decoder.decode<String>();
    auto uploadTotal = decoder.decode<uint64_t>();
    auto uploaded = decoder.decode<uint64_t>();
    auto downloadTotal = decoder.decode<uint64_t>();
    auto downloaded = decoder.decode<uint64_t>();
    auto result = decoder.decode<WebCore::BackgroundFetchResult>();
    auto failureReason = decoder.decode<WebCore::BackgroundFetchFailureReason>();
    auto recordsAvailable = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BackgroundFetchInformation {
            WTFMove(*registrationIdentifier),
            WTFMove(*identifier),
            WTFMove(*uploadTotal),
            WTFMove(*uploaded),
            WTFMove(*downloadTotal),
            WTFMove(*downloaded),
            WTFMove(*result),
            WTFMove(*failureReason),
            WTFMove(*recordsAvailable)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::BackgroundFetchRequest>::encode(Encoder& encoder, const WebCore::BackgroundFetchRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.internalRequest)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.options)>, WebCore::FetchOptions>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.guard)>, WebCore::FetchHeadersGuard>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpHeaders)>, WebCore::HTTPHeaderMap>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrer)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cspResponseHeaders)>, Markable<WebCore::ContentSecurityPolicyResponseHeaders, WebCore::ContentSecurityPolicyResponseHeaders::MarkableTraits>>);
    struct ShouldBeSameSizeAsBackgroundFetchRequest : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::BackgroundFetchRequest>, false> {
        WebCore::ResourceRequest internalRequest;
        WebCore::FetchOptions options;
        WebCore::FetchHeadersGuard guard;
        WebCore::HTTPHeaderMap httpHeaders;
        String referrer;
        Markable<WebCore::ContentSecurityPolicyResponseHeaders, WebCore::ContentSecurityPolicyResponseHeaders::MarkableTraits> cspResponseHeaders;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBackgroundFetchRequest) == sizeof(WebCore::BackgroundFetchRequest));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::BackgroundFetchRequest, internalRequest)
        , offsetof(WebCore::BackgroundFetchRequest, options)
        , offsetof(WebCore::BackgroundFetchRequest, guard)
        , offsetof(WebCore::BackgroundFetchRequest, httpHeaders)
        , offsetof(WebCore::BackgroundFetchRequest, referrer)
        , offsetof(WebCore::BackgroundFetchRequest, cspResponseHeaders)
    >::value);
    encoder << instance.internalRequest;
    encoder << instance.options;
    encoder << instance.guard;
    encoder << instance.httpHeaders;
    encoder << instance.referrer;
    encoder << instance.cspResponseHeaders;
}

std::optional<WebCore::BackgroundFetchRequest> ArgumentCoder<WebCore::BackgroundFetchRequest>::decode(Decoder& decoder)
{
    auto internalRequest = decoder.decode<WebCore::ResourceRequest>();
    auto options = decoder.decode<WebCore::FetchOptions>();
    auto guard = decoder.decode<WebCore::FetchHeadersGuard>();
    auto httpHeaders = decoder.decode<WebCore::HTTPHeaderMap>();
    auto referrer = decoder.decode<String>();
    auto cspResponseHeaders = decoder.decode<Markable<WebCore::ContentSecurityPolicyResponseHeaders, WebCore::ContentSecurityPolicyResponseHeaders::MarkableTraits>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BackgroundFetchRequest {
            WTFMove(*internalRequest),
            WTFMove(*options),
            WTFMove(*guard),
            WTFMove(*httpHeaders),
            WTFMove(*referrer),
            WTFMove(*cspResponseHeaders)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::BackgroundFetchRecordInformation>::encode(Encoder& encoder, const WebCore::BackgroundFetchRecordInformation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::BackgroundFetchRecordIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.internalRequest)>, WebCore::ResourceRequest>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.options)>, WebCore::FetchOptions>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.guard)>, WebCore::FetchHeadersGuard>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpHeaders)>, WebCore::HTTPHeaderMap>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrer)>, String>);
    struct ShouldBeSameSizeAsBackgroundFetchRecordInformation : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::BackgroundFetchRecordInformation>, false> {
        WebCore::BackgroundFetchRecordIdentifier identifier;
        WebCore::ResourceRequest internalRequest;
        WebCore::FetchOptions options;
        WebCore::FetchHeadersGuard guard;
        WebCore::HTTPHeaderMap httpHeaders;
        String referrer;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBackgroundFetchRecordInformation) == sizeof(WebCore::BackgroundFetchRecordInformation));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::BackgroundFetchRecordInformation, identifier)
        , offsetof(WebCore::BackgroundFetchRecordInformation, internalRequest)
        , offsetof(WebCore::BackgroundFetchRecordInformation, options)
        , offsetof(WebCore::BackgroundFetchRecordInformation, guard)
        , offsetof(WebCore::BackgroundFetchRecordInformation, httpHeaders)
        , offsetof(WebCore::BackgroundFetchRecordInformation, referrer)
    >::value);
    encoder << instance.identifier;
    encoder << instance.internalRequest;
    encoder << instance.options;
    encoder << instance.guard;
    encoder << instance.httpHeaders;
    encoder << instance.referrer;
}

std::optional<WebCore::BackgroundFetchRecordInformation> ArgumentCoder<WebCore::BackgroundFetchRecordInformation>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::BackgroundFetchRecordIdentifier>();
    auto internalRequest = decoder.decode<WebCore::ResourceRequest>();
    auto options = decoder.decode<WebCore::FetchOptions>();
    auto guard = decoder.decode<WebCore::FetchHeadersGuard>();
    auto httpHeaders = decoder.decode<WebCore::HTTPHeaderMap>();
    auto referrer = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BackgroundFetchRecordInformation {
            WTFMove(*identifier),
            WTFMove(*internalRequest),
            WTFMove(*options),
            WTFMove(*guard),
            WTFMove(*httpHeaders),
            WTFMove(*referrer)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::BackgroundFetchOptions>::encode(Encoder& encoder, const WebCore::BackgroundFetchOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.icons)>, Vector<WebCore::ImageResource>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.title)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.downloadTotal)>, uint64_t>);
    struct ShouldBeSameSizeAsBackgroundFetchOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::BackgroundFetchOptions>, false> {
        Vector<WebCore::ImageResource> icons;
        String title;
        uint64_t downloadTotal;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBackgroundFetchOptions) == sizeof(WebCore::BackgroundFetchOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::BackgroundFetchOptions, icons)
        , offsetof(WebCore::BackgroundFetchOptions, title)
        , offsetof(WebCore::BackgroundFetchOptions, downloadTotal)
    >::value);
    encoder << instance.icons;
    encoder << instance.title;
    encoder << instance.downloadTotal;
}

std::optional<WebCore::BackgroundFetchOptions> ArgumentCoder<WebCore::BackgroundFetchOptions>::decode(Decoder& decoder)
{
    auto icons = decoder.decode<Vector<WebCore::ImageResource>>();
    auto title = decoder.decode<String>();
    auto downloadTotal = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BackgroundFetchOptions {
            WTFMove(*icons),
            WTFMove(*title),
            WTFMove(*downloadTotal)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ImageResource>::encode(Encoder& encoder, const WebCore::ImageResource& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.src)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sizes)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    struct ShouldBeSameSizeAsImageResource : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ImageResource>, false> {
        String src;
        String sizes;
        String type;
        String label;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageResource) == sizeof(WebCore::ImageResource));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ImageResource, src)
        , offsetof(WebCore::ImageResource, sizes)
        , offsetof(WebCore::ImageResource, type)
        , offsetof(WebCore::ImageResource, label)
    >::value);
    encoder << instance.src;
    encoder << instance.sizes;
    encoder << instance.type;
    encoder << instance.label;
}

std::optional<WebCore::ImageResource> ArgumentCoder<WebCore::ImageResource>::decode(Decoder& decoder)
{
    auto src = decoder.decode<String>();
    auto sizes = decoder.decode<String>();
    auto type = decoder.decode<String>();
    auto label = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ImageResource {
            WTFMove(*src),
            WTFMove(*sizes),
            WTFMove(*type),
            WTFMove(*label)
        }
    };
}

#endif

void ArgumentCoder<WebCore::FourCC>::encode(Encoder& encoder, const WebCore::FourCC& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, uint32_t>);
    struct ShouldBeSameSizeAsFourCC : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FourCC>, false> {
        uint32_t value;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFourCC) == sizeof(WebCore::FourCC));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FourCC, value)
    >::value);
    encoder << instance.value;
}

std::optional<WebCore::FourCC> ArgumentCoder<WebCore::FourCC>::decode(Decoder& decoder)
{
    auto value = decoder.decode<uint32_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FourCC {
            WTFMove(*value)
        }
    };
}

#if ENABLE(VIDEO)
void ArgumentCoder<WebCore::MediaEngineSupportParameters>::encode(Encoder& encoder, const WebCore::MediaEngineSupportParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ContentType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMediaSource)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMediaStream)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requiresRemotePlayback)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentTypesRequiringHardwareSupport)>, Vector<WebCore::ContentType>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaContainerTypes)>, std::optional<Vector<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaCodecTypes)>, std::optional<Vector<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaVideoCodecIDs)>, std::optional<Vector<WebCore::FourCC>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaAudioCodecIDs)>, std::optional<Vector<WebCore::FourCC>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedMediaCaptionFormatTypes)>, std::optional<Vector<WebCore::FourCC>>>);
    struct ShouldBeSameSizeAsMediaEngineSupportParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaEngineSupportParameters>, false> {
        WebCore::ContentType type;
        URL url;
        bool isMediaSource;
        bool isMediaStream;
        bool requiresRemotePlayback;
        Vector<WebCore::ContentType> contentTypesRequiringHardwareSupport;
        std::optional<Vector<String>> allowedMediaContainerTypes;
        std::optional<Vector<String>> allowedMediaCodecTypes;
        std::optional<Vector<WebCore::FourCC>> allowedMediaVideoCodecIDs;
        std::optional<Vector<WebCore::FourCC>> allowedMediaAudioCodecIDs;
        std::optional<Vector<WebCore::FourCC>> allowedMediaCaptionFormatTypes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaEngineSupportParameters) == sizeof(WebCore::MediaEngineSupportParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaEngineSupportParameters, type)
        , offsetof(WebCore::MediaEngineSupportParameters, url)
        , offsetof(WebCore::MediaEngineSupportParameters, isMediaSource)
        , offsetof(WebCore::MediaEngineSupportParameters, isMediaStream)
        , offsetof(WebCore::MediaEngineSupportParameters, requiresRemotePlayback)
        , offsetof(WebCore::MediaEngineSupportParameters, contentTypesRequiringHardwareSupport)
        , offsetof(WebCore::MediaEngineSupportParameters, allowedMediaContainerTypes)
        , offsetof(WebCore::MediaEngineSupportParameters, allowedMediaCodecTypes)
        , offsetof(WebCore::MediaEngineSupportParameters, allowedMediaVideoCodecIDs)
        , offsetof(WebCore::MediaEngineSupportParameters, allowedMediaAudioCodecIDs)
        , offsetof(WebCore::MediaEngineSupportParameters, allowedMediaCaptionFormatTypes)
    >::value);
    encoder << instance.type;
    encoder << instance.url;
    encoder << instance.isMediaSource;
    encoder << instance.isMediaStream;
    encoder << instance.requiresRemotePlayback;
    encoder << instance.contentTypesRequiringHardwareSupport;
    encoder << instance.allowedMediaContainerTypes;
    encoder << instance.allowedMediaCodecTypes;
    encoder << instance.allowedMediaVideoCodecIDs;
    encoder << instance.allowedMediaAudioCodecIDs;
    encoder << instance.allowedMediaCaptionFormatTypes;
}

std::optional<WebCore::MediaEngineSupportParameters> ArgumentCoder<WebCore::MediaEngineSupportParameters>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::ContentType>();
    auto url = decoder.decode<URL>();
    auto isMediaSource = decoder.decode<bool>();
    auto isMediaStream = decoder.decode<bool>();
    auto requiresRemotePlayback = decoder.decode<bool>();
    auto contentTypesRequiringHardwareSupport = decoder.decode<Vector<WebCore::ContentType>>();
    auto allowedMediaContainerTypes = decoder.decode<std::optional<Vector<String>>>();
    auto allowedMediaCodecTypes = decoder.decode<std::optional<Vector<String>>>();
    auto allowedMediaVideoCodecIDs = decoder.decode<std::optional<Vector<WebCore::FourCC>>>();
    auto allowedMediaAudioCodecIDs = decoder.decode<std::optional<Vector<WebCore::FourCC>>>();
    auto allowedMediaCaptionFormatTypes = decoder.decode<std::optional<Vector<WebCore::FourCC>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaEngineSupportParameters {
            WTFMove(*type),
            WTFMove(*url),
            WTFMove(*isMediaSource),
            WTFMove(*isMediaStream),
            WTFMove(*requiresRemotePlayback),
            WTFMove(*contentTypesRequiringHardwareSupport),
            WTFMove(*allowedMediaContainerTypes),
            WTFMove(*allowedMediaCodecTypes),
            WTFMove(*allowedMediaVideoCodecIDs),
            WTFMove(*allowedMediaAudioCodecIDs),
            WTFMove(*allowedMediaCaptionFormatTypes)
        }
    };
}

#endif

void ArgumentCoder<WebCore::GeolocationPositionData>::encode(Encoder& encoder, const WebCore::GeolocationPositionData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.latitude)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.longitude)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.accuracy)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.altitude)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.altitudeAccuracy)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.heading)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.speed)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.floorLevel)>, std::optional<double>>);
    struct ShouldBeSameSizeAsGeolocationPositionData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GeolocationPositionData>, false> {
        double timestamp;
        double latitude;
        double longitude;
        double accuracy;
        std::optional<double> altitude;
        std::optional<double> altitudeAccuracy;
        std::optional<double> heading;
        std::optional<double> speed;
        std::optional<double> floorLevel;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGeolocationPositionData) == sizeof(WebCore::GeolocationPositionData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GeolocationPositionData, timestamp)
        , offsetof(WebCore::GeolocationPositionData, latitude)
        , offsetof(WebCore::GeolocationPositionData, longitude)
        , offsetof(WebCore::GeolocationPositionData, accuracy)
        , offsetof(WebCore::GeolocationPositionData, altitude)
        , offsetof(WebCore::GeolocationPositionData, altitudeAccuracy)
        , offsetof(WebCore::GeolocationPositionData, heading)
        , offsetof(WebCore::GeolocationPositionData, speed)
        , offsetof(WebCore::GeolocationPositionData, floorLevel)
    >::value);
    encoder << instance.timestamp;
    encoder << instance.latitude;
    encoder << instance.longitude;
    encoder << instance.accuracy;
    encoder << instance.altitude;
    encoder << instance.altitudeAccuracy;
    encoder << instance.heading;
    encoder << instance.speed;
    encoder << instance.floorLevel;
}

std::optional<WebCore::GeolocationPositionData> ArgumentCoder<WebCore::GeolocationPositionData>::decode(Decoder& decoder)
{
    auto timestamp = decoder.decode<double>();
    auto latitude = decoder.decode<double>();
    auto longitude = decoder.decode<double>();
    auto accuracy = decoder.decode<double>();
    auto altitude = decoder.decode<std::optional<double>>();
    auto altitudeAccuracy = decoder.decode<std::optional<double>>();
    auto heading = decoder.decode<std::optional<double>>();
    auto speed = decoder.decode<std::optional<double>>();
    auto floorLevel = decoder.decode<std::optional<double>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GeolocationPositionData {
            WTFMove(*timestamp),
            WTFMove(*latitude),
            WTFMove(*longitude),
            WTFMove(*accuracy),
            WTFMove(*altitude),
            WTFMove(*altitudeAccuracy),
            WTFMove(*heading),
            WTFMove(*speed),
            WTFMove(*floorLevel)
        }
    };
}

#if ENABLE(APP_HIGHLIGHTS)
void ArgumentCoder<WebCore::AppHighlight>::encode(Encoder& encoder, const WebCore::AppHighlight& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.highlight)>, Ref<WebCore::FragmentedSharedBuffer>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text)>, std::optional<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isNewGroup)>, WebCore::CreateNewGroupForHighlight>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requestOriginatedInApp)>, WebCore::HighlightRequestOriginatedInApp>);
    struct ShouldBeSameSizeAsAppHighlight : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AppHighlight>, false> {
        Ref<WebCore::FragmentedSharedBuffer> highlight;
        std::optional<String> text;
        WebCore::CreateNewGroupForHighlight isNewGroup;
        WebCore::HighlightRequestOriginatedInApp requestOriginatedInApp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAppHighlight) == sizeof(WebCore::AppHighlight));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AppHighlight, highlight)
        , offsetof(WebCore::AppHighlight, text)
        , offsetof(WebCore::AppHighlight, isNewGroup)
        , offsetof(WebCore::AppHighlight, requestOriginatedInApp)
    >::value);
    encoder << instance.highlight;
    encoder << instance.text;
    encoder << instance.isNewGroup;
    encoder << instance.requestOriginatedInApp;
}

std::optional<WebCore::AppHighlight> ArgumentCoder<WebCore::AppHighlight>::decode(Decoder& decoder)
{
    auto highlight = decoder.decode<Ref<WebCore::FragmentedSharedBuffer>>();
    auto text = decoder.decode<std::optional<String>>();
    auto isNewGroup = decoder.decode<WebCore::CreateNewGroupForHighlight>();
    auto requestOriginatedInApp = decoder.decode<WebCore::HighlightRequestOriginatedInApp>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AppHighlight {
            WTFMove(*highlight),
            WTFMove(*text),
            WTFMove(*isNewGroup),
            WTFMove(*requestOriginatedInApp)
        }
    };
}

#endif

void ArgumentCoder<WebCore::MediaDeviceHashSalts>::encode(Encoder& encoder, const WebCore::MediaDeviceHashSalts& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.persistentDeviceSalt)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ephemeralDeviceSalt)>, String>);
    struct ShouldBeSameSizeAsMediaDeviceHashSalts : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaDeviceHashSalts>, false> {
        String persistentDeviceSalt;
        String ephemeralDeviceSalt;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaDeviceHashSalts) == sizeof(WebCore::MediaDeviceHashSalts));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaDeviceHashSalts, persistentDeviceSalt)
        , offsetof(WebCore::MediaDeviceHashSalts, ephemeralDeviceSalt)
    >::value);
    encoder << instance.persistentDeviceSalt;
    encoder << instance.ephemeralDeviceSalt;
}

std::optional<WebCore::MediaDeviceHashSalts> ArgumentCoder<WebCore::MediaDeviceHashSalts>::decode(Decoder& decoder)
{
    auto persistentDeviceSalt = decoder.decode<String>();
    auto ephemeralDeviceSalt = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaDeviceHashSalts {
            WTFMove(*persistentDeviceSalt),
            WTFMove(*ephemeralDeviceSalt)
        }
    };
}

void ArgumentCoder<WebCore::SpeechRecognitionUpdate>::encode(Encoder& encoder, const WebCore::SpeechRecognitionUpdate& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_clientIdentifier)>, WebCore::SpeechRecognitionConnectionClientIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_type)>, WebCore::SpeechRecognitionUpdateType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_content)>, std::variant<std::monostate, WebCore::SpeechRecognitionError, Vector<WebCore::SpeechRecognitionResultData>>>);
    struct ShouldBeSameSizeAsSpeechRecognitionUpdate : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SpeechRecognitionUpdate>, false> {
        WebCore::SpeechRecognitionConnectionClientIdentifier m_clientIdentifier;
        WebCore::SpeechRecognitionUpdateType m_type;
        std::variant<std::monostate, WebCore::SpeechRecognitionError, Vector<WebCore::SpeechRecognitionResultData>> m_content;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSpeechRecognitionUpdate) == sizeof(WebCore::SpeechRecognitionUpdate));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SpeechRecognitionUpdate, m_clientIdentifier)
        , offsetof(WebCore::SpeechRecognitionUpdate, m_type)
        , offsetof(WebCore::SpeechRecognitionUpdate, m_content)
    >::value);
    encoder << instance.m_clientIdentifier;
    encoder << instance.m_type;
    encoder << instance.m_content;
}

std::optional<WebCore::SpeechRecognitionUpdate> ArgumentCoder<WebCore::SpeechRecognitionUpdate>::decode(Decoder& decoder)
{
    auto m_clientIdentifier = decoder.decode<WebCore::SpeechRecognitionConnectionClientIdentifier>();
    auto m_type = decoder.decode<WebCore::SpeechRecognitionUpdateType>();
    auto m_content = decoder.decode<std::variant<std::monostate, WebCore::SpeechRecognitionError, Vector<WebCore::SpeechRecognitionResultData>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SpeechRecognitionUpdate {
            WTFMove(*m_clientIdentifier),
            WTFMove(*m_type),
            WTFMove(*m_content)
        }
    };
}

#if USE(SYSTEM_PREVIEW)
void ArgumentCoder<WebCore::ARKitBadgeSystemImage>::encode(Encoder& encoder, const WebCore::ARKitBadgeSystemImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.imageIdentifier())>, WebCore::RenderingResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_imageSize)>, WebCore::FloatSize>);
    encoder << instance.imageIdentifier();
    encoder << instance.m_imageSize;
}

void ArgumentCoder<WebCore::ARKitBadgeSystemImage>::encode(StreamConnectionEncoder& encoder, const WebCore::ARKitBadgeSystemImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.imageIdentifier())>, WebCore::RenderingResourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_imageSize)>, WebCore::FloatSize>);
    encoder << instance.imageIdentifier();
    encoder << instance.m_imageSize;
}

std::optional<Ref<WebCore::ARKitBadgeSystemImage>> ArgumentCoder<WebCore::ARKitBadgeSystemImage>::decode(Decoder& decoder)
{
    auto imageIdentifier = decoder.decode<WebCore::RenderingResourceIdentifier>();
    auto m_imageSize = decoder.decode<WebCore::FloatSize>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ARKitBadgeSystemImage::create(
            WTFMove(*imageIdentifier),
            WTFMove(*m_imageSize)
        )
    };
}

#endif

void ArgumentCoder<WebCore::AudioIOPosition>::encode(Encoder& encoder, const WebCore::AudioIOPosition& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position)>, Seconds>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp)>, MonotonicTime>);
    struct ShouldBeSameSizeAsAudioIOPosition : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AudioIOPosition>, false> {
        Seconds position;
        MonotonicTime timestamp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAudioIOPosition) == sizeof(WebCore::AudioIOPosition));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AudioIOPosition, position)
        , offsetof(WebCore::AudioIOPosition, timestamp)
    >::value);
    encoder << instance.position;
    encoder << instance.timestamp;
}

std::optional<WebCore::AudioIOPosition> ArgumentCoder<WebCore::AudioIOPosition>::decode(Decoder& decoder)
{
    auto position = decoder.decode<Seconds>();
    auto timestamp = decoder.decode<MonotonicTime>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AudioIOPosition {
            WTFMove(*position),
            WTFMove(*timestamp)
        }
    };
}

#if ENABLE(VIDEO) && USE(AVFOUNDATION)
void ArgumentCoder<WebCore::VideoFrameCV>::encode(Encoder& encoder, const WebCore::VideoFrameCV& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentationTime())>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMirrored())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rotation())>, WebCore::VideoFrameRotation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_pixelBuffer)>, RetainPtr<CVPixelBufferRef>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace())>, WebCore::PlatformVideoColorSpace>);
    encoder << instance.presentationTime();
    encoder << instance.isMirrored();
    encoder << instance.rotation();
    encoder << instance.m_pixelBuffer;
    encoder << instance.colorSpace();
}

std::optional<Ref<WebCore::VideoFrameCV>> ArgumentCoder<WebCore::VideoFrameCV>::decode(Decoder& decoder)
{
    auto presentationTime = decoder.decode<MediaTime>();
    auto isMirrored = decoder.decode<bool>();
    auto rotation = decoder.decode<WebCore::VideoFrameRotation>();
    auto m_pixelBuffer = decoder.decode<RetainPtr<CVPixelBufferRef>>();
    auto colorSpace = decoder.decode<WebCore::PlatformVideoColorSpace>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::VideoFrameCV::create(
            WTFMove(*presentationTime),
            WTFMove(*isMirrored),
            WTFMove(*rotation),
            WTFMove(*m_pixelBuffer),
            WTFMove(*colorSpace)
        )
    };
}

#endif

void ArgumentCoder<WebCore::GradientColorStop>::encode(Encoder& encoder, const WebCore::GradientColorStop& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    struct ShouldBeSameSizeAsGradientColorStop : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GradientColorStop>, false> {
        float offset;
        WebCore::Color color;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGradientColorStop) == sizeof(WebCore::GradientColorStop));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GradientColorStop, offset)
        , offsetof(WebCore::GradientColorStop, color)
    >::value);
    encoder << instance.offset;
    encoder << instance.color;
}

void ArgumentCoder<WebCore::GradientColorStop>::encode(StreamConnectionEncoder& encoder, const WebCore::GradientColorStop& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    struct ShouldBeSameSizeAsGradientColorStop : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GradientColorStop>, false> {
        float offset;
        WebCore::Color color;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGradientColorStop) == sizeof(WebCore::GradientColorStop));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GradientColorStop, offset)
        , offsetof(WebCore::GradientColorStop, color)
    >::value);
    encoder << instance.offset;
    encoder << instance.color;
}

std::optional<WebCore::GradientColorStop> ArgumentCoder<WebCore::GradientColorStop>::decode(Decoder& decoder)
{
    auto offset = decoder.decode<float>();
    auto color = decoder.decode<WebCore::Color>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GradientColorStop {
            WTFMove(*offset),
            WTFMove(*color)
        }
    };
}

void ArgumentCoder<WebCore::GradientColorStops>::encode(Encoder& encoder, const WebCore::GradientColorStops& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stops())>, Vector<WebCore::GradientColorStop, 2>>);
    encoder << instance.stops();
}

void ArgumentCoder<WebCore::GradientColorStops>::encode(StreamConnectionEncoder& encoder, const WebCore::GradientColorStops& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stops())>, Vector<WebCore::GradientColorStop, 2>>);
    encoder << instance.stops();
}

std::optional<WebCore::GradientColorStops> ArgumentCoder<WebCore::GradientColorStops>::decode(Decoder& decoder)
{
    auto stops = decoder.decode<Vector<WebCore::GradientColorStop, 2>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GradientColorStops {
            WTFMove(*stops)
        }
    };
}

void ArgumentCoder<WebCore::ImageBufferBackendParameters>::encode(Encoder& encoder, const WebCore::ImageBufferBackendParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.logicalSize)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resolutionScale)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::DestinationColorSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pixelFormat)>, WebCore::PixelFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.purpose)>, WebCore::RenderingPurpose>);
    struct ShouldBeSameSizeAsImageBufferBackendParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ImageBufferBackendParameters>, false> {
        WebCore::FloatSize logicalSize;
        float resolutionScale;
        WebCore::DestinationColorSpace colorSpace;
        WebCore::PixelFormat pixelFormat;
        WebCore::RenderingPurpose purpose;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageBufferBackendParameters) == sizeof(WebCore::ImageBufferBackendParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ImageBufferBackendParameters, logicalSize)
        , offsetof(WebCore::ImageBufferBackendParameters, resolutionScale)
        , offsetof(WebCore::ImageBufferBackendParameters, colorSpace)
        , offsetof(WebCore::ImageBufferBackendParameters, pixelFormat)
        , offsetof(WebCore::ImageBufferBackendParameters, purpose)
    >::value);
    encoder << instance.logicalSize;
    encoder << instance.resolutionScale;
    encoder << instance.colorSpace;
    encoder << instance.pixelFormat;
    encoder << instance.purpose;
}

std::optional<WebCore::ImageBufferBackendParameters> ArgumentCoder<WebCore::ImageBufferBackendParameters>::decode(Decoder& decoder)
{
    auto logicalSize = decoder.decode<WebCore::FloatSize>();
    auto resolutionScale = decoder.decode<float>();
    auto colorSpace = decoder.decode<WebCore::DestinationColorSpace>();
    auto pixelFormat = decoder.decode<WebCore::PixelFormat>();
    auto purpose = decoder.decode<WebCore::RenderingPurpose>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ImageBufferBackendParameters {
            WTFMove(*logicalSize),
            WTFMove(*resolutionScale),
            WTFMove(*colorSpace),
            WTFMove(*pixelFormat),
            WTFMove(*purpose)
        }
    };
}

void ArgumentCoder<WebCore::ProcessIdentity>::encode(Encoder& encoder, WebCore::ProcessIdentity&& instance)
{
#if HAVE(TASK_IDENTITY_TOKEN)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_taskIdToken)>, MachSendRight>);
#endif
    struct ShouldBeSameSizeAsProcessIdentity : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ProcessIdentity>, false> {
#if HAVE(TASK_IDENTITY_TOKEN)
        MachSendRight m_taskIdToken;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsProcessIdentity) == sizeof(WebCore::ProcessIdentity));
    static_assert(MembersInCorrectOrder < 0
#if HAVE(TASK_IDENTITY_TOKEN)
        , offsetof(WebCore::ProcessIdentity, m_taskIdToken)
#endif
    >::value);
#if HAVE(TASK_IDENTITY_TOKEN)
    encoder << WTFMove(instance.m_taskIdToken);
#endif
}

std::optional<WebCore::ProcessIdentity> ArgumentCoder<WebCore::ProcessIdentity>::decode(Decoder& decoder)
{
#if HAVE(TASK_IDENTITY_TOKEN)
    auto m_taskIdToken = decoder.decode<MachSendRight>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ProcessIdentity {
#if HAVE(TASK_IDENTITY_TOKEN)
            WTFMove(*m_taskIdToken)
#endif
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::HSL>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::HSL& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsHSL : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::HSL>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHSL) == sizeof(WebCore::ColorInterpolationMethod::HSL));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::HSL, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::HSL>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::HSL& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsHSL : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::HSL>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHSL) == sizeof(WebCore::ColorInterpolationMethod::HSL));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::HSL, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

std::optional<WebCore::ColorInterpolationMethod::HSL> ArgumentCoder<WebCore::ColorInterpolationMethod::HSL>::decode(Decoder& decoder)
{
    auto hueInterpolationMethod = decoder.decode<WebCore::HueInterpolationMethod>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::HSL {
            WTFMove(*hueInterpolationMethod)
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::HWB>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::HWB& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsHWB : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::HWB>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHWB) == sizeof(WebCore::ColorInterpolationMethod::HWB));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::HWB, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::HWB>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::HWB& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsHWB : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::HWB>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHWB) == sizeof(WebCore::ColorInterpolationMethod::HWB));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::HWB, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

std::optional<WebCore::ColorInterpolationMethod::HWB> ArgumentCoder<WebCore::ColorInterpolationMethod::HWB>::decode(Decoder& decoder)
{
    auto hueInterpolationMethod = decoder.decode<WebCore::HueInterpolationMethod>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::HWB {
            WTFMove(*hueInterpolationMethod)
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::LCH>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::LCH& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsLCH : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::LCH>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLCH) == sizeof(WebCore::ColorInterpolationMethod::LCH));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::LCH, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::LCH>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::LCH& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsLCH : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::LCH>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLCH) == sizeof(WebCore::ColorInterpolationMethod::LCH));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::LCH, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

std::optional<WebCore::ColorInterpolationMethod::LCH> ArgumentCoder<WebCore::ColorInterpolationMethod::LCH>::decode(Decoder& decoder)
{
    auto hueInterpolationMethod = decoder.decode<WebCore::HueInterpolationMethod>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::LCH {
            WTFMove(*hueInterpolationMethod)
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::Lab>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::Lab& instance)
{
    struct ShouldBeSameSizeAsLab : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::Lab>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsLab) == sizeof(WebCore::ColorInterpolationMethod::Lab));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::Lab>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::Lab& instance)
{
    struct ShouldBeSameSizeAsLab : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::Lab>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsLab) == sizeof(WebCore::ColorInterpolationMethod::Lab));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::ColorInterpolationMethod::Lab> ArgumentCoder<WebCore::ColorInterpolationMethod::Lab>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::Lab {
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::OKLCH>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::OKLCH& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsOKLCH : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::OKLCH>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsOKLCH) == sizeof(WebCore::ColorInterpolationMethod::OKLCH));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::OKLCH, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::OKLCH>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::OKLCH& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hueInterpolationMethod)>, WebCore::HueInterpolationMethod>);
    struct ShouldBeSameSizeAsOKLCH : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::OKLCH>, false> {
        WebCore::HueInterpolationMethod hueInterpolationMethod;
    };
    static_assert(sizeof(ShouldBeSameSizeAsOKLCH) == sizeof(WebCore::ColorInterpolationMethod::OKLCH));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod::OKLCH, hueInterpolationMethod)
    >::value);
    encoder << instance.hueInterpolationMethod;
}

std::optional<WebCore::ColorInterpolationMethod::OKLCH> ArgumentCoder<WebCore::ColorInterpolationMethod::OKLCH>::decode(Decoder& decoder)
{
    auto hueInterpolationMethod = decoder.decode<WebCore::HueInterpolationMethod>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::OKLCH {
            WTFMove(*hueInterpolationMethod)
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::OKLab>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::OKLab& instance)
{
    struct ShouldBeSameSizeAsOKLab : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::OKLab>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsOKLab) == sizeof(WebCore::ColorInterpolationMethod::OKLab));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::OKLab>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::OKLab& instance)
{
    struct ShouldBeSameSizeAsOKLab : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::OKLab>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsOKLab) == sizeof(WebCore::ColorInterpolationMethod::OKLab));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::ColorInterpolationMethod::OKLab> ArgumentCoder<WebCore::ColorInterpolationMethod::OKLab>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::OKLab {
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::SRGB>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::SRGB& instance)
{
    struct ShouldBeSameSizeAsSRGB : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::SRGB>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsSRGB) == sizeof(WebCore::ColorInterpolationMethod::SRGB));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::SRGB>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::SRGB& instance)
{
    struct ShouldBeSameSizeAsSRGB : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::SRGB>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsSRGB) == sizeof(WebCore::ColorInterpolationMethod::SRGB));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::ColorInterpolationMethod::SRGB> ArgumentCoder<WebCore::ColorInterpolationMethod::SRGB>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::SRGB {
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::SRGBLinear>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::SRGBLinear& instance)
{
    struct ShouldBeSameSizeAsSRGBLinear : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::SRGBLinear>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsSRGBLinear) == sizeof(WebCore::ColorInterpolationMethod::SRGBLinear));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::SRGBLinear>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::SRGBLinear& instance)
{
    struct ShouldBeSameSizeAsSRGBLinear : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::SRGBLinear>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsSRGBLinear) == sizeof(WebCore::ColorInterpolationMethod::SRGBLinear));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::ColorInterpolationMethod::SRGBLinear> ArgumentCoder<WebCore::ColorInterpolationMethod::SRGBLinear>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::SRGBLinear {
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD50>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::XYZD50& instance)
{
    struct ShouldBeSameSizeAsXYZD50 : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::XYZD50>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsXYZD50) == sizeof(WebCore::ColorInterpolationMethod::XYZD50));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD50>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::XYZD50& instance)
{
    struct ShouldBeSameSizeAsXYZD50 : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::XYZD50>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsXYZD50) == sizeof(WebCore::ColorInterpolationMethod::XYZD50));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::ColorInterpolationMethod::XYZD50> ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD50>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::XYZD50 {
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD65>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod::XYZD65& instance)
{
    struct ShouldBeSameSizeAsXYZD65 : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::XYZD65>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsXYZD65) == sizeof(WebCore::ColorInterpolationMethod::XYZD65));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD65>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod::XYZD65& instance)
{
    struct ShouldBeSameSizeAsXYZD65 : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod::XYZD65>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsXYZD65) == sizeof(WebCore::ColorInterpolationMethod::XYZD65));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebCore::ColorInterpolationMethod::XYZD65> ArgumentCoder<WebCore::ColorInterpolationMethod::XYZD65>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod::XYZD65 {
        }
    };
}

void ArgumentCoder<WebCore::ColorInterpolationMethod>::encode(Encoder& encoder, const WebCore::ColorInterpolationMethod& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, std::variant<WebCore::ColorInterpolationMethod::HSL, WebCore::ColorInterpolationMethod::HWB, WebCore::ColorInterpolationMethod::LCH, WebCore::ColorInterpolationMethod::Lab, WebCore::ColorInterpolationMethod::OKLCH, WebCore::ColorInterpolationMethod::OKLab, WebCore::ColorInterpolationMethod::SRGB, WebCore::ColorInterpolationMethod::SRGBLinear, WebCore::ColorInterpolationMethod::XYZD50, WebCore::ColorInterpolationMethod::XYZD65>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaPremultiplication)>, WebCore::AlphaPremultiplication>);
    struct ShouldBeSameSizeAsColorInterpolationMethod : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod>, false> {
        std::variant<WebCore::ColorInterpolationMethod::HSL, WebCore::ColorInterpolationMethod::HWB, WebCore::ColorInterpolationMethod::LCH, WebCore::ColorInterpolationMethod::Lab, WebCore::ColorInterpolationMethod::OKLCH, WebCore::ColorInterpolationMethod::OKLab, WebCore::ColorInterpolationMethod::SRGB, WebCore::ColorInterpolationMethod::SRGBLinear, WebCore::ColorInterpolationMethod::XYZD50, WebCore::ColorInterpolationMethod::XYZD65> colorSpace;
        WebCore::AlphaPremultiplication alphaPremultiplication;
    };
    static_assert(sizeof(ShouldBeSameSizeAsColorInterpolationMethod) == sizeof(WebCore::ColorInterpolationMethod));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod, colorSpace)
        , offsetof(WebCore::ColorInterpolationMethod, alphaPremultiplication)
    >::value);
    encoder << instance.colorSpace;
    encoder << instance.alphaPremultiplication;
}

void ArgumentCoder<WebCore::ColorInterpolationMethod>::encode(StreamConnectionEncoder& encoder, const WebCore::ColorInterpolationMethod& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, std::variant<WebCore::ColorInterpolationMethod::HSL, WebCore::ColorInterpolationMethod::HWB, WebCore::ColorInterpolationMethod::LCH, WebCore::ColorInterpolationMethod::Lab, WebCore::ColorInterpolationMethod::OKLCH, WebCore::ColorInterpolationMethod::OKLab, WebCore::ColorInterpolationMethod::SRGB, WebCore::ColorInterpolationMethod::SRGBLinear, WebCore::ColorInterpolationMethod::XYZD50, WebCore::ColorInterpolationMethod::XYZD65>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaPremultiplication)>, WebCore::AlphaPremultiplication>);
    struct ShouldBeSameSizeAsColorInterpolationMethod : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ColorInterpolationMethod>, false> {
        std::variant<WebCore::ColorInterpolationMethod::HSL, WebCore::ColorInterpolationMethod::HWB, WebCore::ColorInterpolationMethod::LCH, WebCore::ColorInterpolationMethod::Lab, WebCore::ColorInterpolationMethod::OKLCH, WebCore::ColorInterpolationMethod::OKLab, WebCore::ColorInterpolationMethod::SRGB, WebCore::ColorInterpolationMethod::SRGBLinear, WebCore::ColorInterpolationMethod::XYZD50, WebCore::ColorInterpolationMethod::XYZD65> colorSpace;
        WebCore::AlphaPremultiplication alphaPremultiplication;
    };
    static_assert(sizeof(ShouldBeSameSizeAsColorInterpolationMethod) == sizeof(WebCore::ColorInterpolationMethod));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ColorInterpolationMethod, colorSpace)
        , offsetof(WebCore::ColorInterpolationMethod, alphaPremultiplication)
    >::value);
    encoder << instance.colorSpace;
    encoder << instance.alphaPremultiplication;
}

std::optional<WebCore::ColorInterpolationMethod> ArgumentCoder<WebCore::ColorInterpolationMethod>::decode(Decoder& decoder)
{
    auto colorSpace = decoder.decode<std::variant<WebCore::ColorInterpolationMethod::HSL, WebCore::ColorInterpolationMethod::HWB, WebCore::ColorInterpolationMethod::LCH, WebCore::ColorInterpolationMethod::Lab, WebCore::ColorInterpolationMethod::OKLCH, WebCore::ColorInterpolationMethod::OKLab, WebCore::ColorInterpolationMethod::SRGB, WebCore::ColorInterpolationMethod::SRGBLinear, WebCore::ColorInterpolationMethod::XYZD50, WebCore::ColorInterpolationMethod::XYZD65>>();
    auto alphaPremultiplication = decoder.decode<WebCore::AlphaPremultiplication>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ColorInterpolationMethod {
            WTFMove(*colorSpace),
            WTFMove(*alphaPremultiplication)
        }
    };
}

void ArgumentCoder<WebCore::CompositeMode>::encode(Encoder& encoder, const WebCore::CompositeMode& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.operation)>, WebCore::CompositeOperator>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blendMode)>, WebCore::BlendMode>);
    struct ShouldBeSameSizeAsCompositeMode : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CompositeMode>, false> {
        WebCore::CompositeOperator operation;
        WebCore::BlendMode blendMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCompositeMode) == sizeof(WebCore::CompositeMode));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CompositeMode, operation)
        , offsetof(WebCore::CompositeMode, blendMode)
    >::value);
    encoder << instance.operation;
    encoder << instance.blendMode;
}

void ArgumentCoder<WebCore::CompositeMode>::encode(StreamConnectionEncoder& encoder, const WebCore::CompositeMode& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.operation)>, WebCore::CompositeOperator>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blendMode)>, WebCore::BlendMode>);
    struct ShouldBeSameSizeAsCompositeMode : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CompositeMode>, false> {
        WebCore::CompositeOperator operation;
        WebCore::BlendMode blendMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCompositeMode) == sizeof(WebCore::CompositeMode));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CompositeMode, operation)
        , offsetof(WebCore::CompositeMode, blendMode)
    >::value);
    encoder << instance.operation;
    encoder << instance.blendMode;
}

std::optional<WebCore::CompositeMode> ArgumentCoder<WebCore::CompositeMode>::decode(Decoder& decoder)
{
    auto operation = decoder.decode<WebCore::CompositeOperator>();
    auto blendMode = decoder.decode<WebCore::BlendMode>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CompositeMode {
            WTFMove(*operation),
            WTFMove(*blendMode)
        }
    };
}

void ArgumentCoder<WebCore::DocumentMarkerLineStyle>::encode(Encoder& encoder, const WebCore::DocumentMarkerLineStyle& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mode)>, WebCore::DocumentMarkerLineStyleMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    struct ShouldBeSameSizeAsDocumentMarkerLineStyle : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DocumentMarkerLineStyle>, false> {
        WebCore::DocumentMarkerLineStyleMode mode;
        WebCore::Color color;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDocumentMarkerLineStyle) == sizeof(WebCore::DocumentMarkerLineStyle));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DocumentMarkerLineStyle, mode)
        , offsetof(WebCore::DocumentMarkerLineStyle, color)
    >::value);
    encoder << instance.mode;
    encoder << instance.color;
}

void ArgumentCoder<WebCore::DocumentMarkerLineStyle>::encode(StreamConnectionEncoder& encoder, const WebCore::DocumentMarkerLineStyle& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mode)>, WebCore::DocumentMarkerLineStyleMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    struct ShouldBeSameSizeAsDocumentMarkerLineStyle : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::DocumentMarkerLineStyle>, false> {
        WebCore::DocumentMarkerLineStyleMode mode;
        WebCore::Color color;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDocumentMarkerLineStyle) == sizeof(WebCore::DocumentMarkerLineStyle));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DocumentMarkerLineStyle, mode)
        , offsetof(WebCore::DocumentMarkerLineStyle, color)
    >::value);
    encoder << instance.mode;
    encoder << instance.color;
}

std::optional<WebCore::DocumentMarkerLineStyle> ArgumentCoder<WebCore::DocumentMarkerLineStyle>::decode(Decoder& decoder)
{
    auto mode = decoder.decode<WebCore::DocumentMarkerLineStyleMode>();
    auto color = decoder.decode<WebCore::Color>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DocumentMarkerLineStyle {
            WTFMove(*mode),
            WTFMove(*color)
        }
    };
}

void ArgumentCoder<WebCore::GraphicsDropShadow>::encode(Encoder& encoder, const WebCore::GraphicsDropShadow& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusMode)>, WebCore::ShadowRadiusMode>);
    struct ShouldBeSameSizeAsGraphicsDropShadow : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GraphicsDropShadow>, false> {
        WebCore::FloatSize offset;
        float radius;
        WebCore::Color color;
        WebCore::ShadowRadiusMode radiusMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGraphicsDropShadow) == sizeof(WebCore::GraphicsDropShadow));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GraphicsDropShadow, offset)
        , offsetof(WebCore::GraphicsDropShadow, radius)
        , offsetof(WebCore::GraphicsDropShadow, color)
        , offsetof(WebCore::GraphicsDropShadow, radiusMode)
    >::value);
    encoder << instance.offset;
    encoder << instance.radius;
    encoder << instance.color;
    encoder << instance.radiusMode;
}

void ArgumentCoder<WebCore::GraphicsDropShadow>::encode(StreamConnectionEncoder& encoder, const WebCore::GraphicsDropShadow& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusMode)>, WebCore::ShadowRadiusMode>);
    struct ShouldBeSameSizeAsGraphicsDropShadow : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GraphicsDropShadow>, false> {
        WebCore::FloatSize offset;
        float radius;
        WebCore::Color color;
        WebCore::ShadowRadiusMode radiusMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGraphicsDropShadow) == sizeof(WebCore::GraphicsDropShadow));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GraphicsDropShadow, offset)
        , offsetof(WebCore::GraphicsDropShadow, radius)
        , offsetof(WebCore::GraphicsDropShadow, color)
        , offsetof(WebCore::GraphicsDropShadow, radiusMode)
    >::value);
    encoder << instance.offset;
    encoder << instance.radius;
    encoder << instance.color;
    encoder << instance.radiusMode;
}

std::optional<WebCore::GraphicsDropShadow> ArgumentCoder<WebCore::GraphicsDropShadow>::decode(Decoder& decoder)
{
    auto offset = decoder.decode<WebCore::FloatSize>();
    auto radius = decoder.decode<float>();
    auto color = decoder.decode<WebCore::Color>();
    auto radiusMode = decoder.decode<WebCore::ShadowRadiusMode>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GraphicsDropShadow {
            WTFMove(*offset),
            WTFMove(*radius),
            WTFMove(*color),
            WTFMove(*radiusMode)
        }
    };
}

#if ENABLE(VIDEO)
void ArgumentCoder<WebCore::GenericCueData>::encode(Encoder& encoder, const WebCore::GenericCueData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_uniqueId)>, WebCore::InbandGenericCueIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_startTime)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_endTime)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_id)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_content)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_fontName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_line)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_position)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_size)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_baseFontSize)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_relativeFontSize)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_foregroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_backgroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_highlightColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_align)>, WebCore::GenericCueData::Alignment>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_status)>, WebCore::GenericCueData::Status>);
    struct ShouldBeSameSizeAsGenericCueData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GenericCueData>, false> {
        WebCore::InbandGenericCueIdentifier m_uniqueId;
        MediaTime m_startTime;
        MediaTime m_endTime;
        AtomString m_id;
        String m_content;
        String m_fontName;
        double m_line;
        double m_position;
        double m_size;
        double m_baseFontSize;
        double m_relativeFontSize;
        WebCore::Color m_foregroundColor;
        WebCore::Color m_backgroundColor;
        WebCore::Color m_highlightColor;
        WebCore::GenericCueData::Alignment m_align;
        WebCore::GenericCueData::Status m_status;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGenericCueData) == sizeof(WebCore::GenericCueData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GenericCueData, m_uniqueId)
        , offsetof(WebCore::GenericCueData, m_startTime)
        , offsetof(WebCore::GenericCueData, m_endTime)
        , offsetof(WebCore::GenericCueData, m_id)
        , offsetof(WebCore::GenericCueData, m_content)
        , offsetof(WebCore::GenericCueData, m_fontName)
        , offsetof(WebCore::GenericCueData, m_line)
        , offsetof(WebCore::GenericCueData, m_position)
        , offsetof(WebCore::GenericCueData, m_size)
        , offsetof(WebCore::GenericCueData, m_baseFontSize)
        , offsetof(WebCore::GenericCueData, m_relativeFontSize)
        , offsetof(WebCore::GenericCueData, m_foregroundColor)
        , offsetof(WebCore::GenericCueData, m_backgroundColor)
        , offsetof(WebCore::GenericCueData, m_highlightColor)
        , offsetof(WebCore::GenericCueData, m_align)
        , offsetof(WebCore::GenericCueData, m_status)
    >::value);
    encoder << instance.m_uniqueId;
    encoder << instance.m_startTime;
    encoder << instance.m_endTime;
    encoder << instance.m_id;
    encoder << instance.m_content;
    encoder << instance.m_fontName;
    encoder << instance.m_line;
    encoder << instance.m_position;
    encoder << instance.m_size;
    encoder << instance.m_baseFontSize;
    encoder << instance.m_relativeFontSize;
    encoder << instance.m_foregroundColor;
    encoder << instance.m_backgroundColor;
    encoder << instance.m_highlightColor;
    encoder << instance.m_align;
    encoder << instance.m_status;
}

void ArgumentCoder<WebCore::GenericCueData>::encode(StreamConnectionEncoder& encoder, const WebCore::GenericCueData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_uniqueId)>, WebCore::InbandGenericCueIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_startTime)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_endTime)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_id)>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_content)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_fontName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_line)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_position)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_size)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_baseFontSize)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_relativeFontSize)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_foregroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_backgroundColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_highlightColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_align)>, WebCore::GenericCueData::Alignment>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_status)>, WebCore::GenericCueData::Status>);
    struct ShouldBeSameSizeAsGenericCueData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GenericCueData>, false> {
        WebCore::InbandGenericCueIdentifier m_uniqueId;
        MediaTime m_startTime;
        MediaTime m_endTime;
        AtomString m_id;
        String m_content;
        String m_fontName;
        double m_line;
        double m_position;
        double m_size;
        double m_baseFontSize;
        double m_relativeFontSize;
        WebCore::Color m_foregroundColor;
        WebCore::Color m_backgroundColor;
        WebCore::Color m_highlightColor;
        WebCore::GenericCueData::Alignment m_align;
        WebCore::GenericCueData::Status m_status;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGenericCueData) == sizeof(WebCore::GenericCueData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GenericCueData, m_uniqueId)
        , offsetof(WebCore::GenericCueData, m_startTime)
        , offsetof(WebCore::GenericCueData, m_endTime)
        , offsetof(WebCore::GenericCueData, m_id)
        , offsetof(WebCore::GenericCueData, m_content)
        , offsetof(WebCore::GenericCueData, m_fontName)
        , offsetof(WebCore::GenericCueData, m_line)
        , offsetof(WebCore::GenericCueData, m_position)
        , offsetof(WebCore::GenericCueData, m_size)
        , offsetof(WebCore::GenericCueData, m_baseFontSize)
        , offsetof(WebCore::GenericCueData, m_relativeFontSize)
        , offsetof(WebCore::GenericCueData, m_foregroundColor)
        , offsetof(WebCore::GenericCueData, m_backgroundColor)
        , offsetof(WebCore::GenericCueData, m_highlightColor)
        , offsetof(WebCore::GenericCueData, m_align)
        , offsetof(WebCore::GenericCueData, m_status)
    >::value);
    encoder << instance.m_uniqueId;
    encoder << instance.m_startTime;
    encoder << instance.m_endTime;
    encoder << instance.m_id;
    encoder << instance.m_content;
    encoder << instance.m_fontName;
    encoder << instance.m_line;
    encoder << instance.m_position;
    encoder << instance.m_size;
    encoder << instance.m_baseFontSize;
    encoder << instance.m_relativeFontSize;
    encoder << instance.m_foregroundColor;
    encoder << instance.m_backgroundColor;
    encoder << instance.m_highlightColor;
    encoder << instance.m_align;
    encoder << instance.m_status;
}

std::optional<WebCore::GenericCueData> ArgumentCoder<WebCore::GenericCueData>::decode(Decoder& decoder)
{
    auto m_uniqueId = decoder.decode<WebCore::InbandGenericCueIdentifier>();
    auto m_startTime = decoder.decode<MediaTime>();
    auto m_endTime = decoder.decode<MediaTime>();
    auto m_id = decoder.decode<AtomString>();
    auto m_content = decoder.decode<String>();
    auto m_fontName = decoder.decode<String>();
    auto m_line = decoder.decode<double>();
    auto m_position = decoder.decode<double>();
    auto m_size = decoder.decode<double>();
    auto m_baseFontSize = decoder.decode<double>();
    auto m_relativeFontSize = decoder.decode<double>();
    auto m_foregroundColor = decoder.decode<WebCore::Color>();
    auto m_backgroundColor = decoder.decode<WebCore::Color>();
    auto m_highlightColor = decoder.decode<WebCore::Color>();
    auto m_align = decoder.decode<WebCore::GenericCueData::Alignment>();
    auto m_status = decoder.decode<WebCore::GenericCueData::Status>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GenericCueData {
            WTFMove(*m_uniqueId),
            WTFMove(*m_startTime),
            WTFMove(*m_endTime),
            WTFMove(*m_id),
            WTFMove(*m_content),
            WTFMove(*m_fontName),
            WTFMove(*m_line),
            WTFMove(*m_position),
            WTFMove(*m_size),
            WTFMove(*m_baseFontSize),
            WTFMove(*m_relativeFontSize),
            WTFMove(*m_foregroundColor),
            WTFMove(*m_backgroundColor),
            WTFMove(*m_highlightColor),
            WTFMove(*m_align),
            WTFMove(*m_status)
        }
    };
}

#endif

void ArgumentCoder<WebCore::ImageDecoderFrameInfo>::encode(Encoder& encoder, const WebCore::ImageDecoderFrameInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasAlpha)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration)>, Seconds>);
    struct ShouldBeSameSizeAsImageDecoderFrameInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ImageDecoderFrameInfo>, false> {
        bool hasAlpha;
        Seconds duration;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageDecoderFrameInfo) == sizeof(WebCore::ImageDecoderFrameInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ImageDecoderFrameInfo, hasAlpha)
        , offsetof(WebCore::ImageDecoderFrameInfo, duration)
    >::value);
    encoder << instance.hasAlpha;
    encoder << instance.duration;
}

void ArgumentCoder<WebCore::ImageDecoderFrameInfo>::encode(StreamConnectionEncoder& encoder, const WebCore::ImageDecoderFrameInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasAlpha)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration)>, Seconds>);
    struct ShouldBeSameSizeAsImageDecoderFrameInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ImageDecoderFrameInfo>, false> {
        bool hasAlpha;
        Seconds duration;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageDecoderFrameInfo) == sizeof(WebCore::ImageDecoderFrameInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ImageDecoderFrameInfo, hasAlpha)
        , offsetof(WebCore::ImageDecoderFrameInfo, duration)
    >::value);
    encoder << instance.hasAlpha;
    encoder << instance.duration;
}

std::optional<WebCore::ImageDecoderFrameInfo> ArgumentCoder<WebCore::ImageDecoderFrameInfo>::decode(Decoder& decoder)
{
    auto hasAlpha = decoder.decode<bool>();
    auto duration = decoder.decode<Seconds>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ImageDecoderFrameInfo {
            WTFMove(*hasAlpha),
            WTFMove(*duration)
        }
    };
}

void ArgumentCoder<WebCore::DisplayList::DrawDotsForDocumentMarker>::encode(Encoder& encoder, const WebCore::DisplayList::DrawDotsForDocumentMarker& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.style())>, WebCore::DocumentMarkerLineStyle>);
    encoder << instance.rect();
    encoder << instance.style();
}

void ArgumentCoder<WebCore::DisplayList::DrawDotsForDocumentMarker>::encode(StreamConnectionEncoder& encoder, const WebCore::DisplayList::DrawDotsForDocumentMarker& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect())>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.style())>, WebCore::DocumentMarkerLineStyle>);
    encoder << instance.rect();
    encoder << instance.style();
}

std::optional<WebCore::DisplayList::DrawDotsForDocumentMarker> ArgumentCoder<WebCore::DisplayList::DrawDotsForDocumentMarker>::decode(Decoder& decoder)
{
    auto rect = decoder.decode<WebCore::FloatRect>();
    auto style = decoder.decode<WebCore::DocumentMarkerLineStyle>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DisplayList::DrawDotsForDocumentMarker {
            WTFMove(*rect),
            WTFMove(*style)
        }
    };
}

void ArgumentCoder<WebCore::ImageOrientation>::encode(Encoder& encoder, const WebCore::ImageOrientation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.orientation())>, WebCore::ImageOrientation::Orientation>);
    encoder << instance.orientation();
}

void ArgumentCoder<WebCore::ImageOrientation>::encode(StreamConnectionEncoder& encoder, const WebCore::ImageOrientation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.orientation())>, WebCore::ImageOrientation::Orientation>);
    encoder << instance.orientation();
}

std::optional<WebCore::ImageOrientation> ArgumentCoder<WebCore::ImageOrientation>::decode(Decoder& decoder)
{
    auto orientation = decoder.decode<WebCore::ImageOrientation::Orientation>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ImageOrientation {
            WTFMove(*orientation)
        }
    };
}

void ArgumentCoder<WebCore::ImagePaintingOptions>::encode(Encoder& encoder, const WebCore::ImagePaintingOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_compositeOperator)>, WebCore::CompositeOperator>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_blendMode)>, WebCore::BlendMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_decodingMode)>, WebCore::DecodingMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_orientation)>, WebCore::ImageOrientation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_interpolationQuality)>, WebCore::InterpolationQuality>);
    struct ShouldBeSameSizeAsImagePaintingOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ImagePaintingOptions>, false> {
        WebCore::CompositeOperator m_compositeOperator;
        WebCore::BlendMode m_blendMode;
        WebCore::DecodingMode m_decodingMode;
        WebCore::ImageOrientation m_orientation;
        WebCore::InterpolationQuality m_interpolationQuality;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImagePaintingOptions) == sizeof(WebCore::ImagePaintingOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ImagePaintingOptions, m_compositeOperator)
        , offsetof(WebCore::ImagePaintingOptions, m_blendMode)
        , offsetof(WebCore::ImagePaintingOptions, m_decodingMode)
        , offsetof(WebCore::ImagePaintingOptions, m_orientation)
        , offsetof(WebCore::ImagePaintingOptions, m_interpolationQuality)
    >::value);
    encoder << instance.m_compositeOperator;
    encoder << instance.m_blendMode;
    encoder << instance.m_decodingMode;
    encoder << instance.m_orientation;
    encoder << instance.m_interpolationQuality;
}

void ArgumentCoder<WebCore::ImagePaintingOptions>::encode(StreamConnectionEncoder& encoder, const WebCore::ImagePaintingOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_compositeOperator)>, WebCore::CompositeOperator>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_blendMode)>, WebCore::BlendMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_decodingMode)>, WebCore::DecodingMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_orientation)>, WebCore::ImageOrientation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_interpolationQuality)>, WebCore::InterpolationQuality>);
    struct ShouldBeSameSizeAsImagePaintingOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ImagePaintingOptions>, false> {
        WebCore::CompositeOperator m_compositeOperator;
        WebCore::BlendMode m_blendMode;
        WebCore::DecodingMode m_decodingMode;
        WebCore::ImageOrientation m_orientation;
        WebCore::InterpolationQuality m_interpolationQuality;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImagePaintingOptions) == sizeof(WebCore::ImagePaintingOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ImagePaintingOptions, m_compositeOperator)
        , offsetof(WebCore::ImagePaintingOptions, m_blendMode)
        , offsetof(WebCore::ImagePaintingOptions, m_decodingMode)
        , offsetof(WebCore::ImagePaintingOptions, m_orientation)
        , offsetof(WebCore::ImagePaintingOptions, m_interpolationQuality)
    >::value);
    encoder << instance.m_compositeOperator;
    encoder << instance.m_blendMode;
    encoder << instance.m_decodingMode;
    encoder << instance.m_orientation;
    encoder << instance.m_interpolationQuality;
}

std::optional<WebCore::ImagePaintingOptions> ArgumentCoder<WebCore::ImagePaintingOptions>::decode(Decoder& decoder)
{
    auto m_compositeOperator = decoder.decode<WebCore::CompositeOperator>();
    auto m_blendMode = decoder.decode<WebCore::BlendMode>();
    auto m_decodingMode = decoder.decode<WebCore::DecodingMode>();
    auto m_orientation = decoder.decode<WebCore::ImageOrientation>();
    auto m_interpolationQuality = decoder.decode<WebCore::InterpolationQuality>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ImagePaintingOptions {
            WTFMove(*m_compositeOperator),
            WTFMove(*m_blendMode),
            WTFMove(*m_decodingMode),
            WTFMove(*m_orientation),
            WTFMove(*m_interpolationQuality)
        }
    };
}

void ArgumentCoder<WebCore::CanvasActivityRecord>::encode(Encoder& encoder, const WebCore::CanvasActivityRecord& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textWritten)>, HashSet<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wasDataRead)>, bool>);
    struct ShouldBeSameSizeAsCanvasActivityRecord : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CanvasActivityRecord>, false> {
        HashSet<String> textWritten;
        bool wasDataRead;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCanvasActivityRecord) == sizeof(WebCore::CanvasActivityRecord));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CanvasActivityRecord, textWritten)
        , offsetof(WebCore::CanvasActivityRecord, wasDataRead)
    >::value);
    encoder << instance.textWritten;
    encoder << instance.wasDataRead;
}

std::optional<WebCore::CanvasActivityRecord> ArgumentCoder<WebCore::CanvasActivityRecord>::decode(Decoder& decoder)
{
    auto textWritten = decoder.decode<HashSet<String>>();
    auto wasDataRead = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CanvasActivityRecord {
            WTFMove(*textWritten),
            WTFMove(*wasDataRead)
        }
    };
}

void ArgumentCoder<WebCore::PCM::AttributionSecondsUntilSendData>::encode(Encoder& encoder, const WebCore::PCM::AttributionSecondsUntilSendData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceSeconds)>, std::optional<Seconds>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.destinationSeconds)>, std::optional<Seconds>>);
    struct ShouldBeSameSizeAsAttributionSecondsUntilSendData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PCM::AttributionSecondsUntilSendData>, false> {
        std::optional<Seconds> sourceSeconds;
        std::optional<Seconds> destinationSeconds;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAttributionSecondsUntilSendData) == sizeof(WebCore::PCM::AttributionSecondsUntilSendData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PCM::AttributionSecondsUntilSendData, sourceSeconds)
        , offsetof(WebCore::PCM::AttributionSecondsUntilSendData, destinationSeconds)
    >::value);
    encoder << instance.sourceSeconds;
    encoder << instance.destinationSeconds;
}

std::optional<WebCore::PCM::AttributionSecondsUntilSendData> ArgumentCoder<WebCore::PCM::AttributionSecondsUntilSendData>::decode(Decoder& decoder)
{
    auto sourceSeconds = decoder.decode<std::optional<Seconds>>();
    auto destinationSeconds = decoder.decode<std::optional<Seconds>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PCM::AttributionSecondsUntilSendData {
            WTFMove(*sourceSeconds),
            WTFMove(*destinationSeconds)
        }
    };
}

void ArgumentCoder<WebCore::CrossOriginOpenerPolicy>::encode(Encoder& encoder, const WebCore::CrossOriginOpenerPolicy& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value)>, WebCore::CrossOriginOpenerPolicyValue>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.reportOnlyValue)>, WebCore::CrossOriginOpenerPolicyValue>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.reportingEndpoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.reportOnlyReportingEndpoint)>, String>);
    struct ShouldBeSameSizeAsCrossOriginOpenerPolicy : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CrossOriginOpenerPolicy>, false> {
        WebCore::CrossOriginOpenerPolicyValue value;
        WebCore::CrossOriginOpenerPolicyValue reportOnlyValue;
        String reportingEndpoint;
        String reportOnlyReportingEndpoint;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCrossOriginOpenerPolicy) == sizeof(WebCore::CrossOriginOpenerPolicy));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CrossOriginOpenerPolicy, value)
        , offsetof(WebCore::CrossOriginOpenerPolicy, reportOnlyValue)
        , offsetof(WebCore::CrossOriginOpenerPolicy, reportingEndpoint)
        , offsetof(WebCore::CrossOriginOpenerPolicy, reportOnlyReportingEndpoint)
    >::value);
    encoder << instance.value;
    encoder << instance.reportOnlyValue;
    encoder << instance.reportingEndpoint;
    encoder << instance.reportOnlyReportingEndpoint;
}

std::optional<WebCore::CrossOriginOpenerPolicy> ArgumentCoder<WebCore::CrossOriginOpenerPolicy>::decode(Decoder& decoder)
{
    auto value = decoder.decode<WebCore::CrossOriginOpenerPolicyValue>();
    auto reportOnlyValue = decoder.decode<WebCore::CrossOriginOpenerPolicyValue>();
    auto reportingEndpoint = decoder.decode<String>();
    auto reportOnlyReportingEndpoint = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CrossOriginOpenerPolicy {
            WTFMove(*value),
            WTFMove(*reportOnlyValue),
            WTFMove(*reportingEndpoint),
            WTFMove(*reportOnlyReportingEndpoint)
        }
    };
}

void ArgumentCoder<WebCore::CustomHeaderFields>::encode(Encoder& encoder, const WebCore::CustomHeaderFields& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fields)>, Vector<WebCore::HTTPHeaderField>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.thirdPartyDomains)>, Vector<String>>);
    struct ShouldBeSameSizeAsCustomHeaderFields : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CustomHeaderFields>, false> {
        Vector<WebCore::HTTPHeaderField> fields;
        Vector<String> thirdPartyDomains;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCustomHeaderFields) == sizeof(WebCore::CustomHeaderFields));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CustomHeaderFields, fields)
        , offsetof(WebCore::CustomHeaderFields, thirdPartyDomains)
    >::value);
    encoder << instance.fields;
    encoder << instance.thirdPartyDomains;
}

std::optional<WebCore::CustomHeaderFields> ArgumentCoder<WebCore::CustomHeaderFields>::decode(Decoder& decoder)
{
    auto fields = decoder.decode<Vector<WebCore::HTTPHeaderField>>();
    auto thirdPartyDomains = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CustomHeaderFields {
            WTFMove(*fields),
            WTFMove(*thirdPartyDomains)
        }
    };
}

void ArgumentCoder<WebCore::HTTPHeaderField>::encode(Encoder& encoder, const WebCore::HTTPHeaderField& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.value())>, String>);
    encoder << instance.name();
    encoder << instance.value();
}

std::optional<WebCore::HTTPHeaderField> ArgumentCoder<WebCore::HTTPHeaderField>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto value = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::HTTPHeaderField::create(
            WTFMove(*name),
            WTFMove(*value)
        )
    };
}

void ArgumentCoder<WebCore::NavigationRequester>::encode(Encoder& encoder, const WebCore::NavigationRequester& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.securityOrigin)>, Ref<WebCore::SecurityOrigin>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topOrigin)>, Ref<WebCore::SecurityOrigin>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.policyContainer)>, WebCore::PolicyContainer>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, std::optional<WebCore::FrameIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageID)>, std::optional<WebCore::PageIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentIdentifier)>, WebCore::ScriptExecutionContextIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sandboxFlags)>, WebCore::SandboxFlags>);
    struct ShouldBeSameSizeAsNavigationRequester : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::NavigationRequester>, false> {
        URL url;
        Ref<WebCore::SecurityOrigin> securityOrigin;
        Ref<WebCore::SecurityOrigin> topOrigin;
        WebCore::PolicyContainer policyContainer;
        std::optional<WebCore::FrameIdentifier> frameID;
        std::optional<WebCore::PageIdentifier> pageID;
        WebCore::ScriptExecutionContextIdentifier documentIdentifier;
        WebCore::SandboxFlags sandboxFlags;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNavigationRequester) == sizeof(WebCore::NavigationRequester));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::NavigationRequester, url)
        , offsetof(WebCore::NavigationRequester, securityOrigin)
        , offsetof(WebCore::NavigationRequester, topOrigin)
        , offsetof(WebCore::NavigationRequester, policyContainer)
        , offsetof(WebCore::NavigationRequester, frameID)
        , offsetof(WebCore::NavigationRequester, pageID)
        , offsetof(WebCore::NavigationRequester, documentIdentifier)
        , offsetof(WebCore::NavigationRequester, sandboxFlags)
    >::value);
    encoder << instance.url;
    encoder << instance.securityOrigin;
    encoder << instance.topOrigin;
    encoder << instance.policyContainer;
    encoder << instance.frameID;
    encoder << instance.pageID;
    encoder << instance.documentIdentifier;
    encoder << instance.sandboxFlags;
}

std::optional<WebCore::NavigationRequester> ArgumentCoder<WebCore::NavigationRequester>::decode(Decoder& decoder)
{
    auto url = decoder.decode<URL>();
    auto securityOrigin = decoder.decode<Ref<WebCore::SecurityOrigin>>();
    auto topOrigin = decoder.decode<Ref<WebCore::SecurityOrigin>>();
    auto policyContainer = decoder.decode<WebCore::PolicyContainer>();
    auto frameID = decoder.decode<std::optional<WebCore::FrameIdentifier>>();
    auto pageID = decoder.decode<std::optional<WebCore::PageIdentifier>>();
    auto documentIdentifier = decoder.decode<WebCore::ScriptExecutionContextIdentifier>();
    auto sandboxFlags = decoder.decode<WebCore::SandboxFlags>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::NavigationRequester {
            WTFMove(*url),
            WTFMove(*securityOrigin),
            WTFMove(*topOrigin),
            WTFMove(*policyContainer),
            WTFMove(*frameID),
            WTFMove(*pageID),
            WTFMove(*documentIdentifier),
            WTFMove(*sandboxFlags)
        }
    };
}

void ArgumentCoder<WebCore::PolicyContainer>::encode(Encoder& encoder, const WebCore::PolicyContainer& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSecurityPolicyResponseHeaders)>, WebCore::ContentSecurityPolicyResponseHeaders>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginEmbedderPolicy)>, WebCore::CrossOriginEmbedderPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginOpenerPolicy)>, WebCore::CrossOriginOpenerPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrerPolicy)>, WebCore::ReferrerPolicy>);
    struct ShouldBeSameSizeAsPolicyContainer : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PolicyContainer>, false> {
        WebCore::ContentSecurityPolicyResponseHeaders contentSecurityPolicyResponseHeaders;
        WebCore::CrossOriginEmbedderPolicy crossOriginEmbedderPolicy;
        WebCore::CrossOriginOpenerPolicy crossOriginOpenerPolicy;
        WebCore::ReferrerPolicy referrerPolicy;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPolicyContainer) == sizeof(WebCore::PolicyContainer));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PolicyContainer, contentSecurityPolicyResponseHeaders)
        , offsetof(WebCore::PolicyContainer, crossOriginEmbedderPolicy)
        , offsetof(WebCore::PolicyContainer, crossOriginOpenerPolicy)
        , offsetof(WebCore::PolicyContainer, referrerPolicy)
    >::value);
    encoder << instance.contentSecurityPolicyResponseHeaders;
    encoder << instance.crossOriginEmbedderPolicy;
    encoder << instance.crossOriginOpenerPolicy;
    encoder << instance.referrerPolicy;
}

std::optional<WebCore::PolicyContainer> ArgumentCoder<WebCore::PolicyContainer>::decode(Decoder& decoder)
{
    auto contentSecurityPolicyResponseHeaders = decoder.decode<WebCore::ContentSecurityPolicyResponseHeaders>();
    auto crossOriginEmbedderPolicy = decoder.decode<WebCore::CrossOriginEmbedderPolicy>();
    auto crossOriginOpenerPolicy = decoder.decode<WebCore::CrossOriginOpenerPolicy>();
    auto referrerPolicy = decoder.decode<WebCore::ReferrerPolicy>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PolicyContainer {
            WTFMove(*contentSecurityPolicyResponseHeaders),
            WTFMove(*crossOriginEmbedderPolicy),
            WTFMove(*crossOriginOpenerPolicy),
            WTFMove(*referrerPolicy)
        }
    };
}

void ArgumentCoder<WebCore::SubstituteData>::encode(Encoder& encoder, const WebCore::SubstituteData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.content())>, RefPtr<WebCore::FragmentedSharedBuffer>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failingURL())>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.response())>, WebCore::ResourceResponse>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRevealToSessionHistory())>, WebCore::SubstituteData::SessionHistoryVisibility>);
    encoder << instance.content();
    encoder << instance.failingURL();
    encoder << instance.response();
    encoder << instance.shouldRevealToSessionHistory();
}

std::optional<WebCore::SubstituteData> ArgumentCoder<WebCore::SubstituteData>::decode(Decoder& decoder)
{
    auto content = decoder.decode<RefPtr<WebCore::FragmentedSharedBuffer>>();
    auto failingURL = decoder.decode<URL>();
    auto response = decoder.decode<WebCore::ResourceResponse>();
    auto shouldRevealToSessionHistory = decoder.decode<WebCore::SubstituteData::SessionHistoryVisibility>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SubstituteData {
            WTFMove(*content),
            WTFMove(*failingURL),
            WTFMove(*response),
            WTFMove(*shouldRevealToSessionHistory)
        }
    };
}

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerClientData>::encode(Encoder& encoder, const WebCore::ServiceWorkerClientData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::ScriptExecutionContextIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ServiceWorkerClientType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameType)>, WebCore::ServiceWorkerClientFrameType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ownerURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageIdentifier)>, std::optional<WebCore::PageIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameIdentifier)>, std::optional<WebCore::FrameIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastNavigationWasAppInitiated)>, WebCore::LastNavigationWasAppInitiated>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isVisible)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFocused)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.focusOrder)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ancestorOrigins)>, Vector<String>>);
    struct ShouldBeSameSizeAsServiceWorkerClientData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerClientData>, false> {
        WebCore::ScriptExecutionContextIdentifier identifier;
        WebCore::ServiceWorkerClientType type;
        WebCore::ServiceWorkerClientFrameType frameType;
        URL url;
        URL ownerURL;
        std::optional<WebCore::PageIdentifier> pageIdentifier;
        std::optional<WebCore::FrameIdentifier> frameIdentifier;
        WebCore::LastNavigationWasAppInitiated lastNavigationWasAppInitiated;
        bool isVisible;
        bool isFocused;
        uint64_t focusOrder;
        Vector<String> ancestorOrigins;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerClientData) == sizeof(WebCore::ServiceWorkerClientData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerClientData, identifier)
        , offsetof(WebCore::ServiceWorkerClientData, type)
        , offsetof(WebCore::ServiceWorkerClientData, frameType)
        , offsetof(WebCore::ServiceWorkerClientData, url)
        , offsetof(WebCore::ServiceWorkerClientData, ownerURL)
        , offsetof(WebCore::ServiceWorkerClientData, pageIdentifier)
        , offsetof(WebCore::ServiceWorkerClientData, frameIdentifier)
        , offsetof(WebCore::ServiceWorkerClientData, lastNavigationWasAppInitiated)
        , offsetof(WebCore::ServiceWorkerClientData, isVisible)
        , offsetof(WebCore::ServiceWorkerClientData, isFocused)
        , offsetof(WebCore::ServiceWorkerClientData, focusOrder)
        , offsetof(WebCore::ServiceWorkerClientData, ancestorOrigins)
    >::value);
    encoder << instance.identifier;
    encoder << instance.type;
    encoder << instance.frameType;
    encoder << instance.url;
    encoder << instance.ownerURL;
    encoder << instance.pageIdentifier;
    encoder << instance.frameIdentifier;
    encoder << instance.lastNavigationWasAppInitiated;
    encoder << instance.isVisible;
    encoder << instance.isFocused;
    encoder << instance.focusOrder;
    encoder << instance.ancestorOrigins;
}

std::optional<WebCore::ServiceWorkerClientData> ArgumentCoder<WebCore::ServiceWorkerClientData>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::ScriptExecutionContextIdentifier>();
    auto type = decoder.decode<WebCore::ServiceWorkerClientType>();
    auto frameType = decoder.decode<WebCore::ServiceWorkerClientFrameType>();
    auto url = decoder.decode<URL>();
    auto ownerURL = decoder.decode<URL>();
    auto pageIdentifier = decoder.decode<std::optional<WebCore::PageIdentifier>>();
    auto frameIdentifier = decoder.decode<std::optional<WebCore::FrameIdentifier>>();
    auto lastNavigationWasAppInitiated = decoder.decode<WebCore::LastNavigationWasAppInitiated>();
    auto isVisible = decoder.decode<bool>();
    auto isFocused = decoder.decode<bool>();
    auto focusOrder = decoder.decode<uint64_t>();
    auto ancestorOrigins = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerClientData {
            WTFMove(*identifier),
            WTFMove(*type),
            WTFMove(*frameType),
            WTFMove(*url),
            WTFMove(*ownerURL),
            WTFMove(*pageIdentifier),
            WTFMove(*frameIdentifier),
            WTFMove(*lastNavigationWasAppInitiated),
            WTFMove(*isVisible),
            WTFMove(*isFocused),
            WTFMove(*focusOrder),
            WTFMove(*ancestorOrigins)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerClientPendingMessage>::encode(Encoder& encoder, const WebCore::ServiceWorkerClientPendingMessage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, WebCore::MessageWithMessagePorts>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceData)>, WebCore::ServiceWorkerData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceOrigin)>, String>);
    struct ShouldBeSameSizeAsServiceWorkerClientPendingMessage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerClientPendingMessage>, false> {
        WebCore::MessageWithMessagePorts message;
        WebCore::ServiceWorkerData sourceData;
        String sourceOrigin;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerClientPendingMessage) == sizeof(WebCore::ServiceWorkerClientPendingMessage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerClientPendingMessage, message)
        , offsetof(WebCore::ServiceWorkerClientPendingMessage, sourceData)
        , offsetof(WebCore::ServiceWorkerClientPendingMessage, sourceOrigin)
    >::value);
    encoder << instance.message;
    encoder << instance.sourceData;
    encoder << instance.sourceOrigin;
}

std::optional<WebCore::ServiceWorkerClientPendingMessage> ArgumentCoder<WebCore::ServiceWorkerClientPendingMessage>::decode(Decoder& decoder)
{
    auto message = decoder.decode<WebCore::MessageWithMessagePorts>();
    auto sourceData = decoder.decode<WebCore::ServiceWorkerData>();
    auto sourceOrigin = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerClientPendingMessage {
            WTFMove(*message),
            WTFMove(*sourceData),
            WTFMove(*sourceOrigin)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerClientQueryOptions>::encode(Encoder& encoder, const WebCore::ServiceWorkerClientQueryOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.includeUncontrolled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ServiceWorkerClientType>);
    struct ShouldBeSameSizeAsServiceWorkerClientQueryOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerClientQueryOptions>, false> {
        bool includeUncontrolled;
        WebCore::ServiceWorkerClientType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerClientQueryOptions) == sizeof(WebCore::ServiceWorkerClientQueryOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerClientQueryOptions, includeUncontrolled)
        , offsetof(WebCore::ServiceWorkerClientQueryOptions, type)
    >::value);
    encoder << instance.includeUncontrolled;
    encoder << instance.type;
}

std::optional<WebCore::ServiceWorkerClientQueryOptions> ArgumentCoder<WebCore::ServiceWorkerClientQueryOptions>::decode(Decoder& decoder)
{
    auto includeUncontrolled = decoder.decode<bool>();
    auto type = decoder.decode<WebCore::ServiceWorkerClientType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerClientQueryOptions {
            WTFMove(*includeUncontrolled),
            WTFMove(*type)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerData>::encode(Encoder& encoder, const WebCore::ServiceWorkerData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::ServiceWorkerIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrationIdentifier)>, WebCore::ServiceWorkerRegistrationIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scriptURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.state)>, WebCore::ServiceWorkerState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WorkerType>);
    struct ShouldBeSameSizeAsServiceWorkerData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerData>, false> {
        WebCore::ServiceWorkerIdentifier identifier;
        WebCore::ServiceWorkerRegistrationIdentifier registrationIdentifier;
        URL scriptURL;
        WebCore::ServiceWorkerState state;
        WebCore::WorkerType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerData) == sizeof(WebCore::ServiceWorkerData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerData, identifier)
        , offsetof(WebCore::ServiceWorkerData, registrationIdentifier)
        , offsetof(WebCore::ServiceWorkerData, scriptURL)
        , offsetof(WebCore::ServiceWorkerData, state)
        , offsetof(WebCore::ServiceWorkerData, type)
    >::value);
    encoder << instance.identifier;
    encoder << instance.registrationIdentifier;
    encoder << instance.scriptURL;
    encoder << instance.state;
    encoder << instance.type;
}

std::optional<WebCore::ServiceWorkerData> ArgumentCoder<WebCore::ServiceWorkerData>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::ServiceWorkerIdentifier>();
    auto registrationIdentifier = decoder.decode<WebCore::ServiceWorkerRegistrationIdentifier>();
    auto scriptURL = decoder.decode<URL>();
    auto state = decoder.decode<WebCore::ServiceWorkerState>();
    auto type = decoder.decode<WebCore::WorkerType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerData {
            WTFMove(*identifier),
            WTFMove(*registrationIdentifier),
            WTFMove(*scriptURL),
            WTFMove(*state),
            WTFMove(*type)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerJobData>::encode(Encoder& encoder, const WebCore::ServiceWorkerJobData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier())>, WebCore::ServiceWorkerJobDataIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scriptURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientCreationURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topOrigin)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scopeURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceContext)>, WebCore::ServiceWorkerOrClientIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.workerType)>, WebCore::WorkerType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ServiceWorkerJobType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.domainForCachePartition)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isFromServiceWorkerPage)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrationOptions)>, std::optional<WebCore::ServiceWorkerRegistrationOptions>>);
    encoder << instance.identifier();
    encoder << instance.scriptURL;
    encoder << instance.clientCreationURL;
    encoder << instance.topOrigin;
    encoder << instance.scopeURL;
    encoder << instance.sourceContext;
    encoder << instance.workerType;
    encoder << instance.type;
    encoder << instance.domainForCachePartition;
    encoder << instance.isFromServiceWorkerPage;
    encoder << instance.registrationOptions;
}

std::optional<WebCore::ServiceWorkerJobData> ArgumentCoder<WebCore::ServiceWorkerJobData>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<WebCore::ServiceWorkerJobDataIdentifier>();
    auto scriptURL = decoder.decode<URL>();
    auto clientCreationURL = decoder.decode<URL>();
    auto topOrigin = decoder.decode<WebCore::SecurityOriginData>();
    auto scopeURL = decoder.decode<URL>();
    auto sourceContext = decoder.decode<WebCore::ServiceWorkerOrClientIdentifier>();
    auto workerType = decoder.decode<WebCore::WorkerType>();
    auto type = decoder.decode<WebCore::ServiceWorkerJobType>();
    auto domainForCachePartition = decoder.decode<String>();
    auto isFromServiceWorkerPage = decoder.decode<bool>();
    auto registrationOptions = decoder.decode<std::optional<WebCore::ServiceWorkerRegistrationOptions>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(( type == WebCore::ServiceWorkerJobType::Register && registrationOptions->has_value()) || (type != WebCore::ServiceWorkerJobType::Register && !registrationOptions->has_value() )))
        return std::nullopt;
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerJobData {
            WTFMove(*identifier),
            WTFMove(*scriptURL),
            WTFMove(*clientCreationURL),
            WTFMove(*topOrigin),
            WTFMove(*scopeURL),
            WTFMove(*sourceContext),
            WTFMove(*workerType),
            WTFMove(*type),
            WTFMove(*domainForCachePartition),
            WTFMove(*isFromServiceWorkerPage),
            WTFMove(*registrationOptions)
        }
    };
}

#endif

void ArgumentCoder<WebCore::EventInit>::encode(Encoder& encoder, const WebCore::EventInit& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bubbles)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cancelable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.composed)>, bool>);
    struct ShouldBeSameSizeAsEventInit : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::EventInit>, false> {
        bool bubbles;
        bool cancelable;
        bool composed;
    };
    static_assert(sizeof(ShouldBeSameSizeAsEventInit) == sizeof(WebCore::EventInit));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::EventInit, bubbles)
        , offsetof(WebCore::EventInit, cancelable)
        , offsetof(WebCore::EventInit, composed)
    >::value);
    encoder << instance.bubbles;
    encoder << instance.cancelable;
    encoder << instance.composed;
}

std::optional<WebCore::EventInit> ArgumentCoder<WebCore::EventInit>::decode(Decoder& decoder)
{
    auto bubbles = decoder.decode<bool>();
    auto cancelable = decoder.decode<bool>();
    auto composed = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::EventInit {
            WTFMove(*bubbles),
            WTFMove(*cancelable),
            WTFMove(*composed)
        }
    };
}

void ArgumentCoder<WebCore::MessageWithMessagePorts>::encode(Encoder& encoder, const WebCore::MessageWithMessagePorts& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, RefPtr<WebCore::SerializedScriptValue>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transferredPorts)>, Vector<WebCore::TransferredMessagePort>>);
    struct ShouldBeSameSizeAsMessageWithMessagePorts : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MessageWithMessagePorts>, false> {
        RefPtr<WebCore::SerializedScriptValue> message;
        Vector<WebCore::TransferredMessagePort> transferredPorts;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMessageWithMessagePorts) == sizeof(WebCore::MessageWithMessagePorts));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MessageWithMessagePorts, message)
        , offsetof(WebCore::MessageWithMessagePorts, transferredPorts)
    >::value);
    encoder << instance.message;
    encoder << instance.transferredPorts;
}

std::optional<WebCore::MessageWithMessagePorts> ArgumentCoder<WebCore::MessageWithMessagePorts>::decode(Decoder& decoder)
{
    auto message = decoder.decode<RefPtr<WebCore::SerializedScriptValue>>();
    auto transferredPorts = decoder.decode<Vector<WebCore::TransferredMessagePort>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MessageWithMessagePorts {
            WTFMove(*message),
            WTFMove(*transferredPorts)
        }
    };
}

#if ENABLE(TRACKING_PREVENTION)
void ArgumentCoder<WebCore::RequestStorageAccessResult>::encode(Encoder& encoder, const WebCore::RequestStorageAccessResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wasGranted)>, WebCore::StorageAccessWasGranted>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.promptWasShown)>, WebCore::StorageAccessPromptWasShown>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scope)>, WebCore::StorageAccessScope>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topFrameDomain)>, WebCore::RegistrableDomain>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subFrameDomain)>, WebCore::RegistrableDomain>);
    struct ShouldBeSameSizeAsRequestStorageAccessResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RequestStorageAccessResult>, false> {
        WebCore::StorageAccessWasGranted wasGranted;
        WebCore::StorageAccessPromptWasShown promptWasShown;
        WebCore::StorageAccessScope scope;
        WebCore::RegistrableDomain topFrameDomain;
        WebCore::RegistrableDomain subFrameDomain;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRequestStorageAccessResult) == sizeof(WebCore::RequestStorageAccessResult));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RequestStorageAccessResult, wasGranted)
        , offsetof(WebCore::RequestStorageAccessResult, promptWasShown)
        , offsetof(WebCore::RequestStorageAccessResult, scope)
        , offsetof(WebCore::RequestStorageAccessResult, topFrameDomain)
        , offsetof(WebCore::RequestStorageAccessResult, subFrameDomain)
    >::value);
    encoder << instance.wasGranted;
    encoder << instance.promptWasShown;
    encoder << instance.scope;
    encoder << instance.topFrameDomain;
    encoder << instance.subFrameDomain;
}

std::optional<WebCore::RequestStorageAccessResult> ArgumentCoder<WebCore::RequestStorageAccessResult>::decode(Decoder& decoder)
{
    auto wasGranted = decoder.decode<WebCore::StorageAccessWasGranted>();
    auto promptWasShown = decoder.decode<WebCore::StorageAccessPromptWasShown>();
    auto scope = decoder.decode<WebCore::StorageAccessScope>();
    auto topFrameDomain = decoder.decode<WebCore::RegistrableDomain>();
    auto subFrameDomain = decoder.decode<WebCore::RegistrableDomain>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RequestStorageAccessResult {
            WTFMove(*wasGranted),
            WTFMove(*promptWasShown),
            WTFMove(*scope),
            WTFMove(*topFrameDomain),
            WTFMove(*subFrameDomain)
        }
    };
}

#endif

void ArgumentCoder<WebCore::Exception>::encode(Encoder& encoder, const WebCore::Exception& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.code())>, WebCore::ExceptionCode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message())>, String>);
    encoder << instance.code();
    encoder << instance.message();
}

std::optional<WebCore::Exception> ArgumentCoder<WebCore::Exception>::decode(Decoder& decoder)
{
    auto code = decoder.decode<WebCore::ExceptionCode>();
    auto message = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Exception {
            WTFMove(*code),
            WTFMove(*message)
        }
    };
}

void ArgumentCoder<WebCore::ExceptionData>::encode(Encoder& encoder, const WebCore::ExceptionData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.code)>, WebCore::ExceptionCode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, String>);
    struct ShouldBeSameSizeAsExceptionData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ExceptionData>, false> {
        WebCore::ExceptionCode code;
        String message;
    };
    static_assert(sizeof(ShouldBeSameSizeAsExceptionData) == sizeof(WebCore::ExceptionData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ExceptionData, code)
        , offsetof(WebCore::ExceptionData, message)
    >::value);
    encoder << instance.code;
    encoder << instance.message;
}

std::optional<WebCore::ExceptionData> ArgumentCoder<WebCore::ExceptionData>::decode(Decoder& decoder)
{
    auto code = decoder.decode<WebCore::ExceptionCode>();
    auto message = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ExceptionData {
            WTFMove(*code),
            WTFMove(*message)
        }
    };
}

void ArgumentCoder<WebCore::MessagePortIdentifier>::encode(Encoder& encoder, const WebCore::MessagePortIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processIdentifier)>, WebCore::ProcessIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.portIdentifier)>, WebCore::PortIdentifier>);
    struct ShouldBeSameSizeAsMessagePortIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MessagePortIdentifier>, false> {
        WebCore::ProcessIdentifier processIdentifier;
        WebCore::PortIdentifier portIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMessagePortIdentifier) == sizeof(WebCore::MessagePortIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MessagePortIdentifier, processIdentifier)
        , offsetof(WebCore::MessagePortIdentifier, portIdentifier)
    >::value);
    encoder << instance.processIdentifier;
    encoder << instance.portIdentifier;
}

std::optional<WebCore::MessagePortIdentifier> ArgumentCoder<WebCore::MessagePortIdentifier>::decode(Decoder& decoder)
{
    auto processIdentifier = decoder.decode<WebCore::ProcessIdentifier>();
    auto portIdentifier = decoder.decode<WebCore::PortIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MessagePortIdentifier {
            WTFMove(*processIdentifier),
            WTFMove(*portIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::SecurityPolicyViolationEventInit>::encode(Encoder& encoder, const WebCore::SecurityPolicyViolationEventInit& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bubbles)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cancelable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.composed)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::EventInit, bubbles)
        , offsetof(WebCore::EventInit, cancelable)
        , offsetof(WebCore::EventInit, composed)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentURI)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrer)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockedURI)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.violatedDirective)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.effectiveDirective)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originalPolicy)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceFile)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sample)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.disposition)>, WebCore::SecurityPolicyViolationEventDisposition>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.statusCode)>, unsigned short>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lineNumber)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.columnNumber)>, unsigned>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SecurityPolicyViolationEventInit, documentURI)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, referrer)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, blockedURI)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, violatedDirective)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, effectiveDirective)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, originalPolicy)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, sourceFile)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, sample)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, disposition)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, statusCode)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, lineNumber)
        , offsetof(WebCore::SecurityPolicyViolationEventInit, columnNumber)
    >::value);
    encoder << instance.bubbles;
    encoder << instance.cancelable;
    encoder << instance.composed;
    encoder << instance.documentURI;
    encoder << instance.referrer;
    encoder << instance.blockedURI;
    encoder << instance.violatedDirective;
    encoder << instance.effectiveDirective;
    encoder << instance.originalPolicy;
    encoder << instance.sourceFile;
    encoder << instance.sample;
    encoder << instance.disposition;
    encoder << instance.statusCode;
    encoder << instance.lineNumber;
    encoder << instance.columnNumber;
}

std::optional<WebCore::SecurityPolicyViolationEventInit> ArgumentCoder<WebCore::SecurityPolicyViolationEventInit>::decode(Decoder& decoder)
{
    auto bubbles = decoder.decode<bool>();
    auto cancelable = decoder.decode<bool>();
    auto composed = decoder.decode<bool>();
    auto documentURI = decoder.decode<String>();
    auto referrer = decoder.decode<String>();
    auto blockedURI = decoder.decode<String>();
    auto violatedDirective = decoder.decode<String>();
    auto effectiveDirective = decoder.decode<String>();
    auto originalPolicy = decoder.decode<String>();
    auto sourceFile = decoder.decode<String>();
    auto sample = decoder.decode<String>();
    auto disposition = decoder.decode<WebCore::SecurityPolicyViolationEventDisposition>();
    auto statusCode = decoder.decode<unsigned short>();
    auto lineNumber = decoder.decode<unsigned>();
    auto columnNumber = decoder.decode<unsigned>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SecurityPolicyViolationEventInit {
            WebCore::EventInit {
                WTFMove(*bubbles),
                WTFMove(*cancelable),
                WTFMove(*composed)
            },
            WTFMove(*documentURI),
            WTFMove(*referrer),
            WTFMove(*blockedURI),
            WTFMove(*violatedDirective),
            WTFMove(*effectiveDirective),
            WTFMove(*originalPolicy),
            WTFMove(*sourceFile),
            WTFMove(*sample),
            WTFMove(*disposition),
            WTFMove(*statusCode),
            WTFMove(*lineNumber),
            WTFMove(*columnNumber)
        }
    };
}

void ArgumentCoder<WebCore::CSPViolationReportBody>::encode(Encoder& encoder, const WebCore::CSPViolationReportBody& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentURL())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrer())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockedURL())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.effectiveDirective())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originalPolicy())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceFile())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sample())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.disposition())>, WebCore::SecurityPolicyViolationEventDisposition>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.statusCode())>, unsigned short>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lineNumber())>, unsigned long>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.columnNumber())>, unsigned long>);
    encoder << instance.documentURL();
    encoder << instance.referrer();
    encoder << instance.blockedURL();
    encoder << instance.effectiveDirective();
    encoder << instance.originalPolicy();
    encoder << instance.sourceFile();
    encoder << instance.sample();
    encoder << instance.disposition();
    encoder << instance.statusCode();
    encoder << instance.lineNumber();
    encoder << instance.columnNumber();
}

std::optional<Ref<WebCore::CSPViolationReportBody>> ArgumentCoder<WebCore::CSPViolationReportBody>::decode(Decoder& decoder)
{
    auto documentURL = decoder.decode<String>();
    auto referrer = decoder.decode<String>();
    auto blockedURL = decoder.decode<String>();
    auto effectiveDirective = decoder.decode<String>();
    auto originalPolicy = decoder.decode<String>();
    auto sourceFile = decoder.decode<String>();
    auto sample = decoder.decode<String>();
    auto disposition = decoder.decode<WebCore::SecurityPolicyViolationEventDisposition>();
    auto statusCode = decoder.decode<unsigned short>();
    auto lineNumber = decoder.decode<unsigned long>();
    auto columnNumber = decoder.decode<unsigned long>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CSPViolationReportBody::create(
            WTFMove(*documentURL),
            WTFMove(*referrer),
            WTFMove(*blockedURL),
            WTFMove(*effectiveDirective),
            WTFMove(*originalPolicy),
            WTFMove(*sourceFile),
            WTFMove(*sample),
            WTFMove(*disposition),
            WTFMove(*statusCode),
            WTFMove(*lineNumber),
            WTFMove(*columnNumber)
        )
    };
}

void ArgumentCoder<WebCore::SecurityOriginData>::encode(Encoder& encoder, const WebCore::SecurityOriginData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, std::variant<WebCore::SecurityOriginData::Tuple, WebCore::ProcessQualified<WebCore::OpaqueOriginIdentifier>>>);
    encoder << instance.data();
}

std::optional<WebCore::SecurityOriginData> ArgumentCoder<WebCore::SecurityOriginData>::decode(Decoder& decoder)
{
    auto data = decoder.decode<std::variant<WebCore::SecurityOriginData::Tuple, WebCore::ProcessQualified<WebCore::OpaqueOriginIdentifier>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SecurityOriginData {
            WTFMove(*data)
        }
    };
}

void ArgumentCoder<WebCore::SecurityOriginData::Tuple>::encode(Encoder& encoder, const WebCore::SecurityOriginData::Tuple& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.protocol)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.host)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.port)>, std::optional<uint16_t>>);
    struct ShouldBeSameSizeAsTuple : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SecurityOriginData::Tuple>, false> {
        String protocol;
        String host;
        std::optional<uint16_t> port;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTuple) == sizeof(WebCore::SecurityOriginData::Tuple));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SecurityOriginData::Tuple, protocol)
        , offsetof(WebCore::SecurityOriginData::Tuple, host)
        , offsetof(WebCore::SecurityOriginData::Tuple, port)
    >::value);
    encoder << instance.protocol;
    encoder << instance.host;
    encoder << instance.port;
}

std::optional<WebCore::SecurityOriginData::Tuple> ArgumentCoder<WebCore::SecurityOriginData::Tuple>::decode(Decoder& decoder)
{
    auto protocol = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(!protocol->isHashTableDeletedValue()))
        return std::nullopt;
    auto host = decoder.decode<String>();
    auto port = decoder.decode<std::optional<uint16_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SecurityOriginData::Tuple {
            WTFMove(*protocol),
            WTFMove(*host),
            WTFMove(*port)
        }
    };
}

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::MediaConstraint>::encode(Encoder& encoder, const WebCore::MediaConstraint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constraintType())>, WebCore::MediaConstraintType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataType())>, WebCore::MediaConstraint::DataType>);
    encoder << instance.name();
    encoder << instance.constraintType();
    encoder << instance.dataType();
}

std::optional<WebCore::MediaConstraint> ArgumentCoder<WebCore::MediaConstraint>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto constraintType = decoder.decode<WebCore::MediaConstraintType>();
    auto dataType = decoder.decode<WebCore::MediaConstraint::DataType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaConstraint {
            WTFMove(*name),
            WTFMove(*constraintType),
            WTFMove(*dataType)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::IntConstraint>::encode(Encoder& encoder, const WebCore::IntConstraint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constraintType())>, WebCore::MediaConstraintType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataType())>, WebCore::MediaConstraint::DataType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_min)>, std::optional<int>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_max)>, std::optional<int>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_exact)>, std::optional<int>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_ideal)>, std::optional<int>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::IntConstraint, m_min)
        , offsetof(WebCore::IntConstraint, m_max)
        , offsetof(WebCore::IntConstraint, m_exact)
        , offsetof(WebCore::IntConstraint, m_ideal)
    >::value);
    encoder << instance.name();
    encoder << instance.constraintType();
    encoder << instance.dataType();
    encoder << instance.m_min;
    encoder << instance.m_max;
    encoder << instance.m_exact;
    encoder << instance.m_ideal;
}

std::optional<WebCore::IntConstraint> ArgumentCoder<WebCore::IntConstraint>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto constraintType = decoder.decode<WebCore::MediaConstraintType>();
    auto dataType = decoder.decode<WebCore::MediaConstraint::DataType>();
    auto m_min = decoder.decode<std::optional<int>>();
    auto m_max = decoder.decode<std::optional<int>>();
    auto m_exact = decoder.decode<std::optional<int>>();
    auto m_ideal = decoder.decode<std::optional<int>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::IntConstraint {
            WebCore::MediaConstraint {
                WTFMove(*name),
                WTFMove(*constraintType),
                WTFMove(*dataType)
            },
            WTFMove(*m_min),
            WTFMove(*m_max),
            WTFMove(*m_exact),
            WTFMove(*m_ideal)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::DoubleConstraint>::encode(Encoder& encoder, const WebCore::DoubleConstraint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constraintType())>, WebCore::MediaConstraintType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataType())>, WebCore::MediaConstraint::DataType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_min)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_max)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_exact)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_ideal)>, std::optional<double>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::DoubleConstraint, m_min)
        , offsetof(WebCore::DoubleConstraint, m_max)
        , offsetof(WebCore::DoubleConstraint, m_exact)
        , offsetof(WebCore::DoubleConstraint, m_ideal)
    >::value);
    encoder << instance.name();
    encoder << instance.constraintType();
    encoder << instance.dataType();
    encoder << instance.m_min;
    encoder << instance.m_max;
    encoder << instance.m_exact;
    encoder << instance.m_ideal;
}

std::optional<WebCore::DoubleConstraint> ArgumentCoder<WebCore::DoubleConstraint>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto constraintType = decoder.decode<WebCore::MediaConstraintType>();
    auto dataType = decoder.decode<WebCore::MediaConstraint::DataType>();
    auto m_min = decoder.decode<std::optional<double>>();
    auto m_max = decoder.decode<std::optional<double>>();
    auto m_exact = decoder.decode<std::optional<double>>();
    auto m_ideal = decoder.decode<std::optional<double>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::DoubleConstraint {
            WebCore::MediaConstraint {
                WTFMove(*name),
                WTFMove(*constraintType),
                WTFMove(*dataType)
            },
            WTFMove(*m_min),
            WTFMove(*m_max),
            WTFMove(*m_exact),
            WTFMove(*m_ideal)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::BooleanConstraint>::encode(Encoder& encoder, const WebCore::BooleanConstraint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constraintType())>, WebCore::MediaConstraintType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataType())>, WebCore::MediaConstraint::DataType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_exact)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_ideal)>, std::optional<bool>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::BooleanConstraint, m_exact)
        , offsetof(WebCore::BooleanConstraint, m_ideal)
    >::value);
    encoder << instance.name();
    encoder << instance.constraintType();
    encoder << instance.dataType();
    encoder << instance.m_exact;
    encoder << instance.m_ideal;
}

std::optional<WebCore::BooleanConstraint> ArgumentCoder<WebCore::BooleanConstraint>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto constraintType = decoder.decode<WebCore::MediaConstraintType>();
    auto dataType = decoder.decode<WebCore::MediaConstraint::DataType>();
    auto m_exact = decoder.decode<std::optional<bool>>();
    auto m_ideal = decoder.decode<std::optional<bool>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::BooleanConstraint {
            WebCore::MediaConstraint {
                WTFMove(*name),
                WTFMove(*constraintType),
                WTFMove(*dataType)
            },
            WTFMove(*m_exact),
            WTFMove(*m_ideal)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::StringConstraint>::encode(Encoder& encoder, const WebCore::StringConstraint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constraintType())>, WebCore::MediaConstraintType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataType())>, WebCore::MediaConstraint::DataType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_exact)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_ideal)>, Vector<String>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::StringConstraint, m_exact)
        , offsetof(WebCore::StringConstraint, m_ideal)
    >::value);
    encoder << instance.name();
    encoder << instance.constraintType();
    encoder << instance.dataType();
    encoder << instance.m_exact;
    encoder << instance.m_ideal;
}

std::optional<WebCore::StringConstraint> ArgumentCoder<WebCore::StringConstraint>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto constraintType = decoder.decode<WebCore::MediaConstraintType>();
    auto dataType = decoder.decode<WebCore::MediaConstraint::DataType>();
    auto m_exact = decoder.decode<Vector<String>>();
    auto m_ideal = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::StringConstraint {
            WebCore::MediaConstraint {
                WTFMove(*name),
                WTFMove(*constraintType),
                WTFMove(*dataType)
            },
            WTFMove(*m_exact),
            WTFMove(*m_ideal)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::FrameRateRange>::encode(Encoder& encoder, const WebCore::FrameRateRange& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimum)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximum)>, double>);
    struct ShouldBeSameSizeAsFrameRateRange : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::FrameRateRange>, false> {
        double minimum;
        double maximum;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFrameRateRange) == sizeof(WebCore::FrameRateRange));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::FrameRateRange, minimum)
        , offsetof(WebCore::FrameRateRange, maximum)
    >::value);
    encoder << instance.minimum;
    encoder << instance.maximum;
}

std::optional<WebCore::FrameRateRange> ArgumentCoder<WebCore::FrameRateRange>::decode(Decoder& decoder)
{
    auto minimum = decoder.decode<double>();
    auto maximum = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::FrameRateRange {
            WTFMove(*minimum),
            WTFMove(*maximum)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::VideoPresetData>::encode(Encoder& encoder, const WebCore::VideoPresetData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameRateRanges)>, Vector<WebCore::FrameRateRange>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minZoom)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxZoom)>, double>);
    struct ShouldBeSameSizeAsVideoPresetData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::VideoPresetData>, false> {
        WebCore::IntSize size;
        Vector<WebCore::FrameRateRange> frameRateRanges;
        double minZoom;
        double maxZoom;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVideoPresetData) == sizeof(WebCore::VideoPresetData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::VideoPresetData, size)
        , offsetof(WebCore::VideoPresetData, frameRateRanges)
        , offsetof(WebCore::VideoPresetData, minZoom)
        , offsetof(WebCore::VideoPresetData, maxZoom)
    >::value);
    encoder << instance.size;
    encoder << instance.frameRateRanges;
    encoder << instance.minZoom;
    encoder << instance.maxZoom;
}

std::optional<WebCore::VideoPresetData> ArgumentCoder<WebCore::VideoPresetData>::decode(Decoder& decoder)
{
    auto size = decoder.decode<WebCore::IntSize>();
    auto frameRateRanges = decoder.decode<Vector<WebCore::FrameRateRange>>();
    auto minZoom = decoder.decode<double>();
    auto maxZoom = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::VideoPresetData {
            WTFMove(*size),
            WTFMove(*frameRateRanges),
            WTFMove(*minZoom),
            WTFMove(*maxZoom)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::RealtimeMediaSourceSupportedConstraints>::encode(Encoder& encoder, const WebCore::RealtimeMediaSourceSupportedConstraints& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsWidth())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsHeight())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsAspectRatio())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsFrameRate())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsFacingMode())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsVolume())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsSampleRate())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsSampleSize())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsEchoCancellation())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsDeviceId())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsGroupId())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsDisplaySurface())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsLogicalSurface())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsFocusDistance())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsZoom())>, bool>);
    encoder << instance.supportsWidth();
    encoder << instance.supportsHeight();
    encoder << instance.supportsAspectRatio();
    encoder << instance.supportsFrameRate();
    encoder << instance.supportsFacingMode();
    encoder << instance.supportsVolume();
    encoder << instance.supportsSampleRate();
    encoder << instance.supportsSampleSize();
    encoder << instance.supportsEchoCancellation();
    encoder << instance.supportsDeviceId();
    encoder << instance.supportsGroupId();
    encoder << instance.supportsDisplaySurface();
    encoder << instance.supportsLogicalSurface();
    encoder << instance.supportsFocusDistance();
    encoder << instance.supportsZoom();
}

std::optional<WebCore::RealtimeMediaSourceSupportedConstraints> ArgumentCoder<WebCore::RealtimeMediaSourceSupportedConstraints>::decode(Decoder& decoder)
{
    auto supportsWidth = decoder.decode<bool>();
    auto supportsHeight = decoder.decode<bool>();
    auto supportsAspectRatio = decoder.decode<bool>();
    auto supportsFrameRate = decoder.decode<bool>();
    auto supportsFacingMode = decoder.decode<bool>();
    auto supportsVolume = decoder.decode<bool>();
    auto supportsSampleRate = decoder.decode<bool>();
    auto supportsSampleSize = decoder.decode<bool>();
    auto supportsEchoCancellation = decoder.decode<bool>();
    auto supportsDeviceId = decoder.decode<bool>();
    auto supportsGroupId = decoder.decode<bool>();
    auto supportsDisplaySurface = decoder.decode<bool>();
    auto supportsLogicalSurface = decoder.decode<bool>();
    auto supportsFocusDistance = decoder.decode<bool>();
    auto supportsZoom = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RealtimeMediaSourceSupportedConstraints {
            WTFMove(*supportsWidth),
            WTFMove(*supportsHeight),
            WTFMove(*supportsAspectRatio),
            WTFMove(*supportsFrameRate),
            WTFMove(*supportsFacingMode),
            WTFMove(*supportsVolume),
            WTFMove(*supportsSampleRate),
            WTFMove(*supportsSampleSize),
            WTFMove(*supportsEchoCancellation),
            WTFMove(*supportsDeviceId),
            WTFMove(*supportsGroupId),
            WTFMove(*supportsDisplaySurface),
            WTFMove(*supportsLogicalSurface),
            WTFMove(*supportsFocusDistance),
            WTFMove(*supportsZoom)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::RealtimeMediaSourceSettings>::encode(Encoder& encoder, const WebCore::RealtimeMediaSourceSettings& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameRate())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.facingMode())>, WebCore::VideoFacingMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.volume())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleRate())>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleSize())>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.echoCancellation())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deviceId())>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.groupId())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label())>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.displaySurface())>, WebCore::DisplaySurfaceType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.logicalSurface())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.zoom())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportedConstraints())>, WebCore::RealtimeMediaSourceSupportedConstraints>);
    encoder << instance.width();
    encoder << instance.height();
    encoder << instance.frameRate();
    encoder << instance.facingMode();
    encoder << instance.volume();
    encoder << instance.sampleRate();
    encoder << instance.sampleSize();
    encoder << instance.echoCancellation();
    encoder << instance.deviceId();
    encoder << instance.groupId();
    encoder << instance.label();
    encoder << instance.displaySurface();
    encoder << instance.logicalSurface();
    encoder << instance.zoom();
    encoder << instance.supportedConstraints();
}

std::optional<WebCore::RealtimeMediaSourceSettings> ArgumentCoder<WebCore::RealtimeMediaSourceSettings>::decode(Decoder& decoder)
{
    auto width = decoder.decode<uint32_t>();
    auto height = decoder.decode<uint32_t>();
    auto frameRate = decoder.decode<float>();
    auto facingMode = decoder.decode<WebCore::VideoFacingMode>();
    auto volume = decoder.decode<double>();
    auto sampleRate = decoder.decode<uint32_t>();
    auto sampleSize = decoder.decode<uint32_t>();
    auto echoCancellation = decoder.decode<bool>();
    auto deviceId = decoder.decode<AtomString>();
    auto groupId = decoder.decode<String>();
    auto label = decoder.decode<AtomString>();
    auto displaySurface = decoder.decode<WebCore::DisplaySurfaceType>();
    auto logicalSurface = decoder.decode<bool>();
    auto zoom = decoder.decode<double>();
    auto supportedConstraints = decoder.decode<WebCore::RealtimeMediaSourceSupportedConstraints>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RealtimeMediaSourceSettings {
            WTFMove(*width),
            WTFMove(*height),
            WTFMove(*frameRate),
            WTFMove(*facingMode),
            WTFMove(*volume),
            WTFMove(*sampleRate),
            WTFMove(*sampleSize),
            WTFMove(*echoCancellation),
            WTFMove(*deviceId),
            WTFMove(*groupId),
            WTFMove(*label),
            WTFMove(*displaySurface),
            WTFMove(*logicalSurface),
            WTFMove(*zoom),
            WTFMove(*supportedConstraints)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::CaptureDevice>::encode(Encoder& encoder, const WebCore::CaptureDevice& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.persistentId())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebCore::CaptureDevice::DeviceType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.groupId())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enabled())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isDefault())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMockDevice())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isEphemeral())>, bool>);
    encoder << instance.persistentId();
    encoder << instance.type();
    encoder << instance.label();
    encoder << instance.groupId();
    encoder << instance.enabled();
    encoder << instance.isDefault();
    encoder << instance.isMockDevice();
    encoder << instance.isEphemeral();
}

std::optional<WebCore::CaptureDevice> ArgumentCoder<WebCore::CaptureDevice>::decode(Decoder& decoder)
{
    auto persistentId = decoder.decode<String>();
    auto type = decoder.decode<WebCore::CaptureDevice::DeviceType>();
    auto label = decoder.decode<String>();
    auto groupId = decoder.decode<String>();
    auto enabled = decoder.decode<bool>();
    auto isDefault = decoder.decode<bool>();
    auto isMockDevice = decoder.decode<bool>();
    auto isEphemeral = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CaptureDevice {
            WTFMove(*persistentId),
            WTFMove(*type),
            WTFMove(*label),
            WTFMove(*groupId),
            WTFMove(*enabled),
            WTFMove(*isDefault),
            WTFMove(*isMockDevice),
            WTFMove(*isEphemeral)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::CaptureDeviceWithCapabilities>::encode(Encoder& encoder, const WebCore::CaptureDeviceWithCapabilities& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.device)>, WebCore::CaptureDevice>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.capabilities)>, WebCore::RealtimeMediaSourceCapabilities>);
    struct ShouldBeSameSizeAsCaptureDeviceWithCapabilities : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CaptureDeviceWithCapabilities>, false> {
        WebCore::CaptureDevice device;
        WebCore::RealtimeMediaSourceCapabilities capabilities;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCaptureDeviceWithCapabilities) == sizeof(WebCore::CaptureDeviceWithCapabilities));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CaptureDeviceWithCapabilities, device)
        , offsetof(WebCore::CaptureDeviceWithCapabilities, capabilities)
    >::value);
    encoder << instance.device;
    encoder << instance.capabilities;
}

std::optional<WebCore::CaptureDeviceWithCapabilities> ArgumentCoder<WebCore::CaptureDeviceWithCapabilities>::decode(Decoder& decoder)
{
    auto device = decoder.decode<WebCore::CaptureDevice>();
    auto capabilities = decoder.decode<WebCore::RealtimeMediaSourceCapabilities>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CaptureDeviceWithCapabilities {
            WTFMove(*device),
            WTFMove(*capabilities)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::RealtimeMediaSourceCapabilities>::encode(Encoder& encoder, const WebCore::RealtimeMediaSourceCapabilities& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspectRatio())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameRate())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.facingMode())>, Vector<WebCore::VideoFacingMode>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.volume())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleRate())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleSize())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.echoCancellation())>, WebCore::RealtimeMediaSourceCapabilities::EchoCancellation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deviceId())>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.groupId())>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.focusDistance())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.zoom())>, WebCore::CapabilityValueOrRange>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportedConstraints())>, WebCore::RealtimeMediaSourceSupportedConstraints>);
    encoder << instance.width();
    encoder << instance.height();
    encoder << instance.aspectRatio();
    encoder << instance.frameRate();
    encoder << instance.facingMode();
    encoder << instance.volume();
    encoder << instance.sampleRate();
    encoder << instance.sampleSize();
    encoder << instance.echoCancellation();
    encoder << instance.deviceId();
    encoder << instance.groupId();
    encoder << instance.focusDistance();
    encoder << instance.zoom();
    encoder << instance.supportedConstraints();
}

std::optional<WebCore::RealtimeMediaSourceCapabilities> ArgumentCoder<WebCore::RealtimeMediaSourceCapabilities>::decode(Decoder& decoder)
{
    auto width = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto height = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto aspectRatio = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto frameRate = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto facingMode = decoder.decode<Vector<WebCore::VideoFacingMode>>();
    auto volume = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto sampleRate = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto sampleSize = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto echoCancellation = decoder.decode<WebCore::RealtimeMediaSourceCapabilities::EchoCancellation>();
    auto deviceId = decoder.decode<AtomString>();
    auto groupId = decoder.decode<AtomString>();
    auto focusDistance = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto zoom = decoder.decode<WebCore::CapabilityValueOrRange>();
    auto supportedConstraints = decoder.decode<WebCore::RealtimeMediaSourceSupportedConstraints>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RealtimeMediaSourceCapabilities {
            WTFMove(*width),
            WTFMove(*height),
            WTFMove(*aspectRatio),
            WTFMove(*frameRate),
            WTFMove(*facingMode),
            WTFMove(*volume),
            WTFMove(*sampleRate),
            WTFMove(*sampleSize),
            WTFMove(*echoCancellation),
            WTFMove(*deviceId),
            WTFMove(*groupId),
            WTFMove(*focusDistance),
            WTFMove(*zoom),
            WTFMove(*supportedConstraints)
        }
    };
}

#endif

void ArgumentCoder<WebCore::PlatformVideoColorSpace>::encode(Encoder& encoder, const WebCore::PlatformVideoColorSpace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primaries)>, std::optional<WebCore::PlatformVideoColorPrimaries>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transfer)>, std::optional<WebCore::PlatformVideoTransferCharacteristics>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.matrix)>, std::optional<WebCore::PlatformVideoMatrixCoefficients>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fullRange)>, std::optional<bool>>);
    struct ShouldBeSameSizeAsPlatformVideoColorSpace : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PlatformVideoColorSpace>, false> {
        std::optional<WebCore::PlatformVideoColorPrimaries> primaries;
        std::optional<WebCore::PlatformVideoTransferCharacteristics> transfer;
        std::optional<WebCore::PlatformVideoMatrixCoefficients> matrix;
        std::optional<bool> fullRange;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPlatformVideoColorSpace) == sizeof(WebCore::PlatformVideoColorSpace));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformVideoColorSpace, primaries)
        , offsetof(WebCore::PlatformVideoColorSpace, transfer)
        , offsetof(WebCore::PlatformVideoColorSpace, matrix)
        , offsetof(WebCore::PlatformVideoColorSpace, fullRange)
    >::value);
    encoder << instance.primaries;
    encoder << instance.transfer;
    encoder << instance.matrix;
    encoder << instance.fullRange;
}

void ArgumentCoder<WebCore::PlatformVideoColorSpace>::encode(StreamConnectionEncoder& encoder, const WebCore::PlatformVideoColorSpace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primaries)>, std::optional<WebCore::PlatformVideoColorPrimaries>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transfer)>, std::optional<WebCore::PlatformVideoTransferCharacteristics>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.matrix)>, std::optional<WebCore::PlatformVideoMatrixCoefficients>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fullRange)>, std::optional<bool>>);
    struct ShouldBeSameSizeAsPlatformVideoColorSpace : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PlatformVideoColorSpace>, false> {
        std::optional<WebCore::PlatformVideoColorPrimaries> primaries;
        std::optional<WebCore::PlatformVideoTransferCharacteristics> transfer;
        std::optional<WebCore::PlatformVideoMatrixCoefficients> matrix;
        std::optional<bool> fullRange;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPlatformVideoColorSpace) == sizeof(WebCore::PlatformVideoColorSpace));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformVideoColorSpace, primaries)
        , offsetof(WebCore::PlatformVideoColorSpace, transfer)
        , offsetof(WebCore::PlatformVideoColorSpace, matrix)
        , offsetof(WebCore::PlatformVideoColorSpace, fullRange)
    >::value);
    encoder << instance.primaries;
    encoder << instance.transfer;
    encoder << instance.matrix;
    encoder << instance.fullRange;
}

std::optional<WebCore::PlatformVideoColorSpace> ArgumentCoder<WebCore::PlatformVideoColorSpace>::decode(Decoder& decoder)
{
    auto primaries = decoder.decode<std::optional<WebCore::PlatformVideoColorPrimaries>>();
    auto transfer = decoder.decode<std::optional<WebCore::PlatformVideoTransferCharacteristics>>();
    auto matrix = decoder.decode<std::optional<WebCore::PlatformVideoMatrixCoefficients>>();
    auto fullRange = decoder.decode<std::optional<bool>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PlatformVideoColorSpace {
            WTFMove(*primaries),
            WTFMove(*transfer),
            WTFMove(*matrix),
            WTFMove(*fullRange)
        }
    };
}

#if ENABLE(VIDEO)
void ArgumentCoder<WebCore::PlatformTrackConfiguration>::encode(Encoder& encoder, const WebCore::PlatformTrackConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.codec)>, String>);
    struct ShouldBeSameSizeAsPlatformTrackConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PlatformTrackConfiguration>, false> {
        String codec;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPlatformTrackConfiguration) == sizeof(WebCore::PlatformTrackConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformTrackConfiguration, codec)
    >::value);
    encoder << instance.codec;
}

std::optional<WebCore::PlatformTrackConfiguration> ArgumentCoder<WebCore::PlatformTrackConfiguration>::decode(Decoder& decoder)
{
    auto codec = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PlatformTrackConfiguration {
            WTFMove(*codec)
        }
    };
}

#endif

#if ENABLE(VIDEO)
void ArgumentCoder<WebCore::PlatformAudioTrackConfiguration>::encode(Encoder& encoder, const WebCore::PlatformAudioTrackConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.codec)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformTrackConfiguration, codec)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleRate)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.numberOfChannels)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bitrate)>, uint64_t>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformAudioTrackConfiguration, sampleRate)
        , offsetof(WebCore::PlatformAudioTrackConfiguration, numberOfChannels)
        , offsetof(WebCore::PlatformAudioTrackConfiguration, bitrate)
    >::value);
    encoder << instance.codec;
    encoder << instance.sampleRate;
    encoder << instance.numberOfChannels;
    encoder << instance.bitrate;
}

std::optional<WebCore::PlatformAudioTrackConfiguration> ArgumentCoder<WebCore::PlatformAudioTrackConfiguration>::decode(Decoder& decoder)
{
    auto codec = decoder.decode<String>();
    auto sampleRate = decoder.decode<uint32_t>();
    auto numberOfChannels = decoder.decode<uint32_t>();
    auto bitrate = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PlatformAudioTrackConfiguration {
            WebCore::PlatformTrackConfiguration {
                WTFMove(*codec)
            },
            WTFMove(*sampleRate),
            WTFMove(*numberOfChannels),
            WTFMove(*bitrate)
        }
    };
}

#endif

#if ENABLE(VIDEO)
void ArgumentCoder<WebCore::PlatformVideoTrackConfiguration>::encode(Encoder& encoder, const WebCore::PlatformVideoTrackConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.codec)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformTrackConfiguration, codec)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::PlatformVideoColorSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.framerate)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bitrate)>, uint64_t>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PlatformVideoTrackConfiguration, width)
        , offsetof(WebCore::PlatformVideoTrackConfiguration, height)
        , offsetof(WebCore::PlatformVideoTrackConfiguration, colorSpace)
        , offsetof(WebCore::PlatformVideoTrackConfiguration, framerate)
        , offsetof(WebCore::PlatformVideoTrackConfiguration, bitrate)
    >::value);
    encoder << instance.codec;
    encoder << instance.width;
    encoder << instance.height;
    encoder << instance.colorSpace;
    encoder << instance.framerate;
    encoder << instance.bitrate;
}

std::optional<WebCore::PlatformVideoTrackConfiguration> ArgumentCoder<WebCore::PlatformVideoTrackConfiguration>::decode(Decoder& decoder)
{
    auto codec = decoder.decode<String>();
    auto width = decoder.decode<uint32_t>();
    auto height = decoder.decode<uint32_t>();
    auto colorSpace = decoder.decode<WebCore::PlatformVideoColorSpace>();
    auto framerate = decoder.decode<double>();
    auto bitrate = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PlatformVideoTrackConfiguration {
            WebCore::PlatformTrackConfiguration {
                WTFMove(*codec)
            },
            WTFMove(*width),
            WTFMove(*height),
            WTFMove(*colorSpace),
            WTFMove(*framerate),
            WTFMove(*bitrate)
        }
    };
}

#endif

#if ENABLE(WEB_RTC)
void ArgumentCoder<WebCore::RTCDataChannelInit>::encode(Encoder& encoder, const WebCore::RTCDataChannelInit& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ordered)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxPacketLifeTime)>, std::optional<unsigned short>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxRetransmits)>, std::optional<unsigned short>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.protocol)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.negotiated)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id)>, std::optional<unsigned short>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.priority)>, WebCore::RTCPriorityType>);
    struct ShouldBeSameSizeAsRTCDataChannelInit : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RTCDataChannelInit>, false> {
        std::optional<bool> ordered;
        std::optional<unsigned short> maxPacketLifeTime;
        std::optional<unsigned short> maxRetransmits;
        String protocol;
        std::optional<bool> negotiated;
        std::optional<unsigned short> id;
        WebCore::RTCPriorityType priority;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRTCDataChannelInit) == sizeof(WebCore::RTCDataChannelInit));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RTCDataChannelInit, ordered)
        , offsetof(WebCore::RTCDataChannelInit, maxPacketLifeTime)
        , offsetof(WebCore::RTCDataChannelInit, maxRetransmits)
        , offsetof(WebCore::RTCDataChannelInit, protocol)
        , offsetof(WebCore::RTCDataChannelInit, negotiated)
        , offsetof(WebCore::RTCDataChannelInit, id)
        , offsetof(WebCore::RTCDataChannelInit, priority)
    >::value);
    encoder << instance.ordered;
    encoder << instance.maxPacketLifeTime;
    encoder << instance.maxRetransmits;
    encoder << instance.protocol;
    encoder << instance.negotiated;
    encoder << instance.id;
    encoder << instance.priority;
}

std::optional<WebCore::RTCDataChannelInit> ArgumentCoder<WebCore::RTCDataChannelInit>::decode(Decoder& decoder)
{
    auto ordered = decoder.decode<std::optional<bool>>();
    auto maxPacketLifeTime = decoder.decode<std::optional<unsigned short>>();
    auto maxRetransmits = decoder.decode<std::optional<unsigned short>>();
    auto protocol = decoder.decode<String>();
    auto negotiated = decoder.decode<std::optional<bool>>();
    auto id = decoder.decode<std::optional<unsigned short>>();
    auto priority = decoder.decode<WebCore::RTCPriorityType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RTCDataChannelInit {
            WTFMove(*ordered),
            WTFMove(*maxPacketLifeTime),
            WTFMove(*maxRetransmits),
            WTFMove(*protocol),
            WTFMove(*negotiated),
            WTFMove(*id),
            WTFMove(*priority)
        }
    };
}

#endif

#if ENABLE(WEB_RTC)
void ArgumentCoder<WebCore::RTCDataChannelIdentifier>::encode(Encoder& encoder, const WebCore::RTCDataChannelIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processIdentifier)>, WebCore::ProcessIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.channelIdentifier)>, WebCore::RTCDataChannelLocalIdentifier>);
    struct ShouldBeSameSizeAsRTCDataChannelIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::RTCDataChannelIdentifier>, false> {
        WebCore::ProcessIdentifier processIdentifier;
        WebCore::RTCDataChannelLocalIdentifier channelIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRTCDataChannelIdentifier) == sizeof(WebCore::RTCDataChannelIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::RTCDataChannelIdentifier, processIdentifier)
        , offsetof(WebCore::RTCDataChannelIdentifier, channelIdentifier)
    >::value);
    encoder << instance.processIdentifier;
    encoder << instance.channelIdentifier;
}

std::optional<WebCore::RTCDataChannelIdentifier> ArgumentCoder<WebCore::RTCDataChannelIdentifier>::decode(Decoder& decoder)
{
    auto processIdentifier = decoder.decode<WebCore::ProcessIdentifier>();
    auto channelIdentifier = decoder.decode<WebCore::RTCDataChannelLocalIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::RTCDataChannelIdentifier {
            WTFMove(*processIdentifier),
            WTFMove(*channelIdentifier)
        }
    };
}

#endif

#if ENABLE(CONTENT_EXTENSIONS)
void ArgumentCoder<WebCore::ContentRuleListResults::Result>::encode(Encoder& encoder, const WebCore::ContentRuleListResults::Result& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockedLoad)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.madeHTTPS)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockedCookies)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiedHeaders)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.redirected)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.notifications)>, Vector<String>>);
    struct ShouldBeSameSizeAsResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ContentRuleListResults::Result>, false> {
        bool blockedLoad;
        bool madeHTTPS;
        bool blockedCookies;
        bool modifiedHeaders;
        bool redirected;
        Vector<String> notifications;
    };
    static_assert(sizeof(ShouldBeSameSizeAsResult) == sizeof(WebCore::ContentRuleListResults::Result));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ContentRuleListResults::Result, blockedLoad)
        , offsetof(WebCore::ContentRuleListResults::Result, madeHTTPS)
        , offsetof(WebCore::ContentRuleListResults::Result, blockedCookies)
        , offsetof(WebCore::ContentRuleListResults::Result, modifiedHeaders)
        , offsetof(WebCore::ContentRuleListResults::Result, redirected)
        , offsetof(WebCore::ContentRuleListResults::Result, notifications)
    >::value);
    encoder << instance.blockedLoad;
    encoder << instance.madeHTTPS;
    encoder << instance.blockedCookies;
    encoder << instance.modifiedHeaders;
    encoder << instance.redirected;
    encoder << instance.notifications;
}

std::optional<WebCore::ContentRuleListResults::Result> ArgumentCoder<WebCore::ContentRuleListResults::Result>::decode(Decoder& decoder)
{
    auto blockedLoad = decoder.decode<bool>();
    auto madeHTTPS = decoder.decode<bool>();
    auto blockedCookies = decoder.decode<bool>();
    auto modifiedHeaders = decoder.decode<bool>();
    auto redirected = decoder.decode<bool>();
    auto notifications = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ContentRuleListResults::Result {
            WTFMove(*blockedLoad),
            WTFMove(*madeHTTPS),
            WTFMove(*blockedCookies),
            WTFMove(*modifiedHeaders),
            WTFMove(*redirected),
            WTFMove(*notifications)
        }
    };
}

#endif

#if ENABLE(CONTENT_EXTENSIONS)
void ArgumentCoder<WebCore::ContentRuleListResults::Summary>::encode(Encoder& encoder, const WebCore::ContentRuleListResults::Summary& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockedLoad)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.madeHTTPS)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blockedCookies)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasNotifications)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifyHeadersActions)>, Vector<WebCore::ContentExtensions::ModifyHeadersAction>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.redirectActions)>, Vector<std::pair<WebCore::ContentExtensions::RedirectAction, URL>>>);
    struct ShouldBeSameSizeAsSummary : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ContentRuleListResults::Summary>, false> {
        bool blockedLoad;
        bool madeHTTPS;
        bool blockedCookies;
        bool hasNotifications;
        Vector<WebCore::ContentExtensions::ModifyHeadersAction> modifyHeadersActions;
        Vector<std::pair<WebCore::ContentExtensions::RedirectAction, URL>> redirectActions;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSummary) == sizeof(WebCore::ContentRuleListResults::Summary));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ContentRuleListResults::Summary, blockedLoad)
        , offsetof(WebCore::ContentRuleListResults::Summary, madeHTTPS)
        , offsetof(WebCore::ContentRuleListResults::Summary, blockedCookies)
        , offsetof(WebCore::ContentRuleListResults::Summary, hasNotifications)
        , offsetof(WebCore::ContentRuleListResults::Summary, modifyHeadersActions)
        , offsetof(WebCore::ContentRuleListResults::Summary, redirectActions)
    >::value);
    encoder << instance.blockedLoad;
    encoder << instance.madeHTTPS;
    encoder << instance.blockedCookies;
    encoder << instance.hasNotifications;
}

std::optional<WebCore::ContentRuleListResults::Summary> ArgumentCoder<WebCore::ContentRuleListResults::Summary>::decode(Decoder& decoder)
{
    auto blockedLoad = decoder.decode<bool>();
    auto madeHTTPS = decoder.decode<bool>();
    auto blockedCookies = decoder.decode<bool>();
    auto hasNotifications = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ContentRuleListResults::Summary {
            WTFMove(*blockedLoad),
            WTFMove(*madeHTTPS),
            WTFMove(*blockedCookies),
            WTFMove(*hasNotifications)
        }
    };
}

#endif

#if ENABLE(CONTENT_EXTENSIONS)
void ArgumentCoder<WebCore::ContentRuleListResults>::encode(Encoder& encoder, const WebCore::ContentRuleListResults& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.summary)>, WebCore::ContentRuleListResults::Summary>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.results)>, Vector<std::pair<String, WebCore::ContentRuleListResults::Result>>>);
    struct ShouldBeSameSizeAsContentRuleListResults : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ContentRuleListResults>, false> {
        WebCore::ContentRuleListResults::Summary summary;
        Vector<std::pair<String, WebCore::ContentRuleListResults::Result>> results;
    };
    static_assert(sizeof(ShouldBeSameSizeAsContentRuleListResults) == sizeof(WebCore::ContentRuleListResults));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ContentRuleListResults, summary)
        , offsetof(WebCore::ContentRuleListResults, results)
    >::value);
    encoder << instance.summary;
    encoder << instance.results;
}

std::optional<WebCore::ContentRuleListResults> ArgumentCoder<WebCore::ContentRuleListResults>::decode(Decoder& decoder)
{
    auto summary = decoder.decode<WebCore::ContentRuleListResults::Summary>();
    auto results = decoder.decode<Vector<std::pair<String, WebCore::ContentRuleListResults::Result>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ContentRuleListResults {
            WTFMove(*summary),
            WTFMove(*results)
        }
    };
}

#endif

void ArgumentCoder<WebCore::COEPInheritenceViolationReportBody>::encode(Encoder& encoder, const WebCore::COEPInheritenceViolationReportBody& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_disposition)>, WebCore::COEPDisposition>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_blockedURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_type)>, String>);
    struct ShouldBeSameSizeAsCOEPInheritenceViolationReportBody : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::COEPInheritenceViolationReportBody>, true> {
        WebCore::COEPDisposition m_disposition;
        URL m_blockedURL;
        String m_type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCOEPInheritenceViolationReportBody) == sizeof(WebCore::COEPInheritenceViolationReportBody));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::COEPInheritenceViolationReportBody, m_disposition)
        , offsetof(WebCore::COEPInheritenceViolationReportBody, m_blockedURL)
        , offsetof(WebCore::COEPInheritenceViolationReportBody, m_type)
    >::value);
    encoder << instance.m_disposition;
    encoder << instance.m_blockedURL;
    encoder << instance.m_type;
}

std::optional<Ref<WebCore::COEPInheritenceViolationReportBody>> ArgumentCoder<WebCore::COEPInheritenceViolationReportBody>::decode(Decoder& decoder)
{
    auto m_disposition = decoder.decode<WebCore::COEPDisposition>();
    auto m_blockedURL = decoder.decode<URL>();
    auto m_type = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::COEPInheritenceViolationReportBody::create(
            WTFMove(*m_disposition),
            WTFMove(*m_blockedURL),
            WTFMove(*m_type)
        )
    };
}

void ArgumentCoder<WebCore::CORPViolationReportBody>::encode(Encoder& encoder, const WebCore::CORPViolationReportBody& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_disposition)>, WebCore::COEPDisposition>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_blockedURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_destination)>, WebCore::FetchOptionsDestination>);
    struct ShouldBeSameSizeAsCORPViolationReportBody : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CORPViolationReportBody>, true> {
        WebCore::COEPDisposition m_disposition;
        URL m_blockedURL;
        WebCore::FetchOptionsDestination m_destination;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCORPViolationReportBody) == sizeof(WebCore::CORPViolationReportBody));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CORPViolationReportBody, m_disposition)
        , offsetof(WebCore::CORPViolationReportBody, m_blockedURL)
        , offsetof(WebCore::CORPViolationReportBody, m_destination)
    >::value);
    encoder << instance.m_disposition;
    encoder << instance.m_blockedURL;
    encoder << instance.m_destination;
}

std::optional<Ref<WebCore::CORPViolationReportBody>> ArgumentCoder<WebCore::CORPViolationReportBody>::decode(Decoder& decoder)
{
    auto m_disposition = decoder.decode<WebCore::COEPDisposition>();
    auto m_blockedURL = decoder.decode<URL>();
    auto m_destination = decoder.decode<WebCore::FetchOptionsDestination>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CORPViolationReportBody::create(
            WTFMove(*m_disposition),
            WTFMove(*m_blockedURL),
            WTFMove(*m_destination)
        )
    };
}

enum class WebCore_ReportBody_Subclass : IPC::EncodedVariantIndex {
    CSPViolationReportBody,
    COEPInheritenceViolationReportBody,
    CORPViolationReportBody,
    DeprecationReportBody,
    TestReportBody
};

void ArgumentCoder<WebCore::ReportBody>::encode(Encoder& encoder, const WebCore::ReportBody& instance)
{
    if (auto* subclass = dynamicDowncast<WebCore::CSPViolationReportBody>(instance)) {
        encoder << WebCore_ReportBody_Subclass::CSPViolationReportBody;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::COEPInheritenceViolationReportBody>(instance)) {
        encoder << WebCore_ReportBody_Subclass::COEPInheritenceViolationReportBody;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::CORPViolationReportBody>(instance)) {
        encoder << WebCore_ReportBody_Subclass::CORPViolationReportBody;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::DeprecationReportBody>(instance)) {
        encoder << WebCore_ReportBody_Subclass::DeprecationReportBody;
        encoder << *subclass;
        return;
    }
    if (auto* subclass = dynamicDowncast<WebCore::TestReportBody>(instance)) {
        encoder << WebCore_ReportBody_Subclass::TestReportBody;
        encoder << *subclass;
        return;
    }
    ASSERT_NOT_REACHED();
}

std::optional<Ref<WebCore::ReportBody>> ArgumentCoder<WebCore::ReportBody>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore_ReportBody_Subclass>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (type == WebCore_ReportBody_Subclass::CSPViolationReportBody) {
        auto result = decoder.decode<Ref<WebCore::CSPViolationReportBody>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_ReportBody_Subclass::COEPInheritenceViolationReportBody) {
        auto result = decoder.decode<Ref<WebCore::COEPInheritenceViolationReportBody>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_ReportBody_Subclass::CORPViolationReportBody) {
        auto result = decoder.decode<Ref<WebCore::CORPViolationReportBody>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_ReportBody_Subclass::DeprecationReportBody) {
        auto result = decoder.decode<Ref<WebCore::DeprecationReportBody>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    if (type == WebCore_ReportBody_Subclass::TestReportBody) {
        auto result = decoder.decode<Ref<WebCore::TestReportBody>>();
        if (UNLIKELY(!decoder.isValid()))
            return std::nullopt;
        return WTFMove(*result);
    }
    ASSERT_NOT_REACHED();
    return std::nullopt;
}

void ArgumentCoder<WebCore::SourceImage>::encode(Encoder& encoder, const WebCore::SourceImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.imageIdentifier())>, WebCore::RenderingResourceIdentifier>);
    encoder << instance.imageIdentifier();
}

void ArgumentCoder<WebCore::SourceImage>::encode(StreamConnectionEncoder& encoder, const WebCore::SourceImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.imageIdentifier())>, WebCore::RenderingResourceIdentifier>);
    encoder << instance.imageIdentifier();
}

std::optional<WebCore::SourceImage> ArgumentCoder<WebCore::SourceImage>::decode(Decoder& decoder)
{
    auto imageIdentifier = decoder.decode<WebCore::RenderingResourceIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SourceImage {
            WTFMove(*imageIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::LinkIcon>::encode(Encoder& encoder, const WebCore::LinkIcon& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::LinkIconType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mimeType)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, std::optional<unsigned>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.attributes)>, Vector<std::pair<String, String>>>);
    struct ShouldBeSameSizeAsLinkIcon : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::LinkIcon>, false> {
        URL url;
        WebCore::LinkIconType type;
        String mimeType;
        std::optional<unsigned> size;
        Vector<std::pair<String, String>> attributes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLinkIcon) == sizeof(WebCore::LinkIcon));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::LinkIcon, url)
        , offsetof(WebCore::LinkIcon, type)
        , offsetof(WebCore::LinkIcon, mimeType)
        , offsetof(WebCore::LinkIcon, size)
        , offsetof(WebCore::LinkIcon, attributes)
    >::value);
    encoder << instance.url;
    encoder << instance.type;
    encoder << instance.mimeType;
    encoder << instance.size;
    encoder << instance.attributes;
}

std::optional<WebCore::LinkIcon> ArgumentCoder<WebCore::LinkIcon>::decode(Decoder& decoder)
{
    auto url = decoder.decode<URL>();
    auto type = decoder.decode<WebCore::LinkIconType>();
    auto mimeType = decoder.decode<String>();
    auto size = decoder.decode<std::optional<unsigned>>();
    auto attributes = decoder.decode<Vector<std::pair<String, String>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LinkIcon {
            WTFMove(*url),
            WTFMove(*type),
            WTFMove(*mimeType),
            WTFMove(*size),
            WTFMove(*attributes)
        }
    };
}

void ArgumentCoder<WebCore::SupportedPluginIdentifier>::encode(Encoder& encoder, const WebCore::SupportedPluginIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.matchingDomain)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pluginIdentifier)>, String>);
    struct ShouldBeSameSizeAsSupportedPluginIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SupportedPluginIdentifier>, false> {
        String matchingDomain;
        String pluginIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSupportedPluginIdentifier) == sizeof(WebCore::SupportedPluginIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SupportedPluginIdentifier, matchingDomain)
        , offsetof(WebCore::SupportedPluginIdentifier, pluginIdentifier)
    >::value);
    encoder << instance.matchingDomain;
    encoder << instance.pluginIdentifier;
}

std::optional<WebCore::SupportedPluginIdentifier> ArgumentCoder<WebCore::SupportedPluginIdentifier>::decode(Decoder& decoder)
{
    auto matchingDomain = decoder.decode<String>();
    auto pluginIdentifier = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SupportedPluginIdentifier {
            WTFMove(*matchingDomain),
            WTFMove(*pluginIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::CSSValueKey>::encode(Encoder& encoder, const WebCore::CSSValueKey& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cssValueID)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useDarkAppearance)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useElevatedUserInterfaceLevel)>, bool>);
    struct ShouldBeSameSizeAsCSSValueKey : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CSSValueKey>, false> {
        unsigned cssValueID;
        bool useDarkAppearance;
        bool useElevatedUserInterfaceLevel;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCSSValueKey) == sizeof(WebCore::CSSValueKey));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CSSValueKey, cssValueID)
        , offsetof(WebCore::CSSValueKey, useDarkAppearance)
        , offsetof(WebCore::CSSValueKey, useElevatedUserInterfaceLevel)
    >::value);
    encoder << instance.cssValueID;
    encoder << instance.useDarkAppearance;
    encoder << instance.useElevatedUserInterfaceLevel;
}

std::optional<WebCore::CSSValueKey> ArgumentCoder<WebCore::CSSValueKey>::decode(Decoder& decoder)
{
    auto cssValueID = decoder.decode<unsigned>();
    auto useDarkAppearance = decoder.decode<bool>();
    auto useElevatedUserInterfaceLevel = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CSSValueKey {
            WTFMove(*cssValueID),
            WTFMove(*useDarkAppearance),
            WTFMove(*useElevatedUserInterfaceLevel)
        }
    };
}

void ArgumentCoder<WebCore::VideoFrameTimeMetadata>::encode(Encoder& encoder, const WebCore::VideoFrameTimeMetadata& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processingDuration)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.captureTime)>, std::optional<Seconds>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.receiveTime)>, std::optional<Seconds>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rtpTimestamp)>, std::optional<unsigned>>);
    struct ShouldBeSameSizeAsVideoFrameTimeMetadata : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::VideoFrameTimeMetadata>, false> {
        std::optional<double> processingDuration;
        std::optional<Seconds> captureTime;
        std::optional<Seconds> receiveTime;
        std::optional<unsigned> rtpTimestamp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVideoFrameTimeMetadata) == sizeof(WebCore::VideoFrameTimeMetadata));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::VideoFrameTimeMetadata, processingDuration)
        , offsetof(WebCore::VideoFrameTimeMetadata, captureTime)
        , offsetof(WebCore::VideoFrameTimeMetadata, receiveTime)
        , offsetof(WebCore::VideoFrameTimeMetadata, rtpTimestamp)
    >::value);
    encoder << instance.processingDuration;
    encoder << instance.captureTime;
    encoder << instance.receiveTime;
    encoder << instance.rtpTimestamp;
}

std::optional<WebCore::VideoFrameTimeMetadata> ArgumentCoder<WebCore::VideoFrameTimeMetadata>::decode(Decoder& decoder)
{
    auto processingDuration = decoder.decode<std::optional<double>>();
    auto captureTime = decoder.decode<std::optional<Seconds>>();
    auto receiveTime = decoder.decode<std::optional<Seconds>>();
    auto rtpTimestamp = decoder.decode<std::optional<unsigned>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::VideoFrameTimeMetadata {
            WTFMove(*processingDuration),
            WTFMove(*captureTime),
            WTFMove(*receiveTime),
            WTFMove(*rtpTimestamp)
        }
    };
}

void ArgumentCoder<WebCore::PositionedGlyphs>::encode(Encoder& encoder, const WebCore::PositionedGlyphs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.glyphs)>, Vector<WebCore::GlyphBufferGlyph>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.advances)>, Vector<WebCore::GlyphBufferAdvance>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localAnchor)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.smoothingMode)>, WebCore::FontSmoothingMode>);
    struct ShouldBeSameSizeAsPositionedGlyphs : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PositionedGlyphs>, false> {
        Vector<WebCore::GlyphBufferGlyph> glyphs;
        Vector<WebCore::GlyphBufferAdvance> advances;
        WebCore::FloatPoint localAnchor;
        WebCore::FontSmoothingMode smoothingMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPositionedGlyphs) == sizeof(WebCore::PositionedGlyphs));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PositionedGlyphs, glyphs)
        , offsetof(WebCore::PositionedGlyphs, advances)
        , offsetof(WebCore::PositionedGlyphs, localAnchor)
        , offsetof(WebCore::PositionedGlyphs, smoothingMode)
    >::value);
    encoder << instance.glyphs;
    encoder << instance.advances;
    encoder << instance.localAnchor;
    encoder << instance.smoothingMode;
}

void ArgumentCoder<WebCore::PositionedGlyphs>::encode(StreamConnectionEncoder& encoder, const WebCore::PositionedGlyphs& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.glyphs)>, Vector<WebCore::GlyphBufferGlyph>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.advances)>, Vector<WebCore::GlyphBufferAdvance>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localAnchor)>, WebCore::FloatPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.smoothingMode)>, WebCore::FontSmoothingMode>);
    struct ShouldBeSameSizeAsPositionedGlyphs : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PositionedGlyphs>, false> {
        Vector<WebCore::GlyphBufferGlyph> glyphs;
        Vector<WebCore::GlyphBufferAdvance> advances;
        WebCore::FloatPoint localAnchor;
        WebCore::FontSmoothingMode smoothingMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPositionedGlyphs) == sizeof(WebCore::PositionedGlyphs));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PositionedGlyphs, glyphs)
        , offsetof(WebCore::PositionedGlyphs, advances)
        , offsetof(WebCore::PositionedGlyphs, localAnchor)
        , offsetof(WebCore::PositionedGlyphs, smoothingMode)
    >::value);
    encoder << instance.glyphs;
    encoder << instance.advances;
    encoder << instance.localAnchor;
    encoder << instance.smoothingMode;
}

std::optional<WebCore::PositionedGlyphs> ArgumentCoder<WebCore::PositionedGlyphs>::decode(Decoder& decoder)
{
    auto glyphs = decoder.decode<Vector<WebCore::GlyphBufferGlyph>>();
    auto advances = decoder.decode<Vector<WebCore::GlyphBufferAdvance>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(glyphs->size() == advances->size()))
        return std::nullopt;
    auto localAnchor = decoder.decode<WebCore::FloatPoint>();
    auto smoothingMode = decoder.decode<WebCore::FontSmoothingMode>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PositionedGlyphs {
            WTFMove(*glyphs),
            WTFMove(*advances),
            WTFMove(*localAnchor),
            WTFMove(*smoothingMode)
        }
    };
}

void ArgumentCoder<WebCore::EventRegion>::encode(Encoder& encoder, const WebCore::EventRegion& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_region)>, WebCore::Region>);
#if ENABLE(TOUCH_ACTION_REGIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_touchActionRegions)>, Vector<WebCore::Region>>);
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_wheelEventListenerRegion)>, WebCore::Region>);
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_nonPassiveWheelEventListenerRegion)>, WebCore::Region>);
#endif
#if ENABLE(EDITABLE_REGION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_editableRegion)>, std::optional<WebCore::Region>>);
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_interactionRegions)>, Vector<WebCore::InteractionRegion>>);
#endif
    struct ShouldBeSameSizeAsEventRegion : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::EventRegion>, false> {
        WebCore::Region m_region;
#if ENABLE(TOUCH_ACTION_REGIONS)
        Vector<WebCore::Region> m_touchActionRegions;
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
        WebCore::Region m_wheelEventListenerRegion;
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
        WebCore::Region m_nonPassiveWheelEventListenerRegion;
#endif
#if ENABLE(EDITABLE_REGION)
        std::optional<WebCore::Region> m_editableRegion;
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
        Vector<WebCore::InteractionRegion> m_interactionRegions;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsEventRegion) == sizeof(WebCore::EventRegion));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::EventRegion, m_region)
#if ENABLE(TOUCH_ACTION_REGIONS)
        , offsetof(WebCore::EventRegion, m_touchActionRegions)
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
        , offsetof(WebCore::EventRegion, m_wheelEventListenerRegion)
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
        , offsetof(WebCore::EventRegion, m_nonPassiveWheelEventListenerRegion)
#endif
#if ENABLE(EDITABLE_REGION)
        , offsetof(WebCore::EventRegion, m_editableRegion)
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
        , offsetof(WebCore::EventRegion, m_interactionRegions)
#endif
    >::value);
    encoder << instance.m_region;
#if ENABLE(TOUCH_ACTION_REGIONS)
    encoder << instance.m_touchActionRegions;
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
    encoder << instance.m_wheelEventListenerRegion;
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
    encoder << instance.m_nonPassiveWheelEventListenerRegion;
#endif
#if ENABLE(EDITABLE_REGION)
    encoder << instance.m_editableRegion;
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
    encoder << instance.m_interactionRegions;
#endif
}

std::optional<WebCore::EventRegion> ArgumentCoder<WebCore::EventRegion>::decode(Decoder& decoder)
{
    auto m_region = decoder.decode<WebCore::Region>();
#if ENABLE(TOUCH_ACTION_REGIONS)
    auto m_touchActionRegions = decoder.decode<Vector<WebCore::Region>>();
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
    auto m_wheelEventListenerRegion = decoder.decode<WebCore::Region>();
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
    auto m_nonPassiveWheelEventListenerRegion = decoder.decode<WebCore::Region>();
#endif
#if ENABLE(EDITABLE_REGION)
    auto m_editableRegion = decoder.decode<std::optional<WebCore::Region>>();
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
    auto m_interactionRegions = decoder.decode<Vector<WebCore::InteractionRegion>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::EventRegion {
            WTFMove(*m_region),
#if ENABLE(TOUCH_ACTION_REGIONS)
            WTFMove(*m_touchActionRegions),
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
            WTFMove(*m_wheelEventListenerRegion),
#endif
#if ENABLE(WHEEL_EVENT_REGIONS)
            WTFMove(*m_nonPassiveWheelEventListenerRegion),
#endif
#if ENABLE(EDITABLE_REGION)
            WTFMove(*m_editableRegion),
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
            WTFMove(*m_interactionRegions)
#endif
        }
    };
}

void ArgumentCoder<WebCore::PresentationSize>::encode(Encoder& encoder, const WebCore::PresentationSize& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, std::optional<double>>);
    struct ShouldBeSameSizeAsPresentationSize : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PresentationSize>, false> {
        std::optional<double> width;
        std::optional<double> height;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPresentationSize) == sizeof(WebCore::PresentationSize));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PresentationSize, width)
        , offsetof(WebCore::PresentationSize, height)
    >::value);
    encoder << instance.width;
    encoder << instance.height;
}

std::optional<WebCore::PresentationSize> ArgumentCoder<WebCore::PresentationSize>::decode(Decoder& decoder)
{
    auto width = decoder.decode<std::optional<double>>();
    auto height = decoder.decode<std::optional<double>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PresentationSize {
            WTFMove(*width),
            WTFMove(*height)
        }
    };
}

void ArgumentCoder<WebCore::PasteboardItemInfo>::encode(Encoder& encoder, const WebCore::PasteboardItemInfo& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pathsForFileUpload)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.platformTypesForFileUpload)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.platformTypesByFidelity)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.suggestedFileName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preferredPresentationSize)>, WebCore::PresentationSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isNonTextType)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containsFileURLAndFileUploadContent)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webSafeTypesByFidelity)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preferredPresentationStyle)>, WebCore::PasteboardItemPresentationStyle>);
    struct ShouldBeSameSizeAsPasteboardItemInfo : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::PasteboardItemInfo>, false> {
        Vector<String> pathsForFileUpload;
        Vector<String> platformTypesForFileUpload;
        Vector<String> platformTypesByFidelity;
        String suggestedFileName;
        WebCore::PresentationSize preferredPresentationSize;
        bool isNonTextType;
        bool containsFileURLAndFileUploadContent;
        Vector<String> webSafeTypesByFidelity;
        WebCore::PasteboardItemPresentationStyle preferredPresentationStyle;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPasteboardItemInfo) == sizeof(WebCore::PasteboardItemInfo));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::PasteboardItemInfo, pathsForFileUpload)
        , offsetof(WebCore::PasteboardItemInfo, platformTypesForFileUpload)
        , offsetof(WebCore::PasteboardItemInfo, platformTypesByFidelity)
        , offsetof(WebCore::PasteboardItemInfo, suggestedFileName)
        , offsetof(WebCore::PasteboardItemInfo, preferredPresentationSize)
        , offsetof(WebCore::PasteboardItemInfo, isNonTextType)
        , offsetof(WebCore::PasteboardItemInfo, containsFileURLAndFileUploadContent)
        , offsetof(WebCore::PasteboardItemInfo, webSafeTypesByFidelity)
        , offsetof(WebCore::PasteboardItemInfo, preferredPresentationStyle)
    >::value);
    encoder << instance.pathsForFileUpload;
    encoder << instance.platformTypesForFileUpload;
    encoder << instance.platformTypesByFidelity;
    encoder << instance.suggestedFileName;
    encoder << instance.preferredPresentationSize;
    encoder << instance.isNonTextType;
    encoder << instance.containsFileURLAndFileUploadContent;
    encoder << instance.webSafeTypesByFidelity;
    encoder << instance.preferredPresentationStyle;
}

std::optional<WebCore::PasteboardItemInfo> ArgumentCoder<WebCore::PasteboardItemInfo>::decode(Decoder& decoder)
{
    auto pathsForFileUpload = decoder.decode<Vector<String>>();
    auto platformTypesForFileUpload = decoder.decode<Vector<String>>();
    auto platformTypesByFidelity = decoder.decode<Vector<String>>();
    auto suggestedFileName = decoder.decode<String>();
    auto preferredPresentationSize = decoder.decode<WebCore::PresentationSize>();
    auto isNonTextType = decoder.decode<bool>();
    auto containsFileURLAndFileUploadContent = decoder.decode<bool>();
    auto webSafeTypesByFidelity = decoder.decode<Vector<String>>();
    auto preferredPresentationStyle = decoder.decode<WebCore::PasteboardItemPresentationStyle>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::PasteboardItemInfo {
            WTFMove(*pathsForFileUpload),
            WTFMove(*platformTypesForFileUpload),
            WTFMove(*platformTypesByFidelity),
            WTFMove(*suggestedFileName),
            WTFMove(*preferredPresentationSize),
            WTFMove(*isNonTextType),
            WTFMove(*containsFileURLAndFileUploadContent),
            WTFMove(*webSafeTypesByFidelity),
            WTFMove(*preferredPresentationStyle)
        }
    };
}

#if ENABLE(MEDIA_CONTROLS_CONTEXT_MENUS) && USE(UICONTEXTMENU)
void ArgumentCoder<WebCore::MediaControlsContextMenuItem>::encode(Encoder& encoder, const WebCore::MediaControlsContextMenuItem& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.title)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.icon)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.checked)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.children)>, Vector<WebCore::MediaControlsContextMenuItem>>);
    struct ShouldBeSameSizeAsMediaControlsContextMenuItem : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MediaControlsContextMenuItem>, false> {
        uint64_t id;
        String title;
        String icon;
        bool checked;
        Vector<WebCore::MediaControlsContextMenuItem> children;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMediaControlsContextMenuItem) == sizeof(WebCore::MediaControlsContextMenuItem));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MediaControlsContextMenuItem, id)
        , offsetof(WebCore::MediaControlsContextMenuItem, title)
        , offsetof(WebCore::MediaControlsContextMenuItem, icon)
        , offsetof(WebCore::MediaControlsContextMenuItem, checked)
        , offsetof(WebCore::MediaControlsContextMenuItem, children)
    >::value);
    encoder << instance.id;
    encoder << instance.title;
    encoder << instance.icon;
    encoder << instance.checked;
    encoder << instance.children;
}

std::optional<WebCore::MediaControlsContextMenuItem> ArgumentCoder<WebCore::MediaControlsContextMenuItem>::decode(Decoder& decoder)
{
    auto id = decoder.decode<uint64_t>();
    auto title = decoder.decode<String>();
    auto icon = decoder.decode<String>();
    auto checked = decoder.decode<bool>();
    auto children = decoder.decode<Vector<WebCore::MediaControlsContextMenuItem>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MediaControlsContextMenuItem {
            WTFMove(*id),
            WTFMove(*title),
            WTFMove(*icon),
            WTFMove(*checked),
            WTFMove(*children)
        }
    };
}

#endif

void ArgumentCoder<WebCore::WorkerOptions>::encode(Encoder& encoder, const WebCore::WorkerOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WorkerType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.credentials)>, WebCore::FetchRequestCredentials>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    struct ShouldBeSameSizeAsWorkerOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WorkerOptions>, false> {
        WebCore::WorkerType type;
        WebCore::FetchRequestCredentials credentials;
        String name;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWorkerOptions) == sizeof(WebCore::WorkerOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WorkerOptions, type)
        , offsetof(WebCore::WorkerOptions, credentials)
        , offsetof(WebCore::WorkerOptions, name)
    >::value);
    encoder << instance.type;
    encoder << instance.credentials;
    encoder << instance.name;
}

std::optional<WebCore::WorkerOptions> ArgumentCoder<WebCore::WorkerOptions>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::WorkerType>();
    auto credentials = decoder.decode<WebCore::FetchRequestCredentials>();
    auto name = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WorkerOptions {
            WTFMove(*type),
            WTFMove(*credentials),
            WTFMove(*name)
        }
    };
}

void ArgumentCoder<WebCore::WorkerInitializationData>::encode(Encoder& encoder, const WebCore::WorkerInitializationData& instance)
{
#if ENABLE(SERVICE_WORKER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serviceWorkerData)>, std::optional<WebCore::ServiceWorkerData>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clientIdentifier)>, std::optional<WebCore::ScriptExecutionContextIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userAgent)>, String>);
    struct ShouldBeSameSizeAsWorkerInitializationData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WorkerInitializationData>, false> {
#if ENABLE(SERVICE_WORKER)
        std::optional<WebCore::ServiceWorkerData> serviceWorkerData;
#endif
        std::optional<WebCore::ScriptExecutionContextIdentifier> clientIdentifier;
        String userAgent;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWorkerInitializationData) == sizeof(WebCore::WorkerInitializationData));
    static_assert(MembersInCorrectOrder < 0
#if ENABLE(SERVICE_WORKER)
        , offsetof(WebCore::WorkerInitializationData, serviceWorkerData)
#endif
        , offsetof(WebCore::WorkerInitializationData, clientIdentifier)
        , offsetof(WebCore::WorkerInitializationData, userAgent)
    >::value);
#if ENABLE(SERVICE_WORKER)
    encoder << instance.serviceWorkerData;
#endif
    encoder << instance.clientIdentifier;
    encoder << instance.userAgent;
}

std::optional<WebCore::WorkerInitializationData> ArgumentCoder<WebCore::WorkerInitializationData>::decode(Decoder& decoder)
{
#if ENABLE(SERVICE_WORKER)
    auto serviceWorkerData = decoder.decode<std::optional<WebCore::ServiceWorkerData>>();
#endif
    auto clientIdentifier = decoder.decode<std::optional<WebCore::ScriptExecutionContextIdentifier>>();
    auto userAgent = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WorkerInitializationData {
#if ENABLE(SERVICE_WORKER)
            WTFMove(*serviceWorkerData),
#endif
            WTFMove(*clientIdentifier),
            WTFMove(*userAgent)
        }
    };
}

void ArgumentCoder<WebCore::WorkerFetchResult>::encode(Encoder& encoder, const WebCore::WorkerFetchResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.script)>, WebCore::ScriptBuffer>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.responseURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.certificateInfo)>, WebCore::CertificateInfo>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSecurityPolicy)>, WebCore::ContentSecurityPolicyResponseHeaders>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginEmbedderPolicy)>, WebCore::CrossOriginEmbedderPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrerPolicy)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.error)>, WebCore::ResourceError>);
    struct ShouldBeSameSizeAsWorkerFetchResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WorkerFetchResult>, false> {
        WebCore::ScriptBuffer script;
        URL responseURL;
        WebCore::CertificateInfo certificateInfo;
        WebCore::ContentSecurityPolicyResponseHeaders contentSecurityPolicy;
        WebCore::CrossOriginEmbedderPolicy crossOriginEmbedderPolicy;
        String referrerPolicy;
        WebCore::ResourceError error;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWorkerFetchResult) == sizeof(WebCore::WorkerFetchResult));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WorkerFetchResult, script)
        , offsetof(WebCore::WorkerFetchResult, responseURL)
        , offsetof(WebCore::WorkerFetchResult, certificateInfo)
        , offsetof(WebCore::WorkerFetchResult, contentSecurityPolicy)
        , offsetof(WebCore::WorkerFetchResult, crossOriginEmbedderPolicy)
        , offsetof(WebCore::WorkerFetchResult, referrerPolicy)
        , offsetof(WebCore::WorkerFetchResult, error)
    >::value);
    encoder << instance.script;
    encoder << instance.responseURL;
    encoder << instance.certificateInfo;
    encoder << instance.contentSecurityPolicy;
    encoder << instance.crossOriginEmbedderPolicy;
    encoder << instance.referrerPolicy;
    encoder << instance.error;
}

std::optional<WebCore::WorkerFetchResult> ArgumentCoder<WebCore::WorkerFetchResult>::decode(Decoder& decoder)
{
    auto script = decoder.decode<WebCore::ScriptBuffer>();
    auto responseURL = decoder.decode<URL>();
    auto certificateInfo = decoder.decode<WebCore::CertificateInfo>();
    auto contentSecurityPolicy = decoder.decode<WebCore::ContentSecurityPolicyResponseHeaders>();
    auto crossOriginEmbedderPolicy = decoder.decode<WebCore::CrossOriginEmbedderPolicy>();
    auto referrerPolicy = decoder.decode<String>();
    auto error = decoder.decode<WebCore::ResourceError>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WorkerFetchResult {
            WTFMove(*script),
            WTFMove(*responseURL),
            WTFMove(*certificateInfo),
            WTFMove(*contentSecurityPolicy),
            WTFMove(*crossOriginEmbedderPolicy),
            WTFMove(*referrerPolicy),
            WTFMove(*error)
        }
    };
}

void ArgumentCoder<WebCore::SharedWorkerKey>::encode(Encoder& encoder, const WebCore::SharedWorkerKey& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, WebCore::ClientOrigin>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    struct ShouldBeSameSizeAsSharedWorkerKey : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::SharedWorkerKey>, false> {
        WebCore::ClientOrigin origin;
        URL url;
        String name;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSharedWorkerKey) == sizeof(WebCore::SharedWorkerKey));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::SharedWorkerKey, origin)
        , offsetof(WebCore::SharedWorkerKey, url)
        , offsetof(WebCore::SharedWorkerKey, name)
    >::value);
    encoder << instance.origin;
    encoder << instance.url;
    encoder << instance.name;
}

std::optional<WebCore::SharedWorkerKey> ArgumentCoder<WebCore::SharedWorkerKey>::decode(Decoder& decoder)
{
    auto origin = decoder.decode<WebCore::ClientOrigin>();
    auto url = decoder.decode<URL>();
    auto name = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::SharedWorkerKey {
            WTFMove(*origin),
            WTFMove(*url),
            WTFMove(*name)
        }
    };
}

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerRegistrationOptions>::encode(Encoder& encoder, const WebCore::ServiceWorkerRegistrationOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scope)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WorkerType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.updateViaCache)>, WebCore::ServiceWorkerUpdateViaCache>);
    struct ShouldBeSameSizeAsServiceWorkerRegistrationOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerRegistrationOptions>, false> {
        String scope;
        WebCore::WorkerType type;
        WebCore::ServiceWorkerUpdateViaCache updateViaCache;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerRegistrationOptions) == sizeof(WebCore::ServiceWorkerRegistrationOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerRegistrationOptions, scope)
        , offsetof(WebCore::ServiceWorkerRegistrationOptions, type)
        , offsetof(WebCore::ServiceWorkerRegistrationOptions, updateViaCache)
    >::value);
    encoder << instance.scope;
    encoder << instance.type;
    encoder << instance.updateViaCache;
}

std::optional<WebCore::ServiceWorkerRegistrationOptions> ArgumentCoder<WebCore::ServiceWorkerRegistrationOptions>::decode(Decoder& decoder)
{
    auto scope = decoder.decode<String>();
    auto type = decoder.decode<WebCore::WorkerType>();
    auto updateViaCache = decoder.decode<WebCore::ServiceWorkerUpdateViaCache>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerRegistrationOptions {
            WTFMove(*scope),
            WTFMove(*type),
            WTFMove(*updateViaCache)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerRegistrationKey>::encode(Encoder& encoder, const WebCore::ServiceWorkerRegistrationKey& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topOrigin())>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scope())>, URL>);
    encoder << instance.topOrigin();
    encoder << instance.scope();
}

std::optional<WebCore::ServiceWorkerRegistrationKey> ArgumentCoder<WebCore::ServiceWorkerRegistrationKey>::decode(Decoder& decoder)
{
    auto topOrigin = decoder.decode<WebCore::SecurityOriginData>();
    auto scope = decoder.decode<URL>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerRegistrationKey {
            WTFMove(*topOrigin),
            WTFMove(*scope)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerRegistrationData>::encode(Encoder& encoder, const WebCore::ServiceWorkerRegistrationData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.key)>, WebCore::ServiceWorkerRegistrationKey>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebCore::ServiceWorkerRegistrationIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scopeURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.updateViaCache)>, WebCore::ServiceWorkerUpdateViaCache>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastUpdateTime)>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.installingWorker)>, std::optional<WebCore::ServiceWorkerData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.waitingWorker)>, std::optional<WebCore::ServiceWorkerData>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.activeWorker)>, std::optional<WebCore::ServiceWorkerData>>);
    struct ShouldBeSameSizeAsServiceWorkerRegistrationData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerRegistrationData>, false> {
        WebCore::ServiceWorkerRegistrationKey key;
        WebCore::ServiceWorkerRegistrationIdentifier identifier;
        URL scopeURL;
        WebCore::ServiceWorkerUpdateViaCache updateViaCache;
        WallTime lastUpdateTime;
        std::optional<WebCore::ServiceWorkerData> installingWorker;
        std::optional<WebCore::ServiceWorkerData> waitingWorker;
        std::optional<WebCore::ServiceWorkerData> activeWorker;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerRegistrationData) == sizeof(WebCore::ServiceWorkerRegistrationData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerRegistrationData, key)
        , offsetof(WebCore::ServiceWorkerRegistrationData, identifier)
        , offsetof(WebCore::ServiceWorkerRegistrationData, scopeURL)
        , offsetof(WebCore::ServiceWorkerRegistrationData, updateViaCache)
        , offsetof(WebCore::ServiceWorkerRegistrationData, lastUpdateTime)
        , offsetof(WebCore::ServiceWorkerRegistrationData, installingWorker)
        , offsetof(WebCore::ServiceWorkerRegistrationData, waitingWorker)
        , offsetof(WebCore::ServiceWorkerRegistrationData, activeWorker)
    >::value);
    encoder << instance.key;
    encoder << instance.identifier;
    encoder << instance.scopeURL;
    encoder << instance.updateViaCache;
    encoder << instance.lastUpdateTime;
    encoder << instance.installingWorker;
    encoder << instance.waitingWorker;
    encoder << instance.activeWorker;
}

std::optional<WebCore::ServiceWorkerRegistrationData> ArgumentCoder<WebCore::ServiceWorkerRegistrationData>::decode(Decoder& decoder)
{
    auto key = decoder.decode<WebCore::ServiceWorkerRegistrationKey>();
    auto identifier = decoder.decode<WebCore::ServiceWorkerRegistrationIdentifier>();
    auto scopeURL = decoder.decode<URL>();
    auto updateViaCache = decoder.decode<WebCore::ServiceWorkerUpdateViaCache>();
    auto lastUpdateTime = decoder.decode<WallTime>();
    auto installingWorker = decoder.decode<std::optional<WebCore::ServiceWorkerData>>();
    auto waitingWorker = decoder.decode<std::optional<WebCore::ServiceWorkerData>>();
    auto activeWorker = decoder.decode<std::optional<WebCore::ServiceWorkerData>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerRegistrationData {
            WTFMove(*key),
            WTFMove(*identifier),
            WTFMove(*scopeURL),
            WTFMove(*updateViaCache),
            WTFMove(*lastUpdateTime),
            WTFMove(*installingWorker),
            WTFMove(*waitingWorker),
            WTFMove(*activeWorker)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerJobDataIdentifier>::encode(Encoder& encoder, const WebCore::ServiceWorkerJobDataIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.connectionIdentifier)>, WebCore::SWServerConnectionIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.jobIdentifier)>, WebCore::ServiceWorkerJobIdentifier>);
    struct ShouldBeSameSizeAsServiceWorkerJobDataIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerJobDataIdentifier>, false> {
        WebCore::SWServerConnectionIdentifier connectionIdentifier;
        WebCore::ServiceWorkerJobIdentifier jobIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerJobDataIdentifier) == sizeof(WebCore::ServiceWorkerJobDataIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerJobDataIdentifier, connectionIdentifier)
        , offsetof(WebCore::ServiceWorkerJobDataIdentifier, jobIdentifier)
    >::value);
    encoder << instance.connectionIdentifier;
    encoder << instance.jobIdentifier;
}

std::optional<WebCore::ServiceWorkerJobDataIdentifier> ArgumentCoder<WebCore::ServiceWorkerJobDataIdentifier>::decode(Decoder& decoder)
{
    auto connectionIdentifier = decoder.decode<WebCore::SWServerConnectionIdentifier>();
    auto jobIdentifier = decoder.decode<WebCore::ServiceWorkerJobIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerJobDataIdentifier {
            WTFMove(*connectionIdentifier),
            WTFMove(*jobIdentifier)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerImportedScript>::encode(Encoder& encoder, const WebCore::ServiceWorkerImportedScript& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.script)>, WebCore::ScriptBuffer>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.responseURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mimeType)>, String>);
    struct ShouldBeSameSizeAsServiceWorkerImportedScript : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerImportedScript>, false> {
        WebCore::ScriptBuffer script;
        URL responseURL;
        String mimeType;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerImportedScript) == sizeof(WebCore::ServiceWorkerImportedScript));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerImportedScript, script)
        , offsetof(WebCore::ServiceWorkerImportedScript, responseURL)
        , offsetof(WebCore::ServiceWorkerImportedScript, mimeType)
    >::value);
    encoder << instance.script;
    encoder << instance.responseURL;
    encoder << instance.mimeType;
}

std::optional<WebCore::ServiceWorkerImportedScript> ArgumentCoder<WebCore::ServiceWorkerImportedScript>::decode(Decoder& decoder)
{
    auto script = decoder.decode<WebCore::ScriptBuffer>();
    auto responseURL = decoder.decode<URL>();
    auto mimeType = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerImportedScript {
            WTFMove(*script),
            WTFMove(*responseURL),
            WTFMove(*mimeType)
        }
    };
}

#endif

#if ENABLE(SERVICE_WORKER)
void ArgumentCoder<WebCore::ServiceWorkerContextData>::encode(Encoder& encoder, const WebCore::ServiceWorkerContextData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.jobDataIdentifier)>, std::optional<WebCore::ServiceWorkerJobDataIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registration)>, WebCore::ServiceWorkerRegistrationData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serviceWorkerIdentifier)>, WebCore::ServiceWorkerIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.script)>, WebCore::ScriptBuffer>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.certificateInfo)>, WebCore::CertificateInfo>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSecurityPolicy)>, WebCore::ContentSecurityPolicyResponseHeaders>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginEmbedderPolicy)>, WebCore::CrossOriginEmbedderPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.referrerPolicy)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scriptURL)>, URL>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.workerType)>, WebCore::WorkerType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.loadedFromDisk)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastNavigationWasAppInitiated)>, std::optional<WebCore::LastNavigationWasAppInitiated>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scriptResourceMap)>, MemoryCompactRobinHoodHashMap<URL, WebCore::ServiceWorkerImportedScript>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serviceWorkerPageIdentifier)>, std::optional<WebCore::ScriptExecutionContextIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.navigationPreloadState)>, WebCore::NavigationPreloadState>);
    struct ShouldBeSameSizeAsServiceWorkerContextData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ServiceWorkerContextData>, false> {
        std::optional<WebCore::ServiceWorkerJobDataIdentifier> jobDataIdentifier;
        WebCore::ServiceWorkerRegistrationData registration;
        WebCore::ServiceWorkerIdentifier serviceWorkerIdentifier;
        WebCore::ScriptBuffer script;
        WebCore::CertificateInfo certificateInfo;
        WebCore::ContentSecurityPolicyResponseHeaders contentSecurityPolicy;
        WebCore::CrossOriginEmbedderPolicy crossOriginEmbedderPolicy;
        String referrerPolicy;
        URL scriptURL;
        WebCore::WorkerType workerType;
        bool loadedFromDisk;
        std::optional<WebCore::LastNavigationWasAppInitiated> lastNavigationWasAppInitiated;
        MemoryCompactRobinHoodHashMap<URL, WebCore::ServiceWorkerImportedScript> scriptResourceMap;
        std::optional<WebCore::ScriptExecutionContextIdentifier> serviceWorkerPageIdentifier;
        WebCore::NavigationPreloadState navigationPreloadState;
    };
    static_assert(sizeof(ShouldBeSameSizeAsServiceWorkerContextData) == sizeof(WebCore::ServiceWorkerContextData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ServiceWorkerContextData, jobDataIdentifier)
        , offsetof(WebCore::ServiceWorkerContextData, registration)
        , offsetof(WebCore::ServiceWorkerContextData, serviceWorkerIdentifier)
        , offsetof(WebCore::ServiceWorkerContextData, script)
        , offsetof(WebCore::ServiceWorkerContextData, certificateInfo)
        , offsetof(WebCore::ServiceWorkerContextData, contentSecurityPolicy)
        , offsetof(WebCore::ServiceWorkerContextData, crossOriginEmbedderPolicy)
        , offsetof(WebCore::ServiceWorkerContextData, referrerPolicy)
        , offsetof(WebCore::ServiceWorkerContextData, scriptURL)
        , offsetof(WebCore::ServiceWorkerContextData, workerType)
        , offsetof(WebCore::ServiceWorkerContextData, loadedFromDisk)
        , offsetof(WebCore::ServiceWorkerContextData, lastNavigationWasAppInitiated)
        , offsetof(WebCore::ServiceWorkerContextData, scriptResourceMap)
        , offsetof(WebCore::ServiceWorkerContextData, serviceWorkerPageIdentifier)
        , offsetof(WebCore::ServiceWorkerContextData, navigationPreloadState)
    >::value);
    encoder << instance.jobDataIdentifier;
    encoder << instance.registration;
    encoder << instance.serviceWorkerIdentifier;
    encoder << instance.script;
    encoder << instance.certificateInfo;
    encoder << instance.contentSecurityPolicy;
    encoder << instance.crossOriginEmbedderPolicy;
    encoder << instance.referrerPolicy;
    encoder << instance.scriptURL;
    encoder << instance.workerType;
    encoder << instance.loadedFromDisk;
    encoder << instance.lastNavigationWasAppInitiated;
    encoder << instance.scriptResourceMap;
    encoder << instance.serviceWorkerPageIdentifier;
    encoder << instance.navigationPreloadState;
}

std::optional<WebCore::ServiceWorkerContextData> ArgumentCoder<WebCore::ServiceWorkerContextData>::decode(Decoder& decoder)
{
    auto jobDataIdentifier = decoder.decode<std::optional<WebCore::ServiceWorkerJobDataIdentifier>>();
    auto registration = decoder.decode<WebCore::ServiceWorkerRegistrationData>();
    auto serviceWorkerIdentifier = decoder.decode<WebCore::ServiceWorkerIdentifier>();
    auto script = decoder.decode<WebCore::ScriptBuffer>();
    auto certificateInfo = decoder.decode<WebCore::CertificateInfo>();
    auto contentSecurityPolicy = decoder.decode<WebCore::ContentSecurityPolicyResponseHeaders>();
    auto crossOriginEmbedderPolicy = decoder.decode<WebCore::CrossOriginEmbedderPolicy>();
    auto referrerPolicy = decoder.decode<String>();
    auto scriptURL = decoder.decode<URL>();
    auto workerType = decoder.decode<WebCore::WorkerType>();
    auto loadedFromDisk = decoder.decode<bool>();
    auto lastNavigationWasAppInitiated = decoder.decode<std::optional<WebCore::LastNavigationWasAppInitiated>>();
    auto scriptResourceMap = decoder.decode<MemoryCompactRobinHoodHashMap<URL, WebCore::ServiceWorkerImportedScript>>();
    auto serviceWorkerPageIdentifier = decoder.decode<std::optional<WebCore::ScriptExecutionContextIdentifier>>();
    auto navigationPreloadState = decoder.decode<WebCore::NavigationPreloadState>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ServiceWorkerContextData {
            WTFMove(*jobDataIdentifier),
            WTFMove(*registration),
            WTFMove(*serviceWorkerIdentifier),
            WTFMove(*script),
            WTFMove(*certificateInfo),
            WTFMove(*contentSecurityPolicy),
            WTFMove(*crossOriginEmbedderPolicy),
            WTFMove(*referrerPolicy),
            WTFMove(*scriptURL),
            WTFMove(*workerType),
            WTFMove(*loadedFromDisk),
            WTFMove(*lastNavigationWasAppInitiated),
            WTFMove(*scriptResourceMap),
            WTFMove(*serviceWorkerPageIdentifier),
            WTFMove(*navigationPreloadState)
        }
    };
}

#endif

void ArgumentCoder<JSC::ArrayBuffer>::encode(Encoder& encoder, const JSC::ArrayBuffer& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.toVector())>, Vector<uint8_t>>);
    encoder << instance.toVector();
}

std::optional<Ref<JSC::ArrayBuffer>> ArgumentCoder<JSC::ArrayBuffer>::decode(Decoder& decoder)
{
    auto toVector = decoder.decode<Vector<uint8_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        JSC::ArrayBuffer::create(
            WTFMove(*toVector)
        )
    };
}

void ArgumentCoder<WebCore::GlobalFrameIdentifier>::encode(Encoder& encoder, const WebCore::GlobalFrameIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageID)>, WebCore::PageIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID)>, WebCore::FrameIdentifier>);
    struct ShouldBeSameSizeAsGlobalFrameIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GlobalFrameIdentifier>, false> {
        WebCore::PageIdentifier pageID;
        WebCore::FrameIdentifier frameID;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGlobalFrameIdentifier) == sizeof(WebCore::GlobalFrameIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GlobalFrameIdentifier, pageID)
        , offsetof(WebCore::GlobalFrameIdentifier, frameID)
    >::value);
    encoder << instance.pageID;
    encoder << instance.frameID;
}

std::optional<WebCore::GlobalFrameIdentifier> ArgumentCoder<WebCore::GlobalFrameIdentifier>::decode(Decoder& decoder)
{
    auto pageID = decoder.decode<WebCore::PageIdentifier>();
    auto frameID = decoder.decode<WebCore::FrameIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GlobalFrameIdentifier {
            WTFMove(*pageID),
            WTFMove(*frameID)
        }
    };
}

void ArgumentCoder<WebCore::GlobalWindowIdentifier>::encode(Encoder& encoder, const WebCore::GlobalWindowIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processIdentifier)>, WebCore::ProcessIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.windowIdentifier)>, WebCore::WindowIdentifier>);
    struct ShouldBeSameSizeAsGlobalWindowIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::GlobalWindowIdentifier>, false> {
        WebCore::ProcessIdentifier processIdentifier;
        WebCore::WindowIdentifier windowIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsGlobalWindowIdentifier) == sizeof(WebCore::GlobalWindowIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::GlobalWindowIdentifier, processIdentifier)
        , offsetof(WebCore::GlobalWindowIdentifier, windowIdentifier)
    >::value);
    encoder << instance.processIdentifier;
    encoder << instance.windowIdentifier;
}

std::optional<WebCore::GlobalWindowIdentifier> ArgumentCoder<WebCore::GlobalWindowIdentifier>::decode(Decoder& decoder)
{
    auto processIdentifier = decoder.decode<WebCore::ProcessIdentifier>();
    auto windowIdentifier = decoder.decode<WebCore::WindowIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::GlobalWindowIdentifier {
            WTFMove(*processIdentifier),
            WTFMove(*windowIdentifier)
        }
    };
}

void ArgumentCoder<WebCore::InteractionRegion>::encode(Encoder& encoder, const WebCore::InteractionRegion& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::InteractionRegion::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elementIdentifier)>, WebCore::ElementIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rectInLayerCoordinates)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.borderRadius)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maskedCorners)>, OptionSet<WebCore::InteractionRegion::CornerMask>>);
    struct ShouldBeSameSizeAsInteractionRegion : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::InteractionRegion>, false> {
        WebCore::InteractionRegion::Type type;
        WebCore::ElementIdentifier elementIdentifier;
        WebCore::IntRect rectInLayerCoordinates;
        float borderRadius;
        OptionSet<WebCore::InteractionRegion::CornerMask> maskedCorners;
    };
    static_assert(sizeof(ShouldBeSameSizeAsInteractionRegion) == sizeof(WebCore::InteractionRegion));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::InteractionRegion, type)
        , offsetof(WebCore::InteractionRegion, elementIdentifier)
        , offsetof(WebCore::InteractionRegion, rectInLayerCoordinates)
        , offsetof(WebCore::InteractionRegion, borderRadius)
        , offsetof(WebCore::InteractionRegion, maskedCorners)
    >::value);
    encoder << instance.type;
    encoder << instance.elementIdentifier;
    encoder << instance.rectInLayerCoordinates;
    encoder << instance.borderRadius;
    encoder << instance.maskedCorners;
}

std::optional<WebCore::InteractionRegion> ArgumentCoder<WebCore::InteractionRegion>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::InteractionRegion::Type>();
    auto elementIdentifier = decoder.decode<WebCore::ElementIdentifier>();
    auto rectInLayerCoordinates = decoder.decode<WebCore::IntRect>();
    auto borderRadius = decoder.decode<float>();
    auto maskedCorners = decoder.decode<OptionSet<WebCore::InteractionRegion::CornerMask>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::InteractionRegion {
            WTFMove(*type),
            WTFMove(*elementIdentifier),
            WTFMove(*rectInLayerCoordinates),
            WTFMove(*borderRadius),
            WTFMove(*maskedCorners)
        }
    };
}

void ArgumentCoder<WebCore::Region::Span>::encode(Encoder& encoder, const WebCore::Region::Span& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y)>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.segmentIndex)>, size_t>);
    struct ShouldBeSameSizeAsSpan : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Region::Span>, false> {
        int y;
        size_t segmentIndex;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSpan) == sizeof(WebCore::Region::Span));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Region::Span, y)
        , offsetof(WebCore::Region::Span, segmentIndex)
    >::value);
    encoder << instance.y;
    encoder << instance.segmentIndex;
}

std::optional<WebCore::Region::Span> ArgumentCoder<WebCore::Region::Span>::decode(Decoder& decoder)
{
    auto y = decoder.decode<int>();
    auto segmentIndex = decoder.decode<size_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Region::Span {
            WTFMove(*y),
            WTFMove(*segmentIndex)
        }
    };
}

void ArgumentCoder<WebCore::Region::Shape>::encode(Encoder& encoder, const WebCore::Region::Shape& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_segments)>, Vector<int, 32>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_spans)>, Vector<WebCore::Region::Span, 16>>);
    struct ShouldBeSameSizeAsShape : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::Region::Shape>, false> {
        Vector<int, 32> m_segments;
        Vector<WebCore::Region::Span, 16> m_spans;
    };
    static_assert(sizeof(ShouldBeSameSizeAsShape) == sizeof(WebCore::Region::Shape));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::Region::Shape, m_segments)
        , offsetof(WebCore::Region::Shape, m_spans)
    >::value);
    encoder << instance.m_segments;
    encoder << instance.m_spans;
}

std::optional<WebCore::Region::Shape> ArgumentCoder<WebCore::Region::Shape>::decode(Decoder& decoder)
{
    auto m_segments = decoder.decode<Vector<int, 32>>();
    auto m_spans = decoder.decode<Vector<WebCore::Region::Span, 16>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Region::Shape {
            WTFMove(*m_segments),
            WTFMove(*m_spans)
        }
    };
}

void ArgumentCoder<WebCore::Region>::encode(Encoder& encoder, const WebCore::Region& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bounds())>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.copyShape())>, std::unique_ptr<WebCore::Region::Shape>>);
    encoder << instance.bounds();
    encoder << instance.copyShape();
}

std::optional<WebCore::Region> ArgumentCoder<WebCore::Region>::decode(Decoder& decoder)
{
    auto bounds = decoder.decode<WebCore::IntRect>();
    auto copyShape = decoder.decode<std::unique_ptr<WebCore::Region::Shape>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Region {
            WTFMove(*bounds),
            WTFMove(*copyShape)
        }
    };
}

void ArgumentCoder<WebCore::ISOWebVTTCue>::encode(Encoder& encoder, const WebCore::ISOWebVTTCue& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentationTime())>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration())>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id())>, AtomString>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cueText())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.settings())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceID())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.originalStartTime())>, String>);
    encoder << instance.presentationTime();
    encoder << instance.duration();
    encoder << instance.id();
    encoder << instance.cueText();
    encoder << instance.settings();
    encoder << instance.sourceID();
    encoder << instance.originalStartTime();
}

std::optional<WebCore::ISOWebVTTCue> ArgumentCoder<WebCore::ISOWebVTTCue>::decode(Decoder& decoder)
{
    auto presentationTime = decoder.decode<MediaTime>();
    auto duration = decoder.decode<MediaTime>();
    auto id = decoder.decode<AtomString>();
    auto cueText = decoder.decode<String>();
    auto settings = decoder.decode<String>();
    auto sourceID = decoder.decode<String>();
    auto originalStartTime = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ISOWebVTTCue {
            WTFMove(*presentationTime),
            WTFMove(*duration),
            WTFMove(*id),
            WTFMove(*cueText),
            WTFMove(*settings),
            WTFMove(*sourceID),
            WTFMove(*originalStartTime)
        }
    };
}

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration>::encode(Encoder& encoder, const WebCore::MockWebAuthenticationConfiguration::LocalConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userVerification)>, WebCore::MockWebAuthenticationConfiguration::UserVerification>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.acceptAttestation)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.privateKeyBase64)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userCertificateBase64)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.intermediateCACertificateBase64)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preferredCredentialIdBase64)>, String>);
    struct ShouldBeSameSizeAsLocalConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration>, false> {
        WebCore::MockWebAuthenticationConfiguration::UserVerification userVerification;
        bool acceptAttestation;
        String privateKeyBase64;
        String userCertificateBase64;
        String intermediateCACertificateBase64;
        String preferredCredentialIdBase64;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLocalConfiguration) == sizeof(WebCore::MockWebAuthenticationConfiguration::LocalConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MockWebAuthenticationConfiguration::LocalConfiguration, userVerification)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::LocalConfiguration, acceptAttestation)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::LocalConfiguration, privateKeyBase64)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::LocalConfiguration, userCertificateBase64)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::LocalConfiguration, intermediateCACertificateBase64)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::LocalConfiguration, preferredCredentialIdBase64)
    >::value);
    encoder << instance.userVerification;
    encoder << instance.acceptAttestation;
    encoder << instance.privateKeyBase64;
    encoder << instance.userCertificateBase64;
    encoder << instance.intermediateCACertificateBase64;
    encoder << instance.preferredCredentialIdBase64;
}

std::optional<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration> ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration>::decode(Decoder& decoder)
{
    auto userVerification = decoder.decode<WebCore::MockWebAuthenticationConfiguration::UserVerification>();
    auto acceptAttestation = decoder.decode<bool>();
    auto privateKeyBase64 = decoder.decode<String>();
    auto userCertificateBase64 = decoder.decode<String>();
    auto intermediateCACertificateBase64 = decoder.decode<String>();
    auto preferredCredentialIdBase64 = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MockWebAuthenticationConfiguration::LocalConfiguration {
            WTFMove(*userVerification),
            WTFMove(*acceptAttestation),
            WTFMove(*privateKeyBase64),
            WTFMove(*userCertificateBase64),
            WTFMove(*intermediateCACertificateBase64),
            WTFMove(*preferredCredentialIdBase64)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::HidConfiguration>::encode(Encoder& encoder, const WebCore::MockWebAuthenticationConfiguration::HidConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.payloadBase64)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stage)>, WebCore::MockWebAuthenticationConfiguration::HidStage>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subStage)>, WebCore::MockWebAuthenticationConfiguration::HidSubStage>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.error)>, WebCore::MockWebAuthenticationConfiguration::HidError>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isU2f)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keepAlive)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fastDataArrival)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.continueAfterErrorData)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canDowngrade)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.expectCancel)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportClientPin)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportInternalUV)>, bool>);
    struct ShouldBeSameSizeAsHidConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MockWebAuthenticationConfiguration::HidConfiguration>, false> {
        Vector<String> payloadBase64;
        WebCore::MockWebAuthenticationConfiguration::HidStage stage;
        WebCore::MockWebAuthenticationConfiguration::HidSubStage subStage;
        WebCore::MockWebAuthenticationConfiguration::HidError error;
        bool isU2f;
        bool keepAlive;
        bool fastDataArrival;
        bool continueAfterErrorData;
        bool canDowngrade;
        bool expectCancel;
        bool supportClientPin;
        bool supportInternalUV;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHidConfiguration) == sizeof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, payloadBase64)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, stage)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, subStage)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, error)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, isU2f)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, keepAlive)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, fastDataArrival)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, continueAfterErrorData)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, canDowngrade)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, expectCancel)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, supportClientPin)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::HidConfiguration, supportInternalUV)
    >::value);
    encoder << instance.payloadBase64;
    encoder << instance.stage;
    encoder << instance.subStage;
    encoder << instance.error;
    encoder << instance.isU2f;
    encoder << instance.keepAlive;
    encoder << instance.fastDataArrival;
    encoder << instance.continueAfterErrorData;
    encoder << instance.canDowngrade;
    encoder << instance.expectCancel;
    encoder << instance.supportClientPin;
    encoder << instance.supportInternalUV;
}

std::optional<WebCore::MockWebAuthenticationConfiguration::HidConfiguration> ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::HidConfiguration>::decode(Decoder& decoder)
{
    auto payloadBase64 = decoder.decode<Vector<String>>();
    auto stage = decoder.decode<WebCore::MockWebAuthenticationConfiguration::HidStage>();
    auto subStage = decoder.decode<WebCore::MockWebAuthenticationConfiguration::HidSubStage>();
    auto error = decoder.decode<WebCore::MockWebAuthenticationConfiguration::HidError>();
    auto isU2f = decoder.decode<bool>();
    auto keepAlive = decoder.decode<bool>();
    auto fastDataArrival = decoder.decode<bool>();
    auto continueAfterErrorData = decoder.decode<bool>();
    auto canDowngrade = decoder.decode<bool>();
    auto expectCancel = decoder.decode<bool>();
    auto supportClientPin = decoder.decode<bool>();
    auto supportInternalUV = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MockWebAuthenticationConfiguration::HidConfiguration {
            WTFMove(*payloadBase64),
            WTFMove(*stage),
            WTFMove(*subStage),
            WTFMove(*error),
            WTFMove(*isU2f),
            WTFMove(*keepAlive),
            WTFMove(*fastDataArrival),
            WTFMove(*continueAfterErrorData),
            WTFMove(*canDowngrade),
            WTFMove(*expectCancel),
            WTFMove(*supportClientPin),
            WTFMove(*supportInternalUV)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration>::encode(Encoder& encoder, const WebCore::MockWebAuthenticationConfiguration::NfcConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.error)>, WebCore::MockWebAuthenticationConfiguration::NfcError>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.payloadBase64)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multipleTags)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multiplePhysicalTags)>, bool>);
    struct ShouldBeSameSizeAsNfcConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration>, false> {
        WebCore::MockWebAuthenticationConfiguration::NfcError error;
        Vector<String> payloadBase64;
        bool multipleTags;
        bool multiplePhysicalTags;
    };
    static_assert(sizeof(ShouldBeSameSizeAsNfcConfiguration) == sizeof(WebCore::MockWebAuthenticationConfiguration::NfcConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MockWebAuthenticationConfiguration::NfcConfiguration, error)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::NfcConfiguration, payloadBase64)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::NfcConfiguration, multipleTags)
        , offsetof(WebCore::MockWebAuthenticationConfiguration::NfcConfiguration, multiplePhysicalTags)
    >::value);
    encoder << instance.error;
    encoder << instance.payloadBase64;
    encoder << instance.multipleTags;
    encoder << instance.multiplePhysicalTags;
}

std::optional<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration> ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration>::decode(Decoder& decoder)
{
    auto error = decoder.decode<WebCore::MockWebAuthenticationConfiguration::NfcError>();
    auto payloadBase64 = decoder.decode<Vector<String>>();
    auto multipleTags = decoder.decode<bool>();
    auto multiplePhysicalTags = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MockWebAuthenticationConfiguration::NfcConfiguration {
            WTFMove(*error),
            WTFMove(*payloadBase64),
            WTFMove(*multipleTags),
            WTFMove(*multiplePhysicalTags)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration>::encode(Encoder& encoder, const WebCore::MockWebAuthenticationConfiguration::CcidConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.payloadBase64)>, Vector<String>>);
    struct ShouldBeSameSizeAsCcidConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration>, false> {
        Vector<String> payloadBase64;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCcidConfiguration) == sizeof(WebCore::MockWebAuthenticationConfiguration::CcidConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MockWebAuthenticationConfiguration::CcidConfiguration, payloadBase64)
    >::value);
    encoder << instance.payloadBase64;
}

std::optional<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration> ArgumentCoder<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration>::decode(Decoder& decoder)
{
    auto payloadBase64 = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MockWebAuthenticationConfiguration::CcidConfiguration {
            WTFMove(*payloadBase64)
        }
    };
}

#endif

#if ENABLE(WEB_AUTHN)
void ArgumentCoder<WebCore::MockWebAuthenticationConfiguration>::encode(Encoder& encoder, const WebCore::MockWebAuthenticationConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.silentFailure)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.local)>, std::optional<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hid)>, std::optional<WebCore::MockWebAuthenticationConfiguration::HidConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.nfc)>, std::optional<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ccid)>, std::optional<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration>>);
    struct ShouldBeSameSizeAsMockWebAuthenticationConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::MockWebAuthenticationConfiguration>, false> {
        bool silentFailure;
        std::optional<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration> local;
        std::optional<WebCore::MockWebAuthenticationConfiguration::HidConfiguration> hid;
        std::optional<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration> nfc;
        std::optional<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration> ccid;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMockWebAuthenticationConfiguration) == sizeof(WebCore::MockWebAuthenticationConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::MockWebAuthenticationConfiguration, silentFailure)
        , offsetof(WebCore::MockWebAuthenticationConfiguration, local)
        , offsetof(WebCore::MockWebAuthenticationConfiguration, hid)
        , offsetof(WebCore::MockWebAuthenticationConfiguration, nfc)
        , offsetof(WebCore::MockWebAuthenticationConfiguration, ccid)
    >::value);
    encoder << instance.silentFailure;
    encoder << instance.local;
    encoder << instance.hid;
    encoder << instance.nfc;
    encoder << instance.ccid;
}

std::optional<WebCore::MockWebAuthenticationConfiguration> ArgumentCoder<WebCore::MockWebAuthenticationConfiguration>::decode(Decoder& decoder)
{
    auto silentFailure = decoder.decode<bool>();
    auto local = decoder.decode<std::optional<WebCore::MockWebAuthenticationConfiguration::LocalConfiguration>>();
    auto hid = decoder.decode<std::optional<WebCore::MockWebAuthenticationConfiguration::HidConfiguration>>();
    auto nfc = decoder.decode<std::optional<WebCore::MockWebAuthenticationConfiguration::NfcConfiguration>>();
    auto ccid = decoder.decode<std::optional<WebCore::MockWebAuthenticationConfiguration::CcidConfiguration>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::MockWebAuthenticationConfiguration {
            WTFMove(*silentFailure),
            WTFMove(*local),
            WTFMove(*hid),
            WTFMove(*nfc),
            WTFMove(*ccid)
        }
    };
}

#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
void ArgumentCoder<WebCore::LengthPoint>::encode(Encoder& encoder, const WebCore::LengthPoint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x())>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y())>, WebCore::Length>);
    encoder << instance.x();
    encoder << instance.y();
}

std::optional<WebCore::LengthPoint> ArgumentCoder<WebCore::LengthPoint>::decode(Decoder& decoder)
{
    auto x = decoder.decode<WebCore::Length>();
    auto y = decoder.decode<WebCore::Length>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::LengthPoint {
            WTFMove(*x),
            WTFMove(*y)
        }
    };
}

#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
void ArgumentCoder<WebCore::OffsetRotation>::encode(Encoder& encoder, const WebCore::OffsetRotation& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasAuto())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.angle())>, float>);
    encoder << instance.hasAuto();
    encoder << instance.angle();
}

std::optional<WebCore::OffsetRotation> ArgumentCoder<WebCore::OffsetRotation>::decode(Decoder& decoder)
{
    auto hasAuto = decoder.decode<bool>();
    auto angle = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::OffsetRotation {
            WTFMove(*hasAuto),
            WTFMove(*angle)
        }
    };
}

#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
void ArgumentCoder<WebCore::AcceleratedEffectValues>::encode(Encoder& encoder, const WebCore::AcceleratedEffectValues& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.opacity)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transformOrigin)>, WebCore::LengthPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.transform)>, WebCore::TransformOperations>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.translate)>, RefPtr<WebCore::TransformOperation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scale)>, RefPtr<WebCore::TransformOperation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rotate)>, RefPtr<WebCore::TransformOperation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offsetPath)>, RefPtr<WebCore::PathOperation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offsetDistance)>, WebCore::Length>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offsetPosition)>, WebCore::LengthPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offsetAnchor)>, WebCore::LengthPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offsetRotate)>, WebCore::OffsetRotation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.filter)>, WebCore::FilterOperations>);
#if ENABLE(FILTERS_LEVEL_2)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.backdropFilter)>, WebCore::FilterOperations>);
#endif
    struct ShouldBeSameSizeAsAcceleratedEffectValues : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AcceleratedEffectValues>, false> {
        float opacity;
        WebCore::LengthPoint transformOrigin;
        WebCore::TransformOperations transform;
        RefPtr<WebCore::TransformOperation> translate;
        RefPtr<WebCore::TransformOperation> scale;
        RefPtr<WebCore::TransformOperation> rotate;
        RefPtr<WebCore::PathOperation> offsetPath;
        WebCore::Length offsetDistance;
        WebCore::LengthPoint offsetPosition;
        WebCore::LengthPoint offsetAnchor;
        WebCore::OffsetRotation offsetRotate;
        WebCore::FilterOperations filter;
#if ENABLE(FILTERS_LEVEL_2)
        WebCore::FilterOperations backdropFilter;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsAcceleratedEffectValues) == sizeof(WebCore::AcceleratedEffectValues));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AcceleratedEffectValues, opacity)
        , offsetof(WebCore::AcceleratedEffectValues, transformOrigin)
        , offsetof(WebCore::AcceleratedEffectValues, transform)
        , offsetof(WebCore::AcceleratedEffectValues, translate)
        , offsetof(WebCore::AcceleratedEffectValues, scale)
        , offsetof(WebCore::AcceleratedEffectValues, rotate)
        , offsetof(WebCore::AcceleratedEffectValues, offsetPath)
        , offsetof(WebCore::AcceleratedEffectValues, offsetDistance)
        , offsetof(WebCore::AcceleratedEffectValues, offsetPosition)
        , offsetof(WebCore::AcceleratedEffectValues, offsetAnchor)
        , offsetof(WebCore::AcceleratedEffectValues, offsetRotate)
        , offsetof(WebCore::AcceleratedEffectValues, filter)
#if ENABLE(FILTERS_LEVEL_2)
        , offsetof(WebCore::AcceleratedEffectValues, backdropFilter)
#endif
    >::value);
    encoder << instance.opacity;
    encoder << instance.transformOrigin;
    encoder << instance.transform;
    encoder << instance.translate;
    encoder << instance.scale;
    encoder << instance.rotate;
    encoder << instance.offsetPath;
    encoder << instance.offsetDistance;
    encoder << instance.offsetPosition;
    encoder << instance.offsetAnchor;
    encoder << instance.offsetRotate;
    encoder << instance.filter;
#if ENABLE(FILTERS_LEVEL_2)
    encoder << instance.backdropFilter;
#endif
}

std::optional<WebCore::AcceleratedEffectValues> ArgumentCoder<WebCore::AcceleratedEffectValues>::decode(Decoder& decoder)
{
    auto opacity = decoder.decode<float>();
    auto transformOrigin = decoder.decode<WebCore::LengthPoint>();
    auto transform = decoder.decode<WebCore::TransformOperations>();
    auto translate = decoder.decode<RefPtr<WebCore::TransformOperation>>();
    auto scale = decoder.decode<RefPtr<WebCore::TransformOperation>>();
    auto rotate = decoder.decode<RefPtr<WebCore::TransformOperation>>();
    auto offsetPath = decoder.decode<RefPtr<WebCore::PathOperation>>();
    auto offsetDistance = decoder.decode<WebCore::Length>();
    auto offsetPosition = decoder.decode<WebCore::LengthPoint>();
    auto offsetAnchor = decoder.decode<WebCore::LengthPoint>();
    auto offsetRotate = decoder.decode<WebCore::OffsetRotation>();
    auto filter = decoder.decode<WebCore::FilterOperations>();
#if ENABLE(FILTERS_LEVEL_2)
    auto backdropFilter = decoder.decode<WebCore::FilterOperations>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AcceleratedEffectValues {
            WTFMove(*opacity),
            WTFMove(*transformOrigin),
            WTFMove(*transform),
            WTFMove(*translate),
            WTFMove(*scale),
            WTFMove(*rotate),
            WTFMove(*offsetPath),
            WTFMove(*offsetDistance),
            WTFMove(*offsetPosition),
            WTFMove(*offsetAnchor),
            WTFMove(*offsetRotate),
            WTFMove(*filter),
#if ENABLE(FILTERS_LEVEL_2)
            WTFMove(*backdropFilter)
#endif
        }
    };
}

#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
void ArgumentCoder<WebCore::AcceleratedEffectKeyframe>::encode(Encoder& encoder, const WebCore::AcceleratedEffectKeyframe& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.values)>, WebCore::AcceleratedEffectValues>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timingFunction)>, RefPtr<WebCore::TimingFunction>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositeOperation)>, std::optional<WebCore::CompositeOperation>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.animatedProperties)>, OptionSet<WebCore::AcceleratedEffectProperty>>);
    struct ShouldBeSameSizeAsAcceleratedEffectKeyframe : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::AcceleratedEffectKeyframe>, false> {
        double offset;
        WebCore::AcceleratedEffectValues values;
        RefPtr<WebCore::TimingFunction> timingFunction;
        std::optional<WebCore::CompositeOperation> compositeOperation;
        OptionSet<WebCore::AcceleratedEffectProperty> animatedProperties;
    };
    static_assert(sizeof(ShouldBeSameSizeAsAcceleratedEffectKeyframe) == sizeof(WebCore::AcceleratedEffectKeyframe));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::AcceleratedEffectKeyframe, offset)
        , offsetof(WebCore::AcceleratedEffectKeyframe, values)
        , offsetof(WebCore::AcceleratedEffectKeyframe, timingFunction)
        , offsetof(WebCore::AcceleratedEffectKeyframe, compositeOperation)
        , offsetof(WebCore::AcceleratedEffectKeyframe, animatedProperties)
    >::value);
    encoder << instance.offset;
    encoder << instance.values;
    encoder << instance.timingFunction;
    encoder << instance.compositeOperation;
    encoder << instance.animatedProperties;
}

std::optional<WebCore::AcceleratedEffectKeyframe> ArgumentCoder<WebCore::AcceleratedEffectKeyframe>::decode(Decoder& decoder)
{
    auto offset = decoder.decode<double>();
    auto values = decoder.decode<WebCore::AcceleratedEffectValues>();
    auto timingFunction = decoder.decode<RefPtr<WebCore::TimingFunction>>();
    auto compositeOperation = decoder.decode<std::optional<WebCore::CompositeOperation>>();
    auto animatedProperties = decoder.decode<OptionSet<WebCore::AcceleratedEffectProperty>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AcceleratedEffectKeyframe {
            WTFMove(*offset),
            WTFMove(*values),
            WTFMove(*timingFunction),
            WTFMove(*compositeOperation),
            WTFMove(*animatedProperties)
        }
    };
}

#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
void ArgumentCoder<WebCore::AcceleratedEffect>::encode(Encoder& encoder, const WebCore::AcceleratedEffect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyframes())>, Vector<WebCore::AcceleratedEffectKeyframe>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.animationType())>, WebCore::WebAnimationType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fill())>, WebCore::FillMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.direction())>, WebCore::PlaybackDirection>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositeOperation())>, WebCore::CompositeOperation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timingFunction())>, RefPtr<WebCore::TimingFunction>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.defaultKeyframeTimingFunction())>, RefPtr<WebCore::TimingFunction>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.animatedProperties())>, OptionSet<WebCore::AcceleratedEffectProperty>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paused())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.iterationStart())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.iterations())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.playbackRate())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.delay())>, WTF::Seconds>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endDelay())>, WTF::Seconds>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.iterationDuration())>, WTF::Seconds>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.activeDuration())>, WTF::Seconds>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.endTime())>, WTF::Seconds>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startTime())>, std::optional<WTF::Seconds>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.holdTime())>, std::optional<WTF::Seconds>>);
    encoder << instance.keyframes();
    encoder << instance.animationType();
    encoder << instance.fill();
    encoder << instance.direction();
    encoder << instance.compositeOperation();
    encoder << instance.timingFunction();
    encoder << instance.defaultKeyframeTimingFunction();
    encoder << instance.animatedProperties();
    encoder << instance.paused();
    encoder << instance.iterationStart();
    encoder << instance.iterations();
    encoder << instance.playbackRate();
    encoder << instance.delay();
    encoder << instance.endDelay();
    encoder << instance.iterationDuration();
    encoder << instance.activeDuration();
    encoder << instance.endTime();
    encoder << instance.startTime();
    encoder << instance.holdTime();
}

std::optional<Ref<WebCore::AcceleratedEffect>> ArgumentCoder<WebCore::AcceleratedEffect>::decode(Decoder& decoder)
{
    auto keyframes = decoder.decode<Vector<WebCore::AcceleratedEffectKeyframe>>();
    auto animationType = decoder.decode<WebCore::WebAnimationType>();
    auto fill = decoder.decode<WebCore::FillMode>();
    auto direction = decoder.decode<WebCore::PlaybackDirection>();
    auto compositeOperation = decoder.decode<WebCore::CompositeOperation>();
    auto timingFunction = decoder.decode<RefPtr<WebCore::TimingFunction>>();
    auto defaultKeyframeTimingFunction = decoder.decode<RefPtr<WebCore::TimingFunction>>();
    auto animatedProperties = decoder.decode<OptionSet<WebCore::AcceleratedEffectProperty>>();
    auto paused = decoder.decode<bool>();
    auto iterationStart = decoder.decode<double>();
    auto iterations = decoder.decode<double>();
    auto playbackRate = decoder.decode<double>();
    auto delay = decoder.decode<WTF::Seconds>();
    auto endDelay = decoder.decode<WTF::Seconds>();
    auto iterationDuration = decoder.decode<WTF::Seconds>();
    auto activeDuration = decoder.decode<WTF::Seconds>();
    auto endTime = decoder.decode<WTF::Seconds>();
    auto startTime = decoder.decode<std::optional<WTF::Seconds>>();
    auto holdTime = decoder.decode<std::optional<WTF::Seconds>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::AcceleratedEffect::create(
            WTFMove(*keyframes),
            WTFMove(*animationType),
            WTFMove(*fill),
            WTFMove(*direction),
            WTFMove(*compositeOperation),
            WTFMove(*timingFunction),
            WTFMove(*defaultKeyframeTimingFunction),
            WTFMove(*animatedProperties),
            WTFMove(*paused),
            WTFMove(*iterationStart),
            WTFMove(*iterations),
            WTFMove(*playbackRate),
            WTFMove(*delay),
            WTFMove(*endDelay),
            WTFMove(*iterationDuration),
            WTFMove(*activeDuration),
            WTFMove(*endTime),
            WTFMove(*startTime),
            WTFMove(*holdTime)
        )
    };
}

#endif

void ArgumentCoder<WebCore::ProtectionSpace>::encode(Encoder& encoder, const WebCore::ProtectionSpace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.host())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.port())>, int>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.serverType())>, WebCore::ProtectionSpace::ServerType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.realm())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authenticationScheme())>, WebCore::ProtectionSpace::AuthenticationScheme>);
#if USE(CURL)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.certificateInfo())>, WebCore::CertificateInfo>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getPlatformDataToSerialize())>, std::optional<WebCore::ProtectionSpace::PlatformData>>);
#endif
    encoder << instance.host();
    encoder << instance.port();
    encoder << instance.serverType();
    encoder << instance.realm();
    encoder << instance.authenticationScheme();
#if USE(CURL)
    encoder << instance.certificateInfo();
#endif
#if PLATFORM(COCOA)
    encoder << instance.getPlatformDataToSerialize();
#endif
}

std::optional<WebCore::ProtectionSpace> ArgumentCoder<WebCore::ProtectionSpace>::decode(Decoder& decoder)
{
    auto host = decoder.decode<String>();
    auto port = decoder.decode<int>();
    auto serverType = decoder.decode<WebCore::ProtectionSpace::ServerType>();
    auto realm = decoder.decode<String>();
    auto authenticationScheme = decoder.decode<WebCore::ProtectionSpace::AuthenticationScheme>();
#if USE(CURL)
    auto certificateInfo = decoder.decode<WebCore::CertificateInfo>();
#endif
#if PLATFORM(COCOA)
    auto getPlatformDataToSerialize = decoder.decode<std::optional<WebCore::ProtectionSpace::PlatformData>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ProtectionSpace {
            WTFMove(*host),
            WTFMove(*port),
            WTFMove(*serverType),
            WTFMove(*realm),
            WTFMove(*authenticationScheme),
#if USE(CURL)
            WTFMove(*certificateInfo),
#endif
#if PLATFORM(COCOA)
            WTFMove(*getPlatformDataToSerialize)
#endif
        }
    };
}

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::ShapeDetection::BarcodeDetectorOptions>::encode(Encoder& encoder, const WebCore::ShapeDetection::BarcodeDetectorOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.formats)>, Vector<WebCore::ShapeDetection::BarcodeFormat>>);
    struct ShouldBeSameSizeAsBarcodeDetectorOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::BarcodeDetectorOptions>, false> {
        Vector<WebCore::ShapeDetection::BarcodeFormat> formats;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBarcodeDetectorOptions) == sizeof(WebCore::ShapeDetection::BarcodeDetectorOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::BarcodeDetectorOptions, formats)
    >::value);
    encoder << instance.formats;
}

void ArgumentCoder<WebCore::ShapeDetection::BarcodeDetectorOptions>::encode(StreamConnectionEncoder& encoder, const WebCore::ShapeDetection::BarcodeDetectorOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.formats)>, Vector<WebCore::ShapeDetection::BarcodeFormat>>);
    struct ShouldBeSameSizeAsBarcodeDetectorOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::BarcodeDetectorOptions>, false> {
        Vector<WebCore::ShapeDetection::BarcodeFormat> formats;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBarcodeDetectorOptions) == sizeof(WebCore::ShapeDetection::BarcodeDetectorOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::BarcodeDetectorOptions, formats)
    >::value);
    encoder << instance.formats;
}

std::optional<WebCore::ShapeDetection::BarcodeDetectorOptions> ArgumentCoder<WebCore::ShapeDetection::BarcodeDetectorOptions>::decode(Decoder& decoder)
{
    auto formats = decoder.decode<Vector<WebCore::ShapeDetection::BarcodeFormat>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShapeDetection::BarcodeDetectorOptions {
            WTFMove(*formats)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::ShapeDetection::DetectedBarcode>::encode(Encoder& encoder, const WebCore::ShapeDetection::DetectedBarcode& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.boundingBox)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rawValue)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::ShapeDetection::BarcodeFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cornerPoints)>, Vector<WebCore::FloatPoint>>);
    struct ShouldBeSameSizeAsDetectedBarcode : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::DetectedBarcode>, false> {
        WebCore::FloatRect boundingBox;
        String rawValue;
        WebCore::ShapeDetection::BarcodeFormat format;
        Vector<WebCore::FloatPoint> cornerPoints;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDetectedBarcode) == sizeof(WebCore::ShapeDetection::DetectedBarcode));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, boundingBox)
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, rawValue)
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, format)
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, cornerPoints)
    >::value);
    encoder << instance.boundingBox;
    encoder << instance.rawValue;
    encoder << instance.format;
    encoder << instance.cornerPoints;
}

void ArgumentCoder<WebCore::ShapeDetection::DetectedBarcode>::encode(StreamConnectionEncoder& encoder, const WebCore::ShapeDetection::DetectedBarcode& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.boundingBox)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rawValue)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::ShapeDetection::BarcodeFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cornerPoints)>, Vector<WebCore::FloatPoint>>);
    struct ShouldBeSameSizeAsDetectedBarcode : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::DetectedBarcode>, false> {
        WebCore::FloatRect boundingBox;
        String rawValue;
        WebCore::ShapeDetection::BarcodeFormat format;
        Vector<WebCore::FloatPoint> cornerPoints;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDetectedBarcode) == sizeof(WebCore::ShapeDetection::DetectedBarcode));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, boundingBox)
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, rawValue)
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, format)
        , offsetof(WebCore::ShapeDetection::DetectedBarcode, cornerPoints)
    >::value);
    encoder << instance.boundingBox;
    encoder << instance.rawValue;
    encoder << instance.format;
    encoder << instance.cornerPoints;
}

std::optional<WebCore::ShapeDetection::DetectedBarcode> ArgumentCoder<WebCore::ShapeDetection::DetectedBarcode>::decode(Decoder& decoder)
{
    auto boundingBox = decoder.decode<WebCore::FloatRect>();
    auto rawValue = decoder.decode<String>();
    auto format = decoder.decode<WebCore::ShapeDetection::BarcodeFormat>();
    auto cornerPoints = decoder.decode<Vector<WebCore::FloatPoint>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShapeDetection::DetectedBarcode {
            WTFMove(*boundingBox),
            WTFMove(*rawValue),
            WTFMove(*format),
            WTFMove(*cornerPoints)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::ShapeDetection::DetectedFace>::encode(Encoder& encoder, const WebCore::ShapeDetection::DetectedFace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.boundingBox)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.landmarks)>, std::optional<Vector<WebCore::ShapeDetection::Landmark>>>);
    struct ShouldBeSameSizeAsDetectedFace : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::DetectedFace>, false> {
        WebCore::FloatRect boundingBox;
        std::optional<Vector<WebCore::ShapeDetection::Landmark>> landmarks;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDetectedFace) == sizeof(WebCore::ShapeDetection::DetectedFace));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::DetectedFace, boundingBox)
        , offsetof(WebCore::ShapeDetection::DetectedFace, landmarks)
    >::value);
    encoder << instance.boundingBox;
    encoder << instance.landmarks;
}

void ArgumentCoder<WebCore::ShapeDetection::DetectedFace>::encode(StreamConnectionEncoder& encoder, const WebCore::ShapeDetection::DetectedFace& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.boundingBox)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.landmarks)>, std::optional<Vector<WebCore::ShapeDetection::Landmark>>>);
    struct ShouldBeSameSizeAsDetectedFace : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::DetectedFace>, false> {
        WebCore::FloatRect boundingBox;
        std::optional<Vector<WebCore::ShapeDetection::Landmark>> landmarks;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDetectedFace) == sizeof(WebCore::ShapeDetection::DetectedFace));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::DetectedFace, boundingBox)
        , offsetof(WebCore::ShapeDetection::DetectedFace, landmarks)
    >::value);
    encoder << instance.boundingBox;
    encoder << instance.landmarks;
}

std::optional<WebCore::ShapeDetection::DetectedFace> ArgumentCoder<WebCore::ShapeDetection::DetectedFace>::decode(Decoder& decoder)
{
    auto boundingBox = decoder.decode<WebCore::FloatRect>();
    auto landmarks = decoder.decode<std::optional<Vector<WebCore::ShapeDetection::Landmark>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShapeDetection::DetectedFace {
            WTFMove(*boundingBox),
            WTFMove(*landmarks)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::ShapeDetection::DetectedText>::encode(Encoder& encoder, const WebCore::ShapeDetection::DetectedText& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.boundingBox)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rawValue)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cornerPoints)>, Vector<WebCore::FloatPoint>>);
    struct ShouldBeSameSizeAsDetectedText : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::DetectedText>, false> {
        WebCore::FloatRect boundingBox;
        String rawValue;
        Vector<WebCore::FloatPoint> cornerPoints;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDetectedText) == sizeof(WebCore::ShapeDetection::DetectedText));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::DetectedText, boundingBox)
        , offsetof(WebCore::ShapeDetection::DetectedText, rawValue)
        , offsetof(WebCore::ShapeDetection::DetectedText, cornerPoints)
    >::value);
    encoder << instance.boundingBox;
    encoder << instance.rawValue;
    encoder << instance.cornerPoints;
}

void ArgumentCoder<WebCore::ShapeDetection::DetectedText>::encode(StreamConnectionEncoder& encoder, const WebCore::ShapeDetection::DetectedText& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.boundingBox)>, WebCore::FloatRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rawValue)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cornerPoints)>, Vector<WebCore::FloatPoint>>);
    struct ShouldBeSameSizeAsDetectedText : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::DetectedText>, false> {
        WebCore::FloatRect boundingBox;
        String rawValue;
        Vector<WebCore::FloatPoint> cornerPoints;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDetectedText) == sizeof(WebCore::ShapeDetection::DetectedText));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::DetectedText, boundingBox)
        , offsetof(WebCore::ShapeDetection::DetectedText, rawValue)
        , offsetof(WebCore::ShapeDetection::DetectedText, cornerPoints)
    >::value);
    encoder << instance.boundingBox;
    encoder << instance.rawValue;
    encoder << instance.cornerPoints;
}

std::optional<WebCore::ShapeDetection::DetectedText> ArgumentCoder<WebCore::ShapeDetection::DetectedText>::decode(Decoder& decoder)
{
    auto boundingBox = decoder.decode<WebCore::FloatRect>();
    auto rawValue = decoder.decode<String>();
    auto cornerPoints = decoder.decode<Vector<WebCore::FloatPoint>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShapeDetection::DetectedText {
            WTFMove(*boundingBox),
            WTFMove(*rawValue),
            WTFMove(*cornerPoints)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::ShapeDetection::FaceDetectorOptions>::encode(Encoder& encoder, const WebCore::ShapeDetection::FaceDetectorOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxDetectedFaces)>, uint16_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fastMode)>, bool>);
    struct ShouldBeSameSizeAsFaceDetectorOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::FaceDetectorOptions>, false> {
        uint16_t maxDetectedFaces;
        bool fastMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFaceDetectorOptions) == sizeof(WebCore::ShapeDetection::FaceDetectorOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::FaceDetectorOptions, maxDetectedFaces)
        , offsetof(WebCore::ShapeDetection::FaceDetectorOptions, fastMode)
    >::value);
    encoder << instance.maxDetectedFaces;
    encoder << instance.fastMode;
}

void ArgumentCoder<WebCore::ShapeDetection::FaceDetectorOptions>::encode(StreamConnectionEncoder& encoder, const WebCore::ShapeDetection::FaceDetectorOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxDetectedFaces)>, uint16_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fastMode)>, bool>);
    struct ShouldBeSameSizeAsFaceDetectorOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::FaceDetectorOptions>, false> {
        uint16_t maxDetectedFaces;
        bool fastMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsFaceDetectorOptions) == sizeof(WebCore::ShapeDetection::FaceDetectorOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::FaceDetectorOptions, maxDetectedFaces)
        , offsetof(WebCore::ShapeDetection::FaceDetectorOptions, fastMode)
    >::value);
    encoder << instance.maxDetectedFaces;
    encoder << instance.fastMode;
}

std::optional<WebCore::ShapeDetection::FaceDetectorOptions> ArgumentCoder<WebCore::ShapeDetection::FaceDetectorOptions>::decode(Decoder& decoder)
{
    auto maxDetectedFaces = decoder.decode<uint16_t>();
    auto fastMode = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShapeDetection::FaceDetectorOptions {
            WTFMove(*maxDetectedFaces),
            WTFMove(*fastMode)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::ShapeDetection::Landmark>::encode(Encoder& encoder, const WebCore::ShapeDetection::Landmark& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.locations)>, Vector<WebCore::FloatPoint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ShapeDetection::LandmarkType>);
    struct ShouldBeSameSizeAsLandmark : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::Landmark>, false> {
        Vector<WebCore::FloatPoint> locations;
        WebCore::ShapeDetection::LandmarkType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLandmark) == sizeof(WebCore::ShapeDetection::Landmark));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::Landmark, locations)
        , offsetof(WebCore::ShapeDetection::Landmark, type)
    >::value);
    encoder << instance.locations;
    encoder << instance.type;
}

void ArgumentCoder<WebCore::ShapeDetection::Landmark>::encode(StreamConnectionEncoder& encoder, const WebCore::ShapeDetection::Landmark& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.locations)>, Vector<WebCore::FloatPoint>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::ShapeDetection::LandmarkType>);
    struct ShouldBeSameSizeAsLandmark : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::ShapeDetection::Landmark>, false> {
        Vector<WebCore::FloatPoint> locations;
        WebCore::ShapeDetection::LandmarkType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsLandmark) == sizeof(WebCore::ShapeDetection::Landmark));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::ShapeDetection::Landmark, locations)
        , offsetof(WebCore::ShapeDetection::Landmark, type)
    >::value);
    encoder << instance.locations;
    encoder << instance.type;
}

std::optional<WebCore::ShapeDetection::Landmark> ArgumentCoder<WebCore::ShapeDetection::Landmark>::decode(Decoder& decoder)
{
    auto locations = decoder.decode<Vector<WebCore::FloatPoint>>();
    auto type = decoder.decode<WebCore::ShapeDetection::LandmarkType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::ShapeDetection::Landmark {
            WTFMove(*locations),
            WTFMove(*type)
        }
    };
}

#endif

#if ENABLE(MEDIA_STREAM)
void ArgumentCoder<WebCore::CaptureSourceError>::encode(Encoder& encoder, const WebCore::CaptureSourceError& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.errorMessage)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.denialReason)>, WebCore::MediaAccessDenialReason>);
    struct ShouldBeSameSizeAsCaptureSourceError : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CaptureSourceError>, false> {
        String errorMessage;
        WebCore::MediaAccessDenialReason denialReason;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCaptureSourceError) == sizeof(WebCore::CaptureSourceError));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CaptureSourceError, errorMessage)
        , offsetof(WebCore::CaptureSourceError, denialReason)
    >::value);
    encoder << instance.errorMessage;
    encoder << instance.denialReason;
}

std::optional<WebCore::CaptureSourceError> ArgumentCoder<WebCore::CaptureSourceError>::decode(Decoder& decoder)
{
    auto errorMessage = decoder.decode<String>();
    auto denialReason = decoder.decode<WebCore::MediaAccessDenialReason>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CaptureSourceError {
            WTFMove(*errorMessage),
            WTFMove(*denialReason)
        }
    };
}

#endif

void ArgumentCoder<WebCore::CookieStoreGetOptions>::encode(Encoder& encoder, const WebCore::CookieStoreGetOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url)>, String>);
    struct ShouldBeSameSizeAsCookieStoreGetOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::CookieStoreGetOptions>, false> {
        String name;
        String url;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCookieStoreGetOptions) == sizeof(WebCore::CookieStoreGetOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::CookieStoreGetOptions, name)
        , offsetof(WebCore::CookieStoreGetOptions, url)
    >::value);
    encoder << instance.name;
    encoder << instance.url;
}

std::optional<WebCore::CookieStoreGetOptions> ArgumentCoder<WebCore::CookieStoreGetOptions>::decode(Decoder& decoder)
{
    auto name = decoder.decode<String>();
    auto url = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::CookieStoreGetOptions {
            WTFMove(*name),
            WTFMove(*url)
        }
    };
}

void ArgumentCoder<WebCore::Model>::encode(Encoder& encoder, const WebCore::Model& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.data())>, Ref<WebCore::SharedBuffer>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mimeType())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.url())>, URL>);
    encoder << instance.data();
    encoder << instance.mimeType();
    encoder << instance.url();
}

std::optional<Ref<WebCore::Model>> ArgumentCoder<WebCore::Model>::decode(Decoder& decoder)
{
    auto data = decoder.decode<Ref<WebCore::SharedBuffer>>();
    auto mimeType = decoder.decode<String>();
    auto url = decoder.decode<URL>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::Model::create(
            WTFMove(*data),
            WTFMove(*mimeType),
            WTFMove(*url)
        )
    };
}

void ArgumentCoder<WebKit::WebEvent>::encode(Encoder& encoder, const WebKit::WebEvent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebKit::WebEventType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiers())>, OptionSet<WebKit::WebEventModifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authorizationToken())>, WTF::UUID>);
    encoder << instance.type();
    encoder << instance.modifiers();
    encoder << instance.timestamp();
    encoder << instance.authorizationToken();
}

std::optional<WebKit::WebEvent> ArgumentCoder<WebKit::WebEvent>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebKit::WebEventType>();
    auto modifiers = decoder.decode<OptionSet<WebKit::WebEventModifier>>();
    auto timestamp = decoder.decode<WallTime>();
    auto authorizationToken = decoder.decode<WTF::UUID>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebEvent {
            WTFMove(*type),
            WTFMove(*modifiers),
            WTFMove(*timestamp),
            WTFMove(*authorizationToken)
        }
    };
}

void ArgumentCoder<WebKit::WebKeyboardEvent>::encode(Encoder& encoder, const WebKit::WebKeyboardEvent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebKit::WebEventType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiers())>, OptionSet<WebKit::WebEventModifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authorizationToken())>, WTF::UUID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.text())>, String>);
#if !PLATFORM(GTK) && !USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.unmodifiedText())>, String>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.key())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.code())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyIdentifier())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.windowsVirtualKeyCode())>, int32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.nativeVirtualKeyCode())>, int32_t>);
#if !PLATFORM(GTK) && !USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.macCharCode())>, int32_t>);
#endif
#if USE(APPKIT) || PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.handledByInputMethod())>, bool>);
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preeditUnderlines())>, std::optional<Vector<WebCore::CompositionUnderline>>>);
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preeditSelectionRange())>, std::optional<WebKit::EditingRange>>);
#endif
#if USE(APPKIT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.commands())>, Vector<WebCore::KeypressCommand>>);
#endif
#if !USE(APPKIT) && PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.commands())>, Vector<String>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isAutoRepeat())>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isKeypad())>, bool>);
#if !PLATFORM(GTK) && !USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSystemKey())>, bool>);
#endif
    encoder << instance.type();
    encoder << instance.modifiers();
    encoder << instance.timestamp();
    encoder << instance.authorizationToken();
    encoder << instance.text();
#if !PLATFORM(GTK) && !USE(LIBWPE)
    encoder << instance.unmodifiedText();
#endif
    encoder << instance.key();
    encoder << instance.code();
    encoder << instance.keyIdentifier();
    encoder << instance.windowsVirtualKeyCode();
    encoder << instance.nativeVirtualKeyCode();
#if !PLATFORM(GTK) && !USE(LIBWPE)
    encoder << instance.macCharCode();
#endif
#if USE(APPKIT) || PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || USE(LIBWPE)
    encoder << instance.handledByInputMethod();
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
    encoder << instance.preeditUnderlines();
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
    encoder << instance.preeditSelectionRange();
#endif
#if USE(APPKIT)
    encoder << instance.commands();
#endif
#if !USE(APPKIT) && PLATFORM(GTK)
    encoder << instance.commands();
#endif
    encoder << instance.isAutoRepeat();
    encoder << instance.isKeypad();
#if !PLATFORM(GTK) && !USE(LIBWPE)
    encoder << instance.isSystemKey();
#endif
}

std::optional<WebKit::WebKeyboardEvent> ArgumentCoder<WebKit::WebKeyboardEvent>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebKit::WebEventType>();
    auto modifiers = decoder.decode<OptionSet<WebKit::WebEventModifier>>();
    auto timestamp = decoder.decode<WallTime>();
    auto authorizationToken = decoder.decode<WTF::UUID>();
    auto text = decoder.decode<String>();
#if !PLATFORM(GTK) && !USE(LIBWPE)
    auto unmodifiedText = decoder.decode<String>();
#endif
    auto key = decoder.decode<String>();
    auto code = decoder.decode<String>();
    auto keyIdentifier = decoder.decode<String>();
    auto windowsVirtualKeyCode = decoder.decode<int32_t>();
    auto nativeVirtualKeyCode = decoder.decode<int32_t>();
#if !PLATFORM(GTK) && !USE(LIBWPE)
    auto macCharCode = decoder.decode<int32_t>();
#endif
#if USE(APPKIT) || PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || USE(LIBWPE)
    auto handledByInputMethod = decoder.decode<bool>();
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
    auto preeditUnderlines = decoder.decode<std::optional<Vector<WebCore::CompositionUnderline>>>();
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
    auto preeditSelectionRange = decoder.decode<std::optional<WebKit::EditingRange>>();
#endif
#if USE(APPKIT)
    auto commands = decoder.decode<Vector<WebCore::KeypressCommand>>();
#endif
#if !USE(APPKIT) && PLATFORM(GTK)
    auto commands = decoder.decode<Vector<String>>();
#endif
    auto isAutoRepeat = decoder.decode<bool>();
    auto isKeypad = decoder.decode<bool>();
#if !PLATFORM(GTK) && !USE(LIBWPE)
    auto isSystemKey = decoder.decode<bool>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebKeyboardEvent {
            WebKit::WebEvent {
                WTFMove(*type),
                WTFMove(*modifiers),
                WTFMove(*timestamp),
                WTFMove(*authorizationToken)
            },
            WTFMove(*text),
#if !PLATFORM(GTK) && !USE(LIBWPE)
            WTFMove(*unmodifiedText),
#endif
            WTFMove(*key),
            WTFMove(*code),
            WTFMove(*keyIdentifier),
            WTFMove(*windowsVirtualKeyCode),
            WTFMove(*nativeVirtualKeyCode),
#if !PLATFORM(GTK) && !USE(LIBWPE)
            WTFMove(*macCharCode),
#endif
#if USE(APPKIT) || PLATFORM(IOS_FAMILY) || PLATFORM(GTK) || USE(LIBWPE)
            WTFMove(*handledByInputMethod),
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
            WTFMove(*preeditUnderlines),
#endif
#if PLATFORM(GTK) || USE(LIBWPE)
            WTFMove(*preeditSelectionRange),
#endif
#if USE(APPKIT)
            WTFMove(*commands),
#endif
#if !USE(APPKIT) && PLATFORM(GTK)
            WTFMove(*commands),
#endif
            WTFMove(*isAutoRepeat),
            WTFMove(*isKeypad),
#if !PLATFORM(GTK) && !USE(LIBWPE)
            WTFMove(*isSystemKey)
#endif
        }
    };
}

#if ENABLE(TOUCH_EVENTS)
void ArgumentCoder<WebKit::WebTouchEvent>::encode(Encoder& encoder, const WebKit::WebTouchEvent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebKit::WebEventType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiers())>, OptionSet<WebKit::WebEventModifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authorizationToken())>, WTF::UUID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.touchPoints())>, Vector<WebKit::WebPlatformTouchPoint>>);
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::IntPoint>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isPotentialTap())>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isGesture())>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gestureScale())>, float>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gestureRotation())>, float>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canPreventNativeGestures())>, bool>);
#endif
    encoder << instance.type();
    encoder << instance.modifiers();
    encoder << instance.timestamp();
    encoder << instance.authorizationToken();
    encoder << instance.touchPoints();
#if PLATFORM(IOS_FAMILY)
    encoder << instance.position();
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.isPotentialTap();
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.isGesture();
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.gestureScale();
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.gestureRotation();
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.canPreventNativeGestures();
#endif
}

std::optional<WebKit::WebTouchEvent> ArgumentCoder<WebKit::WebTouchEvent>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebKit::WebEventType>();
    auto modifiers = decoder.decode<OptionSet<WebKit::WebEventModifier>>();
    auto timestamp = decoder.decode<WallTime>();
    auto authorizationToken = decoder.decode<WTF::UUID>();
    auto touchPoints = decoder.decode<Vector<WebKit::WebPlatformTouchPoint>>();
#if PLATFORM(IOS_FAMILY)
    auto position = decoder.decode<WebCore::IntPoint>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto isPotentialTap = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto isGesture = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto gestureScale = decoder.decode<float>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto gestureRotation = decoder.decode<float>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto canPreventNativeGestures = decoder.decode<bool>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebTouchEvent {
            WebKit::WebEvent {
                WTFMove(*type),
                WTFMove(*modifiers),
                WTFMove(*timestamp),
                WTFMove(*authorizationToken)
            },
            WTFMove(*touchPoints),
#if PLATFORM(IOS_FAMILY)
            WTFMove(*position),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*isPotentialTap),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*isGesture),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*gestureScale),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*gestureRotation),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*canPreventNativeGestures)
#endif
        }
    };
}

#endif

#if ENABLE(TOUCH_EVENTS)
void ArgumentCoder<WebKit::WebPlatformTouchPoint>::encode(Encoder& encoder, const WebKit::WebPlatformTouchPoint& instance)
{
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier())>, unsigned>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location())>, WebCore::IntPoint>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.phase())>, WebKit::WebPlatformTouchPoint::State>);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusX())>, double>);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radiusY())>, double>);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rotationAngle())>, double>);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.force())>, double>);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.altitudeAngle())>, double>);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.azimuthAngle())>, double>);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.touchType())>, WebKit::WebPlatformTouchPoint::TouchType>);
#endif
#if !PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id())>, uint32_t>);
#endif
#if !PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.state())>, WebKit::WebPlatformTouchPoint::State>);
#endif
#if !PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenPosition())>, WebCore::IntPoint>);
#endif
#if !PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::IntPoint>);
#endif
#if !PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.radius())>, WebCore::IntSize>);
#endif
#if !PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rotationAngle())>, float>);
#endif
#if !PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.force())>, float>);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.identifier();
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.location();
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.phase();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    encoder << instance.radiusX();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    encoder << instance.radiusY();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    encoder << instance.rotationAngle();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    encoder << instance.force();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    encoder << instance.altitudeAngle();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    encoder << instance.azimuthAngle();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    encoder << instance.touchType();
#endif
#if !PLATFORM(IOS_FAMILY)
    encoder << instance.id();
#endif
#if !PLATFORM(IOS_FAMILY)
    encoder << instance.state();
#endif
#if !PLATFORM(IOS_FAMILY)
    encoder << instance.screenPosition();
#endif
#if !PLATFORM(IOS_FAMILY)
    encoder << instance.position();
#endif
#if !PLATFORM(IOS_FAMILY)
    encoder << instance.radius();
#endif
#if !PLATFORM(IOS_FAMILY)
    encoder << instance.rotationAngle();
#endif
#if !PLATFORM(IOS_FAMILY)
    encoder << instance.force();
#endif
}

std::optional<WebKit::WebPlatformTouchPoint> ArgumentCoder<WebKit::WebPlatformTouchPoint>::decode(Decoder& decoder)
{
#if PLATFORM(IOS_FAMILY)
    auto identifier = decoder.decode<unsigned>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto location = decoder.decode<WebCore::IntPoint>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto phase = decoder.decode<WebKit::WebPlatformTouchPoint::State>();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    auto radiusX = decoder.decode<double>();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    auto radiusY = decoder.decode<double>();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    auto rotationAngle = decoder.decode<double>();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    auto force = decoder.decode<double>();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    auto altitudeAngle = decoder.decode<double>();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    auto azimuthAngle = decoder.decode<double>();
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
    auto touchType = decoder.decode<WebKit::WebPlatformTouchPoint::TouchType>();
#endif
#if !PLATFORM(IOS_FAMILY)
    auto id = decoder.decode<uint32_t>();
#endif
#if !PLATFORM(IOS_FAMILY)
    auto state = decoder.decode<WebKit::WebPlatformTouchPoint::State>();
#endif
#if !PLATFORM(IOS_FAMILY)
    auto screenPosition = decoder.decode<WebCore::IntPoint>();
#endif
#if !PLATFORM(IOS_FAMILY)
    auto position = decoder.decode<WebCore::IntPoint>();
#endif
#if !PLATFORM(IOS_FAMILY)
    auto radius = decoder.decode<WebCore::IntSize>();
#endif
#if !PLATFORM(IOS_FAMILY)
    auto rotationAngle = decoder.decode<float>();
#endif
#if !PLATFORM(IOS_FAMILY)
    auto force = decoder.decode<float>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebPlatformTouchPoint {
#if PLATFORM(IOS_FAMILY)
            WTFMove(*identifier),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*location),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*phase),
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
            WTFMove(*radiusX),
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
            WTFMove(*radiusY),
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
            WTFMove(*rotationAngle),
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
            WTFMove(*force),
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
            WTFMove(*altitudeAngle),
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
            WTFMove(*azimuthAngle),
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(IOS_TOUCH_EVENTS)
            WTFMove(*touchType),
#endif
#if !PLATFORM(IOS_FAMILY)
            WTFMove(*id),
#endif
#if !PLATFORM(IOS_FAMILY)
            WTFMove(*state),
#endif
#if !PLATFORM(IOS_FAMILY)
            WTFMove(*screenPosition),
#endif
#if !PLATFORM(IOS_FAMILY)
            WTFMove(*position),
#endif
#if !PLATFORM(IOS_FAMILY)
            WTFMove(*radius),
#endif
#if !PLATFORM(IOS_FAMILY)
            WTFMove(*rotationAngle),
#endif
#if !PLATFORM(IOS_FAMILY)
            WTFMove(*force)
#endif
        }
    };
}

#endif

void ArgumentCoder<WebKit::WebMouseEvent>::encode(Encoder& encoder, const WebKit::WebMouseEvent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebKit::WebEventType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiers())>, OptionSet<WebKit::WebEventModifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authorizationToken())>, WTF::UUID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.button())>, WebKit::WebMouseEventButton>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buttons())>, unsigned short>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.globalPosition())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deltaX())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deltaY())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deltaZ())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clickCount())>, int32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.force())>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.syntheticClickType())>, WebKit::WebMouseEventSyntheticClickType>);
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.eventNumber())>, int32_t>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.menuTypeForEvent())>, int32_t>);
#endif
#if !PLATFORM(MAC) && PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isTouchEvent())>, WebCore::PlatformMouseEvent::IsTouch>);
#endif
#if !PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pointerId())>, WebCore::PointerID>);
#endif
#if !PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pointerType())>, String>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gestureWasCancelled())>, WebKit::GestureWasCancelled>);
    encoder << instance.type();
    encoder << instance.modifiers();
    encoder << instance.timestamp();
    encoder << instance.authorizationToken();
    encoder << instance.button();
    encoder << instance.buttons();
    encoder << instance.position();
    encoder << instance.globalPosition();
    encoder << instance.deltaX();
    encoder << instance.deltaY();
    encoder << instance.deltaZ();
    encoder << instance.clickCount();
    encoder << instance.force();
    encoder << instance.syntheticClickType();
#if PLATFORM(MAC)
    encoder << instance.eventNumber();
#endif
#if PLATFORM(MAC)
    encoder << instance.menuTypeForEvent();
#endif
#if !PLATFORM(MAC) && PLATFORM(GTK)
    encoder << instance.isTouchEvent();
#endif
#if !PLATFORM(MAC)
    encoder << instance.pointerId();
#endif
#if !PLATFORM(MAC)
    encoder << instance.pointerType();
#endif
    encoder << instance.gestureWasCancelled();
}

std::optional<WebKit::WebMouseEvent> ArgumentCoder<WebKit::WebMouseEvent>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebKit::WebEventType>();
    auto modifiers = decoder.decode<OptionSet<WebKit::WebEventModifier>>();
    auto timestamp = decoder.decode<WallTime>();
    auto authorizationToken = decoder.decode<WTF::UUID>();
    auto button = decoder.decode<WebKit::WebMouseEventButton>();
    auto buttons = decoder.decode<unsigned short>();
    auto position = decoder.decode<WebCore::IntPoint>();
    auto globalPosition = decoder.decode<WebCore::IntPoint>();
    auto deltaX = decoder.decode<float>();
    auto deltaY = decoder.decode<float>();
    auto deltaZ = decoder.decode<float>();
    auto clickCount = decoder.decode<int32_t>();
    auto force = decoder.decode<double>();
    auto syntheticClickType = decoder.decode<WebKit::WebMouseEventSyntheticClickType>();
#if PLATFORM(MAC)
    auto eventNumber = decoder.decode<int32_t>();
#endif
#if PLATFORM(MAC)
    auto menuTypeForEvent = decoder.decode<int32_t>();
#endif
#if !PLATFORM(MAC) && PLATFORM(GTK)
    auto isTouchEvent = decoder.decode<WebCore::PlatformMouseEvent::IsTouch>();
#endif
#if !PLATFORM(MAC)
    auto pointerId = decoder.decode<WebCore::PointerID>();
#endif
#if !PLATFORM(MAC)
    auto pointerType = decoder.decode<String>();
#endif
    auto gestureWasCancelled = decoder.decode<WebKit::GestureWasCancelled>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebMouseEvent {
            WebKit::WebEvent {
                WTFMove(*type),
                WTFMove(*modifiers),
                WTFMove(*timestamp),
                WTFMove(*authorizationToken)
            },
            WTFMove(*button),
            WTFMove(*buttons),
            WTFMove(*position),
            WTFMove(*globalPosition),
            WTFMove(*deltaX),
            WTFMove(*deltaY),
            WTFMove(*deltaZ),
            WTFMove(*clickCount),
            WTFMove(*force),
            WTFMove(*syntheticClickType),
#if PLATFORM(MAC)
            WTFMove(*eventNumber),
#endif
#if PLATFORM(MAC)
            WTFMove(*menuTypeForEvent),
#endif
#if !PLATFORM(MAC) && PLATFORM(GTK)
            WTFMove(*isTouchEvent),
#endif
#if !PLATFORM(MAC)
            WTFMove(*pointerId),
#endif
#if !PLATFORM(MAC)
            WTFMove(*pointerType),
#endif
            WTFMove(*gestureWasCancelled)
        }
    };
}

#if ENABLE(MAC_GESTURE_EVENTS)
void ArgumentCoder<WebKit::WebGestureEvent>::encode(Encoder& encoder, const WebKit::WebGestureEvent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebKit::WebEventType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiers())>, OptionSet<WebKit::WebEventModifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authorizationToken())>, WTF::UUID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gestureScale())>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gestureRotation())>, float>);
    encoder << instance.type();
    encoder << instance.modifiers();
    encoder << instance.timestamp();
    encoder << instance.authorizationToken();
    encoder << instance.position();
    encoder << instance.gestureScale();
    encoder << instance.gestureRotation();
}

std::optional<WebKit::WebGestureEvent> ArgumentCoder<WebKit::WebGestureEvent>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebKit::WebEventType>();
    auto modifiers = decoder.decode<OptionSet<WebKit::WebEventModifier>>();
    auto timestamp = decoder.decode<WallTime>();
    auto authorizationToken = decoder.decode<WTF::UUID>();
    auto position = decoder.decode<WebCore::IntPoint>();
    auto gestureScale = decoder.decode<float>();
    auto gestureRotation = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGestureEvent {
            WebKit::WebEvent {
                WTFMove(*type),
                WTFMove(*modifiers),
                WTFMove(*timestamp),
                WTFMove(*authorizationToken)
            },
            WTFMove(*position),
            WTFMove(*gestureScale),
            WTFMove(*gestureRotation)
        }
    };
}

#endif

void ArgumentCoder<WebKit::WebWheelEvent>::encode(Encoder& encoder, const WebKit::WebWheelEvent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type())>, WebKit::WebEventType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modifiers())>, OptionSet<WebKit::WebEventModifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestamp())>, WallTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.authorizationToken())>, WTF::UUID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.position())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.globalPosition())>, WebCore::IntPoint>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.delta())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wheelTicks())>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.granularity())>, WebKit::WebWheelEvent::Granularity>);
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.directionInvertedFromDevice())>, bool>);
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.phase())>, WebKit::WebWheelEvent::Phase>);
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.momentumPhase())>, WebKit::WebWheelEvent::Phase>);
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasPreciseScrollingDeltas())>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollCount())>, uint32_t>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.unacceleratedScrollingDelta())>, WebCore::FloatSize>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ioHIDEventTimestamp())>, WallTime>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rawPlatformDelta())>, std::optional<WebCore::FloatSize>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.momentumEndType())>, WebKit::WebWheelEvent::MomentumEndType>);
#endif
    encoder << instance.type();
    encoder << instance.modifiers();
    encoder << instance.timestamp();
    encoder << instance.authorizationToken();
    encoder << instance.position();
    encoder << instance.globalPosition();
    encoder << instance.delta();
    encoder << instance.wheelTicks();
    encoder << instance.granularity();
#if PLATFORM(COCOA)
    encoder << instance.directionInvertedFromDevice();
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    encoder << instance.phase();
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    encoder << instance.momentumPhase();
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    encoder << instance.hasPreciseScrollingDeltas();
#endif
#if PLATFORM(COCOA)
    encoder << instance.scrollCount();
#endif
#if PLATFORM(COCOA)
    encoder << instance.unacceleratedScrollingDelta();
#endif
#if PLATFORM(COCOA)
    encoder << instance.ioHIDEventTimestamp();
#endif
#if PLATFORM(COCOA)
    encoder << instance.rawPlatformDelta();
#endif
#if PLATFORM(COCOA)
    encoder << instance.momentumEndType();
#endif
}

std::optional<WebKit::WebWheelEvent> ArgumentCoder<WebKit::WebWheelEvent>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebKit::WebEventType>();
    auto modifiers = decoder.decode<OptionSet<WebKit::WebEventModifier>>();
    auto timestamp = decoder.decode<WallTime>();
    auto authorizationToken = decoder.decode<WTF::UUID>();
    auto position = decoder.decode<WebCore::IntPoint>();
    auto globalPosition = decoder.decode<WebCore::IntPoint>();
    auto delta = decoder.decode<WebCore::FloatSize>();
    auto wheelTicks = decoder.decode<WebCore::FloatSize>();
    auto granularity = decoder.decode<WebKit::WebWheelEvent::Granularity>();
#if PLATFORM(COCOA)
    auto directionInvertedFromDevice = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    auto phase = decoder.decode<WebKit::WebWheelEvent::Phase>();
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    auto momentumPhase = decoder.decode<WebKit::WebWheelEvent::Phase>();
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
    auto hasPreciseScrollingDeltas = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto scrollCount = decoder.decode<uint32_t>();
#endif
#if PLATFORM(COCOA)
    auto unacceleratedScrollingDelta = decoder.decode<WebCore::FloatSize>();
#endif
#if PLATFORM(COCOA)
    auto ioHIDEventTimestamp = decoder.decode<WallTime>();
#endif
#if PLATFORM(COCOA)
    auto rawPlatformDelta = decoder.decode<std::optional<WebCore::FloatSize>>();
#endif
#if PLATFORM(COCOA)
    auto momentumEndType = decoder.decode<WebKit::WebWheelEvent::MomentumEndType>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebWheelEvent {
            WebKit::WebEvent {
                WTFMove(*type),
                WTFMove(*modifiers),
                WTFMove(*timestamp),
                WTFMove(*authorizationToken)
            },
            WTFMove(*position),
            WTFMove(*globalPosition),
            WTFMove(*delta),
            WTFMove(*wheelTicks),
            WTFMove(*granularity),
#if PLATFORM(COCOA)
            WTFMove(*directionInvertedFromDevice),
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
            WTFMove(*phase),
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
            WTFMove(*momentumPhase),
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK) || USE(LIBWPE)
            WTFMove(*hasPreciseScrollingDeltas),
#endif
#if PLATFORM(COCOA)
            WTFMove(*scrollCount),
#endif
#if PLATFORM(COCOA)
            WTFMove(*unacceleratedScrollingDelta),
#endif
#if PLATFORM(COCOA)
            WTFMove(*ioHIDEventTimestamp),
#endif
#if PLATFORM(COCOA)
            WTFMove(*rawPlatformDelta),
#endif
#if PLATFORM(COCOA)
            WTFMove(*momentumEndType)
#endif
        }
    };
}

void ArgumentCoder<WebKit::WebHitTestResultData>::encode(Encoder& encoder, const WebKit::WebHitTestResultData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.absoluteImageURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.absolutePDFURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.absoluteLinkURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.absoluteMediaURL)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linkLabel)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linkTitle)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linkSuggestedFilename)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isContentEditable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.elementBoundingBox)>, WebCore::IntRect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isScrollbar)>, WebKit::WebHitTestResultData::IsScrollbar>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isSelected)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isTextNode)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isOverTextInsideFormControlElement)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isDownloadableMedia)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lookupText)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.toolTipText)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.imageText)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.getImageSharedMemoryHandle())>, std::optional<WebKit::SharedMemoryHandle>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.imageBitmap)>, RefPtr<WebKit::ShareableBitmap>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sourceImageMIMEType)>, String>);
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.platformData)>, WebKit::WebHitTestResultPlatformData>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dictionaryPopupInfo)>, WebCore::DictionaryPopupInfo>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linkTextIndicator)>, RefPtr<WebCore::TextIndicator>>);
    encoder << instance.absoluteImageURL;
    encoder << instance.absolutePDFURL;
    encoder << instance.absoluteLinkURL;
    encoder << instance.absoluteMediaURL;
    encoder << instance.linkLabel;
    encoder << instance.linkTitle;
    encoder << instance.linkSuggestedFilename;
    encoder << instance.isContentEditable;
    encoder << instance.elementBoundingBox;
    encoder << instance.isScrollbar;
    encoder << instance.isSelected;
    encoder << instance.isTextNode;
    encoder << instance.isOverTextInsideFormControlElement;
    encoder << instance.isDownloadableMedia;
    encoder << instance.lookupText;
    encoder << instance.toolTipText;
    encoder << instance.imageText;
    encoder << instance.getImageSharedMemoryHandle();
    encoder << instance.imageBitmap;
    encoder << instance.sourceImageMIMEType;
#if PLATFORM(MAC)
    encoder << instance.platformData;
#endif
    encoder << instance.dictionaryPopupInfo;
    encoder << instance.linkTextIndicator;
}

std::optional<WebKit::WebHitTestResultData> ArgumentCoder<WebKit::WebHitTestResultData>::decode(Decoder& decoder)
{
    auto absoluteImageURL = decoder.decode<String>();
    auto absolutePDFURL = decoder.decode<String>();
    auto absoluteLinkURL = decoder.decode<String>();
    auto absoluteMediaURL = decoder.decode<String>();
    auto linkLabel = decoder.decode<String>();
    auto linkTitle = decoder.decode<String>();
    auto linkSuggestedFilename = decoder.decode<String>();
    auto isContentEditable = decoder.decode<bool>();
    auto elementBoundingBox = decoder.decode<WebCore::IntRect>();
    auto isScrollbar = decoder.decode<WebKit::WebHitTestResultData::IsScrollbar>();
    auto isSelected = decoder.decode<bool>();
    auto isTextNode = decoder.decode<bool>();
    auto isOverTextInsideFormControlElement = decoder.decode<bool>();
    auto isDownloadableMedia = decoder.decode<bool>();
    auto lookupText = decoder.decode<String>();
    auto toolTipText = decoder.decode<String>();
    auto imageText = decoder.decode<String>();
    auto getImageSharedMemoryHandle = decoder.decode<std::optional<WebKit::SharedMemoryHandle>>();
    auto imageBitmap = decoder.decode<RefPtr<WebKit::ShareableBitmap>>();
    auto sourceImageMIMEType = decoder.decode<String>();
#if PLATFORM(MAC)
    auto platformData = decoder.decode<WebKit::WebHitTestResultPlatformData>();
#endif
    auto dictionaryPopupInfo = decoder.decode<WebCore::DictionaryPopupInfo>();
    auto linkTextIndicator = decoder.decode<RefPtr<WebCore::TextIndicator>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebHitTestResultData {
            WTFMove(*absoluteImageURL),
            WTFMove(*absolutePDFURL),
            WTFMove(*absoluteLinkURL),
            WTFMove(*absoluteMediaURL),
            WTFMove(*linkLabel),
            WTFMove(*linkTitle),
            WTFMove(*linkSuggestedFilename),
            WTFMove(*isContentEditable),
            WTFMove(*elementBoundingBox),
            WTFMove(*isScrollbar),
            WTFMove(*isSelected),
            WTFMove(*isTextNode),
            WTFMove(*isOverTextInsideFormControlElement),
            WTFMove(*isDownloadableMedia),
            WTFMove(*lookupText),
            WTFMove(*toolTipText),
            WTFMove(*imageText),
            WTFMove(*getImageSharedMemoryHandle),
            WTFMove(*imageBitmap),
            WTFMove(*sourceImageMIMEType),
#if PLATFORM(MAC)
            WTFMove(*platformData),
#endif
            WTFMove(*dictionaryPopupInfo),
            WTFMove(*linkTextIndicator)
        }
    };
}

void ArgumentCoder<WebKit::WebPageCreationParameters>::encode(Encoder& encoder, const WebKit::WebPageCreationParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewSize)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.activityState)>, OptionSet<WebCore::ActivityState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.store)>, WebKit::WebPreferencesStore>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.drawingAreaType)>, WebKit::DrawingAreaType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.drawingAreaIdentifier)>, WebKit::DrawingAreaIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webPageProxyIdentifier)>, WebKit::WebPageProxyIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageGroupData)>, WebKit::WebPageGroupData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isEditable)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.underlayColor)>, WebCore::Color>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useFixedLayout)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fixedLayoutSize)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.defaultUnobscuredSize)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimumUnobscuredSize)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumUnobscuredSize)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewExposedRect)>, std::optional<WebCore::FloatRect>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alwaysShowsHorizontalScroller)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alwaysShowsVerticalScroller)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.suppressScrollbarAnimations)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paginationMode)>, WebCore::Pagination::Mode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paginationBehavesLikeColumns)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageLength)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gapBetweenPages)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userAgent)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.itemStatesWereRestoredByAPIRequest)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.itemStates)>, Vector<WebKit::BackForwardListItemState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.visitedLinkTableID)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canRunBeforeUnloadConfirmPanel)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canRunModal)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deviceScaleFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewScaleFactor)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textZoomFactor)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageZoomFactor)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topContentInset)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaVolume)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.muted)>, WebCore::MediaProducerMutedStateFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.openedByDOM)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mayStartMediaWhenInWindow)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaPlaybackIsSuspended)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minimumSizeForAutoLayout)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sizeToContentAutoSizeMaximumSize)>, WebCore::IntSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.autoSizingShouldExpandToViewHeight)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewportSizeForCSSViewportUnits)>, std::optional<WebCore::FloatSize>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollPinningBehavior)>, WebCore::ScrollPinningBehavior>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.scrollbarOverlayStyle)>, std::optional<uint32_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.backgroundExtendsBeyondPage)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layerHostingMode)>, WebKit::LayerHostingMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasResourceLoadClient)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mimeTypesWithCustomContentProviders)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.controlledByAutomation)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isProcessSwap)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useDarkAppearance)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useElevatedUserInterfaceLevel)>, bool>);
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, std::optional<WebCore::DestinationColorSpace>>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useSystemAppearance)>, bool>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useFormSemanticContext)>, bool>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.headerBannerHeight)>, int>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.footerBannerHeight)>, int>);
#endif
#if ENABLE(META_VIEWPORT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.ignoresViewportScaleLimits)>, bool>);
#endif
#if ENABLE(META_VIEWPORT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewportConfigurationViewLayoutSize)>, WebCore::FloatSize>);
#endif
#if ENABLE(META_VIEWPORT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewportConfigurationLayoutSizeScaleFactor)>, double>);
#endif
#if ENABLE(META_VIEWPORT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewportConfigurationMinimumEffectiveDeviceWidth)>, double>);
#endif
#if ENABLE(META_VIEWPORT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewportConfigurationViewSize)>, WebCore::FloatSize>);
#endif
#if ENABLE(META_VIEWPORT)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.overrideViewportArguments)>, std::optional<WebCore::ViewportArguments>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenSize)>, WebCore::FloatSize>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.availableScreenSize)>, WebCore::FloatSize>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.overrideScreenSize)>, WebCore::FloatSize>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textAutosizingWidth)>, float>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deviceOrientation)>, WebCore::IntDegrees>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.keyboardIsAttached)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canShowWhileLocked)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isCapturingScreen)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.insertionPointColor)>, WebCore::Color>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.smartInsertDeleteEnabled)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.additionalSupportedImageTypes)>, Vector<String>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gpuIOKitExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gpuMachExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
#endif
#if HAVE(STATIC_FONT_REGISTRY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontMachExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
#endif
#if HAVE(APP_ACCENT_COLORS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.accentColor)>, WebCore::Color>);
#endif
#if USE(WPE_RENDERER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hostFileDescriptor)>, UnixFileDescriptor>);
#endif
#if USE(GRAPHICS_LAYER_TEXTURE_MAPPER) || USE(GRAPHICS_LAYER_WC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.nativeWindowHandle)>, uint64_t>);
#endif
#if USE(GRAPHICS_LAYER_WC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usesOffscreenRendering)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldScaleViewToFitDocument)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userInterfaceLayoutDirection)>, WebCore::UserInterfaceLayoutDirection>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.observedLayoutMilestones)>, OptionSet<WebCore::LayoutMilestone>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.overrideContentSecurityPolicy)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cpuLimit)>, std::optional<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemeHandlers)>, HashMap<String, WebKit::WebURLSchemeHandlerIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesWithLegacyCustomProtocolHandlers)>, Vector<String>>);
#if ENABLE(APPLICATION_MANIFEST)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationManifest)>, std::optional<WebCore::ApplicationManifest>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.needsFontAttributes)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.iceCandidateFilteringEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enumeratingAllNetworkInterfacesEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userContentControllerParameters)>, WebKit::UserContentControllerParameters>);
#if ENABLE(WK_WEB_EXTENSIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webExtensionControllerParameters)>, std::optional<WebKit::WebExtensionControllerParameters>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.backgroundColor)>, std::optional<WebCore::Color>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.oldPageID)>, std::optional<WebCore::PageIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.overriddenMediaType)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.corsDisablingPatterns)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maskedURLSchemes)>, HashSet<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.userScriptsShouldWaitUntilNotification)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.loadsSubresources)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedNetworkHosts)>, std::optional<MemoryCompactLookupOnlyRobinHoodHashSet<String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginAccessControlCheckEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.processDisplayName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldCaptureAudioInUIProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldCaptureAudioInGPUProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldCaptureVideoInUIProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldCaptureVideoInGPUProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldCaptureDisplayInUIProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldCaptureDisplayInGPUProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRenderCanvasInGPUProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRenderDOMInGPUProcess)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldPlayMediaInGPUProcess)>, bool>);
#if ENABLE(WEBGL)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRenderWebGLInGPUProcess)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldEnableVP8Decoder)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldEnableVP9Decoder)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldEnableVP9SWDecoder)>, bool>);
#if ENABLE(APP_BOUND_DOMAINS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.limitsNavigationsToAppBoundDomains)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastNavigationWasAppInitiated)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canUseCredentialStorage)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldRelaxThirdPartyCookieBlocking)>, WebCore::ShouldRelaxThirdPartyCookieBlocking>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.httpsUpgradeEnabled)>, bool>);
#if PLATFORM(IOS) || PLATFORM(VISION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsDeprecatedSynchronousXMLHttpRequestDuringUnload)>, bool>);
#endif
#if ENABLE(APP_HIGHLIGHTS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.appHighlightsVisible)>, WebCore::HighlightVisibility>);
#endif
#if HAVE(TOUCH_BAR)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requiresUserActionForEditingControlsManager)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasResizableWindows)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSecurityPolicyModeForExtension)>, WebCore::ContentSecurityPolicyModeForExtension>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.subframeProcessFrameTreeCreationParameters)>, std::optional<WebKit::FrameTreeCreationParameters>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.openerFrameIdentifier)>, std::optional<WebCore::FrameIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mainFrameIdentifier)>, std::optional<WebCore::FrameIdentifier>>);
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linkDecorationFilteringData)>, Vector<WebCore::LinkDecorationFilteringData>>);
#endif
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedQueryParametersForAdvancedPrivacyProtections)>, Vector<WebCore::LinkDecorationFilteringData>>);
#endif
#if HAVE(MACH_BOOTSTRAP_EXTENSION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.machBootstrapHandle)>, WebKit::SandboxExtension::Handle>);
#endif
    struct ShouldBeSameSizeAsWebPageCreationParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebPageCreationParameters>, false> {
        WebCore::IntSize viewSize;
        OptionSet<WebCore::ActivityState> activityState;
        WebKit::WebPreferencesStore store;
        WebKit::DrawingAreaType drawingAreaType;
        WebKit::DrawingAreaIdentifier drawingAreaIdentifier;
        WebKit::WebPageProxyIdentifier webPageProxyIdentifier;
        WebKit::WebPageGroupData pageGroupData;
        bool isEditable;
        WebCore::Color underlayColor;
        bool useFixedLayout;
        WebCore::IntSize fixedLayoutSize;
        WebCore::FloatSize defaultUnobscuredSize;
        WebCore::FloatSize minimumUnobscuredSize;
        WebCore::FloatSize maximumUnobscuredSize;
        std::optional<WebCore::FloatRect> viewExposedRect;
        bool alwaysShowsHorizontalScroller;
        bool alwaysShowsVerticalScroller;
        bool suppressScrollbarAnimations;
        WebCore::Pagination::Mode paginationMode;
        bool paginationBehavesLikeColumns;
        double pageLength;
        double gapBetweenPages;
        String userAgent;
        bool itemStatesWereRestoredByAPIRequest;
        Vector<WebKit::BackForwardListItemState> itemStates;
        uint64_t visitedLinkTableID;
        bool canRunBeforeUnloadConfirmPanel;
        bool canRunModal;
        float deviceScaleFactor;
        float viewScaleFactor;
        double textZoomFactor;
        double pageZoomFactor;
        float topContentInset;
        float mediaVolume;
        WebCore::MediaProducerMutedStateFlags muted;
        bool openedByDOM;
        bool mayStartMediaWhenInWindow;
        bool mediaPlaybackIsSuspended;
        WebCore::IntSize minimumSizeForAutoLayout;
        WebCore::IntSize sizeToContentAutoSizeMaximumSize;
        bool autoSizingShouldExpandToViewHeight;
        std::optional<WebCore::FloatSize> viewportSizeForCSSViewportUnits;
        WebCore::ScrollPinningBehavior scrollPinningBehavior;
        std::optional<uint32_t> scrollbarOverlayStyle;
        bool backgroundExtendsBeyondPage;
        WebKit::LayerHostingMode layerHostingMode;
        bool hasResourceLoadClient;
        Vector<String> mimeTypesWithCustomContentProviders;
        bool controlledByAutomation;
        bool isProcessSwap;
        bool useDarkAppearance;
        bool useElevatedUserInterfaceLevel;
#if PLATFORM(MAC)
        std::optional<WebCore::DestinationColorSpace> colorSpace;
#endif
#if PLATFORM(MAC)
        bool useSystemAppearance;
#endif
#if PLATFORM(MAC)
        bool useFormSemanticContext;
#endif
#if PLATFORM(MAC)
        int headerBannerHeight;
#endif
#if PLATFORM(MAC)
        int footerBannerHeight;
#endif
#if ENABLE(META_VIEWPORT)
        bool ignoresViewportScaleLimits;
#endif
#if ENABLE(META_VIEWPORT)
        WebCore::FloatSize viewportConfigurationViewLayoutSize;
#endif
#if ENABLE(META_VIEWPORT)
        double viewportConfigurationLayoutSizeScaleFactor;
#endif
#if ENABLE(META_VIEWPORT)
        double viewportConfigurationMinimumEffectiveDeviceWidth;
#endif
#if ENABLE(META_VIEWPORT)
        WebCore::FloatSize viewportConfigurationViewSize;
#endif
#if ENABLE(META_VIEWPORT)
        std::optional<WebCore::ViewportArguments> overrideViewportArguments;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::FloatSize screenSize;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::FloatSize availableScreenSize;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::FloatSize overrideScreenSize;
#endif
#if PLATFORM(IOS_FAMILY)
        float textAutosizingWidth;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::IntDegrees deviceOrientation;
#endif
#if PLATFORM(IOS_FAMILY)
        bool keyboardIsAttached;
#endif
#if PLATFORM(IOS_FAMILY)
        bool canShowWhileLocked;
#endif
#if PLATFORM(IOS_FAMILY)
        bool isCapturingScreen;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::Color insertionPointColor;
#endif
#if PLATFORM(COCOA)
        bool smartInsertDeleteEnabled;
#endif
#if PLATFORM(COCOA)
        Vector<String> additionalSupportedImageTypes;
#endif
#if PLATFORM(COCOA)
        Vector<WebKit::SandboxExtension::Handle> gpuIOKitExtensionHandles;
#endif
#if PLATFORM(COCOA)
        Vector<WebKit::SandboxExtension::Handle> gpuMachExtensionHandles;
#endif
#if HAVE(STATIC_FONT_REGISTRY)
        Vector<WebKit::SandboxExtension::Handle> fontMachExtensionHandles;
#endif
#if HAVE(APP_ACCENT_COLORS)
        WebCore::Color accentColor;
#endif
#if USE(WPE_RENDERER)
        UnixFileDescriptor hostFileDescriptor;
#endif
#if USE(GRAPHICS_LAYER_TEXTURE_MAPPER) || USE(GRAPHICS_LAYER_WC)
        uint64_t nativeWindowHandle;
#endif
#if USE(GRAPHICS_LAYER_WC)
        bool usesOffscreenRendering;
#endif
        bool shouldScaleViewToFitDocument;
        WebCore::UserInterfaceLayoutDirection userInterfaceLayoutDirection;
        OptionSet<WebCore::LayoutMilestone> observedLayoutMilestones;
        String overrideContentSecurityPolicy;
        std::optional<double> cpuLimit;
        HashMap<String, WebKit::WebURLSchemeHandlerIdentifier> urlSchemeHandlers;
        Vector<String> urlSchemesWithLegacyCustomProtocolHandlers;
#if ENABLE(APPLICATION_MANIFEST)
        std::optional<WebCore::ApplicationManifest> applicationManifest;
#endif
        bool needsFontAttributes;
        bool iceCandidateFilteringEnabled;
        bool enumeratingAllNetworkInterfacesEnabled;
        WebKit::UserContentControllerParameters userContentControllerParameters;
#if ENABLE(WK_WEB_EXTENSIONS)
        std::optional<WebKit::WebExtensionControllerParameters> webExtensionControllerParameters;
#endif
        std::optional<WebCore::Color> backgroundColor;
        std::optional<WebCore::PageIdentifier> oldPageID;
        String overriddenMediaType;
        Vector<String> corsDisablingPatterns;
        HashSet<String> maskedURLSchemes;
        bool userScriptsShouldWaitUntilNotification;
        bool loadsSubresources;
        std::optional<MemoryCompactLookupOnlyRobinHoodHashSet<String>> allowedNetworkHosts;
        bool crossOriginAccessControlCheckEnabled;
        String processDisplayName;
        bool shouldCaptureAudioInUIProcess;
        bool shouldCaptureAudioInGPUProcess;
        bool shouldCaptureVideoInUIProcess;
        bool shouldCaptureVideoInGPUProcess;
        bool shouldCaptureDisplayInUIProcess;
        bool shouldCaptureDisplayInGPUProcess;
        bool shouldRenderCanvasInGPUProcess;
        bool shouldRenderDOMInGPUProcess;
        bool shouldPlayMediaInGPUProcess;
#if ENABLE(WEBGL)
        bool shouldRenderWebGLInGPUProcess;
#endif
        bool shouldEnableVP8Decoder;
        bool shouldEnableVP9Decoder;
        bool shouldEnableVP9SWDecoder;
#if ENABLE(APP_BOUND_DOMAINS)
        bool limitsNavigationsToAppBoundDomains;
#endif
        bool lastNavigationWasAppInitiated;
        bool canUseCredentialStorage;
        WebCore::ShouldRelaxThirdPartyCookieBlocking shouldRelaxThirdPartyCookieBlocking;
        bool httpsUpgradeEnabled;
#if PLATFORM(IOS) || PLATFORM(VISION)
        bool allowsDeprecatedSynchronousXMLHttpRequestDuringUnload;
#endif
#if ENABLE(APP_HIGHLIGHTS)
        WebCore::HighlightVisibility appHighlightsVisible;
#endif
#if HAVE(TOUCH_BAR)
        bool requiresUserActionForEditingControlsManager;
#endif
        bool hasResizableWindows;
        WebCore::ContentSecurityPolicyModeForExtension contentSecurityPolicyModeForExtension;
        std::optional<WebKit::FrameTreeCreationParameters> subframeProcessFrameTreeCreationParameters;
        std::optional<WebCore::FrameIdentifier> openerFrameIdentifier;
        std::optional<WebCore::FrameIdentifier> mainFrameIdentifier;
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
        Vector<WebCore::LinkDecorationFilteringData> linkDecorationFilteringData;
#endif
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
        Vector<WebCore::LinkDecorationFilteringData> allowedQueryParametersForAdvancedPrivacyProtections;
#endif
#if HAVE(MACH_BOOTSTRAP_EXTENSION)
        WebKit::SandboxExtension::Handle machBootstrapHandle;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebPageCreationParameters) == sizeof(WebKit::WebPageCreationParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebPageCreationParameters, viewSize)
        , offsetof(WebKit::WebPageCreationParameters, activityState)
        , offsetof(WebKit::WebPageCreationParameters, store)
        , offsetof(WebKit::WebPageCreationParameters, drawingAreaType)
        , offsetof(WebKit::WebPageCreationParameters, drawingAreaIdentifier)
        , offsetof(WebKit::WebPageCreationParameters, webPageProxyIdentifier)
        , offsetof(WebKit::WebPageCreationParameters, pageGroupData)
        , offsetof(WebKit::WebPageCreationParameters, isEditable)
        , offsetof(WebKit::WebPageCreationParameters, underlayColor)
        , offsetof(WebKit::WebPageCreationParameters, useFixedLayout)
        , offsetof(WebKit::WebPageCreationParameters, fixedLayoutSize)
        , offsetof(WebKit::WebPageCreationParameters, defaultUnobscuredSize)
        , offsetof(WebKit::WebPageCreationParameters, minimumUnobscuredSize)
        , offsetof(WebKit::WebPageCreationParameters, maximumUnobscuredSize)
        , offsetof(WebKit::WebPageCreationParameters, viewExposedRect)
        , offsetof(WebKit::WebPageCreationParameters, alwaysShowsHorizontalScroller)
        , offsetof(WebKit::WebPageCreationParameters, alwaysShowsVerticalScroller)
        , offsetof(WebKit::WebPageCreationParameters, suppressScrollbarAnimations)
        , offsetof(WebKit::WebPageCreationParameters, paginationMode)
        , offsetof(WebKit::WebPageCreationParameters, paginationBehavesLikeColumns)
        , offsetof(WebKit::WebPageCreationParameters, pageLength)
        , offsetof(WebKit::WebPageCreationParameters, gapBetweenPages)
        , offsetof(WebKit::WebPageCreationParameters, userAgent)
        , offsetof(WebKit::WebPageCreationParameters, itemStatesWereRestoredByAPIRequest)
        , offsetof(WebKit::WebPageCreationParameters, itemStates)
        , offsetof(WebKit::WebPageCreationParameters, visitedLinkTableID)
        , offsetof(WebKit::WebPageCreationParameters, canRunBeforeUnloadConfirmPanel)
        , offsetof(WebKit::WebPageCreationParameters, canRunModal)
        , offsetof(WebKit::WebPageCreationParameters, deviceScaleFactor)
        , offsetof(WebKit::WebPageCreationParameters, viewScaleFactor)
        , offsetof(WebKit::WebPageCreationParameters, textZoomFactor)
        , offsetof(WebKit::WebPageCreationParameters, pageZoomFactor)
        , offsetof(WebKit::WebPageCreationParameters, topContentInset)
        , offsetof(WebKit::WebPageCreationParameters, mediaVolume)
        , offsetof(WebKit::WebPageCreationParameters, muted)
        , offsetof(WebKit::WebPageCreationParameters, openedByDOM)
        , offsetof(WebKit::WebPageCreationParameters, mayStartMediaWhenInWindow)
        , offsetof(WebKit::WebPageCreationParameters, mediaPlaybackIsSuspended)
        , offsetof(WebKit::WebPageCreationParameters, minimumSizeForAutoLayout)
        , offsetof(WebKit::WebPageCreationParameters, sizeToContentAutoSizeMaximumSize)
        , offsetof(WebKit::WebPageCreationParameters, autoSizingShouldExpandToViewHeight)
        , offsetof(WebKit::WebPageCreationParameters, viewportSizeForCSSViewportUnits)
        , offsetof(WebKit::WebPageCreationParameters, scrollPinningBehavior)
        , offsetof(WebKit::WebPageCreationParameters, scrollbarOverlayStyle)
        , offsetof(WebKit::WebPageCreationParameters, backgroundExtendsBeyondPage)
        , offsetof(WebKit::WebPageCreationParameters, layerHostingMode)
        , offsetof(WebKit::WebPageCreationParameters, hasResourceLoadClient)
        , offsetof(WebKit::WebPageCreationParameters, mimeTypesWithCustomContentProviders)
        , offsetof(WebKit::WebPageCreationParameters, controlledByAutomation)
        , offsetof(WebKit::WebPageCreationParameters, isProcessSwap)
        , offsetof(WebKit::WebPageCreationParameters, useDarkAppearance)
        , offsetof(WebKit::WebPageCreationParameters, useElevatedUserInterfaceLevel)
#if PLATFORM(MAC)
        , offsetof(WebKit::WebPageCreationParameters, colorSpace)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::WebPageCreationParameters, useSystemAppearance)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::WebPageCreationParameters, useFormSemanticContext)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::WebPageCreationParameters, headerBannerHeight)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::WebPageCreationParameters, footerBannerHeight)
#endif
#if ENABLE(META_VIEWPORT)
        , offsetof(WebKit::WebPageCreationParameters, ignoresViewportScaleLimits)
#endif
#if ENABLE(META_VIEWPORT)
        , offsetof(WebKit::WebPageCreationParameters, viewportConfigurationViewLayoutSize)
#endif
#if ENABLE(META_VIEWPORT)
        , offsetof(WebKit::WebPageCreationParameters, viewportConfigurationLayoutSizeScaleFactor)
#endif
#if ENABLE(META_VIEWPORT)
        , offsetof(WebKit::WebPageCreationParameters, viewportConfigurationMinimumEffectiveDeviceWidth)
#endif
#if ENABLE(META_VIEWPORT)
        , offsetof(WebKit::WebPageCreationParameters, viewportConfigurationViewSize)
#endif
#if ENABLE(META_VIEWPORT)
        , offsetof(WebKit::WebPageCreationParameters, overrideViewportArguments)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, screenSize)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, availableScreenSize)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, overrideScreenSize)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, textAutosizingWidth)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, deviceOrientation)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, keyboardIsAttached)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, canShowWhileLocked)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, isCapturingScreen)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebPageCreationParameters, insertionPointColor)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebPageCreationParameters, smartInsertDeleteEnabled)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebPageCreationParameters, additionalSupportedImageTypes)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebPageCreationParameters, gpuIOKitExtensionHandles)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebPageCreationParameters, gpuMachExtensionHandles)
#endif
#if HAVE(STATIC_FONT_REGISTRY)
        , offsetof(WebKit::WebPageCreationParameters, fontMachExtensionHandles)
#endif
#if HAVE(APP_ACCENT_COLORS)
        , offsetof(WebKit::WebPageCreationParameters, accentColor)
#endif
#if USE(WPE_RENDERER)
        , offsetof(WebKit::WebPageCreationParameters, hostFileDescriptor)
#endif
#if USE(GRAPHICS_LAYER_TEXTURE_MAPPER) || USE(GRAPHICS_LAYER_WC)
        , offsetof(WebKit::WebPageCreationParameters, nativeWindowHandle)
#endif
#if USE(GRAPHICS_LAYER_WC)
        , offsetof(WebKit::WebPageCreationParameters, usesOffscreenRendering)
#endif
        , offsetof(WebKit::WebPageCreationParameters, shouldScaleViewToFitDocument)
        , offsetof(WebKit::WebPageCreationParameters, userInterfaceLayoutDirection)
        , offsetof(WebKit::WebPageCreationParameters, observedLayoutMilestones)
        , offsetof(WebKit::WebPageCreationParameters, overrideContentSecurityPolicy)
        , offsetof(WebKit::WebPageCreationParameters, cpuLimit)
        , offsetof(WebKit::WebPageCreationParameters, urlSchemeHandlers)
        , offsetof(WebKit::WebPageCreationParameters, urlSchemesWithLegacyCustomProtocolHandlers)
#if ENABLE(APPLICATION_MANIFEST)
        , offsetof(WebKit::WebPageCreationParameters, applicationManifest)
#endif
        , offsetof(WebKit::WebPageCreationParameters, needsFontAttributes)
        , offsetof(WebKit::WebPageCreationParameters, iceCandidateFilteringEnabled)
        , offsetof(WebKit::WebPageCreationParameters, enumeratingAllNetworkInterfacesEnabled)
        , offsetof(WebKit::WebPageCreationParameters, userContentControllerParameters)
#if ENABLE(WK_WEB_EXTENSIONS)
        , offsetof(WebKit::WebPageCreationParameters, webExtensionControllerParameters)
#endif
        , offsetof(WebKit::WebPageCreationParameters, backgroundColor)
        , offsetof(WebKit::WebPageCreationParameters, oldPageID)
        , offsetof(WebKit::WebPageCreationParameters, overriddenMediaType)
        , offsetof(WebKit::WebPageCreationParameters, corsDisablingPatterns)
        , offsetof(WebKit::WebPageCreationParameters, maskedURLSchemes)
        , offsetof(WebKit::WebPageCreationParameters, userScriptsShouldWaitUntilNotification)
        , offsetof(WebKit::WebPageCreationParameters, loadsSubresources)
        , offsetof(WebKit::WebPageCreationParameters, allowedNetworkHosts)
        , offsetof(WebKit::WebPageCreationParameters, crossOriginAccessControlCheckEnabled)
        , offsetof(WebKit::WebPageCreationParameters, processDisplayName)
        , offsetof(WebKit::WebPageCreationParameters, shouldCaptureAudioInUIProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldCaptureAudioInGPUProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldCaptureVideoInUIProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldCaptureVideoInGPUProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldCaptureDisplayInUIProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldCaptureDisplayInGPUProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldRenderCanvasInGPUProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldRenderDOMInGPUProcess)
        , offsetof(WebKit::WebPageCreationParameters, shouldPlayMediaInGPUProcess)
#if ENABLE(WEBGL)
        , offsetof(WebKit::WebPageCreationParameters, shouldRenderWebGLInGPUProcess)
#endif
        , offsetof(WebKit::WebPageCreationParameters, shouldEnableVP8Decoder)
        , offsetof(WebKit::WebPageCreationParameters, shouldEnableVP9Decoder)
        , offsetof(WebKit::WebPageCreationParameters, shouldEnableVP9SWDecoder)
#if ENABLE(APP_BOUND_DOMAINS)
        , offsetof(WebKit::WebPageCreationParameters, limitsNavigationsToAppBoundDomains)
#endif
        , offsetof(WebKit::WebPageCreationParameters, lastNavigationWasAppInitiated)
        , offsetof(WebKit::WebPageCreationParameters, canUseCredentialStorage)
        , offsetof(WebKit::WebPageCreationParameters, shouldRelaxThirdPartyCookieBlocking)
        , offsetof(WebKit::WebPageCreationParameters, httpsUpgradeEnabled)
#if PLATFORM(IOS) || PLATFORM(VISION)
        , offsetof(WebKit::WebPageCreationParameters, allowsDeprecatedSynchronousXMLHttpRequestDuringUnload)
#endif
#if ENABLE(APP_HIGHLIGHTS)
        , offsetof(WebKit::WebPageCreationParameters, appHighlightsVisible)
#endif
#if HAVE(TOUCH_BAR)
        , offsetof(WebKit::WebPageCreationParameters, requiresUserActionForEditingControlsManager)
#endif
        , offsetof(WebKit::WebPageCreationParameters, hasResizableWindows)
        , offsetof(WebKit::WebPageCreationParameters, contentSecurityPolicyModeForExtension)
        , offsetof(WebKit::WebPageCreationParameters, subframeProcessFrameTreeCreationParameters)
        , offsetof(WebKit::WebPageCreationParameters, openerFrameIdentifier)
        , offsetof(WebKit::WebPageCreationParameters, mainFrameIdentifier)
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
        , offsetof(WebKit::WebPageCreationParameters, linkDecorationFilteringData)
#endif
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
        , offsetof(WebKit::WebPageCreationParameters, allowedQueryParametersForAdvancedPrivacyProtections)
#endif
#if HAVE(MACH_BOOTSTRAP_EXTENSION)
        , offsetof(WebKit::WebPageCreationParameters, machBootstrapHandle)
#endif
    >::value);
    encoder << instance.viewSize;
    encoder << instance.activityState;
    encoder << instance.store;
    encoder << instance.drawingAreaType;
    encoder << instance.drawingAreaIdentifier;
    encoder << instance.webPageProxyIdentifier;
    encoder << instance.pageGroupData;
    encoder << instance.isEditable;
    encoder << instance.underlayColor;
    encoder << instance.useFixedLayout;
    encoder << instance.fixedLayoutSize;
    encoder << instance.defaultUnobscuredSize;
    encoder << instance.minimumUnobscuredSize;
    encoder << instance.maximumUnobscuredSize;
    encoder << instance.viewExposedRect;
    encoder << instance.alwaysShowsHorizontalScroller;
    encoder << instance.alwaysShowsVerticalScroller;
    encoder << instance.suppressScrollbarAnimations;
    encoder << instance.paginationMode;
    encoder << instance.paginationBehavesLikeColumns;
    encoder << instance.pageLength;
    encoder << instance.gapBetweenPages;
    encoder << instance.userAgent;
    encoder << instance.itemStatesWereRestoredByAPIRequest;
    encoder << instance.itemStates;
    encoder << instance.visitedLinkTableID;
    encoder << instance.canRunBeforeUnloadConfirmPanel;
    encoder << instance.canRunModal;
    encoder << instance.deviceScaleFactor;
    encoder << instance.viewScaleFactor;
    encoder << instance.textZoomFactor;
    encoder << instance.pageZoomFactor;
    encoder << instance.topContentInset;
    encoder << instance.mediaVolume;
    encoder << instance.muted;
    encoder << instance.openedByDOM;
    encoder << instance.mayStartMediaWhenInWindow;
    encoder << instance.mediaPlaybackIsSuspended;
    encoder << instance.minimumSizeForAutoLayout;
    encoder << instance.sizeToContentAutoSizeMaximumSize;
    encoder << instance.autoSizingShouldExpandToViewHeight;
    encoder << instance.viewportSizeForCSSViewportUnits;
    encoder << instance.scrollPinningBehavior;
    encoder << instance.scrollbarOverlayStyle;
    encoder << instance.backgroundExtendsBeyondPage;
    encoder << instance.layerHostingMode;
    encoder << instance.hasResourceLoadClient;
    encoder << instance.mimeTypesWithCustomContentProviders;
    encoder << instance.controlledByAutomation;
    encoder << instance.isProcessSwap;
    encoder << instance.useDarkAppearance;
    encoder << instance.useElevatedUserInterfaceLevel;
#if PLATFORM(MAC)
    encoder << instance.colorSpace;
#endif
#if PLATFORM(MAC)
    encoder << instance.useSystemAppearance;
#endif
#if PLATFORM(MAC)
    encoder << instance.useFormSemanticContext;
#endif
#if PLATFORM(MAC)
    encoder << instance.headerBannerHeight;
#endif
#if PLATFORM(MAC)
    encoder << instance.footerBannerHeight;
#endif
#if ENABLE(META_VIEWPORT)
    encoder << instance.ignoresViewportScaleLimits;
#endif
#if ENABLE(META_VIEWPORT)
    encoder << instance.viewportConfigurationViewLayoutSize;
#endif
#if ENABLE(META_VIEWPORT)
    encoder << instance.viewportConfigurationLayoutSizeScaleFactor;
#endif
#if ENABLE(META_VIEWPORT)
    encoder << instance.viewportConfigurationMinimumEffectiveDeviceWidth;
#endif
#if ENABLE(META_VIEWPORT)
    encoder << instance.viewportConfigurationViewSize;
#endif
#if ENABLE(META_VIEWPORT)
    encoder << instance.overrideViewportArguments;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.screenSize;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.availableScreenSize;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.overrideScreenSize;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.textAutosizingWidth;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.deviceOrientation;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.keyboardIsAttached;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.canShowWhileLocked;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.isCapturingScreen;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.insertionPointColor;
#endif
#if PLATFORM(COCOA)
    encoder << instance.smartInsertDeleteEnabled;
#endif
#if PLATFORM(COCOA)
    encoder << instance.additionalSupportedImageTypes;
#endif
#if PLATFORM(COCOA)
    encoder << instance.gpuIOKitExtensionHandles;
#endif
#if PLATFORM(COCOA)
    encoder << instance.gpuMachExtensionHandles;
#endif
#if HAVE(STATIC_FONT_REGISTRY)
    encoder << instance.fontMachExtensionHandles;
#endif
#if HAVE(APP_ACCENT_COLORS)
    encoder << instance.accentColor;
#endif
#if USE(WPE_RENDERER)
    encoder << instance.hostFileDescriptor;
#endif
#if USE(GRAPHICS_LAYER_TEXTURE_MAPPER) || USE(GRAPHICS_LAYER_WC)
    encoder << instance.nativeWindowHandle;
#endif
#if USE(GRAPHICS_LAYER_WC)
    encoder << instance.usesOffscreenRendering;
#endif
    encoder << instance.shouldScaleViewToFitDocument;
    encoder << instance.userInterfaceLayoutDirection;
    encoder << instance.observedLayoutMilestones;
    encoder << instance.overrideContentSecurityPolicy;
    encoder << instance.cpuLimit;
    encoder << instance.urlSchemeHandlers;
    encoder << instance.urlSchemesWithLegacyCustomProtocolHandlers;
#if ENABLE(APPLICATION_MANIFEST)
    encoder << instance.applicationManifest;
#endif
    encoder << instance.needsFontAttributes;
    encoder << instance.iceCandidateFilteringEnabled;
    encoder << instance.enumeratingAllNetworkInterfacesEnabled;
    encoder << instance.userContentControllerParameters;
#if ENABLE(WK_WEB_EXTENSIONS)
    encoder << instance.webExtensionControllerParameters;
#endif
    encoder << instance.backgroundColor;
    encoder << instance.oldPageID;
    encoder << instance.overriddenMediaType;
    encoder << instance.corsDisablingPatterns;
    encoder << instance.maskedURLSchemes;
    encoder << instance.userScriptsShouldWaitUntilNotification;
    encoder << instance.loadsSubresources;
    encoder << instance.allowedNetworkHosts;
    encoder << instance.crossOriginAccessControlCheckEnabled;
    encoder << instance.processDisplayName;
    encoder << instance.shouldCaptureAudioInUIProcess;
    encoder << instance.shouldCaptureAudioInGPUProcess;
    encoder << instance.shouldCaptureVideoInUIProcess;
    encoder << instance.shouldCaptureVideoInGPUProcess;
    encoder << instance.shouldCaptureDisplayInUIProcess;
    encoder << instance.shouldCaptureDisplayInGPUProcess;
    encoder << instance.shouldRenderCanvasInGPUProcess;
    encoder << instance.shouldRenderDOMInGPUProcess;
    encoder << instance.shouldPlayMediaInGPUProcess;
#if ENABLE(WEBGL)
    encoder << instance.shouldRenderWebGLInGPUProcess;
#endif
    encoder << instance.shouldEnableVP8Decoder;
    encoder << instance.shouldEnableVP9Decoder;
    encoder << instance.shouldEnableVP9SWDecoder;
#if ENABLE(APP_BOUND_DOMAINS)
    encoder << instance.limitsNavigationsToAppBoundDomains;
#endif
    encoder << instance.lastNavigationWasAppInitiated;
    encoder << instance.canUseCredentialStorage;
    encoder << instance.shouldRelaxThirdPartyCookieBlocking;
    encoder << instance.httpsUpgradeEnabled;
#if PLATFORM(IOS) || PLATFORM(VISION)
    encoder << instance.allowsDeprecatedSynchronousXMLHttpRequestDuringUnload;
#endif
#if ENABLE(APP_HIGHLIGHTS)
    encoder << instance.appHighlightsVisible;
#endif
#if HAVE(TOUCH_BAR)
    encoder << instance.requiresUserActionForEditingControlsManager;
#endif
    encoder << instance.hasResizableWindows;
    encoder << instance.contentSecurityPolicyModeForExtension;
    encoder << instance.subframeProcessFrameTreeCreationParameters;
    encoder << instance.openerFrameIdentifier;
    encoder << instance.mainFrameIdentifier;
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
    encoder << instance.linkDecorationFilteringData;
#endif
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
    encoder << instance.allowedQueryParametersForAdvancedPrivacyProtections;
#endif
#if HAVE(MACH_BOOTSTRAP_EXTENSION)
    encoder << instance.machBootstrapHandle;
#endif
}

std::optional<WebKit::WebPageCreationParameters> ArgumentCoder<WebKit::WebPageCreationParameters>::decode(Decoder& decoder)
{
    auto viewSize = decoder.decode<WebCore::IntSize>();
    auto activityState = decoder.decode<OptionSet<WebCore::ActivityState>>();
    auto store = decoder.decode<WebKit::WebPreferencesStore>();
    auto drawingAreaType = decoder.decode<WebKit::DrawingAreaType>();
    auto drawingAreaIdentifier = decoder.decode<WebKit::DrawingAreaIdentifier>();
    auto webPageProxyIdentifier = decoder.decode<WebKit::WebPageProxyIdentifier>();
    auto pageGroupData = decoder.decode<WebKit::WebPageGroupData>();
    auto isEditable = decoder.decode<bool>();
    auto underlayColor = decoder.decode<WebCore::Color>();
    auto useFixedLayout = decoder.decode<bool>();
    auto fixedLayoutSize = decoder.decode<WebCore::IntSize>();
    auto defaultUnobscuredSize = decoder.decode<WebCore::FloatSize>();
    auto minimumUnobscuredSize = decoder.decode<WebCore::FloatSize>();
    auto maximumUnobscuredSize = decoder.decode<WebCore::FloatSize>();
    auto viewExposedRect = decoder.decode<std::optional<WebCore::FloatRect>>();
    auto alwaysShowsHorizontalScroller = decoder.decode<bool>();
    auto alwaysShowsVerticalScroller = decoder.decode<bool>();
    auto suppressScrollbarAnimations = decoder.decode<bool>();
    auto paginationMode = decoder.decode<WebCore::Pagination::Mode>();
    auto paginationBehavesLikeColumns = decoder.decode<bool>();
    auto pageLength = decoder.decode<double>();
    auto gapBetweenPages = decoder.decode<double>();
    auto userAgent = decoder.decode<String>();
    auto itemStatesWereRestoredByAPIRequest = decoder.decode<bool>();
    auto itemStates = decoder.decode<Vector<WebKit::BackForwardListItemState>>();
    auto visitedLinkTableID = decoder.decode<uint64_t>();
    auto canRunBeforeUnloadConfirmPanel = decoder.decode<bool>();
    auto canRunModal = decoder.decode<bool>();
    auto deviceScaleFactor = decoder.decode<float>();
    auto viewScaleFactor = decoder.decode<float>();
    auto textZoomFactor = decoder.decode<double>();
    auto pageZoomFactor = decoder.decode<double>();
    auto topContentInset = decoder.decode<float>();
    auto mediaVolume = decoder.decode<float>();
    auto muted = decoder.decode<WebCore::MediaProducerMutedStateFlags>();
    auto openedByDOM = decoder.decode<bool>();
    auto mayStartMediaWhenInWindow = decoder.decode<bool>();
    auto mediaPlaybackIsSuspended = decoder.decode<bool>();
    auto minimumSizeForAutoLayout = decoder.decode<WebCore::IntSize>();
    auto sizeToContentAutoSizeMaximumSize = decoder.decode<WebCore::IntSize>();
    auto autoSizingShouldExpandToViewHeight = decoder.decode<bool>();
    auto viewportSizeForCSSViewportUnits = decoder.decode<std::optional<WebCore::FloatSize>>();
    auto scrollPinningBehavior = decoder.decode<WebCore::ScrollPinningBehavior>();
    auto scrollbarOverlayStyle = decoder.decode<std::optional<uint32_t>>();
    auto backgroundExtendsBeyondPage = decoder.decode<bool>();
    auto layerHostingMode = decoder.decode<WebKit::LayerHostingMode>();
    auto hasResourceLoadClient = decoder.decode<bool>();
    auto mimeTypesWithCustomContentProviders = decoder.decode<Vector<String>>();
    auto controlledByAutomation = decoder.decode<bool>();
    auto isProcessSwap = decoder.decode<bool>();
    auto useDarkAppearance = decoder.decode<bool>();
    auto useElevatedUserInterfaceLevel = decoder.decode<bool>();
#if PLATFORM(MAC)
    auto colorSpace = decoder.decode<std::optional<WebCore::DestinationColorSpace>>();
#endif
#if PLATFORM(MAC)
    auto useSystemAppearance = decoder.decode<bool>();
#endif
#if PLATFORM(MAC)
    auto useFormSemanticContext = decoder.decode<bool>();
#endif
#if PLATFORM(MAC)
    auto headerBannerHeight = decoder.decode<int>();
#endif
#if PLATFORM(MAC)
    auto footerBannerHeight = decoder.decode<int>();
#endif
#if ENABLE(META_VIEWPORT)
    auto ignoresViewportScaleLimits = decoder.decode<bool>();
#endif
#if ENABLE(META_VIEWPORT)
    auto viewportConfigurationViewLayoutSize = decoder.decode<WebCore::FloatSize>();
#endif
#if ENABLE(META_VIEWPORT)
    auto viewportConfigurationLayoutSizeScaleFactor = decoder.decode<double>();
#endif
#if ENABLE(META_VIEWPORT)
    auto viewportConfigurationMinimumEffectiveDeviceWidth = decoder.decode<double>();
#endif
#if ENABLE(META_VIEWPORT)
    auto viewportConfigurationViewSize = decoder.decode<WebCore::FloatSize>();
#endif
#if ENABLE(META_VIEWPORT)
    auto overrideViewportArguments = decoder.decode<std::optional<WebCore::ViewportArguments>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto screenSize = decoder.decode<WebCore::FloatSize>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto availableScreenSize = decoder.decode<WebCore::FloatSize>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto overrideScreenSize = decoder.decode<WebCore::FloatSize>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto textAutosizingWidth = decoder.decode<float>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto deviceOrientation = decoder.decode<WebCore::IntDegrees>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto keyboardIsAttached = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto canShowWhileLocked = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto isCapturingScreen = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto insertionPointColor = decoder.decode<WebCore::Color>();
#endif
#if PLATFORM(COCOA)
    auto smartInsertDeleteEnabled = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto additionalSupportedImageTypes = decoder.decode<Vector<String>>();
#endif
#if PLATFORM(COCOA)
    auto gpuIOKitExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(COCOA)
    auto gpuMachExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
#endif
#if HAVE(STATIC_FONT_REGISTRY)
    auto fontMachExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
#endif
#if HAVE(APP_ACCENT_COLORS)
    auto accentColor = decoder.decode<WebCore::Color>();
#endif
#if USE(WPE_RENDERER)
    auto hostFileDescriptor = decoder.decode<UnixFileDescriptor>();
#endif
#if USE(GRAPHICS_LAYER_TEXTURE_MAPPER) || USE(GRAPHICS_LAYER_WC)
    auto nativeWindowHandle = decoder.decode<uint64_t>();
#endif
#if USE(GRAPHICS_LAYER_WC)
    auto usesOffscreenRendering = decoder.decode<bool>();
#endif
    auto shouldScaleViewToFitDocument = decoder.decode<bool>();
    auto userInterfaceLayoutDirection = decoder.decode<WebCore::UserInterfaceLayoutDirection>();
    auto observedLayoutMilestones = decoder.decode<OptionSet<WebCore::LayoutMilestone>>();
    auto overrideContentSecurityPolicy = decoder.decode<String>();
    auto cpuLimit = decoder.decode<std::optional<double>>();
    auto urlSchemeHandlers = decoder.decode<HashMap<String, WebKit::WebURLSchemeHandlerIdentifier>>();
    auto urlSchemesWithLegacyCustomProtocolHandlers = decoder.decode<Vector<String>>();
#if ENABLE(APPLICATION_MANIFEST)
    auto applicationManifest = decoder.decode<std::optional<WebCore::ApplicationManifest>>();
#endif
    auto needsFontAttributes = decoder.decode<bool>();
    auto iceCandidateFilteringEnabled = decoder.decode<bool>();
    auto enumeratingAllNetworkInterfacesEnabled = decoder.decode<bool>();
    auto userContentControllerParameters = decoder.decode<WebKit::UserContentControllerParameters>();
#if ENABLE(WK_WEB_EXTENSIONS)
    auto webExtensionControllerParameters = decoder.decode<std::optional<WebKit::WebExtensionControllerParameters>>();
#endif
    auto backgroundColor = decoder.decode<std::optional<WebCore::Color>>();
    auto oldPageID = decoder.decode<std::optional<WebCore::PageIdentifier>>();
    auto overriddenMediaType = decoder.decode<String>();
    auto corsDisablingPatterns = decoder.decode<Vector<String>>();
    auto maskedURLSchemes = decoder.decode<HashSet<String>>();
    auto userScriptsShouldWaitUntilNotification = decoder.decode<bool>();
    auto loadsSubresources = decoder.decode<bool>();
    auto allowedNetworkHosts = decoder.decode<std::optional<MemoryCompactLookupOnlyRobinHoodHashSet<String>>>();
    auto crossOriginAccessControlCheckEnabled = decoder.decode<bool>();
    auto processDisplayName = decoder.decode<String>();
    auto shouldCaptureAudioInUIProcess = decoder.decode<bool>();
    auto shouldCaptureAudioInGPUProcess = decoder.decode<bool>();
    auto shouldCaptureVideoInUIProcess = decoder.decode<bool>();
    auto shouldCaptureVideoInGPUProcess = decoder.decode<bool>();
    auto shouldCaptureDisplayInUIProcess = decoder.decode<bool>();
    auto shouldCaptureDisplayInGPUProcess = decoder.decode<bool>();
    auto shouldRenderCanvasInGPUProcess = decoder.decode<bool>();
    auto shouldRenderDOMInGPUProcess = decoder.decode<bool>();
    auto shouldPlayMediaInGPUProcess = decoder.decode<bool>();
#if ENABLE(WEBGL)
    auto shouldRenderWebGLInGPUProcess = decoder.decode<bool>();
#endif
    auto shouldEnableVP8Decoder = decoder.decode<bool>();
    auto shouldEnableVP9Decoder = decoder.decode<bool>();
    auto shouldEnableVP9SWDecoder = decoder.decode<bool>();
#if ENABLE(APP_BOUND_DOMAINS)
    auto limitsNavigationsToAppBoundDomains = decoder.decode<bool>();
#endif
    auto lastNavigationWasAppInitiated = decoder.decode<bool>();
    auto canUseCredentialStorage = decoder.decode<bool>();
    auto shouldRelaxThirdPartyCookieBlocking = decoder.decode<WebCore::ShouldRelaxThirdPartyCookieBlocking>();
    auto httpsUpgradeEnabled = decoder.decode<bool>();
#if PLATFORM(IOS) || PLATFORM(VISION)
    auto allowsDeprecatedSynchronousXMLHttpRequestDuringUnload = decoder.decode<bool>();
#endif
#if ENABLE(APP_HIGHLIGHTS)
    auto appHighlightsVisible = decoder.decode<WebCore::HighlightVisibility>();
#endif
#if HAVE(TOUCH_BAR)
    auto requiresUserActionForEditingControlsManager = decoder.decode<bool>();
#endif
    auto hasResizableWindows = decoder.decode<bool>();
    auto contentSecurityPolicyModeForExtension = decoder.decode<WebCore::ContentSecurityPolicyModeForExtension>();
    auto subframeProcessFrameTreeCreationParameters = decoder.decode<std::optional<WebKit::FrameTreeCreationParameters>>();
    auto openerFrameIdentifier = decoder.decode<std::optional<WebCore::FrameIdentifier>>();
    auto mainFrameIdentifier = decoder.decode<std::optional<WebCore::FrameIdentifier>>();
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
    auto linkDecorationFilteringData = decoder.decode<Vector<WebCore::LinkDecorationFilteringData>>();
#endif
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
    auto allowedQueryParametersForAdvancedPrivacyProtections = decoder.decode<Vector<WebCore::LinkDecorationFilteringData>>();
#endif
#if HAVE(MACH_BOOTSTRAP_EXTENSION)
    auto machBootstrapHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebPageCreationParameters {
            WTFMove(*viewSize),
            WTFMove(*activityState),
            WTFMove(*store),
            WTFMove(*drawingAreaType),
            WTFMove(*drawingAreaIdentifier),
            WTFMove(*webPageProxyIdentifier),
            WTFMove(*pageGroupData),
            WTFMove(*isEditable),
            WTFMove(*underlayColor),
            WTFMove(*useFixedLayout),
            WTFMove(*fixedLayoutSize),
            WTFMove(*defaultUnobscuredSize),
            WTFMove(*minimumUnobscuredSize),
            WTFMove(*maximumUnobscuredSize),
            WTFMove(*viewExposedRect),
            WTFMove(*alwaysShowsHorizontalScroller),
            WTFMove(*alwaysShowsVerticalScroller),
            WTFMove(*suppressScrollbarAnimations),
            WTFMove(*paginationMode),
            WTFMove(*paginationBehavesLikeColumns),
            WTFMove(*pageLength),
            WTFMove(*gapBetweenPages),
            WTFMove(*userAgent),
            WTFMove(*itemStatesWereRestoredByAPIRequest),
            WTFMove(*itemStates),
            WTFMove(*visitedLinkTableID),
            WTFMove(*canRunBeforeUnloadConfirmPanel),
            WTFMove(*canRunModal),
            WTFMove(*deviceScaleFactor),
            WTFMove(*viewScaleFactor),
            WTFMove(*textZoomFactor),
            WTFMove(*pageZoomFactor),
            WTFMove(*topContentInset),
            WTFMove(*mediaVolume),
            WTFMove(*muted),
            WTFMove(*openedByDOM),
            WTFMove(*mayStartMediaWhenInWindow),
            WTFMove(*mediaPlaybackIsSuspended),
            WTFMove(*minimumSizeForAutoLayout),
            WTFMove(*sizeToContentAutoSizeMaximumSize),
            WTFMove(*autoSizingShouldExpandToViewHeight),
            WTFMove(*viewportSizeForCSSViewportUnits),
            WTFMove(*scrollPinningBehavior),
            WTFMove(*scrollbarOverlayStyle),
            WTFMove(*backgroundExtendsBeyondPage),
            WTFMove(*layerHostingMode),
            WTFMove(*hasResourceLoadClient),
            WTFMove(*mimeTypesWithCustomContentProviders),
            WTFMove(*controlledByAutomation),
            WTFMove(*isProcessSwap),
            WTFMove(*useDarkAppearance),
            WTFMove(*useElevatedUserInterfaceLevel),
#if PLATFORM(MAC)
            WTFMove(*colorSpace),
#endif
#if PLATFORM(MAC)
            WTFMove(*useSystemAppearance),
#endif
#if PLATFORM(MAC)
            WTFMove(*useFormSemanticContext),
#endif
#if PLATFORM(MAC)
            WTFMove(*headerBannerHeight),
#endif
#if PLATFORM(MAC)
            WTFMove(*footerBannerHeight),
#endif
#if ENABLE(META_VIEWPORT)
            WTFMove(*ignoresViewportScaleLimits),
#endif
#if ENABLE(META_VIEWPORT)
            WTFMove(*viewportConfigurationViewLayoutSize),
#endif
#if ENABLE(META_VIEWPORT)
            WTFMove(*viewportConfigurationLayoutSizeScaleFactor),
#endif
#if ENABLE(META_VIEWPORT)
            WTFMove(*viewportConfigurationMinimumEffectiveDeviceWidth),
#endif
#if ENABLE(META_VIEWPORT)
            WTFMove(*viewportConfigurationViewSize),
#endif
#if ENABLE(META_VIEWPORT)
            WTFMove(*overrideViewportArguments),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*screenSize),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*availableScreenSize),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*overrideScreenSize),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*textAutosizingWidth),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*deviceOrientation),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*keyboardIsAttached),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*canShowWhileLocked),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*isCapturingScreen),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*insertionPointColor),
#endif
#if PLATFORM(COCOA)
            WTFMove(*smartInsertDeleteEnabled),
#endif
#if PLATFORM(COCOA)
            WTFMove(*additionalSupportedImageTypes),
#endif
#if PLATFORM(COCOA)
            WTFMove(*gpuIOKitExtensionHandles),
#endif
#if PLATFORM(COCOA)
            WTFMove(*gpuMachExtensionHandles),
#endif
#if HAVE(STATIC_FONT_REGISTRY)
            WTFMove(*fontMachExtensionHandles),
#endif
#if HAVE(APP_ACCENT_COLORS)
            WTFMove(*accentColor),
#endif
#if USE(WPE_RENDERER)
            WTFMove(*hostFileDescriptor),
#endif
#if USE(GRAPHICS_LAYER_TEXTURE_MAPPER) || USE(GRAPHICS_LAYER_WC)
            WTFMove(*nativeWindowHandle),
#endif
#if USE(GRAPHICS_LAYER_WC)
            WTFMove(*usesOffscreenRendering),
#endif
            WTFMove(*shouldScaleViewToFitDocument),
            WTFMove(*userInterfaceLayoutDirection),
            WTFMove(*observedLayoutMilestones),
            WTFMove(*overrideContentSecurityPolicy),
            WTFMove(*cpuLimit),
            WTFMove(*urlSchemeHandlers),
            WTFMove(*urlSchemesWithLegacyCustomProtocolHandlers),
#if ENABLE(APPLICATION_MANIFEST)
            WTFMove(*applicationManifest),
#endif
            WTFMove(*needsFontAttributes),
            WTFMove(*iceCandidateFilteringEnabled),
            WTFMove(*enumeratingAllNetworkInterfacesEnabled),
            WTFMove(*userContentControllerParameters),
#if ENABLE(WK_WEB_EXTENSIONS)
            WTFMove(*webExtensionControllerParameters),
#endif
            WTFMove(*backgroundColor),
            WTFMove(*oldPageID),
            WTFMove(*overriddenMediaType),
            WTFMove(*corsDisablingPatterns),
            WTFMove(*maskedURLSchemes),
            WTFMove(*userScriptsShouldWaitUntilNotification),
            WTFMove(*loadsSubresources),
            WTFMove(*allowedNetworkHosts),
            WTFMove(*crossOriginAccessControlCheckEnabled),
            WTFMove(*processDisplayName),
            WTFMove(*shouldCaptureAudioInUIProcess),
            WTFMove(*shouldCaptureAudioInGPUProcess),
            WTFMove(*shouldCaptureVideoInUIProcess),
            WTFMove(*shouldCaptureVideoInGPUProcess),
            WTFMove(*shouldCaptureDisplayInUIProcess),
            WTFMove(*shouldCaptureDisplayInGPUProcess),
            WTFMove(*shouldRenderCanvasInGPUProcess),
            WTFMove(*shouldRenderDOMInGPUProcess),
            WTFMove(*shouldPlayMediaInGPUProcess),
#if ENABLE(WEBGL)
            WTFMove(*shouldRenderWebGLInGPUProcess),
#endif
            WTFMove(*shouldEnableVP8Decoder),
            WTFMove(*shouldEnableVP9Decoder),
            WTFMove(*shouldEnableVP9SWDecoder),
#if ENABLE(APP_BOUND_DOMAINS)
            WTFMove(*limitsNavigationsToAppBoundDomains),
#endif
            WTFMove(*lastNavigationWasAppInitiated),
            WTFMove(*canUseCredentialStorage),
            WTFMove(*shouldRelaxThirdPartyCookieBlocking),
            WTFMove(*httpsUpgradeEnabled),
#if PLATFORM(IOS) || PLATFORM(VISION)
            WTFMove(*allowsDeprecatedSynchronousXMLHttpRequestDuringUnload),
#endif
#if ENABLE(APP_HIGHLIGHTS)
            WTFMove(*appHighlightsVisible),
#endif
#if HAVE(TOUCH_BAR)
            WTFMove(*requiresUserActionForEditingControlsManager),
#endif
            WTFMove(*hasResizableWindows),
            WTFMove(*contentSecurityPolicyModeForExtension),
            WTFMove(*subframeProcessFrameTreeCreationParameters),
            WTFMove(*openerFrameIdentifier),
            WTFMove(*mainFrameIdentifier),
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
            WTFMove(*linkDecorationFilteringData),
#endif
#if ENABLE(ADVANCED_PRIVACY_PROTECTIONS)
            WTFMove(*allowedQueryParametersForAdvancedPrivacyProtections),
#endif
#if HAVE(MACH_BOOTSTRAP_EXTENSION)
            WTFMove(*machBootstrapHandle)
#endif
        }
    };
}

void ArgumentCoder<WebKit::WebPopupItem>::encode(Encoder& encoder, const WebKit::WebPopupItem& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_type)>, WebKit::WebPopupItem::Type>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_text)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_textDirection)>, WebCore::TextDirection>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_hasTextDirectionOverride)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_toolTip)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_accessibilityText)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isLabel)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_isSelected)>, bool>);
    struct ShouldBeSameSizeAsWebPopupItem : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebPopupItem>, false> {
        WebKit::WebPopupItem::Type m_type;
        String m_text;
        WebCore::TextDirection m_textDirection;
        bool m_hasTextDirectionOverride;
        String m_toolTip;
        String m_accessibilityText;
        bool m_isEnabled;
        bool m_isLabel;
        bool m_isSelected;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebPopupItem) == sizeof(WebKit::WebPopupItem));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebPopupItem, m_type)
        , offsetof(WebKit::WebPopupItem, m_text)
        , offsetof(WebKit::WebPopupItem, m_textDirection)
        , offsetof(WebKit::WebPopupItem, m_hasTextDirectionOverride)
        , offsetof(WebKit::WebPopupItem, m_toolTip)
        , offsetof(WebKit::WebPopupItem, m_accessibilityText)
        , offsetof(WebKit::WebPopupItem, m_isEnabled)
        , offsetof(WebKit::WebPopupItem, m_isLabel)
        , offsetof(WebKit::WebPopupItem, m_isSelected)
    >::value);
    encoder << instance.m_type;
    encoder << instance.m_text;
    encoder << instance.m_textDirection;
    encoder << instance.m_hasTextDirectionOverride;
    encoder << instance.m_toolTip;
    encoder << instance.m_accessibilityText;
    encoder << instance.m_isEnabled;
    encoder << instance.m_isLabel;
    encoder << instance.m_isSelected;
}

std::optional<WebKit::WebPopupItem> ArgumentCoder<WebKit::WebPopupItem>::decode(Decoder& decoder)
{
    auto m_type = decoder.decode<WebKit::WebPopupItem::Type>();
    auto m_text = decoder.decode<String>();
    auto m_textDirection = decoder.decode<WebCore::TextDirection>();
    auto m_hasTextDirectionOverride = decoder.decode<bool>();
    auto m_toolTip = decoder.decode<String>();
    auto m_accessibilityText = decoder.decode<String>();
    auto m_isEnabled = decoder.decode<bool>();
    auto m_isLabel = decoder.decode<bool>();
    auto m_isSelected = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebPopupItem {
            WTFMove(*m_type),
            WTFMove(*m_text),
            WTFMove(*m_textDirection),
            WTFMove(*m_hasTextDirectionOverride),
            WTFMove(*m_toolTip),
            WTFMove(*m_accessibilityText),
            WTFMove(*m_isEnabled),
            WTFMove(*m_isLabel),
            WTFMove(*m_isSelected)
        }
    };
}

void ArgumentCoder<WebKit::WebProcessCreationParameters>::encode(Encoder& encoder, WebKit::WebProcessCreationParameters&& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.auxiliaryProcessParameters)>, WebKit::AuxiliaryProcessCreationParameters>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.injectedBundlePath)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.injectedBundlePathExtensionHandle)>, WebKit::SandboxExtension::Handle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.additionalSandboxExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.initializationUserData)>, WebKit::UserData>);
#if PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enableRemoteWebInspectorExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsEmptyDocument)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsSecure)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsBypassingContentSecurityPolicy)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesForWhichDomainRelaxationIsForbidden)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsLocal)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsNoAccess)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsDisplayIsolated)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsCORSEnabled)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsAlwaysRevalidated)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsCachePartitioned)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.urlSchemesRegisteredAsCanDisplayOnlyIfCanRequest)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fontAllowList)>, Vector<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.overrideLanguages)>, Vector<String>>);
#if USE(GSTREAMER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gstreamerOptions)>, Vector<String>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cacheModel)>, WebKit::CacheModel>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.defaultRequestTimeoutInterval)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.backForwardCacheCapacity)>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldAlwaysUseComplexTextCodePath)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldEnableMemoryPressureReliefLogging)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldSuppressMemoryPressureHandler)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.disableFontSubpixelAntialiasingForTesting)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fullKeyboardAccessEnabled)>, bool>);
#if HAVE(MOUSE_DEVICE_OBSERVATION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasMouseDevice)>, bool>);
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasStylusDevice)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.memoryCacheDisabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.attrStyleEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldThrowExceptionForGlobalConstantRedeclaration)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.crossOriginMode)>, WebCore::CrossOriginMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isLockdownModeEnabled)>, bool>);
#if ENABLE(SERVICE_CONTROLS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasImageServices)>, bool>);
#endif
#if ENABLE(SERVICE_CONTROLS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasSelectionServices)>, bool>);
#endif
#if ENABLE(SERVICE_CONTROLS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasRichContentServices)>, bool>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.textCheckerState)>, WebKit::TextCheckerState>);
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uiProcessBundleIdentifier)>, String>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.latencyQOS)>, int>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.throughputQOS)>, int>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentingApplicationBundleIdentifier)>, String>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.presentingApplicationPID)>, ProcessID>);
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.acceleratedCompositingPort)>, WTF::MachSendRight>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uiProcessBundleResourcePath)>, String>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uiProcessBundleResourcePathExtensionHandle)>, WebKit::SandboxExtension::Handle>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldEnableJIT)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldEnableFTLJIT)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.accessibilityEnhancedUserInterfaceEnabled)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bundleParameterData)>, RefPtr<API::Data>>);
#endif
#if ENABLE(NOTIFICATIONS)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.notificationPermissions)>, HashMap<String, bool>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkATSContext)>, RetainPtr<CFDataRef>>);
#endif
#if PLATFORM(WAYLAND)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.waylandCompositorDisplayName)>, String>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaMIMETypes)>, Vector<String>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.screenProperties)>, WebCore::ScreenProperties>);
#endif
#if ENABLE(TRACKING_PREVENTION) && !RELEASE_LOG_DISABLED
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldLogUserInteraction)>, bool>);
#endif
#if PLATFORM(MAC)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useOverlayScrollbars)>, bool>);
#endif
#if USE(WPE_RENDERER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isServiceWorkerProcess)>, bool>);
#endif
#if USE(WPE_RENDERER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hostClientFileDescriptor)>, UnixFileDescriptor>);
#endif
#if USE(WPE_RENDERER)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.implementationLibraryName)>, CString>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.websiteDataStoreParameters)>, std::optional<WebKit::WebProcessDataStoreParameters>>);
#if PLATFORM(IOS) || PLATFORM(VISION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compilerServiceExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mobileGestaltExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.launchServicesExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#if PLATFORM(MAC) && HAVE(VIDEO_RESTRICTED_DECODING)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trustdExtensionHandle)>, WebKit::SandboxExtension::Handle>);
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enableDecodingHEIC)>, bool>);
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.enableDecodingAVIF)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dynamicIOKitExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(VISION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.metalCacheDirectoryExtensionHandles)>, Vector<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.systemHasBattery)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.systemHasAC)>, bool>);
#endif
#if PLATFORM(COCOA)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.strictSecureDecodingForAllObjCEnabled)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.currentUserInterfaceIdiom)>, PAL::UserInterfaceIdiom>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsPictureInPicture)>, bool>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cssValueToSystemColorMap)>, WebCore::RenderThemeIOS::CSSValueToSystemColorMap>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.focusRingColor)>, WebCore::Color>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localizedDeviceModel)>, String>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentSizeCategory)>, String>);
#endif
#if USE(GBM)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.renderDeviceFile)>, String>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dmaBufRendererBufferMode)>, OptionSet<WebKit::DMABufRendererBufferMode>>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useSystemAppearanceForScrollbars)>, bool>);
#endif
#if PLATFORM(GTK)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.gtkSettings)>, WebKit::GtkSettingsState>);
#endif
#if HAVE(CATALYST_USER_INTERFACE_IDIOM_AND_SCALE_FACTOR)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.overrideUserInterfaceIdiomAndScale)>, std::pair<int64_t, double>>);
#endif
#if HAVE(IOSURFACE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumIOSurfaceSize)>, WebCore::IntSize>);
#endif
#if HAVE(IOSURFACE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bytesPerRowIOSurfaceAlignment)>, size_t>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.accessibilityPreferences)>, WebKit::AccessibilityPreferences>);
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationAccessibilityEnabled)>, bool>);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.memoryPressureHandlerConfiguration)>, std::optional<WTF::MemoryPressureHandler::Configuration>>);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.disableFontHintingForTesting)>, bool>);
#endif
#if USE(GLIB)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationID)>, String>);
#endif
#if USE(GLIB)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationName)>, String>);
#endif
#if USE(GLIB) && ENABLE(REMOTE_INSPECTOR)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.inspectorServerAddress)>, CString>);
#endif
#if USE(ATSPI)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.accessibilityBusAddress)>, String>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timeZoneOverride)>, String>);
    struct ShouldBeSameSizeAsWebProcessCreationParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebProcessCreationParameters>, false> {
        WebKit::AuxiliaryProcessCreationParameters auxiliaryProcessParameters;
        String injectedBundlePath;
        WebKit::SandboxExtension::Handle injectedBundlePathExtensionHandle;
        Vector<WebKit::SandboxExtension::Handle> additionalSandboxExtensionHandles;
        WebKit::UserData initializationUserData;
#if PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR)
        Vector<WebKit::SandboxExtension::Handle> enableRemoteWebInspectorExtensionHandles;
#endif
        Vector<String> urlSchemesRegisteredAsEmptyDocument;
        Vector<String> urlSchemesRegisteredAsSecure;
        Vector<String> urlSchemesRegisteredAsBypassingContentSecurityPolicy;
        Vector<String> urlSchemesForWhichDomainRelaxationIsForbidden;
        Vector<String> urlSchemesRegisteredAsLocal;
        Vector<String> urlSchemesRegisteredAsNoAccess;
        Vector<String> urlSchemesRegisteredAsDisplayIsolated;
        Vector<String> urlSchemesRegisteredAsCORSEnabled;
        Vector<String> urlSchemesRegisteredAsAlwaysRevalidated;
        Vector<String> urlSchemesRegisteredAsCachePartitioned;
        Vector<String> urlSchemesRegisteredAsCanDisplayOnlyIfCanRequest;
        Vector<String> fontAllowList;
        Vector<String> overrideLanguages;
#if USE(GSTREAMER)
        Vector<String> gstreamerOptions;
#endif
        WebKit::CacheModel cacheModel;
        double defaultRequestTimeoutInterval;
        unsigned backForwardCacheCapacity;
        bool shouldAlwaysUseComplexTextCodePath;
        bool shouldEnableMemoryPressureReliefLogging;
        bool shouldSuppressMemoryPressureHandler;
        bool disableFontSubpixelAntialiasingForTesting;
        bool fullKeyboardAccessEnabled;
#if HAVE(MOUSE_DEVICE_OBSERVATION)
        bool hasMouseDevice;
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
        bool hasStylusDevice;
#endif
        bool memoryCacheDisabled;
        bool attrStyleEnabled;
        bool shouldThrowExceptionForGlobalConstantRedeclaration;
        WebCore::CrossOriginMode crossOriginMode;
        bool isLockdownModeEnabled;
#if ENABLE(SERVICE_CONTROLS)
        bool hasImageServices;
#endif
#if ENABLE(SERVICE_CONTROLS)
        bool hasSelectionServices;
#endif
#if ENABLE(SERVICE_CONTROLS)
        bool hasRichContentServices;
#endif
        WebKit::TextCheckerState textCheckerState;
#if PLATFORM(COCOA)
        String uiProcessBundleIdentifier;
#endif
#if PLATFORM(COCOA)
        int latencyQOS;
#endif
#if PLATFORM(COCOA)
        int throughputQOS;
#endif
#if PLATFORM(COCOA)
        String presentingApplicationBundleIdentifier;
#endif
        ProcessID presentingApplicationPID;
#if PLATFORM(COCOA)
        WTF::MachSendRight acceleratedCompositingPort;
#endif
#if PLATFORM(COCOA)
        String uiProcessBundleResourcePath;
#endif
#if PLATFORM(COCOA)
        WebKit::SandboxExtension::Handle uiProcessBundleResourcePathExtensionHandle;
#endif
#if PLATFORM(COCOA)
        bool shouldEnableJIT;
#endif
#if PLATFORM(COCOA)
        bool shouldEnableFTLJIT;
#endif
#if PLATFORM(COCOA)
        bool accessibilityEnhancedUserInterfaceEnabled;
#endif
#if PLATFORM(COCOA)
        RefPtr<API::Data> bundleParameterData;
#endif
#if ENABLE(NOTIFICATIONS)
        HashMap<String, bool> notificationPermissions;
#endif
#if PLATFORM(COCOA)
        RetainPtr<CFDataRef> networkATSContext;
#endif
#if PLATFORM(WAYLAND)
        String waylandCompositorDisplayName;
#endif
#if PLATFORM(COCOA)
        Vector<String> mediaMIMETypes;
#endif
#if PLATFORM(COCOA)
        WebCore::ScreenProperties screenProperties;
#endif
#if ENABLE(TRACKING_PREVENTION) && !RELEASE_LOG_DISABLED
        bool shouldLogUserInteraction;
#endif
#if PLATFORM(MAC)
        bool useOverlayScrollbars;
#endif
#if USE(WPE_RENDERER)
        bool isServiceWorkerProcess;
#endif
#if USE(WPE_RENDERER)
        UnixFileDescriptor hostClientFileDescriptor;
#endif
#if USE(WPE_RENDERER)
        CString implementationLibraryName;
#endif
        std::optional<WebKit::WebProcessDataStoreParameters> websiteDataStoreParameters;
#if PLATFORM(IOS) || PLATFORM(VISION)
        Vector<WebKit::SandboxExtension::Handle> compilerServiceExtensionHandles;
#endif
        std::optional<WebKit::SandboxExtension::Handle> mobileGestaltExtensionHandle;
        std::optional<WebKit::SandboxExtension::Handle> launchServicesExtensionHandle;
#if PLATFORM(MAC) && HAVE(VIDEO_RESTRICTED_DECODING)
        WebKit::SandboxExtension::Handle trustdExtensionHandle;
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
        bool enableDecodingHEIC;
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
        bool enableDecodingAVIF;
#endif
#if PLATFORM(IOS_FAMILY)
        Vector<WebKit::SandboxExtension::Handle> dynamicIOKitExtensionHandles;
#endif
#if PLATFORM(VISION)
        Vector<WebKit::SandboxExtension::Handle> metalCacheDirectoryExtensionHandles;
#endif
#if PLATFORM(COCOA)
        bool systemHasBattery;
#endif
#if PLATFORM(COCOA)
        bool systemHasAC;
#endif
#if PLATFORM(COCOA)
        bool strictSecureDecodingForAllObjCEnabled;
#endif
#if PLATFORM(IOS_FAMILY)
        PAL::UserInterfaceIdiom currentUserInterfaceIdiom;
#endif
#if PLATFORM(IOS_FAMILY)
        bool supportsPictureInPicture;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::RenderThemeIOS::CSSValueToSystemColorMap cssValueToSystemColorMap;
#endif
#if PLATFORM(IOS_FAMILY)
        WebCore::Color focusRingColor;
#endif
#if PLATFORM(IOS_FAMILY)
        String localizedDeviceModel;
#endif
#if PLATFORM(IOS_FAMILY)
        String contentSizeCategory;
#endif
#if USE(GBM)
        String renderDeviceFile;
#endif
#if PLATFORM(GTK)
        OptionSet<WebKit::DMABufRendererBufferMode> dmaBufRendererBufferMode;
#endif
#if PLATFORM(GTK)
        bool useSystemAppearanceForScrollbars;
#endif
#if PLATFORM(GTK)
        WebKit::GtkSettingsState gtkSettings;
#endif
#if HAVE(CATALYST_USER_INTERFACE_IDIOM_AND_SCALE_FACTOR)
        std::pair<int64_t, double> overrideUserInterfaceIdiomAndScale;
#endif
#if HAVE(IOSURFACE)
        WebCore::IntSize maximumIOSurfaceSize;
#endif
#if HAVE(IOSURFACE)
        size_t bytesPerRowIOSurfaceAlignment;
#endif
        WebKit::AccessibilityPreferences accessibilityPreferences;
#if PLATFORM(IOS_FAMILY)
        bool applicationAccessibilityEnabled;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        std::optional<WTF::MemoryPressureHandler::Configuration> memoryPressureHandlerConfiguration;
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        bool disableFontHintingForTesting;
#endif
#if USE(GLIB)
        String applicationID;
#endif
#if USE(GLIB)
        String applicationName;
#endif
#if USE(GLIB) && ENABLE(REMOTE_INSPECTOR)
        CString inspectorServerAddress;
#endif
#if USE(ATSPI)
        String accessibilityBusAddress;
#endif
        String timeZoneOverride;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebProcessCreationParameters) == sizeof(WebKit::WebProcessCreationParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebProcessCreationParameters, auxiliaryProcessParameters)
        , offsetof(WebKit::WebProcessCreationParameters, injectedBundlePath)
        , offsetof(WebKit::WebProcessCreationParameters, injectedBundlePathExtensionHandle)
        , offsetof(WebKit::WebProcessCreationParameters, additionalSandboxExtensionHandles)
        , offsetof(WebKit::WebProcessCreationParameters, initializationUserData)
#if PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR)
        , offsetof(WebKit::WebProcessCreationParameters, enableRemoteWebInspectorExtensionHandles)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsEmptyDocument)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsSecure)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsBypassingContentSecurityPolicy)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesForWhichDomainRelaxationIsForbidden)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsLocal)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsNoAccess)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsDisplayIsolated)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsCORSEnabled)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsAlwaysRevalidated)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsCachePartitioned)
        , offsetof(WebKit::WebProcessCreationParameters, urlSchemesRegisteredAsCanDisplayOnlyIfCanRequest)
        , offsetof(WebKit::WebProcessCreationParameters, fontAllowList)
        , offsetof(WebKit::WebProcessCreationParameters, overrideLanguages)
#if USE(GSTREAMER)
        , offsetof(WebKit::WebProcessCreationParameters, gstreamerOptions)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, cacheModel)
        , offsetof(WebKit::WebProcessCreationParameters, defaultRequestTimeoutInterval)
        , offsetof(WebKit::WebProcessCreationParameters, backForwardCacheCapacity)
        , offsetof(WebKit::WebProcessCreationParameters, shouldAlwaysUseComplexTextCodePath)
        , offsetof(WebKit::WebProcessCreationParameters, shouldEnableMemoryPressureReliefLogging)
        , offsetof(WebKit::WebProcessCreationParameters, shouldSuppressMemoryPressureHandler)
        , offsetof(WebKit::WebProcessCreationParameters, disableFontSubpixelAntialiasingForTesting)
        , offsetof(WebKit::WebProcessCreationParameters, fullKeyboardAccessEnabled)
#if HAVE(MOUSE_DEVICE_OBSERVATION)
        , offsetof(WebKit::WebProcessCreationParameters, hasMouseDevice)
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
        , offsetof(WebKit::WebProcessCreationParameters, hasStylusDevice)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, memoryCacheDisabled)
        , offsetof(WebKit::WebProcessCreationParameters, attrStyleEnabled)
        , offsetof(WebKit::WebProcessCreationParameters, shouldThrowExceptionForGlobalConstantRedeclaration)
        , offsetof(WebKit::WebProcessCreationParameters, crossOriginMode)
        , offsetof(WebKit::WebProcessCreationParameters, isLockdownModeEnabled)
#if ENABLE(SERVICE_CONTROLS)
        , offsetof(WebKit::WebProcessCreationParameters, hasImageServices)
#endif
#if ENABLE(SERVICE_CONTROLS)
        , offsetof(WebKit::WebProcessCreationParameters, hasSelectionServices)
#endif
#if ENABLE(SERVICE_CONTROLS)
        , offsetof(WebKit::WebProcessCreationParameters, hasRichContentServices)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, textCheckerState)
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, uiProcessBundleIdentifier)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, latencyQOS)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, throughputQOS)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, presentingApplicationBundleIdentifier)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, presentingApplicationPID)
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, acceleratedCompositingPort)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, uiProcessBundleResourcePath)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, uiProcessBundleResourcePathExtensionHandle)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, shouldEnableJIT)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, shouldEnableFTLJIT)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, accessibilityEnhancedUserInterfaceEnabled)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, bundleParameterData)
#endif
#if ENABLE(NOTIFICATIONS)
        , offsetof(WebKit::WebProcessCreationParameters, notificationPermissions)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, networkATSContext)
#endif
#if PLATFORM(WAYLAND)
        , offsetof(WebKit::WebProcessCreationParameters, waylandCompositorDisplayName)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, mediaMIMETypes)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, screenProperties)
#endif
#if ENABLE(TRACKING_PREVENTION) && !RELEASE_LOG_DISABLED
        , offsetof(WebKit::WebProcessCreationParameters, shouldLogUserInteraction)
#endif
#if PLATFORM(MAC)
        , offsetof(WebKit::WebProcessCreationParameters, useOverlayScrollbars)
#endif
#if USE(WPE_RENDERER)
        , offsetof(WebKit::WebProcessCreationParameters, isServiceWorkerProcess)
#endif
#if USE(WPE_RENDERER)
        , offsetof(WebKit::WebProcessCreationParameters, hostClientFileDescriptor)
#endif
#if USE(WPE_RENDERER)
        , offsetof(WebKit::WebProcessCreationParameters, implementationLibraryName)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, websiteDataStoreParameters)
#if PLATFORM(IOS) || PLATFORM(VISION)
        , offsetof(WebKit::WebProcessCreationParameters, compilerServiceExtensionHandles)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, mobileGestaltExtensionHandle)
        , offsetof(WebKit::WebProcessCreationParameters, launchServicesExtensionHandle)
#if PLATFORM(MAC) && HAVE(VIDEO_RESTRICTED_DECODING)
        , offsetof(WebKit::WebProcessCreationParameters, trustdExtensionHandle)
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
        , offsetof(WebKit::WebProcessCreationParameters, enableDecodingHEIC)
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
        , offsetof(WebKit::WebProcessCreationParameters, enableDecodingAVIF)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, dynamicIOKitExtensionHandles)
#endif
#if PLATFORM(VISION)
        , offsetof(WebKit::WebProcessCreationParameters, metalCacheDirectoryExtensionHandles)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, systemHasBattery)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, systemHasAC)
#endif
#if PLATFORM(COCOA)
        , offsetof(WebKit::WebProcessCreationParameters, strictSecureDecodingForAllObjCEnabled)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, currentUserInterfaceIdiom)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, supportsPictureInPicture)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, cssValueToSystemColorMap)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, focusRingColor)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, localizedDeviceModel)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, contentSizeCategory)
#endif
#if USE(GBM)
        , offsetof(WebKit::WebProcessCreationParameters, renderDeviceFile)
#endif
#if PLATFORM(GTK)
        , offsetof(WebKit::WebProcessCreationParameters, dmaBufRendererBufferMode)
#endif
#if PLATFORM(GTK)
        , offsetof(WebKit::WebProcessCreationParameters, useSystemAppearanceForScrollbars)
#endif
#if PLATFORM(GTK)
        , offsetof(WebKit::WebProcessCreationParameters, gtkSettings)
#endif
#if HAVE(CATALYST_USER_INTERFACE_IDIOM_AND_SCALE_FACTOR)
        , offsetof(WebKit::WebProcessCreationParameters, overrideUserInterfaceIdiomAndScale)
#endif
#if HAVE(IOSURFACE)
        , offsetof(WebKit::WebProcessCreationParameters, maximumIOSurfaceSize)
#endif
#if HAVE(IOSURFACE)
        , offsetof(WebKit::WebProcessCreationParameters, bytesPerRowIOSurfaceAlignment)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, accessibilityPreferences)
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessCreationParameters, applicationAccessibilityEnabled)
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        , offsetof(WebKit::WebProcessCreationParameters, memoryPressureHandlerConfiguration)
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
        , offsetof(WebKit::WebProcessCreationParameters, disableFontHintingForTesting)
#endif
#if USE(GLIB)
        , offsetof(WebKit::WebProcessCreationParameters, applicationID)
#endif
#if USE(GLIB)
        , offsetof(WebKit::WebProcessCreationParameters, applicationName)
#endif
#if USE(GLIB) && ENABLE(REMOTE_INSPECTOR)
        , offsetof(WebKit::WebProcessCreationParameters, inspectorServerAddress)
#endif
#if USE(ATSPI)
        , offsetof(WebKit::WebProcessCreationParameters, accessibilityBusAddress)
#endif
        , offsetof(WebKit::WebProcessCreationParameters, timeZoneOverride)
    >::value);
    encoder << WTFMove(instance.auxiliaryProcessParameters);
    encoder << WTFMove(instance.injectedBundlePath);
    encoder << WTFMove(instance.injectedBundlePathExtensionHandle);
    encoder << WTFMove(instance.additionalSandboxExtensionHandles);
    encoder << WTFMove(instance.initializationUserData);
#if PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR)
    encoder << WTFMove(instance.enableRemoteWebInspectorExtensionHandles);
#endif
    encoder << WTFMove(instance.urlSchemesRegisteredAsEmptyDocument);
    encoder << WTFMove(instance.urlSchemesRegisteredAsSecure);
    encoder << WTFMove(instance.urlSchemesRegisteredAsBypassingContentSecurityPolicy);
    encoder << WTFMove(instance.urlSchemesForWhichDomainRelaxationIsForbidden);
    encoder << WTFMove(instance.urlSchemesRegisteredAsLocal);
    encoder << WTFMove(instance.urlSchemesRegisteredAsNoAccess);
    encoder << WTFMove(instance.urlSchemesRegisteredAsDisplayIsolated);
    encoder << WTFMove(instance.urlSchemesRegisteredAsCORSEnabled);
    encoder << WTFMove(instance.urlSchemesRegisteredAsAlwaysRevalidated);
    encoder << WTFMove(instance.urlSchemesRegisteredAsCachePartitioned);
    encoder << WTFMove(instance.urlSchemesRegisteredAsCanDisplayOnlyIfCanRequest);
    encoder << WTFMove(instance.fontAllowList);
    encoder << WTFMove(instance.overrideLanguages);
#if USE(GSTREAMER)
    encoder << WTFMove(instance.gstreamerOptions);
#endif
    encoder << WTFMove(instance.cacheModel);
    encoder << WTFMove(instance.defaultRequestTimeoutInterval);
    encoder << WTFMove(instance.backForwardCacheCapacity);
    encoder << WTFMove(instance.shouldAlwaysUseComplexTextCodePath);
    encoder << WTFMove(instance.shouldEnableMemoryPressureReliefLogging);
    encoder << WTFMove(instance.shouldSuppressMemoryPressureHandler);
    encoder << WTFMove(instance.disableFontSubpixelAntialiasingForTesting);
    encoder << WTFMove(instance.fullKeyboardAccessEnabled);
#if HAVE(MOUSE_DEVICE_OBSERVATION)
    encoder << WTFMove(instance.hasMouseDevice);
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
    encoder << WTFMove(instance.hasStylusDevice);
#endif
    encoder << WTFMove(instance.memoryCacheDisabled);
    encoder << WTFMove(instance.attrStyleEnabled);
    encoder << WTFMove(instance.shouldThrowExceptionForGlobalConstantRedeclaration);
    encoder << WTFMove(instance.crossOriginMode);
    encoder << WTFMove(instance.isLockdownModeEnabled);
#if ENABLE(SERVICE_CONTROLS)
    encoder << WTFMove(instance.hasImageServices);
#endif
#if ENABLE(SERVICE_CONTROLS)
    encoder << WTFMove(instance.hasSelectionServices);
#endif
#if ENABLE(SERVICE_CONTROLS)
    encoder << WTFMove(instance.hasRichContentServices);
#endif
    encoder << WTFMove(instance.textCheckerState);
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.uiProcessBundleIdentifier);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.latencyQOS);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.throughputQOS);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.presentingApplicationBundleIdentifier);
#endif
    encoder << WTFMove(instance.presentingApplicationPID);
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.acceleratedCompositingPort);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.uiProcessBundleResourcePath);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.uiProcessBundleResourcePathExtensionHandle);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.shouldEnableJIT);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.shouldEnableFTLJIT);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.accessibilityEnhancedUserInterfaceEnabled);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.bundleParameterData);
#endif
#if ENABLE(NOTIFICATIONS)
    encoder << WTFMove(instance.notificationPermissions);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.networkATSContext);
#endif
#if PLATFORM(WAYLAND)
    encoder << WTFMove(instance.waylandCompositorDisplayName);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.mediaMIMETypes);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.screenProperties);
#endif
#if ENABLE(TRACKING_PREVENTION) && !RELEASE_LOG_DISABLED
    encoder << WTFMove(instance.shouldLogUserInteraction);
#endif
#if PLATFORM(MAC)
    encoder << WTFMove(instance.useOverlayScrollbars);
#endif
#if USE(WPE_RENDERER)
    encoder << WTFMove(instance.isServiceWorkerProcess);
#endif
#if USE(WPE_RENDERER)
    encoder << WTFMove(instance.hostClientFileDescriptor);
#endif
#if USE(WPE_RENDERER)
    encoder << WTFMove(instance.implementationLibraryName);
#endif
    encoder << WTFMove(instance.websiteDataStoreParameters);
#if PLATFORM(IOS) || PLATFORM(VISION)
    encoder << WTFMove(instance.compilerServiceExtensionHandles);
#endif
    encoder << WTFMove(instance.mobileGestaltExtensionHandle);
    encoder << WTFMove(instance.launchServicesExtensionHandle);
#if PLATFORM(MAC) && HAVE(VIDEO_RESTRICTED_DECODING)
    encoder << WTFMove(instance.trustdExtensionHandle);
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
    encoder << WTFMove(instance.enableDecodingHEIC);
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
    encoder << WTFMove(instance.enableDecodingAVIF);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.dynamicIOKitExtensionHandles);
#endif
#if PLATFORM(VISION)
    encoder << WTFMove(instance.metalCacheDirectoryExtensionHandles);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.systemHasBattery);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.systemHasAC);
#endif
#if PLATFORM(COCOA)
    encoder << WTFMove(instance.strictSecureDecodingForAllObjCEnabled);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.currentUserInterfaceIdiom);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.supportsPictureInPicture);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.cssValueToSystemColorMap);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.focusRingColor);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.localizedDeviceModel);
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.contentSizeCategory);
#endif
#if USE(GBM)
    encoder << WTFMove(instance.renderDeviceFile);
#endif
#if PLATFORM(GTK)
    encoder << WTFMove(instance.dmaBufRendererBufferMode);
#endif
#if PLATFORM(GTK)
    encoder << WTFMove(instance.useSystemAppearanceForScrollbars);
#endif
#if PLATFORM(GTK)
    encoder << WTFMove(instance.gtkSettings);
#endif
#if HAVE(CATALYST_USER_INTERFACE_IDIOM_AND_SCALE_FACTOR)
    encoder << WTFMove(instance.overrideUserInterfaceIdiomAndScale);
#endif
#if HAVE(IOSURFACE)
    encoder << WTFMove(instance.maximumIOSurfaceSize);
#endif
#if HAVE(IOSURFACE)
    encoder << WTFMove(instance.bytesPerRowIOSurfaceAlignment);
#endif
    encoder << WTFMove(instance.accessibilityPreferences);
#if PLATFORM(IOS_FAMILY)
    encoder << WTFMove(instance.applicationAccessibilityEnabled);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    encoder << WTFMove(instance.memoryPressureHandlerConfiguration);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    encoder << WTFMove(instance.disableFontHintingForTesting);
#endif
#if USE(GLIB)
    encoder << WTFMove(instance.applicationID);
#endif
#if USE(GLIB)
    encoder << WTFMove(instance.applicationName);
#endif
#if USE(GLIB) && ENABLE(REMOTE_INSPECTOR)
    encoder << WTFMove(instance.inspectorServerAddress);
#endif
#if USE(ATSPI)
    encoder << WTFMove(instance.accessibilityBusAddress);
#endif
    encoder << WTFMove(instance.timeZoneOverride);
}

std::optional<WebKit::WebProcessCreationParameters> ArgumentCoder<WebKit::WebProcessCreationParameters>::decode(Decoder& decoder)
{
    auto auxiliaryProcessParameters = decoder.decode<WebKit::AuxiliaryProcessCreationParameters>();
    auto injectedBundlePath = decoder.decode<String>();
    auto injectedBundlePathExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
    auto additionalSandboxExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
    auto initializationUserData = decoder.decode<WebKit::UserData>();
#if PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR)
    auto enableRemoteWebInspectorExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
#endif
    auto urlSchemesRegisteredAsEmptyDocument = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsSecure = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsBypassingContentSecurityPolicy = decoder.decode<Vector<String>>();
    auto urlSchemesForWhichDomainRelaxationIsForbidden = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsLocal = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsNoAccess = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsDisplayIsolated = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsCORSEnabled = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsAlwaysRevalidated = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsCachePartitioned = decoder.decode<Vector<String>>();
    auto urlSchemesRegisteredAsCanDisplayOnlyIfCanRequest = decoder.decode<Vector<String>>();
    auto fontAllowList = decoder.decode<Vector<String>>();
    auto overrideLanguages = decoder.decode<Vector<String>>();
#if USE(GSTREAMER)
    auto gstreamerOptions = decoder.decode<Vector<String>>();
#endif
    auto cacheModel = decoder.decode<WebKit::CacheModel>();
    auto defaultRequestTimeoutInterval = decoder.decode<double>();
    auto backForwardCacheCapacity = decoder.decode<unsigned>();
    auto shouldAlwaysUseComplexTextCodePath = decoder.decode<bool>();
    auto shouldEnableMemoryPressureReliefLogging = decoder.decode<bool>();
    auto shouldSuppressMemoryPressureHandler = decoder.decode<bool>();
    auto disableFontSubpixelAntialiasingForTesting = decoder.decode<bool>();
    auto fullKeyboardAccessEnabled = decoder.decode<bool>();
#if HAVE(MOUSE_DEVICE_OBSERVATION)
    auto hasMouseDevice = decoder.decode<bool>();
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
    auto hasStylusDevice = decoder.decode<bool>();
#endif
    auto memoryCacheDisabled = decoder.decode<bool>();
    auto attrStyleEnabled = decoder.decode<bool>();
    auto shouldThrowExceptionForGlobalConstantRedeclaration = decoder.decode<bool>();
    auto crossOriginMode = decoder.decode<WebCore::CrossOriginMode>();
    auto isLockdownModeEnabled = decoder.decode<bool>();
#if ENABLE(SERVICE_CONTROLS)
    auto hasImageServices = decoder.decode<bool>();
#endif
#if ENABLE(SERVICE_CONTROLS)
    auto hasSelectionServices = decoder.decode<bool>();
#endif
#if ENABLE(SERVICE_CONTROLS)
    auto hasRichContentServices = decoder.decode<bool>();
#endif
    auto textCheckerState = decoder.decode<WebKit::TextCheckerState>();
#if PLATFORM(COCOA)
    auto uiProcessBundleIdentifier = decoder.decode<String>();
#endif
#if PLATFORM(COCOA)
    auto latencyQOS = decoder.decode<int>();
#endif
#if PLATFORM(COCOA)
    auto throughputQOS = decoder.decode<int>();
#endif
#if PLATFORM(COCOA)
    auto presentingApplicationBundleIdentifier = decoder.decode<String>();
#endif
    auto presentingApplicationPID = decoder.decode<ProcessID>();
#if PLATFORM(COCOA)
    auto acceleratedCompositingPort = decoder.decode<WTF::MachSendRight>();
#endif
#if PLATFORM(COCOA)
    auto uiProcessBundleResourcePath = decoder.decode<String>();
#endif
#if PLATFORM(COCOA)
    auto uiProcessBundleResourcePathExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
#endif
#if PLATFORM(COCOA)
    auto shouldEnableJIT = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto shouldEnableFTLJIT = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto accessibilityEnhancedUserInterfaceEnabled = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto bundleParameterData = decoder.decode<RefPtr<API::Data>>();
#endif
#if ENABLE(NOTIFICATIONS)
    auto notificationPermissions = decoder.decode<HashMap<String, bool>>();
#endif
#if PLATFORM(COCOA)
    auto networkATSContext = decoder.decode<RetainPtr<CFDataRef>>();
#endif
#if PLATFORM(WAYLAND)
    auto waylandCompositorDisplayName = decoder.decode<String>();
#endif
#if PLATFORM(COCOA)
    auto mediaMIMETypes = decoder.decode<Vector<String>>();
#endif
#if PLATFORM(COCOA)
    auto screenProperties = decoder.decode<WebCore::ScreenProperties>();
#endif
#if ENABLE(TRACKING_PREVENTION) && !RELEASE_LOG_DISABLED
    auto shouldLogUserInteraction = decoder.decode<bool>();
#endif
#if PLATFORM(MAC)
    auto useOverlayScrollbars = decoder.decode<bool>();
#endif
#if USE(WPE_RENDERER)
    auto isServiceWorkerProcess = decoder.decode<bool>();
#endif
#if USE(WPE_RENDERER)
    auto hostClientFileDescriptor = decoder.decode<UnixFileDescriptor>();
#endif
#if USE(WPE_RENDERER)
    auto implementationLibraryName = decoder.decode<CString>();
#endif
    auto websiteDataStoreParameters = decoder.decode<std::optional<WebKit::WebProcessDataStoreParameters>>();
#if PLATFORM(IOS) || PLATFORM(VISION)
    auto compilerServiceExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
#endif
    auto mobileGestaltExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
    auto launchServicesExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#if PLATFORM(MAC) && HAVE(VIDEO_RESTRICTED_DECODING)
    auto trustdExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
    auto enableDecodingHEIC = decoder.decode<bool>();
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
    auto enableDecodingAVIF = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto dynamicIOKitExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(VISION)
    auto metalCacheDirectoryExtensionHandles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(COCOA)
    auto systemHasBattery = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto systemHasAC = decoder.decode<bool>();
#endif
#if PLATFORM(COCOA)
    auto strictSecureDecodingForAllObjCEnabled = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto currentUserInterfaceIdiom = decoder.decode<PAL::UserInterfaceIdiom>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto supportsPictureInPicture = decoder.decode<bool>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto cssValueToSystemColorMap = decoder.decode<WebCore::RenderThemeIOS::CSSValueToSystemColorMap>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto focusRingColor = decoder.decode<WebCore::Color>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto localizedDeviceModel = decoder.decode<String>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto contentSizeCategory = decoder.decode<String>();
#endif
#if USE(GBM)
    auto renderDeviceFile = decoder.decode<String>();
#endif
#if PLATFORM(GTK)
    auto dmaBufRendererBufferMode = decoder.decode<OptionSet<WebKit::DMABufRendererBufferMode>>();
#endif
#if PLATFORM(GTK)
    auto useSystemAppearanceForScrollbars = decoder.decode<bool>();
#endif
#if PLATFORM(GTK)
    auto gtkSettings = decoder.decode<WebKit::GtkSettingsState>();
#endif
#if HAVE(CATALYST_USER_INTERFACE_IDIOM_AND_SCALE_FACTOR)
    auto overrideUserInterfaceIdiomAndScale = decoder.decode<std::pair<int64_t, double>>();
#endif
#if HAVE(IOSURFACE)
    auto maximumIOSurfaceSize = decoder.decode<WebCore::IntSize>();
#endif
#if HAVE(IOSURFACE)
    auto bytesPerRowIOSurfaceAlignment = decoder.decode<size_t>();
#endif
    auto accessibilityPreferences = decoder.decode<WebKit::AccessibilityPreferences>();
#if PLATFORM(IOS_FAMILY)
    auto applicationAccessibilityEnabled = decoder.decode<bool>();
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    auto memoryPressureHandlerConfiguration = decoder.decode<std::optional<WTF::MemoryPressureHandler::Configuration>>();
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    auto disableFontHintingForTesting = decoder.decode<bool>();
#endif
#if USE(GLIB)
    auto applicationID = decoder.decode<String>();
#endif
#if USE(GLIB)
    auto applicationName = decoder.decode<String>();
#endif
#if USE(GLIB) && ENABLE(REMOTE_INSPECTOR)
    auto inspectorServerAddress = decoder.decode<CString>();
#endif
#if USE(ATSPI)
    auto accessibilityBusAddress = decoder.decode<String>();
#endif
    auto timeZoneOverride = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebProcessCreationParameters {
            WTFMove(*auxiliaryProcessParameters),
            WTFMove(*injectedBundlePath),
            WTFMove(*injectedBundlePathExtensionHandle),
            WTFMove(*additionalSandboxExtensionHandles),
            WTFMove(*initializationUserData),
#if PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR)
            WTFMove(*enableRemoteWebInspectorExtensionHandles),
#endif
            WTFMove(*urlSchemesRegisteredAsEmptyDocument),
            WTFMove(*urlSchemesRegisteredAsSecure),
            WTFMove(*urlSchemesRegisteredAsBypassingContentSecurityPolicy),
            WTFMove(*urlSchemesForWhichDomainRelaxationIsForbidden),
            WTFMove(*urlSchemesRegisteredAsLocal),
            WTFMove(*urlSchemesRegisteredAsNoAccess),
            WTFMove(*urlSchemesRegisteredAsDisplayIsolated),
            WTFMove(*urlSchemesRegisteredAsCORSEnabled),
            WTFMove(*urlSchemesRegisteredAsAlwaysRevalidated),
            WTFMove(*urlSchemesRegisteredAsCachePartitioned),
            WTFMove(*urlSchemesRegisteredAsCanDisplayOnlyIfCanRequest),
            WTFMove(*fontAllowList),
            WTFMove(*overrideLanguages),
#if USE(GSTREAMER)
            WTFMove(*gstreamerOptions),
#endif
            WTFMove(*cacheModel),
            WTFMove(*defaultRequestTimeoutInterval),
            WTFMove(*backForwardCacheCapacity),
            WTFMove(*shouldAlwaysUseComplexTextCodePath),
            WTFMove(*shouldEnableMemoryPressureReliefLogging),
            WTFMove(*shouldSuppressMemoryPressureHandler),
            WTFMove(*disableFontSubpixelAntialiasingForTesting),
            WTFMove(*fullKeyboardAccessEnabled),
#if HAVE(MOUSE_DEVICE_OBSERVATION)
            WTFMove(*hasMouseDevice),
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
            WTFMove(*hasStylusDevice),
#endif
            WTFMove(*memoryCacheDisabled),
            WTFMove(*attrStyleEnabled),
            WTFMove(*shouldThrowExceptionForGlobalConstantRedeclaration),
            WTFMove(*crossOriginMode),
            WTFMove(*isLockdownModeEnabled),
#if ENABLE(SERVICE_CONTROLS)
            WTFMove(*hasImageServices),
#endif
#if ENABLE(SERVICE_CONTROLS)
            WTFMove(*hasSelectionServices),
#endif
#if ENABLE(SERVICE_CONTROLS)
            WTFMove(*hasRichContentServices),
#endif
            WTFMove(*textCheckerState),
#if PLATFORM(COCOA)
            WTFMove(*uiProcessBundleIdentifier),
#endif
#if PLATFORM(COCOA)
            WTFMove(*latencyQOS),
#endif
#if PLATFORM(COCOA)
            WTFMove(*throughputQOS),
#endif
#if PLATFORM(COCOA)
            WTFMove(*presentingApplicationBundleIdentifier),
#endif
            WTFMove(*presentingApplicationPID),
#if PLATFORM(COCOA)
            WTFMove(*acceleratedCompositingPort),
#endif
#if PLATFORM(COCOA)
            WTFMove(*uiProcessBundleResourcePath),
#endif
#if PLATFORM(COCOA)
            WTFMove(*uiProcessBundleResourcePathExtensionHandle),
#endif
#if PLATFORM(COCOA)
            WTFMove(*shouldEnableJIT),
#endif
#if PLATFORM(COCOA)
            WTFMove(*shouldEnableFTLJIT),
#endif
#if PLATFORM(COCOA)
            WTFMove(*accessibilityEnhancedUserInterfaceEnabled),
#endif
#if PLATFORM(COCOA)
            WTFMove(*bundleParameterData),
#endif
#if ENABLE(NOTIFICATIONS)
            WTFMove(*notificationPermissions),
#endif
#if PLATFORM(COCOA)
            WTFMove(*networkATSContext),
#endif
#if PLATFORM(WAYLAND)
            WTFMove(*waylandCompositorDisplayName),
#endif
#if PLATFORM(COCOA)
            WTFMove(*mediaMIMETypes),
#endif
#if PLATFORM(COCOA)
            WTFMove(*screenProperties),
#endif
#if ENABLE(TRACKING_PREVENTION) && !RELEASE_LOG_DISABLED
            WTFMove(*shouldLogUserInteraction),
#endif
#if PLATFORM(MAC)
            WTFMove(*useOverlayScrollbars),
#endif
#if USE(WPE_RENDERER)
            WTFMove(*isServiceWorkerProcess),
#endif
#if USE(WPE_RENDERER)
            WTFMove(*hostClientFileDescriptor),
#endif
#if USE(WPE_RENDERER)
            WTFMove(*implementationLibraryName),
#endif
            WTFMove(*websiteDataStoreParameters),
#if PLATFORM(IOS) || PLATFORM(VISION)
            WTFMove(*compilerServiceExtensionHandles),
#endif
            WTFMove(*mobileGestaltExtensionHandle),
            WTFMove(*launchServicesExtensionHandle),
#if PLATFORM(MAC) && HAVE(VIDEO_RESTRICTED_DECODING)
            WTFMove(*trustdExtensionHandle),
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
            WTFMove(*enableDecodingHEIC),
#endif
#if HAVE(VIDEO_RESTRICTED_DECODING)
            WTFMove(*enableDecodingAVIF),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*dynamicIOKitExtensionHandles),
#endif
#if PLATFORM(VISION)
            WTFMove(*metalCacheDirectoryExtensionHandles),
#endif
#if PLATFORM(COCOA)
            WTFMove(*systemHasBattery),
#endif
#if PLATFORM(COCOA)
            WTFMove(*systemHasAC),
#endif
#if PLATFORM(COCOA)
            WTFMove(*strictSecureDecodingForAllObjCEnabled),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*currentUserInterfaceIdiom),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*supportsPictureInPicture),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*cssValueToSystemColorMap),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*focusRingColor),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*localizedDeviceModel),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*contentSizeCategory),
#endif
#if USE(GBM)
            WTFMove(*renderDeviceFile),
#endif
#if PLATFORM(GTK)
            WTFMove(*dmaBufRendererBufferMode),
#endif
#if PLATFORM(GTK)
            WTFMove(*useSystemAppearanceForScrollbars),
#endif
#if PLATFORM(GTK)
            WTFMove(*gtkSettings),
#endif
#if HAVE(CATALYST_USER_INTERFACE_IDIOM_AND_SCALE_FACTOR)
            WTFMove(*overrideUserInterfaceIdiomAndScale),
#endif
#if HAVE(IOSURFACE)
            WTFMove(*maximumIOSurfaceSize),
#endif
#if HAVE(IOSURFACE)
            WTFMove(*bytesPerRowIOSurfaceAlignment),
#endif
            WTFMove(*accessibilityPreferences),
#if PLATFORM(IOS_FAMILY)
            WTFMove(*applicationAccessibilityEnabled),
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
            WTFMove(*memoryPressureHandlerConfiguration),
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
            WTFMove(*disableFontHintingForTesting),
#endif
#if USE(GLIB)
            WTFMove(*applicationID),
#endif
#if USE(GLIB)
            WTFMove(*applicationName),
#endif
#if USE(GLIB) && ENABLE(REMOTE_INSPECTOR)
            WTFMove(*inspectorServerAddress),
#endif
#if USE(ATSPI)
            WTFMove(*accessibilityBusAddress),
#endif
            WTFMove(*timeZoneOverride)
        }
    };
}

void ArgumentCoder<WebKit::WebProcessDataStoreParameters>::encode(Encoder& encoder, const WebKit::WebProcessDataStoreParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sessionID)>, PAL::SessionID>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationCacheDirectory)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationCacheDirectoryExtensionHandle)>, WebKit::SandboxExtension::Handle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.applicationCacheFlatFileSubdirectoryName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaCacheDirectory)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaCacheDirectoryExtensionHandle)>, WebKit::SandboxExtension::Handle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaKeyStorageDirectory)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaKeyStorageDirectoryExtensionHandle)>, WebKit::SandboxExtension::Handle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.javaScriptConfigurationDirectory)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.javaScriptConfigurationDirectoryExtensionHandle)>, WebKit::SandboxExtension::Handle>);
#if ENABLE(TRACKING_PREVENTION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.thirdPartyCookieBlockingMode)>, WebCore::ThirdPartyCookieBlockingMode>);
#endif
#if ENABLE(TRACKING_PREVENTION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.domainsWithUserInteraction)>, HashSet<WebCore::RegistrableDomain>>);
#endif
#if ENABLE(TRACKING_PREVENTION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.domainsWithStorageAccessQuirk)>, HashMap<WebCore::RegistrableDomain, WebCore::RegistrableDomain>>);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modelElementCacheDirectory)>, String>);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modelElementCacheDirectoryExtensionHandle)>, WebKit::SandboxExtension::Handle>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cookieStorageDirectoryExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containerCachesDirectoryExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containerTemporaryDirectoryExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.trackingPreventionEnabled)>, bool>);
    struct ShouldBeSameSizeAsWebProcessDataStoreParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebProcessDataStoreParameters>, false> {
        PAL::SessionID sessionID;
        String applicationCacheDirectory;
        WebKit::SandboxExtension::Handle applicationCacheDirectoryExtensionHandle;
        String applicationCacheFlatFileSubdirectoryName;
        String mediaCacheDirectory;
        WebKit::SandboxExtension::Handle mediaCacheDirectoryExtensionHandle;
        String mediaKeyStorageDirectory;
        WebKit::SandboxExtension::Handle mediaKeyStorageDirectoryExtensionHandle;
        String javaScriptConfigurationDirectory;
        WebKit::SandboxExtension::Handle javaScriptConfigurationDirectoryExtensionHandle;
#if ENABLE(TRACKING_PREVENTION)
        WebCore::ThirdPartyCookieBlockingMode thirdPartyCookieBlockingMode;
#endif
#if ENABLE(TRACKING_PREVENTION)
        HashSet<WebCore::RegistrableDomain> domainsWithUserInteraction;
#endif
#if ENABLE(TRACKING_PREVENTION)
        HashMap<WebCore::RegistrableDomain, WebCore::RegistrableDomain> domainsWithStorageAccessQuirk;
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
        String modelElementCacheDirectory;
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
        WebKit::SandboxExtension::Handle modelElementCacheDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
        std::optional<WebKit::SandboxExtension::Handle> cookieStorageDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
        std::optional<WebKit::SandboxExtension::Handle> containerCachesDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
        std::optional<WebKit::SandboxExtension::Handle> containerTemporaryDirectoryExtensionHandle;
#endif
        bool trackingPreventionEnabled;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebProcessDataStoreParameters) == sizeof(WebKit::WebProcessDataStoreParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebProcessDataStoreParameters, sessionID)
        , offsetof(WebKit::WebProcessDataStoreParameters, applicationCacheDirectory)
        , offsetof(WebKit::WebProcessDataStoreParameters, applicationCacheDirectoryExtensionHandle)
        , offsetof(WebKit::WebProcessDataStoreParameters, applicationCacheFlatFileSubdirectoryName)
        , offsetof(WebKit::WebProcessDataStoreParameters, mediaCacheDirectory)
        , offsetof(WebKit::WebProcessDataStoreParameters, mediaCacheDirectoryExtensionHandle)
        , offsetof(WebKit::WebProcessDataStoreParameters, mediaKeyStorageDirectory)
        , offsetof(WebKit::WebProcessDataStoreParameters, mediaKeyStorageDirectoryExtensionHandle)
        , offsetof(WebKit::WebProcessDataStoreParameters, javaScriptConfigurationDirectory)
        , offsetof(WebKit::WebProcessDataStoreParameters, javaScriptConfigurationDirectoryExtensionHandle)
#if ENABLE(TRACKING_PREVENTION)
        , offsetof(WebKit::WebProcessDataStoreParameters, thirdPartyCookieBlockingMode)
#endif
#if ENABLE(TRACKING_PREVENTION)
        , offsetof(WebKit::WebProcessDataStoreParameters, domainsWithUserInteraction)
#endif
#if ENABLE(TRACKING_PREVENTION)
        , offsetof(WebKit::WebProcessDataStoreParameters, domainsWithStorageAccessQuirk)
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
        , offsetof(WebKit::WebProcessDataStoreParameters, modelElementCacheDirectory)
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
        , offsetof(WebKit::WebProcessDataStoreParameters, modelElementCacheDirectoryExtensionHandle)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessDataStoreParameters, cookieStorageDirectoryExtensionHandle)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessDataStoreParameters, containerCachesDirectoryExtensionHandle)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebProcessDataStoreParameters, containerTemporaryDirectoryExtensionHandle)
#endif
        , offsetof(WebKit::WebProcessDataStoreParameters, trackingPreventionEnabled)
    >::value);
    encoder << instance.sessionID;
    encoder << instance.applicationCacheDirectory;
    encoder << instance.applicationCacheDirectoryExtensionHandle;
    encoder << instance.applicationCacheFlatFileSubdirectoryName;
    encoder << instance.mediaCacheDirectory;
    encoder << instance.mediaCacheDirectoryExtensionHandle;
    encoder << instance.mediaKeyStorageDirectory;
    encoder << instance.mediaKeyStorageDirectoryExtensionHandle;
    encoder << instance.javaScriptConfigurationDirectory;
    encoder << instance.javaScriptConfigurationDirectoryExtensionHandle;
#if ENABLE(TRACKING_PREVENTION)
    encoder << instance.thirdPartyCookieBlockingMode;
#endif
#if ENABLE(TRACKING_PREVENTION)
    encoder << instance.domainsWithUserInteraction;
#endif
#if ENABLE(TRACKING_PREVENTION)
    encoder << instance.domainsWithStorageAccessQuirk;
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    encoder << instance.modelElementCacheDirectory;
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    encoder << instance.modelElementCacheDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.cookieStorageDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.containerCachesDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.containerTemporaryDirectoryExtensionHandle;
#endif
    encoder << instance.trackingPreventionEnabled;
}

std::optional<WebKit::WebProcessDataStoreParameters> ArgumentCoder<WebKit::WebProcessDataStoreParameters>::decode(Decoder& decoder)
{
    auto sessionID = decoder.decode<PAL::SessionID>();
    auto applicationCacheDirectory = decoder.decode<String>();
    auto applicationCacheDirectoryExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
    auto applicationCacheFlatFileSubdirectoryName = decoder.decode<String>();
    auto mediaCacheDirectory = decoder.decode<String>();
    auto mediaCacheDirectoryExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
    auto mediaKeyStorageDirectory = decoder.decode<String>();
    auto mediaKeyStorageDirectoryExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
    auto javaScriptConfigurationDirectory = decoder.decode<String>();
    auto javaScriptConfigurationDirectoryExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
#if ENABLE(TRACKING_PREVENTION)
    auto thirdPartyCookieBlockingMode = decoder.decode<WebCore::ThirdPartyCookieBlockingMode>();
#endif
#if ENABLE(TRACKING_PREVENTION)
    auto domainsWithUserInteraction = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
#endif
#if ENABLE(TRACKING_PREVENTION)
    auto domainsWithStorageAccessQuirk = decoder.decode<HashMap<WebCore::RegistrableDomain, WebCore::RegistrableDomain>>();
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    auto modelElementCacheDirectory = decoder.decode<String>();
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    auto modelElementCacheDirectoryExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto cookieStorageDirectoryExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto containerCachesDirectoryExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto containerTemporaryDirectoryExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#endif
    auto trackingPreventionEnabled = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebProcessDataStoreParameters {
            WTFMove(*sessionID),
            WTFMove(*applicationCacheDirectory),
            WTFMove(*applicationCacheDirectoryExtensionHandle),
            WTFMove(*applicationCacheFlatFileSubdirectoryName),
            WTFMove(*mediaCacheDirectory),
            WTFMove(*mediaCacheDirectoryExtensionHandle),
            WTFMove(*mediaKeyStorageDirectory),
            WTFMove(*mediaKeyStorageDirectoryExtensionHandle),
            WTFMove(*javaScriptConfigurationDirectory),
            WTFMove(*javaScriptConfigurationDirectoryExtensionHandle),
#if ENABLE(TRACKING_PREVENTION)
            WTFMove(*thirdPartyCookieBlockingMode),
#endif
#if ENABLE(TRACKING_PREVENTION)
            WTFMove(*domainsWithUserInteraction),
#endif
#if ENABLE(TRACKING_PREVENTION)
            WTFMove(*domainsWithStorageAccessQuirk),
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
            WTFMove(*modelElementCacheDirectory),
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
            WTFMove(*modelElementCacheDirectoryExtensionHandle),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*cookieStorageDirectoryExtensionHandle),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*containerCachesDirectoryExtensionHandle),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*containerTemporaryDirectoryExtensionHandle),
#endif
            WTFMove(*trackingPreventionEnabled)
        }
    };
}

void ArgumentCoder<WebKit::WebPushD::WebPushDaemonConnectionConfiguration>::encode(Encoder& encoder, const WebKit::WebPushD::WebPushDaemonConnectionConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.useMockBundlesForTesting)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hostAppAuditTokenData)>, std::optional<Vector<uint8_t>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pushPartitionString)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataStoreIdentifier)>, std::optional<WTF::UUID>>);
    struct ShouldBeSameSizeAsWebPushDaemonConnectionConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebPushD::WebPushDaemonConnectionConfiguration>, false> {
        bool useMockBundlesForTesting;
        std::optional<Vector<uint8_t>> hostAppAuditTokenData;
        String pushPartitionString;
        std::optional<WTF::UUID> dataStoreIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebPushDaemonConnectionConfiguration) == sizeof(WebKit::WebPushD::WebPushDaemonConnectionConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebPushD::WebPushDaemonConnectionConfiguration, useMockBundlesForTesting)
        , offsetof(WebKit::WebPushD::WebPushDaemonConnectionConfiguration, hostAppAuditTokenData)
        , offsetof(WebKit::WebPushD::WebPushDaemonConnectionConfiguration, pushPartitionString)
        , offsetof(WebKit::WebPushD::WebPushDaemonConnectionConfiguration, dataStoreIdentifier)
    >::value);
    encoder << instance.useMockBundlesForTesting;
    encoder << instance.hostAppAuditTokenData;
    encoder << instance.pushPartitionString;
    encoder << instance.dataStoreIdentifier;
}

std::optional<WebKit::WebPushD::WebPushDaemonConnectionConfiguration> ArgumentCoder<WebKit::WebPushD::WebPushDaemonConnectionConfiguration>::decode(Decoder& decoder)
{
    auto useMockBundlesForTesting = decoder.decode<bool>();
    auto hostAppAuditTokenData = decoder.decode<std::optional<Vector<uint8_t>>>();
    auto pushPartitionString = decoder.decode<String>();
    auto dataStoreIdentifier = decoder.decode<std::optional<WTF::UUID>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebPushD::WebPushDaemonConnectionConfiguration {
            WTFMove(*useMockBundlesForTesting),
            WTFMove(*hostAppAuditTokenData),
            WTFMove(*pushPartitionString),
            WTFMove(*dataStoreIdentifier)
        }
    };
}

void ArgumentCoder<WebKit::WebPushMessage>::encode(Encoder& encoder, const WebKit::WebPushMessage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pushData)>, std::optional<Vector<uint8_t>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pushPartitionString)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrationURL)>, URL>);
    struct ShouldBeSameSizeAsWebPushMessage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebPushMessage>, false> {
        std::optional<Vector<uint8_t>> pushData;
        String pushPartitionString;
        URL registrationURL;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebPushMessage) == sizeof(WebKit::WebPushMessage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebPushMessage, pushData)
        , offsetof(WebKit::WebPushMessage, pushPartitionString)
        , offsetof(WebKit::WebPushMessage, registrationURL)
    >::value);
    encoder << instance.pushData;
    encoder << instance.pushPartitionString;
    encoder << instance.registrationURL;
}

std::optional<WebKit::WebPushMessage> ArgumentCoder<WebKit::WebPushMessage>::decode(Decoder& decoder)
{
    auto pushData = decoder.decode<std::optional<Vector<uint8_t>>>();
    auto pushPartitionString = decoder.decode<String>();
    auto registrationURL = decoder.decode<URL>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebPushMessage {
            WTFMove(*pushData),
            WTFMove(*pushPartitionString),
            WTFMove(*registrationURL)
        }
    };
}

void ArgumentCoder<WebKit::WebsiteDataStoreParameters>::encode(Encoder& encoder, const WebKit::WebsiteDataStoreParameters& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.uiProcessCookieStorageIdentifier)>, Vector<uint8_t>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cookieStoragePathExtensionHandle)>, WebKit::SandboxExtension::Handle>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkSessionParameters)>, WebKit::NetworkSessionCreationParameters>);
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cookieStorageDirectoryExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.containerCachesDirectoryExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.parentBundleDirectoryExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#endif
#if PLATFORM(IOS_FAMILY)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.tempDirectoryExtensionHandle)>, std::optional<WebKit::SandboxExtension::Handle>>);
#endif
    struct ShouldBeSameSizeAsWebsiteDataStoreParameters : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebsiteDataStoreParameters>, false> {
        Vector<uint8_t> uiProcessCookieStorageIdentifier;
        WebKit::SandboxExtension::Handle cookieStoragePathExtensionHandle;
        WebKit::NetworkSessionCreationParameters networkSessionParameters;
#if PLATFORM(IOS_FAMILY)
        std::optional<WebKit::SandboxExtension::Handle> cookieStorageDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
        std::optional<WebKit::SandboxExtension::Handle> containerCachesDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
        std::optional<WebKit::SandboxExtension::Handle> parentBundleDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
        std::optional<WebKit::SandboxExtension::Handle> tempDirectoryExtensionHandle;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebsiteDataStoreParameters) == sizeof(WebKit::WebsiteDataStoreParameters));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebsiteDataStoreParameters, uiProcessCookieStorageIdentifier)
        , offsetof(WebKit::WebsiteDataStoreParameters, cookieStoragePathExtensionHandle)
        , offsetof(WebKit::WebsiteDataStoreParameters, networkSessionParameters)
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebsiteDataStoreParameters, cookieStorageDirectoryExtensionHandle)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebsiteDataStoreParameters, containerCachesDirectoryExtensionHandle)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebsiteDataStoreParameters, parentBundleDirectoryExtensionHandle)
#endif
#if PLATFORM(IOS_FAMILY)
        , offsetof(WebKit::WebsiteDataStoreParameters, tempDirectoryExtensionHandle)
#endif
    >::value);
    encoder << instance.uiProcessCookieStorageIdentifier;
    encoder << instance.cookieStoragePathExtensionHandle;
    encoder << instance.networkSessionParameters;
#if PLATFORM(IOS_FAMILY)
    encoder << instance.cookieStorageDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.containerCachesDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.parentBundleDirectoryExtensionHandle;
#endif
#if PLATFORM(IOS_FAMILY)
    encoder << instance.tempDirectoryExtensionHandle;
#endif
}

std::optional<WebKit::WebsiteDataStoreParameters> ArgumentCoder<WebKit::WebsiteDataStoreParameters>::decode(Decoder& decoder)
{
    auto uiProcessCookieStorageIdentifier = decoder.decode<Vector<uint8_t>>();
    auto cookieStoragePathExtensionHandle = decoder.decode<WebKit::SandboxExtension::Handle>();
    auto networkSessionParameters = decoder.decode<WebKit::NetworkSessionCreationParameters>();
#if PLATFORM(IOS_FAMILY)
    auto cookieStorageDirectoryExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto containerCachesDirectoryExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto parentBundleDirectoryExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#endif
#if PLATFORM(IOS_FAMILY)
    auto tempDirectoryExtensionHandle = decoder.decode<std::optional<WebKit::SandboxExtension::Handle>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebsiteDataStoreParameters {
            WTFMove(*uiProcessCookieStorageIdentifier),
            WTFMove(*cookieStoragePathExtensionHandle),
            WTFMove(*networkSessionParameters),
#if PLATFORM(IOS_FAMILY)
            WTFMove(*cookieStorageDirectoryExtensionHandle),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*containerCachesDirectoryExtensionHandle),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*parentBundleDirectoryExtensionHandle),
#endif
#if PLATFORM(IOS_FAMILY)
            WTFMove(*tempDirectoryExtensionHandle)
#endif
        }
    };
}

void ArgumentCoder<WebKit::WebsitePoliciesData>::encode(Encoder& encoder, const WebKit::WebsitePoliciesData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.contentExtensionEnablement)>, WebCore::ContentExtensionEnablement>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.activeContentRuleListActionPatterns)>, HashMap<WTF::String, Vector<WTF::String>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowedAutoplayQuirks)>, OptionSet<WebKit::WebsiteAutoplayQuirk>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.autoplayPolicy)>, WebKit::WebsiteAutoplayPolicy>);
#if ENABLE(DEVICE_ORIENTATION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.deviceOrientationAndMotionAccessState)>, WebCore::DeviceOrientationOrMotionPermissionState>);
#endif
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.customHeaderFields)>, Vector<WebCore::CustomHeaderFields>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.popUpPolicy)>, WebKit::WebsitePopUpPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.customUserAgent)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.customUserAgentAsSiteSpecificQuirks)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.customNavigatorPlatform)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.metaViewportPolicy)>, WebKit::WebsiteMetaViewportPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaSourcePolicy)>, WebKit::WebsiteMediaSourcePolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.simulatedMouseEventsDispatchPolicy)>, WebKit::WebsiteSimulatedMouseEventsDispatchPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.legacyOverflowScrollingTouchPolicy)>, WebKit::WebsiteLegacyOverflowScrollingTouchPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowContentChangeObserverQuirk)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowsContentJavaScript)>, WebCore::AllowsContentJavaScript>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mouseEventPolicy)>, WebCore::MouseEventPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.modalContainerObservationPolicy)>, WebCore::ModalContainerObservationPolicy>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSchemePreference)>, WebCore::ColorSchemePreference>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.advancedPrivacyProtections)>, OptionSet<WebCore::AdvancedPrivacyProtections>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.idempotentModeAutosizingOnlyHonorsPercentages)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.allowPrivacyProxy)>, bool>);
    struct ShouldBeSameSizeAsWebsitePoliciesData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebsitePoliciesData>, false> {
        WebCore::ContentExtensionEnablement contentExtensionEnablement;
        HashMap<WTF::String, Vector<WTF::String>> activeContentRuleListActionPatterns;
        OptionSet<WebKit::WebsiteAutoplayQuirk> allowedAutoplayQuirks;
        WebKit::WebsiteAutoplayPolicy autoplayPolicy;
#if ENABLE(DEVICE_ORIENTATION)
        WebCore::DeviceOrientationOrMotionPermissionState deviceOrientationAndMotionAccessState;
#endif
        Vector<WebCore::CustomHeaderFields> customHeaderFields;
        WebKit::WebsitePopUpPolicy popUpPolicy;
        String customUserAgent;
        String customUserAgentAsSiteSpecificQuirks;
        String customNavigatorPlatform;
        WebKit::WebsiteMetaViewportPolicy metaViewportPolicy;
        WebKit::WebsiteMediaSourcePolicy mediaSourcePolicy;
        WebKit::WebsiteSimulatedMouseEventsDispatchPolicy simulatedMouseEventsDispatchPolicy;
        WebKit::WebsiteLegacyOverflowScrollingTouchPolicy legacyOverflowScrollingTouchPolicy;
        bool allowContentChangeObserverQuirk;
        WebCore::AllowsContentJavaScript allowsContentJavaScript;
        WebCore::MouseEventPolicy mouseEventPolicy;
        WebCore::ModalContainerObservationPolicy modalContainerObservationPolicy;
        WebCore::ColorSchemePreference colorSchemePreference;
        OptionSet<WebCore::AdvancedPrivacyProtections> advancedPrivacyProtections;
        bool idempotentModeAutosizingOnlyHonorsPercentages;
        bool allowPrivacyProxy;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebsitePoliciesData) == sizeof(WebKit::WebsitePoliciesData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebsitePoliciesData, contentExtensionEnablement)
        , offsetof(WebKit::WebsitePoliciesData, activeContentRuleListActionPatterns)
        , offsetof(WebKit::WebsitePoliciesData, allowedAutoplayQuirks)
        , offsetof(WebKit::WebsitePoliciesData, autoplayPolicy)
#if ENABLE(DEVICE_ORIENTATION)
        , offsetof(WebKit::WebsitePoliciesData, deviceOrientationAndMotionAccessState)
#endif
        , offsetof(WebKit::WebsitePoliciesData, customHeaderFields)
        , offsetof(WebKit::WebsitePoliciesData, popUpPolicy)
        , offsetof(WebKit::WebsitePoliciesData, customUserAgent)
        , offsetof(WebKit::WebsitePoliciesData, customUserAgentAsSiteSpecificQuirks)
        , offsetof(WebKit::WebsitePoliciesData, customNavigatorPlatform)
        , offsetof(WebKit::WebsitePoliciesData, metaViewportPolicy)
        , offsetof(WebKit::WebsitePoliciesData, mediaSourcePolicy)
        , offsetof(WebKit::WebsitePoliciesData, simulatedMouseEventsDispatchPolicy)
        , offsetof(WebKit::WebsitePoliciesData, legacyOverflowScrollingTouchPolicy)
        , offsetof(WebKit::WebsitePoliciesData, allowContentChangeObserverQuirk)
        , offsetof(WebKit::WebsitePoliciesData, allowsContentJavaScript)
        , offsetof(WebKit::WebsitePoliciesData, mouseEventPolicy)
        , offsetof(WebKit::WebsitePoliciesData, modalContainerObservationPolicy)
        , offsetof(WebKit::WebsitePoliciesData, colorSchemePreference)
        , offsetof(WebKit::WebsitePoliciesData, advancedPrivacyProtections)
        , offsetof(WebKit::WebsitePoliciesData, idempotentModeAutosizingOnlyHonorsPercentages)
        , offsetof(WebKit::WebsitePoliciesData, allowPrivacyProxy)
    >::value);
    encoder << instance.contentExtensionEnablement;
    encoder << instance.activeContentRuleListActionPatterns;
    encoder << instance.allowedAutoplayQuirks;
    encoder << instance.autoplayPolicy;
#if ENABLE(DEVICE_ORIENTATION)
    encoder << instance.deviceOrientationAndMotionAccessState;
#endif
    encoder << instance.customHeaderFields;
    encoder << instance.popUpPolicy;
    encoder << instance.customUserAgent;
    encoder << instance.customUserAgentAsSiteSpecificQuirks;
    encoder << instance.customNavigatorPlatform;
    encoder << instance.metaViewportPolicy;
    encoder << instance.mediaSourcePolicy;
    encoder << instance.simulatedMouseEventsDispatchPolicy;
    encoder << instance.legacyOverflowScrollingTouchPolicy;
    encoder << instance.allowContentChangeObserverQuirk;
    encoder << instance.allowsContentJavaScript;
    encoder << instance.mouseEventPolicy;
    encoder << instance.modalContainerObservationPolicy;
    encoder << instance.colorSchemePreference;
    encoder << instance.advancedPrivacyProtections;
    encoder << instance.idempotentModeAutosizingOnlyHonorsPercentages;
    encoder << instance.allowPrivacyProxy;
}

std::optional<WebKit::WebsitePoliciesData> ArgumentCoder<WebKit::WebsitePoliciesData>::decode(Decoder& decoder)
{
    auto contentExtensionEnablement = decoder.decode<WebCore::ContentExtensionEnablement>();
    auto activeContentRuleListActionPatterns = decoder.decode<HashMap<WTF::String, Vector<WTF::String>>>();
    auto allowedAutoplayQuirks = decoder.decode<OptionSet<WebKit::WebsiteAutoplayQuirk>>();
    auto autoplayPolicy = decoder.decode<WebKit::WebsiteAutoplayPolicy>();
#if ENABLE(DEVICE_ORIENTATION)
    auto deviceOrientationAndMotionAccessState = decoder.decode<WebCore::DeviceOrientationOrMotionPermissionState>();
#endif
    auto customHeaderFields = decoder.decode<Vector<WebCore::CustomHeaderFields>>();
    auto popUpPolicy = decoder.decode<WebKit::WebsitePopUpPolicy>();
    auto customUserAgent = decoder.decode<String>();
    auto customUserAgentAsSiteSpecificQuirks = decoder.decode<String>();
    auto customNavigatorPlatform = decoder.decode<String>();
    auto metaViewportPolicy = decoder.decode<WebKit::WebsiteMetaViewportPolicy>();
    auto mediaSourcePolicy = decoder.decode<WebKit::WebsiteMediaSourcePolicy>();
    auto simulatedMouseEventsDispatchPolicy = decoder.decode<WebKit::WebsiteSimulatedMouseEventsDispatchPolicy>();
    auto legacyOverflowScrollingTouchPolicy = decoder.decode<WebKit::WebsiteLegacyOverflowScrollingTouchPolicy>();
    auto allowContentChangeObserverQuirk = decoder.decode<bool>();
    auto allowsContentJavaScript = decoder.decode<WebCore::AllowsContentJavaScript>();
    auto mouseEventPolicy = decoder.decode<WebCore::MouseEventPolicy>();
    auto modalContainerObservationPolicy = decoder.decode<WebCore::ModalContainerObservationPolicy>();
    auto colorSchemePreference = decoder.decode<WebCore::ColorSchemePreference>();
    auto advancedPrivacyProtections = decoder.decode<OptionSet<WebCore::AdvancedPrivacyProtections>>();
    auto idempotentModeAutosizingOnlyHonorsPercentages = decoder.decode<bool>();
    auto allowPrivacyProxy = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebsitePoliciesData {
            WTFMove(*contentExtensionEnablement),
            WTFMove(*activeContentRuleListActionPatterns),
            WTFMove(*allowedAutoplayQuirks),
            WTFMove(*autoplayPolicy),
#if ENABLE(DEVICE_ORIENTATION)
            WTFMove(*deviceOrientationAndMotionAccessState),
#endif
            WTFMove(*customHeaderFields),
            WTFMove(*popUpPolicy),
            WTFMove(*customUserAgent),
            WTFMove(*customUserAgentAsSiteSpecificQuirks),
            WTFMove(*customNavigatorPlatform),
            WTFMove(*metaViewportPolicy),
            WTFMove(*mediaSourcePolicy),
            WTFMove(*simulatedMouseEventsDispatchPolicy),
            WTFMove(*legacyOverflowScrollingTouchPolicy),
            WTFMove(*allowContentChangeObserverQuirk),
            WTFMove(*allowsContentJavaScript),
            WTFMove(*mouseEventPolicy),
            WTFMove(*modalContainerObservationPolicy),
            WTFMove(*colorSchemePreference),
            WTFMove(*advancedPrivacyProtections),
            WTFMove(*idempotentModeAutosizingOnlyHonorsPercentages),
            WTFMove(*allowPrivacyProxy)
        }
    };
}

void ArgumentCoder<API::Data>::encode(Encoder& encoder, const API::Data& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dataReference())>, IPC::DataReference>);
    encoder << instance.dataReference();
}

std::optional<Ref<API::Data>> ArgumentCoder<API::Data>::decode(Decoder& decoder)
{
    auto dataReference = decoder.decode<IPC::DataReference>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::Data::create(
            WTFMove(*dataReference)
        )
    };
}

void ArgumentCoder<API::Error>::encode(Encoder& encoder, const API::Error& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.platformError())>, WebCore::ResourceError>);
    encoder << instance.platformError();
}

std::optional<Ref<API::Error>> ArgumentCoder<API::Error>::decode(Decoder& decoder)
{
    auto platformError = decoder.decode<WebCore::ResourceError>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::Error::create(
            WTFMove(*platformError)
        )
    };
}

void ArgumentCoder<API::FrameHandle>::encode(Encoder& encoder, const API::FrameHandle& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frameID())>, WebCore::FrameIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isAutoconverting())>, bool>);
    encoder << instance.frameID();
    encoder << instance.isAutoconverting();
}

std::optional<Ref<API::FrameHandle>> ArgumentCoder<API::FrameHandle>::decode(Decoder& decoder)
{
    auto frameID = decoder.decode<WebCore::FrameIdentifier>();
    auto isAutoconverting = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::FrameHandle::create(
            WTFMove(*frameID),
            WTFMove(*isAutoconverting)
        )
    };
}

void ArgumentCoder<API::Size>::encode(Encoder& encoder, const API::Size& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size().width)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size().height)>, double>);
    encoder << instance.size().width;
    encoder << instance.size().height;
}

std::optional<Ref<API::Size>> ArgumentCoder<API::Size>::decode(Decoder& decoder)
{
    auto sizewidth = decoder.decode<double>();
    auto sizeheight = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::Size::create(
            WTFMove(*sizewidth),
            WTFMove(*sizeheight)
        )
    };
}

void ArgumentCoder<API::Point>::encode(Encoder& encoder, const API::Point& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point().x)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.point().y)>, double>);
    encoder << instance.point().x;
    encoder << instance.point().y;
}

std::optional<Ref<API::Point>> ArgumentCoder<API::Point>::decode(Decoder& decoder)
{
    auto pointx = decoder.decode<double>();
    auto pointy = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::Point::create(
            WTFMove(*pointx),
            WTFMove(*pointy)
        )
    };
}

void ArgumentCoder<API::Rect>::encode(Encoder& encoder, const API::Rect& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect().origin.x)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect().origin.y)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect().size.width)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rect().size.height)>, double>);
    encoder << instance.rect().origin.x;
    encoder << instance.rect().origin.y;
    encoder << instance.rect().size.width;
    encoder << instance.rect().size.height;
}

std::optional<Ref<API::Rect>> ArgumentCoder<API::Rect>::decode(Decoder& decoder)
{
    auto rectoriginx = decoder.decode<double>();
    auto rectoriginy = decoder.decode<double>();
    auto rectsizewidth = decoder.decode<double>();
    auto rectsizeheight = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::Rect::create(
            WTFMove(*rectoriginx),
            WTFMove(*rectoriginy),
            WTFMove(*rectsizewidth),
            WTFMove(*rectsizeheight)
        )
    };
}

void ArgumentCoder<API::PageHandle>::encode(Encoder& encoder, const API::PageHandle& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pageProxyID())>, WebKit::WebPageProxyIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.webPageID())>, WebCore::PageIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isAutoconverting())>, bool>);
    encoder << instance.pageProxyID();
    encoder << instance.webPageID();
    encoder << instance.isAutoconverting();
}

std::optional<Ref<API::PageHandle>> ArgumentCoder<API::PageHandle>::decode(Decoder& decoder)
{
    auto pageProxyID = decoder.decode<WebKit::WebPageProxyIdentifier>();
    auto webPageID = decoder.decode<WebCore::PageIdentifier>();
    auto isAutoconverting = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::PageHandle::create(
            WTFMove(*pageProxyID),
            WTFMove(*webPageID),
            WTFMove(*isAutoconverting)
        )
    };
}

void ArgumentCoder<API::URL>::encode(Encoder& encoder, const API::URL& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.string())>, WTF::String>);
    encoder << instance.string();
}

std::optional<Ref<API::URL>> ArgumentCoder<API::URL>::decode(Decoder& decoder)
{
    auto string = decoder.decode<WTF::String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::URL::create(
            WTFMove(*string)
        )
    };
}

void ArgumentCoder<API::URLRequest>::encode(Encoder& encoder, const API::URLRequest& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resourceRequest())>, WebCore::ResourceRequest>);
    encoder << instance.resourceRequest();
}

std::optional<Ref<API::URLRequest>> ArgumentCoder<API::URLRequest>::decode(Decoder& decoder)
{
    auto resourceRequest = decoder.decode<WebCore::ResourceRequest>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::URLRequest::create(
            WTFMove(*resourceRequest)
        )
    };
}

void ArgumentCoder<API::URLResponse>::encode(Encoder& encoder, const API::URLResponse& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resourceResponse())>, WebCore::ResourceResponse>);
    encoder << instance.resourceResponse();
}

std::optional<Ref<API::URLResponse>> ArgumentCoder<API::URLResponse>::decode(Decoder& decoder)
{
    auto resourceResponse = decoder.decode<WebCore::ResourceResponse>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        API::URLResponse::create(
            WTFMove(*resourceResponse)
        )
    };
}

void ArgumentCoder<WebKit::WebIDBResult>::encode(Encoder& encoder, const WebKit::WebIDBResult& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_resultData)>, WebCore::IDBResultData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.m_handles)>, Vector<WebKit::SandboxExtension::Handle>>);
    struct ShouldBeSameSizeAsWebIDBResult : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebIDBResult>, false> {
        WebCore::IDBResultData m_resultData;
        Vector<WebKit::SandboxExtension::Handle> m_handles;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebIDBResult) == sizeof(WebKit::WebIDBResult));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebIDBResult, m_resultData)
        , offsetof(WebKit::WebIDBResult, m_handles)
    >::value);
    encoder << instance.m_resultData;
    encoder << instance.m_handles;
}

std::optional<WebKit::WebIDBResult> ArgumentCoder<WebKit::WebIDBResult>::decode(Decoder& decoder)
{
    auto m_resultData = decoder.decode<WebCore::IDBResultData>();
    auto m_handles = decoder.decode<Vector<WebKit::SandboxExtension::Handle>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebIDBResult {
            WTFMove(*m_resultData),
            WTFMove(*m_handles)
        }
    };
}

#if ENABLE(GAMEPAD)
void ArgumentCoder<WebKit::GamepadData>::encode(Encoder& encoder, const WebKit::GamepadData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.index())>, unsigned>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.id())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mapping())>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.axisValues())>, Vector<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buttonValues())>, Vector<double>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lastUpdateTime())>, MonotonicTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportedEffectTypes())>, HashSet<WebCore::GamepadHapticEffectType, IntHash<WebCore::GamepadHapticEffectType>, WTF::StrongEnumHashTraits<WebCore::GamepadHapticEffectType>>>);
    encoder << instance.index();
    encoder << instance.id();
    encoder << instance.mapping();
    encoder << instance.axisValues();
    encoder << instance.buttonValues();
    encoder << instance.lastUpdateTime();
    encoder << instance.supportedEffectTypes();
}

std::optional<WebKit::GamepadData> ArgumentCoder<WebKit::GamepadData>::decode(Decoder& decoder)
{
    auto index = decoder.decode<unsigned>();
    auto id = decoder.decode<String>();
    auto mapping = decoder.decode<String>();
    auto axisValues = decoder.decode<Vector<double>>();
    auto buttonValues = decoder.decode<Vector<double>>();
    auto lastUpdateTime = decoder.decode<MonotonicTime>();
    auto supportedEffectTypes = decoder.decode<HashSet<WebCore::GamepadHapticEffectType, IntHash<WebCore::GamepadHapticEffectType>, WTF::StrongEnumHashTraits<WebCore::GamepadHapticEffectType>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::GamepadData {
            WTFMove(*index),
            WTFMove(*id),
            WTFMove(*mapping),
            WTFMove(*axisValues),
            WTFMove(*buttonValues),
            WTFMove(*lastUpdateTime),
            WTFMove(*supportedEffectTypes)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BindGroupDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::BindGroupDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bindGroupLayout)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entries)>, Vector<WebKit::WebGPU::BindGroupEntry>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupDescriptor, bindGroupLayout)
        , offsetof(WebKit::WebGPU::BindGroupDescriptor, entries)
    >::value);
    encoder << instance.label;
    encoder << instance.bindGroupLayout;
    encoder << instance.entries;
}

void ArgumentCoder<WebKit::WebGPU::BindGroupDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BindGroupDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bindGroupLayout)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entries)>, Vector<WebKit::WebGPU::BindGroupEntry>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupDescriptor, bindGroupLayout)
        , offsetof(WebKit::WebGPU::BindGroupDescriptor, entries)
    >::value);
    encoder << instance.label;
    encoder << instance.bindGroupLayout;
    encoder << instance.entries;
}

std::optional<WebKit::WebGPU::BindGroupDescriptor> ArgumentCoder<WebKit::WebGPU::BindGroupDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto bindGroupLayout = decoder.decode<WebKit::WebGPUIdentifier>();
    auto entries = decoder.decode<Vector<WebKit::WebGPU::BindGroupEntry>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BindGroupDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*bindGroupLayout),
            WTFMove(*entries)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BindGroupEntry>::encode(Encoder& encoder, const WebKit::WebGPU::BindGroupEntry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.binding)>, WebCore::WebGPU::Index32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bufferBinding)>, WebKit::WebGPU::BufferBinding>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebKit::WebGPU::BindingResourceType>);
    struct ShouldBeSameSizeAsBindGroupEntry : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BindGroupEntry>, false> {
        WebCore::WebGPU::Index32 binding;
        WebKit::WebGPU::BufferBinding bufferBinding;
        WebKit::WebGPUIdentifier identifier;
        WebKit::WebGPU::BindingResourceType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBindGroupEntry) == sizeof(WebKit::WebGPU::BindGroupEntry));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupEntry, binding)
        , offsetof(WebKit::WebGPU::BindGroupEntry, bufferBinding)
        , offsetof(WebKit::WebGPU::BindGroupEntry, identifier)
        , offsetof(WebKit::WebGPU::BindGroupEntry, type)
    >::value);
    encoder << instance.binding;
    encoder << instance.bufferBinding;
    encoder << instance.identifier;
    encoder << instance.type;
}

void ArgumentCoder<WebKit::WebGPU::BindGroupEntry>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BindGroupEntry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.binding)>, WebCore::WebGPU::Index32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bufferBinding)>, WebKit::WebGPU::BufferBinding>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebKit::WebGPU::BindingResourceType>);
    struct ShouldBeSameSizeAsBindGroupEntry : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BindGroupEntry>, false> {
        WebCore::WebGPU::Index32 binding;
        WebKit::WebGPU::BufferBinding bufferBinding;
        WebKit::WebGPUIdentifier identifier;
        WebKit::WebGPU::BindingResourceType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBindGroupEntry) == sizeof(WebKit::WebGPU::BindGroupEntry));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupEntry, binding)
        , offsetof(WebKit::WebGPU::BindGroupEntry, bufferBinding)
        , offsetof(WebKit::WebGPU::BindGroupEntry, identifier)
        , offsetof(WebKit::WebGPU::BindGroupEntry, type)
    >::value);
    encoder << instance.binding;
    encoder << instance.bufferBinding;
    encoder << instance.identifier;
    encoder << instance.type;
}

std::optional<WebKit::WebGPU::BindGroupEntry> ArgumentCoder<WebKit::WebGPU::BindGroupEntry>::decode(Decoder& decoder)
{
    auto binding = decoder.decode<WebCore::WebGPU::Index32>();
    auto bufferBinding = decoder.decode<WebKit::WebGPU::BufferBinding>();
    auto identifier = decoder.decode<WebKit::WebGPUIdentifier>();
    auto type = decoder.decode<WebKit::WebGPU::BindingResourceType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BindGroupEntry {
            WTFMove(*binding),
            WTFMove(*bufferBinding),
            WTFMove(*identifier),
            WTFMove(*type)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BindGroupLayoutDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::BindGroupLayoutDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entries)>, Vector<WebKit::WebGPU::BindGroupLayoutEntry>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupLayoutDescriptor, entries)
    >::value);
    encoder << instance.label;
    encoder << instance.entries;
}

void ArgumentCoder<WebKit::WebGPU::BindGroupLayoutDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BindGroupLayoutDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entries)>, Vector<WebKit::WebGPU::BindGroupLayoutEntry>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupLayoutDescriptor, entries)
    >::value);
    encoder << instance.label;
    encoder << instance.entries;
}

std::optional<WebKit::WebGPU::BindGroupLayoutDescriptor> ArgumentCoder<WebKit::WebGPU::BindGroupLayoutDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto entries = decoder.decode<Vector<WebKit::WebGPU::BindGroupLayoutEntry>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BindGroupLayoutDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*entries)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BindGroupLayoutEntry>::encode(Encoder& encoder, const WebKit::WebGPU::BindGroupLayoutEntry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.binding)>, WebCore::WebGPU::Index32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.visibility)>, WebCore::WebGPU::ShaderStageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffer)>, std::optional<WebKit::WebGPU::BufferBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampler)>, std::optional<WebKit::WebGPU::SamplerBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.texture)>, std::optional<WebKit::WebGPU::TextureBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.storageTexture)>, std::optional<WebKit::WebGPU::StorageTextureBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.externalTexture)>, std::optional<WebKit::WebGPU::ExternalTextureBindingLayout>>);
    struct ShouldBeSameSizeAsBindGroupLayoutEntry : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BindGroupLayoutEntry>, false> {
        WebCore::WebGPU::Index32 binding;
        WebCore::WebGPU::ShaderStageFlags visibility;
        std::optional<WebKit::WebGPU::BufferBindingLayout> buffer;
        std::optional<WebKit::WebGPU::SamplerBindingLayout> sampler;
        std::optional<WebKit::WebGPU::TextureBindingLayout> texture;
        std::optional<WebKit::WebGPU::StorageTextureBindingLayout> storageTexture;
        std::optional<WebKit::WebGPU::ExternalTextureBindingLayout> externalTexture;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBindGroupLayoutEntry) == sizeof(WebKit::WebGPU::BindGroupLayoutEntry));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, binding)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, visibility)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, buffer)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, sampler)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, texture)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, storageTexture)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, externalTexture)
    >::value);
    encoder << instance.binding;
    encoder << instance.visibility;
    encoder << instance.buffer;
    encoder << instance.sampler;
    encoder << instance.texture;
    encoder << instance.storageTexture;
    encoder << instance.externalTexture;
}

void ArgumentCoder<WebKit::WebGPU::BindGroupLayoutEntry>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BindGroupLayoutEntry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.binding)>, WebCore::WebGPU::Index32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.visibility)>, WebCore::WebGPU::ShaderStageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffer)>, std::optional<WebKit::WebGPU::BufferBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampler)>, std::optional<WebKit::WebGPU::SamplerBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.texture)>, std::optional<WebKit::WebGPU::TextureBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.storageTexture)>, std::optional<WebKit::WebGPU::StorageTextureBindingLayout>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.externalTexture)>, std::optional<WebKit::WebGPU::ExternalTextureBindingLayout>>);
    struct ShouldBeSameSizeAsBindGroupLayoutEntry : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BindGroupLayoutEntry>, false> {
        WebCore::WebGPU::Index32 binding;
        WebCore::WebGPU::ShaderStageFlags visibility;
        std::optional<WebKit::WebGPU::BufferBindingLayout> buffer;
        std::optional<WebKit::WebGPU::SamplerBindingLayout> sampler;
        std::optional<WebKit::WebGPU::TextureBindingLayout> texture;
        std::optional<WebKit::WebGPU::StorageTextureBindingLayout> storageTexture;
        std::optional<WebKit::WebGPU::ExternalTextureBindingLayout> externalTexture;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBindGroupLayoutEntry) == sizeof(WebKit::WebGPU::BindGroupLayoutEntry));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, binding)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, visibility)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, buffer)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, sampler)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, texture)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, storageTexture)
        , offsetof(WebKit::WebGPU::BindGroupLayoutEntry, externalTexture)
    >::value);
    encoder << instance.binding;
    encoder << instance.visibility;
    encoder << instance.buffer;
    encoder << instance.sampler;
    encoder << instance.texture;
    encoder << instance.storageTexture;
    encoder << instance.externalTexture;
}

std::optional<WebKit::WebGPU::BindGroupLayoutEntry> ArgumentCoder<WebKit::WebGPU::BindGroupLayoutEntry>::decode(Decoder& decoder)
{
    auto binding = decoder.decode<WebCore::WebGPU::Index32>();
    auto visibility = decoder.decode<WebCore::WebGPU::ShaderStageFlags>();
    auto buffer = decoder.decode<std::optional<WebKit::WebGPU::BufferBindingLayout>>();
    auto sampler = decoder.decode<std::optional<WebKit::WebGPU::SamplerBindingLayout>>();
    auto texture = decoder.decode<std::optional<WebKit::WebGPU::TextureBindingLayout>>();
    auto storageTexture = decoder.decode<std::optional<WebKit::WebGPU::StorageTextureBindingLayout>>();
    auto externalTexture = decoder.decode<std::optional<WebKit::WebGPU::ExternalTextureBindingLayout>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BindGroupLayoutEntry {
            WTFMove(*binding),
            WTFMove(*visibility),
            WTFMove(*buffer),
            WTFMove(*sampler),
            WTFMove(*texture),
            WTFMove(*storageTexture),
            WTFMove(*externalTexture)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BlendComponent>::encode(Encoder& encoder, const WebKit::WebGPU::BlendComponent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.operation)>, WebCore::WebGPU::BlendOperation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.srcFactor)>, WebCore::WebGPU::BlendFactor>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dstFactor)>, WebCore::WebGPU::BlendFactor>);
    struct ShouldBeSameSizeAsBlendComponent : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BlendComponent>, false> {
        WebCore::WebGPU::BlendOperation operation;
        WebCore::WebGPU::BlendFactor srcFactor;
        WebCore::WebGPU::BlendFactor dstFactor;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBlendComponent) == sizeof(WebKit::WebGPU::BlendComponent));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BlendComponent, operation)
        , offsetof(WebKit::WebGPU::BlendComponent, srcFactor)
        , offsetof(WebKit::WebGPU::BlendComponent, dstFactor)
    >::value);
    encoder << instance.operation;
    encoder << instance.srcFactor;
    encoder << instance.dstFactor;
}

void ArgumentCoder<WebKit::WebGPU::BlendComponent>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BlendComponent& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.operation)>, WebCore::WebGPU::BlendOperation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.srcFactor)>, WebCore::WebGPU::BlendFactor>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dstFactor)>, WebCore::WebGPU::BlendFactor>);
    struct ShouldBeSameSizeAsBlendComponent : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BlendComponent>, false> {
        WebCore::WebGPU::BlendOperation operation;
        WebCore::WebGPU::BlendFactor srcFactor;
        WebCore::WebGPU::BlendFactor dstFactor;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBlendComponent) == sizeof(WebKit::WebGPU::BlendComponent));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BlendComponent, operation)
        , offsetof(WebKit::WebGPU::BlendComponent, srcFactor)
        , offsetof(WebKit::WebGPU::BlendComponent, dstFactor)
    >::value);
    encoder << instance.operation;
    encoder << instance.srcFactor;
    encoder << instance.dstFactor;
}

std::optional<WebKit::WebGPU::BlendComponent> ArgumentCoder<WebKit::WebGPU::BlendComponent>::decode(Decoder& decoder)
{
    auto operation = decoder.decode<WebCore::WebGPU::BlendOperation>();
    auto srcFactor = decoder.decode<WebCore::WebGPU::BlendFactor>();
    auto dstFactor = decoder.decode<WebCore::WebGPU::BlendFactor>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BlendComponent {
            WTFMove(*operation),
            WTFMove(*srcFactor),
            WTFMove(*dstFactor)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BlendState>::encode(Encoder& encoder, const WebKit::WebGPU::BlendState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebKit::WebGPU::BlendComponent>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alpha)>, WebKit::WebGPU::BlendComponent>);
    struct ShouldBeSameSizeAsBlendState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BlendState>, false> {
        WebKit::WebGPU::BlendComponent color;
        WebKit::WebGPU::BlendComponent alpha;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBlendState) == sizeof(WebKit::WebGPU::BlendState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BlendState, color)
        , offsetof(WebKit::WebGPU::BlendState, alpha)
    >::value);
    encoder << instance.color;
    encoder << instance.alpha;
}

void ArgumentCoder<WebKit::WebGPU::BlendState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BlendState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.color)>, WebKit::WebGPU::BlendComponent>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alpha)>, WebKit::WebGPU::BlendComponent>);
    struct ShouldBeSameSizeAsBlendState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BlendState>, false> {
        WebKit::WebGPU::BlendComponent color;
        WebKit::WebGPU::BlendComponent alpha;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBlendState) == sizeof(WebKit::WebGPU::BlendState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BlendState, color)
        , offsetof(WebKit::WebGPU::BlendState, alpha)
    >::value);
    encoder << instance.color;
    encoder << instance.alpha;
}

std::optional<WebKit::WebGPU::BlendState> ArgumentCoder<WebKit::WebGPU::BlendState>::decode(Decoder& decoder)
{
    auto color = decoder.decode<WebKit::WebGPU::BlendComponent>();
    auto alpha = decoder.decode<WebKit::WebGPU::BlendComponent>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BlendState {
            WTFMove(*color),
            WTFMove(*alpha)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BufferBinding>::encode(Encoder& encoder, const WebKit::WebGPU::BufferBinding& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffer)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, std::optional<WebCore::WebGPU::Size64>>);
    struct ShouldBeSameSizeAsBufferBinding : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BufferBinding>, false> {
        WebKit::WebGPUIdentifier buffer;
        WebCore::WebGPU::Size64 offset;
        std::optional<WebCore::WebGPU::Size64> size;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBufferBinding) == sizeof(WebKit::WebGPU::BufferBinding));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BufferBinding, buffer)
        , offsetof(WebKit::WebGPU::BufferBinding, offset)
        , offsetof(WebKit::WebGPU::BufferBinding, size)
    >::value);
    encoder << instance.buffer;
    encoder << instance.offset;
    encoder << instance.size;
}

void ArgumentCoder<WebKit::WebGPU::BufferBinding>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BufferBinding& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffer)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, std::optional<WebCore::WebGPU::Size64>>);
    struct ShouldBeSameSizeAsBufferBinding : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BufferBinding>, false> {
        WebKit::WebGPUIdentifier buffer;
        WebCore::WebGPU::Size64 offset;
        std::optional<WebCore::WebGPU::Size64> size;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBufferBinding) == sizeof(WebKit::WebGPU::BufferBinding));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BufferBinding, buffer)
        , offsetof(WebKit::WebGPU::BufferBinding, offset)
        , offsetof(WebKit::WebGPU::BufferBinding, size)
    >::value);
    encoder << instance.buffer;
    encoder << instance.offset;
    encoder << instance.size;
}

std::optional<WebKit::WebGPU::BufferBinding> ArgumentCoder<WebKit::WebGPU::BufferBinding>::decode(Decoder& decoder)
{
    auto buffer = decoder.decode<WebKit::WebGPUIdentifier>();
    auto offset = decoder.decode<WebCore::WebGPU::Size64>();
    auto size = decoder.decode<std::optional<WebCore::WebGPU::Size64>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BufferBinding {
            WTFMove(*buffer),
            WTFMove(*offset),
            WTFMove(*size)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BufferBindingLayout>::encode(Encoder& encoder, const WebKit::WebGPU::BufferBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::BufferBindingType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasDynamicOffset)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minBindingSize)>, WebCore::WebGPU::Size64>);
    struct ShouldBeSameSizeAsBufferBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BufferBindingLayout>, false> {
        WebCore::WebGPU::BufferBindingType type;
        bool hasDynamicOffset;
        WebCore::WebGPU::Size64 minBindingSize;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBufferBindingLayout) == sizeof(WebKit::WebGPU::BufferBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BufferBindingLayout, type)
        , offsetof(WebKit::WebGPU::BufferBindingLayout, hasDynamicOffset)
        , offsetof(WebKit::WebGPU::BufferBindingLayout, minBindingSize)
    >::value);
    encoder << instance.type;
    encoder << instance.hasDynamicOffset;
    encoder << instance.minBindingSize;
}

void ArgumentCoder<WebKit::WebGPU::BufferBindingLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BufferBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::BufferBindingType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasDynamicOffset)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minBindingSize)>, WebCore::WebGPU::Size64>);
    struct ShouldBeSameSizeAsBufferBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::BufferBindingLayout>, false> {
        WebCore::WebGPU::BufferBindingType type;
        bool hasDynamicOffset;
        WebCore::WebGPU::Size64 minBindingSize;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBufferBindingLayout) == sizeof(WebKit::WebGPU::BufferBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BufferBindingLayout, type)
        , offsetof(WebKit::WebGPU::BufferBindingLayout, hasDynamicOffset)
        , offsetof(WebKit::WebGPU::BufferBindingLayout, minBindingSize)
    >::value);
    encoder << instance.type;
    encoder << instance.hasDynamicOffset;
    encoder << instance.minBindingSize;
}

std::optional<WebKit::WebGPU::BufferBindingLayout> ArgumentCoder<WebKit::WebGPU::BufferBindingLayout>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::WebGPU::BufferBindingType>();
    auto hasDynamicOffset = decoder.decode<bool>();
    auto minBindingSize = decoder.decode<WebCore::WebGPU::Size64>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BufferBindingLayout {
            WTFMove(*type),
            WTFMove(*hasDynamicOffset),
            WTFMove(*minBindingSize)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::BufferDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::BufferDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usage)>, WebCore::WebGPU::BufferUsageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mappedAtCreation)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BufferDescriptor, size)
        , offsetof(WebKit::WebGPU::BufferDescriptor, usage)
        , offsetof(WebKit::WebGPU::BufferDescriptor, mappedAtCreation)
    >::value);
    encoder << instance.label;
    encoder << instance.size;
    encoder << instance.usage;
    encoder << instance.mappedAtCreation;
}

void ArgumentCoder<WebKit::WebGPU::BufferDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::BufferDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usage)>, WebCore::WebGPU::BufferUsageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mappedAtCreation)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::BufferDescriptor, size)
        , offsetof(WebKit::WebGPU::BufferDescriptor, usage)
        , offsetof(WebKit::WebGPU::BufferDescriptor, mappedAtCreation)
    >::value);
    encoder << instance.label;
    encoder << instance.size;
    encoder << instance.usage;
    encoder << instance.mappedAtCreation;
}

std::optional<WebKit::WebGPU::BufferDescriptor> ArgumentCoder<WebKit::WebGPU::BufferDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto size = decoder.decode<WebCore::WebGPU::Size64>();
    auto usage = decoder.decode<WebCore::WebGPU::BufferUsageFlags>();
    auto mappedAtCreation = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::BufferDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*size),
            WTFMove(*usage),
            WTFMove(*mappedAtCreation)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::CanvasConfiguration>::encode(Encoder& encoder, const WebKit::WebGPU::CanvasConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.device)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usage)>, WebCore::WebGPU::TextureUsageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewFormats)>, Vector<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::WebGPU::PredefinedColorSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositingAlphaMode)>, WebCore::WebGPU::CanvasCompositingAlphaMode>);
    struct ShouldBeSameSizeAsCanvasConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::CanvasConfiguration>, false> {
        WebKit::WebGPUIdentifier device;
        WebCore::WebGPU::TextureFormat format;
        WebCore::WebGPU::TextureUsageFlags usage;
        Vector<WebCore::WebGPU::TextureFormat> viewFormats;
        WebCore::WebGPU::PredefinedColorSpace colorSpace;
        WebCore::WebGPU::CanvasCompositingAlphaMode compositingAlphaMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCanvasConfiguration) == sizeof(WebKit::WebGPU::CanvasConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::CanvasConfiguration, device)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, format)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, usage)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, viewFormats)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, colorSpace)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, compositingAlphaMode)
    >::value);
    encoder << instance.device;
    encoder << instance.format;
    encoder << instance.usage;
    encoder << instance.viewFormats;
    encoder << instance.colorSpace;
    encoder << instance.compositingAlphaMode;
}

void ArgumentCoder<WebKit::WebGPU::CanvasConfiguration>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::CanvasConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.device)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usage)>, WebCore::WebGPU::TextureUsageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewFormats)>, Vector<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::WebGPU::PredefinedColorSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositingAlphaMode)>, WebCore::WebGPU::CanvasCompositingAlphaMode>);
    struct ShouldBeSameSizeAsCanvasConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::CanvasConfiguration>, false> {
        WebKit::WebGPUIdentifier device;
        WebCore::WebGPU::TextureFormat format;
        WebCore::WebGPU::TextureUsageFlags usage;
        Vector<WebCore::WebGPU::TextureFormat> viewFormats;
        WebCore::WebGPU::PredefinedColorSpace colorSpace;
        WebCore::WebGPU::CanvasCompositingAlphaMode compositingAlphaMode;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCanvasConfiguration) == sizeof(WebKit::WebGPU::CanvasConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::CanvasConfiguration, device)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, format)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, usage)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, viewFormats)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, colorSpace)
        , offsetof(WebKit::WebGPU::CanvasConfiguration, compositingAlphaMode)
    >::value);
    encoder << instance.device;
    encoder << instance.format;
    encoder << instance.usage;
    encoder << instance.viewFormats;
    encoder << instance.colorSpace;
    encoder << instance.compositingAlphaMode;
}

std::optional<WebKit::WebGPU::CanvasConfiguration> ArgumentCoder<WebKit::WebGPU::CanvasConfiguration>::decode(Decoder& decoder)
{
    auto device = decoder.decode<WebKit::WebGPUIdentifier>();
    auto format = decoder.decode<WebCore::WebGPU::TextureFormat>();
    auto usage = decoder.decode<WebCore::WebGPU::TextureUsageFlags>();
    auto viewFormats = decoder.decode<Vector<WebCore::WebGPU::TextureFormat>>();
    auto colorSpace = decoder.decode<WebCore::WebGPU::PredefinedColorSpace>();
    auto compositingAlphaMode = decoder.decode<WebCore::WebGPU::CanvasCompositingAlphaMode>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::CanvasConfiguration {
            WTFMove(*device),
            WTFMove(*format),
            WTFMove(*usage),
            WTFMove(*viewFormats),
            WTFMove(*colorSpace),
            WTFMove(*compositingAlphaMode)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ColorDict>::encode(Encoder& encoder, const WebKit::WebGPU::ColorDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.r)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.g)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.b)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.a)>, double>);
    struct ShouldBeSameSizeAsColorDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ColorDict>, false> {
        double r;
        double g;
        double b;
        double a;
    };
    static_assert(sizeof(ShouldBeSameSizeAsColorDict) == sizeof(WebKit::WebGPU::ColorDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ColorDict, r)
        , offsetof(WebKit::WebGPU::ColorDict, g)
        , offsetof(WebKit::WebGPU::ColorDict, b)
        , offsetof(WebKit::WebGPU::ColorDict, a)
    >::value);
    encoder << instance.r;
    encoder << instance.g;
    encoder << instance.b;
    encoder << instance.a;
}

void ArgumentCoder<WebKit::WebGPU::ColorDict>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ColorDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.r)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.g)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.b)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.a)>, double>);
    struct ShouldBeSameSizeAsColorDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ColorDict>, false> {
        double r;
        double g;
        double b;
        double a;
    };
    static_assert(sizeof(ShouldBeSameSizeAsColorDict) == sizeof(WebKit::WebGPU::ColorDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ColorDict, r)
        , offsetof(WebKit::WebGPU::ColorDict, g)
        , offsetof(WebKit::WebGPU::ColorDict, b)
        , offsetof(WebKit::WebGPU::ColorDict, a)
    >::value);
    encoder << instance.r;
    encoder << instance.g;
    encoder << instance.b;
    encoder << instance.a;
}

std::optional<WebKit::WebGPU::ColorDict> ArgumentCoder<WebKit::WebGPU::ColorDict>::decode(Decoder& decoder)
{
    auto r = decoder.decode<double>();
    auto g = decoder.decode<double>();
    auto b = decoder.decode<double>();
    auto a = decoder.decode<double>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ColorDict {
            WTFMove(*r),
            WTFMove(*g),
            WTFMove(*b),
            WTFMove(*a)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ColorTargetState>::encode(Encoder& encoder, const WebKit::WebGPU::ColorTargetState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blend)>, std::optional<WebKit::WebGPU::BlendState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.writeMask)>, WebCore::WebGPU::ColorWriteFlags>);
    struct ShouldBeSameSizeAsColorTargetState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ColorTargetState>, false> {
        WebCore::WebGPU::TextureFormat format;
        std::optional<WebKit::WebGPU::BlendState> blend;
        WebCore::WebGPU::ColorWriteFlags writeMask;
    };
    static_assert(sizeof(ShouldBeSameSizeAsColorTargetState) == sizeof(WebKit::WebGPU::ColorTargetState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ColorTargetState, format)
        , offsetof(WebKit::WebGPU::ColorTargetState, blend)
        , offsetof(WebKit::WebGPU::ColorTargetState, writeMask)
    >::value);
    encoder << instance.format;
    encoder << instance.blend;
    encoder << instance.writeMask;
}

void ArgumentCoder<WebKit::WebGPU::ColorTargetState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ColorTargetState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.blend)>, std::optional<WebKit::WebGPU::BlendState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.writeMask)>, WebCore::WebGPU::ColorWriteFlags>);
    struct ShouldBeSameSizeAsColorTargetState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ColorTargetState>, false> {
        WebCore::WebGPU::TextureFormat format;
        std::optional<WebKit::WebGPU::BlendState> blend;
        WebCore::WebGPU::ColorWriteFlags writeMask;
    };
    static_assert(sizeof(ShouldBeSameSizeAsColorTargetState) == sizeof(WebKit::WebGPU::ColorTargetState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ColorTargetState, format)
        , offsetof(WebKit::WebGPU::ColorTargetState, blend)
        , offsetof(WebKit::WebGPU::ColorTargetState, writeMask)
    >::value);
    encoder << instance.format;
    encoder << instance.blend;
    encoder << instance.writeMask;
}

std::optional<WebKit::WebGPU::ColorTargetState> ArgumentCoder<WebKit::WebGPU::ColorTargetState>::decode(Decoder& decoder)
{
    auto format = decoder.decode<WebCore::WebGPU::TextureFormat>();
    auto blend = decoder.decode<std::optional<WebKit::WebGPU::BlendState>>();
    auto writeMask = decoder.decode<WebCore::WebGPU::ColorWriteFlags>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ColorTargetState {
            WTFMove(*format),
            WTFMove(*blend),
            WTFMove(*writeMask)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::CommandBufferDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::CommandBufferDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(MembersInCorrectOrder < 0
    >::value);
    encoder << instance.label;
}

void ArgumentCoder<WebKit::WebGPU::CommandBufferDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::CommandBufferDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(MembersInCorrectOrder < 0
    >::value);
    encoder << instance.label;
}

std::optional<WebKit::WebGPU::CommandBufferDescriptor> ArgumentCoder<WebKit::WebGPU::CommandBufferDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::CommandBufferDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            }
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::CommandEncoderDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::CommandEncoderDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(MembersInCorrectOrder < 0
    >::value);
    encoder << instance.label;
}

void ArgumentCoder<WebKit::WebGPU::CommandEncoderDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::CommandEncoderDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(MembersInCorrectOrder < 0
    >::value);
    encoder << instance.label;
}

std::optional<WebKit::WebGPU::CommandEncoderDescriptor> ArgumentCoder<WebKit::WebGPU::CommandEncoderDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::CommandEncoderDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            }
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::CompilationMessage>::encode(Encoder& encoder, const WebKit::WebGPU::CompilationMessage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::CompilationMessageType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lineNum)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linePos)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.length)>, uint64_t>);
    struct ShouldBeSameSizeAsCompilationMessage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::CompilationMessage>, false> {
        String message;
        WebCore::WebGPU::CompilationMessageType type;
        uint64_t lineNum;
        uint64_t linePos;
        uint64_t offset;
        uint64_t length;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCompilationMessage) == sizeof(WebKit::WebGPU::CompilationMessage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::CompilationMessage, message)
        , offsetof(WebKit::WebGPU::CompilationMessage, type)
        , offsetof(WebKit::WebGPU::CompilationMessage, lineNum)
        , offsetof(WebKit::WebGPU::CompilationMessage, linePos)
        , offsetof(WebKit::WebGPU::CompilationMessage, offset)
        , offsetof(WebKit::WebGPU::CompilationMessage, length)
    >::value);
    encoder << instance.message;
    encoder << instance.type;
    encoder << instance.lineNum;
    encoder << instance.linePos;
    encoder << instance.offset;
    encoder << instance.length;
}

void ArgumentCoder<WebKit::WebGPU::CompilationMessage>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::CompilationMessage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::CompilationMessageType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lineNum)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.linePos)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.length)>, uint64_t>);
    struct ShouldBeSameSizeAsCompilationMessage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::CompilationMessage>, false> {
        String message;
        WebCore::WebGPU::CompilationMessageType type;
        uint64_t lineNum;
        uint64_t linePos;
        uint64_t offset;
        uint64_t length;
    };
    static_assert(sizeof(ShouldBeSameSizeAsCompilationMessage) == sizeof(WebKit::WebGPU::CompilationMessage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::CompilationMessage, message)
        , offsetof(WebKit::WebGPU::CompilationMessage, type)
        , offsetof(WebKit::WebGPU::CompilationMessage, lineNum)
        , offsetof(WebKit::WebGPU::CompilationMessage, linePos)
        , offsetof(WebKit::WebGPU::CompilationMessage, offset)
        , offsetof(WebKit::WebGPU::CompilationMessage, length)
    >::value);
    encoder << instance.message;
    encoder << instance.type;
    encoder << instance.lineNum;
    encoder << instance.linePos;
    encoder << instance.offset;
    encoder << instance.length;
}

std::optional<WebKit::WebGPU::CompilationMessage> ArgumentCoder<WebKit::WebGPU::CompilationMessage>::decode(Decoder& decoder)
{
    auto message = decoder.decode<String>();
    auto type = decoder.decode<WebCore::WebGPU::CompilationMessageType>();
    auto lineNum = decoder.decode<uint64_t>();
    auto linePos = decoder.decode<uint64_t>();
    auto offset = decoder.decode<uint64_t>();
    auto length = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::CompilationMessage {
            WTFMove(*message),
            WTFMove(*type),
            WTFMove(*lineNum),
            WTFMove(*linePos),
            WTFMove(*offset),
            WTFMove(*length)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ComputePassDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::ComputePassDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestampWrites)>, WebKit::WebGPU::ComputePassTimestampWrites>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ComputePassDescriptor, timestampWrites)
    >::value);
    encoder << instance.label;
    encoder << instance.timestampWrites;
}

void ArgumentCoder<WebKit::WebGPU::ComputePassDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ComputePassDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestampWrites)>, WebKit::WebGPU::ComputePassTimestampWrites>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ComputePassDescriptor, timestampWrites)
    >::value);
    encoder << instance.label;
    encoder << instance.timestampWrites;
}

std::optional<WebKit::WebGPU::ComputePassDescriptor> ArgumentCoder<WebKit::WebGPU::ComputePassDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto timestampWrites = decoder.decode<WebKit::WebGPU::ComputePassTimestampWrites>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ComputePassDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*timestampWrites)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ComputePassTimestampWrite>::encode(Encoder& encoder, const WebKit::WebGPU::ComputePassTimestampWrite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.querySet)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.queryIndex)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location)>, WebCore::WebGPU::ComputePassTimestampLocation>);
    struct ShouldBeSameSizeAsComputePassTimestampWrite : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ComputePassTimestampWrite>, false> {
        WebKit::WebGPUIdentifier querySet;
        WebCore::WebGPU::Size32 queryIndex;
        WebCore::WebGPU::ComputePassTimestampLocation location;
    };
    static_assert(sizeof(ShouldBeSameSizeAsComputePassTimestampWrite) == sizeof(WebKit::WebGPU::ComputePassTimestampWrite));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ComputePassTimestampWrite, querySet)
        , offsetof(WebKit::WebGPU::ComputePassTimestampWrite, queryIndex)
        , offsetof(WebKit::WebGPU::ComputePassTimestampWrite, location)
    >::value);
    encoder << instance.querySet;
    encoder << instance.queryIndex;
    encoder << instance.location;
}

void ArgumentCoder<WebKit::WebGPU::ComputePassTimestampWrite>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ComputePassTimestampWrite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.querySet)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.queryIndex)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location)>, WebCore::WebGPU::ComputePassTimestampLocation>);
    struct ShouldBeSameSizeAsComputePassTimestampWrite : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ComputePassTimestampWrite>, false> {
        WebKit::WebGPUIdentifier querySet;
        WebCore::WebGPU::Size32 queryIndex;
        WebCore::WebGPU::ComputePassTimestampLocation location;
    };
    static_assert(sizeof(ShouldBeSameSizeAsComputePassTimestampWrite) == sizeof(WebKit::WebGPU::ComputePassTimestampWrite));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ComputePassTimestampWrite, querySet)
        , offsetof(WebKit::WebGPU::ComputePassTimestampWrite, queryIndex)
        , offsetof(WebKit::WebGPU::ComputePassTimestampWrite, location)
    >::value);
    encoder << instance.querySet;
    encoder << instance.queryIndex;
    encoder << instance.location;
}

std::optional<WebKit::WebGPU::ComputePassTimestampWrite> ArgumentCoder<WebKit::WebGPU::ComputePassTimestampWrite>::decode(Decoder& decoder)
{
    auto querySet = decoder.decode<WebKit::WebGPUIdentifier>();
    auto queryIndex = decoder.decode<WebCore::WebGPU::Size32>();
    auto location = decoder.decode<WebCore::WebGPU::ComputePassTimestampLocation>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ComputePassTimestampWrite {
            WTFMove(*querySet),
            WTFMove(*queryIndex),
            WTFMove(*location)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ComputePipelineDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::ComputePipelineDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layout)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineDescriptorBase, layout)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compute)>, WebKit::WebGPU::ProgrammableStage>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ComputePipelineDescriptor, compute)
    >::value);
    encoder << instance.label;
    encoder << instance.layout;
    encoder << instance.compute;
}

void ArgumentCoder<WebKit::WebGPU::ComputePipelineDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ComputePipelineDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layout)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineDescriptorBase, layout)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compute)>, WebKit::WebGPU::ProgrammableStage>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ComputePipelineDescriptor, compute)
    >::value);
    encoder << instance.label;
    encoder << instance.layout;
    encoder << instance.compute;
}

std::optional<WebKit::WebGPU::ComputePipelineDescriptor> ArgumentCoder<WebKit::WebGPU::ComputePipelineDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto layout = decoder.decode<WebKit::WebGPUIdentifier>();
    auto compute = decoder.decode<WebKit::WebGPU::ProgrammableStage>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ComputePipelineDescriptor {
            WebKit::WebGPU::PipelineDescriptorBase {
                WebKit::WebGPU::ObjectDescriptorBase {
                    WTFMove(*label)
                },
                WTFMove(*layout)
            },
            WTFMove(*compute)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::DepthStencilState>::encode(Encoder& encoder, const WebKit::WebGPU::DepthStencilState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthWriteEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthCompare)>, WebCore::WebGPU::CompareFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilFront)>, WebKit::WebGPU::StencilFaceState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilBack)>, WebKit::WebGPU::StencilFaceState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilReadMask)>, std::optional<WebCore::WebGPU::StencilValue>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilWriteMask)>, std::optional<WebCore::WebGPU::StencilValue>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthBias)>, WebCore::WebGPU::DepthBias>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthBiasSlopeScale)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthBiasClamp)>, float>);
    struct ShouldBeSameSizeAsDepthStencilState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::DepthStencilState>, false> {
        WebCore::WebGPU::TextureFormat format;
        bool depthWriteEnabled;
        WebCore::WebGPU::CompareFunction depthCompare;
        WebKit::WebGPU::StencilFaceState stencilFront;
        WebKit::WebGPU::StencilFaceState stencilBack;
        std::optional<WebCore::WebGPU::StencilValue> stencilReadMask;
        std::optional<WebCore::WebGPU::StencilValue> stencilWriteMask;
        WebCore::WebGPU::DepthBias depthBias;
        float depthBiasSlopeScale;
        float depthBiasClamp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDepthStencilState) == sizeof(WebKit::WebGPU::DepthStencilState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::DepthStencilState, format)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthWriteEnabled)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthCompare)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilFront)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilBack)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilReadMask)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilWriteMask)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthBias)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthBiasSlopeScale)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthBiasClamp)
    >::value);
    encoder << instance.format;
    encoder << instance.depthWriteEnabled;
    encoder << instance.depthCompare;
    encoder << instance.stencilFront;
    encoder << instance.stencilBack;
    encoder << instance.stencilReadMask;
    encoder << instance.stencilWriteMask;
    encoder << instance.depthBias;
    encoder << instance.depthBiasSlopeScale;
    encoder << instance.depthBiasClamp;
}

void ArgumentCoder<WebKit::WebGPU::DepthStencilState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::DepthStencilState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthWriteEnabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthCompare)>, WebCore::WebGPU::CompareFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilFront)>, WebKit::WebGPU::StencilFaceState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilBack)>, WebKit::WebGPU::StencilFaceState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilReadMask)>, std::optional<WebCore::WebGPU::StencilValue>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilWriteMask)>, std::optional<WebCore::WebGPU::StencilValue>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthBias)>, WebCore::WebGPU::DepthBias>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthBiasSlopeScale)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthBiasClamp)>, float>);
    struct ShouldBeSameSizeAsDepthStencilState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::DepthStencilState>, false> {
        WebCore::WebGPU::TextureFormat format;
        bool depthWriteEnabled;
        WebCore::WebGPU::CompareFunction depthCompare;
        WebKit::WebGPU::StencilFaceState stencilFront;
        WebKit::WebGPU::StencilFaceState stencilBack;
        std::optional<WebCore::WebGPU::StencilValue> stencilReadMask;
        std::optional<WebCore::WebGPU::StencilValue> stencilWriteMask;
        WebCore::WebGPU::DepthBias depthBias;
        float depthBiasSlopeScale;
        float depthBiasClamp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsDepthStencilState) == sizeof(WebKit::WebGPU::DepthStencilState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::DepthStencilState, format)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthWriteEnabled)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthCompare)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilFront)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilBack)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilReadMask)
        , offsetof(WebKit::WebGPU::DepthStencilState, stencilWriteMask)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthBias)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthBiasSlopeScale)
        , offsetof(WebKit::WebGPU::DepthStencilState, depthBiasClamp)
    >::value);
    encoder << instance.format;
    encoder << instance.depthWriteEnabled;
    encoder << instance.depthCompare;
    encoder << instance.stencilFront;
    encoder << instance.stencilBack;
    encoder << instance.stencilReadMask;
    encoder << instance.stencilWriteMask;
    encoder << instance.depthBias;
    encoder << instance.depthBiasSlopeScale;
    encoder << instance.depthBiasClamp;
}

std::optional<WebKit::WebGPU::DepthStencilState> ArgumentCoder<WebKit::WebGPU::DepthStencilState>::decode(Decoder& decoder)
{
    auto format = decoder.decode<WebCore::WebGPU::TextureFormat>();
    auto depthWriteEnabled = decoder.decode<bool>();
    auto depthCompare = decoder.decode<WebCore::WebGPU::CompareFunction>();
    auto stencilFront = decoder.decode<WebKit::WebGPU::StencilFaceState>();
    auto stencilBack = decoder.decode<WebKit::WebGPU::StencilFaceState>();
    auto stencilReadMask = decoder.decode<std::optional<WebCore::WebGPU::StencilValue>>();
    auto stencilWriteMask = decoder.decode<std::optional<WebCore::WebGPU::StencilValue>>();
    auto depthBias = decoder.decode<WebCore::WebGPU::DepthBias>();
    auto depthBiasSlopeScale = decoder.decode<float>();
    auto depthBiasClamp = decoder.decode<float>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::DepthStencilState {
            WTFMove(*format),
            WTFMove(*depthWriteEnabled),
            WTFMove(*depthCompare),
            WTFMove(*stencilFront),
            WTFMove(*stencilBack),
            WTFMove(*stencilReadMask),
            WTFMove(*stencilWriteMask),
            WTFMove(*depthBias),
            WTFMove(*depthBiasSlopeScale),
            WTFMove(*depthBiasClamp)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::DeviceDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::DeviceDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requiredFeatures)>, Vector<WebCore::WebGPU::FeatureName>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requiredLimits)>, Vector<KeyValuePair<String, uint64_t>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::DeviceDescriptor, requiredFeatures)
        , offsetof(WebKit::WebGPU::DeviceDescriptor, requiredLimits)
    >::value);
    encoder << instance.label;
    encoder << instance.requiredFeatures;
    encoder << instance.requiredLimits;
}

void ArgumentCoder<WebKit::WebGPU::DeviceDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::DeviceDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requiredFeatures)>, Vector<WebCore::WebGPU::FeatureName>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.requiredLimits)>, Vector<KeyValuePair<String, uint64_t>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::DeviceDescriptor, requiredFeatures)
        , offsetof(WebKit::WebGPU::DeviceDescriptor, requiredLimits)
    >::value);
    encoder << instance.label;
    encoder << instance.requiredFeatures;
    encoder << instance.requiredLimits;
}

std::optional<WebKit::WebGPU::DeviceDescriptor> ArgumentCoder<WebKit::WebGPU::DeviceDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto requiredFeatures = decoder.decode<Vector<WebCore::WebGPU::FeatureName>>();
    auto requiredLimits = decoder.decode<Vector<KeyValuePair<String, uint64_t>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::DeviceDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*requiredFeatures),
            WTFMove(*requiredLimits)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::Extent3DDict>::encode(Encoder& encoder, const WebKit::WebGPU::Extent3DDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthOrArrayLayers)>, WebCore::WebGPU::IntegerCoordinate>);
    struct ShouldBeSameSizeAsExtent3DDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::Extent3DDict>, false> {
        WebCore::WebGPU::IntegerCoordinate width;
        WebCore::WebGPU::IntegerCoordinate height;
        WebCore::WebGPU::IntegerCoordinate depthOrArrayLayers;
    };
    static_assert(sizeof(ShouldBeSameSizeAsExtent3DDict) == sizeof(WebKit::WebGPU::Extent3DDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::Extent3DDict, width)
        , offsetof(WebKit::WebGPU::Extent3DDict, height)
        , offsetof(WebKit::WebGPU::Extent3DDict, depthOrArrayLayers)
    >::value);
    encoder << instance.width;
    encoder << instance.height;
    encoder << instance.depthOrArrayLayers;
}

void ArgumentCoder<WebKit::WebGPU::Extent3DDict>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::Extent3DDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.width)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.height)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthOrArrayLayers)>, WebCore::WebGPU::IntegerCoordinate>);
    struct ShouldBeSameSizeAsExtent3DDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::Extent3DDict>, false> {
        WebCore::WebGPU::IntegerCoordinate width;
        WebCore::WebGPU::IntegerCoordinate height;
        WebCore::WebGPU::IntegerCoordinate depthOrArrayLayers;
    };
    static_assert(sizeof(ShouldBeSameSizeAsExtent3DDict) == sizeof(WebKit::WebGPU::Extent3DDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::Extent3DDict, width)
        , offsetof(WebKit::WebGPU::Extent3DDict, height)
        , offsetof(WebKit::WebGPU::Extent3DDict, depthOrArrayLayers)
    >::value);
    encoder << instance.width;
    encoder << instance.height;
    encoder << instance.depthOrArrayLayers;
}

std::optional<WebKit::WebGPU::Extent3DDict> ArgumentCoder<WebKit::WebGPU::Extent3DDict>::decode(Decoder& decoder)
{
    auto width = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto height = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto depthOrArrayLayers = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::Extent3DDict {
            WTFMove(*width),
            WTFMove(*height),
            WTFMove(*depthOrArrayLayers)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ExternalTextureBindingLayout>::encode(Encoder& encoder, const WebKit::WebGPU::ExternalTextureBindingLayout& instance)
{
    struct ShouldBeSameSizeAsExternalTextureBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ExternalTextureBindingLayout>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsExternalTextureBindingLayout) == sizeof(WebKit::WebGPU::ExternalTextureBindingLayout));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebKit::WebGPU::ExternalTextureBindingLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ExternalTextureBindingLayout& instance)
{
    struct ShouldBeSameSizeAsExternalTextureBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ExternalTextureBindingLayout>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsExternalTextureBindingLayout) == sizeof(WebKit::WebGPU::ExternalTextureBindingLayout));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebKit::WebGPU::ExternalTextureBindingLayout> ArgumentCoder<WebKit::WebGPU::ExternalTextureBindingLayout>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ExternalTextureBindingLayout {
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::WebGPU::HTMLVideoElementIdentifier>::encode(Encoder& encoder, const WebCore::WebGPU::HTMLVideoElementIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, uint64_t>);
    struct ShouldBeSameSizeAsHTMLVideoElementIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WebGPU::HTMLVideoElementIdentifier>, false> {
        uint64_t identifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHTMLVideoElementIdentifier) == sizeof(WebCore::WebGPU::HTMLVideoElementIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WebGPU::HTMLVideoElementIdentifier, identifier)
    >::value);
    encoder << instance.identifier;
}

void ArgumentCoder<WebCore::WebGPU::HTMLVideoElementIdentifier>::encode(StreamConnectionEncoder& encoder, const WebCore::WebGPU::HTMLVideoElementIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, uint64_t>);
    struct ShouldBeSameSizeAsHTMLVideoElementIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WebGPU::HTMLVideoElementIdentifier>, false> {
        uint64_t identifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsHTMLVideoElementIdentifier) == sizeof(WebCore::WebGPU::HTMLVideoElementIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WebGPU::HTMLVideoElementIdentifier, identifier)
    >::value);
    encoder << instance.identifier;
}

std::optional<WebCore::WebGPU::HTMLVideoElementIdentifier> ArgumentCoder<WebCore::WebGPU::HTMLVideoElementIdentifier>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WebGPU::HTMLVideoElementIdentifier {
            WTFMove(*identifier)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebCore::WebGPU::WebCodecsVideoFrameIdentifier>::encode(Encoder& encoder, const WebCore::WebGPU::WebCodecsVideoFrameIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, std::pair<uint64_t, uint64_t>>);
    struct ShouldBeSameSizeAsWebCodecsVideoFrameIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WebGPU::WebCodecsVideoFrameIdentifier>, false> {
        std::pair<uint64_t, uint64_t> identifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebCodecsVideoFrameIdentifier) == sizeof(WebCore::WebGPU::WebCodecsVideoFrameIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WebGPU::WebCodecsVideoFrameIdentifier, identifier)
    >::value);
    encoder << instance.identifier;
}

void ArgumentCoder<WebCore::WebGPU::WebCodecsVideoFrameIdentifier>::encode(StreamConnectionEncoder& encoder, const WebCore::WebGPU::WebCodecsVideoFrameIdentifier& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.identifier)>, std::pair<uint64_t, uint64_t>>);
    struct ShouldBeSameSizeAsWebCodecsVideoFrameIdentifier : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebCore::WebGPU::WebCodecsVideoFrameIdentifier>, false> {
        std::pair<uint64_t, uint64_t> identifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebCodecsVideoFrameIdentifier) == sizeof(WebCore::WebGPU::WebCodecsVideoFrameIdentifier));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebCore::WebGPU::WebCodecsVideoFrameIdentifier, identifier)
    >::value);
    encoder << instance.identifier;
}

std::optional<WebCore::WebGPU::WebCodecsVideoFrameIdentifier> ArgumentCoder<WebCore::WebGPU::WebCodecsVideoFrameIdentifier>::decode(Decoder& decoder)
{
    auto identifier = decoder.decode<std::pair<uint64_t, uint64_t>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebCore::WebGPU::WebCodecsVideoFrameIdentifier {
            WTFMove(*identifier)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ExternalTextureDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::ExternalTextureDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaIdentifier)>, WebCore::WebGPU::VideoSourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::WebGPU::PredefinedColorSpace>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ExternalTextureDescriptor, mediaIdentifier)
        , offsetof(WebKit::WebGPU::ExternalTextureDescriptor, colorSpace)
    >::value);
    encoder << instance.label;
    encoder << instance.mediaIdentifier;
    encoder << instance.colorSpace;
}

void ArgumentCoder<WebKit::WebGPU::ExternalTextureDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ExternalTextureDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mediaIdentifier)>, WebCore::WebGPU::VideoSourceIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::WebGPU::PredefinedColorSpace>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ExternalTextureDescriptor, mediaIdentifier)
        , offsetof(WebKit::WebGPU::ExternalTextureDescriptor, colorSpace)
    >::value);
    encoder << instance.label;
    encoder << instance.mediaIdentifier;
    encoder << instance.colorSpace;
}

std::optional<WebKit::WebGPU::ExternalTextureDescriptor> ArgumentCoder<WebKit::WebGPU::ExternalTextureDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto mediaIdentifier = decoder.decode<WebCore::WebGPU::VideoSourceIdentifier>();
    auto colorSpace = decoder.decode<WebCore::WebGPU::PredefinedColorSpace>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ExternalTextureDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*mediaIdentifier),
            WTFMove(*colorSpace)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::FragmentState>::encode(Encoder& encoder, const WebKit::WebGPU::FragmentState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.module)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entryPoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constants)>, Vector<KeyValuePair<String, double>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ProgrammableStage, module)
        , offsetof(WebKit::WebGPU::ProgrammableStage, entryPoint)
        , offsetof(WebKit::WebGPU::ProgrammableStage, constants)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.targets)>, Vector<std::optional<WebKit::WebGPU::ColorTargetState>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::FragmentState, targets)
    >::value);
    encoder << instance.module;
    encoder << instance.entryPoint;
    encoder << instance.constants;
    encoder << instance.targets;
}

void ArgumentCoder<WebKit::WebGPU::FragmentState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::FragmentState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.module)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entryPoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constants)>, Vector<KeyValuePair<String, double>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ProgrammableStage, module)
        , offsetof(WebKit::WebGPU::ProgrammableStage, entryPoint)
        , offsetof(WebKit::WebGPU::ProgrammableStage, constants)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.targets)>, Vector<std::optional<WebKit::WebGPU::ColorTargetState>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::FragmentState, targets)
    >::value);
    encoder << instance.module;
    encoder << instance.entryPoint;
    encoder << instance.constants;
    encoder << instance.targets;
}

std::optional<WebKit::WebGPU::FragmentState> ArgumentCoder<WebKit::WebGPU::FragmentState>::decode(Decoder& decoder)
{
    auto module = decoder.decode<WebKit::WebGPUIdentifier>();
    auto entryPoint = decoder.decode<String>();
    auto constants = decoder.decode<Vector<KeyValuePair<String, double>>>();
    auto targets = decoder.decode<Vector<std::optional<WebKit::WebGPU::ColorTargetState>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::FragmentState {
            WebKit::WebGPU::ProgrammableStage {
                WTFMove(*module),
                WTFMove(*entryPoint),
                WTFMove(*constants)
            },
            WTFMove(*targets)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ImageCopyBuffer>::encode(Encoder& encoder, const WebKit::WebGPU::ImageCopyBuffer& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bytesPerRow)>, std::optional<WebCore::WebGPU::Size32>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rowsPerImage)>, std::optional<WebCore::WebGPU::Size32>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageDataLayout, offset)
        , offsetof(WebKit::WebGPU::ImageDataLayout, bytesPerRow)
        , offsetof(WebKit::WebGPU::ImageDataLayout, rowsPerImage)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffer)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyBuffer, buffer)
    >::value);
    encoder << instance.offset;
    encoder << instance.bytesPerRow;
    encoder << instance.rowsPerImage;
    encoder << instance.buffer;
}

void ArgumentCoder<WebKit::WebGPU::ImageCopyBuffer>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ImageCopyBuffer& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bytesPerRow)>, std::optional<WebCore::WebGPU::Size32>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rowsPerImage)>, std::optional<WebCore::WebGPU::Size32>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageDataLayout, offset)
        , offsetof(WebKit::WebGPU::ImageDataLayout, bytesPerRow)
        , offsetof(WebKit::WebGPU::ImageDataLayout, rowsPerImage)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffer)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyBuffer, buffer)
    >::value);
    encoder << instance.offset;
    encoder << instance.bytesPerRow;
    encoder << instance.rowsPerImage;
    encoder << instance.buffer;
}

std::optional<WebKit::WebGPU::ImageCopyBuffer> ArgumentCoder<WebKit::WebGPU::ImageCopyBuffer>::decode(Decoder& decoder)
{
    auto offset = decoder.decode<WebCore::WebGPU::Size64>();
    auto bytesPerRow = decoder.decode<std::optional<WebCore::WebGPU::Size32>>();
    auto rowsPerImage = decoder.decode<std::optional<WebCore::WebGPU::Size32>>();
    auto buffer = decoder.decode<WebKit::WebGPUIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ImageCopyBuffer {
            WebKit::WebGPU::ImageDataLayout {
                WTFMove(*offset),
                WTFMove(*bytesPerRow),
                WTFMove(*rowsPerImage)
            },
            WTFMove(*buffer)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ImageCopyExternalImage>::encode(Encoder& encoder, const WebKit::WebGPU::ImageCopyExternalImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, std::optional<WebKit::WebGPU::Origin2D>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.flipY)>, bool>);
    struct ShouldBeSameSizeAsImageCopyExternalImage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ImageCopyExternalImage>, false> {
        std::optional<WebKit::WebGPU::Origin2D> origin;
        bool flipY;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageCopyExternalImage) == sizeof(WebKit::WebGPU::ImageCopyExternalImage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyExternalImage, origin)
        , offsetof(WebKit::WebGPU::ImageCopyExternalImage, flipY)
    >::value);
    encoder << instance.origin;
    encoder << instance.flipY;
}

void ArgumentCoder<WebKit::WebGPU::ImageCopyExternalImage>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ImageCopyExternalImage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, std::optional<WebKit::WebGPU::Origin2D>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.flipY)>, bool>);
    struct ShouldBeSameSizeAsImageCopyExternalImage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ImageCopyExternalImage>, false> {
        std::optional<WebKit::WebGPU::Origin2D> origin;
        bool flipY;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageCopyExternalImage) == sizeof(WebKit::WebGPU::ImageCopyExternalImage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyExternalImage, origin)
        , offsetof(WebKit::WebGPU::ImageCopyExternalImage, flipY)
    >::value);
    encoder << instance.origin;
    encoder << instance.flipY;
}

std::optional<WebKit::WebGPU::ImageCopyExternalImage> ArgumentCoder<WebKit::WebGPU::ImageCopyExternalImage>::decode(Decoder& decoder)
{
    auto origin = decoder.decode<std::optional<WebKit::WebGPU::Origin2D>>();
    auto flipY = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ImageCopyExternalImage {
            WTFMove(*origin),
            WTFMove(*flipY)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ImageCopyTexture>::encode(Encoder& encoder, const WebKit::WebGPU::ImageCopyTexture& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.texture)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevel)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, std::optional<WebKit::WebGPU::Origin3D>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspect)>, WebCore::WebGPU::TextureAspect>);
    struct ShouldBeSameSizeAsImageCopyTexture : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ImageCopyTexture>, false> {
        WebKit::WebGPUIdentifier texture;
        WebCore::WebGPU::IntegerCoordinate mipLevel;
        std::optional<WebKit::WebGPU::Origin3D> origin;
        WebCore::WebGPU::TextureAspect aspect;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageCopyTexture) == sizeof(WebKit::WebGPU::ImageCopyTexture));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyTexture, texture)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, mipLevel)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, origin)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, aspect)
    >::value);
    encoder << instance.texture;
    encoder << instance.mipLevel;
    encoder << instance.origin;
    encoder << instance.aspect;
}

void ArgumentCoder<WebKit::WebGPU::ImageCopyTexture>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ImageCopyTexture& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.texture)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevel)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, std::optional<WebKit::WebGPU::Origin3D>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspect)>, WebCore::WebGPU::TextureAspect>);
    struct ShouldBeSameSizeAsImageCopyTexture : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ImageCopyTexture>, false> {
        WebKit::WebGPUIdentifier texture;
        WebCore::WebGPU::IntegerCoordinate mipLevel;
        std::optional<WebKit::WebGPU::Origin3D> origin;
        WebCore::WebGPU::TextureAspect aspect;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageCopyTexture) == sizeof(WebKit::WebGPU::ImageCopyTexture));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyTexture, texture)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, mipLevel)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, origin)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, aspect)
    >::value);
    encoder << instance.texture;
    encoder << instance.mipLevel;
    encoder << instance.origin;
    encoder << instance.aspect;
}

std::optional<WebKit::WebGPU::ImageCopyTexture> ArgumentCoder<WebKit::WebGPU::ImageCopyTexture>::decode(Decoder& decoder)
{
    auto texture = decoder.decode<WebKit::WebGPUIdentifier>();
    auto mipLevel = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto origin = decoder.decode<std::optional<WebKit::WebGPU::Origin3D>>();
    auto aspect = decoder.decode<WebCore::WebGPU::TextureAspect>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ImageCopyTexture {
            WTFMove(*texture),
            WTFMove(*mipLevel),
            WTFMove(*origin),
            WTFMove(*aspect)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ImageCopyTextureTagged>::encode(Encoder& encoder, const WebKit::WebGPU::ImageCopyTextureTagged& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.texture)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevel)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, std::optional<WebKit::WebGPU::Origin3D>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspect)>, WebCore::WebGPU::TextureAspect>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyTexture, texture)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, mipLevel)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, origin)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, aspect)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::WebGPU::PredefinedColorSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.premultipliedAlpha)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyTextureTagged, colorSpace)
        , offsetof(WebKit::WebGPU::ImageCopyTextureTagged, premultipliedAlpha)
    >::value);
    encoder << instance.texture;
    encoder << instance.mipLevel;
    encoder << instance.origin;
    encoder << instance.aspect;
    encoder << instance.colorSpace;
    encoder << instance.premultipliedAlpha;
}

void ArgumentCoder<WebKit::WebGPU::ImageCopyTextureTagged>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ImageCopyTextureTagged& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.texture)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevel)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, std::optional<WebKit::WebGPU::Origin3D>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspect)>, WebCore::WebGPU::TextureAspect>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyTexture, texture)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, mipLevel)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, origin)
        , offsetof(WebKit::WebGPU::ImageCopyTexture, aspect)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorSpace)>, WebCore::WebGPU::PredefinedColorSpace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.premultipliedAlpha)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageCopyTextureTagged, colorSpace)
        , offsetof(WebKit::WebGPU::ImageCopyTextureTagged, premultipliedAlpha)
    >::value);
    encoder << instance.texture;
    encoder << instance.mipLevel;
    encoder << instance.origin;
    encoder << instance.aspect;
    encoder << instance.colorSpace;
    encoder << instance.premultipliedAlpha;
}

std::optional<WebKit::WebGPU::ImageCopyTextureTagged> ArgumentCoder<WebKit::WebGPU::ImageCopyTextureTagged>::decode(Decoder& decoder)
{
    auto texture = decoder.decode<WebKit::WebGPUIdentifier>();
    auto mipLevel = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto origin = decoder.decode<std::optional<WebKit::WebGPU::Origin3D>>();
    auto aspect = decoder.decode<WebCore::WebGPU::TextureAspect>();
    auto colorSpace = decoder.decode<WebCore::WebGPU::PredefinedColorSpace>();
    auto premultipliedAlpha = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ImageCopyTextureTagged {
            WebKit::WebGPU::ImageCopyTexture {
                WTFMove(*texture),
                WTFMove(*mipLevel),
                WTFMove(*origin),
                WTFMove(*aspect)
            },
            WTFMove(*colorSpace),
            WTFMove(*premultipliedAlpha)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ImageDataLayout>::encode(Encoder& encoder, const WebKit::WebGPU::ImageDataLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bytesPerRow)>, std::optional<WebCore::WebGPU::Size32>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rowsPerImage)>, std::optional<WebCore::WebGPU::Size32>>);
    struct ShouldBeSameSizeAsImageDataLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ImageDataLayout>, false> {
        WebCore::WebGPU::Size64 offset;
        std::optional<WebCore::WebGPU::Size32> bytesPerRow;
        std::optional<WebCore::WebGPU::Size32> rowsPerImage;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageDataLayout) == sizeof(WebKit::WebGPU::ImageDataLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageDataLayout, offset)
        , offsetof(WebKit::WebGPU::ImageDataLayout, bytesPerRow)
        , offsetof(WebKit::WebGPU::ImageDataLayout, rowsPerImage)
    >::value);
    encoder << instance.offset;
    encoder << instance.bytesPerRow;
    encoder << instance.rowsPerImage;
}

void ArgumentCoder<WebKit::WebGPU::ImageDataLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ImageDataLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bytesPerRow)>, std::optional<WebCore::WebGPU::Size32>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.rowsPerImage)>, std::optional<WebCore::WebGPU::Size32>>);
    struct ShouldBeSameSizeAsImageDataLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ImageDataLayout>, false> {
        WebCore::WebGPU::Size64 offset;
        std::optional<WebCore::WebGPU::Size32> bytesPerRow;
        std::optional<WebCore::WebGPU::Size32> rowsPerImage;
    };
    static_assert(sizeof(ShouldBeSameSizeAsImageDataLayout) == sizeof(WebKit::WebGPU::ImageDataLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ImageDataLayout, offset)
        , offsetof(WebKit::WebGPU::ImageDataLayout, bytesPerRow)
        , offsetof(WebKit::WebGPU::ImageDataLayout, rowsPerImage)
    >::value);
    encoder << instance.offset;
    encoder << instance.bytesPerRow;
    encoder << instance.rowsPerImage;
}

std::optional<WebKit::WebGPU::ImageDataLayout> ArgumentCoder<WebKit::WebGPU::ImageDataLayout>::decode(Decoder& decoder)
{
    auto offset = decoder.decode<WebCore::WebGPU::Size64>();
    auto bytesPerRow = decoder.decode<std::optional<WebCore::WebGPU::Size32>>();
    auto rowsPerImage = decoder.decode<std::optional<WebCore::WebGPU::Size32>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ImageDataLayout {
            WTFMove(*offset),
            WTFMove(*bytesPerRow),
            WTFMove(*rowsPerImage)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::MultisampleState>::encode(Encoder& encoder, const WebKit::WebGPU::MultisampleState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.count)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mask)>, WebCore::WebGPU::SampleMask>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaToCoverageEnabled)>, bool>);
    struct ShouldBeSameSizeAsMultisampleState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::MultisampleState>, false> {
        WebCore::WebGPU::Size32 count;
        WebCore::WebGPU::SampleMask mask;
        bool alphaToCoverageEnabled;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMultisampleState) == sizeof(WebKit::WebGPU::MultisampleState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::MultisampleState, count)
        , offsetof(WebKit::WebGPU::MultisampleState, mask)
        , offsetof(WebKit::WebGPU::MultisampleState, alphaToCoverageEnabled)
    >::value);
    encoder << instance.count;
    encoder << instance.mask;
    encoder << instance.alphaToCoverageEnabled;
}

void ArgumentCoder<WebKit::WebGPU::MultisampleState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::MultisampleState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.count)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mask)>, WebCore::WebGPU::SampleMask>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.alphaToCoverageEnabled)>, bool>);
    struct ShouldBeSameSizeAsMultisampleState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::MultisampleState>, false> {
        WebCore::WebGPU::Size32 count;
        WebCore::WebGPU::SampleMask mask;
        bool alphaToCoverageEnabled;
    };
    static_assert(sizeof(ShouldBeSameSizeAsMultisampleState) == sizeof(WebKit::WebGPU::MultisampleState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::MultisampleState, count)
        , offsetof(WebKit::WebGPU::MultisampleState, mask)
        , offsetof(WebKit::WebGPU::MultisampleState, alphaToCoverageEnabled)
    >::value);
    encoder << instance.count;
    encoder << instance.mask;
    encoder << instance.alphaToCoverageEnabled;
}

std::optional<WebKit::WebGPU::MultisampleState> ArgumentCoder<WebKit::WebGPU::MultisampleState>::decode(Decoder& decoder)
{
    auto count = decoder.decode<WebCore::WebGPU::Size32>();
    auto mask = decoder.decode<WebCore::WebGPU::SampleMask>();
    auto alphaToCoverageEnabled = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::MultisampleState {
            WTFMove(*count),
            WTFMove(*mask),
            WTFMove(*alphaToCoverageEnabled)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ObjectDescriptorBase>::encode(Encoder& encoder, const WebKit::WebGPU::ObjectDescriptorBase& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    struct ShouldBeSameSizeAsObjectDescriptorBase : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ObjectDescriptorBase>, false> {
        String label;
    };
    static_assert(sizeof(ShouldBeSameSizeAsObjectDescriptorBase) == sizeof(WebKit::WebGPU::ObjectDescriptorBase));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    encoder << instance.label;
}

void ArgumentCoder<WebKit::WebGPU::ObjectDescriptorBase>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ObjectDescriptorBase& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    struct ShouldBeSameSizeAsObjectDescriptorBase : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ObjectDescriptorBase>, false> {
        String label;
    };
    static_assert(sizeof(ShouldBeSameSizeAsObjectDescriptorBase) == sizeof(WebKit::WebGPU::ObjectDescriptorBase));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    encoder << instance.label;
}

std::optional<WebKit::WebGPU::ObjectDescriptorBase> ArgumentCoder<WebKit::WebGPU::ObjectDescriptorBase>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ObjectDescriptorBase {
            WTFMove(*label)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::Origin2DDict>::encode(Encoder& encoder, const WebKit::WebGPU::Origin2DDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y)>, WebCore::WebGPU::IntegerCoordinate>);
    struct ShouldBeSameSizeAsOrigin2DDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::Origin2DDict>, false> {
        WebCore::WebGPU::IntegerCoordinate x;
        WebCore::WebGPU::IntegerCoordinate y;
    };
    static_assert(sizeof(ShouldBeSameSizeAsOrigin2DDict) == sizeof(WebKit::WebGPU::Origin2DDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::Origin2DDict, x)
        , offsetof(WebKit::WebGPU::Origin2DDict, y)
    >::value);
    encoder << instance.x;
    encoder << instance.y;
}

void ArgumentCoder<WebKit::WebGPU::Origin2DDict>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::Origin2DDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y)>, WebCore::WebGPU::IntegerCoordinate>);
    struct ShouldBeSameSizeAsOrigin2DDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::Origin2DDict>, false> {
        WebCore::WebGPU::IntegerCoordinate x;
        WebCore::WebGPU::IntegerCoordinate y;
    };
    static_assert(sizeof(ShouldBeSameSizeAsOrigin2DDict) == sizeof(WebKit::WebGPU::Origin2DDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::Origin2DDict, x)
        , offsetof(WebKit::WebGPU::Origin2DDict, y)
    >::value);
    encoder << instance.x;
    encoder << instance.y;
}

std::optional<WebKit::WebGPU::Origin2DDict> ArgumentCoder<WebKit::WebGPU::Origin2DDict>::decode(Decoder& decoder)
{
    auto x = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto y = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::Origin2DDict {
            WTFMove(*x),
            WTFMove(*y)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::Origin3DDict>::encode(Encoder& encoder, const WebKit::WebGPU::Origin3DDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.z)>, WebCore::WebGPU::IntegerCoordinate>);
    struct ShouldBeSameSizeAsOrigin3DDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::Origin3DDict>, false> {
        WebCore::WebGPU::IntegerCoordinate x;
        WebCore::WebGPU::IntegerCoordinate y;
        WebCore::WebGPU::IntegerCoordinate z;
    };
    static_assert(sizeof(ShouldBeSameSizeAsOrigin3DDict) == sizeof(WebKit::WebGPU::Origin3DDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::Origin3DDict, x)
        , offsetof(WebKit::WebGPU::Origin3DDict, y)
        , offsetof(WebKit::WebGPU::Origin3DDict, z)
    >::value);
    encoder << instance.x;
    encoder << instance.y;
    encoder << instance.z;
}

void ArgumentCoder<WebKit::WebGPU::Origin3DDict>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::Origin3DDict& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.x)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.y)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.z)>, WebCore::WebGPU::IntegerCoordinate>);
    struct ShouldBeSameSizeAsOrigin3DDict : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::Origin3DDict>, false> {
        WebCore::WebGPU::IntegerCoordinate x;
        WebCore::WebGPU::IntegerCoordinate y;
        WebCore::WebGPU::IntegerCoordinate z;
    };
    static_assert(sizeof(ShouldBeSameSizeAsOrigin3DDict) == sizeof(WebKit::WebGPU::Origin3DDict));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::Origin3DDict, x)
        , offsetof(WebKit::WebGPU::Origin3DDict, y)
        , offsetof(WebKit::WebGPU::Origin3DDict, z)
    >::value);
    encoder << instance.x;
    encoder << instance.y;
    encoder << instance.z;
}

std::optional<WebKit::WebGPU::Origin3DDict> ArgumentCoder<WebKit::WebGPU::Origin3DDict>::decode(Decoder& decoder)
{
    auto x = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto y = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto z = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::Origin3DDict {
            WTFMove(*x),
            WTFMove(*y),
            WTFMove(*z)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::OutOfMemoryError>::encode(Encoder& encoder, const WebKit::WebGPU::OutOfMemoryError& instance)
{
    struct ShouldBeSameSizeAsOutOfMemoryError : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::OutOfMemoryError>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsOutOfMemoryError) == sizeof(WebKit::WebGPU::OutOfMemoryError));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

void ArgumentCoder<WebKit::WebGPU::OutOfMemoryError>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::OutOfMemoryError& instance)
{
    struct ShouldBeSameSizeAsOutOfMemoryError : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::OutOfMemoryError>, false> {
    };
    static_assert(sizeof(ShouldBeSameSizeAsOutOfMemoryError) == sizeof(WebKit::WebGPU::OutOfMemoryError));
    static_assert(MembersInCorrectOrder < 0
    >::value);
}

std::optional<WebKit::WebGPU::OutOfMemoryError> ArgumentCoder<WebKit::WebGPU::OutOfMemoryError>::decode(Decoder& decoder)
{
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::OutOfMemoryError {
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::PipelineDescriptorBase>::encode(Encoder& encoder, const WebKit::WebGPU::PipelineDescriptorBase& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layout)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineDescriptorBase, layout)
    >::value);
    encoder << instance.label;
    encoder << instance.layout;
}

void ArgumentCoder<WebKit::WebGPU::PipelineDescriptorBase>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::PipelineDescriptorBase& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layout)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineDescriptorBase, layout)
    >::value);
    encoder << instance.label;
    encoder << instance.layout;
}

std::optional<WebKit::WebGPU::PipelineDescriptorBase> ArgumentCoder<WebKit::WebGPU::PipelineDescriptorBase>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto layout = decoder.decode<WebKit::WebGPUIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::PipelineDescriptorBase {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*layout)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::PipelineLayoutDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::PipelineLayoutDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bindGroupLayouts)>, std::optional<Vector<WebKit::WebGPUIdentifier>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineLayoutDescriptor, bindGroupLayouts)
    >::value);
    encoder << instance.label;
    encoder << instance.bindGroupLayouts;
}

void ArgumentCoder<WebKit::WebGPU::PipelineLayoutDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::PipelineLayoutDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bindGroupLayouts)>, std::optional<Vector<WebKit::WebGPUIdentifier>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineLayoutDescriptor, bindGroupLayouts)
    >::value);
    encoder << instance.label;
    encoder << instance.bindGroupLayouts;
}

std::optional<WebKit::WebGPU::PipelineLayoutDescriptor> ArgumentCoder<WebKit::WebGPU::PipelineLayoutDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto bindGroupLayouts = decoder.decode<std::optional<Vector<WebKit::WebGPUIdentifier>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::PipelineLayoutDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*bindGroupLayouts)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::PresentationContextDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::PresentationContextDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositorIntegration)>, WebKit::WebGPUIdentifier>);
    struct ShouldBeSameSizeAsPresentationContextDescriptor : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::PresentationContextDescriptor>, false> {
        WebKit::WebGPUIdentifier compositorIntegration;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPresentationContextDescriptor) == sizeof(WebKit::WebGPU::PresentationContextDescriptor));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PresentationContextDescriptor, compositorIntegration)
    >::value);
    encoder << instance.compositorIntegration;
}

void ArgumentCoder<WebKit::WebGPU::PresentationContextDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::PresentationContextDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compositorIntegration)>, WebKit::WebGPUIdentifier>);
    struct ShouldBeSameSizeAsPresentationContextDescriptor : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::PresentationContextDescriptor>, false> {
        WebKit::WebGPUIdentifier compositorIntegration;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPresentationContextDescriptor) == sizeof(WebKit::WebGPU::PresentationContextDescriptor));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PresentationContextDescriptor, compositorIntegration)
    >::value);
    encoder << instance.compositorIntegration;
}

std::optional<WebKit::WebGPU::PresentationContextDescriptor> ArgumentCoder<WebKit::WebGPU::PresentationContextDescriptor>::decode(Decoder& decoder)
{
    auto compositorIntegration = decoder.decode<WebKit::WebGPUIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::PresentationContextDescriptor {
            WTFMove(*compositorIntegration)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::PrimitiveState>::encode(Encoder& encoder, const WebKit::WebGPU::PrimitiveState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topology)>, WebCore::WebGPU::PrimitiveTopology>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stripIndexFormat)>, std::optional<WebCore::WebGPU::IndexFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frontFace)>, WebCore::WebGPU::FrontFace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cullMode)>, WebCore::WebGPU::CullMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.unclippedDepth)>, bool>);
    struct ShouldBeSameSizeAsPrimitiveState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::PrimitiveState>, false> {
        WebCore::WebGPU::PrimitiveTopology topology;
        std::optional<WebCore::WebGPU::IndexFormat> stripIndexFormat;
        WebCore::WebGPU::FrontFace frontFace;
        WebCore::WebGPU::CullMode cullMode;
        bool unclippedDepth;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPrimitiveState) == sizeof(WebKit::WebGPU::PrimitiveState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PrimitiveState, topology)
        , offsetof(WebKit::WebGPU::PrimitiveState, stripIndexFormat)
        , offsetof(WebKit::WebGPU::PrimitiveState, frontFace)
        , offsetof(WebKit::WebGPU::PrimitiveState, cullMode)
        , offsetof(WebKit::WebGPU::PrimitiveState, unclippedDepth)
    >::value);
    encoder << instance.topology;
    encoder << instance.stripIndexFormat;
    encoder << instance.frontFace;
    encoder << instance.cullMode;
    encoder << instance.unclippedDepth;
}

void ArgumentCoder<WebKit::WebGPU::PrimitiveState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::PrimitiveState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.topology)>, WebCore::WebGPU::PrimitiveTopology>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stripIndexFormat)>, std::optional<WebCore::WebGPU::IndexFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frontFace)>, WebCore::WebGPU::FrontFace>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.cullMode)>, WebCore::WebGPU::CullMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.unclippedDepth)>, bool>);
    struct ShouldBeSameSizeAsPrimitiveState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::PrimitiveState>, false> {
        WebCore::WebGPU::PrimitiveTopology topology;
        std::optional<WebCore::WebGPU::IndexFormat> stripIndexFormat;
        WebCore::WebGPU::FrontFace frontFace;
        WebCore::WebGPU::CullMode cullMode;
        bool unclippedDepth;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPrimitiveState) == sizeof(WebKit::WebGPU::PrimitiveState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PrimitiveState, topology)
        , offsetof(WebKit::WebGPU::PrimitiveState, stripIndexFormat)
        , offsetof(WebKit::WebGPU::PrimitiveState, frontFace)
        , offsetof(WebKit::WebGPU::PrimitiveState, cullMode)
        , offsetof(WebKit::WebGPU::PrimitiveState, unclippedDepth)
    >::value);
    encoder << instance.topology;
    encoder << instance.stripIndexFormat;
    encoder << instance.frontFace;
    encoder << instance.cullMode;
    encoder << instance.unclippedDepth;
}

std::optional<WebKit::WebGPU::PrimitiveState> ArgumentCoder<WebKit::WebGPU::PrimitiveState>::decode(Decoder& decoder)
{
    auto topology = decoder.decode<WebCore::WebGPU::PrimitiveTopology>();
    auto stripIndexFormat = decoder.decode<std::optional<WebCore::WebGPU::IndexFormat>>();
    auto frontFace = decoder.decode<WebCore::WebGPU::FrontFace>();
    auto cullMode = decoder.decode<WebCore::WebGPU::CullMode>();
    auto unclippedDepth = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::PrimitiveState {
            WTFMove(*topology),
            WTFMove(*stripIndexFormat),
            WTFMove(*frontFace),
            WTFMove(*cullMode),
            WTFMove(*unclippedDepth)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ProgrammableStage>::encode(Encoder& encoder, const WebKit::WebGPU::ProgrammableStage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.module)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entryPoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constants)>, Vector<KeyValuePair<String, double>>>);
    struct ShouldBeSameSizeAsProgrammableStage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ProgrammableStage>, false> {
        WebKit::WebGPUIdentifier module;
        String entryPoint;
        Vector<KeyValuePair<String, double>> constants;
    };
    static_assert(sizeof(ShouldBeSameSizeAsProgrammableStage) == sizeof(WebKit::WebGPU::ProgrammableStage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ProgrammableStage, module)
        , offsetof(WebKit::WebGPU::ProgrammableStage, entryPoint)
        , offsetof(WebKit::WebGPU::ProgrammableStage, constants)
    >::value);
    encoder << instance.module;
    encoder << instance.entryPoint;
    encoder << instance.constants;
}

void ArgumentCoder<WebKit::WebGPU::ProgrammableStage>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ProgrammableStage& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.module)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entryPoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constants)>, Vector<KeyValuePair<String, double>>>);
    struct ShouldBeSameSizeAsProgrammableStage : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ProgrammableStage>, false> {
        WebKit::WebGPUIdentifier module;
        String entryPoint;
        Vector<KeyValuePair<String, double>> constants;
    };
    static_assert(sizeof(ShouldBeSameSizeAsProgrammableStage) == sizeof(WebKit::WebGPU::ProgrammableStage));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ProgrammableStage, module)
        , offsetof(WebKit::WebGPU::ProgrammableStage, entryPoint)
        , offsetof(WebKit::WebGPU::ProgrammableStage, constants)
    >::value);
    encoder << instance.module;
    encoder << instance.entryPoint;
    encoder << instance.constants;
}

std::optional<WebKit::WebGPU::ProgrammableStage> ArgumentCoder<WebKit::WebGPU::ProgrammableStage>::decode(Decoder& decoder)
{
    auto module = decoder.decode<WebKit::WebGPUIdentifier>();
    auto entryPoint = decoder.decode<String>();
    auto constants = decoder.decode<Vector<KeyValuePair<String, double>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ProgrammableStage {
            WTFMove(*module),
            WTFMove(*entryPoint),
            WTFMove(*constants)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::QuerySetDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::QuerySetDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::QueryType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.count)>, WebCore::WebGPU::Size32>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::QuerySetDescriptor, type)
        , offsetof(WebKit::WebGPU::QuerySetDescriptor, count)
    >::value);
    encoder << instance.label;
    encoder << instance.type;
    encoder << instance.count;
}

void ArgumentCoder<WebKit::WebGPU::QuerySetDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::QuerySetDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::QueryType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.count)>, WebCore::WebGPU::Size32>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::QuerySetDescriptor, type)
        , offsetof(WebKit::WebGPU::QuerySetDescriptor, count)
    >::value);
    encoder << instance.label;
    encoder << instance.type;
    encoder << instance.count;
}

std::optional<WebKit::WebGPU::QuerySetDescriptor> ArgumentCoder<WebKit::WebGPU::QuerySetDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto type = decoder.decode<WebCore::WebGPU::QueryType>();
    auto count = decoder.decode<WebCore::WebGPU::Size32>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::QuerySetDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*type),
            WTFMove(*count)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderBundleDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::RenderBundleDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(MembersInCorrectOrder < 0
    >::value);
    encoder << instance.label;
}

void ArgumentCoder<WebKit::WebGPU::RenderBundleDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderBundleDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(MembersInCorrectOrder < 0
    >::value);
    encoder << instance.label;
}

std::optional<WebKit::WebGPU::RenderBundleDescriptor> ArgumentCoder<WebKit::WebGPU::RenderBundleDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderBundleDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            }
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderBundleEncoderDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::RenderBundleEncoderDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorFormats)>, Vector<std::optional<WebCore::WebGPU::TextureFormat>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencilFormat)>, std::optional<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleCount)>, WebCore::WebGPU::Size32>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassLayout, colorFormats)
        , offsetof(WebKit::WebGPU::RenderPassLayout, depthStencilFormat)
        , offsetof(WebKit::WebGPU::RenderPassLayout, sampleCount)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthReadOnly)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilReadOnly)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderBundleEncoderDescriptor, depthReadOnly)
        , offsetof(WebKit::WebGPU::RenderBundleEncoderDescriptor, stencilReadOnly)
    >::value);
    encoder << instance.label;
    encoder << instance.colorFormats;
    encoder << instance.depthStencilFormat;
    encoder << instance.sampleCount;
    encoder << instance.depthReadOnly;
    encoder << instance.stencilReadOnly;
}

void ArgumentCoder<WebKit::WebGPU::RenderBundleEncoderDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderBundleEncoderDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorFormats)>, Vector<std::optional<WebCore::WebGPU::TextureFormat>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencilFormat)>, std::optional<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleCount)>, WebCore::WebGPU::Size32>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassLayout, colorFormats)
        , offsetof(WebKit::WebGPU::RenderPassLayout, depthStencilFormat)
        , offsetof(WebKit::WebGPU::RenderPassLayout, sampleCount)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthReadOnly)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilReadOnly)>, bool>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderBundleEncoderDescriptor, depthReadOnly)
        , offsetof(WebKit::WebGPU::RenderBundleEncoderDescriptor, stencilReadOnly)
    >::value);
    encoder << instance.label;
    encoder << instance.colorFormats;
    encoder << instance.depthStencilFormat;
    encoder << instance.sampleCount;
    encoder << instance.depthReadOnly;
    encoder << instance.stencilReadOnly;
}

std::optional<WebKit::WebGPU::RenderBundleEncoderDescriptor> ArgumentCoder<WebKit::WebGPU::RenderBundleEncoderDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto colorFormats = decoder.decode<Vector<std::optional<WebCore::WebGPU::TextureFormat>>>();
    auto depthStencilFormat = decoder.decode<std::optional<WebCore::WebGPU::TextureFormat>>();
    auto sampleCount = decoder.decode<WebCore::WebGPU::Size32>();
    auto depthReadOnly = decoder.decode<bool>();
    auto stencilReadOnly = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderBundleEncoderDescriptor {
            WebKit::WebGPU::RenderPassLayout {
                WebKit::WebGPU::ObjectDescriptorBase {
                    WTFMove(*label)
                },
                WTFMove(*colorFormats),
                WTFMove(*depthStencilFormat),
                WTFMove(*sampleCount)
            },
            WTFMove(*depthReadOnly),
            WTFMove(*stencilReadOnly)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderPassColorAttachment>::encode(Encoder& encoder, const WebKit::WebGPU::RenderPassColorAttachment& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.view)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resolveTarget)>, std::optional<WebKit::WebGPUIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clearValue)>, std::optional<WebKit::WebGPU::Color>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.loadOp)>, WebCore::WebGPU::LoadOp>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.storeOp)>, WebCore::WebGPU::StoreOp>);
    struct ShouldBeSameSizeAsRenderPassColorAttachment : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RenderPassColorAttachment>, false> {
        WebKit::WebGPUIdentifier view;
        std::optional<WebKit::WebGPUIdentifier> resolveTarget;
        std::optional<WebKit::WebGPU::Color> clearValue;
        WebCore::WebGPU::LoadOp loadOp;
        WebCore::WebGPU::StoreOp storeOp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRenderPassColorAttachment) == sizeof(WebKit::WebGPU::RenderPassColorAttachment));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, view)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, resolveTarget)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, clearValue)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, loadOp)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, storeOp)
    >::value);
    encoder << instance.view;
    encoder << instance.resolveTarget;
    encoder << instance.clearValue;
    encoder << instance.loadOp;
    encoder << instance.storeOp;
}

void ArgumentCoder<WebKit::WebGPU::RenderPassColorAttachment>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderPassColorAttachment& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.view)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.resolveTarget)>, std::optional<WebKit::WebGPUIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.clearValue)>, std::optional<WebKit::WebGPU::Color>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.loadOp)>, WebCore::WebGPU::LoadOp>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.storeOp)>, WebCore::WebGPU::StoreOp>);
    struct ShouldBeSameSizeAsRenderPassColorAttachment : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RenderPassColorAttachment>, false> {
        WebKit::WebGPUIdentifier view;
        std::optional<WebKit::WebGPUIdentifier> resolveTarget;
        std::optional<WebKit::WebGPU::Color> clearValue;
        WebCore::WebGPU::LoadOp loadOp;
        WebCore::WebGPU::StoreOp storeOp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRenderPassColorAttachment) == sizeof(WebKit::WebGPU::RenderPassColorAttachment));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, view)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, resolveTarget)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, clearValue)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, loadOp)
        , offsetof(WebKit::WebGPU::RenderPassColorAttachment, storeOp)
    >::value);
    encoder << instance.view;
    encoder << instance.resolveTarget;
    encoder << instance.clearValue;
    encoder << instance.loadOp;
    encoder << instance.storeOp;
}

std::optional<WebKit::WebGPU::RenderPassColorAttachment> ArgumentCoder<WebKit::WebGPU::RenderPassColorAttachment>::decode(Decoder& decoder)
{
    auto view = decoder.decode<WebKit::WebGPUIdentifier>();
    auto resolveTarget = decoder.decode<std::optional<WebKit::WebGPUIdentifier>>();
    auto clearValue = decoder.decode<std::optional<WebKit::WebGPU::Color>>();
    auto loadOp = decoder.decode<WebCore::WebGPU::LoadOp>();
    auto storeOp = decoder.decode<WebCore::WebGPU::StoreOp>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderPassColorAttachment {
            WTFMove(*view),
            WTFMove(*resolveTarget),
            WTFMove(*clearValue),
            WTFMove(*loadOp),
            WTFMove(*storeOp)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderPassDepthStencilAttachment>::encode(Encoder& encoder, const WebKit::WebGPU::RenderPassDepthStencilAttachment& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.view)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthClearValue)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthLoadOp)>, std::optional<WebCore::WebGPU::LoadOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStoreOp)>, std::optional<WebCore::WebGPU::StoreOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthReadOnly)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilClearValue)>, WebCore::WebGPU::StencilValue>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilLoadOp)>, std::optional<WebCore::WebGPU::LoadOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilStoreOp)>, std::optional<WebCore::WebGPU::StoreOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilReadOnly)>, bool>);
    struct ShouldBeSameSizeAsRenderPassDepthStencilAttachment : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RenderPassDepthStencilAttachment>, false> {
        WebKit::WebGPUIdentifier view;
        float depthClearValue;
        std::optional<WebCore::WebGPU::LoadOp> depthLoadOp;
        std::optional<WebCore::WebGPU::StoreOp> depthStoreOp;
        bool depthReadOnly;
        WebCore::WebGPU::StencilValue stencilClearValue;
        std::optional<WebCore::WebGPU::LoadOp> stencilLoadOp;
        std::optional<WebCore::WebGPU::StoreOp> stencilStoreOp;
        bool stencilReadOnly;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRenderPassDepthStencilAttachment) == sizeof(WebKit::WebGPU::RenderPassDepthStencilAttachment));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, view)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthClearValue)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthLoadOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthStoreOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthReadOnly)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilClearValue)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilLoadOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilStoreOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilReadOnly)
    >::value);
    encoder << instance.view;
    encoder << instance.depthClearValue;
    encoder << instance.depthLoadOp;
    encoder << instance.depthStoreOp;
    encoder << instance.depthReadOnly;
    encoder << instance.stencilClearValue;
    encoder << instance.stencilLoadOp;
    encoder << instance.stencilStoreOp;
    encoder << instance.stencilReadOnly;
}

void ArgumentCoder<WebKit::WebGPU::RenderPassDepthStencilAttachment>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderPassDepthStencilAttachment& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.view)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthClearValue)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthLoadOp)>, std::optional<WebCore::WebGPU::LoadOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStoreOp)>, std::optional<WebCore::WebGPU::StoreOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthReadOnly)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilClearValue)>, WebCore::WebGPU::StencilValue>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilLoadOp)>, std::optional<WebCore::WebGPU::LoadOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilStoreOp)>, std::optional<WebCore::WebGPU::StoreOp>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stencilReadOnly)>, bool>);
    struct ShouldBeSameSizeAsRenderPassDepthStencilAttachment : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RenderPassDepthStencilAttachment>, false> {
        WebKit::WebGPUIdentifier view;
        float depthClearValue;
        std::optional<WebCore::WebGPU::LoadOp> depthLoadOp;
        std::optional<WebCore::WebGPU::StoreOp> depthStoreOp;
        bool depthReadOnly;
        WebCore::WebGPU::StencilValue stencilClearValue;
        std::optional<WebCore::WebGPU::LoadOp> stencilLoadOp;
        std::optional<WebCore::WebGPU::StoreOp> stencilStoreOp;
        bool stencilReadOnly;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRenderPassDepthStencilAttachment) == sizeof(WebKit::WebGPU::RenderPassDepthStencilAttachment));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, view)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthClearValue)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthLoadOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthStoreOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, depthReadOnly)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilClearValue)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilLoadOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilStoreOp)
        , offsetof(WebKit::WebGPU::RenderPassDepthStencilAttachment, stencilReadOnly)
    >::value);
    encoder << instance.view;
    encoder << instance.depthClearValue;
    encoder << instance.depthLoadOp;
    encoder << instance.depthStoreOp;
    encoder << instance.depthReadOnly;
    encoder << instance.stencilClearValue;
    encoder << instance.stencilLoadOp;
    encoder << instance.stencilStoreOp;
    encoder << instance.stencilReadOnly;
}

std::optional<WebKit::WebGPU::RenderPassDepthStencilAttachment> ArgumentCoder<WebKit::WebGPU::RenderPassDepthStencilAttachment>::decode(Decoder& decoder)
{
    auto view = decoder.decode<WebKit::WebGPUIdentifier>();
    auto depthClearValue = decoder.decode<float>();
    auto depthLoadOp = decoder.decode<std::optional<WebCore::WebGPU::LoadOp>>();
    auto depthStoreOp = decoder.decode<std::optional<WebCore::WebGPU::StoreOp>>();
    auto depthReadOnly = decoder.decode<bool>();
    auto stencilClearValue = decoder.decode<WebCore::WebGPU::StencilValue>();
    auto stencilLoadOp = decoder.decode<std::optional<WebCore::WebGPU::LoadOp>>();
    auto stencilStoreOp = decoder.decode<std::optional<WebCore::WebGPU::StoreOp>>();
    auto stencilReadOnly = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderPassDepthStencilAttachment {
            WTFMove(*view),
            WTFMove(*depthClearValue),
            WTFMove(*depthLoadOp),
            WTFMove(*depthStoreOp),
            WTFMove(*depthReadOnly),
            WTFMove(*stencilClearValue),
            WTFMove(*stencilLoadOp),
            WTFMove(*stencilStoreOp),
            WTFMove(*stencilReadOnly)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderPassDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::RenderPassDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorAttachments)>, Vector<std::optional<WebKit::WebGPU::RenderPassColorAttachment>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencilAttachment)>, std::optional<WebKit::WebGPU::RenderPassDepthStencilAttachment>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.occlusionQuerySet)>, std::optional<WebKit::WebGPUIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestampWrites)>, WebKit::WebGPU::RenderPassTimestampWrites>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, colorAttachments)
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, depthStencilAttachment)
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, occlusionQuerySet)
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, timestampWrites)
    >::value);
    encoder << instance.label;
    encoder << instance.colorAttachments;
    encoder << instance.depthStencilAttachment;
    encoder << instance.occlusionQuerySet;
    encoder << instance.timestampWrites;
}

void ArgumentCoder<WebKit::WebGPU::RenderPassDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderPassDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorAttachments)>, Vector<std::optional<WebKit::WebGPU::RenderPassColorAttachment>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencilAttachment)>, std::optional<WebKit::WebGPU::RenderPassDepthStencilAttachment>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.occlusionQuerySet)>, std::optional<WebKit::WebGPUIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.timestampWrites)>, WebKit::WebGPU::RenderPassTimestampWrites>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, colorAttachments)
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, depthStencilAttachment)
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, occlusionQuerySet)
        , offsetof(WebKit::WebGPU::RenderPassDescriptor, timestampWrites)
    >::value);
    encoder << instance.label;
    encoder << instance.colorAttachments;
    encoder << instance.depthStencilAttachment;
    encoder << instance.occlusionQuerySet;
    encoder << instance.timestampWrites;
}

std::optional<WebKit::WebGPU::RenderPassDescriptor> ArgumentCoder<WebKit::WebGPU::RenderPassDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto colorAttachments = decoder.decode<Vector<std::optional<WebKit::WebGPU::RenderPassColorAttachment>>>();
    auto depthStencilAttachment = decoder.decode<std::optional<WebKit::WebGPU::RenderPassDepthStencilAttachment>>();
    auto occlusionQuerySet = decoder.decode<std::optional<WebKit::WebGPUIdentifier>>();
    auto timestampWrites = decoder.decode<WebKit::WebGPU::RenderPassTimestampWrites>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderPassDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*colorAttachments),
            WTFMove(*depthStencilAttachment),
            WTFMove(*occlusionQuerySet),
            WTFMove(*timestampWrites)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderPassLayout>::encode(Encoder& encoder, const WebKit::WebGPU::RenderPassLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorFormats)>, Vector<std::optional<WebCore::WebGPU::TextureFormat>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencilFormat)>, std::optional<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleCount)>, WebCore::WebGPU::Size32>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassLayout, colorFormats)
        , offsetof(WebKit::WebGPU::RenderPassLayout, depthStencilFormat)
        , offsetof(WebKit::WebGPU::RenderPassLayout, sampleCount)
    >::value);
    encoder << instance.label;
    encoder << instance.colorFormats;
    encoder << instance.depthStencilFormat;
    encoder << instance.sampleCount;
}

void ArgumentCoder<WebKit::WebGPU::RenderPassLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderPassLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.colorFormats)>, Vector<std::optional<WebCore::WebGPU::TextureFormat>>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencilFormat)>, std::optional<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleCount)>, WebCore::WebGPU::Size32>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassLayout, colorFormats)
        , offsetof(WebKit::WebGPU::RenderPassLayout, depthStencilFormat)
        , offsetof(WebKit::WebGPU::RenderPassLayout, sampleCount)
    >::value);
    encoder << instance.label;
    encoder << instance.colorFormats;
    encoder << instance.depthStencilFormat;
    encoder << instance.sampleCount;
}

std::optional<WebKit::WebGPU::RenderPassLayout> ArgumentCoder<WebKit::WebGPU::RenderPassLayout>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto colorFormats = decoder.decode<Vector<std::optional<WebCore::WebGPU::TextureFormat>>>();
    auto depthStencilFormat = decoder.decode<std::optional<WebCore::WebGPU::TextureFormat>>();
    auto sampleCount = decoder.decode<WebCore::WebGPU::Size32>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderPassLayout {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*colorFormats),
            WTFMove(*depthStencilFormat),
            WTFMove(*sampleCount)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderPassTimestampWrite>::encode(Encoder& encoder, const WebKit::WebGPU::RenderPassTimestampWrite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.querySet)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.queryIndex)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location)>, WebCore::WebGPU::RenderPassTimestampLocation>);
    struct ShouldBeSameSizeAsRenderPassTimestampWrite : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RenderPassTimestampWrite>, false> {
        WebKit::WebGPUIdentifier querySet;
        WebCore::WebGPU::Size32 queryIndex;
        WebCore::WebGPU::RenderPassTimestampLocation location;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRenderPassTimestampWrite) == sizeof(WebKit::WebGPU::RenderPassTimestampWrite));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassTimestampWrite, querySet)
        , offsetof(WebKit::WebGPU::RenderPassTimestampWrite, queryIndex)
        , offsetof(WebKit::WebGPU::RenderPassTimestampWrite, location)
    >::value);
    encoder << instance.querySet;
    encoder << instance.queryIndex;
    encoder << instance.location;
}

void ArgumentCoder<WebKit::WebGPU::RenderPassTimestampWrite>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderPassTimestampWrite& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.querySet)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.queryIndex)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.location)>, WebCore::WebGPU::RenderPassTimestampLocation>);
    struct ShouldBeSameSizeAsRenderPassTimestampWrite : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RenderPassTimestampWrite>, false> {
        WebKit::WebGPUIdentifier querySet;
        WebCore::WebGPU::Size32 queryIndex;
        WebCore::WebGPU::RenderPassTimestampLocation location;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRenderPassTimestampWrite) == sizeof(WebKit::WebGPU::RenderPassTimestampWrite));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPassTimestampWrite, querySet)
        , offsetof(WebKit::WebGPU::RenderPassTimestampWrite, queryIndex)
        , offsetof(WebKit::WebGPU::RenderPassTimestampWrite, location)
    >::value);
    encoder << instance.querySet;
    encoder << instance.queryIndex;
    encoder << instance.location;
}

std::optional<WebKit::WebGPU::RenderPassTimestampWrite> ArgumentCoder<WebKit::WebGPU::RenderPassTimestampWrite>::decode(Decoder& decoder)
{
    auto querySet = decoder.decode<WebKit::WebGPUIdentifier>();
    auto queryIndex = decoder.decode<WebCore::WebGPU::Size32>();
    auto location = decoder.decode<WebCore::WebGPU::RenderPassTimestampLocation>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderPassTimestampWrite {
            WTFMove(*querySet),
            WTFMove(*queryIndex),
            WTFMove(*location)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RenderPipelineDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::RenderPipelineDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layout)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineDescriptorBase, layout)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.vertex)>, WebKit::WebGPU::VertexState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primitive)>, std::optional<WebKit::WebGPU::PrimitiveState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencil)>, std::optional<WebKit::WebGPU::DepthStencilState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multisample)>, std::optional<WebKit::WebGPU::MultisampleState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fragment)>, std::optional<WebKit::WebGPU::FragmentState>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, vertex)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, primitive)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, depthStencil)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, multisample)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, fragment)
    >::value);
    encoder << instance.label;
    encoder << instance.layout;
    encoder << instance.vertex;
    encoder << instance.primitive;
    encoder << instance.depthStencil;
    encoder << instance.multisample;
    encoder << instance.fragment;
}

void ArgumentCoder<WebKit::WebGPU::RenderPipelineDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RenderPipelineDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.layout)>, WebKit::WebGPUIdentifier>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::PipelineDescriptorBase, layout)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.vertex)>, WebKit::WebGPU::VertexState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.primitive)>, std::optional<WebKit::WebGPU::PrimitiveState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthStencil)>, std::optional<WebKit::WebGPU::DepthStencilState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multisample)>, std::optional<WebKit::WebGPU::MultisampleState>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.fragment)>, std::optional<WebKit::WebGPU::FragmentState>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, vertex)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, primitive)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, depthStencil)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, multisample)
        , offsetof(WebKit::WebGPU::RenderPipelineDescriptor, fragment)
    >::value);
    encoder << instance.label;
    encoder << instance.layout;
    encoder << instance.vertex;
    encoder << instance.primitive;
    encoder << instance.depthStencil;
    encoder << instance.multisample;
    encoder << instance.fragment;
}

std::optional<WebKit::WebGPU::RenderPipelineDescriptor> ArgumentCoder<WebKit::WebGPU::RenderPipelineDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto layout = decoder.decode<WebKit::WebGPUIdentifier>();
    auto vertex = decoder.decode<WebKit::WebGPU::VertexState>();
    auto primitive = decoder.decode<std::optional<WebKit::WebGPU::PrimitiveState>>();
    auto depthStencil = decoder.decode<std::optional<WebKit::WebGPU::DepthStencilState>>();
    auto multisample = decoder.decode<std::optional<WebKit::WebGPU::MultisampleState>>();
    auto fragment = decoder.decode<std::optional<WebKit::WebGPU::FragmentState>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RenderPipelineDescriptor {
            WebKit::WebGPU::PipelineDescriptorBase {
                WebKit::WebGPU::ObjectDescriptorBase {
                    WTFMove(*label)
                },
                WTFMove(*layout)
            },
            WTFMove(*vertex),
            WTFMove(*primitive),
            WTFMove(*depthStencil),
            WTFMove(*multisample),
            WTFMove(*fragment)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::RequestAdapterOptions>::encode(Encoder& encoder, const WebKit::WebGPU::RequestAdapterOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.powerPreference)>, std::optional<WebCore::WebGPU::PowerPreference>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.forceFallbackAdapter)>, bool>);
    struct ShouldBeSameSizeAsRequestAdapterOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RequestAdapterOptions>, false> {
        std::optional<WebCore::WebGPU::PowerPreference> powerPreference;
        bool forceFallbackAdapter;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRequestAdapterOptions) == sizeof(WebKit::WebGPU::RequestAdapterOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RequestAdapterOptions, powerPreference)
        , offsetof(WebKit::WebGPU::RequestAdapterOptions, forceFallbackAdapter)
    >::value);
    encoder << instance.powerPreference;
    encoder << instance.forceFallbackAdapter;
}

void ArgumentCoder<WebKit::WebGPU::RequestAdapterOptions>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::RequestAdapterOptions& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.powerPreference)>, std::optional<WebCore::WebGPU::PowerPreference>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.forceFallbackAdapter)>, bool>);
    struct ShouldBeSameSizeAsRequestAdapterOptions : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::RequestAdapterOptions>, false> {
        std::optional<WebCore::WebGPU::PowerPreference> powerPreference;
        bool forceFallbackAdapter;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRequestAdapterOptions) == sizeof(WebKit::WebGPU::RequestAdapterOptions));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::RequestAdapterOptions, powerPreference)
        , offsetof(WebKit::WebGPU::RequestAdapterOptions, forceFallbackAdapter)
    >::value);
    encoder << instance.powerPreference;
    encoder << instance.forceFallbackAdapter;
}

std::optional<WebKit::WebGPU::RequestAdapterOptions> ArgumentCoder<WebKit::WebGPU::RequestAdapterOptions>::decode(Decoder& decoder)
{
    auto powerPreference = decoder.decode<std::optional<WebCore::WebGPU::PowerPreference>>();
    auto forceFallbackAdapter = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::RequestAdapterOptions {
            WTFMove(*powerPreference),
            WTFMove(*forceFallbackAdapter)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::SamplerBindingLayout>::encode(Encoder& encoder, const WebKit::WebGPU::SamplerBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::SamplerBindingType>);
    struct ShouldBeSameSizeAsSamplerBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::SamplerBindingLayout>, false> {
        WebCore::WebGPU::SamplerBindingType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSamplerBindingLayout) == sizeof(WebKit::WebGPU::SamplerBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SamplerBindingLayout, type)
    >::value);
    encoder << instance.type;
}

void ArgumentCoder<WebKit::WebGPU::SamplerBindingLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::SamplerBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.type)>, WebCore::WebGPU::SamplerBindingType>);
    struct ShouldBeSameSizeAsSamplerBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::SamplerBindingLayout>, false> {
        WebCore::WebGPU::SamplerBindingType type;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSamplerBindingLayout) == sizeof(WebKit::WebGPU::SamplerBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SamplerBindingLayout, type)
    >::value);
    encoder << instance.type;
}

std::optional<WebKit::WebGPU::SamplerBindingLayout> ArgumentCoder<WebKit::WebGPU::SamplerBindingLayout>::decode(Decoder& decoder)
{
    auto type = decoder.decode<WebCore::WebGPU::SamplerBindingType>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::SamplerBindingLayout {
            WTFMove(*type)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::SamplerDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::SamplerDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.addressModeU)>, WebCore::WebGPU::AddressMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.addressModeV)>, WebCore::WebGPU::AddressMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.addressModeW)>, WebCore::WebGPU::AddressMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.magFilter)>, WebCore::WebGPU::FilterMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minFilter)>, WebCore::WebGPU::FilterMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipmapFilter)>, WebCore::WebGPU::MipmapFilterMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lodMinClamp)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lodMaxClamp)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compare)>, std::optional<WebCore::WebGPU::CompareFunction>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxAnisotropy)>, uint16_t>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SamplerDescriptor, addressModeU)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, addressModeV)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, addressModeW)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, magFilter)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, minFilter)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, mipmapFilter)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, lodMinClamp)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, lodMaxClamp)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, compare)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, maxAnisotropy)
    >::value);
    encoder << instance.label;
    encoder << instance.addressModeU;
    encoder << instance.addressModeV;
    encoder << instance.addressModeW;
    encoder << instance.magFilter;
    encoder << instance.minFilter;
    encoder << instance.mipmapFilter;
    encoder << instance.lodMinClamp;
    encoder << instance.lodMaxClamp;
    encoder << instance.compare;
    encoder << instance.maxAnisotropy;
}

void ArgumentCoder<WebKit::WebGPU::SamplerDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::SamplerDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.addressModeU)>, WebCore::WebGPU::AddressMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.addressModeV)>, WebCore::WebGPU::AddressMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.addressModeW)>, WebCore::WebGPU::AddressMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.magFilter)>, WebCore::WebGPU::FilterMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minFilter)>, WebCore::WebGPU::FilterMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipmapFilter)>, WebCore::WebGPU::MipmapFilterMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lodMinClamp)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lodMaxClamp)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compare)>, std::optional<WebCore::WebGPU::CompareFunction>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxAnisotropy)>, uint16_t>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SamplerDescriptor, addressModeU)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, addressModeV)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, addressModeW)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, magFilter)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, minFilter)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, mipmapFilter)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, lodMinClamp)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, lodMaxClamp)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, compare)
        , offsetof(WebKit::WebGPU::SamplerDescriptor, maxAnisotropy)
    >::value);
    encoder << instance.label;
    encoder << instance.addressModeU;
    encoder << instance.addressModeV;
    encoder << instance.addressModeW;
    encoder << instance.magFilter;
    encoder << instance.minFilter;
    encoder << instance.mipmapFilter;
    encoder << instance.lodMinClamp;
    encoder << instance.lodMaxClamp;
    encoder << instance.compare;
    encoder << instance.maxAnisotropy;
}

std::optional<WebKit::WebGPU::SamplerDescriptor> ArgumentCoder<WebKit::WebGPU::SamplerDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto addressModeU = decoder.decode<WebCore::WebGPU::AddressMode>();
    auto addressModeV = decoder.decode<WebCore::WebGPU::AddressMode>();
    auto addressModeW = decoder.decode<WebCore::WebGPU::AddressMode>();
    auto magFilter = decoder.decode<WebCore::WebGPU::FilterMode>();
    auto minFilter = decoder.decode<WebCore::WebGPU::FilterMode>();
    auto mipmapFilter = decoder.decode<WebCore::WebGPU::MipmapFilterMode>();
    auto lodMinClamp = decoder.decode<float>();
    auto lodMaxClamp = decoder.decode<float>();
    auto compare = decoder.decode<std::optional<WebCore::WebGPU::CompareFunction>>();
    auto maxAnisotropy = decoder.decode<uint16_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::SamplerDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*addressModeU),
            WTFMove(*addressModeV),
            WTFMove(*addressModeW),
            WTFMove(*magFilter),
            WTFMove(*minFilter),
            WTFMove(*mipmapFilter),
            WTFMove(*lodMinClamp),
            WTFMove(*lodMaxClamp),
            WTFMove(*compare),
            WTFMove(*maxAnisotropy)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ShaderModuleCompilationHint>::encode(Encoder& encoder, const WebKit::WebGPU::ShaderModuleCompilationHint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pipelineLayout)>, WebKit::WebGPUIdentifier>);
    struct ShouldBeSameSizeAsShaderModuleCompilationHint : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ShaderModuleCompilationHint>, false> {
        WebKit::WebGPUIdentifier pipelineLayout;
    };
    static_assert(sizeof(ShouldBeSameSizeAsShaderModuleCompilationHint) == sizeof(WebKit::WebGPU::ShaderModuleCompilationHint));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ShaderModuleCompilationHint, pipelineLayout)
    >::value);
    encoder << instance.pipelineLayout;
}

void ArgumentCoder<WebKit::WebGPU::ShaderModuleCompilationHint>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ShaderModuleCompilationHint& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.pipelineLayout)>, WebKit::WebGPUIdentifier>);
    struct ShouldBeSameSizeAsShaderModuleCompilationHint : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ShaderModuleCompilationHint>, false> {
        WebKit::WebGPUIdentifier pipelineLayout;
    };
    static_assert(sizeof(ShouldBeSameSizeAsShaderModuleCompilationHint) == sizeof(WebKit::WebGPU::ShaderModuleCompilationHint));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ShaderModuleCompilationHint, pipelineLayout)
    >::value);
    encoder << instance.pipelineLayout;
}

std::optional<WebKit::WebGPU::ShaderModuleCompilationHint> ArgumentCoder<WebKit::WebGPU::ShaderModuleCompilationHint>::decode(Decoder& decoder)
{
    auto pipelineLayout = decoder.decode<WebKit::WebGPUIdentifier>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ShaderModuleCompilationHint {
            WTFMove(*pipelineLayout)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ShaderModuleDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::ShaderModuleDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.code)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hints)>, Vector<KeyValuePair<String, WebKit::WebGPU::ShaderModuleCompilationHint>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ShaderModuleDescriptor, code)
        , offsetof(WebKit::WebGPU::ShaderModuleDescriptor, hints)
    >::value);
    encoder << instance.label;
    encoder << instance.code;
    encoder << instance.hints;
}

void ArgumentCoder<WebKit::WebGPU::ShaderModuleDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ShaderModuleDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.code)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hints)>, Vector<KeyValuePair<String, WebKit::WebGPU::ShaderModuleCompilationHint>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ShaderModuleDescriptor, code)
        , offsetof(WebKit::WebGPU::ShaderModuleDescriptor, hints)
    >::value);
    encoder << instance.label;
    encoder << instance.code;
    encoder << instance.hints;
}

std::optional<WebKit::WebGPU::ShaderModuleDescriptor> ArgumentCoder<WebKit::WebGPU::ShaderModuleDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto code = decoder.decode<String>();
    auto hints = decoder.decode<Vector<KeyValuePair<String, WebKit::WebGPU::ShaderModuleCompilationHint>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ShaderModuleDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*code),
            WTFMove(*hints)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::StencilFaceState>::encode(Encoder& encoder, const WebKit::WebGPU::StencilFaceState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compare)>, WebCore::WebGPU::CompareFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failOp)>, WebCore::WebGPU::StencilOperation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthFailOp)>, WebCore::WebGPU::StencilOperation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.passOp)>, WebCore::WebGPU::StencilOperation>);
    struct ShouldBeSameSizeAsStencilFaceState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::StencilFaceState>, false> {
        WebCore::WebGPU::CompareFunction compare;
        WebCore::WebGPU::StencilOperation failOp;
        WebCore::WebGPU::StencilOperation depthFailOp;
        WebCore::WebGPU::StencilOperation passOp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsStencilFaceState) == sizeof(WebKit::WebGPU::StencilFaceState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::StencilFaceState, compare)
        , offsetof(WebKit::WebGPU::StencilFaceState, failOp)
        , offsetof(WebKit::WebGPU::StencilFaceState, depthFailOp)
        , offsetof(WebKit::WebGPU::StencilFaceState, passOp)
    >::value);
    encoder << instance.compare;
    encoder << instance.failOp;
    encoder << instance.depthFailOp;
    encoder << instance.passOp;
}

void ArgumentCoder<WebKit::WebGPU::StencilFaceState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::StencilFaceState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.compare)>, WebCore::WebGPU::CompareFunction>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.failOp)>, WebCore::WebGPU::StencilOperation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.depthFailOp)>, WebCore::WebGPU::StencilOperation>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.passOp)>, WebCore::WebGPU::StencilOperation>);
    struct ShouldBeSameSizeAsStencilFaceState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::StencilFaceState>, false> {
        WebCore::WebGPU::CompareFunction compare;
        WebCore::WebGPU::StencilOperation failOp;
        WebCore::WebGPU::StencilOperation depthFailOp;
        WebCore::WebGPU::StencilOperation passOp;
    };
    static_assert(sizeof(ShouldBeSameSizeAsStencilFaceState) == sizeof(WebKit::WebGPU::StencilFaceState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::StencilFaceState, compare)
        , offsetof(WebKit::WebGPU::StencilFaceState, failOp)
        , offsetof(WebKit::WebGPU::StencilFaceState, depthFailOp)
        , offsetof(WebKit::WebGPU::StencilFaceState, passOp)
    >::value);
    encoder << instance.compare;
    encoder << instance.failOp;
    encoder << instance.depthFailOp;
    encoder << instance.passOp;
}

std::optional<WebKit::WebGPU::StencilFaceState> ArgumentCoder<WebKit::WebGPU::StencilFaceState>::decode(Decoder& decoder)
{
    auto compare = decoder.decode<WebCore::WebGPU::CompareFunction>();
    auto failOp = decoder.decode<WebCore::WebGPU::StencilOperation>();
    auto depthFailOp = decoder.decode<WebCore::WebGPU::StencilOperation>();
    auto passOp = decoder.decode<WebCore::WebGPU::StencilOperation>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::StencilFaceState {
            WTFMove(*compare),
            WTFMove(*failOp),
            WTFMove(*depthFailOp),
            WTFMove(*passOp)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::StorageTextureBindingLayout>::encode(Encoder& encoder, const WebKit::WebGPU::StorageTextureBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.access)>, WebCore::WebGPU::StorageTextureAccess>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewDimension)>, WebCore::WebGPU::TextureViewDimension>);
    struct ShouldBeSameSizeAsStorageTextureBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::StorageTextureBindingLayout>, false> {
        WebCore::WebGPU::StorageTextureAccess access;
        WebCore::WebGPU::TextureFormat format;
        WebCore::WebGPU::TextureViewDimension viewDimension;
    };
    static_assert(sizeof(ShouldBeSameSizeAsStorageTextureBindingLayout) == sizeof(WebKit::WebGPU::StorageTextureBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::StorageTextureBindingLayout, access)
        , offsetof(WebKit::WebGPU::StorageTextureBindingLayout, format)
        , offsetof(WebKit::WebGPU::StorageTextureBindingLayout, viewDimension)
    >::value);
    encoder << instance.access;
    encoder << instance.format;
    encoder << instance.viewDimension;
}

void ArgumentCoder<WebKit::WebGPU::StorageTextureBindingLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::StorageTextureBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.access)>, WebCore::WebGPU::StorageTextureAccess>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewDimension)>, WebCore::WebGPU::TextureViewDimension>);
    struct ShouldBeSameSizeAsStorageTextureBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::StorageTextureBindingLayout>, false> {
        WebCore::WebGPU::StorageTextureAccess access;
        WebCore::WebGPU::TextureFormat format;
        WebCore::WebGPU::TextureViewDimension viewDimension;
    };
    static_assert(sizeof(ShouldBeSameSizeAsStorageTextureBindingLayout) == sizeof(WebKit::WebGPU::StorageTextureBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::StorageTextureBindingLayout, access)
        , offsetof(WebKit::WebGPU::StorageTextureBindingLayout, format)
        , offsetof(WebKit::WebGPU::StorageTextureBindingLayout, viewDimension)
    >::value);
    encoder << instance.access;
    encoder << instance.format;
    encoder << instance.viewDimension;
}

std::optional<WebKit::WebGPU::StorageTextureBindingLayout> ArgumentCoder<WebKit::WebGPU::StorageTextureBindingLayout>::decode(Decoder& decoder)
{
    auto access = decoder.decode<WebCore::WebGPU::StorageTextureAccess>();
    auto format = decoder.decode<WebCore::WebGPU::TextureFormat>();
    auto viewDimension = decoder.decode<WebCore::WebGPU::TextureViewDimension>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::StorageTextureBindingLayout {
            WTFMove(*access),
            WTFMove(*format),
            WTFMove(*viewDimension)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::SupportedFeatures>::encode(Encoder& encoder, const WebKit::WebGPU::SupportedFeatures& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.features)>, Vector<String>>);
    struct ShouldBeSameSizeAsSupportedFeatures : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::SupportedFeatures>, false> {
        Vector<String> features;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSupportedFeatures) == sizeof(WebKit::WebGPU::SupportedFeatures));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SupportedFeatures, features)
    >::value);
    encoder << instance.features;
}

void ArgumentCoder<WebKit::WebGPU::SupportedFeatures>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::SupportedFeatures& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.features)>, Vector<String>>);
    struct ShouldBeSameSizeAsSupportedFeatures : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::SupportedFeatures>, false> {
        Vector<String> features;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSupportedFeatures) == sizeof(WebKit::WebGPU::SupportedFeatures));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SupportedFeatures, features)
    >::value);
    encoder << instance.features;
}

std::optional<WebKit::WebGPU::SupportedFeatures> ArgumentCoder<WebKit::WebGPU::SupportedFeatures>::decode(Decoder& decoder)
{
    auto features = decoder.decode<Vector<String>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::SupportedFeatures {
            WTFMove(*features)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::SupportedLimits>::encode(Encoder& encoder, const WebKit::WebGPU::SupportedLimits& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureDimension1D)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureDimension2D)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureDimension3D)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureArrayLayers)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxBindGroups)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxBindingsPerBindGroup)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxDynamicUniformBuffersPerPipelineLayout)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxDynamicStorageBuffersPerPipelineLayout)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxSampledTexturesPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxSamplersPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxStorageBuffersPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxStorageTexturesPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxUniformBuffersPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxUniformBufferBindingSize)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxStorageBufferBindingSize)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minUniformBufferOffsetAlignment)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minStorageBufferOffsetAlignment)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxVertexBuffers)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxBufferSize)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxVertexAttributes)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxVertexBufferArrayStride)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxInterStageShaderComponents)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxInterStageShaderVariables)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxColorAttachments)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxColorAttachmentBytesPerSample)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupStorageSize)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeInvocationsPerWorkgroup)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupSizeX)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupSizeY)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupSizeZ)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupsPerDimension)>, uint32_t>);
    struct ShouldBeSameSizeAsSupportedLimits : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::SupportedLimits>, false> {
        uint32_t maxTextureDimension1D;
        uint32_t maxTextureDimension2D;
        uint32_t maxTextureDimension3D;
        uint32_t maxTextureArrayLayers;
        uint32_t maxBindGroups;
        uint32_t maxBindingsPerBindGroup;
        uint32_t maxDynamicUniformBuffersPerPipelineLayout;
        uint32_t maxDynamicStorageBuffersPerPipelineLayout;
        uint32_t maxSampledTexturesPerShaderStage;
        uint32_t maxSamplersPerShaderStage;
        uint32_t maxStorageBuffersPerShaderStage;
        uint32_t maxStorageTexturesPerShaderStage;
        uint32_t maxUniformBuffersPerShaderStage;
        uint64_t maxUniformBufferBindingSize;
        uint64_t maxStorageBufferBindingSize;
        uint32_t minUniformBufferOffsetAlignment;
        uint32_t minStorageBufferOffsetAlignment;
        uint32_t maxVertexBuffers;
        uint64_t maxBufferSize;
        uint32_t maxVertexAttributes;
        uint32_t maxVertexBufferArrayStride;
        uint32_t maxInterStageShaderComponents;
        uint32_t maxInterStageShaderVariables;
        uint32_t maxColorAttachments;
        uint32_t maxColorAttachmentBytesPerSample;
        uint32_t maxComputeWorkgroupStorageSize;
        uint32_t maxComputeInvocationsPerWorkgroup;
        uint32_t maxComputeWorkgroupSizeX;
        uint32_t maxComputeWorkgroupSizeY;
        uint32_t maxComputeWorkgroupSizeZ;
        uint32_t maxComputeWorkgroupsPerDimension;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSupportedLimits) == sizeof(WebKit::WebGPU::SupportedLimits));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureDimension1D)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureDimension2D)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureDimension3D)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureArrayLayers)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxBindGroups)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxBindingsPerBindGroup)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxDynamicUniformBuffersPerPipelineLayout)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxDynamicStorageBuffersPerPipelineLayout)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxSampledTexturesPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxSamplersPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxStorageBuffersPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxStorageTexturesPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxUniformBuffersPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxUniformBufferBindingSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxStorageBufferBindingSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, minUniformBufferOffsetAlignment)
        , offsetof(WebKit::WebGPU::SupportedLimits, minStorageBufferOffsetAlignment)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxVertexBuffers)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxBufferSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxVertexAttributes)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxVertexBufferArrayStride)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxInterStageShaderComponents)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxInterStageShaderVariables)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxColorAttachments)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxColorAttachmentBytesPerSample)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupStorageSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeInvocationsPerWorkgroup)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupSizeX)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupSizeY)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupSizeZ)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupsPerDimension)
    >::value);
    encoder << instance.maxTextureDimension1D;
    encoder << instance.maxTextureDimension2D;
    encoder << instance.maxTextureDimension3D;
    encoder << instance.maxTextureArrayLayers;
    encoder << instance.maxBindGroups;
    encoder << instance.maxBindingsPerBindGroup;
    encoder << instance.maxDynamicUniformBuffersPerPipelineLayout;
    encoder << instance.maxDynamicStorageBuffersPerPipelineLayout;
    encoder << instance.maxSampledTexturesPerShaderStage;
    encoder << instance.maxSamplersPerShaderStage;
    encoder << instance.maxStorageBuffersPerShaderStage;
    encoder << instance.maxStorageTexturesPerShaderStage;
    encoder << instance.maxUniformBuffersPerShaderStage;
    encoder << instance.maxUniformBufferBindingSize;
    encoder << instance.maxStorageBufferBindingSize;
    encoder << instance.minUniformBufferOffsetAlignment;
    encoder << instance.minStorageBufferOffsetAlignment;
    encoder << instance.maxVertexBuffers;
    encoder << instance.maxBufferSize;
    encoder << instance.maxVertexAttributes;
    encoder << instance.maxVertexBufferArrayStride;
    encoder << instance.maxInterStageShaderComponents;
    encoder << instance.maxInterStageShaderVariables;
    encoder << instance.maxColorAttachments;
    encoder << instance.maxColorAttachmentBytesPerSample;
    encoder << instance.maxComputeWorkgroupStorageSize;
    encoder << instance.maxComputeInvocationsPerWorkgroup;
    encoder << instance.maxComputeWorkgroupSizeX;
    encoder << instance.maxComputeWorkgroupSizeY;
    encoder << instance.maxComputeWorkgroupSizeZ;
    encoder << instance.maxComputeWorkgroupsPerDimension;
}

void ArgumentCoder<WebKit::WebGPU::SupportedLimits>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::SupportedLimits& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureDimension1D)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureDimension2D)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureDimension3D)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTextureArrayLayers)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxBindGroups)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxBindingsPerBindGroup)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxDynamicUniformBuffersPerPipelineLayout)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxDynamicStorageBuffersPerPipelineLayout)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxSampledTexturesPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxSamplersPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxStorageBuffersPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxStorageTexturesPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxUniformBuffersPerShaderStage)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxUniformBufferBindingSize)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxStorageBufferBindingSize)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minUniformBufferOffsetAlignment)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minStorageBufferOffsetAlignment)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxVertexBuffers)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxBufferSize)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxVertexAttributes)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxVertexBufferArrayStride)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxInterStageShaderComponents)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxInterStageShaderVariables)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxColorAttachments)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxColorAttachmentBytesPerSample)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupStorageSize)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeInvocationsPerWorkgroup)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupSizeX)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupSizeY)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupSizeZ)>, uint32_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxComputeWorkgroupsPerDimension)>, uint32_t>);
    struct ShouldBeSameSizeAsSupportedLimits : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::SupportedLimits>, false> {
        uint32_t maxTextureDimension1D;
        uint32_t maxTextureDimension2D;
        uint32_t maxTextureDimension3D;
        uint32_t maxTextureArrayLayers;
        uint32_t maxBindGroups;
        uint32_t maxBindingsPerBindGroup;
        uint32_t maxDynamicUniformBuffersPerPipelineLayout;
        uint32_t maxDynamicStorageBuffersPerPipelineLayout;
        uint32_t maxSampledTexturesPerShaderStage;
        uint32_t maxSamplersPerShaderStage;
        uint32_t maxStorageBuffersPerShaderStage;
        uint32_t maxStorageTexturesPerShaderStage;
        uint32_t maxUniformBuffersPerShaderStage;
        uint64_t maxUniformBufferBindingSize;
        uint64_t maxStorageBufferBindingSize;
        uint32_t minUniformBufferOffsetAlignment;
        uint32_t minStorageBufferOffsetAlignment;
        uint32_t maxVertexBuffers;
        uint64_t maxBufferSize;
        uint32_t maxVertexAttributes;
        uint32_t maxVertexBufferArrayStride;
        uint32_t maxInterStageShaderComponents;
        uint32_t maxInterStageShaderVariables;
        uint32_t maxColorAttachments;
        uint32_t maxColorAttachmentBytesPerSample;
        uint32_t maxComputeWorkgroupStorageSize;
        uint32_t maxComputeInvocationsPerWorkgroup;
        uint32_t maxComputeWorkgroupSizeX;
        uint32_t maxComputeWorkgroupSizeY;
        uint32_t maxComputeWorkgroupSizeZ;
        uint32_t maxComputeWorkgroupsPerDimension;
    };
    static_assert(sizeof(ShouldBeSameSizeAsSupportedLimits) == sizeof(WebKit::WebGPU::SupportedLimits));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureDimension1D)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureDimension2D)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureDimension3D)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxTextureArrayLayers)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxBindGroups)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxBindingsPerBindGroup)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxDynamicUniformBuffersPerPipelineLayout)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxDynamicStorageBuffersPerPipelineLayout)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxSampledTexturesPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxSamplersPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxStorageBuffersPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxStorageTexturesPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxUniformBuffersPerShaderStage)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxUniformBufferBindingSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxStorageBufferBindingSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, minUniformBufferOffsetAlignment)
        , offsetof(WebKit::WebGPU::SupportedLimits, minStorageBufferOffsetAlignment)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxVertexBuffers)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxBufferSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxVertexAttributes)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxVertexBufferArrayStride)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxInterStageShaderComponents)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxInterStageShaderVariables)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxColorAttachments)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxColorAttachmentBytesPerSample)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupStorageSize)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeInvocationsPerWorkgroup)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupSizeX)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupSizeY)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupSizeZ)
        , offsetof(WebKit::WebGPU::SupportedLimits, maxComputeWorkgroupsPerDimension)
    >::value);
    encoder << instance.maxTextureDimension1D;
    encoder << instance.maxTextureDimension2D;
    encoder << instance.maxTextureDimension3D;
    encoder << instance.maxTextureArrayLayers;
    encoder << instance.maxBindGroups;
    encoder << instance.maxBindingsPerBindGroup;
    encoder << instance.maxDynamicUniformBuffersPerPipelineLayout;
    encoder << instance.maxDynamicStorageBuffersPerPipelineLayout;
    encoder << instance.maxSampledTexturesPerShaderStage;
    encoder << instance.maxSamplersPerShaderStage;
    encoder << instance.maxStorageBuffersPerShaderStage;
    encoder << instance.maxStorageTexturesPerShaderStage;
    encoder << instance.maxUniformBuffersPerShaderStage;
    encoder << instance.maxUniformBufferBindingSize;
    encoder << instance.maxStorageBufferBindingSize;
    encoder << instance.minUniformBufferOffsetAlignment;
    encoder << instance.minStorageBufferOffsetAlignment;
    encoder << instance.maxVertexBuffers;
    encoder << instance.maxBufferSize;
    encoder << instance.maxVertexAttributes;
    encoder << instance.maxVertexBufferArrayStride;
    encoder << instance.maxInterStageShaderComponents;
    encoder << instance.maxInterStageShaderVariables;
    encoder << instance.maxColorAttachments;
    encoder << instance.maxColorAttachmentBytesPerSample;
    encoder << instance.maxComputeWorkgroupStorageSize;
    encoder << instance.maxComputeInvocationsPerWorkgroup;
    encoder << instance.maxComputeWorkgroupSizeX;
    encoder << instance.maxComputeWorkgroupSizeY;
    encoder << instance.maxComputeWorkgroupSizeZ;
    encoder << instance.maxComputeWorkgroupsPerDimension;
}

std::optional<WebKit::WebGPU::SupportedLimits> ArgumentCoder<WebKit::WebGPU::SupportedLimits>::decode(Decoder& decoder)
{
    auto maxTextureDimension1D = decoder.decode<uint32_t>();
    auto maxTextureDimension2D = decoder.decode<uint32_t>();
    auto maxTextureDimension3D = decoder.decode<uint32_t>();
    auto maxTextureArrayLayers = decoder.decode<uint32_t>();
    auto maxBindGroups = decoder.decode<uint32_t>();
    auto maxBindingsPerBindGroup = decoder.decode<uint32_t>();
    auto maxDynamicUniformBuffersPerPipelineLayout = decoder.decode<uint32_t>();
    auto maxDynamicStorageBuffersPerPipelineLayout = decoder.decode<uint32_t>();
    auto maxSampledTexturesPerShaderStage = decoder.decode<uint32_t>();
    auto maxSamplersPerShaderStage = decoder.decode<uint32_t>();
    auto maxStorageBuffersPerShaderStage = decoder.decode<uint32_t>();
    auto maxStorageTexturesPerShaderStage = decoder.decode<uint32_t>();
    auto maxUniformBuffersPerShaderStage = decoder.decode<uint32_t>();
    auto maxUniformBufferBindingSize = decoder.decode<uint64_t>();
    auto maxStorageBufferBindingSize = decoder.decode<uint64_t>();
    auto minUniformBufferOffsetAlignment = decoder.decode<uint32_t>();
    auto minStorageBufferOffsetAlignment = decoder.decode<uint32_t>();
    auto maxVertexBuffers = decoder.decode<uint32_t>();
    auto maxBufferSize = decoder.decode<uint64_t>();
    auto maxVertexAttributes = decoder.decode<uint32_t>();
    auto maxVertexBufferArrayStride = decoder.decode<uint32_t>();
    auto maxInterStageShaderComponents = decoder.decode<uint32_t>();
    auto maxInterStageShaderVariables = decoder.decode<uint32_t>();
    auto maxColorAttachments = decoder.decode<uint32_t>();
    auto maxColorAttachmentBytesPerSample = decoder.decode<uint32_t>();
    auto maxComputeWorkgroupStorageSize = decoder.decode<uint32_t>();
    auto maxComputeInvocationsPerWorkgroup = decoder.decode<uint32_t>();
    auto maxComputeWorkgroupSizeX = decoder.decode<uint32_t>();
    auto maxComputeWorkgroupSizeY = decoder.decode<uint32_t>();
    auto maxComputeWorkgroupSizeZ = decoder.decode<uint32_t>();
    auto maxComputeWorkgroupsPerDimension = decoder.decode<uint32_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::SupportedLimits {
            WTFMove(*maxTextureDimension1D),
            WTFMove(*maxTextureDimension2D),
            WTFMove(*maxTextureDimension3D),
            WTFMove(*maxTextureArrayLayers),
            WTFMove(*maxBindGroups),
            WTFMove(*maxBindingsPerBindGroup),
            WTFMove(*maxDynamicUniformBuffersPerPipelineLayout),
            WTFMove(*maxDynamicStorageBuffersPerPipelineLayout),
            WTFMove(*maxSampledTexturesPerShaderStage),
            WTFMove(*maxSamplersPerShaderStage),
            WTFMove(*maxStorageBuffersPerShaderStage),
            WTFMove(*maxStorageTexturesPerShaderStage),
            WTFMove(*maxUniformBuffersPerShaderStage),
            WTFMove(*maxUniformBufferBindingSize),
            WTFMove(*maxStorageBufferBindingSize),
            WTFMove(*minUniformBufferOffsetAlignment),
            WTFMove(*minStorageBufferOffsetAlignment),
            WTFMove(*maxVertexBuffers),
            WTFMove(*maxBufferSize),
            WTFMove(*maxVertexAttributes),
            WTFMove(*maxVertexBufferArrayStride),
            WTFMove(*maxInterStageShaderComponents),
            WTFMove(*maxInterStageShaderVariables),
            WTFMove(*maxColorAttachments),
            WTFMove(*maxColorAttachmentBytesPerSample),
            WTFMove(*maxComputeWorkgroupStorageSize),
            WTFMove(*maxComputeInvocationsPerWorkgroup),
            WTFMove(*maxComputeWorkgroupSizeX),
            WTFMove(*maxComputeWorkgroupSizeY),
            WTFMove(*maxComputeWorkgroupSizeZ),
            WTFMove(*maxComputeWorkgroupsPerDimension)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::TextureBindingLayout>::encode(Encoder& encoder, const WebKit::WebGPU::TextureBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleType)>, WebCore::WebGPU::TextureSampleType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewDimension)>, WebCore::WebGPU::TextureViewDimension>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multisampled)>, bool>);
    struct ShouldBeSameSizeAsTextureBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::TextureBindingLayout>, false> {
        WebCore::WebGPU::TextureSampleType sampleType;
        WebCore::WebGPU::TextureViewDimension viewDimension;
        bool multisampled;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextureBindingLayout) == sizeof(WebKit::WebGPU::TextureBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::TextureBindingLayout, sampleType)
        , offsetof(WebKit::WebGPU::TextureBindingLayout, viewDimension)
        , offsetof(WebKit::WebGPU::TextureBindingLayout, multisampled)
    >::value);
    encoder << instance.sampleType;
    encoder << instance.viewDimension;
    encoder << instance.multisampled;
}

void ArgumentCoder<WebKit::WebGPU::TextureBindingLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::TextureBindingLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleType)>, WebCore::WebGPU::TextureSampleType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewDimension)>, WebCore::WebGPU::TextureViewDimension>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.multisampled)>, bool>);
    struct ShouldBeSameSizeAsTextureBindingLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::TextureBindingLayout>, false> {
        WebCore::WebGPU::TextureSampleType sampleType;
        WebCore::WebGPU::TextureViewDimension viewDimension;
        bool multisampled;
    };
    static_assert(sizeof(ShouldBeSameSizeAsTextureBindingLayout) == sizeof(WebKit::WebGPU::TextureBindingLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::TextureBindingLayout, sampleType)
        , offsetof(WebKit::WebGPU::TextureBindingLayout, viewDimension)
        , offsetof(WebKit::WebGPU::TextureBindingLayout, multisampled)
    >::value);
    encoder << instance.sampleType;
    encoder << instance.viewDimension;
    encoder << instance.multisampled;
}

std::optional<WebKit::WebGPU::TextureBindingLayout> ArgumentCoder<WebKit::WebGPU::TextureBindingLayout>::decode(Decoder& decoder)
{
    auto sampleType = decoder.decode<WebCore::WebGPU::TextureSampleType>();
    auto viewDimension = decoder.decode<WebCore::WebGPU::TextureViewDimension>();
    auto multisampled = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::TextureBindingLayout {
            WTFMove(*sampleType),
            WTFMove(*viewDimension),
            WTFMove(*multisampled)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::TextureDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::TextureDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, WebKit::WebGPU::Extent3D>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevelCount)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleCount)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dimension)>, WebCore::WebGPU::TextureDimension>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usage)>, WebCore::WebGPU::TextureUsageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewFormats)>, Vector<WebCore::WebGPU::TextureFormat>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::TextureDescriptor, size)
        , offsetof(WebKit::WebGPU::TextureDescriptor, mipLevelCount)
        , offsetof(WebKit::WebGPU::TextureDescriptor, sampleCount)
        , offsetof(WebKit::WebGPU::TextureDescriptor, dimension)
        , offsetof(WebKit::WebGPU::TextureDescriptor, format)
        , offsetof(WebKit::WebGPU::TextureDescriptor, usage)
        , offsetof(WebKit::WebGPU::TextureDescriptor, viewFormats)
    >::value);
    encoder << instance.label;
    encoder << instance.size;
    encoder << instance.mipLevelCount;
    encoder << instance.sampleCount;
    encoder << instance.dimension;
    encoder << instance.format;
    encoder << instance.usage;
    encoder << instance.viewFormats;
}

void ArgumentCoder<WebKit::WebGPU::TextureDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::TextureDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, WebKit::WebGPU::Extent3D>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevelCount)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleCount)>, WebCore::WebGPU::Size32>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dimension)>, WebCore::WebGPU::TextureDimension>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::TextureFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.usage)>, WebCore::WebGPU::TextureUsageFlags>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.viewFormats)>, Vector<WebCore::WebGPU::TextureFormat>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::TextureDescriptor, size)
        , offsetof(WebKit::WebGPU::TextureDescriptor, mipLevelCount)
        , offsetof(WebKit::WebGPU::TextureDescriptor, sampleCount)
        , offsetof(WebKit::WebGPU::TextureDescriptor, dimension)
        , offsetof(WebKit::WebGPU::TextureDescriptor, format)
        , offsetof(WebKit::WebGPU::TextureDescriptor, usage)
        , offsetof(WebKit::WebGPU::TextureDescriptor, viewFormats)
    >::value);
    encoder << instance.label;
    encoder << instance.size;
    encoder << instance.mipLevelCount;
    encoder << instance.sampleCount;
    encoder << instance.dimension;
    encoder << instance.format;
    encoder << instance.usage;
    encoder << instance.viewFormats;
}

std::optional<WebKit::WebGPU::TextureDescriptor> ArgumentCoder<WebKit::WebGPU::TextureDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto size = decoder.decode<WebKit::WebGPU::Extent3D>();
    auto mipLevelCount = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto sampleCount = decoder.decode<WebCore::WebGPU::Size32>();
    auto dimension = decoder.decode<WebCore::WebGPU::TextureDimension>();
    auto format = decoder.decode<WebCore::WebGPU::TextureFormat>();
    auto usage = decoder.decode<WebCore::WebGPU::TextureUsageFlags>();
    auto viewFormats = decoder.decode<Vector<WebCore::WebGPU::TextureFormat>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::TextureDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*size),
            WTFMove(*mipLevelCount),
            WTFMove(*sampleCount),
            WTFMove(*dimension),
            WTFMove(*format),
            WTFMove(*usage),
            WTFMove(*viewFormats)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::TextureViewDescriptor>::encode(Encoder& encoder, const WebKit::WebGPU::TextureViewDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, std::optional<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dimension)>, std::optional<WebCore::WebGPU::TextureViewDimension>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspect)>, WebCore::WebGPU::TextureAspect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseMipLevel)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevelCount)>, std::optional<WebCore::WebGPU::IntegerCoordinate>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseArrayLayer)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.arrayLayerCount)>, std::optional<WebCore::WebGPU::IntegerCoordinate>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, format)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, dimension)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, aspect)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, baseMipLevel)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, mipLevelCount)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, baseArrayLayer)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, arrayLayerCount)
    >::value);
    encoder << instance.label;
    encoder << instance.format;
    encoder << instance.dimension;
    encoder << instance.aspect;
    encoder << instance.baseMipLevel;
    encoder << instance.mipLevelCount;
    encoder << instance.baseArrayLayer;
    encoder << instance.arrayLayerCount;
}

void ArgumentCoder<WebKit::WebGPU::TextureViewDescriptor>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::TextureViewDescriptor& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.label)>, String>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ObjectDescriptorBase, label)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, std::optional<WebCore::WebGPU::TextureFormat>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.dimension)>, std::optional<WebCore::WebGPU::TextureViewDimension>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.aspect)>, WebCore::WebGPU::TextureAspect>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseMipLevel)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.mipLevelCount)>, std::optional<WebCore::WebGPU::IntegerCoordinate>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.baseArrayLayer)>, WebCore::WebGPU::IntegerCoordinate>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.arrayLayerCount)>, std::optional<WebCore::WebGPU::IntegerCoordinate>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, format)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, dimension)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, aspect)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, baseMipLevel)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, mipLevelCount)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, baseArrayLayer)
        , offsetof(WebKit::WebGPU::TextureViewDescriptor, arrayLayerCount)
    >::value);
    encoder << instance.label;
    encoder << instance.format;
    encoder << instance.dimension;
    encoder << instance.aspect;
    encoder << instance.baseMipLevel;
    encoder << instance.mipLevelCount;
    encoder << instance.baseArrayLayer;
    encoder << instance.arrayLayerCount;
}

std::optional<WebKit::WebGPU::TextureViewDescriptor> ArgumentCoder<WebKit::WebGPU::TextureViewDescriptor>::decode(Decoder& decoder)
{
    auto label = decoder.decode<String>();
    auto format = decoder.decode<std::optional<WebCore::WebGPU::TextureFormat>>();
    auto dimension = decoder.decode<std::optional<WebCore::WebGPU::TextureViewDimension>>();
    auto aspect = decoder.decode<WebCore::WebGPU::TextureAspect>();
    auto baseMipLevel = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto mipLevelCount = decoder.decode<std::optional<WebCore::WebGPU::IntegerCoordinate>>();
    auto baseArrayLayer = decoder.decode<WebCore::WebGPU::IntegerCoordinate>();
    auto arrayLayerCount = decoder.decode<std::optional<WebCore::WebGPU::IntegerCoordinate>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::TextureViewDescriptor {
            WebKit::WebGPU::ObjectDescriptorBase {
                WTFMove(*label)
            },
            WTFMove(*format),
            WTFMove(*dimension),
            WTFMove(*aspect),
            WTFMove(*baseMipLevel),
            WTFMove(*mipLevelCount),
            WTFMove(*baseArrayLayer),
            WTFMove(*arrayLayerCount)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::ValidationError>::encode(Encoder& encoder, const WebKit::WebGPU::ValidationError& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, String>);
    struct ShouldBeSameSizeAsValidationError : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ValidationError>, false> {
        String message;
    };
    static_assert(sizeof(ShouldBeSameSizeAsValidationError) == sizeof(WebKit::WebGPU::ValidationError));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ValidationError, message)
    >::value);
    encoder << instance.message;
}

void ArgumentCoder<WebKit::WebGPU::ValidationError>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::ValidationError& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.message)>, String>);
    struct ShouldBeSameSizeAsValidationError : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::ValidationError>, false> {
        String message;
    };
    static_assert(sizeof(ShouldBeSameSizeAsValidationError) == sizeof(WebKit::WebGPU::ValidationError));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ValidationError, message)
    >::value);
    encoder << instance.message;
}

std::optional<WebKit::WebGPU::ValidationError> ArgumentCoder<WebKit::WebGPU::ValidationError>::decode(Decoder& decoder)
{
    auto message = decoder.decode<String>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::ValidationError {
            WTFMove(*message)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::VertexAttribute>::encode(Encoder& encoder, const WebKit::WebGPU::VertexAttribute& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::VertexFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shaderLocation)>, WebCore::WebGPU::Index32>);
    struct ShouldBeSameSizeAsVertexAttribute : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::VertexAttribute>, false> {
        WebCore::WebGPU::VertexFormat format;
        WebCore::WebGPU::Size64 offset;
        WebCore::WebGPU::Index32 shaderLocation;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVertexAttribute) == sizeof(WebKit::WebGPU::VertexAttribute));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::VertexAttribute, format)
        , offsetof(WebKit::WebGPU::VertexAttribute, offset)
        , offsetof(WebKit::WebGPU::VertexAttribute, shaderLocation)
    >::value);
    encoder << instance.format;
    encoder << instance.offset;
    encoder << instance.shaderLocation;
}

void ArgumentCoder<WebKit::WebGPU::VertexAttribute>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::VertexAttribute& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.format)>, WebCore::WebGPU::VertexFormat>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.offset)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shaderLocation)>, WebCore::WebGPU::Index32>);
    struct ShouldBeSameSizeAsVertexAttribute : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::VertexAttribute>, false> {
        WebCore::WebGPU::VertexFormat format;
        WebCore::WebGPU::Size64 offset;
        WebCore::WebGPU::Index32 shaderLocation;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVertexAttribute) == sizeof(WebKit::WebGPU::VertexAttribute));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::VertexAttribute, format)
        , offsetof(WebKit::WebGPU::VertexAttribute, offset)
        , offsetof(WebKit::WebGPU::VertexAttribute, shaderLocation)
    >::value);
    encoder << instance.format;
    encoder << instance.offset;
    encoder << instance.shaderLocation;
}

std::optional<WebKit::WebGPU::VertexAttribute> ArgumentCoder<WebKit::WebGPU::VertexAttribute>::decode(Decoder& decoder)
{
    auto format = decoder.decode<WebCore::WebGPU::VertexFormat>();
    auto offset = decoder.decode<WebCore::WebGPU::Size64>();
    auto shaderLocation = decoder.decode<WebCore::WebGPU::Index32>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::VertexAttribute {
            WTFMove(*format),
            WTFMove(*offset),
            WTFMove(*shaderLocation)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::VertexBufferLayout>::encode(Encoder& encoder, const WebKit::WebGPU::VertexBufferLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.arrayStride)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stepMode)>, WebCore::WebGPU::VertexStepMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.attributes)>, Vector<WebKit::WebGPU::VertexAttribute>>);
    struct ShouldBeSameSizeAsVertexBufferLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::VertexBufferLayout>, false> {
        WebCore::WebGPU::Size64 arrayStride;
        WebCore::WebGPU::VertexStepMode stepMode;
        Vector<WebKit::WebGPU::VertexAttribute> attributes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVertexBufferLayout) == sizeof(WebKit::WebGPU::VertexBufferLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::VertexBufferLayout, arrayStride)
        , offsetof(WebKit::WebGPU::VertexBufferLayout, stepMode)
        , offsetof(WebKit::WebGPU::VertexBufferLayout, attributes)
    >::value);
    encoder << instance.arrayStride;
    encoder << instance.stepMode;
    encoder << instance.attributes;
}

void ArgumentCoder<WebKit::WebGPU::VertexBufferLayout>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::VertexBufferLayout& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.arrayStride)>, WebCore::WebGPU::Size64>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.stepMode)>, WebCore::WebGPU::VertexStepMode>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.attributes)>, Vector<WebKit::WebGPU::VertexAttribute>>);
    struct ShouldBeSameSizeAsVertexBufferLayout : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebGPU::VertexBufferLayout>, false> {
        WebCore::WebGPU::Size64 arrayStride;
        WebCore::WebGPU::VertexStepMode stepMode;
        Vector<WebKit::WebGPU::VertexAttribute> attributes;
    };
    static_assert(sizeof(ShouldBeSameSizeAsVertexBufferLayout) == sizeof(WebKit::WebGPU::VertexBufferLayout));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::VertexBufferLayout, arrayStride)
        , offsetof(WebKit::WebGPU::VertexBufferLayout, stepMode)
        , offsetof(WebKit::WebGPU::VertexBufferLayout, attributes)
    >::value);
    encoder << instance.arrayStride;
    encoder << instance.stepMode;
    encoder << instance.attributes;
}

std::optional<WebKit::WebGPU::VertexBufferLayout> ArgumentCoder<WebKit::WebGPU::VertexBufferLayout>::decode(Decoder& decoder)
{
    auto arrayStride = decoder.decode<WebCore::WebGPU::Size64>();
    auto stepMode = decoder.decode<WebCore::WebGPU::VertexStepMode>();
    auto attributes = decoder.decode<Vector<WebKit::WebGPU::VertexAttribute>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::VertexBufferLayout {
            WTFMove(*arrayStride),
            WTFMove(*stepMode),
            WTFMove(*attributes)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::WebGPU::VertexState>::encode(Encoder& encoder, const WebKit::WebGPU::VertexState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.module)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entryPoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constants)>, Vector<KeyValuePair<String, double>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ProgrammableStage, module)
        , offsetof(WebKit::WebGPU::ProgrammableStage, entryPoint)
        , offsetof(WebKit::WebGPU::ProgrammableStage, constants)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffers)>, Vector<std::optional<WebKit::WebGPU::VertexBufferLayout>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::VertexState, buffers)
    >::value);
    encoder << instance.module;
    encoder << instance.entryPoint;
    encoder << instance.constants;
    encoder << instance.buffers;
}

void ArgumentCoder<WebKit::WebGPU::VertexState>::encode(StreamConnectionEncoder& encoder, const WebKit::WebGPU::VertexState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.module)>, WebKit::WebGPUIdentifier>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entryPoint)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.constants)>, Vector<KeyValuePair<String, double>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::ProgrammableStage, module)
        , offsetof(WebKit::WebGPU::ProgrammableStage, entryPoint)
        , offsetof(WebKit::WebGPU::ProgrammableStage, constants)
    >::value);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.buffers)>, Vector<std::optional<WebKit::WebGPU::VertexBufferLayout>>>);
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebGPU::VertexState, buffers)
    >::value);
    encoder << instance.module;
    encoder << instance.entryPoint;
    encoder << instance.constants;
    encoder << instance.buffers;
}

std::optional<WebKit::WebGPU::VertexState> ArgumentCoder<WebKit::WebGPU::VertexState>::decode(Decoder& decoder)
{
    auto module = decoder.decode<WebKit::WebGPUIdentifier>();
    auto entryPoint = decoder.decode<String>();
    auto constants = decoder.decode<Vector<KeyValuePair<String, double>>>();
    auto buffers = decoder.decode<Vector<std::optional<WebKit::WebGPU::VertexBufferLayout>>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebGPU::VertexState {
            WebKit::WebGPU::ProgrammableStage {
                WTFMove(*module),
                WTFMove(*entryPoint),
                WTFMove(*constants)
            },
            WTFMove(*buffers)
        }
    };
}

#endif

void ArgumentCoder<WebKit::WebsiteData::Entry>::encode(Encoder& encoder, const WebKit::WebsiteData::Entry& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.origin)>, WebCore::SecurityOriginData>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.typeAsOptionSet())>, OptionSet<WebKit::WebsiteDataType>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.size)>, uint64_t>);
    encoder << instance.origin;
    encoder << instance.typeAsOptionSet();
    encoder << instance.size;
}

std::optional<WebKit::WebsiteData::Entry> ArgumentCoder<WebKit::WebsiteData::Entry>::decode(Decoder& decoder)
{
    auto origin = decoder.decode<WebCore::SecurityOriginData>();
    auto typeAsOptionSet = decoder.decode<OptionSet<WebKit::WebsiteDataType>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;

    if (!(typeAsOptionSet->hasExactlyOneBitSet()))
        return std::nullopt;
    auto size = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebsiteData::Entry {
            WTFMove(*origin),
            WTFMove(*typeAsOptionSet),
            WTFMove(*size)
        }
    };
}

void ArgumentCoder<WebKit::WebsiteData>::encode(Encoder& encoder, const WebKit::WebsiteData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.entries)>, Vector<WebKit::WebsiteData::Entry>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hostNamesWithCookies)>, HashSet<String>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hostNamesWithHSTSCache)>, HashSet<String>>);
#if ENABLE(TRACKING_PREVENTION)
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.registrableDomainsWithResourceLoadStatistics)>, HashSet<WebCore::RegistrableDomain>>);
#endif
    struct ShouldBeSameSizeAsWebsiteData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebsiteData>, false> {
        Vector<WebKit::WebsiteData::Entry> entries;
        HashSet<String> hostNamesWithCookies;
        HashSet<String> hostNamesWithHSTSCache;
#if ENABLE(TRACKING_PREVENTION)
        HashSet<WebCore::RegistrableDomain> registrableDomainsWithResourceLoadStatistics;
#endif
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebsiteData) == sizeof(WebKit::WebsiteData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebsiteData, entries)
        , offsetof(WebKit::WebsiteData, hostNamesWithCookies)
        , offsetof(WebKit::WebsiteData, hostNamesWithHSTSCache)
#if ENABLE(TRACKING_PREVENTION)
        , offsetof(WebKit::WebsiteData, registrableDomainsWithResourceLoadStatistics)
#endif
    >::value);
    encoder << instance.entries;
    encoder << instance.hostNamesWithCookies;
    encoder << instance.hostNamesWithHSTSCache;
#if ENABLE(TRACKING_PREVENTION)
    encoder << instance.registrableDomainsWithResourceLoadStatistics;
#endif
}

std::optional<WebKit::WebsiteData> ArgumentCoder<WebKit::WebsiteData>::decode(Decoder& decoder)
{
    auto entries = decoder.decode<Vector<WebKit::WebsiteData::Entry>>();
    auto hostNamesWithCookies = decoder.decode<HashSet<String>>();
    auto hostNamesWithHSTSCache = decoder.decode<HashSet<String>>();
#if ENABLE(TRACKING_PREVENTION)
    auto registrableDomainsWithResourceLoadStatistics = decoder.decode<HashSet<WebCore::RegistrableDomain>>();
#endif
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebsiteData {
            WTFMove(*entries),
            WTFMove(*hostNamesWithCookies),
            WTFMove(*hostNamesWithHSTSCache),
#if ENABLE(TRACKING_PREVENTION)
            WTFMove(*registrableDomainsWithResourceLoadStatistics)
#endif
        }
    };
}

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::BufferIdentifierSet>::encode(Encoder& encoder, const WebKit::BufferIdentifierSet& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.front)>, std::optional<WebCore::RenderingResourceIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.back)>, std::optional<WebCore::RenderingResourceIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.secondaryBack)>, std::optional<WebCore::RenderingResourceIdentifier>>);
    struct ShouldBeSameSizeAsBufferIdentifierSet : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::BufferIdentifierSet>, false> {
        std::optional<WebCore::RenderingResourceIdentifier> front;
        std::optional<WebCore::RenderingResourceIdentifier> back;
        std::optional<WebCore::RenderingResourceIdentifier> secondaryBack;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBufferIdentifierSet) == sizeof(WebKit::BufferIdentifierSet));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::BufferIdentifierSet, front)
        , offsetof(WebKit::BufferIdentifierSet, back)
        , offsetof(WebKit::BufferIdentifierSet, secondaryBack)
    >::value);
    encoder << instance.front;
    encoder << instance.back;
    encoder << instance.secondaryBack;
}

void ArgumentCoder<WebKit::BufferIdentifierSet>::encode(StreamConnectionEncoder& encoder, const WebKit::BufferIdentifierSet& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.front)>, std::optional<WebCore::RenderingResourceIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.back)>, std::optional<WebCore::RenderingResourceIdentifier>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.secondaryBack)>, std::optional<WebCore::RenderingResourceIdentifier>>);
    struct ShouldBeSameSizeAsBufferIdentifierSet : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::BufferIdentifierSet>, false> {
        std::optional<WebCore::RenderingResourceIdentifier> front;
        std::optional<WebCore::RenderingResourceIdentifier> back;
        std::optional<WebCore::RenderingResourceIdentifier> secondaryBack;
    };
    static_assert(sizeof(ShouldBeSameSizeAsBufferIdentifierSet) == sizeof(WebKit::BufferIdentifierSet));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::BufferIdentifierSet, front)
        , offsetof(WebKit::BufferIdentifierSet, back)
        , offsetof(WebKit::BufferIdentifierSet, secondaryBack)
    >::value);
    encoder << instance.front;
    encoder << instance.back;
    encoder << instance.secondaryBack;
}

std::optional<WebKit::BufferIdentifierSet> ArgumentCoder<WebKit::BufferIdentifierSet>::decode(Decoder& decoder)
{
    auto front = decoder.decode<std::optional<WebCore::RenderingResourceIdentifier>>();
    auto back = decoder.decode<std::optional<WebCore::RenderingResourceIdentifier>>();
    auto secondaryBack = decoder.decode<std::optional<WebCore::RenderingResourceIdentifier>>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::BufferIdentifierSet {
            WTFMove(*front),
            WTFMove(*back),
            WTFMove(*secondaryBack)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && PLATFORM(COCOA)
void ArgumentCoder<WebKit::PrepareBackingStoreBuffersInputData>::encode(Encoder& encoder, const WebKit::PrepareBackingStoreBuffersInputData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bufferSet)>, WebKit::BufferIdentifierSet>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsPartialRepaint)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasEmptyDirtyRegion)>, bool>);
    struct ShouldBeSameSizeAsPrepareBackingStoreBuffersInputData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::PrepareBackingStoreBuffersInputData>, false> {
        WebKit::BufferIdentifierSet bufferSet;
        bool supportsPartialRepaint;
        bool hasEmptyDirtyRegion;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPrepareBackingStoreBuffersInputData) == sizeof(WebKit::PrepareBackingStoreBuffersInputData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::PrepareBackingStoreBuffersInputData, bufferSet)
        , offsetof(WebKit::PrepareBackingStoreBuffersInputData, supportsPartialRepaint)
        , offsetof(WebKit::PrepareBackingStoreBuffersInputData, hasEmptyDirtyRegion)
    >::value);
    encoder << instance.bufferSet;
    encoder << instance.supportsPartialRepaint;
    encoder << instance.hasEmptyDirtyRegion;
}

void ArgumentCoder<WebKit::PrepareBackingStoreBuffersInputData>::encode(StreamConnectionEncoder& encoder, const WebKit::PrepareBackingStoreBuffersInputData& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bufferSet)>, WebKit::BufferIdentifierSet>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsPartialRepaint)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasEmptyDirtyRegion)>, bool>);
    struct ShouldBeSameSizeAsPrepareBackingStoreBuffersInputData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::PrepareBackingStoreBuffersInputData>, false> {
        WebKit::BufferIdentifierSet bufferSet;
        bool supportsPartialRepaint;
        bool hasEmptyDirtyRegion;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPrepareBackingStoreBuffersInputData) == sizeof(WebKit::PrepareBackingStoreBuffersInputData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::PrepareBackingStoreBuffersInputData, bufferSet)
        , offsetof(WebKit::PrepareBackingStoreBuffersInputData, supportsPartialRepaint)
        , offsetof(WebKit::PrepareBackingStoreBuffersInputData, hasEmptyDirtyRegion)
    >::value);
    encoder << instance.bufferSet;
    encoder << instance.supportsPartialRepaint;
    encoder << instance.hasEmptyDirtyRegion;
}

std::optional<WebKit::PrepareBackingStoreBuffersInputData> ArgumentCoder<WebKit::PrepareBackingStoreBuffersInputData>::decode(Decoder& decoder)
{
    auto bufferSet = decoder.decode<WebKit::BufferIdentifierSet>();
    auto supportsPartialRepaint = decoder.decode<bool>();
    auto hasEmptyDirtyRegion = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::PrepareBackingStoreBuffersInputData {
            WTFMove(*bufferSet),
            WTFMove(*supportsPartialRepaint),
            WTFMove(*hasEmptyDirtyRegion)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && PLATFORM(COCOA)
void ArgumentCoder<WebKit::PrepareBackingStoreBuffersOutputData>::encode(Encoder& encoder, WebKit::PrepareBackingStoreBuffersOutputData&& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bufferSet)>, WebKit::BufferIdentifierSet>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.frontBufferHandle)>, std::optional<WebKit::ImageBufferBackendHandle>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.displayRequirement)>, WebKit::SwapBuffersDisplayRequirement>);
    struct ShouldBeSameSizeAsPrepareBackingStoreBuffersOutputData : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::PrepareBackingStoreBuffersOutputData>, false> {
        WebKit::BufferIdentifierSet bufferSet;
        std::optional<WebKit::ImageBufferBackendHandle> frontBufferHandle;
        WebKit::SwapBuffersDisplayRequirement displayRequirement;
    };
    static_assert(sizeof(ShouldBeSameSizeAsPrepareBackingStoreBuffersOutputData) == sizeof(WebKit::PrepareBackingStoreBuffersOutputData));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::PrepareBackingStoreBuffersOutputData, bufferSet)
        , offsetof(WebKit::PrepareBackingStoreBuffersOutputData, frontBufferHandle)
        , offsetof(WebKit::PrepareBackingStoreBuffersOutputData, displayRequirement)
    >::value);
    encoder << WTFMove(instance.bufferSet);
    encoder << WTFMove(instance.frontBufferHandle);
    encoder << WTFMove(instance.displayRequirement);
}

std::optional<WebKit::PrepareBackingStoreBuffersOutputData> ArgumentCoder<WebKit::PrepareBackingStoreBuffersOutputData>::decode(Decoder& decoder)
{
    auto bufferSet = decoder.decode<WebKit::BufferIdentifierSet>();
    auto frontBufferHandle = decoder.decode<std::optional<WebKit::ImageBufferBackendHandle>>();
    auto displayRequirement = decoder.decode<WebKit::SwapBuffersDisplayRequirement>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::PrepareBackingStoreBuffersOutputData {
            WTFMove(*bufferSet),
            WTFMove(*frontBufferHandle),
            WTFMove(*displayRequirement)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
void ArgumentCoder<WebKit::RemoteAudioSessionConfiguration>::encode(Encoder& encoder, const WebKit::RemoteAudioSessionConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.routingContextUID)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.sampleRate)>, float>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bufferSize)>, size_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.numberOfOutputChannels)>, size_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumNumberOfOutputChannels)>, size_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.preferredBufferSize)>, size_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isMuted)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.isActive)>, bool>);
    struct ShouldBeSameSizeAsRemoteAudioSessionConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteAudioSessionConfiguration>, false> {
        String routingContextUID;
        float sampleRate;
        size_t bufferSize;
        size_t numberOfOutputChannels;
        size_t maximumNumberOfOutputChannels;
        size_t preferredBufferSize;
        bool isMuted;
        bool isActive;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteAudioSessionConfiguration) == sizeof(WebKit::RemoteAudioSessionConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteAudioSessionConfiguration, routingContextUID)
        , offsetof(WebKit::RemoteAudioSessionConfiguration, sampleRate)
        , offsetof(WebKit::RemoteAudioSessionConfiguration, bufferSize)
        , offsetof(WebKit::RemoteAudioSessionConfiguration, numberOfOutputChannels)
        , offsetof(WebKit::RemoteAudioSessionConfiguration, maximumNumberOfOutputChannels)
        , offsetof(WebKit::RemoteAudioSessionConfiguration, preferredBufferSize)
        , offsetof(WebKit::RemoteAudioSessionConfiguration, isMuted)
        , offsetof(WebKit::RemoteAudioSessionConfiguration, isActive)
    >::value);
    encoder << instance.routingContextUID;
    encoder << instance.sampleRate;
    encoder << instance.bufferSize;
    encoder << instance.numberOfOutputChannels;
    encoder << instance.maximumNumberOfOutputChannels;
    encoder << instance.preferredBufferSize;
    encoder << instance.isMuted;
    encoder << instance.isActive;
}

std::optional<WebKit::RemoteAudioSessionConfiguration> ArgumentCoder<WebKit::RemoteAudioSessionConfiguration>::decode(Decoder& decoder)
{
    auto routingContextUID = decoder.decode<String>();
    auto sampleRate = decoder.decode<float>();
    auto bufferSize = decoder.decode<size_t>();
    auto numberOfOutputChannels = decoder.decode<size_t>();
    auto maximumNumberOfOutputChannels = decoder.decode<size_t>();
    auto preferredBufferSize = decoder.decode<size_t>();
    auto isMuted = decoder.decode<bool>();
    auto isActive = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::RemoteAudioSessionConfiguration {
            WTFMove(*routingContextUID),
            WTFMove(*sampleRate),
            WTFMove(*bufferSize),
            WTFMove(*numberOfOutputChannels),
            WTFMove(*maximumNumberOfOutputChannels),
            WTFMove(*preferredBufferSize),
            WTFMove(*isMuted),
            WTFMove(*isActive)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
void ArgumentCoder<WebKit::RemoteCDMConfiguration>::encode(Encoder& encoder, const WebKit::RemoteCDMConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportedInitDataTypes)>, Vector<AtomString>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportedRobustnesses)>, Vector<AtomString>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsServerCertificates)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsSessions)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.logIdentifier)>, uint64_t>);
    struct ShouldBeSameSizeAsRemoteCDMConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteCDMConfiguration>, false> {
        Vector<AtomString> supportedInitDataTypes;
        Vector<AtomString> supportedRobustnesses;
        bool supportsServerCertificates;
        bool supportsSessions;
        uint64_t logIdentifier;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteCDMConfiguration) == sizeof(WebKit::RemoteCDMConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteCDMConfiguration, supportedInitDataTypes)
        , offsetof(WebKit::RemoteCDMConfiguration, supportedRobustnesses)
        , offsetof(WebKit::RemoteCDMConfiguration, supportsServerCertificates)
        , offsetof(WebKit::RemoteCDMConfiguration, supportsSessions)
        , offsetof(WebKit::RemoteCDMConfiguration, logIdentifier)
    >::value);
    encoder << instance.supportedInitDataTypes;
    encoder << instance.supportedRobustnesses;
    encoder << instance.supportsServerCertificates;
    encoder << instance.supportsSessions;
    encoder << instance.logIdentifier;
}

std::optional<WebKit::RemoteCDMConfiguration> ArgumentCoder<WebKit::RemoteCDMConfiguration>::decode(Decoder& decoder)
{
    auto supportedInitDataTypes = decoder.decode<Vector<AtomString>>();
    auto supportedRobustnesses = decoder.decode<Vector<AtomString>>();
    auto supportsServerCertificates = decoder.decode<bool>();
    auto supportsSessions = decoder.decode<bool>();
    auto logIdentifier = decoder.decode<uint64_t>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::RemoteCDMConfiguration {
            WTFMove(*supportedInitDataTypes),
            WTFMove(*supportedRobustnesses),
            WTFMove(*supportsServerCertificates),
            WTFMove(*supportsSessions),
            WTFMove(*logIdentifier)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS)
void ArgumentCoder<WebKit::RemoteMediaPlayerConfiguration>::encode(Encoder& encoder, const WebKit::RemoteMediaPlayerConfiguration& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.engineDescription)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maximumDurationToCacheMediaTime)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsScanning)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsFullscreen)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsPictureInPicture)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsAcceleratedRendering)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsPlayAtHostTime)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.supportsPauseAtHostTime)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canPlayToWirelessPlaybackTarget)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.shouldIgnoreIntrinsicSize)>, bool>);
    struct ShouldBeSameSizeAsRemoteMediaPlayerConfiguration : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteMediaPlayerConfiguration>, false> {
        String engineDescription;
        double maximumDurationToCacheMediaTime;
        bool supportsScanning;
        bool supportsFullscreen;
        bool supportsPictureInPicture;
        bool supportsAcceleratedRendering;
        bool supportsPlayAtHostTime;
        bool supportsPauseAtHostTime;
        bool canPlayToWirelessPlaybackTarget;
        bool shouldIgnoreIntrinsicSize;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteMediaPlayerConfiguration) == sizeof(WebKit::RemoteMediaPlayerConfiguration));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, engineDescription)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, maximumDurationToCacheMediaTime)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, supportsScanning)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, supportsFullscreen)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, supportsPictureInPicture)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, supportsAcceleratedRendering)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, supportsPlayAtHostTime)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, supportsPauseAtHostTime)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, canPlayToWirelessPlaybackTarget)
        , offsetof(WebKit::RemoteMediaPlayerConfiguration, shouldIgnoreIntrinsicSize)
    >::value);
    encoder << instance.engineDescription;
    encoder << instance.maximumDurationToCacheMediaTime;
    encoder << instance.supportsScanning;
    encoder << instance.supportsFullscreen;
    encoder << instance.supportsPictureInPicture;
    encoder << instance.supportsAcceleratedRendering;
    encoder << instance.supportsPlayAtHostTime;
    encoder << instance.supportsPauseAtHostTime;
    encoder << instance.canPlayToWirelessPlaybackTarget;
    encoder << instance.shouldIgnoreIntrinsicSize;
}

std::optional<WebKit::RemoteMediaPlayerConfiguration> ArgumentCoder<WebKit::RemoteMediaPlayerConfiguration>::decode(Decoder& decoder)
{
    auto engineDescription = decoder.decode<String>();
    auto maximumDurationToCacheMediaTime = decoder.decode<double>();
    auto supportsScanning = decoder.decode<bool>();
    auto supportsFullscreen = decoder.decode<bool>();
    auto supportsPictureInPicture = decoder.decode<bool>();
    auto supportsAcceleratedRendering = decoder.decode<bool>();
    auto supportsPlayAtHostTime = decoder.decode<bool>();
    auto supportsPauseAtHostTime = decoder.decode<bool>();
    auto canPlayToWirelessPlaybackTarget = decoder.decode<bool>();
    auto shouldIgnoreIntrinsicSize = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::RemoteMediaPlayerConfiguration {
            WTFMove(*engineDescription),
            WTFMove(*maximumDurationToCacheMediaTime),
            WTFMove(*supportsScanning),
            WTFMove(*supportsFullscreen),
            WTFMove(*supportsPictureInPicture),
            WTFMove(*supportsAcceleratedRendering),
            WTFMove(*supportsPlayAtHostTime),
            WTFMove(*supportsPauseAtHostTime),
            WTFMove(*canPlayToWirelessPlaybackTarget),
            WTFMove(*shouldIgnoreIntrinsicSize)
        }
    };
}

#endif

#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
void ArgumentCoder<WebKit::RemoteMediaPlayerState>::encode(Encoder& encoder, const WebKit::RemoteMediaPlayerState& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.duration)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minTimeSeekable)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxTimeSeekable)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startDate)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.startTime)>, MediaTime>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.languageOfPrimaryAudioTrack)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wirelessPlaybackTargetName)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.bufferedRanges)>, std::optional<WebCore::PlatformTimeRanges>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.networkState)>, WebCore::MediaPlayerEnums::NetworkState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.readyState)>, WebCore::MediaPlayerEnums::ReadyState>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.movieLoadType)>, WebCore::MediaPlayerEnums::MovieLoadType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wirelessPlaybackTargetType)>, WebCore::MediaPlayerEnums::WirelessPlaybackTargetType>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.naturalSize)>, WebCore::FloatSize>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.maxFastForwardRate)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.minFastReverseRate)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.seekableTimeRangesLastModifiedTime)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.liveUpdateInterval)>, double>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.totalBytes)>, uint64_t>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.videoMetrics)>, std::optional<WebCore::VideoPlaybackQualityMetrics>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.documentIsCrossOrigin)>, std::optional<bool>>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.paused)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.canSaveMediaData)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasAudio)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasVideo)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasClosedCaptions)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.hasAvailableVideoFrame)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.wirelessVideoPlaybackDisabled)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.didPassCORSAccessCheck)>, bool>);
    struct ShouldBeSameSizeAsRemoteMediaPlayerState : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::RemoteMediaPlayerState>, false> {
        MediaTime duration;
        MediaTime minTimeSeekable;
        MediaTime maxTimeSeekable;
        MediaTime startDate;
        MediaTime startTime;
        String languageOfPrimaryAudioTrack;
        String wirelessPlaybackTargetName;
        std::optional<WebCore::PlatformTimeRanges> bufferedRanges;
        WebCore::MediaPlayerEnums::NetworkState networkState;
        WebCore::MediaPlayerEnums::ReadyState readyState;
        WebCore::MediaPlayerEnums::MovieLoadType movieLoadType;
        WebCore::MediaPlayerEnums::WirelessPlaybackTargetType wirelessPlaybackTargetType;
        WebCore::FloatSize naturalSize;
        double maxFastForwardRate;
        double minFastReverseRate;
        double seekableTimeRangesLastModifiedTime;
        double liveUpdateInterval;
        uint64_t totalBytes;
        std::optional<WebCore::VideoPlaybackQualityMetrics> videoMetrics;
        std::optional<bool> documentIsCrossOrigin;
        bool paused;
        bool canSaveMediaData;
        bool hasAudio;
        bool hasVideo;
        bool hasClosedCaptions;
        bool hasAvailableVideoFrame;
        bool wirelessVideoPlaybackDisabled;
        bool didPassCORSAccessCheck;
    };
    static_assert(sizeof(ShouldBeSameSizeAsRemoteMediaPlayerState) == sizeof(WebKit::RemoteMediaPlayerState));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::RemoteMediaPlayerState, duration)
        , offsetof(WebKit::RemoteMediaPlayerState, minTimeSeekable)
        , offsetof(WebKit::RemoteMediaPlayerState, maxTimeSeekable)
        , offsetof(WebKit::RemoteMediaPlayerState, startDate)
        , offsetof(WebKit::RemoteMediaPlayerState, startTime)
        , offsetof(WebKit::RemoteMediaPlayerState, languageOfPrimaryAudioTrack)
        , offsetof(WebKit::RemoteMediaPlayerState, wirelessPlaybackTargetName)
        , offsetof(WebKit::RemoteMediaPlayerState, bufferedRanges)
        , offsetof(WebKit::RemoteMediaPlayerState, networkState)
        , offsetof(WebKit::RemoteMediaPlayerState, readyState)
        , offsetof(WebKit::RemoteMediaPlayerState, movieLoadType)
        , offsetof(WebKit::RemoteMediaPlayerState, wirelessPlaybackTargetType)
        , offsetof(WebKit::RemoteMediaPlayerState, naturalSize)
        , offsetof(WebKit::RemoteMediaPlayerState, maxFastForwardRate)
        , offsetof(WebKit::RemoteMediaPlayerState, minFastReverseRate)
        , offsetof(WebKit::RemoteMediaPlayerState, seekableTimeRangesLastModifiedTime)
        , offsetof(WebKit::RemoteMediaPlayerState, liveUpdateInterval)
        , offsetof(WebKit::RemoteMediaPlayerState, totalBytes)
        , offsetof(WebKit::RemoteMediaPlayerState, videoMetrics)
        , offsetof(WebKit::RemoteMediaPlayerState, documentIsCrossOrigin)
        , offsetof(WebKit::RemoteMediaPlayerState, paused)
        , offsetof(WebKit::RemoteMediaPlayerState, canSaveMediaData)
        , offsetof(WebKit::RemoteMediaPlayerState, hasAudio)
        , offsetof(WebKit::RemoteMediaPlayerState, hasVideo)
        , offsetof(WebKit::RemoteMediaPlayerState, hasClosedCaptions)
        , offsetof(WebKit::RemoteMediaPlayerState, hasAvailableVideoFrame)
        , offsetof(WebKit::RemoteMediaPlayerState, wirelessVideoPlaybackDisabled)
        , offsetof(WebKit::RemoteMediaPlayerState, didPassCORSAccessCheck)
    >::value);
    encoder << instance.duration;
    encoder << instance.minTimeSeekable;
    encoder << instance.maxTimeSeekable;
    encoder << instance.startDate;
    encoder << instance.startTime;
    encoder << instance.languageOfPrimaryAudioTrack;
    encoder << instance.wirelessPlaybackTargetName;
    encoder << instance.bufferedRanges;
    encoder << instance.networkState;
    encoder << instance.readyState;
    encoder << instance.movieLoadType;
    encoder << instance.wirelessPlaybackTargetType;
    encoder << instance.naturalSize;
    encoder << instance.maxFastForwardRate;
    encoder << instance.minFastReverseRate;
    encoder << instance.seekableTimeRangesLastModifiedTime;
    encoder << instance.liveUpdateInterval;
    encoder << instance.totalBytes;
    encoder << instance.videoMetrics;
    encoder << instance.documentIsCrossOrigin;
    encoder << instance.paused;
    encoder << instance.canSaveMediaData;
    encoder << instance.hasAudio;
    encoder << instance.hasVideo;
    encoder << instance.hasClosedCaptions;
    encoder << instance.hasAvailableVideoFrame;
    encoder << instance.wirelessVideoPlaybackDisabled;
    encoder << instance.didPassCORSAccessCheck;
}

std::optional<WebKit::RemoteMediaPlayerState> ArgumentCoder<WebKit::RemoteMediaPlayerState>::decode(Decoder& decoder)
{
    auto duration = decoder.decode<MediaTime>();
    auto minTimeSeekable = decoder.decode<MediaTime>();
    auto maxTimeSeekable = decoder.decode<MediaTime>();
    auto startDate = decoder.decode<MediaTime>();
    auto startTime = decoder.decode<MediaTime>();
    auto languageOfPrimaryAudioTrack = decoder.decode<String>();
    auto wirelessPlaybackTargetName = decoder.decode<String>();
    auto bufferedRanges = decoder.decode<std::optional<WebCore::PlatformTimeRanges>>();
    auto networkState = decoder.decode<WebCore::MediaPlayerEnums::NetworkState>();
    auto readyState = decoder.decode<WebCore::MediaPlayerEnums::ReadyState>();
    auto movieLoadType = decoder.decode<WebCore::MediaPlayerEnums::MovieLoadType>();
    auto wirelessPlaybackTargetType = decoder.decode<WebCore::MediaPlayerEnums::WirelessPlaybackTargetType>();
    auto naturalSize = decoder.decode<WebCore::FloatSize>();
    auto maxFastForwardRate = decoder.decode<double>();
    auto minFastReverseRate = decoder.decode<double>();
    auto seekableTimeRangesLastModifiedTime = decoder.decode<double>();
    auto liveUpdateInterval = decoder.decode<double>();
    auto totalBytes = decoder.decode<uint64_t>();
    auto videoMetrics = decoder.decode<std::optional<WebCore::VideoPlaybackQualityMetrics>>();
    auto documentIsCrossOrigin = decoder.decode<std::optional<bool>>();
    auto paused = decoder.decode<bool>();
    auto canSaveMediaData = decoder.decode<bool>();
    auto hasAudio = decoder.decode<bool>();
    auto hasVideo = decoder.decode<bool>();
    auto hasClosedCaptions = decoder.decode<bool>();
    auto hasAvailableVideoFrame = decoder.decode<bool>();
    auto wirelessVideoPlaybackDisabled = decoder.decode<bool>();
    auto didPassCORSAccessCheck = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::RemoteMediaPlayerState {
            WTFMove(*duration),
            WTFMove(*minTimeSeekable),
            WTFMove(*maxTimeSeekable),
            WTFMove(*startDate),
            WTFMove(*startTime),
            WTFMove(*languageOfPrimaryAudioTrack),
            WTFMove(*wirelessPlaybackTargetName),
            WTFMove(*bufferedRanges),
            WTFMove(*networkState),
            WTFMove(*readyState),
            WTFMove(*movieLoadType),
            WTFMove(*wirelessPlaybackTargetType),
            WTFMove(*naturalSize),
            WTFMove(*maxFastForwardRate),
            WTFMove(*minFastReverseRate),
            WTFMove(*seekableTimeRangesLastModifiedTime),
            WTFMove(*liveUpdateInterval),
            WTFMove(*totalBytes),
            WTFMove(*videoMetrics),
            WTFMove(*documentIsCrossOrigin),
            WTFMove(*paused),
            WTFMove(*canSaveMediaData),
            WTFMove(*hasAudio),
            WTFMove(*hasVideo),
            WTFMove(*hasClosedCaptions),
            WTFMove(*hasAvailableVideoFrame),
            WTFMove(*wirelessVideoPlaybackDisabled),
            WTFMove(*didPassCORSAccessCheck)
        }
    };
}

#endif

#if ENABLE(SPEECH_SYNTHESIS)
void ArgumentCoder<WebKit::WebSpeechSynthesisVoice>::encode(Encoder& encoder, const WebKit::WebSpeechSynthesisVoice& instance)
{
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.voiceURI)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.name)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.lang)>, String>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.localService)>, bool>);
    static_assert(std::is_same_v<std::remove_cvref_t<decltype(instance.defaultLang)>, bool>);
    struct ShouldBeSameSizeAsWebSpeechSynthesisVoice : public VirtualTableAndRefCountOverhead<std::is_polymorphic_v<WebKit::WebSpeechSynthesisVoice>, false> {
        String voiceURI;
        String name;
        String lang;
        bool localService;
        bool defaultLang;
    };
    static_assert(sizeof(ShouldBeSameSizeAsWebSpeechSynthesisVoice) == sizeof(WebKit::WebSpeechSynthesisVoice));
    static_assert(MembersInCorrectOrder < 0
        , offsetof(WebKit::WebSpeechSynthesisVoice, voiceURI)
        , offsetof(WebKit::WebSpeechSynthesisVoice, name)
        , offsetof(WebKit::WebSpeechSynthesisVoice, lang)
        , offsetof(WebKit::WebSpeechSynthesisVoice, localService)
        , offsetof(WebKit::WebSpeechSynthesisVoice, defaultLang)
    >::value);
    encoder << instance.voiceURI;
    encoder << instance.name;
    encoder << instance.lang;
    encoder << instance.localService;
    encoder << instance.defaultLang;
}

std::optional<WebKit::WebSpeechSynthesisVoice> ArgumentCoder<WebKit::WebSpeechSynthesisVoice>::decode(Decoder& decoder)
{
    auto voiceURI = decoder.decode<String>();
    auto name = decoder.decode<String>();
    auto lang = decoder.decode<String>();
    auto localService = decoder.decode<bool>();
    auto defaultLang = decoder.decode<bool>();
    if (UNLIKELY(!decoder.isValid()))
        return std::nullopt;
    return {
        WebKit::WebSpeechSynthesisVoice {
            WTFMove(*voiceURI),
            WTFMove(*name),
            WTFMove(*lang),
            WTFMove(*localService),
            WTFMove(*defaultLang)
        }
    };
}

#endif

} // namespace IPC

namespace WTF {

template<> bool isValidEnum<IPC::WebCore_TimingFunction_Subclass, void>(IPC::EncodedVariantIndex value)
{
    switch (static_cast<IPC::WebCore_TimingFunction_Subclass>(value)) {
    case IPC::WebCore_TimingFunction_Subclass::LinearTimingFunction:
    case IPC::WebCore_TimingFunction_Subclass::CubicBezierTimingFunction:
    case IPC::WebCore_TimingFunction_Subclass::StepsTimingFunction:
    case IPC::WebCore_TimingFunction_Subclass::SpringTimingFunction:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<IPC::WebCore_TransformOperation_Subclass, void>(IPC::EncodedVariantIndex value)
{
    switch (static_cast<IPC::WebCore_TransformOperation_Subclass>(value)) {
    case IPC::WebCore_TransformOperation_Subclass::TranslateTransformOperation:
    case IPC::WebCore_TransformOperation_Subclass::ScaleTransformOperation:
    case IPC::WebCore_TransformOperation_Subclass::RotateTransformOperation:
    case IPC::WebCore_TransformOperation_Subclass::SkewTransformOperation:
    case IPC::WebCore_TransformOperation_Subclass::MatrixTransformOperation:
    case IPC::WebCore_TransformOperation_Subclass::Matrix3DTransformOperation:
    case IPC::WebCore_TransformOperation_Subclass::PerspectiveTransformOperation:
    case IPC::WebCore_TransformOperation_Subclass::IdentityTransformOperation:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<IPC::WebCore_BasicShape_Subclass, void>(IPC::EncodedVariantIndex value)
{
    switch (static_cast<IPC::WebCore_BasicShape_Subclass>(value)) {
    case IPC::WebCore_BasicShape_Subclass::BasicShapeCircle:
    case IPC::WebCore_BasicShape_Subclass::BasicShapeEllipse:
    case IPC::WebCore_BasicShape_Subclass::BasicShapePolygon:
    case IPC::WebCore_BasicShape_Subclass::BasicShapePath:
    case IPC::WebCore_BasicShape_Subclass::BasicShapeInset:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<IPC::WebCore_PathOperation_Subclass, void>(IPC::EncodedVariantIndex value)
{
    switch (static_cast<IPC::WebCore_PathOperation_Subclass>(value)) {
    case IPC::WebCore_PathOperation_Subclass::ReferencePathOperation:
    case IPC::WebCore_PathOperation_Subclass::ShapePathOperation:
    case IPC::WebCore_PathOperation_Subclass::BoxPathOperation:
    case IPC::WebCore_PathOperation_Subclass::RayPathOperation:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<IPC::WebCore_ReportBody_Subclass, void>(IPC::EncodedVariantIndex value)
{
    switch (static_cast<IPC::WebCore_ReportBody_Subclass>(value)) {
    case IPC::WebCore_ReportBody_Subclass::CSPViolationReportBody:
    case IPC::WebCore_ReportBody_Subclass::COEPInheritenceViolationReportBody:
    case IPC::WebCore_ReportBody_Subclass::CORPViolationReportBody:
    case IPC::WebCore_ReportBody_Subclass::DeprecationReportBody:
    case IPC::WebCore_ReportBody_Subclass::TestReportBody:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PathRoundedRect::Strategy, void>(uint8_t value)
{
    switch (static_cast<WebCore::PathRoundedRect::Strategy>(value)) {
    case WebCore::PathRoundedRect::Strategy::PreferNative:
    case WebCore::PathRoundedRect::Strategy::PreferBezier:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::ListType, void>(uint8_t value)
{
    switch (static_cast<WebKit::ListType>(value)) {
    case WebKit::ListType::None:
    case WebKit::ListType::OrderedList:
    case WebKit::ListType::UnorderedList:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::TextAlignment, void>(uint8_t value)
{
    switch (static_cast<WebKit::TextAlignment>(value)) {
    case WebKit::TextAlignment::Natural:
    case WebKit::TextAlignment::Left:
    case WebKit::TextAlignment::Right:
    case WebKit::TextAlignment::Center:
    case WebKit::TextAlignment::Justified:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebKit::TypingAttribute>(OptionSet<WebKit::TypingAttribute> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebKit::TypingAttribute::Bold)
        | static_cast<uint8_t>(WebKit::TypingAttribute::Italics)
        | static_cast<uint8_t>(WebKit::TypingAttribute::Underline)
        | static_cast<uint8_t>(WebKit::TypingAttribute::StrikeThrough)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebKit::RemoteWorkerType, void>(uint8_t value)
{
    switch (static_cast<WebKit::RemoteWorkerType>(value)) {
    case WebKit::RemoteWorkerType::ServiceWorker:
    case WebKit::RemoteWorkerType::SharedWorker:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::SameDocumentNavigationType, void>(uint8_t value)
{
    switch (static_cast<WebKit::SameDocumentNavigationType>(value)) {
    case WebKit::SameDocumentNavigationType::AnchorNavigation:
    case WebKit::SameDocumentNavigationType::SessionStatePush:
    case WebKit::SameDocumentNavigationType::SessionStateReplace:
    case WebKit::SameDocumentNavigationType::SessionStatePop:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontWidthVariant, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontWidthVariant>(value)) {
    case WebCore::FontWidthVariant::RegularWidth:
    case WebCore::FontWidthVariant::HalfWidth:
    case WebCore::FontWidthVariant::ThirdWidth:
    case WebCore::FontWidthVariant::QuarterWidth:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::Kerning, void>(uint8_t value)
{
    switch (static_cast<WebCore::Kerning>(value)) {
    case WebCore::Kerning::Auto:
    case WebCore::Kerning::Normal:
    case WebCore::Kerning::NoShift:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontOpticalSizing, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontOpticalSizing>(value)) {
    case WebCore::FontOpticalSizing::Enabled:
    case WebCore::FontOpticalSizing::Disabled:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontStyleAxis, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontStyleAxis>(value)) {
    case WebCore::FontStyleAxis::slnt:
    case WebCore::FontStyleAxis::ital:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantEastAsianVariant, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantEastAsianVariant>(value)) {
    case WebCore::FontVariantEastAsianVariant::Normal:
    case WebCore::FontVariantEastAsianVariant::Jis78:
    case WebCore::FontVariantEastAsianVariant::Jis83:
    case WebCore::FontVariantEastAsianVariant::Jis90:
    case WebCore::FontVariantEastAsianVariant::Jis04:
    case WebCore::FontVariantEastAsianVariant::Simplified:
    case WebCore::FontVariantEastAsianVariant::Traditional:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantEastAsianWidth, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantEastAsianWidth>(value)) {
    case WebCore::FontVariantEastAsianWidth::Normal:
    case WebCore::FontVariantEastAsianWidth::Full:
    case WebCore::FontVariantEastAsianWidth::Proportional:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantEastAsianRuby, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantEastAsianRuby>(value)) {
    case WebCore::FontVariantEastAsianRuby::Normal:
    case WebCore::FontVariantEastAsianRuby::Yes:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantLigatures, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantLigatures>(value)) {
    case WebCore::FontVariantLigatures::Normal:
    case WebCore::FontVariantLigatures::Yes:
    case WebCore::FontVariantLigatures::No:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantPosition, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantPosition>(value)) {
    case WebCore::FontVariantPosition::Normal:
    case WebCore::FontVariantPosition::Subscript:
    case WebCore::FontVariantPosition::Superscript:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantCaps, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantCaps>(value)) {
    case WebCore::FontVariantCaps::Normal:
    case WebCore::FontVariantCaps::Small:
    case WebCore::FontVariantCaps::AllSmall:
    case WebCore::FontVariantCaps::Petite:
    case WebCore::FontVariantCaps::AllPetite:
    case WebCore::FontVariantCaps::Unicase:
    case WebCore::FontVariantCaps::Titling:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantNumericFigure, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantNumericFigure>(value)) {
    case WebCore::FontVariantNumericFigure::Normal:
    case WebCore::FontVariantNumericFigure::LiningNumbers:
    case WebCore::FontVariantNumericFigure::OldStyleNumbers:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantNumericSpacing, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantNumericSpacing>(value)) {
    case WebCore::FontVariantNumericSpacing::Normal:
    case WebCore::FontVariantNumericSpacing::ProportionalNumbers:
    case WebCore::FontVariantNumericSpacing::TabularNumbers:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontVariantNumericFraction, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontVariantNumericFraction>(value)) {
    case WebCore::FontVariantNumericFraction::Normal:
    case WebCore::FontVariantNumericFraction::DiagonalFractions:
    case WebCore::FontVariantNumericFraction::StackedFractions:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::TextRenderingMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::TextRenderingMode>(value)) {
    case WebCore::TextRenderingMode::AutoTextRendering:
    case WebCore::TextRenderingMode::OptimizeSpeed:
    case WebCore::TextRenderingMode::OptimizeLegibility:
    case WebCore::TextRenderingMode::GeometricPrecision:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontSmoothingMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontSmoothingMode>(value)) {
    case WebCore::FontSmoothingMode::AutoSmoothing:
    case WebCore::FontSmoothingMode::NoSmoothing:
    case WebCore::FontSmoothingMode::Antialiased:
    case WebCore::FontSmoothingMode::SubpixelAntialiased:
        return true;
    default:
        return false;
    }
}

#if ENABLE(APPLE_PAY)
template<> bool isValidEnum<WebCore::ApplePayFeature, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayFeature>(value)) {
#if ENABLE(APPLE_PAY_LATER)
    case WebCore::ApplePayFeature::ApplePayLater:
#endif
#if ENABLE(APPLE_PAY_LATER_AVAILABILITY)
    case WebCore::ApplePayFeature::ApplePayLaterAvailability:
#endif
#if ENABLE(APPLE_PAY_PAYMENT_ORDER_DETAILS)
    case WebCore::ApplePayFeature::AuthorizationResultOrderDetails:
#endif
    case WebCore::ApplePayFeature::LineItemPaymentTiming:
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_PAYMENTS)
    case WebCore::ApplePayFeature::PaymentRequestAutomaticReload:
#endif
#if ENABLE(APPLE_PAY_COUPON_CODE)
    case WebCore::ApplePayFeature::PaymentRequestCouponCode:
#endif
#if ENABLE(APPLE_PAY_MULTI_MERCHANT_PAYMENTS)
    case WebCore::ApplePayFeature::PaymentRequestMultiTokenContexts:
#endif
#if ENABLE(APPLE_PAY_RECURRING_PAYMENTS)
    case WebCore::ApplePayFeature::PaymentRequestRecurring:
#endif
#if ENABLE(APPLE_PAY_SHIPPING_CONTACT_EDITING_MODE)
    case WebCore::ApplePayFeature::PaymentRequestShippingContactEditingMode:
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
    case WebCore::ApplePayFeature::PaymentTimingAutomaticReload:
#endif
#if ENABLE(APPLE_PAY_DEFERRED_PAYMENTS)
    case WebCore::ApplePayFeature::PaymentRequestDeferred:
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
    case WebCore::ApplePayFeature::PaymentTimingDeferred:
#endif
    case WebCore::ApplePayFeature::PaymentTimingImmediate:
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    case WebCore::ApplePayFeature::PaymentTimingRecurring:
#endif
#if ENABLE(APPLE_PAY_SHIPPING_CONTACT_EDITING_MODE)
    case WebCore::ApplePayFeature::ShippingContactEditingModeEnabled:
#endif
#if ENABLE(APPLE_PAY_SHIPPING_CONTACT_EDITING_MODE)
    case WebCore::ApplePayFeature::ShippingContactEditingModeStorePickup:
#endif
#if ENABLE(APPLE_PAY_SHIPPING_METHOD_DATE_COMPONENTS_RANGE)
    case WebCore::ApplePayFeature::ShippingMethodDateComponentsRange:
#endif
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(APPLE_PAY)
template<> bool isValidEnum<WebCore::ApplePayPaymentTiming, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayPaymentTiming>(value)) {
    case WebCore::ApplePayPaymentTiming::Immediate:
#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
    case WebCore::ApplePayPaymentTiming::Recurring:
#endif
#if ENABLE(APPLE_PAY_DEFERRED_LINE_ITEM)
    case WebCore::ApplePayPaymentTiming::Deferred:
#endif
#if ENABLE(APPLE_PAY_AUTOMATIC_RELOAD_LINE_ITEM)
    case WebCore::ApplePayPaymentTiming::AutomaticReload:
#endif
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(APPLE_PAY_RECURRING_LINE_ITEM)
template<> bool isValidEnum<WebCore::ApplePayRecurringPaymentDateUnit, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayRecurringPaymentDateUnit>(value)) {
    case WebCore::ApplePayRecurringPaymentDateUnit::Year:
    case WebCore::ApplePayRecurringPaymentDateUnit::Month:
    case WebCore::ApplePayRecurringPaymentDateUnit::Day:
    case WebCore::ApplePayRecurringPaymentDateUnit::Hour:
    case WebCore::ApplePayRecurringPaymentDateUnit::Minute:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::ContactProperty, void>(uint8_t value)
{
    switch (static_cast<WebCore::ContactProperty>(value)) {
    case WebCore::ContactProperty::Email:
    case WebCore::ContactProperty::Name:
    case WebCore::ContactProperty::Tel:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::CookieConsentDecisionResult, void>(uint8_t value)
{
    switch (static_cast<WebCore::CookieConsentDecisionResult>(value)) {
    case WebCore::CookieConsentDecisionResult::NotSupported:
    case WebCore::CookieConsentDecisionResult::Consent:
    case WebCore::CookieConsentDecisionResult::Dissent:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::IDBResultType, void>(uint8_t value)
{
    switch (static_cast<WebCore::IDBResultType>(value)) {
    case WebCore::IDBResultType::Error:
    case WebCore::IDBResultType::OpenDatabaseSuccess:
    case WebCore::IDBResultType::OpenDatabaseUpgradeNeeded:
    case WebCore::IDBResultType::DeleteDatabaseSuccess:
    case WebCore::IDBResultType::CreateObjectStoreSuccess:
    case WebCore::IDBResultType::DeleteObjectStoreSuccess:
    case WebCore::IDBResultType::ClearObjectStoreSuccess:
    case WebCore::IDBResultType::PutOrAddSuccess:
    case WebCore::IDBResultType::GetRecordSuccess:
    case WebCore::IDBResultType::GetAllRecordsSuccess:
    case WebCore::IDBResultType::GetCountSuccess:
    case WebCore::IDBResultType::DeleteRecordSuccess:
    case WebCore::IDBResultType::CreateIndexSuccess:
    case WebCore::IDBResultType::DeleteIndexSuccess:
    case WebCore::IDBResultType::OpenCursorSuccess:
    case WebCore::IDBResultType::IterateCursorSuccess:
    case WebCore::IDBResultType::RenameObjectStoreSuccess:
    case WebCore::IDBResultType::RenameIndexSuccess:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::IDBTransactionDurability, void>(uint8_t value)
{
    switch (static_cast<WebCore::IDBTransactionDurability>(value)) {
    case WebCore::IDBTransactionDurability::Strict:
    case WebCore::IDBTransactionDurability::Relaxed:
    case WebCore::IDBTransactionDurability::Default:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::IDBTransactionMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::IDBTransactionMode>(value)) {
    case WebCore::IDBTransactionMode::Readonly:
    case WebCore::IDBTransactionMode::Readwrite:
    case WebCore::IDBTransactionMode::Versionchange:
        return true;
    default:
        return false;
    }
}

#if ENABLE(MEDIA_SESSION)
template<> bool isValidEnum<WebCore::MediaSessionAction, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaSessionAction>(value)) {
    case WebCore::MediaSessionAction::Play:
    case WebCore::MediaSessionAction::Pause:
    case WebCore::MediaSessionAction::Seekbackward:
    case WebCore::MediaSessionAction::Seekforward:
    case WebCore::MediaSessionAction::Previoustrack:
    case WebCore::MediaSessionAction::Nexttrack:
    case WebCore::MediaSessionAction::Skipad:
    case WebCore::MediaSessionAction::Stop:
    case WebCore::MediaSessionAction::Seekto:
    case WebCore::MediaSessionAction::Settrack:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_SESSION)
template<> bool isValidEnum<WebCore::MediaSessionPlaybackState, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaSessionPlaybackState>(value)) {
    case WebCore::MediaSessionPlaybackState::None:
    case WebCore::MediaSessionPlaybackState::Paused:
    case WebCore::MediaSessionPlaybackState::Playing:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_SESSION_COORDINATOR)
template<> bool isValidEnum<WebCore::MediaSessionCoordinatorState, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaSessionCoordinatorState>(value)) {
    case WebCore::MediaSessionCoordinatorState::Waiting:
    case WebCore::MediaSessionCoordinatorState::Joined:
    case WebCore::MediaSessionCoordinatorState::Closed:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_SESSION_COORDINATOR)
template<> bool isValidEnum<WebCore::MediaSessionReadyState, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaSessionReadyState>(value)) {
    case WebCore::MediaSessionReadyState::Havenothing:
    case WebCore::MediaSessionReadyState::Havemetadata:
    case WebCore::MediaSessionReadyState::Havecurrentdata:
    case WebCore::MediaSessionReadyState::Havefuturedata:
    case WebCore::MediaSessionReadyState::Haveenoughdata:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::RTCErrorDetailType, void>(uint8_t value)
{
    switch (static_cast<WebCore::RTCErrorDetailType>(value)) {
    case WebCore::RTCErrorDetailType::DataChannelFailure:
    case WebCore::RTCErrorDetailType::DtlsFailure:
    case WebCore::RTCErrorDetailType::FingerprintFailure:
    case WebCore::RTCErrorDetailType::SctpFailure:
    case WebCore::RTCErrorDetailType::SdpSyntaxError:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PermissionName, void>(uint8_t value)
{
    switch (static_cast<WebCore::PermissionName>(value)) {
    case WebCore::PermissionName::Accelerometer:
    case WebCore::PermissionName::BackgroundFetch:
    case WebCore::PermissionName::Bluetooth:
    case WebCore::PermissionName::Camera:
    case WebCore::PermissionName::DisplayCapture:
    case WebCore::PermissionName::Geolocation:
    case WebCore::PermissionName::Gyroscope:
    case WebCore::PermissionName::Magnetometer:
    case WebCore::PermissionName::Microphone:
    case WebCore::PermissionName::Midi:
    case WebCore::PermissionName::Nfc:
    case WebCore::PermissionName::Notifications:
    case WebCore::PermissionName::Push:
    case WebCore::PermissionName::ScreenWakeLock:
    case WebCore::PermissionName::SpeakerSelection:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PermissionQuerySource, void>(uint8_t value)
{
    switch (static_cast<WebCore::PermissionQuerySource>(value)) {
    case WebCore::PermissionQuerySource::Window:
    case WebCore::PermissionQuerySource::DedicatedWorker:
    case WebCore::PermissionQuerySource::SharedWorker:
    case WebCore::PermissionQuerySource::ServiceWorker:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PermissionState, void>(uint8_t value)
{
    switch (static_cast<WebCore::PermissionState>(value)) {
    case WebCore::PermissionState::Granted:
    case WebCore::PermissionState::Denied:
    case WebCore::PermissionState::Prompt:
        return true;
    default:
        return false;
    }
}

#if ENABLE(SERVICE_WORKER)
template<> bool isValidEnum<WebCore::PushPermissionState, void>(uint8_t value)
{
    switch (static_cast<WebCore::PushPermissionState>(value)) {
    case WebCore::PushPermissionState::Denied:
    case WebCore::PushPermissionState::Granted:
    case WebCore::PushPermissionState::Prompt:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::AutofillFieldName, void>(uint8_t value)
{
    switch (static_cast<WebCore::AutofillFieldName>(value)) {
    case WebCore::AutofillFieldName::None:
    case WebCore::AutofillFieldName::Name:
    case WebCore::AutofillFieldName::HonorificPrefix:
    case WebCore::AutofillFieldName::GivenName:
    case WebCore::AutofillFieldName::AdditionalName:
    case WebCore::AutofillFieldName::FamilyName:
    case WebCore::AutofillFieldName::HonorificSuffix:
    case WebCore::AutofillFieldName::Nickname:
    case WebCore::AutofillFieldName::Username:
    case WebCore::AutofillFieldName::NewPassword:
    case WebCore::AutofillFieldName::CurrentPassword:
    case WebCore::AutofillFieldName::OrganizationTitle:
    case WebCore::AutofillFieldName::Organization:
    case WebCore::AutofillFieldName::StreetAddress:
    case WebCore::AutofillFieldName::AddressLine1:
    case WebCore::AutofillFieldName::AddressLine2:
    case WebCore::AutofillFieldName::AddressLine3:
    case WebCore::AutofillFieldName::AddressLevel4:
    case WebCore::AutofillFieldName::AddressLevel3:
    case WebCore::AutofillFieldName::AddressLevel2:
    case WebCore::AutofillFieldName::AddressLevel1:
    case WebCore::AutofillFieldName::Country:
    case WebCore::AutofillFieldName::CountryName:
    case WebCore::AutofillFieldName::PostalCode:
    case WebCore::AutofillFieldName::CcName:
    case WebCore::AutofillFieldName::CcGivenName:
    case WebCore::AutofillFieldName::CcAdditionalName:
    case WebCore::AutofillFieldName::CcFamilyName:
    case WebCore::AutofillFieldName::CcNumber:
    case WebCore::AutofillFieldName::CcExp:
    case WebCore::AutofillFieldName::CcExpMonth:
    case WebCore::AutofillFieldName::CcExpYear:
    case WebCore::AutofillFieldName::CcCsc:
    case WebCore::AutofillFieldName::CcType:
    case WebCore::AutofillFieldName::TransactionCurrency:
    case WebCore::AutofillFieldName::TransactionAmount:
    case WebCore::AutofillFieldName::Language:
    case WebCore::AutofillFieldName::Bday:
    case WebCore::AutofillFieldName::BdayDay:
    case WebCore::AutofillFieldName::BdayMonth:
    case WebCore::AutofillFieldName::BdayYear:
    case WebCore::AutofillFieldName::Sex:
    case WebCore::AutofillFieldName::URL:
    case WebCore::AutofillFieldName::Photo:
    case WebCore::AutofillFieldName::Tel:
    case WebCore::AutofillFieldName::TelCountryCode:
    case WebCore::AutofillFieldName::TelNational:
    case WebCore::AutofillFieldName::TelAreaCode:
    case WebCore::AutofillFieldName::TelLocal:
    case WebCore::AutofillFieldName::TelLocalPrefix:
    case WebCore::AutofillFieldName::TelLocalSuffix:
    case WebCore::AutofillFieldName::TelExtension:
    case WebCore::AutofillFieldName::Email:
    case WebCore::AutofillFieldName::Impp:
    case WebCore::AutofillFieldName::WebAuthn:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::EnterKeyHint, void>(uint8_t value)
{
    switch (static_cast<WebCore::EnterKeyHint>(value)) {
    case WebCore::EnterKeyHint::Unspecified:
    case WebCore::EnterKeyHint::Enter:
    case WebCore::EnterKeyHint::Done:
    case WebCore::EnterKeyHint::Go:
    case WebCore::EnterKeyHint::Next:
    case WebCore::EnterKeyHint::Previous:
    case WebCore::EnterKeyHint::Search:
    case WebCore::EnterKeyHint::Send:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::DownloadableBinaryFontAllowedTypes, void>(uint8_t value)
{
    switch (static_cast<WebCore::DownloadableBinaryFontAllowedTypes>(value)) {
    case WebCore::DownloadableBinaryFontAllowedTypes::Any:
    case WebCore::DownloadableBinaryFontAllowedTypes::Restricted:
    case WebCore::DownloadableBinaryFontAllowedTypes::None:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::LoadSchedulingMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::LoadSchedulingMode>(value)) {
    case WebCore::LoadSchedulingMode::Direct:
    case WebCore::LoadSchedulingMode::Prioritized:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ServiceWorkersMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::ServiceWorkersMode>(value)) {
    case WebCore::ServiceWorkersMode::All:
    case WebCore::ServiceWorkersMode::None:
    case WebCore::ServiceWorkersMode::Only:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ShouldTreatAsContinuingLoad, void>(uint8_t value)
{
    switch (static_cast<WebCore::ShouldTreatAsContinuingLoad>(value)) {
    case WebCore::ShouldTreatAsContinuingLoad::No:
    case WebCore::ShouldTreatAsContinuingLoad::YesAfterNavigationPolicyDecision:
    case WebCore::ShouldTreatAsContinuingLoad::YesAfterProvisionalLoadStarted:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ContentSecurityPolicyModeForExtension, void>(uint8_t value)
{
    switch (static_cast<WebCore::ContentSecurityPolicyModeForExtension>(value)) {
    case WebCore::ContentSecurityPolicyModeForExtension::None:
    case WebCore::ContentSecurityPolicyModeForExtension::ManifestV2:
    case WebCore::ContentSecurityPolicyModeForExtension::ManifestV3:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaProducerMediaCaptureKind, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaProducerMediaCaptureKind>(value)) {
    case WebCore::MediaProducerMediaCaptureKind::Microphone:
    case WebCore::MediaProducerMediaCaptureKind::Camera:
    case WebCore::MediaProducerMediaCaptureKind::Display:
    case WebCore::MediaProducerMediaCaptureKind::SystemAudio:
    case WebCore::MediaProducerMediaCaptureKind::EveryKind:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ScreenOrientationType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScreenOrientationType>(value)) {
    case WebCore::ScreenOrientationType::PortraitPrimary:
    case WebCore::ScreenOrientationType::PortraitSecondary:
    case WebCore::ScreenOrientationType::LandscapePrimary:
    case WebCore::ScreenOrientationType::LandscapeSecondary:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::StorageBlockingPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::StorageBlockingPolicy>(value)) {
    case WebCore::StorageBlockingPolicy::AllowAll:
    case WebCore::StorageBlockingPolicy::BlockThirdParty:
    case WebCore::StorageBlockingPolicy::BlockAll:
        return true;
    default:
        return false;
    }
}

#if USE(APPKIT)
template<> bool isValidEnum<WebCore::AppKitControlSystemImageType, void>(uint8_t value)
{
    switch (static_cast<WebCore::AppKitControlSystemImageType>(value)) {
    case WebCore::AppKitControlSystemImageType::ScrollbarTrackCorner:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::ListStyleType::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::ListStyleType::Type>(value)) {
    case WebCore::ListStyleType::Type::CounterStyle:
    case WebCore::ListStyleType::Type::String:
    case WebCore::ListStyleType::Type::None:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ScrollGranularity, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScrollGranularity>(value)) {
    case WebCore::ScrollGranularity::Line:
    case WebCore::ScrollGranularity::Page:
    case WebCore::ScrollGranularity::Document:
    case WebCore::ScrollGranularity::Pixel:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::IndexedDB::KeyType, void>(int8_t value)
{
    switch (static_cast<WebCore::IndexedDB::KeyType>(value)) {
    case WebCore::IndexedDB::KeyType::Max:
    case WebCore::IndexedDB::KeyType::Invalid:
    case WebCore::IndexedDB::KeyType::Array:
    case WebCore::IndexedDB::KeyType::Binary:
    case WebCore::IndexedDB::KeyType::String:
    case WebCore::IndexedDB::KeyType::Date:
    case WebCore::IndexedDB::KeyType::Number:
    case WebCore::IndexedDB::KeyType::Min:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::IndexedDB::CursorDirection, void>(uint8_t value)
{
    switch (static_cast<WebCore::IndexedDB::CursorDirection>(value)) {
    case WebCore::IndexedDB::CursorDirection::Next:
    case WebCore::IndexedDB::CursorDirection::Nextunique:
    case WebCore::IndexedDB::CursorDirection::Prev:
    case WebCore::IndexedDB::CursorDirection::Prevunique:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::IndexedDB::ObjectStoreOverwriteMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::IndexedDB::ObjectStoreOverwriteMode>(value)) {
    case WebCore::IndexedDB::ObjectStoreOverwriteMode::Overwrite:
    case WebCore::IndexedDB::ObjectStoreOverwriteMode::OverwriteForCursor:
    case WebCore::IndexedDB::ObjectStoreOverwriteMode::NoOverwrite:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::IndexedDB::RequestType, void>(uint8_t value)
{
    switch (static_cast<WebCore::IndexedDB::RequestType>(value)) {
    case WebCore::IndexedDB::RequestType::Open:
    case WebCore::IndexedDB::RequestType::Delete:
    case WebCore::IndexedDB::RequestType::Other:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::CubicBezierTimingFunction::TimingFunctionPreset, void>(uint8_t value)
{
    switch (static_cast<WebCore::CubicBezierTimingFunction::TimingFunctionPreset>(value)) {
    case WebCore::CubicBezierTimingFunction::TimingFunctionPreset::Ease:
    case WebCore::CubicBezierTimingFunction::TimingFunctionPreset::EaseIn:
    case WebCore::CubicBezierTimingFunction::TimingFunctionPreset::EaseOut:
    case WebCore::CubicBezierTimingFunction::TimingFunctionPreset::EaseInOut:
    case WebCore::CubicBezierTimingFunction::TimingFunctionPreset::Custom:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::StepsTimingFunction::StepPosition, void>(uint8_t value)
{
    switch (static_cast<WebCore::StepsTimingFunction::StepPosition>(value)) {
    case WebCore::StepsTimingFunction::StepPosition::JumpStart:
    case WebCore::StepsTimingFunction::StepPosition::JumpEnd:
    case WebCore::StepsTimingFunction::StepPosition::JumpNone:
    case WebCore::StepsTimingFunction::StepPosition::JumpBoth:
    case WebCore::StepsTimingFunction::StepPosition::Start:
    case WebCore::StepsTimingFunction::StepPosition::End:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::NavigatorAPIsAccessed>(OptionSet<WebCore::NavigatorAPIsAccessed> value)
{
    constexpr uint64_t allValidBitsValue = 0
        | static_cast<uint64_t>(WebCore::NavigatorAPIsAccessed::AppVersion)
        | static_cast<uint64_t>(WebCore::NavigatorAPIsAccessed::UserAgent)
        | static_cast<uint64_t>(WebCore::NavigatorAPIsAccessed::Plugins)
        | static_cast<uint64_t>(WebCore::NavigatorAPIsAccessed::MimeTypes)
        | static_cast<uint64_t>(WebCore::NavigatorAPIsAccessed::CookieEnabled)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidOptionSet<WebCore::ScreenAPIsAccessed>(OptionSet<WebCore::ScreenAPIsAccessed> value)
{
    constexpr uint64_t allValidBitsValue = 0
        | static_cast<uint64_t>(WebCore::ScreenAPIsAccessed::Height)
        | static_cast<uint64_t>(WebCore::ScreenAPIsAccessed::Width)
        | static_cast<uint64_t>(WebCore::ScreenAPIsAccessed::ColorDepth)
        | static_cast<uint64_t>(WebCore::ScreenAPIsAccessed::AvailLeft)
        | static_cast<uint64_t>(WebCore::ScreenAPIsAccessed::AvailTop)
        | static_cast<uint64_t>(WebCore::ScreenAPIsAccessed::AvailHeight)
        | static_cast<uint64_t>(WebCore::ScreenAPIsAccessed::AvailWidth)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::ViewportFit, void>(uint8_t value)
{
    switch (static_cast<WebCore::ViewportFit>(value)) {
    case WebCore::ViewportFit::Auto:
    case WebCore::ViewportFit::Contain:
    case WebCore::ViewportFit::Cover:
        return true;
    default:
        return false;
    }
}

#if ENABLE(APPLE_PAY)
template<> bool isValidEnum<WebCore::ApplePayErrorContactField, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayErrorContactField>(value)) {
    case WebCore::ApplePayErrorContactField::PhoneNumber:
    case WebCore::ApplePayErrorContactField::EmailAddress:
    case WebCore::ApplePayErrorContactField::Name:
    case WebCore::ApplePayErrorContactField::PhoneticName:
    case WebCore::ApplePayErrorContactField::PostalAddress:
    case WebCore::ApplePayErrorContactField::AddressLines:
    case WebCore::ApplePayErrorContactField::SubLocality:
    case WebCore::ApplePayErrorContactField::Locality:
    case WebCore::ApplePayErrorContactField::PostalCode:
    case WebCore::ApplePayErrorContactField::SubAdministrativeArea:
    case WebCore::ApplePayErrorContactField::AdministrativeArea:
    case WebCore::ApplePayErrorContactField::Country:
    case WebCore::ApplePayErrorContactField::CountryCode:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(APPLE_PAY)
template<> bool isValidEnum<WebCore::ApplePayErrorCode, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayErrorCode>(value)) {
    case WebCore::ApplePayErrorCode::Unknown:
    case WebCore::ApplePayErrorCode::ShippingContactInvalid:
    case WebCore::ApplePayErrorCode::BillingContactInvalid:
    case WebCore::ApplePayErrorCode::AddressUnserviceable:
#if ENABLE(APPLE_PAY_COUPON_CODE)
    case WebCore::ApplePayErrorCode::CouponCodeInvalid:
#endif
#if ENABLE(APPLE_PAY_COUPON_CODE)
    case WebCore::ApplePayErrorCode::CouponCodeExpired:
#endif
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(APPLE_PAY_INSTALLMENTS)
template<> bool isValidEnum<WebCore::ApplePayInstallmentItemType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayInstallmentItemType>(value)) {
    case WebCore::ApplePayInstallmentItemType::Generic:
    case WebCore::ApplePayInstallmentItemType::Phone:
    case WebCore::ApplePayInstallmentItemType::Pad:
    case WebCore::ApplePayInstallmentItemType::Watch:
    case WebCore::ApplePayInstallmentItemType::Mac:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(APPLE_PAY_INSTALLMENTS)
template<> bool isValidEnum<WebCore::ApplePayInstallmentRetailChannel, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayInstallmentRetailChannel>(value)) {
    case WebCore::ApplePayInstallmentRetailChannel::Unknown:
    case WebCore::ApplePayInstallmentRetailChannel::App:
    case WebCore::ApplePayInstallmentRetailChannel::Web:
    case WebCore::ApplePayInstallmentRetailChannel::InStore:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::ScreenOrientationLockType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScreenOrientationLockType>(value)) {
    case WebCore::ScreenOrientationLockType::Any:
    case WebCore::ScreenOrientationLockType::Natural:
    case WebCore::ScreenOrientationLockType::Landscape:
    case WebCore::ScreenOrientationLockType::Portrait:
    case WebCore::ScreenOrientationLockType::PortraitPrimary:
    case WebCore::ScreenOrientationLockType::PortraitSecondary:
    case WebCore::ScreenOrientationLockType::LandscapePrimary:
    case WebCore::ScreenOrientationLockType::LandscapeSecondary:
        return true;
    default:
        return false;
    }
}

#if ENABLE(APPLICATION_MANIFEST)
template<> bool isValidEnum<WebCore::ApplicationManifest::Display, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplicationManifest::Display>(value)) {
    case WebCore::ApplicationManifest::Display::Browser:
    case WebCore::ApplicationManifest::Display::MinimalUI:
    case WebCore::ApplicationManifest::Display::Standalone:
    case WebCore::ApplicationManifest::Display::Fullscreen:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(APPLICATION_MANIFEST)
template<> bool isValidOptionSet<WebCore::ApplicationManifest::Icon::Purpose>(OptionSet<WebCore::ApplicationManifest::Icon::Purpose> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::ApplicationManifest::Icon::Purpose::Any)
        | static_cast<uint8_t>(WebCore::ApplicationManifest::Icon::Purpose::Monochrome)
        | static_cast<uint8_t>(WebCore::ApplicationManifest::Icon::Purpose::Maskable)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}
#endif

template<> bool isValidEnum<WebCore::DOMCacheEngine::Error, void>(uint8_t value)
{
    switch (static_cast<WebCore::DOMCacheEngine::Error>(value)) {
    case WebCore::DOMCacheEngine::Error::NotImplemented:
    case WebCore::DOMCacheEngine::Error::ReadDisk:
    case WebCore::DOMCacheEngine::Error::WriteDisk:
    case WebCore::DOMCacheEngine::Error::QuotaExceeded:
    case WebCore::DOMCacheEngine::Error::Internal:
    case WebCore::DOMCacheEngine::Error::Stopped:
    case WebCore::DOMCacheEngine::Error::CORP:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::SpeechRecognitionErrorType, void>(uint8_t value)
{
    switch (static_cast<WebCore::SpeechRecognitionErrorType>(value)) {
    case WebCore::SpeechRecognitionErrorType::NoSpeech:
    case WebCore::SpeechRecognitionErrorType::Aborted:
    case WebCore::SpeechRecognitionErrorType::AudioCapture:
    case WebCore::SpeechRecognitionErrorType::Network:
    case WebCore::SpeechRecognitionErrorType::NotAllowed:
    case WebCore::SpeechRecognitionErrorType::ServiceNotAllowed:
    case WebCore::SpeechRecognitionErrorType::BadGrammar:
    case WebCore::SpeechRecognitionErrorType::LanguageNotSupported:
        return true;
    default:
        return false;
    }
}

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::AttestationConveyancePreference, void>(uint8_t value)
{
    switch (static_cast<WebCore::AttestationConveyancePreference>(value)) {
    case WebCore::AttestationConveyancePreference::None:
    case WebCore::AttestationConveyancePreference::Indirect:
    case WebCore::AttestationConveyancePreference::Direct:
    case WebCore::AttestationConveyancePreference::Enterprise:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::Cookie::SameSitePolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::Cookie::SameSitePolicy>(value)) {
    case WebCore::Cookie::SameSitePolicy::None:
    case WebCore::Cookie::SameSitePolicy::Lax:
    case WebCore::Cookie::SameSitePolicy::Strict:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::RenderingPurpose, void>(uint8_t value)
{
    switch (static_cast<WebCore::RenderingPurpose>(value)) {
    case WebCore::RenderingPurpose::Unspecified:
    case WebCore::RenderingPurpose::Canvas:
    case WebCore::RenderingPurpose::DOM:
    case WebCore::RenderingPurpose::LayerBacking:
    case WebCore::RenderingPurpose::Snapshot:
    case WebCore::RenderingPurpose::ShareableSnapshot:
    case WebCore::RenderingPurpose::ShareableLocalSnapshot:
    case WebCore::RenderingPurpose::MediaPainting:
        return true;
    default:
        return false;
    }
}

#if ENABLE(CONTENT_FILTERING)
template<> bool isValidEnum<WebCore::MockContentFilterSettings::DecisionPoint, void>(uint8_t value)
{
    switch (static_cast<WebCore::MockContentFilterSettings::DecisionPoint>(value)) {
    case WebCore::MockContentFilterSettings::DecisionPoint::AfterWillSendRequest:
    case WebCore::MockContentFilterSettings::DecisionPoint::AfterRedirect:
    case WebCore::MockContentFilterSettings::DecisionPoint::AfterResponse:
    case WebCore::MockContentFilterSettings::DecisionPoint::AfterAddData:
    case WebCore::MockContentFilterSettings::DecisionPoint::AfterFinishedAddingData:
    case WebCore::MockContentFilterSettings::DecisionPoint::Never:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::AutoplayEvent, void>(uint8_t value)
{
    switch (static_cast<WebCore::AutoplayEvent>(value)) {
    case WebCore::AutoplayEvent::DidPreventMediaFromPlaying:
    case WebCore::AutoplayEvent::DidPlayMediaWithUserGesture:
    case WebCore::AutoplayEvent::DidAutoplayMediaPastThresholdWithoutUserInterference:
    case WebCore::AutoplayEvent::UserDidInterfereWithPlayback:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::InputMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::InputMode>(value)) {
    case WebCore::InputMode::Unspecified:
    case WebCore::InputMode::None:
    case WebCore::InputMode::Text:
    case WebCore::InputMode::Telephone:
    case WebCore::InputMode::Url:
    case WebCore::InputMode::Email:
    case WebCore::InputMode::Numeric:
    case WebCore::InputMode::Decimal:
    case WebCore::InputMode::Search:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::NotificationDirection, void>(uint8_t value)
{
    switch (static_cast<WebCore::NotificationDirection>(value)) {
    case WebCore::NotificationDirection::Auto:
    case WebCore::NotificationDirection::Ltr:
    case WebCore::NotificationDirection::Rtl:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::StoredCredentialsPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::StoredCredentialsPolicy>(value)) {
    case WebCore::StoredCredentialsPolicy::DoNotUse:
    case WebCore::StoredCredentialsPolicy::Use:
    case WebCore::StoredCredentialsPolicy::EphemeralStateless:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PreflightPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::PreflightPolicy>(value)) {
    case WebCore::PreflightPolicy::Consider:
    case WebCore::PreflightPolicy::Force:
    case WebCore::PreflightPolicy::Prevent:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WTFLogChannelState, void>(uint8_t value)
{
    switch (static_cast<WTFLogChannelState>(value)) {
    case WTFLogChannelState::Off:
    case WTFLogChannelState::On:
    case WTFLogChannelState::OnWithAccumulation:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ScrollSnapStrictness, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScrollSnapStrictness>(value)) {
    case WebCore::ScrollSnapStrictness::None:
    case WebCore::ScrollSnapStrictness::Proximity:
    case WebCore::ScrollSnapStrictness::Mandatory:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::LengthType, void>(uint8_t value)
{
    switch (static_cast<WebCore::LengthType>(value)) {
    case WebCore::LengthType::Auto:
    case WebCore::LengthType::Relative:
    case WebCore::LengthType::Percent:
    case WebCore::LengthType::Fixed:
    case WebCore::LengthType::Intrinsic:
    case WebCore::LengthType::MinIntrinsic:
    case WebCore::LengthType::MinContent:
    case WebCore::LengthType::MaxContent:
    case WebCore::LengthType::FillAvailable:
    case WebCore::LengthType::FitContent:
    case WebCore::LengthType::Calculated:
    case WebCore::LengthType::Undefined:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::OverscrollBehavior, void>(uint8_t value)
{
    switch (static_cast<WebCore::OverscrollBehavior>(value)) {
    case WebCore::OverscrollBehavior::Auto:
    case WebCore::OverscrollBehavior::Contain:
    case WebCore::OverscrollBehavior::None:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ScrollPinningBehavior, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScrollPinningBehavior>(value)) {
    case WebCore::ScrollPinningBehavior::DoNotPin:
    case WebCore::ScrollPinningBehavior::PinToTop:
    case WebCore::ScrollPinningBehavior::PinToBottom:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::WheelScrollGestureState, void>(uint8_t value)
{
    switch (static_cast<WebCore::WheelScrollGestureState>(value)) {
    case WebCore::WheelScrollGestureState::Blocking:
    case WebCore::WheelScrollGestureState::NonBlocking:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::DiagnosticLoggingDomain, void>(uint8_t value)
{
    switch (static_cast<WebCore::DiagnosticLoggingDomain>(value)) {
    case WebCore::DiagnosticLoggingDomain::Media:
        return true;
    default:
        return false;
    }
}

#if !USE(CG)
template<> bool isValidEnum<WebCore::PlatformColorSpace::Name, void>(uint8_t value)
{
    switch (static_cast<WebCore::PlatformColorSpace::Name>(value)) {
    case WebCore::PlatformColorSpace::Name::SRGB:
#if ENABLE(DESTINATION_COLOR_SPACE_LINEAR_SRGB)
    case WebCore::PlatformColorSpace::Name::LinearSRGB:
#endif
#if ENABLE(DESTINATION_COLOR_SPACE_DISPLAY_P3)
    case WebCore::PlatformColorSpace::Name::DisplayP3:
#endif
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::MediaCaptureType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaCaptureType>(value)) {
    case WebCore::MediaCaptureType::MediaCaptureTypeNone:
    case WebCore::MediaCaptureType::MediaCaptureTypeUser:
    case WebCore::MediaCaptureType::MediaCaptureTypeEnvironment:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaStreamRequest::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaStreamRequest::Type>(value)) {
    case WebCore::MediaStreamRequest::Type::UserMedia:
    case WebCore::MediaStreamRequest::Type::DisplayMedia:
    case WebCore::MediaStreamRequest::Type::DisplayMediaWithAudio:
        return true;
    default:
        return false;
    }
}

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::AuthenticatorTransport, void>(uint8_t value)
{
    switch (static_cast<WebCore::AuthenticatorTransport>(value)) {
    case WebCore::AuthenticatorTransport::Usb:
    case WebCore::AuthenticatorTransport::Nfc:
    case WebCore::AuthenticatorTransport::Ble:
    case WebCore::AuthenticatorTransport::Internal:
    case WebCore::AuthenticatorTransport::Cable:
    case WebCore::AuthenticatorTransport::Hybrid:
    case WebCore::AuthenticatorTransport::SmartCard:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(DATALIST_ELEMENT)
template<> bool isValidEnum<WebCore::DataListSuggestionActivationType, void>(uint8_t value)
{
    switch (static_cast<WebCore::DataListSuggestionActivationType>(value)) {
    case WebCore::DataListSuggestionActivationType::ControlClicked:
    case WebCore::DataListSuggestionActivationType::IndicatorClicked:
    case WebCore::DataListSuggestionActivationType::TextChanged:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::AlphaPremultiplication, void>(uint8_t value)
{
    switch (static_cast<WebCore::AlphaPremultiplication>(value)) {
    case WebCore::AlphaPremultiplication::Premultiplied:
    case WebCore::AlphaPremultiplication::Unpremultiplied:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PixelFormat, void>(uint8_t value)
{
    switch (static_cast<WebCore::PixelFormat>(value)) {
    case WebCore::PixelFormat::RGBA8:
    case WebCore::PixelFormat::BGRX8:
    case WebCore::PixelFormat::BGRA8:
    case WebCore::PixelFormat::RGB10:
    case WebCore::PixelFormat::RGB10A8:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::EventTrackingRegions::EventType, void>(uint8_t value)
{
    switch (static_cast<WebCore::EventTrackingRegions::EventType>(value)) {
    case WebCore::EventTrackingRegions::EventType::Mousedown:
    case WebCore::EventTrackingRegions::EventType::Mousemove:
    case WebCore::EventTrackingRegions::EventType::Mouseup:
    case WebCore::EventTrackingRegions::EventType::Mousewheel:
    case WebCore::EventTrackingRegions::EventType::Pointerdown:
    case WebCore::EventTrackingRegions::EventType::Pointerenter:
    case WebCore::EventTrackingRegions::EventType::Pointerleave:
    case WebCore::EventTrackingRegions::EventType::Pointermove:
    case WebCore::EventTrackingRegions::EventType::Pointerout:
    case WebCore::EventTrackingRegions::EventType::Pointerover:
    case WebCore::EventTrackingRegions::EventType::Pointerup:
    case WebCore::EventTrackingRegions::EventType::Touchend:
    case WebCore::EventTrackingRegions::EventType::Touchforcechange:
    case WebCore::EventTrackingRegions::EventType::Touchmove:
    case WebCore::EventTrackingRegions::EventType::Touchstart:
    case WebCore::EventTrackingRegions::EventType::Wheel:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ScrollElasticity, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScrollElasticity>(value)) {
    case WebCore::ScrollElasticity::Automatic:
    case WebCore::ScrollElasticity::None:
    case WebCore::ScrollElasticity::Allowed:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ScrollbarMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScrollbarMode>(value)) {
    case WebCore::ScrollbarMode::Auto:
    case WebCore::ScrollbarMode::AlwaysOff:
    case WebCore::ScrollbarMode::AlwaysOn:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::NativeScrollbarVisibility, void>(uint8_t value)
{
    switch (static_cast<WebCore::NativeScrollbarVisibility>(value)) {
    case WebCore::NativeScrollbarVisibility::Visible:
    case WebCore::NativeScrollbarVisibility::HiddenByStyle:
    case WebCore::NativeScrollbarVisibility::ReplacedByCustomScrollbar:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ScrollbarWidth, void>(uint8_t value)
{
    switch (static_cast<WebCore::ScrollbarWidth>(value)) {
    case WebCore::ScrollbarWidth::Auto:
    case WebCore::ScrollbarWidth::Thin:
    case WebCore::ScrollbarWidth::None:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::WheelEventProcessingSteps>(OptionSet<WebCore::WheelEventProcessingSteps> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::WheelEventProcessingSteps::AsyncScrolling)
        | static_cast<uint8_t>(WebCore::WheelEventProcessingSteps::SynchronousScrolling)
        | static_cast<uint8_t>(WebCore::WheelEventProcessingSteps::NonBlockingDOMEventDispatch)
        | static_cast<uint8_t>(WebCore::WheelEventProcessingSteps::BlockingDOMEventDispatch)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
template<> bool isValidEnum<WebCore::GraphicsContextGLPowerPreference, void>(uint8_t value)
{
    switch (static_cast<WebCore::GraphicsContextGLPowerPreference>(value)) {
    case WebCore::GraphicsContextGLPowerPreference::Default:
    case WebCore::GraphicsContextGLPowerPreference::LowPower:
    case WebCore::GraphicsContextGLPowerPreference::HighPerformance:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
template<> bool isValidEnum<WebCore::GraphicsContextGLWebGLVersion, void>(uint8_t value)
{
    switch (static_cast<WebCore::GraphicsContextGLWebGLVersion>(value)) {
    case WebCore::GraphicsContextGLWebGLVersion::WebGL1:
    case WebCore::GraphicsContextGLWebGLVersion::WebGL2:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::BasicShapeRadius::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::BasicShapeRadius::Type>(value)) {
    case WebCore::BasicShapeRadius::Type::Value:
    case WebCore::BasicShapeRadius::Type::ClosestSide:
    case WebCore::BasicShapeRadius::Type::FarthestSide:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::CSSBoxType, void>(uint8_t value)
{
    switch (static_cast<WebCore::CSSBoxType>(value)) {
    case WebCore::CSSBoxType::BoxMissing:
    case WebCore::CSSBoxType::MarginBox:
    case WebCore::CSSBoxType::BorderBox:
    case WebCore::CSSBoxType::PaddingBox:
    case WebCore::CSSBoxType::ContentBox:
    case WebCore::CSSBoxType::FillBox:
    case WebCore::CSSBoxType::StrokeBox:
    case WebCore::CSSBoxType::ViewBox:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::RayPathOperation::Size, void>(uint8_t value)
{
    switch (static_cast<WebCore::RayPathOperation::Size>(value)) {
    case WebCore::RayPathOperation::Size::ClosestSide:
    case WebCore::RayPathOperation::Size::ClosestCorner:
    case WebCore::RayPathOperation::Size::FarthestSide:
    case WebCore::RayPathOperation::Size::FarthestCorner:
    case WebCore::RayPathOperation::Size::Sides:
        return true;
    default:
        return false;
    }
}

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::UserVerificationRequirement, void>(uint8_t value)
{
    switch (static_cast<WebCore::UserVerificationRequirement>(value)) {
    case WebCore::UserVerificationRequirement::Required:
    case WebCore::UserVerificationRequirement::Preferred:
    case WebCore::UserVerificationRequirement::Discouraged:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::ResidentKeyRequirement, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResidentKeyRequirement>(value)) {
    case WebCore::ResidentKeyRequirement::Required:
    case WebCore::ResidentKeyRequirement::Preferred:
    case WebCore::ResidentKeyRequirement::Discouraged:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::AuthenticatorAttachment, void>(uint8_t value)
{
    switch (static_cast<WebCore::AuthenticatorAttachment>(value)) {
    case WebCore::AuthenticatorAttachment::Platform:
    case WebCore::AuthenticatorAttachment::CrossPlatform:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::ResourceRequestCachePolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResourceRequestCachePolicy>(value)) {
    case WebCore::ResourceRequestCachePolicy::UseProtocolCachePolicy:
    case WebCore::ResourceRequestCachePolicy::ReloadIgnoringCacheData:
    case WebCore::ResourceRequestCachePolicy::ReturnCacheDataElseLoad:
    case WebCore::ResourceRequestCachePolicy::ReturnCacheDataDontLoad:
    case WebCore::ResourceRequestCachePolicy::DoNotUseAnyCache:
    case WebCore::ResourceRequestCachePolicy::RefreshAnyCacheData:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ResourceRequestBase::SameSiteDisposition, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResourceRequestBase::SameSiteDisposition>(value)) {
    case WebCore::ResourceRequestBase::SameSiteDisposition::Unspecified:
    case WebCore::ResourceRequestBase::SameSiteDisposition::SameSite:
    case WebCore::ResourceRequestBase::SameSiteDisposition::CrossSite:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ResourceRequestRequester, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResourceRequestRequester>(value)) {
    case WebCore::ResourceRequestRequester::Unspecified:
    case WebCore::ResourceRequestRequester::Main:
    case WebCore::ResourceRequestRequester::XHR:
    case WebCore::ResourceRequestRequester::Fetch:
    case WebCore::ResourceRequestRequester::Media:
    case WebCore::ResourceRequestRequester::Model:
    case WebCore::ResourceRequestRequester::ImportScripts:
    case WebCore::ResourceRequestRequester::Ping:
    case WebCore::ResourceRequestRequester::Beacon:
    case WebCore::ResourceRequestRequester::EventSource:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ResourceLoadPriority, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResourceLoadPriority>(value)) {
    case WebCore::ResourceLoadPriority::VeryLow:
    case WebCore::ResourceLoadPriority::Low:
    case WebCore::ResourceLoadPriority::Medium:
    case WebCore::ResourceLoadPriority::High:
    case WebCore::ResourceLoadPriority::VeryHigh:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ResourceResponseBase::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResourceResponseBase::Type>(value)) {
    case WebCore::ResourceResponseBase::Type::Basic:
    case WebCore::ResourceResponseBase::Type::Cors:
    case WebCore::ResourceResponseBase::Type::Default:
    case WebCore::ResourceResponseBase::Type::Error:
    case WebCore::ResourceResponseBase::Type::Opaque:
    case WebCore::ResourceResponseBase::Type::Opaqueredirect:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ResourceResponseBase::Tainting, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResourceResponseBase::Tainting>(value)) {
    case WebCore::ResourceResponseBase::Tainting::Basic:
    case WebCore::ResourceResponseBase::Tainting::Cors:
    case WebCore::ResourceResponseBase::Tainting::Opaque:
    case WebCore::ResourceResponseBase::Tainting::Opaqueredirect:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ResourceResponseBase::Source, void>(uint8_t value)
{
    switch (static_cast<WebCore::ResourceResponseBase::Source>(value)) {
    case WebCore::ResourceResponseBase::Source::Unknown:
    case WebCore::ResourceResponseBase::Source::Network:
    case WebCore::ResourceResponseBase::Source::DiskCache:
    case WebCore::ResourceResponseBase::Source::DiskCacheAfterValidation:
    case WebCore::ResourceResponseBase::Source::MemoryCache:
    case WebCore::ResourceResponseBase::Source::MemoryCacheAfterValidation:
    case WebCore::ResourceResponseBase::Source::ServiceWorker:
    case WebCore::ResourceResponseBase::Source::ApplicationCache:
    case WebCore::ResourceResponseBase::Source::DOMCache:
    case WebCore::ResourceResponseBase::Source::InspectorOverride:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ReferrerPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::ReferrerPolicy>(value)) {
    case WebCore::ReferrerPolicy::EmptyString:
    case WebCore::ReferrerPolicy::NoReferrer:
    case WebCore::ReferrerPolicy::NoReferrerWhenDowngrade:
    case WebCore::ReferrerPolicy::SameOrigin:
    case WebCore::ReferrerPolicy::Origin:
    case WebCore::ReferrerPolicy::StrictOrigin:
    case WebCore::ReferrerPolicy::OriginWhenCrossOrigin:
    case WebCore::ReferrerPolicy::StrictOriginWhenCrossOrigin:
    case WebCore::ReferrerPolicy::UnsafeUrl:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FetchOptionsDestination, void>(uint8_t value)
{
    switch (static_cast<WebCore::FetchOptionsDestination>(value)) {
    case WebCore::FetchOptionsDestination::EmptyString:
    case WebCore::FetchOptionsDestination::Audio:
    case WebCore::FetchOptionsDestination::Audioworklet:
    case WebCore::FetchOptionsDestination::Document:
    case WebCore::FetchOptionsDestination::Embed:
    case WebCore::FetchOptionsDestination::Font:
    case WebCore::FetchOptionsDestination::Image:
    case WebCore::FetchOptionsDestination::Iframe:
    case WebCore::FetchOptionsDestination::Manifest:
    case WebCore::FetchOptionsDestination::Model:
    case WebCore::FetchOptionsDestination::Object:
    case WebCore::FetchOptionsDestination::Paintworklet:
    case WebCore::FetchOptionsDestination::Report:
    case WebCore::FetchOptionsDestination::Script:
    case WebCore::FetchOptionsDestination::Serviceworker:
    case WebCore::FetchOptionsDestination::Sharedworker:
    case WebCore::FetchOptionsDestination::Style:
    case WebCore::FetchOptionsDestination::Track:
    case WebCore::FetchOptionsDestination::Video:
    case WebCore::FetchOptionsDestination::Worker:
    case WebCore::FetchOptionsDestination::Xslt:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FetchOptionsMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::FetchOptionsMode>(value)) {
    case WebCore::FetchOptionsMode::Navigate:
    case WebCore::FetchOptionsMode::SameOrigin:
    case WebCore::FetchOptionsMode::NoCors:
    case WebCore::FetchOptionsMode::Cors:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FetchOptionsCredentials, void>(uint8_t value)
{
    switch (static_cast<WebCore::FetchOptionsCredentials>(value)) {
    case WebCore::FetchOptionsCredentials::Omit:
    case WebCore::FetchOptionsCredentials::SameOrigin:
    case WebCore::FetchOptionsCredentials::Include:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FetchOptionsCache, void>(uint8_t value)
{
    switch (static_cast<WebCore::FetchOptionsCache>(value)) {
    case WebCore::FetchOptionsCache::Default:
    case WebCore::FetchOptionsCache::NoStore:
    case WebCore::FetchOptionsCache::Reload:
    case WebCore::FetchOptionsCache::NoCache:
    case WebCore::FetchOptionsCache::ForceCache:
    case WebCore::FetchOptionsCache::OnlyIfCached:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FetchOptionsRedirect, void>(uint8_t value)
{
    switch (static_cast<WebCore::FetchOptionsRedirect>(value)) {
    case WebCore::FetchOptionsRedirect::Follow:
    case WebCore::FetchOptionsRedirect::Error:
    case WebCore::FetchOptionsRedirect::Manual:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FetchHeadersGuard, void>(uint8_t value)
{
    switch (static_cast<WebCore::FetchHeadersGuard>(value)) {
    case WebCore::FetchHeadersGuard::None:
    case WebCore::FetchHeadersGuard::Immutable:
    case WebCore::FetchHeadersGuard::Request:
    case WebCore::FetchHeadersGuard::RequestNoCors:
    case WebCore::FetchHeadersGuard::Response:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::ControlStyle::State>(OptionSet<WebCore::ControlStyle::State> value)
{
    constexpr unsigned allValidBitsValue = 0
        | static_cast<unsigned>(WebCore::ControlStyle::State::Hovered)
        | static_cast<unsigned>(WebCore::ControlStyle::State::Pressed)
        | static_cast<unsigned>(WebCore::ControlStyle::State::Focused)
        | static_cast<unsigned>(WebCore::ControlStyle::State::Enabled)
        | static_cast<unsigned>(WebCore::ControlStyle::State::Checked)
        | static_cast<unsigned>(WebCore::ControlStyle::State::Default)
        | static_cast<unsigned>(WebCore::ControlStyle::State::WindowActive)
        | static_cast<unsigned>(WebCore::ControlStyle::State::Indeterminate)
        | static_cast<unsigned>(WebCore::ControlStyle::State::SpinUp)
        | static_cast<unsigned>(WebCore::ControlStyle::State::Presenting)
        | static_cast<unsigned>(WebCore::ControlStyle::State::FormSemanticContext)
        | static_cast<unsigned>(WebCore::ControlStyle::State::DarkAppearance)
        | static_cast<unsigned>(WebCore::ControlStyle::State::RightToLeft)
        | static_cast<unsigned>(WebCore::ControlStyle::State::LargeControls)
        | static_cast<unsigned>(WebCore::ControlStyle::State::ReadOnly)
        | static_cast<unsigned>(WebCore::ControlStyle::State::ListButton)
        | static_cast<unsigned>(WebCore::ControlStyle::State::ListButtonPressed)
        | static_cast<unsigned>(WebCore::ControlStyle::State::VerticalWritingMode)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::StyleAppearance, void>(uint8_t value)
{
    switch (static_cast<WebCore::StyleAppearance>(value)) {
    case WebCore::StyleAppearance::None:
    case WebCore::StyleAppearance::Auto:
    case WebCore::StyleAppearance::Checkbox:
    case WebCore::StyleAppearance::Radio:
    case WebCore::StyleAppearance::PushButton:
    case WebCore::StyleAppearance::SquareButton:
    case WebCore::StyleAppearance::Button:
    case WebCore::StyleAppearance::DefaultButton:
    case WebCore::StyleAppearance::Listbox:
    case WebCore::StyleAppearance::Menulist:
    case WebCore::StyleAppearance::MenulistButton:
    case WebCore::StyleAppearance::Meter:
    case WebCore::StyleAppearance::ProgressBar:
    case WebCore::StyleAppearance::SliderHorizontal:
    case WebCore::StyleAppearance::SliderVertical:
    case WebCore::StyleAppearance::SearchField:
#if ENABLE(APPLE_PAY)
    case WebCore::StyleAppearance::ApplePayButton:
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case WebCore::StyleAppearance::Attachment:
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case WebCore::StyleAppearance::BorderlessAttachment:
#endif
    case WebCore::StyleAppearance::TextArea:
    case WebCore::StyleAppearance::TextField:
    case WebCore::StyleAppearance::CapsLockIndicator:
#if ENABLE(INPUT_TYPE_COLOR)
    case WebCore::StyleAppearance::ColorWell:
#endif
#if ENABLE(SERVICE_CONTROLS)
    case WebCore::StyleAppearance::ImageControlsButton:
#endif
    case WebCore::StyleAppearance::InnerSpinButton:
#if ENABLE(DATALIST_ELEMENT)
    case WebCore::StyleAppearance::ListButton:
#endif
    case WebCore::StyleAppearance::SearchFieldDecoration:
    case WebCore::StyleAppearance::SearchFieldResultsDecoration:
    case WebCore::StyleAppearance::SearchFieldResultsButton:
    case WebCore::StyleAppearance::SearchFieldCancelButton:
    case WebCore::StyleAppearance::SliderThumbHorizontal:
    case WebCore::StyleAppearance::SliderThumbVertical:
        return true;
    default:
        return false;
    }
}

#if ENABLE(APPLE_PAY)
template<> bool isValidEnum<WebCore::ApplePayButtonType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayButtonType>(value)) {
    case WebCore::ApplePayButtonType::Plain:
    case WebCore::ApplePayButtonType::Buy:
    case WebCore::ApplePayButtonType::SetUp:
    case WebCore::ApplePayButtonType::Donate:
    case WebCore::ApplePayButtonType::CheckOut:
    case WebCore::ApplePayButtonType::Book:
    case WebCore::ApplePayButtonType::Subscribe:
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::Reload:
#endif
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::AddMoney:
#endif
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::TopUp:
#endif
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::Order:
#endif
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::Rent:
#endif
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::Support:
#endif
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::Contribute:
#endif
#if ENABLE(APPLE_PAY_NEW_BUTTON_TYPES)
    case WebCore::ApplePayButtonType::Tip:
#endif
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(APPLE_PAY)
template<> bool isValidEnum<WebCore::ApplePayButtonStyle, void>(uint8_t value)
{
    switch (static_cast<WebCore::ApplePayButtonStyle>(value)) {
    case WebCore::ApplePayButtonStyle::White:
    case WebCore::ApplePayButtonStyle::WhiteOutline:
    case WebCore::ApplePayButtonStyle::Black:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::MeterPart::GaugeRegion, void>(uint8_t value)
{
    switch (static_cast<WebCore::MeterPart::GaugeRegion>(value)) {
    case WebCore::MeterPart::GaugeRegion::Optimum:
    case WebCore::MeterPart::GaugeRegion::Suboptimal:
    case WebCore::MeterPart::GaugeRegion::EvenLessGood:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::FilterEffectGeometry::Flags>(OptionSet<WebCore::FilterEffectGeometry::Flags> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::FilterEffectGeometry::Flags::HasX)
        | static_cast<uint8_t>(WebCore::FilterEffectGeometry::Flags::HasY)
        | static_cast<uint8_t>(WebCore::FilterEffectGeometry::Flags::HasWidth)
        | static_cast<uint8_t>(WebCore::FilterEffectGeometry::Flags::HasHeight)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::ModalContainerDecision, void>(uint8_t value)
{
    switch (static_cast<WebCore::ModalContainerDecision>(value)) {
    case WebCore::ModalContainerDecision::Show:
    case WebCore::ModalContainerDecision::HideAndIgnore:
    case WebCore::ModalContainerDecision::HideAndAllow:
    case WebCore::ModalContainerDecision::HideAndDisallow:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::AutoplayEventFlags>(OptionSet<WebCore::AutoplayEventFlags> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::AutoplayEventFlags::HasAudio)
        | static_cast<uint8_t>(WebCore::AutoplayEventFlags::PlaybackWasPrevented)
        | static_cast<uint8_t>(WebCore::AutoplayEventFlags::MediaIsMainContent)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::FocusDirection, void>(uint8_t value)
{
    switch (static_cast<WebCore::FocusDirection>(value)) {
    case WebCore::FocusDirection::None:
    case WebCore::FocusDirection::Forward:
    case WebCore::FocusDirection::Backward:
    case WebCore::FocusDirection::Up:
    case WebCore::FocusDirection::Down:
    case WebCore::FocusDirection::Left:
    case WebCore::FocusDirection::Right:
        return true;
    default:
        return false;
    }
}

#if ENABLE(GAMEPAD)
template<> bool isValidEnum<WebCore::GamepadHapticEffectType, void>(uint8_t value)
{
    switch (static_cast<WebCore::GamepadHapticEffectType>(value)) {
    case WebCore::GamepadHapticEffectType::DualRumble:
    case WebCore::GamepadHapticEffectType::TriggerRumble:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::DOMPasteAccessCategory, void>(uint8_t value)
{
    switch (static_cast<WebCore::DOMPasteAccessCategory>(value)) {
    case WebCore::DOMPasteAccessCategory::General:
    case WebCore::DOMPasteAccessCategory::Fonts:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::DOMPasteAccessPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::DOMPasteAccessPolicy>(value)) {
    case WebCore::DOMPasteAccessPolicy::NotRequestedYet:
    case WebCore::DOMPasteAccessPolicy::Denied:
    case WebCore::DOMPasteAccessPolicy::Granted:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::DOMPasteAccessResponse, void>(uint8_t value)
{
    switch (static_cast<WebCore::DOMPasteAccessResponse>(value)) {
    case WebCore::DOMPasteAccessResponse::DeniedForGesture:
    case WebCore::DOMPasteAccessResponse::GrantedForCommand:
    case WebCore::DOMPasteAccessResponse::GrantedForGesture:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::DeviceOrientationOrMotionPermissionState, void>(uint8_t value)
{
    switch (static_cast<WebCore::DeviceOrientationOrMotionPermissionState>(value)) {
    case WebCore::DeviceOrientationOrMotionPermissionState::Granted:
    case WebCore::DeviceOrientationOrMotionPermissionState::Denied:
    case WebCore::DeviceOrientationOrMotionPermissionState::Prompt:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ExceptionDetails::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::ExceptionDetails::Type>(value)) {
    case WebCore::ExceptionDetails::Type::Script:
    case WebCore::ExceptionDetails::Type::InvalidTargetFrame:
    case WebCore::ExceptionDetails::Type::AppBoundDomain:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::VerticalAlignChange, void>(uint8_t value)
{
    switch (static_cast<WebCore::VerticalAlignChange>(value)) {
    case WebCore::VerticalAlignChange::Superscript:
    case WebCore::VerticalAlignChange::Baseline:
    case WebCore::VerticalAlignChange::Subscript:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::TextGranularity, void>(uint8_t value)
{
    switch (static_cast<WebCore::TextGranularity>(value)) {
    case WebCore::TextGranularity::CharacterGranularity:
    case WebCore::TextGranularity::WordGranularity:
    case WebCore::TextGranularity::SentenceGranularity:
    case WebCore::TextGranularity::LineGranularity:
    case WebCore::TextGranularity::ParagraphGranularity:
    case WebCore::TextGranularity::DocumentGranularity:
    case WebCore::TextGranularity::SentenceBoundary:
    case WebCore::TextGranularity::LineBoundary:
    case WebCore::TextGranularity::ParagraphBoundary:
    case WebCore::TextGranularity::DocumentBoundary:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontAttributes::SubscriptOrSuperscript, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontAttributes::SubscriptOrSuperscript>(value)) {
    case WebCore::FontAttributes::SubscriptOrSuperscript::None:
    case WebCore::FontAttributes::SubscriptOrSuperscript::Subscript:
    case WebCore::FontAttributes::SubscriptOrSuperscript::Superscript:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FontAttributes::HorizontalAlignment, void>(uint8_t value)
{
    switch (static_cast<WebCore::FontAttributes::HorizontalAlignment>(value)) {
    case WebCore::FontAttributes::HorizontalAlignment::Left:
    case WebCore::FontAttributes::HorizontalAlignment::Center:
    case WebCore::FontAttributes::HorizontalAlignment::Right:
    case WebCore::FontAttributes::HorizontalAlignment::Justify:
    case WebCore::FontAttributes::HorizontalAlignment::Natural:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::WritingDirection, void>(uint8_t value)
{
    switch (static_cast<WebCore::WritingDirection>(value)) {
    case WebCore::WritingDirection::Natural:
    case WebCore::WritingDirection::LeftToRight:
    case WebCore::WritingDirection::RightToLeft:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::SelectionDirection, void>(uint8_t value)
{
    switch (static_cast<WebCore::SelectionDirection>(value)) {
    case WebCore::SelectionDirection::Forward:
    case WebCore::SelectionDirection::Backward:
    case WebCore::SelectionDirection::Right:
    case WebCore::SelectionDirection::Left:
        return true;
    default:
        return false;
    }
}

#if ENABLE(DATA_DETECTION)
template<> bool isValidOptionSet<WebCore::DataDetectorType>(OptionSet<WebCore::DataDetectorType> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::DataDetectorType::PhoneNumber)
        | static_cast<uint8_t>(WebCore::DataDetectorType::Link)
        | static_cast<uint8_t>(WebCore::DataDetectorType::Address)
        | static_cast<uint8_t>(WebCore::DataDetectorType::CalendarEvent)
        | static_cast<uint8_t>(WebCore::DataDetectorType::TrackingNumber)
        | static_cast<uint8_t>(WebCore::DataDetectorType::FlightNumber)
        | static_cast<uint8_t>(WebCore::DataDetectorType::LookupSuggestion)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}
#endif

template<> bool isValidEnum<WebCore::AutocapitalizeType, void>(uint8_t value)
{
    switch (static_cast<WebCore::AutocapitalizeType>(value)) {
    case WebCore::AutocapitalizeType::Default:
    case WebCore::AutocapitalizeType::None:
    case WebCore::AutocapitalizeType::Words:
    case WebCore::AutocapitalizeType::Sentences:
    case WebCore::AutocapitalizeType::AllCharacters:
        return true;
    default:
        return false;
    }
}

#if ENABLE(ASYNC_SCROLLING)
template<> bool isValidOptionSet<WebCore::ScrollingStateNodeProperty>(OptionSet<WebCore::ScrollingStateNodeProperty> value)
{
    constexpr uint64_t allValidBitsValue = 0
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::Layer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ChildNodes)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollableAreaSize)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::TotalContentsSize)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ReachableContentsSize)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollPosition)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollOrigin)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollableAreaParams)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ReasonsForSynchronousScrolling)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::RequestedScrollPosition)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::SnapOffsetsInfo)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::CurrentHorizontalSnapOffsetIndex)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::CurrentVerticalSnapOffsetIndex)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::IsMonitoringWheelEvents)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollContainerLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrolledContentsLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::HorizontalScrollbarLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::VerticalScrollbarLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::PainterForScrollbar)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ContentAreaHoverState)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::MouseActivityState)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollbarHoverState)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollbarEnabledState)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::FrameScaleFactor)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::EventTrackingRegion)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::RootContentsLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::CounterScrollingLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::InsetClipLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ContentShadowLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::HeaderHeight)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::FooterHeight)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::HeaderLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::FooterLayer)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::BehaviorForFixedElements)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::TopContentInset)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::FixedElementsLayoutRelativeToFrame)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::VisualViewportIsSmallerThanLayoutViewport)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::AsyncFrameOrOverflowScrollingEnabled)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::WheelEventGesturesBecomeNonBlocking)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ScrollingPerformanceTestingEnabled)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::LayoutViewport)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::MinLayoutViewportOrigin)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::MaxLayoutViewportOrigin)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::OverrideVisualViewportSize)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::RelatedOverflowScrollingNodes)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::LayoutConstraintData)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::ViewportConstraints)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::OverflowScrollingNode)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::KeyboardScrollData)
        | static_cast<uint64_t>(WebCore::ScrollingStateNodeProperty::OverlayScrollbarsEnabled)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}
#endif

#if ENABLE(ENCRYPTED_MEDIA)
template<> bool isValidEnum<WebCore::CDMSessionType, void>(uint8_t value)
{
    switch (static_cast<WebCore::CDMSessionType>(value)) {
    case WebCore::CDMSessionType::Temporary:
    case WebCore::CDMSessionType::PersistentUsageRecord:
    case WebCore::CDMSessionType::PersistentLicense:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::MediaSelectionOption::MediaType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaSelectionOption::MediaType>(value)) {
    case WebCore::MediaSelectionOption::MediaType::Unknown:
    case WebCore::MediaSelectionOption::MediaType::Audio:
    case WebCore::MediaSelectionOption::MediaType::Subtitles:
    case WebCore::MediaSelectionOption::MediaType::Captions:
    case WebCore::MediaSelectionOption::MediaType::Metadata:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaSelectionOption::LegibleType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaSelectionOption::LegibleType>(value)) {
    case WebCore::MediaSelectionOption::LegibleType::Regular:
    case WebCore::MediaSelectionOption::LegibleType::LegibleOff:
    case WebCore::MediaSelectionOption::LegibleType::LegibleAuto:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaDecodingType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaDecodingType>(value)) {
    case WebCore::MediaDecodingType::File:
    case WebCore::MediaDecodingType::MediaSource:
    case WebCore::MediaDecodingType::WebRTC:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::TextManipulationControllerExclusionRule::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::TextManipulationControllerExclusionRule::Type>(value)) {
    case WebCore::TextManipulationControllerExclusionRule::Type::Exclude:
    case WebCore::TextManipulationControllerExclusionRule::Type::Include:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::TextManipulationControllerManipulationFailure::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::TextManipulationControllerManipulationFailure::Type>(value)) {
    case WebCore::TextManipulationControllerManipulationFailure::Type::NotAvailable:
    case WebCore::TextManipulationControllerManipulationFailure::Type::ContentChanged:
    case WebCore::TextManipulationControllerManipulationFailure::Type::InvalidItem:
    case WebCore::TextManipulationControllerManipulationFailure::Type::InvalidToken:
    case WebCore::TextManipulationControllerManipulationFailure::Type::ExclusionViolation:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerNetworkState, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerNetworkState>(value)) {
    case WebCore::MediaPlayerNetworkState::Empty:
    case WebCore::MediaPlayerNetworkState::Idle:
    case WebCore::MediaPlayerNetworkState::Loading:
    case WebCore::MediaPlayerNetworkState::Loaded:
    case WebCore::MediaPlayerNetworkState::FormatError:
    case WebCore::MediaPlayerNetworkState::NetworkError:
    case WebCore::MediaPlayerNetworkState::DecodeError:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerReadyState, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerReadyState>(value)) {
    case WebCore::MediaPlayerReadyState::HaveNothing:
    case WebCore::MediaPlayerReadyState::HaveMetadata:
    case WebCore::MediaPlayerReadyState::HaveCurrentData:
    case WebCore::MediaPlayerReadyState::HaveFutureData:
    case WebCore::MediaPlayerReadyState::HaveEnoughData:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerMovieLoadType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerMovieLoadType>(value)) {
    case WebCore::MediaPlayerMovieLoadType::Unknown:
    case WebCore::MediaPlayerMovieLoadType::Download:
    case WebCore::MediaPlayerMovieLoadType::StoredStream:
    case WebCore::MediaPlayerMovieLoadType::LiveStream:
    case WebCore::MediaPlayerMovieLoadType::HttpLiveStream:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerPreload, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerPreload>(value)) {
    case WebCore::MediaPlayerPreload::None:
    case WebCore::MediaPlayerPreload::MetaData:
    case WebCore::MediaPlayerPreload::Auto:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerVideoGravity, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerVideoGravity>(value)) {
    case WebCore::MediaPlayerVideoGravity::Resize:
    case WebCore::MediaPlayerVideoGravity::ResizeAspect:
    case WebCore::MediaPlayerVideoGravity::ResizeAspectFill:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerSupportsType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerSupportsType>(value)) {
    case WebCore::MediaPlayerSupportsType::IsNotSupported:
    case WebCore::MediaPlayerSupportsType::IsSupported:
    case WebCore::MediaPlayerSupportsType::MayBeSupported:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerBufferingPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerBufferingPolicy>(value)) {
    case WebCore::MediaPlayerBufferingPolicy::Default:
    case WebCore::MediaPlayerBufferingPolicy::LimitReadAhead:
    case WebCore::MediaPlayerBufferingPolicy::MakeResourcesPurgeable:
    case WebCore::MediaPlayerBufferingPolicy::PurgeResources:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerMediaEngineIdentifier, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerMediaEngineIdentifier>(value)) {
    case WebCore::MediaPlayerMediaEngineIdentifier::AVFoundation:
    case WebCore::MediaPlayerMediaEngineIdentifier::AVFoundationMSE:
    case WebCore::MediaPlayerMediaEngineIdentifier::AVFoundationMediaStream:
    case WebCore::MediaPlayerMediaEngineIdentifier::AVFoundationCF:
    case WebCore::MediaPlayerMediaEngineIdentifier::GStreamer:
    case WebCore::MediaPlayerMediaEngineIdentifier::GStreamerMSE:
    case WebCore::MediaPlayerMediaEngineIdentifier::HolePunch:
    case WebCore::MediaPlayerMediaEngineIdentifier::MediaFoundation:
    case WebCore::MediaPlayerMediaEngineIdentifier::MockMSE:
    case WebCore::MediaPlayerMediaEngineIdentifier::CocoaWebM:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerWirelessPlaybackTargetType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerWirelessPlaybackTargetType>(value)) {
    case WebCore::MediaPlayerWirelessPlaybackTargetType::TargetTypeNone:
    case WebCore::MediaPlayerWirelessPlaybackTargetType::TargetTypeAirPlay:
    case WebCore::MediaPlayerWirelessPlaybackTargetType::TargetTypeTVOut:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MediaPlayerPitchCorrectionAlgorithm, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaPlayerPitchCorrectionAlgorithm>(value)) {
    case WebCore::MediaPlayerPitchCorrectionAlgorithm::BestAllAround:
    case WebCore::MediaPlayerPitchCorrectionAlgorithm::BestForMusic:
    case WebCore::MediaPlayerPitchCorrectionAlgorithm::BestForSpeech:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::SpeechRecognitionUpdateType, void>(uint8_t value)
{
    switch (static_cast<WebCore::SpeechRecognitionUpdateType>(value)) {
    case WebCore::SpeechRecognitionUpdateType::Start:
    case WebCore::SpeechRecognitionUpdateType::AudioStart:
    case WebCore::SpeechRecognitionUpdateType::SoundStart:
    case WebCore::SpeechRecognitionUpdateType::SpeechStart:
    case WebCore::SpeechRecognitionUpdateType::SpeechEnd:
    case WebCore::SpeechRecognitionUpdateType::SoundEnd:
    case WebCore::SpeechRecognitionUpdateType::AudioEnd:
    case WebCore::SpeechRecognitionUpdateType::Result:
    case WebCore::SpeechRecognitionUpdateType::NoMatch:
    case WebCore::SpeechRecognitionUpdateType::Error:
    case WebCore::SpeechRecognitionUpdateType::End:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::CompositeOperator, void>(uint8_t value)
{
    switch (static_cast<WebCore::CompositeOperator>(value)) {
    case WebCore::CompositeOperator::Clear:
    case WebCore::CompositeOperator::Copy:
    case WebCore::CompositeOperator::SourceOver:
    case WebCore::CompositeOperator::SourceIn:
    case WebCore::CompositeOperator::SourceOut:
    case WebCore::CompositeOperator::SourceAtop:
    case WebCore::CompositeOperator::DestinationOver:
    case WebCore::CompositeOperator::DestinationIn:
    case WebCore::CompositeOperator::DestinationOut:
    case WebCore::CompositeOperator::DestinationAtop:
    case WebCore::CompositeOperator::XOR:
    case WebCore::CompositeOperator::PlusDarker:
    case WebCore::CompositeOperator::PlusLighter:
    case WebCore::CompositeOperator::Difference:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::BlendMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::BlendMode>(value)) {
    case WebCore::BlendMode::Normal:
    case WebCore::BlendMode::Multiply:
    case WebCore::BlendMode::Screen:
    case WebCore::BlendMode::Darken:
    case WebCore::BlendMode::Lighten:
    case WebCore::BlendMode::Overlay:
    case WebCore::BlendMode::ColorDodge:
    case WebCore::BlendMode::ColorBurn:
    case WebCore::BlendMode::HardLight:
    case WebCore::BlendMode::SoftLight:
    case WebCore::BlendMode::Difference:
    case WebCore::BlendMode::Exclusion:
    case WebCore::BlendMode::Hue:
    case WebCore::BlendMode::Saturation:
    case WebCore::BlendMode::Color:
    case WebCore::BlendMode::Luminosity:
    case WebCore::BlendMode::PlusDarker:
    case WebCore::BlendMode::PlusLighter:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::GradientSpreadMethod, void>(uint8_t value)
{
    switch (static_cast<WebCore::GradientSpreadMethod>(value)) {
    case WebCore::GradientSpreadMethod::Pad:
    case WebCore::GradientSpreadMethod::Reflect:
    case WebCore::GradientSpreadMethod::Repeat:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::InterpolationQuality, void>(uint8_t value)
{
    switch (static_cast<WebCore::InterpolationQuality>(value)) {
    case WebCore::InterpolationQuality::Default:
    case WebCore::InterpolationQuality::DoNotInterpolate:
    case WebCore::InterpolationQuality::Low:
    case WebCore::InterpolationQuality::Medium:
    case WebCore::InterpolationQuality::High:
        return true;
    default:
        return false;
    }
}

#if ENABLE(VIDEO)
template<> bool isValidEnum<WebCore::VideoFrameRotation, void>(uint16_t value)
{
    switch (static_cast<WebCore::VideoFrameRotation>(value)) {
    case WebCore::VideoFrameRotation::None:
    case WebCore::VideoFrameRotation::UpsideDown:
    case WebCore::VideoFrameRotation::Right:
    case WebCore::VideoFrameRotation::Left:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::ColorInterpolationColorSpace, void>(uint8_t value)
{
    switch (static_cast<WebCore::ColorInterpolationColorSpace>(value)) {
    case WebCore::ColorInterpolationColorSpace::HSL:
    case WebCore::ColorInterpolationColorSpace::HWB:
    case WebCore::ColorInterpolationColorSpace::LCH:
    case WebCore::ColorInterpolationColorSpace::Lab:
    case WebCore::ColorInterpolationColorSpace::OKLCH:
    case WebCore::ColorInterpolationColorSpace::OKLab:
    case WebCore::ColorInterpolationColorSpace::SRGB:
    case WebCore::ColorInterpolationColorSpace::SRGBLinear:
    case WebCore::ColorInterpolationColorSpace::XYZD50:
    case WebCore::ColorInterpolationColorSpace::XYZD65:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::HueInterpolationMethod, void>(uint8_t value)
{
    switch (static_cast<WebCore::HueInterpolationMethod>(value)) {
    case WebCore::HueInterpolationMethod::Shorter:
    case WebCore::HueInterpolationMethod::Longer:
    case WebCore::HueInterpolationMethod::Increasing:
    case WebCore::HueInterpolationMethod::Decreasing:
        return true;
    default:
        return false;
    }
}

#if USE(AUDIO_SESSION)
template<> bool isValidEnum<WebCore::RouteSharingPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::RouteSharingPolicy>(value)) {
    case WebCore::RouteSharingPolicy::Default:
    case WebCore::RouteSharingPolicy::LongFormAudio:
    case WebCore::RouteSharingPolicy::Independent:
    case WebCore::RouteSharingPolicy::LongFormVideo:
        return true;
    default:
        return false;
    }
}
#endif

#if USE(AUDIO_SESSION)
template<> bool isValidEnum<WebCore::AudioSessionCategory, void>(uint8_t value)
{
    switch (static_cast<WebCore::AudioSessionCategory>(value)) {
    case WebCore::AudioSessionCategory::None:
    case WebCore::AudioSessionCategory::AmbientSound:
    case WebCore::AudioSessionCategory::SoloAmbientSound:
    case WebCore::AudioSessionCategory::MediaPlayback:
    case WebCore::AudioSessionCategory::RecordAudio:
    case WebCore::AudioSessionCategory::PlayAndRecord:
    case WebCore::AudioSessionCategory::AudioProcessing:
        return true;
    default:
        return false;
    }
}
#endif

#if USE(AUDIO_SESSION)
template<> bool isValidEnum<WebCore::AudioSessionMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::AudioSessionMode>(value)) {
    case WebCore::AudioSessionMode::Default:
    case WebCore::AudioSessionMode::VideoChat:
    case WebCore::AudioSessionMode::MoviePlayback:
        return true;
    default:
        return false;
    }
}
#endif

#if USE(AUDIO_SESSION)
template<> bool isValidEnum<WebCore::AudioSessionRoutingArbitrationError, void>(uint8_t value)
{
    switch (static_cast<WebCore::AudioSessionRoutingArbitrationError>(value)) {
    case WebCore::AudioSessionRoutingArbitrationError::None:
    case WebCore::AudioSessionRoutingArbitrationError::Failed:
    case WebCore::AudioSessionRoutingArbitrationError::Cancelled:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(ENCRYPTED_MEDIA)
template<> bool isValidEnum<WebCore::CDMKeyStatus, void>(uint8_t value)
{
    switch (static_cast<WebCore::CDMKeyStatus>(value)) {
    case WebCore::CDMKeyStatus::Usable:
    case WebCore::CDMKeyStatus::Expired:
    case WebCore::CDMKeyStatus::Released:
    case WebCore::CDMKeyStatus::OutputRestricted:
    case WebCore::CDMKeyStatus::OutputDownscaled:
    case WebCore::CDMKeyStatus::StatusPending:
    case WebCore::CDMKeyStatus::InternalError:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(ENCRYPTED_MEDIA)
template<> bool isValidEnum<WebCore::CDMMessageType, void>(uint8_t value)
{
    switch (static_cast<WebCore::CDMMessageType>(value)) {
    case WebCore::CDMMessageType::LicenseRequest:
    case WebCore::CDMMessageType::LicenseRenewal:
    case WebCore::CDMMessageType::LicenseRelease:
    case WebCore::CDMMessageType::IndividualizationRequest:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(ENCRYPTED_MEDIA)
template<> bool isValidEnum<WebCore::CDMRequirement, void>(uint8_t value)
{
    switch (static_cast<WebCore::CDMRequirement>(value)) {
    case WebCore::CDMRequirement::Required:
    case WebCore::CDMRequirement::Optional:
    case WebCore::CDMRequirement::NotAllowed:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(WEB_RTC)
template<> bool isValidEnum<WebCore::RTCDataChannelState, void>(uint8_t value)
{
    switch (static_cast<WebCore::RTCDataChannelState>(value)) {
    case WebCore::RTCDataChannelState::Connecting:
    case WebCore::RTCDataChannelState::Open:
    case WebCore::RTCDataChannelState::Closing:
    case WebCore::RTCDataChannelState::Closed:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(WEB_RTC)
template<> bool isValidEnum<WebCore::RTCPriorityType, void>(uint8_t value)
{
    switch (static_cast<WebCore::RTCPriorityType>(value)) {
    case WebCore::RTCPriorityType::VeryLow:
    case WebCore::RTCPriorityType::Low:
    case WebCore::RTCPriorityType::Medium:
    case WebCore::RTCPriorityType::High:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::LineCap, void>(uint8_t value)
{
    switch (static_cast<WebCore::LineCap>(value)) {
    case WebCore::LineCap::Butt:
    case WebCore::LineCap::Round:
    case WebCore::LineCap::Square:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::LineJoin, void>(uint8_t value)
{
    switch (static_cast<WebCore::LineJoin>(value)) {
    case WebCore::LineJoin::Miter:
    case WebCore::LineJoin::Round:
    case WebCore::LineJoin::Bevel:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::StrokeStyle, void>(uint8_t value)
{
    switch (static_cast<WebCore::StrokeStyle>(value)) {
    case WebCore::StrokeStyle::NoStroke:
    case WebCore::StrokeStyle::SolidStroke:
    case WebCore::StrokeStyle::DottedStroke:
    case WebCore::StrokeStyle::DashedStroke:
    case WebCore::StrokeStyle::DoubleStroke:
    case WebCore::StrokeStyle::WavyStroke:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::TextDrawingMode>(OptionSet<WebCore::TextDrawingMode> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::TextDrawingMode::Fill)
        | static_cast<uint8_t>(WebCore::TextDrawingMode::Stroke)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::DecodingMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::DecodingMode>(value)) {
    case WebCore::DecodingMode::Auto:
    case WebCore::DecodingMode::Synchronous:
    case WebCore::DecodingMode::Asynchronous:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::DocumentMarkerLineStyleMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::DocumentMarkerLineStyleMode>(value)) {
    case WebCore::DocumentMarkerLineStyleMode::TextCheckingDictationPhraseWithAlternatives:
    case WebCore::DocumentMarkerLineStyleMode::Spelling:
    case WebCore::DocumentMarkerLineStyleMode::Grammar:
    case WebCore::DocumentMarkerLineStyleMode::AutocorrectionReplacement:
    case WebCore::DocumentMarkerLineStyleMode::DictationAlternatives:
        return true;
    default:
        return false;
    }
}

#if ENABLE(VIDEO)
template<> bool isValidEnum<WebCore::GenericCueData::Alignment, void>(uint8_t value)
{
    switch (static_cast<WebCore::GenericCueData::Alignment>(value)) {
    case WebCore::GenericCueData::Alignment::None:
    case WebCore::GenericCueData::Alignment::Start:
    case WebCore::GenericCueData::Alignment::Middle:
    case WebCore::GenericCueData::Alignment::End:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(VIDEO)
template<> bool isValidEnum<WebCore::GenericCueData::Status, void>(uint8_t value)
{
    switch (static_cast<WebCore::GenericCueData::Status>(value)) {
    case WebCore::GenericCueData::Status::Uninitialized:
    case WebCore::GenericCueData::Status::Partial:
    case WebCore::GenericCueData::Status::Complete:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::ImageOrientation::Orientation, void>(int value)
{
    switch (static_cast<WebCore::ImageOrientation::Orientation>(value)) {
    case WebCore::ImageOrientation::Orientation::FromImage:
    case WebCore::ImageOrientation::Orientation::OriginTopLeft:
    case WebCore::ImageOrientation::Orientation::OriginTopRight:
    case WebCore::ImageOrientation::Orientation::OriginBottomRight:
    case WebCore::ImageOrientation::Orientation::OriginBottomLeft:
    case WebCore::ImageOrientation::Orientation::OriginLeftTop:
    case WebCore::ImageOrientation::Orientation::OriginRightTop:
    case WebCore::ImageOrientation::Orientation::OriginRightBottom:
    case WebCore::ImageOrientation::Orientation::OriginLeftBottom:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::CrossOriginOpenerPolicyValue, void>(uint8_t value)
{
    switch (static_cast<WebCore::CrossOriginOpenerPolicyValue>(value)) {
    case WebCore::CrossOriginOpenerPolicyValue::UnsafeNone:
    case WebCore::CrossOriginOpenerPolicyValue::SameOrigin:
    case WebCore::CrossOriginOpenerPolicyValue::SameOriginPlusCOEP:
    case WebCore::CrossOriginOpenerPolicyValue::SameOriginAllowPopups:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::HTTPCookieAcceptPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::HTTPCookieAcceptPolicy>(value)) {
    case WebCore::HTTPCookieAcceptPolicy::AlwaysAccept:
    case WebCore::HTTPCookieAcceptPolicy::Never:
    case WebCore::HTTPCookieAcceptPolicy::OnlyFromMainDocumentDomain:
    case WebCore::HTTPCookieAcceptPolicy::ExclusivelyFromMainDocumentDomain:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::NetworkLoadPriority, void>(uint8_t value)
{
    switch (static_cast<WebCore::NetworkLoadPriority>(value)) {
    case WebCore::NetworkLoadPriority::Low:
    case WebCore::NetworkLoadPriority::Medium:
    case WebCore::NetworkLoadPriority::High:
    case WebCore::NetworkLoadPriority::Unknown:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PrivacyStance, void>(uint8_t value)
{
    switch (static_cast<WebCore::PrivacyStance>(value)) {
    case WebCore::PrivacyStance::Unknown:
    case WebCore::PrivacyStance::NotEligible:
    case WebCore::PrivacyStance::Proxied:
    case WebCore::PrivacyStance::Failed:
    case WebCore::PrivacyStance::Direct:
    case WebCore::PrivacyStance::FailedUnreachable:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ThirdPartyCookieBlockingMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::ThirdPartyCookieBlockingMode>(value)) {
    case WebCore::ThirdPartyCookieBlockingMode::All:
    case WebCore::ThirdPartyCookieBlockingMode::AllExceptBetweenAppBoundDomains:
    case WebCore::ThirdPartyCookieBlockingMode::AllExceptManagedDomains:
    case WebCore::ThirdPartyCookieBlockingMode::AllOnSitesWithoutUserInteraction:
    case WebCore::ThirdPartyCookieBlockingMode::OnlyAccordingToPerDomainPolicy:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::FirstPartyWebsiteDataRemovalMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::FirstPartyWebsiteDataRemovalMode>(value)) {
    case WebCore::FirstPartyWebsiteDataRemovalMode::AllButCookies:
    case WebCore::FirstPartyWebsiteDataRemovalMode::None:
    case WebCore::FirstPartyWebsiteDataRemovalMode::AllButCookiesLiveOnTestingTimeout:
    case WebCore::FirstPartyWebsiteDataRemovalMode::AllButCookiesReproTestingTimeout:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::DragApplicationFlags>(OptionSet<WebCore::DragApplicationFlags> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::DragApplicationFlags::IsModal)
        | static_cast<uint8_t>(WebCore::DragApplicationFlags::IsSource)
        | static_cast<uint8_t>(WebCore::DragApplicationFlags::HasAttachedSheet)
        | static_cast<uint8_t>(WebCore::DragApplicationFlags::IsCopyKeyDown)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidOptionSet<WebCore::TextCheckingType>(OptionSet<WebCore::TextCheckingType> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::TextCheckingType::None)
        | static_cast<uint8_t>(WebCore::TextCheckingType::Spelling)
        | static_cast<uint8_t>(WebCore::TextCheckingType::Grammar)
        | static_cast<uint8_t>(WebCore::TextCheckingType::Link)
        | static_cast<uint8_t>(WebCore::TextCheckingType::Quote)
        | static_cast<uint8_t>(WebCore::TextCheckingType::Dash)
        | static_cast<uint8_t>(WebCore::TextCheckingType::Replacement)
        | static_cast<uint8_t>(WebCore::TextCheckingType::Correction)
        | static_cast<uint8_t>(WebCore::TextCheckingType::ShowCorrectionPanel)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

#if ENABLE(SERVICE_WORKER)
template<> bool isValidEnum<WebCore::ServiceWorkerClientType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ServiceWorkerClientType>(value)) {
    case WebCore::ServiceWorkerClientType::Window:
    case WebCore::ServiceWorkerClientType::Worker:
    case WebCore::ServiceWorkerClientType::Sharedworker:
    case WebCore::ServiceWorkerClientType::All:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(SERVICE_WORKER)
template<> bool isValidEnum<WebCore::ServiceWorkerJobType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ServiceWorkerJobType>(value)) {
    case WebCore::ServiceWorkerJobType::Register:
    case WebCore::ServiceWorkerJobType::Unregister:
    case WebCore::ServiceWorkerJobType::Update:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(SERVICE_WORKER)
template<> bool isValidEnum<WebCore::ServiceWorkerRegistrationState, void>(uint8_t value)
{
    switch (static_cast<WebCore::ServiceWorkerRegistrationState>(value)) {
    case WebCore::ServiceWorkerRegistrationState::Installing:
    case WebCore::ServiceWorkerRegistrationState::Waiting:
    case WebCore::ServiceWorkerRegistrationState::Active:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(SERVICE_WORKER)
template<> bool isValidEnum<WebCore::ServiceWorkerState, void>(uint8_t value)
{
    switch (static_cast<WebCore::ServiceWorkerState>(value)) {
    case WebCore::ServiceWorkerState::Parsed:
    case WebCore::ServiceWorkerState::Installing:
    case WebCore::ServiceWorkerState::Installed:
    case WebCore::ServiceWorkerState::Activating:
    case WebCore::ServiceWorkerState::Activated:
    case WebCore::ServiceWorkerState::Redundant:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(SERVICE_WORKER)
template<> bool isValidEnum<WebCore::ServiceWorkerClientFrameType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ServiceWorkerClientFrameType>(value)) {
    case WebCore::ServiceWorkerClientFrameType::Auxiliary:
    case WebCore::ServiceWorkerClientFrameType::TopLevel:
    case WebCore::ServiceWorkerClientFrameType::Nested:
    case WebCore::ServiceWorkerClientFrameType::None:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(SERVICE_WORKER)
template<> bool isValidEnum<WebCore::ServiceWorkerUpdateViaCache, void>(uint8_t value)
{
    switch (static_cast<WebCore::ServiceWorkerUpdateViaCache>(value)) {
    case WebCore::ServiceWorkerUpdateViaCache::Imports:
    case WebCore::ServiceWorkerUpdateViaCache::All:
    case WebCore::ServiceWorkerUpdateViaCache::None:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_STREAM)
template<> bool isValidEnum<WebCore::MediaConstraint::DataType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaConstraint::DataType>(value)) {
    case WebCore::MediaConstraint::DataType::None:
    case WebCore::MediaConstraint::DataType::Integer:
    case WebCore::MediaConstraint::DataType::Double:
    case WebCore::MediaConstraint::DataType::Boolean:
    case WebCore::MediaConstraint::DataType::String:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_STREAM)
template<> bool isValidEnum<WebCore::MediaConstraintType, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaConstraintType>(value)) {
    case WebCore::MediaConstraintType::Unknown:
    case WebCore::MediaConstraintType::Width:
    case WebCore::MediaConstraintType::Height:
    case WebCore::MediaConstraintType::AspectRatio:
    case WebCore::MediaConstraintType::FrameRate:
    case WebCore::MediaConstraintType::FacingMode:
    case WebCore::MediaConstraintType::Volume:
    case WebCore::MediaConstraintType::SampleRate:
    case WebCore::MediaConstraintType::SampleSize:
    case WebCore::MediaConstraintType::EchoCancellation:
    case WebCore::MediaConstraintType::DeviceId:
    case WebCore::MediaConstraintType::GroupId:
    case WebCore::MediaConstraintType::DisplaySurface:
    case WebCore::MediaConstraintType::LogicalSurface:
    case WebCore::MediaConstraintType::Zoom:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_STREAM)
template<> bool isValidEnum<WebCore::VideoFacingMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::VideoFacingMode>(value)) {
    case WebCore::VideoFacingMode::Unknown:
    case WebCore::VideoFacingMode::User:
    case WebCore::VideoFacingMode::Environment:
    case WebCore::VideoFacingMode::Left:
    case WebCore::VideoFacingMode::Right:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_STREAM)
template<> bool isValidEnum<WebCore::DisplaySurfaceType, void>(uint8_t value)
{
    switch (static_cast<WebCore::DisplaySurfaceType>(value)) {
    case WebCore::DisplaySurfaceType::Monitor:
    case WebCore::DisplaySurfaceType::Window:
    case WebCore::DisplaySurfaceType::Application:
    case WebCore::DisplaySurfaceType::Browser:
    case WebCore::DisplaySurfaceType::Invalid:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_STREAM)
template<> bool isValidEnum<WebCore::CaptureDevice::DeviceType, void>(uint8_t value)
{
    switch (static_cast<WebCore::CaptureDevice::DeviceType>(value)) {
    case WebCore::CaptureDevice::DeviceType::Unknown:
    case WebCore::CaptureDevice::DeviceType::Microphone:
    case WebCore::CaptureDevice::DeviceType::Speaker:
    case WebCore::CaptureDevice::DeviceType::Camera:
    case WebCore::CaptureDevice::DeviceType::Screen:
    case WebCore::CaptureDevice::DeviceType::Window:
    case WebCore::CaptureDevice::DeviceType::SystemAudio:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(MEDIA_STREAM)
template<> bool isValidOptionSet<WebCore::MockMediaDevice::Flag>(OptionSet<WebCore::MockMediaDevice::Flag> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::MockMediaDevice::Flag::Ephemeral)
        | static_cast<uint8_t>(WebCore::MockMediaDevice::Flag::Invalid)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}
#endif

template<> bool isValidEnum<WebCore::PlatformVideoColorPrimaries, void>(uint8_t value)
{
    switch (static_cast<WebCore::PlatformVideoColorPrimaries>(value)) {
    case WebCore::PlatformVideoColorPrimaries::Bt709:
    case WebCore::PlatformVideoColorPrimaries::Bt470bg:
    case WebCore::PlatformVideoColorPrimaries::Smpte170m:
    case WebCore::PlatformVideoColorPrimaries::Bt470m:
    case WebCore::PlatformVideoColorPrimaries::Smpte240m:
    case WebCore::PlatformVideoColorPrimaries::Film:
    case WebCore::PlatformVideoColorPrimaries::Bt2020:
    case WebCore::PlatformVideoColorPrimaries::SmpteSt4281:
    case WebCore::PlatformVideoColorPrimaries::SmpteRp431:
    case WebCore::PlatformVideoColorPrimaries::SmpteEg432:
    case WebCore::PlatformVideoColorPrimaries::JedecP22Phosphors:
    case WebCore::PlatformVideoColorPrimaries::Unspecified:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PlatformVideoTransferCharacteristics, void>(uint8_t value)
{
    switch (static_cast<WebCore::PlatformVideoTransferCharacteristics>(value)) {
    case WebCore::PlatformVideoTransferCharacteristics::Bt709:
    case WebCore::PlatformVideoTransferCharacteristics::Smpte170m:
    case WebCore::PlatformVideoTransferCharacteristics::Iec6196621:
    case WebCore::PlatformVideoTransferCharacteristics::Gamma22curve:
    case WebCore::PlatformVideoTransferCharacteristics::Gamma28curve:
    case WebCore::PlatformVideoTransferCharacteristics::Smpte240m:
    case WebCore::PlatformVideoTransferCharacteristics::Linear:
    case WebCore::PlatformVideoTransferCharacteristics::Log:
    case WebCore::PlatformVideoTransferCharacteristics::LogSqrt:
    case WebCore::PlatformVideoTransferCharacteristics::Iec6196624:
    case WebCore::PlatformVideoTransferCharacteristics::Bt1361ExtendedColourGamut:
    case WebCore::PlatformVideoTransferCharacteristics::Bt2020_10bit:
    case WebCore::PlatformVideoTransferCharacteristics::Bt2020_12bit:
    case WebCore::PlatformVideoTransferCharacteristics::SmpteSt2084:
    case WebCore::PlatformVideoTransferCharacteristics::SmpteSt4281:
    case WebCore::PlatformVideoTransferCharacteristics::AribStdB67Hlg:
    case WebCore::PlatformVideoTransferCharacteristics::Unspecified:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PlatformVideoMatrixCoefficients, void>(uint8_t value)
{
    switch (static_cast<WebCore::PlatformVideoMatrixCoefficients>(value)) {
    case WebCore::PlatformVideoMatrixCoefficients::Rgb:
    case WebCore::PlatformVideoMatrixCoefficients::Bt709:
    case WebCore::PlatformVideoMatrixCoefficients::Bt470bg:
    case WebCore::PlatformVideoMatrixCoefficients::Smpte170m:
    case WebCore::PlatformVideoMatrixCoefficients::Smpte240m:
    case WebCore::PlatformVideoMatrixCoefficients::Fcc:
    case WebCore::PlatformVideoMatrixCoefficients::YCgCo:
    case WebCore::PlatformVideoMatrixCoefficients::Bt2020NonconstantLuminance:
    case WebCore::PlatformVideoMatrixCoefficients::Bt2020ConstantLuminance:
    case WebCore::PlatformVideoMatrixCoefficients::Unspecified:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PolicyAction, void>(uint8_t value)
{
    switch (static_cast<WebCore::PolicyAction>(value)) {
    case WebCore::PolicyAction::Use:
    case WebCore::PolicyAction::Download:
    case WebCore::PolicyAction::Ignore:
    case WebCore::PolicyAction::LoadWillContinueInAnotherProcess:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::ReloadOption>(OptionSet<WebCore::ReloadOption> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::ReloadOption::ExpiredOnly)
        | static_cast<uint8_t>(WebCore::ReloadOption::FromOrigin)
        | static_cast<uint8_t>(WebCore::ReloadOption::DisableContentBlockers)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::FrameLoadType, void>(uint8_t value)
{
    switch (static_cast<WebCore::FrameLoadType>(value)) {
    case WebCore::FrameLoadType::Standard:
    case WebCore::FrameLoadType::Back:
    case WebCore::FrameLoadType::Forward:
    case WebCore::FrameLoadType::IndexedBackForward:
    case WebCore::FrameLoadType::Reload:
    case WebCore::FrameLoadType::Same:
    case WebCore::FrameLoadType::RedirectWithLockedBackForwardList:
    case WebCore::FrameLoadType::Replace:
    case WebCore::FrameLoadType::ReloadFromOrigin:
    case WebCore::FrameLoadType::ReloadExpiredOnly:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::NavigationType, void>(uint8_t value)
{
    switch (static_cast<WebCore::NavigationType>(value)) {
    case WebCore::NavigationType::LinkClicked:
    case WebCore::NavigationType::FormSubmitted:
    case WebCore::NavigationType::BackForward:
    case WebCore::NavigationType::Reload:
    case WebCore::NavigationType::FormResubmitted:
    case WebCore::NavigationType::Other:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::BrowsingContextGroupSwitchDecision, void>(uint8_t value)
{
    switch (static_cast<WebCore::BrowsingContextGroupSwitchDecision>(value)) {
    case WebCore::BrowsingContextGroupSwitchDecision::StayInGroup:
    case WebCore::BrowsingContextGroupSwitchDecision::NewSharedGroup:
    case WebCore::BrowsingContextGroupSwitchDecision::NewIsolatedGroup:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ShouldOpenExternalURLsPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::ShouldOpenExternalURLsPolicy>(value)) {
    case WebCore::ShouldOpenExternalURLsPolicy::ShouldNotAllow:
    case WebCore::ShouldOpenExternalURLsPolicy::ShouldAllowExternalSchemesButNotAppLinks:
    case WebCore::ShouldOpenExternalURLsPolicy::ShouldAllow:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::MouseEventPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::MouseEventPolicy>(value)) {
    case WebCore::MouseEventPolicy::Default:
#if ENABLE(IOS_TOUCH_EVENTS)
    case WebCore::MouseEventPolicy::SynthesizeTouchEvents:
#endif
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ColorSchemePreference, void>(uint8_t value)
{
    switch (static_cast<WebCore::ColorSchemePreference>(value)) {
    case WebCore::ColorSchemePreference::NoPreference:
    case WebCore::ColorSchemePreference::Light:
    case WebCore::ColorSchemePreference::Dark:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::AdvancedPrivacyProtections>(OptionSet<WebCore::AdvancedPrivacyProtections> value)
{
    constexpr uint16_t allValidBitsValue = 0
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::BaselineProtections)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::HTTPSFirst)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::HTTPSOnly)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::HTTPSOnlyExplicitlyBypassedForDomain)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::FailClosed)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::WebSearchContent)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::FingerprintingProtections)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::EnhancedNetworkPrivacy)
        | static_cast<uint16_t>(WebCore::AdvancedPrivacyProtections::LinkDecorationFiltering)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::ViolationReportType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ViolationReportType>(value)) {
    case WebCore::ViolationReportType::COEPInheritenceViolation:
    case WebCore::ViolationReportType::CORPViolation:
    case WebCore::ViolationReportType::ContentSecurityPolicy:
    case WebCore::ViolationReportType::CrossOriginOpenerPolicy:
    case WebCore::ViolationReportType::Deprecation:
    case WebCore::ViolationReportType::StandardReportingAPIViolation:
    case WebCore::ViolationReportType::Test:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::LinkIconType, void>(uint8_t value)
{
    switch (static_cast<WebCore::LinkIconType>(value)) {
    case WebCore::LinkIconType::Favicon:
    case WebCore::LinkIconType::TouchIcon:
    case WebCore::LinkIconType::TouchPrecomposedIcon:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PluginLoadClientPolicy, void>(uint8_t value)
{
    switch (static_cast<WebCore::PluginLoadClientPolicy>(value)) {
    case WebCore::PluginLoadClientPolicy::Undefined:
    case WebCore::PluginLoadClientPolicy::Block:
    case WebCore::PluginLoadClientPolicy::Ask:
    case WebCore::PluginLoadClientPolicy::Allow:
    case WebCore::PluginLoadClientPolicy::AllowAlways:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::PasteboardItemPresentationStyle, void>(uint8_t value)
{
    switch (static_cast<WebCore::PasteboardItemPresentationStyle>(value)) {
    case WebCore::PasteboardItemPresentationStyle::Unspecified:
    case WebCore::PasteboardItemPresentationStyle::Inline:
    case WebCore::PasteboardItemPresentationStyle::Attachment:
        return true;
    default:
        return false;
    }
}

#if HAVE(SCREEN_CAPTURE_KIT)
template<> bool isValidEnum<WebCore::DisplayCapturePromptType, void>(uint8_t value)
{
    switch (static_cast<WebCore::DisplayCapturePromptType>(value)) {
    case WebCore::DisplayCapturePromptType::Window:
    case WebCore::DisplayCapturePromptType::Screen:
    case WebCore::DisplayCapturePromptType::UserChoose:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::TextIndicatorPresentationTransition, void>(uint8_t value)
{
    switch (static_cast<WebCore::TextIndicatorPresentationTransition>(value)) {
    case WebCore::TextIndicatorPresentationTransition::None:
    case WebCore::TextIndicatorPresentationTransition::Bounce:
    case WebCore::TextIndicatorPresentationTransition::BounceAndCrossfade:
    case WebCore::TextIndicatorPresentationTransition::FadeIn:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::TextIndicatorOption>(OptionSet<WebCore::TextIndicatorOption> value)
{
    constexpr uint16_t allValidBitsValue = 0
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::RespectTextColor)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::PaintBackgrounds)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::PaintAllContent)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::IncludeSnapshotWithSelectionHighlight)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::TightlyFitContent)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::UseBoundingRectAndPaintAllContentForComplexRanges)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::IncludeMarginIfRangeMatchesSelection)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::ExpandClipBeyondVisibleRect)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::DoNotClipToVisibleRect)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::IncludeSnapshotOfAllVisibleContentWithoutSelection)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::UseSelectionRectForSizing)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::ComputeEstimatedBackgroundColor)
        | static_cast<uint16_t>(WebCore::TextIndicatorOption::UseUserSelectAllCommonAncestor)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::InteractionRegion::Type, void>(uint8_t value)
{
    switch (static_cast<WebCore::InteractionRegion::Type>(value)) {
    case WebCore::InteractionRegion::Type::Interaction:
    case WebCore::InteractionRegion::Type::Occlusion:
    case WebCore::InteractionRegion::Type::Guard:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebCore::InteractionRegion::CornerMask>(OptionSet<WebCore::InteractionRegion::CornerMask> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebCore::InteractionRegion::CornerMask::MinXMinYCorner)
        | static_cast<uint8_t>(WebCore::InteractionRegion::CornerMask::MaxXMinYCorner)
        | static_cast<uint8_t>(WebCore::InteractionRegion::CornerMask::MinXMaxYCorner)
        | static_cast<uint8_t>(WebCore::InteractionRegion::CornerMask::MaxXMaxYCorner)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::MockWebAuthenticationConfiguration::UserVerification, void>(uint8_t value)
{
    switch (static_cast<WebCore::MockWebAuthenticationConfiguration::UserVerification>(value)) {
    case WebCore::MockWebAuthenticationConfiguration::UserVerification::No:
    case WebCore::MockWebAuthenticationConfiguration::UserVerification::Yes:
    case WebCore::MockWebAuthenticationConfiguration::UserVerification::Cancel:
    case WebCore::MockWebAuthenticationConfiguration::UserVerification::Presence:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::MockWebAuthenticationConfiguration::HidError, void>(uint8_t value)
{
    switch (static_cast<WebCore::MockWebAuthenticationConfiguration::HidError>(value)) {
    case WebCore::MockWebAuthenticationConfiguration::HidError::Success:
    case WebCore::MockWebAuthenticationConfiguration::HidError::DataNotSent:
    case WebCore::MockWebAuthenticationConfiguration::HidError::EmptyReport:
    case WebCore::MockWebAuthenticationConfiguration::HidError::WrongChannelId:
    case WebCore::MockWebAuthenticationConfiguration::HidError::MaliciousPayload:
    case WebCore::MockWebAuthenticationConfiguration::HidError::UnsupportedOptions:
    case WebCore::MockWebAuthenticationConfiguration::HidError::WrongNonce:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(WEB_AUTHN)
template<> bool isValidEnum<WebCore::MockWebAuthenticationConfiguration::NfcError, void>(uint8_t value)
{
    switch (static_cast<WebCore::MockWebAuthenticationConfiguration::NfcError>(value)) {
    case WebCore::MockWebAuthenticationConfiguration::NfcError::Success:
    case WebCore::MockWebAuthenticationConfiguration::NfcError::NoTags:
    case WebCore::MockWebAuthenticationConfiguration::NfcError::WrongTagType:
    case WebCore::MockWebAuthenticationConfiguration::NfcError::NoConnections:
    case WebCore::MockWebAuthenticationConfiguration::NfcError::MaliciousPayload:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::StorageType, void>(uint8_t value)
{
    switch (static_cast<WebCore::StorageType>(value)) {
    case WebCore::StorageType::Session:
    case WebCore::StorageType::Local:
    case WebCore::StorageType::TransientLocal:
        return true;
    default:
        return false;
    }
}

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
template<> bool isValidOptionSet<WebCore::AcceleratedEffectProperty>(OptionSet<WebCore::AcceleratedEffectProperty> value)
{
    constexpr uint16_t allValidBitsValue = 0
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::Opacity)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::Transform)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::Translate)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::Rotate)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::Scale)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::OffsetPath)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::OffsetDistance)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::OffsetPosition)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::OffsetAnchor)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::OffsetRotate)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::Filter)
#if ENABLE(FILTERS_LEVEL_2)
        | static_cast<uint16_t>(WebCore::AcceleratedEffectProperty::BackdropFilter)
#endif
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}
#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
template<> bool isValidEnum<WebCore::WebAnimationType, void>(uint8_t value)
{
    switch (static_cast<WebCore::WebAnimationType>(value)) {
    case WebCore::WebAnimationType::CSSAnimation:
    case WebCore::WebAnimationType::CSSTransition:
    case WebCore::WebAnimationType::WebAnimation:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
template<> bool isValidEnum<WebCore::FillMode, void>(uint8_t value)
{
    switch (static_cast<WebCore::FillMode>(value)) {
    case WebCore::FillMode::None:
    case WebCore::FillMode::Forwards:
    case WebCore::FillMode::Backwards:
    case WebCore::FillMode::Both:
    case WebCore::FillMode::Auto:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
template<> bool isValidEnum<WebCore::PlaybackDirection, void>(uint8_t value)
{
    switch (static_cast<WebCore::PlaybackDirection>(value)) {
    case WebCore::PlaybackDirection::Normal:
    case WebCore::PlaybackDirection::Reverse:
    case WebCore::PlaybackDirection::Alternate:
    case WebCore::PlaybackDirection::AlternateReverse:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(THREADED_ANIMATION_RESOLUTION)
template<> bool isValidEnum<WebCore::CompositeOperation, void>(uint8_t value)
{
    switch (static_cast<WebCore::CompositeOperation>(value)) {
    case WebCore::CompositeOperation::Replace:
    case WebCore::CompositeOperation::Add:
    case WebCore::CompositeOperation::Accumulate:
        return true;
    default:
        return false;
    }
}
#endif

#if PLATFORM(MAC)
template<> bool isValidEnum<WebCore::CaretAnimatorType, void>(uint8_t value)
{
    switch (static_cast<WebCore::CaretAnimatorType>(value)) {
    case WebCore::CaretAnimatorType::Default:
    case WebCore::CaretAnimatorType::Dictation:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::ProtectionSpace::ServerType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ProtectionSpace::ServerType>(value)) {
    case WebCore::ProtectionSpace::ServerType::HTTP:
    case WebCore::ProtectionSpace::ServerType::HTTPS:
    case WebCore::ProtectionSpace::ServerType::FTP:
    case WebCore::ProtectionSpace::ServerType::FTPS:
    case WebCore::ProtectionSpace::ServerType::ProxyHTTP:
    case WebCore::ProtectionSpace::ServerType::ProxyHTTPS:
    case WebCore::ProtectionSpace::ServerType::ProxyFTP:
    case WebCore::ProtectionSpace::ServerType::ProxySOCKS:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::ProtectionSpace::AuthenticationScheme, void>(uint8_t value)
{
    switch (static_cast<WebCore::ProtectionSpace::AuthenticationScheme>(value)) {
    case WebCore::ProtectionSpace::AuthenticationScheme::Default:
    case WebCore::ProtectionSpace::AuthenticationScheme::HTTPBasic:
    case WebCore::ProtectionSpace::AuthenticationScheme::HTTPDigest:
    case WebCore::ProtectionSpace::AuthenticationScheme::HTMLForm:
    case WebCore::ProtectionSpace::AuthenticationScheme::NTLM:
    case WebCore::ProtectionSpace::AuthenticationScheme::Negotiate:
    case WebCore::ProtectionSpace::AuthenticationScheme::ClientCertificateRequested:
    case WebCore::ProtectionSpace::AuthenticationScheme::ServerTrustEvaluationRequested:
    case WebCore::ProtectionSpace::AuthenticationScheme::OAuth:
#if USE(GLIB)
    case WebCore::ProtectionSpace::AuthenticationScheme::ClientCertificatePINRequested:
#endif
    case WebCore::ProtectionSpace::AuthenticationScheme::Unknown:
        return true;
    default:
        return false;
    }
}

#if ENABLE(GPU_PROCESS)
template<> bool isValidEnum<WebCore::ShapeDetection::BarcodeFormat, void>(uint8_t value)
{
    switch (static_cast<WebCore::ShapeDetection::BarcodeFormat>(value)) {
    case WebCore::ShapeDetection::BarcodeFormat::Aztec:
    case WebCore::ShapeDetection::BarcodeFormat::Code_128:
    case WebCore::ShapeDetection::BarcodeFormat::Code_39:
    case WebCore::ShapeDetection::BarcodeFormat::Code_93:
    case WebCore::ShapeDetection::BarcodeFormat::Codabar:
    case WebCore::ShapeDetection::BarcodeFormat::Data_matrix:
    case WebCore::ShapeDetection::BarcodeFormat::Ean_13:
    case WebCore::ShapeDetection::BarcodeFormat::Ean_8:
    case WebCore::ShapeDetection::BarcodeFormat::Itf:
    case WebCore::ShapeDetection::BarcodeFormat::Pdf417:
    case WebCore::ShapeDetection::BarcodeFormat::Qr_code:
    case WebCore::ShapeDetection::BarcodeFormat::Unknown:
    case WebCore::ShapeDetection::BarcodeFormat::Upc_a:
    case WebCore::ShapeDetection::BarcodeFormat::Upc_e:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(GPU_PROCESS)
template<> bool isValidEnum<WebCore::ShapeDetection::LandmarkType, void>(uint8_t value)
{
    switch (static_cast<WebCore::ShapeDetection::LandmarkType>(value)) {
    case WebCore::ShapeDetection::LandmarkType::Mouth:
    case WebCore::ShapeDetection::LandmarkType::Eye:
    case WebCore::ShapeDetection::LandmarkType::Nose:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidOptionSet<WebCore::RenderAsTextFlag>(OptionSet<WebCore::RenderAsTextFlag> value)
{
    constexpr uint16_t allValidBitsValue = 0
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowAllLayers)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowLayerNesting)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowCompositedLayers)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowOverflow)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowSVGGeometry)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowLayerFragments)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowAddresses)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowIDAndClass)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::PrintingMode)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::DontUpdateLayout)
        | static_cast<uint16_t>(WebCore::RenderAsTextFlag::ShowLayoutState)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebCore::WebGPU::TextureFormat, void>(uint8_t value)
{
    switch (static_cast<WebCore::WebGPU::TextureFormat>(value)) {
    case WebCore::WebGPU::TextureFormat::R8unorm:
    case WebCore::WebGPU::TextureFormat::R8snorm:
    case WebCore::WebGPU::TextureFormat::R8uint:
    case WebCore::WebGPU::TextureFormat::R8sint:
    case WebCore::WebGPU::TextureFormat::R16uint:
    case WebCore::WebGPU::TextureFormat::R16sint:
    case WebCore::WebGPU::TextureFormat::R16float:
    case WebCore::WebGPU::TextureFormat::Rg8unorm:
    case WebCore::WebGPU::TextureFormat::Rg8snorm:
    case WebCore::WebGPU::TextureFormat::Rg8uint:
    case WebCore::WebGPU::TextureFormat::Rg8sint:
    case WebCore::WebGPU::TextureFormat::R32uint:
    case WebCore::WebGPU::TextureFormat::R32sint:
    case WebCore::WebGPU::TextureFormat::R32float:
    case WebCore::WebGPU::TextureFormat::Rg16uint:
    case WebCore::WebGPU::TextureFormat::Rg16sint:
    case WebCore::WebGPU::TextureFormat::Rg16float:
    case WebCore::WebGPU::TextureFormat::Rgba8unorm:
    case WebCore::WebGPU::TextureFormat::Rgba8unormSRGB:
    case WebCore::WebGPU::TextureFormat::Rgba8snorm:
    case WebCore::WebGPU::TextureFormat::Rgba8uint:
    case WebCore::WebGPU::TextureFormat::Rgba8sint:
    case WebCore::WebGPU::TextureFormat::Bgra8unorm:
    case WebCore::WebGPU::TextureFormat::Bgra8unormSRGB:
    case WebCore::WebGPU::TextureFormat::Rgb9e5ufloat:
    case WebCore::WebGPU::TextureFormat::Rgb10a2unorm:
    case WebCore::WebGPU::TextureFormat::Rg11b10ufloat:
    case WebCore::WebGPU::TextureFormat::Rg32uint:
    case WebCore::WebGPU::TextureFormat::Rg32sint:
    case WebCore::WebGPU::TextureFormat::Rg32float:
    case WebCore::WebGPU::TextureFormat::Rgba16uint:
    case WebCore::WebGPU::TextureFormat::Rgba16sint:
    case WebCore::WebGPU::TextureFormat::Rgba16float:
    case WebCore::WebGPU::TextureFormat::Rgba32uint:
    case WebCore::WebGPU::TextureFormat::Rgba32sint:
    case WebCore::WebGPU::TextureFormat::Rgba32float:
    case WebCore::WebGPU::TextureFormat::Stencil8:
    case WebCore::WebGPU::TextureFormat::Depth16unorm:
    case WebCore::WebGPU::TextureFormat::Depth24plus:
    case WebCore::WebGPU::TextureFormat::Depth24plusStencil8:
    case WebCore::WebGPU::TextureFormat::Depth32float:
    case WebCore::WebGPU::TextureFormat::Depth32floatStencil8:
    case WebCore::WebGPU::TextureFormat::Bc1RgbaUnorm:
    case WebCore::WebGPU::TextureFormat::Bc1RgbaUnormSRGB:
    case WebCore::WebGPU::TextureFormat::Bc2RgbaUnorm:
    case WebCore::WebGPU::TextureFormat::Bc2RgbaUnormSRGB:
    case WebCore::WebGPU::TextureFormat::Bc3RgbaUnorm:
    case WebCore::WebGPU::TextureFormat::Bc3RgbaUnormSRGB:
    case WebCore::WebGPU::TextureFormat::Bc4RUnorm:
    case WebCore::WebGPU::TextureFormat::Bc4RSnorm:
    case WebCore::WebGPU::TextureFormat::Bc5RgUnorm:
    case WebCore::WebGPU::TextureFormat::Bc5RgSnorm:
    case WebCore::WebGPU::TextureFormat::Bc6hRgbUfloat:
    case WebCore::WebGPU::TextureFormat::Bc6hRgbFloat:
    case WebCore::WebGPU::TextureFormat::Bc7RgbaUnorm:
    case WebCore::WebGPU::TextureFormat::Bc7RgbaUnormSRGB:
    case WebCore::WebGPU::TextureFormat::Etc2Rgb8unorm:
    case WebCore::WebGPU::TextureFormat::Etc2Rgb8unormSRGB:
    case WebCore::WebGPU::TextureFormat::Etc2Rgb8a1unorm:
    case WebCore::WebGPU::TextureFormat::Etc2Rgb8a1unormSRGB:
    case WebCore::WebGPU::TextureFormat::Etc2Rgba8unorm:
    case WebCore::WebGPU::TextureFormat::Etc2Rgba8unormSRGB:
    case WebCore::WebGPU::TextureFormat::EacR11unorm:
    case WebCore::WebGPU::TextureFormat::EacR11snorm:
    case WebCore::WebGPU::TextureFormat::EacRg11unorm:
    case WebCore::WebGPU::TextureFormat::EacRg11snorm:
    case WebCore::WebGPU::TextureFormat::Astc4x4Unorm:
    case WebCore::WebGPU::TextureFormat::Astc4x4UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc5x4Unorm:
    case WebCore::WebGPU::TextureFormat::Astc5x4UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc5x5Unorm:
    case WebCore::WebGPU::TextureFormat::Astc5x5UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc6x5Unorm:
    case WebCore::WebGPU::TextureFormat::Astc6x5UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc6x6Unorm:
    case WebCore::WebGPU::TextureFormat::Astc6x6UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc8x5Unorm:
    case WebCore::WebGPU::TextureFormat::Astc8x5UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc8x6Unorm:
    case WebCore::WebGPU::TextureFormat::Astc8x6UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc8x8Unorm:
    case WebCore::WebGPU::TextureFormat::Astc8x8UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc10x5Unorm:
    case WebCore::WebGPU::TextureFormat::Astc10x5UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc10x6Unorm:
    case WebCore::WebGPU::TextureFormat::Astc10x6UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc10x8Unorm:
    case WebCore::WebGPU::TextureFormat::Astc10x8UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc10x10Unorm:
    case WebCore::WebGPU::TextureFormat::Astc10x10UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc12x10Unorm:
    case WebCore::WebGPU::TextureFormat::Astc12x10UnormSRGB:
    case WebCore::WebGPU::TextureFormat::Astc12x12Unorm:
    case WebCore::WebGPU::TextureFormat::Astc12x12UnormSRGB:
        return true;
    default:
        return false;
    }
}

#if ENABLE(MEDIA_STREAM)
template<> bool isValidEnum<WebCore::MediaAccessDenialReason, void>(uint8_t value)
{
    switch (static_cast<WebCore::MediaAccessDenialReason>(value)) {
    case WebCore::MediaAccessDenialReason::NoReason:
    case WebCore::MediaAccessDenialReason::NoConstraints:
    case WebCore::MediaAccessDenialReason::UserMediaDisabled:
    case WebCore::MediaAccessDenialReason::NoCaptureDevices:
    case WebCore::MediaAccessDenialReason::InvalidConstraint:
    case WebCore::MediaAccessDenialReason::HardwareError:
    case WebCore::MediaAccessDenialReason::PermissionDenied:
    case WebCore::MediaAccessDenialReason::InvalidAccess:
    case WebCore::MediaAccessDenialReason::OtherFailure:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebCore::WebGPU::TextureAspect, void>(uint8_t value)
{
    switch (static_cast<WebCore::WebGPU::TextureAspect>(value)) {
    case WebCore::WebGPU::TextureAspect::All:
    case WebCore::WebGPU::TextureAspect::StencilOnly:
    case WebCore::WebGPU::TextureAspect::DepthOnly:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebCore::WebGPU::PredefinedColorSpace, void>(uint8_t value)
{
    switch (static_cast<WebCore::WebGPU::PredefinedColorSpace>(value)) {
    case WebCore::WebGPU::PredefinedColorSpace::SRGB:
        return true;
    default:
        return false;
    }
}

template<> bool isValidOptionSet<WebKit::WebEventModifier>(OptionSet<WebKit::WebEventModifier> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebKit::WebEventModifier::ShiftKey)
        | static_cast<uint8_t>(WebKit::WebEventModifier::ControlKey)
        | static_cast<uint8_t>(WebKit::WebEventModifier::AltKey)
        | static_cast<uint8_t>(WebKit::WebEventModifier::MetaKey)
        | static_cast<uint8_t>(WebKit::WebEventModifier::CapsLockKey)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidEnum<WebKit::WebEventType, void>(int8_t value)
{
    switch (static_cast<WebKit::WebEventType>(value)) {
    case WebKit::WebEventType::NoType:
    case WebKit::WebEventType::MouseDown:
    case WebKit::WebEventType::MouseUp:
    case WebKit::WebEventType::MouseMove:
    case WebKit::WebEventType::MouseForceChanged:
    case WebKit::WebEventType::MouseForceDown:
    case WebKit::WebEventType::MouseForceUp:
    case WebKit::WebEventType::Wheel:
    case WebKit::WebEventType::KeyDown:
    case WebKit::WebEventType::KeyUp:
    case WebKit::WebEventType::RawKeyDown:
    case WebKit::WebEventType::Char:
#if ENABLE(TOUCH_EVENTS)
    case WebKit::WebEventType::TouchStart:
#endif
#if ENABLE(TOUCH_EVENTS)
    case WebKit::WebEventType::TouchMove:
#endif
#if ENABLE(TOUCH_EVENTS)
    case WebKit::WebEventType::TouchEnd:
#endif
#if ENABLE(TOUCH_EVENTS)
    case WebKit::WebEventType::TouchCancel:
#endif
#if ENABLE(MAC_GESTURE_EVENTS)
    case WebKit::WebEventType::GestureStart:
#endif
#if ENABLE(MAC_GESTURE_EVENTS)
    case WebKit::WebEventType::GestureChange:
#endif
#if ENABLE(MAC_GESTURE_EVENTS)
    case WebKit::WebEventType::GestureEnd:
#endif
        return true;
    default:
        return false;
    }
}

#if ENABLE(TOUCH_EVENTS)
template<> bool isValidEnum<WebKit::WebPlatformTouchPoint::State, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebPlatformTouchPoint::State>(value)) {
    case WebKit::WebPlatformTouchPoint::State::Released:
    case WebKit::WebPlatformTouchPoint::State::Pressed:
    case WebKit::WebPlatformTouchPoint::State::Moved:
    case WebKit::WebPlatformTouchPoint::State::Stationary:
    case WebKit::WebPlatformTouchPoint::State::Cancelled:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidEnum<WebKit::WebMouseEventButton, void>(int32_t value)
{
    switch (static_cast<WebKit::WebMouseEventButton>(value)) {
    case WebKit::WebMouseEventButton::LeftButton:
    case WebKit::WebMouseEventButton::MiddleButton:
    case WebKit::WebMouseEventButton::RightButton:
    case WebKit::WebMouseEventButton::NoButton:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebMouseEventSyntheticClickType, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebMouseEventSyntheticClickType>(value)) {
    case WebKit::WebMouseEventSyntheticClickType::NoTap:
    case WebKit::WebMouseEventSyntheticClickType::OneFingerTap:
    case WebKit::WebMouseEventSyntheticClickType::TwoFingerTap:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebWheelEvent::Granularity, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebWheelEvent::Granularity>(value)) {
    case WebKit::WebWheelEvent::Granularity::ScrollByPageWheelEvent:
    case WebKit::WebWheelEvent::Granularity::ScrollByPixelWheelEvent:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebWheelEvent::Phase, void>(uint32_t value)
{
    switch (static_cast<WebKit::WebWheelEvent::Phase>(value)) {
    case WebKit::WebWheelEvent::Phase::PhaseNone:
    case WebKit::WebWheelEvent::Phase::PhaseBegan:
    case WebKit::WebWheelEvent::Phase::PhaseStationary:
    case WebKit::WebWheelEvent::Phase::PhaseChanged:
    case WebKit::WebWheelEvent::Phase::PhaseEnded:
    case WebKit::WebWheelEvent::Phase::PhaseCancelled:
    case WebKit::WebWheelEvent::Phase::PhaseMayBegin:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebWheelEvent::MomentumEndType, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebWheelEvent::MomentumEndType>(value)) {
    case WebKit::WebWheelEvent::MomentumEndType::Unknown:
    case WebKit::WebWheelEvent::MomentumEndType::Interrupted:
    case WebKit::WebWheelEvent::MomentumEndType::Natural:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebHitTestResultData::IsScrollbar, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebHitTestResultData::IsScrollbar>(value)) {
    case WebKit::WebHitTestResultData::IsScrollbar::No:
    case WebKit::WebHitTestResultData::IsScrollbar::Vertical:
    case WebKit::WebHitTestResultData::IsScrollbar::Horizontal:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebsiteMetaViewportPolicy, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebsiteMetaViewportPolicy>(value)) {
    case WebKit::WebsiteMetaViewportPolicy::Default:
    case WebKit::WebsiteMetaViewportPolicy::Respect:
    case WebKit::WebsiteMetaViewportPolicy::Ignore:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebsiteMediaSourcePolicy, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebsiteMediaSourcePolicy>(value)) {
    case WebKit::WebsiteMediaSourcePolicy::Default:
    case WebKit::WebsiteMediaSourcePolicy::Disable:
    case WebKit::WebsiteMediaSourcePolicy::Enable:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebsiteSimulatedMouseEventsDispatchPolicy, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebsiteSimulatedMouseEventsDispatchPolicy>(value)) {
    case WebKit::WebsiteSimulatedMouseEventsDispatchPolicy::Default:
    case WebKit::WebsiteSimulatedMouseEventsDispatchPolicy::Allow:
    case WebKit::WebsiteSimulatedMouseEventsDispatchPolicy::Deny:
        return true;
    default:
        return false;
    }
}

template<> bool isValidEnum<WebKit::WebsiteLegacyOverflowScrollingTouchPolicy, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebsiteLegacyOverflowScrollingTouchPolicy>(value)) {
    case WebKit::WebsiteLegacyOverflowScrollingTouchPolicy::Default:
    case WebKit::WebsiteLegacyOverflowScrollingTouchPolicy::Disable:
    case WebKit::WebsiteLegacyOverflowScrollingTouchPolicy::Enable:
        return true;
    default:
        return false;
    }
}

#if ENABLE(WK_WEB_EXTENSIONS)
template<> bool isValidEnum<WebKit::WebExtensionEventListenerType, void>(uint8_t value)
{
    switch (static_cast<WebKit::WebExtensionEventListenerType>(value)) {
    case WebKit::WebExtensionEventListenerType::Unknown:
    case WebKit::WebExtensionEventListenerType::ActionOnClicked:
    case WebKit::WebExtensionEventListenerType::AlarmsOnAlarm:
    case WebKit::WebExtensionEventListenerType::CommandsOnCommand:
    case WebKit::WebExtensionEventListenerType::ContextMenusOnClicked:
    case WebKit::WebExtensionEventListenerType::CookiesOnChanged:
    case WebKit::WebExtensionEventListenerType::DevToolsElementsPanelOnSelectionChanged:
    case WebKit::WebExtensionEventListenerType::DevToolsExtensionPanelOnShown:
    case WebKit::WebExtensionEventListenerType::DevToolsExtensionPanelOnHidden:
    case WebKit::WebExtensionEventListenerType::DevToolsExtensionSidebarPaneOnShown:
    case WebKit::WebExtensionEventListenerType::DevToolsExtensionSidebarPaneOnHidden:
    case WebKit::WebExtensionEventListenerType::DevToolsExtensionPanelOnSearch:
    case WebKit::WebExtensionEventListenerType::DevToolsNetworkOnNavigated:
    case WebKit::WebExtensionEventListenerType::DevToolsNetworkOnRequestFinished:
    case WebKit::WebExtensionEventListenerType::DevToolsPanelsOnThemeChanged:
    case WebKit::WebExtensionEventListenerType::NotificationsOnButtonClicked:
    case WebKit::WebExtensionEventListenerType::NotificationsOnClicked:
    case WebKit::WebExtensionEventListenerType::PermissionsOnAdded:
    case WebKit::WebExtensionEventListenerType::PermissionsOnRemoved:
    case WebKit::WebExtensionEventListenerType::PortOnDisconnect:
    case WebKit::WebExtensionEventListenerType::PortOnMessage:
    case WebKit::WebExtensionEventListenerType::RuntimeOnConnect:
    case WebKit::WebExtensionEventListenerType::RuntimeOnConnectExternal:
    case WebKit::WebExtensionEventListenerType::RuntimeOnInstalled:
    case WebKit::WebExtensionEventListenerType::RuntimeOnMessage:
    case WebKit::WebExtensionEventListenerType::RuntimeOnMessageExternal:
    case WebKit::WebExtensionEventListenerType::RuntimeOnStartup:
    case WebKit::WebExtensionEventListenerType::StorageOnChanged:
    case WebKit::WebExtensionEventListenerType::TabsOnActivated:
    case WebKit::WebExtensionEventListenerType::TabsOnAttached:
    case WebKit::WebExtensionEventListenerType::TabsOnCreated:
    case WebKit::WebExtensionEventListenerType::TabsOnDetached:
    case WebKit::WebExtensionEventListenerType::TabsOnHighlighted:
    case WebKit::WebExtensionEventListenerType::TabsOnMoved:
    case WebKit::WebExtensionEventListenerType::TabsOnRemoved:
    case WebKit::WebExtensionEventListenerType::TabsOnReplaced:
    case WebKit::WebExtensionEventListenerType::TabsOnUpdated:
    case WebKit::WebExtensionEventListenerType::WebNavigationOnBeforeNavigate:
    case WebKit::WebExtensionEventListenerType::WebNavigationOnCommitted:
    case WebKit::WebExtensionEventListenerType::WebNavigationOnCompleted:
    case WebKit::WebExtensionEventListenerType::WebNavigationOnDOMContentLoaded:
    case WebKit::WebExtensionEventListenerType::WebNavigationOnErrorOccurred:
    case WebKit::WebExtensionEventListenerType::WebRequestOnAuthRequired:
    case WebKit::WebExtensionEventListenerType::WebRequestOnBeforeRedirect:
    case WebKit::WebExtensionEventListenerType::WebRequestOnBeforeRequest:
    case WebKit::WebExtensionEventListenerType::WebRequestOnBeforeSendHeaders:
    case WebKit::WebExtensionEventListenerType::WebRequestOnCompleted:
    case WebKit::WebExtensionEventListenerType::WebRequestOnErrorOccurred:
    case WebKit::WebExtensionEventListenerType::WebRequestOnHeadersReceived:
    case WebKit::WebExtensionEventListenerType::WebRequestOnResponseStarted:
    case WebKit::WebExtensionEventListenerType::WebRequestOnSendHeaders:
    case WebKit::WebExtensionEventListenerType::WindowsOnCreated:
    case WebKit::WebExtensionEventListenerType::WindowsOnFocusChanged:
    case WebKit::WebExtensionEventListenerType::WindowsOnRemoved:
    case WebKit::WebExtensionEventListenerType::DevToolsInspectedWindowOnResourceAdded:
    case WebKit::WebExtensionEventListenerType::DownloadsOnCreated:
    case WebKit::WebExtensionEventListenerType::DownloadsOnChanged:
        return true;
    default:
        return false;
    }
}
#endif

#if ENABLE(GPU_PROCESS)
template<> bool isValidEnum<WebCore::WebGPU::FeatureName, void>(uint8_t value)
{
    switch (static_cast<WebCore::WebGPU::FeatureName>(value)) {
    case WebCore::WebGPU::FeatureName::DepthClipControl:
    case WebCore::WebGPU::FeatureName::Depth32floatStencil8:
    case WebCore::WebGPU::FeatureName::TextureCompressionBc:
    case WebCore::WebGPU::FeatureName::TextureCompressionEtc2:
    case WebCore::WebGPU::FeatureName::TextureCompressionAstc:
    case WebCore::WebGPU::FeatureName::TimestampQuery:
    case WebCore::WebGPU::FeatureName::IndirectFirstInstance:
    case WebCore::WebGPU::FeatureName::ShaderF16:
    case WebCore::WebGPU::FeatureName::Rg11b10ufloatRenderable:
    case WebCore::WebGPU::FeatureName::Bgra8unormStorage:
        return true;
    default:
        return false;
    }
}
#endif

template<> bool isValidOptionSet<WebKit::WebsiteDataFetchOption>(OptionSet<WebKit::WebsiteDataFetchOption> value)
{
    constexpr uint8_t allValidBitsValue = 0
        | static_cast<uint8_t>(WebKit::WebsiteDataFetchOption::ComputeSizes)
        | static_cast<uint8_t>(WebKit::WebsiteDataFetchOption::DoNotCreateProcesses)
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

template<> bool isValidOptionSet<WebKit::WebsiteDataType>(OptionSet<WebKit::WebsiteDataType> value)
{
    constexpr uint32_t allValidBitsValue = 0
        | static_cast<uint32_t>(WebKit::WebsiteDataType::Cookies)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::DiskCache)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::MemoryCache)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::OfflineWebApplicationCache)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::SessionStorage)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::LocalStorage)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::WebSQLDatabases)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::IndexedDBDatabases)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::MediaKeys)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::HSTSCache)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::SearchFieldRecentSearches)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::ResourceLoadStatistics)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::Credentials)
#if ENABLE(SERVICE_WORKER)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::ServiceWorkerRegistrations)
#endif
        | static_cast<uint32_t>(WebKit::WebsiteDataType::DOMCache)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::DeviceIdHashSalt)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::PrivateClickMeasurements)
#if HAVE(ALTERNATIVE_SERVICE)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::AlternativeServices)
#endif
        | static_cast<uint32_t>(WebKit::WebsiteDataType::FileSystem)
#if ENABLE(SERVICE_WORKER)
        | static_cast<uint32_t>(WebKit::WebsiteDataType::BackgroundFetchStorage)
#endif
        | 0;
    return (value.toRaw() | allValidBitsValue) == allValidBitsValue;
}

#if ENABLE(WEBXR)
template<> bool isValidEnum<PlatformXR::SessionFeature, void>(uint8_t value)
{
    switch (static_cast<PlatformXR::SessionFeature>(value)) {
    case PlatformXR::SessionFeature::ReferenceSpaceTypeViewer:
    case PlatformXR::SessionFeature::ReferenceSpaceTypeLocal:
    case PlatformXR::SessionFeature::ReferenceSpaceTypeLocalFloor:
    case PlatformXR::SessionFeature::ReferenceSpaceTypeBoundedFloor:
    case PlatformXR::SessionFeature::ReferenceSpaceTypeUnbounded:
#if ENABLE(WEBXR_HANDS)
    case PlatformXR::SessionFeature::HandTracking:
#endif
        return true;
    default:
        return false;
    }
}
#endif

} // namespace WTF

#if COMPILER(GCC)
IGNORE_WARNINGS_END
#endif
